#------------------------------------------------------------------------------
# snackAmp - HttpD package 1.0 for snackAmp
# Integrated web server for documents and streaming audio
# web server.
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
# Simple Sample httpd/1.[01] server  mini1.1.tcl
# Stephen Uhler (c) 1996-1997 Sun Microsystems
# Httpd is a global array containing the global server state
#  root: the root of the document directory
#  port: The port this server is serving
#  listen:  the main listening socket id
#  accepts: a count of accepted connections so far
#  maxtime:     The max time (msec) allowed to complete an http request
#  maxused:     The max # of requests for a socket
# Modified: Tom Wilkason
#           Namespaced
#           support tml substitution
#           support vfs files
#           support streaming audio
#------------------------------------------------------------------------------

set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/HttpD.tcl,v 1.53 2007/04/15 19:18:58 wilkason Exp $
}
set todo {
   - Keep alives don't work
   - POST is broken (never worked well)
   - migrate puts respond to fcopy and centralize code, should use memchan to make it work easier
}
# HTTP/1.[01] error codes (the ones we use)
namespace eval HttpD {
   variable Httpd
   variable httpConfig
   variable HttpdUrlCache
   variable HttpMimeType
   variable HttpdErrors
   variable HttpdMimeType
   variable map
   variable localURL "/local"
   # Generic error response.
   variable HttpdErrorFormat {
      <title>Error: %1$s</title>
      Got the error: <b>%2$s</b><br>
      while trying to obtain <b>%3$s</b>
   }

   # Errors
   array set HttpdErrors {
      200 {Data follows}
      204 {No Content}
      302 {Found}
      400 {Bad Request}
      401 {Authorization Required}
      403 {Permission denied}
      404 {Not Found}
      408 {Request Timeout}
      411 {Length Required}
      419 {Expectation Failed}
      500 {Server Internal Error}
      501 {Server Busy}
      503 {Service Unavailable}
      504 {Service Temporarily Unavailable}
   }

   # Default Settings
   array set Httpd {
      maxtime  6000000
      maxused  100
   }
   # convert the file suffix into a mime type
   # add your own types as needed
   array set HttpdMimeType {
      {}    text/plain
      .txt  text/plain
	   .tml	application/x-tcl-template
      .html text/html
      .htm  text/html
      .css  text/css
      .gif  image/gif
      .jpg  image/jpeg
      .mp3  audio/mpeg
      .ogg  application/x-ogg
      .wav  audio/x-wav
      .rm   audio/x-pn-realaudio
      .aifc audio/x-aiff
      .ra   audio/x-realaudio
      .png  image/png
      .m3u  audio/x-mpegurl
      .pls  audio/x-scpls
      .js   application/x-javascript
   }
   # Initialize cgi map for encode/decode
   #  doesn't exist
   for {set i 1} {$i <= 256} {incr i} {
      set c [format %c $i]
      if {![string match \[a-zA-Z0-9\] $c]} {
         set map($c) %[format %.2X $i]
      }
   }
# These are handled specially
   array set map {
       " " +   \n %0d%0a
   }
   variable bufferSize
   upvar #0 ::snackAmpSettings(bufSize) bufferSize


}
# Start the server by listening for connections on the desired port.

proc HttpD::Server {root {port 80} {default index.html}} {
   Trace
   variable Httpd
   if {[catch {package require Iocpsock}]} {
      ### Use the stock socket command.
      set cmd socket
   } else {
      ### Use the field patch.
      set cmd socket2
      saLog "using new socket driver $cmd"
   }
   catch {close $Httpd(port)}  ;# it might already be running
   array set Httpd [list root $root default $default port $port]
   array set Httpd [list accepts 0 requests 0 errors 0]
   set Httpd(listen) [$cmd -server HttpD::Accept $port]
   return $Httpd(port)
}

# Accept a new connection from the server and set up a handler
# to read the request from the client. Called automatically
# On incoming request.

proc HttpD::Accept {sock ipaddr {port {}}} {
   Trace
   variable Httpd
   upvar #0 Httpd$sock data
   if {$::snackAmpSettings(localOnly) && [iptools::meetsMask $ipaddr $::snackAmpSettings(ipAddress) $::snackAmpSettings(subNetMask)]==0} {
      if {$::TA(webLog)!=0} {
         if {[catch {puts $::TA(webLog) "[timestamp] non-local address $ipaddr rejected"} result] } then {
            #socket was closed or some error occured
            Debug $sock $result
         }
      }
      close $sock
      array unset data
      return
   }
   incr Httpd(accepts)
   set ::snackAmpSettings(HttpHits) $Httpd(accepts)
   Reset $sock $::snackAmpSettings(httpMaxused)
   Debug $sock Connect $ipaddr $port
   set data(ip) $ipaddr
   ##
   # Log access if needed
   #
   if {$::TA(webLog)!=0} {
      if {[catch {puts $::TA(webLog) "[timestamp] $ipaddr Connect on $sock"} result] } then {
         #socket was closed or some error occured
         Debug $sock $result
      }
   }
}

# Initialize or reset the socket state

proc HttpD::Reset {sock left} {
   Trace
   variable Httpd
   variable bufferSize
   upvar #0 Httpd$sock data

   array set data [list state start linemode 1 version 0 left $left]
   HttpD::pushCancel $sock
   fconfigure $sock -blocking 0 -buffering full -buffersize $bufferSize \
         -translation {auto crlf}
   fileevent $sock readable [list HttpD::Read $sock]
   fileevent $sock writable {}
}

# Read data from a client request. Called from a fileevent on the incoming
# socket.
# 1) read the request line
# 2) read the mime headers
# 3) read the additional data (if post && content-length not satisfied)

proc HttpD::Read {sock} {
   Trace
   variable Httpd
   upvar #0 Httpd$sock data
   # Use line mode to read the request and the mime headers
   if {$data(linemode)} {
      if {[catch {gets $sock line} readCount] } then {
         # Can get this during an fcopy when another read
         # request comes in
         #
         Debug $sock "socket read error: $readCount"
         #htLog "$sock socket read error: $readCount [fblocked $sock]"
         if {[streamBuffer::hasMore $sock]} {
            SocketDone $sock 0
         } else {
            SocketDone $sock 1   ;# TFW: See if this can also be 0 for http
         }
         return
      }
      set state [string compare $readCount 0],$data(state)
      switch -glob -- $state {
         1,start {
            if {[regexp {(HEAD|POST|GET) ([^?]+)\??([^ ]*) HTTP/1.([01])} $line x data(proto) data(url) data(query) data(version)]} {
               set data(state) mime
               incr Httpd(requests)
               Debug $sock Request $data(left) $line
            } else {
               Debug $sock "REGEXP error: $state ($readCount) '$line'"
               Error $sock 400 $line
            }
         }
         0,start {
            Debug $sock Warning "Initial blank line fetching request"
         }
         1,mime {
            ##
            # Start collecting the headers of the request
            #
            if {[regexp {([^:]+):[  ]*(.*)}  $line {} key value]} {
               set key [string tolower $key]
               set data(key) $key
               if {[info exists data(mime,$key)]} {
                  append data(mime,$key) ", $value"
               } else {
                  set data(mime,$key) $value
               }
            } elseif {[regexp {^[   ]+(.+)} $line {} value] && \
                     [info exists data(key)]} {
               append data(mime,$data($key)) " " $value
            } else {
               Error $sock 400 $line
            }
         }
         0,mime {
            ##
            # Headers are done (blank line) respond to request
            #
            if {$data(proto) == "POST" && [info exists data(mime,content-length)]} {
               set data(linemode) 0
               set data(count) $data(mime,content-length)
               if {$data(version) && [info exists data(mime,expect]} {
                  if {$data(mime,expect) == "100-continue"} {
                     puts $sock "100 Continue HTTP/1.1\n"
                     flush $sock
                  } else {
                     Error $sock 419 $data(mime,expect)
                  }
               }
               fconfigure $sock -translation {binary crlf}
            } elseif {$data(proto) != "POST"}  {
               Respond $sock
            } else {
               Error $sock 411 "Confusing mime headers"
            }
         }
         -1,* {
            if {[eof $sock]} {
               Debug $sock Error "Broken connection fetching request, client may have disconnected"
               SocketDone $sock 1
            } else {
               Debug $sock "Partial read, retrying"
            }
         }
         default {
            Error $sock 404 "Invalid http state: $state,[eof $sock]"
         }
      }

      # Use counted mode to get the post data

   } elseif {![eof $sock]} {
      if {[catch {
         append data(postdata) [read $sock $data(count)]
         set data(count) [expr {$data(mime,content-length) - \
                  [string length $data(postdata)]}]
         if {$data(count) == 0} {
            Debug $sock "Respond 0 data count"
            Respond $sock
         }
      } result] } then {
         Log $sock Error $result
         SocketDone $sock 1
      }
   } else {
      Log $sock Error "Broken connection reading POST data"
      SocketDone $sock 1
   }
}
#----------------------------------------------------------------------------
#  Done with the socket, either close it, or set up for next fetch
#  sock:  The socket I'm done with
#  close: If true, close the socket, otherwise set up for reuse
#----------------------------------------------------------------------------
proc HttpD::SocketDone {sock close} {
   Trace
   variable Httpd
   upvar #0 Httpd$sock data
   if {[info exists data(cancel)]} {
      after cancel $data(cancel)
   }
   # Close any open files and remove temporary ones
   foreach {index} [array names data fid,*] {
      foreach {xx fid} [split $index ,] {break}
      if {[catch {close $fid} result] } then {
         Debug $sock $result
      } else {
         Debug $sock "Closing leftover fid=$fid"
         htLog "Closing leftover fid=$fid"
      }
      # If a file was tagged, remove it too (it was temp)
      if {[string length $data($index)]} {
         if {[catch {file delete -force $data($index)} result] } then {
            Debug $sock $result
         } else {
            Debug $sock "Deleting leftover file $data($index)"
            htLog "Deleting leftover file $data($index)"
         }
      }
   }
   # See if socket is to be closed or just reset
   if {[info exists data(left)]} {
      array set xxx [array get data ip]
      set left [incr data(left) -1]
      unset data
      array set data [array get xxx]
   }
   if {$close} {
      if {[catch {close $sock} result]} then {
         Debug $sock "Didn't close $sock"
      } else {
         Debug $sock "Closing socket $sock"
      }
      array unset data
   } else {
      Reset $sock $left
      Debug $sock "Resetting socket $sock - $left connections left"
   }
   return ""
}

# A timeout happened

proc HttpD::Timeout {sock} {
   Trace
   variable Httpd
   upvar #0 Httpd$sock data
   SocketDone $sock 1
   #Error $sock 408 "Timeout"
}
##
# Header for dynamically generated files like .pls and some images
#
proc HttpD::dynamicContentHeader {sock myPath _close} {
   Trace
   upvar #0 Httpd$sock data
   upvar $_close close
   append result "HTTP/1.$data(version) 200 OK\n"
   append result "Date: [Date [expr {[clock seconds]-10}]]\n"
   append result "X-Powered-By: Snackamp $::SAVersion\n"
   #append result "Last-Modified: [Date [file mtime $mypath]]\n"
   switch -- [string tolower [file extension $myPath]] {
      .css -
      .js -
      .jpg -
      .png -
      .gif {
         append result "Cache-control: max-age=86400\n"
      }
      default {
#         No Cache-control causes streams not to play on windows!
#         append result "Cache-control: no-cache\n"
      }
   }
   append result "Content-Type: [ContentType $myPath]\n"
   ## Should also close socket if recvd connection close header
   set close [expr {$data(left) == 0}]
   if {$close} {
      append result "Connection: close\n"
   } elseif {$data(version) == 0} {
      if {[info exists data(mime,connection)] && \
          [string equal -nocase $data(mime,connection) "Keep-Alive"]} {
         set close 0
         append result "Connection: Keep-Alive\n"
      }
   }
   append result "Content-Length: [string length $data(content)]\n"
   append result "\n"
   return $result
}
#------------------------------------------------------------------------------
# Header, either for binary or non-binary data
#------------------------------------------------------------------------------
;# * icy-name is the name of the stations
;# * icy-genre is the genre that the station resides in
;# * icy-pub is basically a switch to either allow the server to publish itself in the directory or not (1 meaning yes and 0 meaning no)
;# * icy-br is the bitrate of the stream
;# * icy-url is the homepage for the stream
;# * icy-irc is yp shoutcast specific (used for contact information)
;# * icy-icq is yp shoutcast specific (used for contact information)
;# * icy-aim is yp shoutcast specific (used for contact information)
proc HttpD::ContentHeader {sock mypath _close} {
   Trace
   upvar #0 Httpd$sock data
   upvar $_close close
   # for IceCast -> instead of 200 ICY 200 OK
   if {[streamBuffer::isStream $sock]} {
      #append result "ICY 200 OK\n"
      append result "HTTP/1.0 200 OK\n"
      append result "Server: Snackamp\n"
      if {[info exists data(Title)]} {
         append result "icy-name: $data(Title)\n"
      } else {
         append result "icy-name: SnackAmp Streaming Music\n"
      }
      if {[info exists data(Album)]} {
         append result "icy-url: $data(Album)\n"
      } else {
         append result "icy-url: Unknown\n"
      }
      if {[info exists data(Artist)]} {
         append result "icy-artist: $data(Artist)\n"
      }
      if {[info exists data(Genre)]} {
         append result "icy-genre: $data(Genre)\n"
      } else {
         append result "icy-genre: Various\n"
      }
      if {[info exists data(rate)]} {
         append result "icy-br: $data(rate)\n"
      } else {
         append result "icy-br: Mixed\n"
      }
      append result "icy-metadata: 0\n"
      append result "Content-Type: [ContentType $mypath]\n"
      append result "Cache-control: no-cache\n"
      append result "Pragma: no-cache\n"
      append result "Connection: close\n"
      append result "Content-Length: $data(streamingLength)\n"
      set close 1  ;# will reuse this socket for next track
   } else {
      append result "HTTP/1.$data(version) 200 OK\n"
      append result "X-Powered-By: Snackamp $::SAVersion\n"
      append result "Date: [Date [expr {[clock seconds]-10}]]\n"
      #append result "Last-Modified: [Date [file mtime $mypath]]\n"
      #append result "Cache-control: max-age=86400\n"
      switch -- [string tolower [file extension $mypath]] {
         .css -
         .js -
         .jpg -
         .png -
         .gif {
            append result "Cache-control: max-age=86400\n"
         }
         default {
            append result "Cache-control: no-cache\n"
         }
      }
      append result "Content-Type: [ContentType $mypath]\n"
      ## Should also close socket if recvd connection close header
      set close [expr {$data(left) == 0}]
      if {$close} {
         append result "Connection: close\n"
      } elseif {$data(version) == 0} {
         if {[info exists data(mime,connection)] && \
             [string equal -nocase $data(mime,connection) "Keep-Alive"]} {
            set close 0
            append result "Connection: Keep-Alive\n"
         }
      }
      append result "Content-Length: [file size $mypath]\n"
   }
   append result "\n"
   return $result
}
#------------------------------------------------------------------------------
# Handle file system and tml/dhtml dispatch queries.
# Copy resultant data to client
#------------------------------------------------------------------------------
proc HttpD::Respond {sock} {
   Trace
   variable Httpd
   variable HttpdUrlCache
   variable localURL
   set tfile {}
   upvar #0 Httpd$sock data
   if {![info exists data(query)]} {
      set data(query) {}
   }
   # Get url with http stuff striped off
   regsub  -- {(^http://[^/]+)?} $data(url) {} url
   set type file
   set ext html
   # Patch by Wojciech Kocjan to add authorization
   # not working for streaming, should just be for
   # some pages maybe.
   if {[string match "*remote/*" $url] && $::snackAmpSettings(EnableRemoteControl)==0} {
      Error $sock 403 "$url Service Disabled"
      return
   } elseif {![string match -nocase "*/streaming/*" $url] && [llength $::snackAmpSettings(HttpUsers)] > 0} {
      set ok 0
      if {[info exists data(mime,authorization)]} {
         package require base64
         set auth [split $data(mime,authorization)]
         set cup [base64::decode [lindex $auth 1]]
         foreach up $::snackAmpSettings(HttpUsers) {
            if {[string equal $up $cup]} {
               set ok 1
               break
            }
         }
      }
      if {!$ok} {
         set content "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">\n<HTML>\n<HEAD>\n<TITLE>Access Denied</TITLE>\n</HEAD>\n<BODY>\n<H2>Access Denied</H2>\nThe requested URL cannot be accessed because a valid username and password are required.\n</BODY></HTML>\n"
         eval {
            fconfigure $sock -translation binary
            puts $sock "HTTP/1.0 401 Unauthorized\r\nWWW-Authenticate: Basic realm=\"snackAmp\"\r\nMIME-Version: 1.0\r\nDate: [Date [clock seconds]]\r\nContent-Type: text/html\r\nContent-Length: [string length $content]\r\nConnection: close\r\n\r\n$content"
         }
         Error $sock 401 "'$url' is Unauthorized"
         return
      }
   }
   # end patch
   ##
   # Process the url converting to local files
   # as necessary. If a translation for that URL
   # has already occured, just lookup it up.
   #
   if {[info exists HttpdUrlCache($url)]} {
      set mypath $HttpdUrlCache($url)
      Debug $sock "Using cached '$mypath' for '$data(url)'"
   } else {
      if {[string length $::snackAmpSettings(HttpLocal)] && [regsub "^${localURL}(.*)" $url {\1} tail]} {
         # Translate any "/local" requests to a user specified location
         set mypath [Url2File $::snackAmpSettings(HttpLocal) $tail]
      } else {
         # Otherwise do a simple mapping
         set mypath [Url2File $Httpd(root) $url]
      }
      Debug $sock "New path:'$mypath' for 'data($url)'"
      ##
      # If the path is a folder, then use the appropriate default file for it
      #
      if {[file isdirectory $mypath]} {
         ##
         # Attempt default file names
         #
         foreach {attempt} {
            index.tml
            default.tml
            default.html
            index.html
            default.htm
            index.htm
         } {
            set try [file join $mypath $attempt]
            if {[file exists $try]} {
               set mypath $try
               Debug $sock "Using default file '$try' for '$data(url)'"
               # TFW: Need to set relative folder to this one
               break
            }
         }
      }
      set HttpdUrlCache($url) $mypath
   }
   ##
   # Now check the path and process files like this:
   # 1) If no path, throw an error
   # 2) If the file not readable, assume it is a dispatch request
   #    and try to process it. If that fails, throw an error
   # 3) If the file is readable, then check if it is a tml file
   #    if so then subst the tml file, write it out to the
   #    corresponding html file.
   # 4) Prococess the html file by copying it from the file
   #    to the socket with appropriate headers attached.
   #
   if {[string length $mypath] == 0} {
      # For security
      Error $sock 404 "No Path Specified"
   } elseif {![file readable $mypath] && [regexp {/(.+[^/])} $data(url) xx name]} {
      ##
      # Treat request as internal dispatch if the
      # url decoded into a local file did not exist.
      #/control/?GLA=K106&Date=2000-04&submit=Query

      Debug "Query: '$data(query)'=''$mypath"
      # Grab the appropriate query parameters
      ##
      # Get any post or get data
      #
      switch -- [string toupper $data(proto)] {
         GET {
            set param $data(query)
         }
         POST {
            # this is no longer working (never really did)
            set param $data(postdata)
         }
         default {set param ""}
      }
      ##
      # If a dynamic url $name, was specified process it.
      # it can return either dhtml content or a file name
      # from a local system. Verify the service is enabled
      #
      # TODO: Clean up logic, including check for streaming
      #       use namespace exists and service call to see
      #       if particular url base is enabled.
      #
      if {[info command ::${name}::init] ne ""} {
         if {[string match *streaming/* $name] && $::snackAmpSettings(EnableStreaming)==0} {
            set result "url:$url\npath:$mypath\nStreaming is Disabled"
            Error $sock 403 "$name Service Disabled"
            return
          } elseif {[info command ::${name}::process] eq ""} {
             # Try and load the service, we have an init but no process yet.
#debug loading ::${name}
             if {[catch {::${name}::init} result] } then {
               Error $sock 403 "$name Service Not Available"
               return
             }
         }
         ##
         # Attempt to process the dhtml, file type is returned
         # and contentheader uses it to determine file type
         # the data itself is returned.
         # e.g. m3u, html, ...
         #
         if {[catch {::${name}::process $sock $param} result] } then {
            # We had an error processing the dispatch, notify the user
            set result "url:$url\npath:$mypath\nUnable to process: $result\n$::errorInfo"
            Error $sock 503 $result
            return
         } else {
            # Expects the type of data, extension and content/file name
            if {[llength $result]==3} {
               foreach {type ext content} $result {break}
               switch -- $type {
                  file {
                     set mypath $content
                  }
                  dhtml {
                     # handled later
                  }
                  default {}
               }
            } else {
               saLog "Process Completion: $result"
               Error $sock 503 $result
            }
         }
      } else {
         # Page not found
         Error $sock 404 "$mypath"
         return
      }
   } elseif {[string equal -nocase [file extension $mypath] ".tml"]} {
      ##
      # File exists, if it is a tml macro file, process it
      # and write out corresponding html file.
      if {[catch {set fid [open $mypath r]} error] } then {
         Log $sock "TML Open Error: $::errorInfo"
         Error $sock 503 "$mypath: $error"
         return
      } else {
         set contents [read $fid]
         close $fid
         # Substitute and rename output file
         # TFW: If any $xxx stuff in file, the subst will fail.
         if {[catch {subst $contents} contents] } then {
            Log $sock "TML Processing Error: $::errorInfo"
            Error $sock 503 "$mypath: $contents"
            return
         } else {
            # Everything looks good after the processing
            # translate to html file
            set mypath "[file join $::TA(TEMP) [file rootname [file tail $mypath]].html]"
            set tfile $mypath
            # Write out to html file
            set fid [open $mypath w]
            fconfigure $fid -translation binary -buffering full -buffersize 250000
            puts -nonewline $fid $contents
            close $fid
            Debug $sock "Processed TML File '$mypath'"
            # Note: Need to remove this files in the future, otherwise they build up
         }
      }
   } else {
      Debug $sock "Processed Static: '$mypath'"
   }
   ##
   # If this is still a folder, we throw an error
   #
   if {[file isdirectory $mypath]} {
      Error $sock 404 "No default file found in $mypath"
      return
   }
   ##
   # Process complete html files here copying from file
   # to the socket.
   #
   if {[info exists data(proto)] && ($data(proto) != "HEAD")} {
      if {$type=="dhtml"} {
         ##
         # Handle dynamic content by using the file writeable
         # instead of creating a temp file and fcopy, keep
         # writing this content to this socket until we have
         # no more to push.
         set data(content) $content
         puts -nonewline $sock [dynamicContentHeader $sock "dummy.$ext" close]
         fconfigure $sock -translation binary
         #fileevent $sock readable {}
         fileevent $sock writable [list HttpD::PushContent $sock $close]
      } else {
         if {![file exists $mypath]} {
            Error $sock 404 "$mypath no found"
            return
         }
         ##
         # At this point we should have a valid file, either existing or
         # newly created from a tml
         #
         puts -nonewline $sock [ContentHeader $sock $mypath close]
         flush $sock
         if {[catch {
            set in [open $mypath r]
            fconfigure $sock -translation binary
            fconfigure $in -translation binary -buffering full -buffersize [fconfigure $sock -buffersize]
            Debug $sock "Starting Copy of '$mypath' fid=$in sock=$sock close=$close"
            # Disable any reads on this socket while a copy is in progress
            # to prevent busy errors on the read.
            fileevent $sock writable {}
            # The readable will be re-enabled once the file is copied
            set size [file size $mypath]    ;# -size needed on win2k/xp workaround for bug #719790
            if {[streamBuffer::isStream $sock]} {
               set offset [streamBuffer::seekPastId3 $in]
               incr size -$offset
            }
#puts "[me] started copy of $size bytes from $mypath [fconfigure $sock -buffersize]"
            HttpD::queueCloseFile $sock $in $tfile  ;# win2k starkit bug, fcopies don't always complete
            #fconfigure $sock -translation binary
            # This will also copy images, css, js or any files, -size fixes win2k starkit bug on fcopy
            fcopy $in $sock -size $size -command [list HttpD::CopyDone $in $sock $close $tfile]
         } error] } then {
            Log $sock "Respond: $error"
         }
      }
   } else {
      # Can add more header data as needed for HEAD requests "HTTP/1.$data(version) 200 OK\n"
      if {[info exists data(version)]} {
         set v $data(version)
      } else {
         set v 0
      }
      if {[catch {puts $sock "HTTP/1.$v 200 OK\n"} result] } then {
# BUG: pressing Play icon on web interface will cause this error on firefox
         Log $sock "[me] Respond: $result"
      }
      SocketDone $sock 1
   }
}
#------------------------------------------------------------------------------
# Push timeout cancel into the future
#------------------------------------------------------------------------------
proc HttpD::pushCancel {sock} {
   variable Httpd
   Trace
   upvar #0 config$sock data
   if {[info exists data(cancel)]} {
      after cancel $data(cancel)
   }
   set data(cancel) [after $Httpd(maxtime) [list HttpD::Timeout $sock]]
}
#------------------------------------------------------------------------------
# Load
#------------------------------------------------------------------------------
proc HttpD::LoadService { sock service } {
   Trace
   upvar #0 config$sock data
   variable httpConfig
   if {$service == ""} {
      return 0
   }
   if {![info exists httpConfig(services)] || [lsearch -exact $httpConfig(services) $service ] == -1 } {
      Debug "$sock Info Loading service: $service"
      if {[catch {namespace eval $service init} result] } then {
         Error $sock 503 "unable to start service \"$service\"\n$result"
         return 0
      } else {
         lappend httpConfig(services) $service
      }
   }
   return 1
}

#------------------------------------------------------------------------------
# Function   : HttpD::queueCloseFile
# Description: Save the fid to close it later if the CopyDone didn't complete
# Author     : Tom Wilkason
# Date       : 2/1/2003
#------------------------------------------------------------------------------
proc HttpD::queueCloseFile {sock fid file} {
   upvar #0 Httpd$sock data
   variable httpConfig
   set data(fid,$fid) $file
}
#------------------------------------------------------------------------------
# Callback when the socket becomes writable
# This should be replaced by an fcopy using Memchan for more effecency
#------------------------------------------------------------------------------
proc HttpD::PushContent {sock close} {
   Trace
   variable bufferSize
   upvar #0 Httpd$sock data
   # Pull off one socket buffers worth of data and push it out
   set outBuf [string range $data(content) 0 [expr {$bufferSize-1}]]
   set data(content) [string range $data(content) $bufferSize end]
   # This can block if it takes too long on a slow net
   # not sure if this is better than fcopy performance
   if {[catch {
      puts -nonewline $sock $outBuf
      flush $sock
   } result] } then {
      SocketDone $sock $close
   } else {
      if {[string length $data(content)]==0} {
         fileevent $sock writable {}
   #      set close 1  ;# TFW: Won't seem to work at 0 if fileevent $sock readable {}
         fileevent $sock readable [list HttpD::Read $sock]
         SocketDone $sock $close
         Debug $sock "PushContent Complete"
      }
   }
}
# Callback when file is done being output to client
# in:  The fd for the file being copied
# sock: The client socket
# close: close the socket if true
# bytes: The # of bytes copied
# error:  The error message (if any)

proc HttpD::CopyDone {in sock close tfile bytes {error {}}} {
   Trace
   variable Httpd
   upvar #0 Httpd$sock data
   close $in
   if {[catch {unset data(fid,$in)} result] } then {
      Debug $sock $result
   }
   if {[streamBuffer::hasMore $sock]} {
      # Attempt to stuff more data in the socket
      # Get next track
      set mypath [streamBuffer::popFile $sock]  ;# expand
      if {[catch {open $mypath r} in] } then {
         saLog "Open Failure on $mypath, $in"
         fileevent $sock readable [list HttpD::Read $sock]
         SocketDone $sock $close
      } else {
         fconfigure $in -translation binary -buffering full -buffersize [fconfigure $sock -buffersize]
         # Bug: If client closes socket, the CopyDone is not called
         # the socket is left hanging. queueCloseFile will clean it up later
         puts -nonewline $sock [ContentHeader $sock $mypath close]
         HttpD::queueCloseFile $sock $in {}
         #HttpD::pushCancel $sock
         if {[info exists data(cancel)]} {
            after cancel $data(cancel)
         }
         ##
         # startkit/win2k bug, fcopy won't complete without a -size
         #
         set size [file size $mypath]    ;# -size needed on win2k/xp workaround for bug #719790
         set offset [streamBuffer::seekPastId3 $in]
         incr size -$offset
         fconfigure $sock -translation binary
#puts "[me] started copy next of $size bytes from $mypath"
         fcopy $in $sock -size $size -command [list HttpD::CopyDone $in $sock $close $tfile]
         Debug $sock "Copying Next File '$mypath'"
      }
   } else {
      if {[string length $tfile]} {
         if {[catch {file delete -force $tfile} result] } then {
            Debug $sock $result
         }
      }
      Debug $sock Done $bytes bytes
      fileevent $sock readable [list HttpD::Read $sock]
      SocketDone $sock $close
   }
}
#
# Return the content type based on file name extension
#
proc HttpD::ContentType {path} {
   variable HttpdMimeType

   set type text/plain
   catch {set type $HttpdMimeType([file extension $path])}
   return $type
}

# Respond with an error reply
# sock:  The socket handle to the client
# code:  The httpd error code
# args:  Additional information for error logging

proc HttpD::Error {sock code mess} {
   Trace
   upvar #0 Httpd$sock data
   variable Httpd
   variable HttpdErrors
   variable HttpdErrorFormat
   append data(url) ""
   incr Httpd(errors)
   set message [format $HttpdErrorFormat $code $HttpdErrors($code) $data(url)]
   append message "<pre>\n$mess</pre>"
   append head "HTTP/1.$data(version) $code $HttpdErrors($code)"  \n
   append head "Date: [Date [clock seconds]]"  \n
   append head "Connection: close"  \n
   append head "Content-Length: [string length $message]"  \n

   # Because there is an error condition, the socket may be "dead"

   catch {
      fconfigure $sock  -translation {auto crlf}
      puts -nonewline $sock $head\n$message
      flush $sock
   } reason
   SocketDone $sock 1
   Debug $sock "Error $code $HttpdErrors($code) $mess $reason"
}
# Log an Httpd transaction.
# This should be replaced as needed.

proc HttpD::Log {sock args} {
   #Trace
   saLog "LOG: $sock $args"
}
proc HttpD::Debug {sock args} {
   if {$::snackAmpSettings(Trace)} {
      saLog "Debug: $sock $args"
   }
}
proc HttpD::htLog {message} {
   if {[info exists ::TA(log)]} {
      saLog $message
   }
}

# Generate a date string in HTTP format.

proc HttpD::Date {seconds} {
   return [clock format $seconds -format {%a, %d %b %Y %T %Z} -gmt 1]
#   return [clock format $seconds -format {%a, %d %b %Y %T %Z}]
}


# Convert a url into a pathname. (UNIX version only)
# This is probably not right, and belongs somewhere else.
# - Remove leading http://... if any
# - Collapse all /./ and /../ constructs
# - expand %xx sequences -> disallow "/"'s  and "."'s due to expansions
proc HttpD::Url2File {root url} {
   Trace
   # Do not allow mapping that won't affect the url
   if {[string length [string trim $root]]==0} {
      return ""
   }
   regsub -all  -- {//+} $url / url     ;# collapse multiple /'s
   while {[regsub -all  -- {/\./} $url / url]} {}   ;# collapse /./
   while {[regsub -all  -- {/\.\.(/|$)} $url /\x81\\1 url]} {} ;# mark /../
   while {[regsub  -- "/\[^/\x81]+/\x81/" $url / url]} {} ;# collapse /../
   if {![regexp "\x81|%2\[eEfF]" $url]} { ;# invalid /../, / or . ?
      set result $root[CgiMap $url]
   } else {
      set result ""
   }
   return $result
}

# Decode url-encoded strings.

proc HttpD::CgiMap {data} {
   Trace
   regsub -all  -- {([][$\\])} $data {\\\1} data
   regsub -all  -- {%([0-9a-fA-F][0-9a-fA-F])} $data  {[format %c 0x\1]} data
   return [subst $data]
}
# HttpD::GetParams
#
#  This decodes parameters
#
# Arguments:
#  An encoded value
#
# Results:
#  The decoded value
proc HttpD::GetParams { vars args } {
   Trace
   upvar args arglist

   if { [ llength $arglist ] == 1 } {
      # braced set of args
      eval set arglist $arglist
   }
   foreach { var val } $vars {
      uplevel 1 set $var \"$val\"   ;#"
   }
   foreach arg $arglist {
      set param [ split $arg "=" ]
      set var [ lindex $param 0 ]
      set val [ join [ lreplace $param 0 0 ] "=" ]
      if { [ lsearch -exact $vars $var ] != -1 } {
         uplevel 1 set $var \{$val\}
      }
   }
}
# HttpD::Decode
#
#  This decodes data in www-url-encoded format.
#
# Arguments:
#  An encoded value
#
# Results:
#  The decoded value
proc HttpD::Decode {data} {
   Trace
   regsub -all {\+} $data " " data
   regsub -all {([][$\\])} $data {\\\1} data
   regsub -all {%([0-9a-fA-F][0-9a-fA-F])} $data  {[format %c 0x\1]} data
   # convert result back from utf-8 to match local file system
   return [encoding convertfrom utf-8 [subst $data]]
}

#------------------------------------------------------------------------------
# Replace all html codes with normal letters
#------------------------------------------------------------------------------
proc HttpD::unSub {content} {
  set XML_MAP {
     &lt;   <
     &gt;   >
     &amp;  &
     &quot; \"
     &apos; '
  } ;#"
  return [string map $XML_MAP $content]
}
# HttpD::Encode
#
#  This encodes data in www-url-encoded format.
#
# Arguments:
#  A string
#
# Results:
#  The encoded value
proc HttpD::Encode {string {depth 1}} {
   variable map
   # Convert into utf-8 so we have valid url characters
   set string [encoding convertto utf-8 $string]
   set saveString $string
   regsub -all \[^a-zA-Z0-9\] $string {$map(&)} string
   regsub -all \n $string {\\n} string
   regsub -all \t $string {\\t} string
   regsub -all {[][{})\\]\)} $string {\\&} string
   # This shouldn't happen anymore since converting from utf-8
   if {[catch {subst -nocommands $string} result] } then {
      # encodings for utf/unicode map give map substitute errors
      # convert to ascii then try again, prevent loops
      if {$depth < 2} {
         return [Encode [encoding convertto ascii $saveString] [incr depth]]
      } else {
         saLog "Encode Error:'$saveString'->$result"
         return $saveString
      }
   } else {
      return $result
   }
}
#------------------------------------------------------------------------------
# Flush the log file every once in a while
#------------------------------------------------------------------------------
proc HttpD::FlushLog {fid time} {
   flush $fid
   after $time HttpD::FlushLog $fid $time
}
#------------------------------------------------------------------------------
# Function   : HttpD::Start
# Description: Start up a local webserver on the next free port starting at 8080
#               Save the port number
# Author     : Tom Wilkason
# Date       : 4/14/2002
#------------------------------------------------------------------------------
proc HttpD::Start {port root default} {
   Trace
   global TA
   set tries 0
   set TA(webLog) 0
   if {$::snackAmpSettings(webLog)} {
      set fileName "[clock format [clock seconds] -format %y-%m-%d].log"
      if {[catch {set TA(webLog) [open [file join $::HOME $fileName] a+]} result] } then {
         debug $result
      } else {
         puts $TA(webLog) "[timestamp] Log File opened"
      }
      HttpD::FlushLog $TA(webLog) 25000
   }
   while {[catch {HttpD::Server $root $port $default}] && $tries <= 20} {
      incr port
      incr tries
   }
   if {$tries == 20} {
      Log "Failed to start the server"
   } else {
      Log "Starting Tcl httpd server on [info hostname] port $port"
   }
   return $port
}
#------------------------------------------------------------------------------
# Function   : HttpD::Shutdown
# Description: Close all open sockets and shutdown
# Author     : Tom Wilkason
# Date       : 1/30/2003
#------------------------------------------------------------------------------
proc HttpD::Shutdown {args} {
   variable Httpd
   catch {
      close $Httpd(listen)
      foreach glb [info globals Httpdsock*] {
         set socket [string range $glb 5 end]
         catch {close $::TA(webLog)}
         saLog "Closing ::$glb"
         HttpD::SocketDone $socket 1
      }
   }
}
proc HttpD::dump {args} {
   variable Httpd
   variable bufferSize
   foreach glb [info globals Httpdsock*] {
      set socket [string range $glb 5 end]
      if {[catch {
         puts "$socket\n[fconfigure $socket]\n[fileevent $socket readable]"
      } result] } then {
         continue
      }
      if {[catch {
      fconfigure $socket -blocking 0 -buffering full -buffersize $bufferSize \
         -translation {auto crlf}
      } result]} {
         saLog "$socket $result"
      }
      parray ::$glb
      update
   }
}
#------------------------------------------------------------------------------
# Helper functions to manage the stream buffer
#------------------------------------------------------------------------------
namespace eval streamBuffer {
   proc allStreamInfo {args} {
      Trace
      set result [list]
      foreach sock [file channels sock*] {
         if {[hasMore $sock]} {
            upvar #0 Httpd$sock data
            lappend result [lindex [db::sortedList] [lindex $data(streaming) 0]]
            lappend result $data(ip)
         }
      }
      return $result
   }
   proc streamInfo {sock} {
      Trace
      set result [list]
      upvar #0 Httpd$sock data
      if {[hasMore $sock]} {
         lappend result [lindex [db::sortedList] [lindex $data(streaming)] 0]
         lappend result $data(ip)
      }
      return result
   }
   proc hasMore {sock} {
      upvar #0 Httpd$sock data
      if {[info exists data(streaming)] && [llength $data(streaming)] > 0} {
         return 1
      } else {
         return 0
      }
   }
   proc isStream {sock} {
      Trace
      upvar #0 Httpd$sock data
      if {[info exists data(streaming)]} {
         return 1
      } else {
         return 0
      }
   }
   proc pop {sock} {
      Trace
      upvar #0 Httpd$sock data
      if {[hasMore $sock]} {
         return [lpop left data(streaming)]
      } else {
         return {}
      }
   }
   proc popFile {sock} {
      Trace
      upvar #0 Httpd$sock data
      if {[hasMore $sock]} {
         set front [lpop left data(streaming)]
         # TODO: This is wrong here, we need to have saved it
         return [db::TrackFromDb $front]
      } else {
         return {}
      }
   }
   proc setList {sock List} {
      upvar #0 Httpd$sock data
      set data(streaming) $List
   }
   proc appendList {sock List} {
      upvar #0 Httpd$sock data
      foreach {item} $List {
         lpush right data(streaming) $item
      }
   }
   proc seekPastId3 {in} {
      Trace
      set offset [id3Tag::id3V2Offset $in]
      if {$offset > 0} {
         if {[catch {seek $in $offset start} result] } then {
            debug $result
         }
      }
      return $offset
   }
}

#------------------------------------------------------------------------------
# Setup the internal web server
#------------------------------------------------------------------------------
proc HttpD::startup {args} {
   global TA snackAmpSettings
   if {$snackAmpSettings(EnableWebServer)} {
      package require HttpD
      set snackAmpSettings(webRoot) $TA(webRoot)
      set snackAmpSettings(usedPort) [HttpD::Start $snackAmpSettings(port) $::snackAmpSettings(webRoot) ""]
      set snackAmpSettings(internalWebBase) "http://[info hostname]:$snackAmpSettings(usedPort)"
      browser::init
      if {[info exists browser::helpFile]} {
         set browser::helpFile "$snackAmpSettings(internalWebBase)/$browser::helpFile"
         #set browser::playControl "$snackAmpSettings(internalWebBase)/$::snackAmpSettings(HttpDefault)"
         set browser::Jukebox "$snackAmpSettings(internalWebBase)/remote/jukebox"
      }
   } else {
      set snackAmpSettings(webRoot) "disabled"
      set snackAmpSettings(usedPort) "disabled"
      set snackAmpSettings(internalWebBase) "disabled"
      saLog "Web server is disabled"
   }
   set snackAmpSettings(HttpHits) 0
}
#set ::TA(log) 1
package provide HttpD 1.0

