#------------------------------------------------------------------------------
# Support functions for windows based icons
# Mostly copied from winico.tcl part of winico package
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/IconSupport.tcl,v 1.20 2007/02/08 01:55:50 wilkason Exp $
}
if {$::tcl_platform(platform)=="windows"} {
   if {[catch {package require Winico} result]} {
      puts "No built-in or available Winco, using local DLL"
      load [file join [file dirname [info script]] myWinico winico.dll]
   }
}
#------------------------------------------------------------------------------
# Routines below are in the winico distribution
#------------------------------------------------------------------------------
proc winico_seticon { w icofile } {
   set icofile [winico_file $icofile]   ;# calls either wrapped or stub file
   set ico [winico create $icofile]
   set screendepth [winfo screendepth .]
   set bigsize "32x32"
   set bigpos -1
   set bigdepth 0
   set smallsize "16x16"
   set smallpos -1
   set smalldepth 0
   foreach i [winico info $ico] {
      array set opts $i
      set depth    $opts(-bpp)
      set pos      $opts(-pos)
      set geometry $opts(-geometry)
      if { $geometry=="$bigsize" && $depth<=$screendepth } {
         if { $depth>$bigdepth } {
            set bigpos $pos
            set bigdepth $depth
         }
      } elseif { $geometry=="$smallsize" && $depth<=$screendepth } {
         if { $depth>$smalldepth } {
            set smallpos $pos
            set smalldepth $depth
         }
      }
   }
   if { $bigpos==-1 && $smallpos==-1 } {
      puts stderr "couldn't find $bigsize and $smallsize icons in $icofile"
      return $ico
   } elseif { $bigpos==-1 } {
      set bigpos $smallpos
      puts stderr "couldn't find $bigsize icons in $icofile"
   } elseif { $smallpos==-1 } {
      set smallpos $bigpos
      puts stderr "couldn't find $smallsize icons in $icofile"
   }
   winico setwindow $w $ico big   $bigpos
   winico setwindow $w $ico small $smallpos
   return $ico
}
proc winico_delall {} {
   foreach i [winico info] { winico delete $i }
}
proc winico_loadicon { w symbol } {
   set ico [winico load $symbol]
   winico setwindow $w $ico big
   winico setwindow $w $ico small
}
#------------------------------------------------------------------------------
# Function   : winico_file
# Description: Dummy proc to just return the file passed in
#              replaced with wrap code if wrapped
#              wrapped code will write then read icon file to disk
# Author     : Tom Wilkason
# Date       : 1/27/2001
#------------------------------------------------------------------------------
if {[hasNamespace ::starkit]} {
   proc winico_file {icofile} {
      # Icon file must be located in substDir (i.e. same as main script)
      # Just read it from the wrapped location and write it out to
      # the TEMP folder for use by the icon utilities.
      if {[catch {
         # In
         set fin [open $icofile r]
         fconfigure $fin -translation binary
         # Out
         set tmpfile [file join $::env(TEMP) [file tail $icofile]]
         # puts "icon: $icofile with temp file as $tmpfile"
         if {! [file exists $tmpfile] } {
            set fout [open $tmpfile w]
            fconfigure $fout -translation binary
            # Move file to temp folder.
            puts -nonewline $fout [read $fin]
            close $fout
            set libMtime [file mtime $tmpfile]
            puts "Created $tmpfile with time [clock format $libMtime]"
         }
         close $fin
      } result]} {
         puts stderr "getIconScript error: $result\n$::errorInfo"
         return {}
      }
      return "$tmpfile"
   }
} else {
   proc winico_file {name} {
      # The snackAmp.ico HAS to be in the same folder
      # as the snackAmp.tcl file.
      return [file join $::TA(scriptHome) $name]
   }
}
#------------------------------------------------------------------------------
# Function   : winTaskBar::init
# Description: Add icons in taskbar for Windows
# Author     : Tom Wilkason
# Date       : 1/9/2001
#------------------------------------------------------------------------------
set ::TA(iconsVisible) 0
namespace eval winTaskBar {
   variable iconCache
   variable lastCommand ""
   variable lastMessage ""
}
proc winTaskBar::init {args} {
   global TA snackAmpSettings
   variable iconCache
   variable iconDef
   if {[catch {
      set TA(IconFile) [file join $TA(snackRoot) icons snackAmp.ico]
      #saLog "Installing icons '$TA(IconFile)'"
      ##
      # Mini player in icon bar
      #
      set TA(sysTrayIcon) [winico_seticon $::SA $TA(IconFile)]
      set iconDef($TA(sysTrayIcon)) "LMB=Minimize SnackAmp, RMB=Toggle Mini-player"
      set iconCache($TA(sysTrayIcon)) "<Init>"
      # No menu decoration, force a tray icon
      if {$snackAmpSettings(fancyMenu)} {
         winico taskbar add $TA(sysTrayIcon) \
             -callback [list winTaskBar::Callback %m $::TA(sysTrayIcon) [list gui::Toggle]] \
             -text $winTaskBar::iconDef($TA(sysTrayIcon)) \
             -pos 1
      }
      ##
      # Add all the play control icons that are specified in settings
      #
      foreach icon $settings::taskBarIconList {
         if {$snackAmpSettings(icon,$icon)==1} {
            set iconFile [file join $TA(snackRoot) icons $icon.ico]

            if {[file exists $iconFile]} {
               set Icon [winico_seticon $::SA $iconFile]
               set iconCache($Icon) "<Init>"
               set iconDef($Icon) "LMB=$icon, RMB=Toggle Mini-player"
               winico taskbar add $Icon \
                  -callback [list winTaskBar::Callback %m $Icon $icon] \
                  -text $winTaskBar::iconDef($Icon) \
                  -pos 1
               incr ::TA(iconsVisible)
            }
         }
      }
   } result]} {
      debug $result
   }
}
##
# Callback fired when mouse over or clicked on one of the icons (windows only)
#

proc winTaskBar::Callback {message icon command args} {
   # Todo, handle volume control on certain button, can even do a winicon taskbar modify $icon ...
   global TA
   variable iconCache
   variable lastCommand
   variable lastMessage
   switch -- $message {
      WM_LBUTTONDOWN {
         eval $command
      }
      WM_LBUTTONUP {}
      WM_MOUSEMOVE {
         # called alot, only update if text has changed
         set dur [soundControl::duration]
         set state [soundControl::playState]
         switch -- $state {
            "Playing" {
               if {$dur > 0} {
                  set pos "[expr {round([soundControl::position]/[soundControl::duration]*1000.)/10.}]% - "
                } else {
                  set pos "Unknown - "
                }
            }
            default {
               set pos "$state - "
            }
         }
         set text "$pos$::songArray(Artist) - $::songArray(Title)"
         if {$iconCache($icon) ne $text} {
            winico taskbar modify $icon \
            -text $text \
            -pos 1
            set iconCache($icon) $text
         }
#          This a hack
#          if {$command eq "gui::Toggle" && $lastCommand ne $command && $::snackAmpSettings(fadePopup) && [trackTransition::visible]==0} {
#             trackTransition::show -1 "Right Bottom"
#          }
      }
      WM_RBUTTONUP {}
      WM_RBUTTONDOWN {
         mpToggle
      }
      default {}
   }
   set lastCommand $command
   set lastMessage $message
}

#------------------------------------------------------------------------------
#  Toggle the minplayer
#------------------------------------------------------------------------------
proc mpToggle {args} {
   package require miniPlayer
   foreach {x y} $::snackAmpSettings(positionMiniPlayer) {break}
   miniPlayer::toggler $x $y
}
#----------------------------------------------------------------------------
# tkTray icon support for later linux's
#----------------------------------------------------------------------------
namespace eval linuxTaskBar {
}
proc linuxTaskBar::init {args} {
   global snackAmpSettings
   set ::TA(iconsVisible) 0
   set i 0
   set g [list]
   foreach icon $settings::taskBarIconList {
      if {$snackAmpSettings(icon,$icon)==1} {
         set iconFile $settings::taskBarIconImages($icon)
         set l .tray[incr i]
         tktray::icon $l \
         -visible 1 -class Icon \
         -image $iconFile \
         -balloon "Click to $icon, RMB to toggle mini-player"
         incr ::TA(iconsVisible)
         #imageTools::button $l "Click to $icon, RMB to toggle mini-player"
         bind $l <1> $icon
         bind $l <3> mpToggle
         bind $l <Enter> {focus %W}
         bind $l <Button-5> [list bumpLevel -2]
         bind $l <Button-4> [list bumpLevel 2]
      }
   }
}
package provide IconSupport 1.0


