#------------------------------------------------------------------------------
# autoPlayList Package 1.0
#
# Auto Playlist functions for snackAmp Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/autoPlayList.tcl,v 1.118 2007/10/01 01:32:52 wilkason Exp $
}
namespace eval APL {
   package require snID3
   package require trackList
   variable folderList
   variable folderPos
   upvar #0 snackAmpSettings(folderList) folderList
   upvar #0 snackAmpSettings(folderPos) folderPos
#   variable showFolderTypes
#   upvar #0 snackAmpSettings(showFolderTypes) showFolderTypes

   variable folderNext
   variable folderNext
   variable folderPrev
   variable icons
   variable Table
   variable paneOrient "horizontal"
   variable lastAfter 0
   variable coverLabel
   variable coverPane
   # TODO: Needs to be decremented by first position to shift down starting at 0
   set icons [list "saImages.Tree.Folder" "saImages.Tree.ROM" "saImages.mp3_player" "saImages.Disk.Folder"]
   option add *FolderTree*Font  treeFont startup
   option add *FolderTree*Menu.font  menu startup
   # Trace variables used during settings when adding/removing
   # a column from the table.
   foreach {key} [array names ::snackAmpSettings col,*,show] {
      trace variable ::snackAmpSettings($key) w APL::rebuildTableAfterIdle
   }
}
#------------------------------------------------------------------------------
# Function   : APL::rebuildTableAfterIdle
# Description: When setting the snackAmpSettings array, this callback will
#              be called for each column. We only want to rebuild once, so
#              we keep pushing it off until the last call.
# Author     : Tom Wilkason
# Date       : 8/7/2003
#------------------------------------------------------------------------------
proc APL::rebuildTableAfterIdle {variable index mode} {
   variable lastAfter
   global $variable
   after cancel $lastAfter
   set lastAfter [after 200 [list APL::rebuildTable .wFC $variable $index $mode]]
}
#------------------------------------------------------------------------------
# Function   : saveSize
# Description: Save the APL window sizes
# Author     : Tom Wilkason
# Date       : 1/9/2001
#------------------------------------------------------------------------------
proc APL::saveSize {args} {
   if {![catch {winfo toplevel $::TREE} tl]} then {
      set ::snackAmpSettings(treeSize) [wm geometry $tl]
      catch {set ::snackAmpSettings(FileListSize) [lindex [pane $tl.sv.lb -size] 0]}
   }
}
##
# Handlers for entry widgets
#
proc APL::invokeFindNextFile {w} {
   APL::findNextSong $::TA(SAFind)
   entryHistory::push $w
}
proc APL::invokeFindNextFolder {w} {
   APL::findNextFolder $::TA(SAFind)
   entryHistory::push $w
}
proc APL::invokeFilterTracks {w} {
   APL::setSearchFilter
   entryHistory::push $w
}
#------------------------------------------------------------------------------
# Function   : folderTree
# Description: Toss up a folder tree linked to a variable
# Author     : Tom Wilkason
# Date       : 1/9/2001
#------------------------------------------------------------------------------
proc APL::FolderTree {args} {
   package require saTree
   global TA snackAmpSettings
   variable folderNext
   variable folderPrev
   variable paneOrient
   variable coverLabel
   variable coverPane
   ##
   # Create the folder tree entries if needed
   #
   set W .wFC
   if {[winfo exists $W]} {
      wm deiconify $W
      #raise $W      ;# This causes a noticible delay on KDE
      focus $W
      return
   }
   Wait 1
   if {[aplFolder::entries]==0} {
      readAPLFile $snackAmpSettings(lastFolderFile)
   }

   toplevel $W -cursor $::TA(arrow) -class FolderTree
   wm withdraw $W
   set TA(highlightPath) ""
   trace variable ::TA(highlightPath) w {after 10 [list APLcover::packCoverArt $::TA(highlightPath)]}
   wm protocol $W WM_DELETE_WINDOW {
      APL::close
   }
   ##
   # Button bar
   #
   frame $W.mb -relief groove -bd 2
   pack $W.mb -side top -fill x -pady 3

   # Save Button
   pack [button $W.mb.save -image saImages.Save -command {macros::saveCurrentAPL 0}]\
      -padx 3 -side left -fill y
   balloon::button $W.mb.save "Save the current settings"

   # Save As button
   pack [button $W.mb.saveas -image saImages.SaveAs -command {macros::SaveFolderFile}]\
    -padx 3 -side left -fill y
   balloon::button $W.mb.saveas "Save AutoPlaylist under a new file name"

   # Apply button
   pack [button $W.mb.apply -image saImages.Apply -command {macros::applyCurrentAPL}]\
     -padx 3 -side left -fill y
   balloon::button $W.mb.apply "Update the playlist with the current settings"

   menubutton $W.mb.files -relief flat -menu $W.mb.files.list
   menubutton $W.mb.favs -relief flat -menu $W.mb.favs.list

   # Refresh button
   pack [button $W.mb.refresh -image saImages.refresh -command {APL::indexFolderChildren $TA(highlightPath)}]\
     -padx 3 -side left -fill y
   balloon::button $W.mb.refresh "Re-Index selected folder and sub-folders"
   if {$gui::hasImg} {
      # Coverart Toggle
      pack [button $W.mb.covart -image saImages.coverart \
         -command APLcover::ToggleArt] \
         -padx 3 -side left -fill y
      balloon::button $W.mb.covart "Show or hide cover art for selected folder"
      set TA(aplArtButton) $W.mb.covart
   }

   # Filelist menu
   pack [button $W.mb.togpl -image saImages.FileList \
      -command "APL::setFileListToShow $W.sv.lb $W.mb.togpl"] \
      -padx 3 -side left -fill y
   balloon::button $W.mb.togpl "Depress to show files under the selected folder"

   # Initially, we toggle the setting to start the statemachine
   set ::snackAmpSettings(hideFileList) [expr {! ($::snackAmpSettings(hideFileList)&0x1)}]

   # Play button
   pack [button $W.mb.play -image saImages.player_play -command {aplLB::playSelectedFiles}] \
      -padx 3 -side left -fill y
   balloon::button $W.mb.play "Play selection(s)"

   # Play button
   pack [button $W.mb.queue -image saImages.queue -command {aplLB::insertSelectedFiles}] \
      -padx 3 -side left -fill y
   balloon::button $W.mb.queue "Queue selection(s)"

   # Level Buttons
   foreach {level} {0 1 2 3 4 5} {
      pack [radiobutton $W.mb.level$level -width 2 -text [expr {$level+1}] -value $level \
         -padx 0 -pady 0 -bd 1 -variable ::snackAmpSettings(treeDepth) -indicatoron 0 -relief flat \
          -relief flat -command {trackList::populateHighighted 1 1}] \
         -padx 0 -fill both -side left
      balloon::button $W.mb.level$level "Show an additional $level levels below the selected folder"
   }

   # Edit cell button
   pack [button $W.mb.edit -image saImages.edit-16 \
      -command trackList::changeEdit] \
      -padx 3 -side left
   balloon::button $W.mb.edit "Toggle cell editing"
   # Rating +/-
   pack [button $W.mb.plus -image saImages.iconthumbsup \
      -command {aplLB::bumpRating $::TA(highlightPath) 100 1}] \
      -padx 0 -side left -fill y
   balloon::button $W.mb.plus "Increase Rating of selected tracks or folders"
   pack [button $W.mb.minus -image saImages.iconthumbsdown \
      -command {aplLB::bumpRating $::TA(highlightPath) 100 -1}] \
      -padx 0 -side left -fill y
   balloon::button $W.mb.minus "Decrease Rating of selected tracks or folders"


   set trackList::editCellButton $W.mb.edit

   # Back button
   pack [button $W.mb.prev -image saImages.NavBack -command APL::prevSelection] \
      -padx 3 -side left -fill y
   balloon::button $W.mb.prev "Change to prior selected folder"


   # Dropdown history button (redundant with menu)
   pack [menubutton $W.mb.prevHistory -relief flat -image saImages.dropdownImage -menu $W.mb.prevHistory.list] -side left -fill y
   balloon::button $W.mb.prevHistory  "List of previous locations"
   # History menu, one for button and one for RMB
   menu $W.mb.prevHistory.list -postcommand [list APL::cbHistoryMenu $W.mb.prevHistory.list]

   # Forward Button
   pack [button $W.mb.next -image saImages.NavForward -command APL::nextSelection] \
      -padx 3 -side left -fill y
   balloon::button $W.mb.next "Change to next selected folder"

   ##
   # Find next entry
   #
   pack [button $W.mb.findFile -image saImages.findfile \
      -command [list APL::invokeFindNextFile $W.mb.find]] -side left -fill y
   balloon::button $W.mb.findFile "Press to search for next file containing text to the right."

   pack [button $W.mb.findLab -image saImages.findfolder \
      -command [list APL::invokeFindNextFolder $W.mb.find]] -side left -fill y
   balloon::button $W.mb.findLab "Press to search for next folder containing text to the right."

   pack [entryHistory::create $W.mb.find ::TA(SAFind) ::snackAmpSettings(findHistory)] \
      -side left -expand yes -fill x
   set ::SAFind [entryHistory::hull $W.mb.find]

   bind $::SAFind <Control-Return> [list APL::invokeFindNextFile $W.mb.find]
   bind $::SAFind <Return> [list APL::invokeFindNextFolder $W.mb.find]

   ##
   # Filter tracks
   #
   pack [button $W.mb.ckfilter -image saImages.filterOff \
      -command [list APL::invokeFilterTracks $W.mb.filter]] \
      -padx 3 -side left -fill y
   balloon::button $W.mb.ckfilter "Press search all media for tracks matching pattern to the right"

   pack [listOptionMenu $W.mb.searchWhat ::snackAmpSettings(FilterWhat) {Any File Title Artist Comp Album Note Genre Folder Enc Track Rate Tag Year} 6] \
      -side left -expand no -fill x
   balloon::define $W.mb.searchWhat.e "Select what attribute to filter on"

   pack [entryHistory::create $W.mb.filter ::TA(SAFilter) ::snackAmpSettings(filterHistory)] \
      -side left -expand yes -fill x
   set ::SAFilter [entryHistory::hull $W.mb.filter]

   pack $::SAFilter -side left -expand yes -fill x
   bind $::SAFilter <Return> {APL::setSearchFilter}
   bind $::SAFilter <Key> {APL::incrSearchFilter %K}
   pack [button $W.mb.clfilter -text Reset \
      -command {APL::SelectFile $::TA(highlightPath)}] \
      -padx 3 -side left -fill y
   balloon::button $W.mb.clfilter "Clear search entry and return to tree view of tracks"

   balloon::define $::SAFind "Enter text then hit return to search for next folder with matching text. Enter Ctrl-Return to search for next song with matching text. You may use wildcards for searching."
   balloon::define $::SAFilter "Enter text then depress the filter button to the left to restrict track names that contain this text."
   # Save the button names
   set folderNext $W.mb.next
   set folderPrev $W.mb.prev

   set Favorites [menu $W.mb.favs.list -tearoff 1 -postcommand [list APL::FavList $W.mb.favs.list]]
   set FavoritesRem [menu $W.mb.favs.list.remove -tearoff 0 -postcommand [list APL::remFavorite $W.mb.favs.list.remove]]
   $Favorites add command -label "Add Current Folder" -command [list APL::addFavorite]
   $Favorites add command -label "Clear All Favorites" -command [list APL::clearFavorites] -compound left -image saImages.delete
   $Favorites add cascade -label "Remove..." -menu $FavoritesRem -compound left -image saImages.delete
   $Favorites add separator

   set Folders [menu $W.mb.files.list -tearoff 1 -postcommand [list gui::APLFiles $W.mb.files.list] ]
   set Removal   [menu $W.mb.files.list.remove -postcommand [list gui::APLFilesRemove $W.mb.files.list.remove]]
   $Folders add command -label "Save AutoPlaylist As..." -command SaveFolderFile -compound left -image saImages.SaveAs
   $Folders add cascade -label "Remove..." -menu $Removal -compound left -image saImages.delete
   $Folders add separator

   ##
   # Frame to hold the tree/filelist
   #
   pack [frame $W.sv -bd 2] -side top -fill both -expand true
   #
   set TA(ck,selected)     "Checked: 0"
   set TA(ck,unselected)   "Unchecked: 0"
   set TA(ck,total)        "Total: 0"
   set TA(ck,tableTracks)  "Tracks: 0"
   set TA(ck,tableTime)    "Time: 00:00:00"
   set TA(ck,tableSize)    "Size: 0"
   set TA(ck,checkSize)    "Check Size: 0"
   set TA(ck,totalTree)    "Tree Total: 0"

   pack [frame $W.file -relief flat] -fill x
   set i 0
   # TFW: Can't get grid to force them to the right like I would think -sticky e would
   set statusWidgets {
      selected   12 left
      unselected 14 left
      total      14 left
      totalTree  14 left
   }
   # Fast free space
   if {[string length [info command ntFreeSpace]] > 0} {
      set TA(ck,free)         "Free 0"
      lappend statusWidgets free 13 left
   }
   lappend statusWidgets tableTracks 10 right tableSize 16 right tableTime 14 right
   foreach {item wid where} $statusWidgets {
      pack [label $W.file.$item \
         -textvariable TA(ck,$item) -width $wid  -bd 1 -relief sunken -padx 2 -pady 1] \
         -side $where -padx 1
   }
   # todo: Bindsource on the checked label to drag checked selected folder tree files
   # todo: Bindsource on the total label to drag entire selected folder tree files
   if {![catch {package require tkdnd}] } then {

      dnd bindsource $W.file.selected text/uri-list {APL::dndSrcCheckedTreeDrag}
      bind $W.file.selected <Alt-1> {dnd drag %W -actions copy;break}
      balloon::define $W.file.selected "Alt-Drag tracks from highlighted folder(s)"

      dnd bindsource $W.file.total text/uri-list {APL::dndSrcTotalTreeDrag}
      bind $W.file.total <Alt-1> {dnd drag %W -actions copy;break}
      balloon::define $W.file.total "Alt-Drag selected tracks from highlighted folder(s)"
      bindtags $W [concat dndTag [bindtags $W]]
   }
   ##
   # Frame to hold the tree and scrollbars, pane packs it
   #
   frame $W.sv.tree -bd 1 -relief sunken ;# "placed" by pane'r

   # Two frames that will be pained
   frame $W.sv.tree.tv -bd 0
   frame $W.sv.tree.cover -bd 0
   ##
   # Cover art
   #
   image create photo "saImages.APLCoverArt"
   set coverLabel $W.sv.tree.cover.art
   label $coverLabel -image saImages.APLCoverArt -background $::snackAmpSettings(bg)
   settings::regColor $coverLabel -background ::snackAmpSettings(bg) \

   balloon::define $coverLabel "Double click to query cover art for this album. Drag desired image from web browser to here to save it."
   if {![catch {package require tkdnd} result] } then {
      dnd bindtarget $coverLabel "UniformResourceLocator" <Drop> {APLcover::dndDropCoverArt %D}
      dnd bindtarget $coverLabel "text/uri-list"          <Drop> {APLcover::dndDropCoverArt %D}
   }
   bind $coverLabel <Double-ButtonPress-1> {
      APLcover::searchCoverArt
   }
   bind $coverLabel <ButtonPress-1> {
      albumart::clear
   }

   pack $W.sv.tree.cover.art -side top -expand true -fill both
   # pane top and bottom
   package require saPane
   set coverPane $W.sv.tree.cover
   pane $W.sv.tree.tv -orient vertical -dynamic 0


   set ::TREE $W.sv.tree.tv.w  ;# Canvas
   scrollbar $W.sv.tree.tv.sb -orient vertical -command [list $::TREE yview]

   pack $W.sv.tree.tv.sb -side right -fill y
   ##
   # Build the guts of the tree (menus are created here)
   #
   APL::buildAPL $::TREE
   mscroll $::TREE
   ##
   # File listbox/Table
   #
   set ::PLB [trackList::buildTable $W.sv.lb ::snackAmpSettings]
   ##
   # Create the appropriate bindings
   #
   aplLB::bindings
   bindings::assignHotKeys
   APL::treeBind
   # Keys for the APL search box
   APLFilterAssign <Return> {
      APL::findNextFolder $::TA(SAFind)
   }
   APLFilterAssign <Control-Return> {
      APL::findNextSong $::TA(SAFind)
   }
   ##
   # Top Menus
   #
   APL::aplMenu $W
   ##
   # Pane them
   #
   package require saPane
   pane $W.sv.tree -orient $paneOrient -dynamic 0
   APL::setFileListToShow $W.sv.lb $W.mb.togpl
   ##
   # Table/Listbox Menus
   #
   aplLB::menus
   # Expand initially to folder level on menu
   if {[info exists ::snackAmpSettings(treeSize)]} {
      wm geometry $W $::snackAmpSettings(treeSize)
   }
   # resize automatically
   bind $coverLabel <Configure> {APL::showCoverArtWindow}
   # Open from the top n layers down
   Tree::open $::TREE / 1     ;# TODO: Make 1 user adjustable
   wm deiconify $W
   focus $W
   after 100 [list myIcon $W]
   Wait 0
   after 100 APL::bindTkDND
   macros::Title
}

#------------------------------------------------------------------------------
# When cover art window shown, pack it and make sure node is still visible
#------------------------------------------------------------------------------
proc APL::showCoverArtWindow {args} {
   APLcover::packCoverArt $::TA(highlightPath)
   APL::viewHighlighted $::TREE
}
#------------------------------------------------------------------------------
# Bindings for TKDND
#------------------------------------------------------------------------------
proc APL::bindTkDND {args} {
   ##
   # DND
   #
   if {![catch {package require tkdnd}] } then {

      # When active over the tree, the following <Drag> event will be called
      # This kills the drop operation however
      #dnd bindtarget $::TREE text/uri-list <Drag> {
      #   Tree::setselection %W [Tree::labelat %W %x %y]
      #}
      # When dropping files or album covers in a tree...
      dnd bindtarget $::TREE "UniformResourceLocator" <Drop> [list APL::dndDropOnTree %x %y %D %A %T %d %m]
      dnd bindtarget $::TREE "text/uri-list"          <Drop> [list APL::dndDropOnTree %x %y %D %A %T %d %m]
      #%m is modifier key pressed
      # Define the APL Listbox as a source, to drag to one of
      #- LB -> Insert files
      #- TREE -> Move files
      #- Explorer -> copy files
      dnd bindsource [$::PLB bodypath] text/uri-list {APL::dndSrcLBDrag %W}
      bind [$::PLB bodypath] <Alt-1> {dnd drag %W -actions copy;break}

      # Bind the drag to the mouse, for Alt must release Alt prior to the
      # drop (for windows anyway)
      # Bind the tree as a valid source to drag data from
      # relying on mouse binding to put the folder/tracks in the array
      dnd bindsource $::TREE text/uri-list {APL::dndSrcTreeDrag}
      bind $::TREE <Alt-1> {dnd drag %W -actions copy;break}
   }
}
#------------------------------------------------------------------------------
# Function   : APL::rebuildTable
# Description: Rebuild the table when snackAmpSettings columns change, called
#              from a trace writing to one of the column settings.
#              Rebuild if the col changed from the current value.
# Author     : Tom Wilkason
# Date       : 5/27/2002
#------------------------------------------------------------------------------
proc APL::rebuildTable {W _variable index mode} {
   Trace
   global TA
   if {[catch {
      variable paneOrient
      upvar #0 $_variable settings
      set what $W.sv.lb
      # Add check to see if cols really need to be reconfigured
      if {[winfo exists $what]} {
         ##
         # Make sure we have the columns in the correct order
         #
         set col2Use [list]
         foreach {xx col xx} [$::PLB cget -columns] {
            lappend col2Use $col
         }
         set ::TA(cols) [orderList $::TA(cols) $col2Use]
         set wasOpen [winfo ismapped $what]
         foreach {xx col mode} [split $index ,] {break} ;# mode s/b "show"
         set geom [wm geometry $W]
         set ::snackAmpSettings(FileListSize) [lindex [pane $what -size] 0]
         destroy $what
         set trackList::TABLEDATA [list]

         set ::PLB [trackList::buildTable $what ::snackAmpSettings]
         if {$trackList::editCellState} {
            trackList::changeEdit
         }
         ##
         # Table/Listbox Menus
         #
         aplLB::menus
         ##
         # Create the appropriate bindings
         #
         aplLB::bindings
         if {$wasOpen} {
            pane $what -orient $paneOrient -dynamic 0 \
               -size $::snackAmpSettings(FileListSize)
            aplLB::UpdateList $::TA(highlightPath) 1
         }
         wm geometry $W $geom
         APL::bindTkDND
      }
   } result] } then {
      saLog "$::errorInfo"
   }
}
#------------------------------------------------------------------------------
# Function   : APL::close
# Description: Close the window asking to save changes if needed
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc APL::close {args} {
   $::TA(treeButton) configure -relief flat
   if {[Tree::isDirty $::TREE]} {
      if {[tk_messageBox -type yesno -icon question -message "Do you want to save your current AutoPlaylist?"] == "yes"} {
         saveCurrentAPL 0
      }
   }
   wm withdraw .wFC
   APL::saveSize
   db::commit
   trackList::saveColumnWidths
}
#----------------------------------------------------------------------------
# Grab some cover art and dump it into the selected folder
# Also handles native files dropped (called once per file) -> not implemented yet
#----------------------------------------------------------------------------
proc APL::dndDropOnTree {x y urls action type action1 modifier args} {
   Trace
   #type=UniformResourceLocator if it is a url
   #type=text/uri-list if it is a file or folder
   set fromList [list]
   set toList [list]
   set lbl [Tree::labelat $::TREE $x $y]
   set act "none"
   # Process each one
   foreach {url} $urls {
      if {[string length $lbl]} {
         if {$modifier eq "Control"} {
            set act "copy"
         } else {
            set act "move"
         }
         if {[catch {APL::GetTreeFolder $::TREE $lbl} folderName]} {
            saLog $folderName
            return
         }
         switch -- [string tolower [file extension $url]] {
            .jpg -
            .png -
            .gif -
            .jpeg -
            .bmp
            {
               macros::copyImageToFolder $url $folderName
            }
            default {
               # general files
               # Below code is the same, can be combined if works
               set tail [file tail $url]
               set newName [file join $folderName $tail]
               if {[file isdirectory $folderName] && ($newName ne $folderName)} {
                  lappend fromList $url
                  lappend toList $newName
               }
            }
         }
      }
   };#foreach
   ##
   # Do any post processing on the dropped files
   # Note: May have file case issues on windows
   #
   if {[llength $fromList]} {
      set ans [tk_messageBox -type yesno -icon question \
            -message "Are you sure you want to $act\n'[join $fromList \n]'\n  to\n'$folderName'"]
      if {$ans=="yes"} {
         switch -- $act {
            "copy" {
               return [catalog::copyTrackFolderPairs $fromList $toList]
            }
            "move" {
               return [catalog::moveTrackFolderPairs $fromList $toList]
            }
            default {
            saLog "Unknown action $"
            }
         }
      }
   }
}

#------------------------------------------------------------------------------
# Function   : APL::cbDndPLBToTree
# Description: Callback on drag-n-drop target from PLB to Tree
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc APL::cbDndPLBToTree {X Y} {
   Trace
   set x [expr {$X - [winfo rootx $::TREE]}]
   set y [expr {$Y - [winfo rooty $::TREE]}]
   set lbl [Tree::labelat $::TREE $x $y]
   if {$lbl != {}} {
      if {![catch {APL::GetTreeFolder $::TREE $lbl} folderName]} {
         # We let go over a node, ask to copy files
         if {[catalog::isMediaPlayer $folderName $::TA(highlightPath)]} {
            if {[catalog::copyClipboardFiles $folderName] > 0} {
               catch {
                  aplLB::UpdateList $::TA(highlightPath) [winfo ismapped $::PLB]
                  APL::treeSelect $::TA(highlightPath)
               }
            }
         } else {
            if {[catalog::moveClipboardFiles $folderName] > 0} {
               catch {
                  aplLB::UpdateList $::TA(highlightPath) [winfo ismapped $::PLB]
                  APL::treeSelect $::TA(highlightPath)
               }
            }
         }
      }
   }
}
#------------------------------------------------------------------------------
# Function   : APL::cbDndPLBToLB
# Description: Callback on drag-n-drop target from PLB to LB
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc APL::cbDndPLBToLB {x y} {
   global LB
   Trace
   if {$x>0} {
      set new [$LB nearest [expr {$y - [winfo rooty $LB]}]]
   } else {
      set new end
   }
   if {[catch {selection get -selection CLIPBOARD} tryFiles]} {
      return 0
   }
   clipboard clear
   set files [list]
   # Ignore errors due to funky clipboard data
   foreach {file} $tryFiles {
      if {[lsearch -exact $files $file] < 0} {
         lappend files $file
      }
   }
   LB::insertList $new $files
   return 1
}
#------------------------------------------------------------------------------
# Function   : APL::cbDndTreeToClipboard
# Description: Callback on drag-n-drop target from Tree to PlayList
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc APL::cbDndTreeToClipboard {X Y} {
   Trace
   set src $::TA(aplDNDSource)
   clipboard clear
   set data [list]
   ##
   # Would be nice to grab the data in the sort order of the tablelist
   #
   foreach {entry} [db::filesMatchingPattern "$src/*"] {
      lappend data $entry
   }
   clipboard append [lsort -dictionary $data]
}
#------------------------------------------------------------------------------
# Function   : APL::cbDndTreeToTree
# Description: Callback when moving folders in the APL tree
# Author     : Tom Wilkason
# Date       : 6/3/2002
#------------------------------------------------------------------------------
proc APL::cbDndTreeToTree {W lbl} {
   if {![catch {APL::GetTreeFolder $W $lbl} folderName]} {
      set newName [file join $folderName [file tail $::TA(aplDNDSource)]]
      if {$::TA(aplDNDSource) ne $folderName && $::TA(aplDNDSource) ne $newName} {
         set ans [tk_messageBox -type yesno -icon question \
               -message "Are you sure you want to move/copy \n'$::TA(aplDNDSource)'\n  to\n'$newName'"]
         if {$ans=="yes"} {
            if {[catch {
                  if {[catalog::isMediaPlayer $newName $::TA(aplDNDSource)]} {
                     catalog::copyFolder $::TA(aplDNDSource) $newName
                  } else {
                     catalog::moveFolder $::TA(aplDNDSource) $newName
                  }
                  if {$::snackAmpSettings(dndExpand)} {
                     APL::goAndShow $newName
                  }
               } result] \
                  } then {
               tk_messageBox -type ok -icon warning \
                     -message "Failed to move/copy $::TA(aplDNDSource) under $folderName\n$result\n$::errorInfo"
            }
         } else {
            APL::treeSelect $::TA(aplDNDSource)
         }
         set ::TA(aplDND) 0
      }
   }
}
#------------------------------------------------------------------------------
# Function   : APL::findNextSong
# Description: Highlight the next song in tree with matching text
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc APL::findNextSong {text} {
   if {[catch {aplLB::getActiveFullName} fullName]} {
      set pos 0
   } else {
      # Locate where song is now, and search from there
      set pos [db::whereInDb $fullName]
      if {$pos < 0} {
         set pos 0
      }
   }
   # Search from current place down
   set npos [db::matchInDb "*$text*" [incr pos]]
   if {$npos >= 0} {
      SelectFile [db::TrackFromDb $npos] 1
      return 1
   } else {
      # search from top
      set npos [db::matchInDb "*$text*" 0]
      if {$npos >= 0} {
         SelectFile [db::TrackFromDb $npos] 1
         return 1
      }
   }
   return 0
}
#------------------------------------------------------------------------------
# Function   : APL::findAllSongs
# Description: Locate all tracks containing the text
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc APL::findAllSongs {text} {
   return 0
}
#------------------------------------------------------------------------------
# Function   : APL::positionFolder
# Description: Bound to cursor keys, this moves the selection within the tree
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc APL::positionFolder {how} {
   set current $::TA(highlightPath)
   set new ""
   # Move to the appropriate folder
   switch -- $how {
      prev {
         set folders [aplFolder::getFolders [file dirname $current] 1]
         set pos [sortedSearch $folders $current]
         if {$pos > -1} {
            set new [lindex $folders [incr pos -1]]
         }
      }
      next {
         set folders [aplFolder::getFolders [file dirname $current] 1]
         set pos [sortedSearch $folders $current]
         if {$pos > -1} {
            set new [lindex $folders [incr pos 1]]
         }
      }
      left {
         # Go up one
         #set new [file dirname $current]
         set folders [aplFolder::foldersMatchingPattern * 1]
         set pos [sortedSearch $folders $current]
         if {$pos > -1} {
            set new [lindex $folders [incr pos -1]]
         }
      }
      right {
         #set folders [aplFolder::getFolders $current 1]
         set folders [aplFolder::foldersMatchingPattern * 1]
         set pos [sortedSearch $folders $current]
         if {$pos > -1} {
            set new [lindex $folders [incr pos 1]]
         }
      }
      default {
      }
   }
   if {[file exists $new]} {
      SelectFile $new 1
   }
}
#------------------------------------------------------------------------------
# Function   : APL::findNextFolder
# Description: Highlight the next folder in tree with matching text
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc APL::findNextFolder {text} {
   set folders [aplFolder::foldersMatchingPattern * 1]
   if {[catch {sortedSearch $folders $::TA(highlightPath)} pos]} {
      set pos -1
   }
   # Now search from current place
   # TODO: Verify found folder is mounted, if not continue search
   # check in TA(mediaDirs)
   foreach {folder} [lrange $folders [incr pos] end] {
      if {[string match -nocase "*$text*" [file tail $folder]] && [catalog::isMountedFolder $folder]} {
         SelectFile $folder 1
         return 1
      }
   }
   ##
   # Didn't find from current pos, start from top (wrap)
   #
   foreach {folder} [lrange $folders 0 $pos] {
      if {[string match -nocase "*$text*" [file tail $folder]] && [catalog::isMountedFolder $folder]} {
         SelectFile $folder 1
         return 1
      }
   }
}
#------------------------------------------------------------------------------
# Function   : APL::prevSelection
# Description: Move to the last know selected folder
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc APL::prevSelection {args} {
   variable folderList
   variable folderPos
   variable folderNext
   variable folderPrev
   if {$folderPos > 0} {
      incr folderPos -1
      set folder [lindex $folderList $folderPos]
      SelectFile $folder 0
   }
   APL::setNavButtons $folderNext $folderPrev
}
#------------------------------------------------------------------------------
# Function   : APL::prevSelection
# Description: Move to the next know selected folder
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc APL::nextSelection {args} {
   variable folderList
   variable folderPos
   variable folderNext
   variable folderPrev
   if {($folderPos+1)  < [llength $folderList] && $folderPos >= 0} {
      incr folderPos 1
      set folder [lindex $folderList $folderPos]
      SelectFile $folder 0
   }
   APL::setNavButtons $folderNext $folderPrev
}
#------------------------------------------------------------------------------
# Function   : insertUnique
# Description: Insert an item in the list if it is unique
#              folderPos points to location in list of current selection
#              if list is empty, it will be -1 (initial condition)
#              When called, the next entry will be inserted into the list
#              after the current item and the pointer incremented.
#              This only will occur if the item at the current position
#              does not equal the item to be inserted
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc APL::insertUnique {item} {
   Trace
   variable folderList
   variable folderPos
   variable folderNext
   variable folderPrev
   if {$item eq "."} {
      return
   }
   set curItem [lindex $folderList $folderPos]
   set curLen [llength $folderList]
   if {$curItem ne $item} {
      incr folderPos  ;# Point to next item
      if {$folderPos > $curLen} {
         set folderPos $curLen
      }
      lipinsert folderList $folderPos $item
   }
   APL::setNavButtons $folderNext $folderPrev
}
#------------------------------------------------------------------------------
# Function   : APL::historyMenu
# Description: Display a history menu
# Author     : Tom Wilkason
# Date       : 3/5/2002
#------------------------------------------------------------------------------
proc APL::cbHistoryMenu {menu} {
   variable folderList
   variable folderPos
   variable folderNext
   variable folderPrev

   $menu delete 0 end
   # Clamp menu size to 20 high
   set mh 20
   set pos $folderPos
   set end [expr {[llength $folderList]-1}]
   set min [expr {$pos - $mh/2}]
   # If closer than mh/2 from end, offset min back by delta
   if {($end-$pos) < $mh/2} {
      incr min [expr {($end-$pos) - $mh/2}]
   }
   # Clamp min
   if {$end < $mh || $min < 0} {
      set min 0
   }
   # Display up to $mh menus
   set i 0
   set cnt 0
   foreach item $folderList {
      if {$i >= $min && $cnt <= $mh} {
         $menu add radiobutton -variable APL::folderPos -value $i \
            -label "$i  $item" -command [list APL::positionHistory $i]
         incr cnt
      }
      incr i
   }
}
#------------------------------------------------------------------------------
# Function   : APL::positionHistory
# Description: Position the folder at some location and update the buttons
# Author     : Tom Wilkason
# Date       : 3/5/2002
#------------------------------------------------------------------------------
proc APL::positionHistory {location} {
   variable folderList
   variable folderPos
   variable folderNext
   variable folderPrev
   set folder [lindex $folderList $folderPos]
   SelectFile $folder 0
   APL::setNavButtons $folderNext $folderPrev

}
#------------------------------------------------------------------------------
# Function   : APL::setNavButtons
# Description: Set the state of the navigation buttons
# Author     : Tom Wilkason
# Date       : 3/5/2002
#------------------------------------------------------------------------------
proc APL::setNavButtons {next prev} {
   variable folderList
   variable folderPos
   # Configure previous button
   if {$folderPos > 0} {
      $prev config -state normal
   } else {
      $prev config -state disabled
   }
   # configuration next button
   if {$folderPos < ([llength $folderList]-1)} {
      $next config -state normal
   } else {
      $next config -state disabled
   }
}
#------------------------------------------------------------------------------
# Function   : APL::popWatch
# Description: Tempoaraly put a busy watch up
# Author     : Tom Wilkason
# Date       : 3/5/2002
#------------------------------------------------------------------------------
proc APL::popWatch {{time 300}} {
   APL::busyCursor 1
   update idletasks
   after 250 APL::busyCursor 0
}

#------------------------------------------------------------------------------
# Fire a filter action on incoming key strokes, only entered characters
# will fire it.
#------------------------------------------------------------------------------
proc APL::incrSearchFilter {Key} {
   if {$::snackAmpSettings(incrKeyFilter)} {
      after cancel APL::setSearchFilter
      if {[string length $Key]==1 && [string length $::TA(SAFilter)]>1} {
         after 1000 APL::setSearchFilter
      }
   }
}
#------------------------------------------------------------------------------
# Toggle the filter button and update the track listing
#------------------------------------------------------------------------------
proc APL::setSearchFilter {args} {
   after cancel APL::setSearchFilter
   APL::setFilter 1
   trackList::populate "" $::PLB 1 1
}
#----------------------------------------------------------------------------
# Set the filter to some mode
#----------------------------------------------------------------------------
proc APL::setFilter {how} {
   set ::TA(useSAFilter) $how
}

#------------------------------------------------------------------------------
# Function   : APL::aplMenu
# Description: Create the main menu for the APL Window, reuse
#              pop-up menus for the tree and table.
# Author     : Tom Wilkason
# Date       : 5/27/2002
#------------------------------------------------------------------------------
proc APL::aplMenu {W} {
   global SA
   set M $W.aplMenu
   bindings::init $W menuBindings
   # This forces a moden L&F, top menus are show
   # as raised, sub-menus are hightlight selected
   menu $M \
      -activebackground [$SA cget -background] \
      -activeforeground [$SA cget -highlightcolor] \
      -activeborderwidth 1 -font menu

   if {$::tcl_platform(platform)=="unix"} {
      $M configure -borderwidth 0   ;# for linux
   }
   # attach it to the main window
   $W config -menu $M
   # Create more cascade menus
   foreach {m u} {File 0 Edit 0 Folders 1 Tracks 0 AutoPlaylists 0 Columns 2 Favorites 2 Help 0} {
      switch -- $m {
         Folders {
            # Use existing menu
            $M add cascade -label $m -menu $::TREE.popup  -underline $u
         }
         Tracks {
            # Use existing menu
            $M add cascade -label $m -menu $::PLB.popup  -underline $u
         }
         AutoPlaylists {
            # Use existing menu
            $M add cascade -label $m -menu $W.mb.files.list -underline $u
         }
         Favorites {
            # Use existing menu
            $M add cascade -label $m -menu $W.mb.favs.list -underline $u
         }
         Columns {
            set $m [menu $M.m$m -tearoff 1]
            $M add cascade -label $m -menu $M.m$m  -underline $u
            foreach {item} [lrange $::TA(cols) 1 end] {
               $Columns add checkbutton -label $item -variable ::snackAmpSettings(col,$item,show)
            }
         }
         Subfolders {
            set $m [menu $M.m$m]
            $M add cascade -label $m -menu $M.m$m  -underline $u
            foreach {subfolders} {0 1 2 3 4 5 6 7 8} {
               $Subfolders add radiobutton -label "Show [expr {$subfolders+1}] levels" -value $subfolders -variable ::snackAmpSettings(treeDepth)
            }
         }
         History {
            # Use existing menu
            $M add cascade -label $m -menu $W.mb.prevHistory.list  -underline $u
         }
         default {
            set $m [menu $M.m$m]
            $M add cascade -label $m -menu $M.m$m  -underline $u
         }
      }
   } ;#end foreach
   ##
   # Here is how to add a menu to the "system" menu
   # Use it to put "pinned" command
   #
   ontop::pinOnSystemMenu $M .wFC
   ##
   # File Menu
   #
   bindings::addCommand $File "Manage Media Library..."  {catalog::manageMediaFolders} 0
   bindings::addCommand $File "Manage Portable Player Library..."  {catalog::managePlayerFolders} 7
   bindings::addCommand $File "Index Media Library" {catalog::IndexAskFirst} 1
   bindings::addCommand $File "Check for Media Changes" {catalog::rebuildIfNewMedia} 1
   $File add separator
   bindings::addCommand $File "Save AutoPlaylist" {saveCurrentAPL 0} 0
   bindings::addCommand $File "Save AutoPlaylist As..." {SaveFolderFile}
   bindings::addCommand $File "Apply AutoPlaylist" {applyCurrentAPL} 0
   $File add separator
   bindings::addCommand $File "Sync to Playing" {SelectFile 1} 1
   bindings::addCommand $File "Sync to Playlist Selected" {SelectFile 0}
   $File add separator
   bindings::addCommand $File "Close" {APL::close} 0
   # Edit menu
   bindings::addCommand $Edit "Copy Selection to Cliboard" {clipboard clear;APL::cut} 0
   bindings::addCommand $Edit "Append Selection to Cliboard" {APL::cut} 0
   $Edit add separator
   bindings::addCommand $Edit "Copy Clipboard to Selected Node" {APL::PasteFolderToTree copy} 0
   bindings::addCommand $Edit "Move Clipboard to Selected Node" {APL::PasteFolderToTree move} 0
   $Edit add separator
   bindings::addCommand $Edit "ID3 Group Tag Editor for Selected Node"  {::APL::treeEditId3}
   bindings::addCommand $Edit "ID3 Table Editor for Selected Node"  {::APL::treeNameToTag 0} 0
   bindings::addCommand $Edit "ID3 Tag To Name for Selected Node" {::APL::treeTagToName}
   $Edit add separator
   bindings::addCommand $Edit "ID3 Tag Editor for Selected Tracks"  {::APL::EditId3} 0
   bindings::addCommand $Edit "ID3 Table Editor for Selected Tracks"  {::APL::NameToTag 0}
   bindings::addCommand $Edit "ID3 Tag To Name Editor for Selected Tracks" {::APL::TagToName}

   $Edit add separator
   $Edit add checkbutton -label "Edit Cell Mode" -variable trackList::editCellCheckState \
      -command {trackList::changeEdit} -compound left -image saImages.edit-16
   $Help add command -label "About" -command \
   {
      tk_messageBox -type ok -icon info -title "Help" -message \
"The Autoplaylist Manager allow allows you to set/save/restore preferred folders.
You can also rename, remove and move files. With the mouse you can:
- Select a folder to include (checked) or exclude (unchecked)
- Control Click: Check selected and all subfolders
- Alt Click: Uncheck selected and all subfolders
- Double Click: Toggle expand/collapse selected folder
- Right MB Click: Menu allowing you do above actions, apply and/or
  save the settings.
   - Rescan folder, sync internal list with any changes that has taken place
     on the disk.
   - Add folder to playlist, copy all of this folders files to the main
     playlist.
   - Move files copied to clipboard to highlighted folder
   - Create a new folder or rename and existing folder
   - Remove a folder and all of its subfolders and files.
- Using drag and drop you can move an entire tree limb from one
  folder to another.

From the menu and menu bar you can:
- Apply the settings to the playlist (does not save to file)
- Save, applies and saves the settings to the current AutoPlaylist file
- Save as, save the settings to a new file name and make it current.
- File List checkbox, toggle a view of the files contained within the
  selected folder at the bottom.
- Close the window

You can drag highlighted files from the playlist to a folder
in the tree view to move them.
"
   }

}
##
# Prepend the common year to the album name
#
proc APL::prependFolderYear {args} {
   macros::prependFolderYear $::TA(highlightPath)
}
proc APL::fixFolderChildren {args} {
   macros::fixFolderChildren $::TA(highlightPath)
}

#------------------------------------------------------------------------------
#  Correct the folder name garbling
#------------------------------------------------------------------------------
proc APL::FixFolderName {args} {
   global TA
   set oldName $TA(highlightPath)
   regsub -all -- {_} $oldName { } tpath
   set newName [file join  [file dirname $TA(highlightPath)]\
      [stringTitle [file tail $tpath]]]
   if {$oldName ne $newName} {
      catalog::moveFolder $oldName $newName
      APL::treeSelect $newName
      APL::viewHighlighted $::TREE
      set TA(highlightPath) $newName
   }

}
proc APL::SelectedFolderRenameInPlace {{select 0}} {
   set W $::TREE
   # Note access into tree global here
   set tag [Tree::activetag $::TREE]
   if {[string length $tag] == 0} {
      return
   }
   set w  .canRndialog
   destroy $w
   toplevel $w -class SnackAmp

   set oldName [file tail $::TA(highlightPath)]
   set folder $::TA(highlightPath)
   set base [file dirname $folder]
   foreach {x y - -} [$W bbox $tag] {break}
   # Need to account for scrolled region
   # bu subtracting out canvas displayed 0 value
   set rx [winfo rootx $W]
   set ry [winfo rooty $W]
   set X [expr {$rx+$x-round([$W canvasx 0])}]
   set Y [expr {$ry+$y-round([$W canvasy 0])}]
   set font [$W itemcget $tag -font]
   set bg [$W cget -background]
   set fg $Tree::fg
   set ::TA(rename) $oldName
   set wid [expr {[string length $oldName]+12}]
   if {$wid<60} {
      set wid 60
   }
   pack [entry $w.rn -bd 2 -relief groove -font $font -insertbackground $fg \
      -width $wid -bg $bg -fg $fg -textvariable ::TA(rename)]
   wm withdraw $w
   wm overrideredirect $w 1
   wm transient $w [winfo toplevel $W]
   if {$select} {
      $w.rn selection range 0 end
   }
   focus -force $w.rn
   # with it hiden, let the layout take place so we know its geometry
   update idletasks
   wm geometry $w "+$X+$Y"
   bind $w <FocusOut> {
      destroy .canRndialog
      set ::TA(rename) ""
   }
   bind $w <Escape> {
      destroy .canRndialog
      set ::TA(rename) ""
   }
   bind $w <Return> [list destroy .canRndialog]
   wm deiconify $w
   raise $w
   # Force on top if possible
   update idletasks
   ontop::ifpossible $w $w
   ##
   # Wait for window to close
   #
   tkwait window $w
   ##
   # Back from window closure
   #
   focus -force [winfo parent $::TREE]
   if {($::TA(rename) ne "") && ($::TA(rename) ne $oldName)} {
      if {[catch {
         set newFullName [file join $base $::TA(rename)]
         catalog::moveFolder $folder $newFullName
         APL::treeSelect $newFullName
         APL::viewHighlighted $::TREE
         set ::TA(highlightPath) $newFullName
         set res 1
      } result] \
      } then {
         tk_messageBox -type ok -icon warning -message "Could not rename folder \n$folder to \n$newFullName\n$result"
         set res 0
      }
   }
}
#------------------------------------------------------------------------------
# Function   : APL::treeBind
# Description: Add bindings to the autoplay list tree
# Author     : Tom Wilkason
# Date       : 12/23/2001
#------------------------------------------------------------------------------
proc APL::treeBind {args} {
   Trace
   ##
   # Pop-up menu, save some data for use by callbacks
   #
   bind $::TREE <ButtonPress-3> {
      focus -force [winfo parent %W]
      set lbl [Tree::labelat %W %x %y]
      Tree::setselection %W $lbl
      if {$lbl != ""} {
         if {[catch {APL::GetTreeFolder %W $lbl} folderName]} {
            return
         }
         APL::insertUnique $folderName
         set TA(highlightPath) $folderName
         #set TA(highlightEntry) $lbl
         #set TA(treeID) %W
         aplLB::UpdateList $folderName [winfo ismapped $::PLB]
         tk_popup %W.popup %X %Y
      }
      unset -nocomplain lbl
   }
   set ::TA(aplDND) 0
   ##
   # macros::Highlight the current folder, save away the selected folder
   #
   APL_TreeNodeAssign <ButtonPress-1> {
      catch {
         focus -force [winfo parent %W]
         set lbl [Tree::labelat %W %x %y]
         Tree::setselection %W $lbl
         if {![catch {APL::GetTreeFolder %W $lbl} folderName]} {
            set ::TA(highlightPath) $folderName
            set ::TA(aplDNDSource) $folderName
            set ::TA(aplDragTarget) $folderName
            APL::insertUnique $folderName
            APL::setFilter 0
# This interfears with the tcl only DND, I need to change completely over to tkdnd
# Once I am sure it runs on linux.
            after idle [list aplLB::UpdateList $folderName [winfo ismapped $::PLB]]
         }
         # Update file listbox if it exists.
      }
   }
   ##
   # During a drag, highlight the current folder under the cursor
   #
   APL_TreeNodeAssign <B1-Motion> {
      set lbl [Tree::labelat %W %x %y]
      if {![catch {APL::GetTreeFolder %W $lbl} folderName]} {
         if {![string match $folderName $::TA(aplDragTarget)]} {
            Tree::setselection %W $lbl
            set ::TA(aplDragTarget) $folderName
         }
      }
      $::TREE config -cursor hand2
      set ::TA(aplDND) 1
      unset -nocomplain lbl
   }
   ##
   # Drop action coming from the Tree
   #
   APL_TreeNodeAssign <ButtonRelease-1> {
      APL::busyCursor 0
      set lbl [Tree::labelat %W %x %y]
      set targetWin [winfo containing %X %Y]
      if {[winfo exists $targetWin] && $::TA(aplDND)} {
         APL::busyCursor 1
         if {$targetWin == $::TREE} {   ;# move the folder
            APL::cbDndTreeToTree %W $lbl
         } elseif {$targetWin == [LB::body]} {
            APL::cbDndTreeToClipboard %X %Y   ;# get to clipboard
            APL::cbDndPLBToLB %X %Y    ;# now works like PLB to LB
         } elseif {[winfo toplevel $targetWin] == [winfo toplevel $::LB]} {
            # Improve to append to bottom here
            APL::cbDndTreeToClipboard %X %Y   ;# get to clipboard
            APL::cbDndPLBToLB 0 0    ;# now works like PLB to LB
         }
         APL::busyCursor 0
      }
      unset -nocomplain lbl
   }
   ##
   # Paste Files/folders (not on canvas)
   #
   APL_TreeAssign <<Paste>> {APL::PasteFolderToTree}
   ##
   # Append selected folder to clipboard
   #
   APL_TreeAssign <<Copy>> {APL::popWatch;clipboard append "[list $TA(highlightPath)] "}
   APL_TreeAssign <<Cut>> {APL::popWatch;clipboard clear;clipboard append "[list $TA(highlightPath)] "}
   ##
   # Cursor movement keys
   #
   APL_TreeAssign <Left> {APL::positionFolder left}
   APL_TreeAssign <Right> {APL::positionFolder right}
   APL_TreeAssign <Up>   {APL::positionFolder prev}
   APL_TreeAssign <Down> {APL::positionFolder next}
   ##
   # Assign hotkey to rename
   #
   APL_TreeAssign <F2> {APL::SelectedFolderRenameInPlace 1}
   APL_TreeAssign <Delete> {::catalog::removeFolder $::TA(highlightPath)}
   # Toggle open/close on Return key
   APL_TreeAssign <Return> {Tree::toggleExpand $::TREE [APL::pathToTreeName $::TA(highlightPath)]}
   APL_ListAssign <Escape> {
      if {$trackList::editCellState} {
         trackList::changeEdit
      }
   }
   # Rename full folder name
   APL_TreeAssign <Alt-F2> {::APL::renameFolder $::TA(highlightPath)}
   ##
   # Set this and all child checkboxes
   #
#    APL_TreeNodeAssign <Control-1> {
#       set lbl [Tree::labelat %W %x %y]
#       Tree::setselection %W $lbl
#       Tree::setDirty $::TREE
#       if {![catch {APL::GetTreeFolder %W $lbl} folderName]} {
#          APL::setSomeFolders $folderName 1
#       }
#       APL::insertUnique $folderName
#    }
   ##
   # Clear this and all child checkboxes
   #
#    APL_TreeNodeAssign <Alt-1> {
#       set lbl [Tree::labelat %W %x %y]
#       Tree::setDirty $::TREE
#       Tree::setselection %W $lbl
#       if {![catch {APL::GetTreeFolder %W $lbl} folderName]} {
#          APL::setSomeFolders $folderName 0
#       }
#       APL::insertUnique $folderName
#    }
   ##
   # Recurse expand on double click
   #
   APL_TreeNodeAssign <Double-1> {
      APL::busyCursor 1
      Tree::toggleExpand %W [Tree::labelat %W %x %y] 10
      APL::busyCursor 0
   }
   ##
   # Now build the pop-up menu for the tree
   #
   set m $::TREE.popup
   menu $m
   set P [winfo parent $::TREE]
   bindings::init $P menuBindings
   menu $m.folder
   $m add cascade -label "Folder" -menu $m.folder -underline 0 \
      -compound left -image saImages.Tree.Folder.close

   ##
   # Folder submenu
   #
   bindings::addCommand $m.folder "Create" {APL::createFolder $TA(highlightPath)}
   bindings::addCommand $m.folder "Rename" {APL::SelectedFolderRenameInPlace 1} 0
   bindings::addCommand $m.folder "Rename Path" {::APL::renameFolder $::TA(highlightPath)} 0
   bindings::addCommand $m.folder "Remove" {::catalog::removeFolder $::TA(highlightPath)}
   $m.folder add separator
   bindings::addCommand $m.folder "Correct Folder Name" {APL::FixFolderName} 0
   $m.folder add separator
   bindings::addCommand $m.folder "Expand Children"  {APL::expandChildren 1} 0
   bindings::addCommand $m.folder "Collapse Children" {APL::expandChildren 0}
   $m.folder add separator
   bindings::addCommand $m.folder "Copy Folder to Clipboard" {APL::popWatch;clipboard clear "[list $TA(highlightPath)] "}
   bindings::addCommand $m.folder "Append Folder to Clipboard" {APL::popWatch;clipboard append "[list $TA(highlightPath)] "}
   $m.folder add separator
   bindings::addCommand $m.folder "Move Clipboard Files Here" {APL::PasteFolderToTree move}
   bindings::addCommand $m.folder "Copy Clipboard Files Here" {APL::PasteFolderToTree copy}
   $m.folder add separator
   bindings::addCommand $m.folder "Create Stream Shortcut Here" {
      if {[playurl::checkuskfolder]} {
         playurl::addPromptUrl $TA(highlightPath)
         aplLB::UpdateList $::TA(highlightPath) 1
      }
   }
   $m.folder add separator
   bindings::addCommand $m.folder "Move Folder Tracks under Artist Tag" {::APL::insertTagName tree {Artist}}
   bindings::addCommand $m.folder "Move Folder Tracks under Artist/Album Tags" {::APL::insertTagName tree {Artist Album}}
   bindings::addCommand $m.folder "Move Folder Tracks under Album Tag" {::APL::insertTagName tree {Album}}

   ##
   # Selected submenu
   #
   menu $m.selected
   $m add cascade -label "Viewed Tracks" -menu $m.selected -underline 0
   bindings::addCommand $m.selected "Correct File Names" {fixplus::correctCase [aplLB::getAllNames 1]}
   ##
   # Media Players
   #
   menu $m.mplayer
   $m add cascade -label "Media Player" -menu $m.mplayer -underline 0 \
      -compound left -image saImages.mp3_player.close
   set pm1 [menu $m.mplayer.medplay1 \
      -postcommand [list gui::mediaPlayerPopup $m.mplayer.medplay1 {APL::getDepthFiles $::snackAmpSettings(treeDepth)}]]
   set pm2 [menu $m.mplayer.medplay2 \
      -postcommand [list gui::mediaPlayerPopup $m.mplayer.medplay2 {APL::getDepthFiles 100}]]
   $m.mplayer add cascade -label "Copy Shown Levels to Media Player" -menu $pm1 \
      -underline 0 -compound left -image saImages.copy
   $m.mplayer add cascade -label "Copy Branch to Media Player" -menu $pm2 \
      -underline 0 -compound left -image saImages.copy

   ##
   # Menus for ratings
   #
   menu $m.rating -tearoff 1
   $m add cascade -label "Ratings" -menu $m.rating -underline 0 \
      -compound left -image saImages.iconthumbsup
   bindings::addCommand $m.rating "+ Folder Rating" {::db::bumpFolderRating $::TA(highlightPath) 0 1}
   bindings::addCommand $m.rating "- Folder Rating" {::db::bumpFolderRating $::TA(highlightPath) 0 -1}
   bindings::addCommand $m.rating "Reset Ratings"   {::db::setFolderRating $::TA(highlightPath) 0 "--"}
   $m.rating add separator
   bindings::addCommand $m.rating "+ Viewed Tracks Rating" {::db::bumpFolderRating $::TA(highlightPath) $snackAmpSettings(treeDepth) 1}
   bindings::addCommand $m.rating "- Viewed Tracks Rating" {::db::bumpFolderRating $::TA(highlightPath) $snackAmpSettings(treeDepth) -1}
   bindings::addCommand $m.rating "Reset Ratings"   {::db::setFolderRating $::TA(highlightPath) $snackAmpSettings(treeDepth) "--"}
   $m.rating add separator
   bindings::addCommand $m.rating "+ Tree Rating" {::db::bumpFolderRating $::TA(highlightPath) 50 1}
   bindings::addCommand $m.rating "- Tree Rating" {::db::bumpFolderRating $::TA(highlightPath) 50 -1}
   bindings::addCommand $m.rating "Reset Ratings"   {::db::setFolderRating $::TA(highlightPath) 50 "--"}
   ##
   # Catalog submenu
   #
   menu $m.catalog
   $m add cascade -label "Media Library" -menu $m.catalog -underline 0 \
      -compound left -image saImages.Disk.Folder.close
      bindings::addCommand $m.catalog "Index folder and children" {
         APL::indexFolderChildren $TA(highlightPath)
      } 0
      bindings::addCommand $m.catalog "Catalog folder and children" {
         db::catalog $TA(highlightPath) 1 100
         waitWindow::close
      } 8
      bindings::addCommand $m.catalog "Catalog Viewed Tracks" {
         APL::catalogFolder $TA(highlightPath)
      } 8
      $m.catalog add separator
      bindings::addCommand $m.catalog "Export database for this node" {folderManager::exportCatalog $TA(highlightPath)}
      bindings::addCommand $m.catalog "Import database for this node" {folderManager::importCatalog $TA(highlightPath)}

   ##
   # Playlist
   #
   menu $m.playlist -tearoff 1
   $m add cascade -label "Playlist" -menu $m.playlist -underline 0 -compound left -image saImages.playlist
   bindings::addCommand $m.playlist "Append viewed tracks" {APL::AddToPlayList $TA(highlightPath)} 0
   bindings::addCommand $m.playlist "Queue viewed tracks" {APL::InsertToPlayList $TA(highlightPath)} 0
   bindings::addCommand $m.playlist "Queue then play viewed tracks" {aplLB::playSelectedFiles}
   $m.playlist add separator
   bindings::addCommand $m.playlist "Export Playlist from AutoPlaylist" {folderManager::buildPlaylist $TA(highlightPath)}
   ##
   # AutoPlaylist
   #
   menu $m.apl
   $m add cascade -label "AutoPlaylist" -menu $m.apl -underline 0
   bindings::addCommand $m.apl "Set this folder and children"  {APL::setSomeFolders $TA(highlightPath) 1} 0
   bindings::addCommand $m.apl "Clear this folder and children" {APL::setSomeFolders $TA(highlightPath) 0} 1
   $m.apl add separator
   bindings::addCommand $m.apl "Apply" {
      macros::applyCurrentAPL
   }
   bindings::addCommand $m.apl "Save" {
      macros::saveCurrentAPL 0
   }
   ##
   # ID3
   #
   menu $m.id3
   $m add cascade -label "ID3" -menu $m.id3 -underline 0 -compound left -image saImages.edit-16
   bindings::addCommand $m.id3 "ID3 Group Tag Editor"  {::APL::treeEditId3}
   bindings::addCommand $m.id3 "ID3 Table Editor"  {::APL::treeNameToTag 0} 0
   bindings::addCommand $m.id3 "ID3 Tag To Name" {::APL::treeTagToName}
   ##
   # Auto-Level
   #
   menu $m.gain
   $m add cascade -label "Auto-Level" -menu $m.gain -underline 0  -compound left -image saImages.autoLevel
   # Highlighted Folder Tracks
   menu $m.gain.folder
   $m.gain add cascade -label "Selected Folder Tracks" -menu $m.gain.folder -underline 0
   bindings::addCommand $m.gain.folder "Compute Track Gain" {gainControl::computeSome [db::filesInFolder $::TA(highlightPath)]}
   bindings::addCommand $m.gain.folder "Compute Average Gain" {gainControl::averageSome [db::filesInFolder $::TA(highlightPath)]}
   bindings::addCommand $m.gain.folder "Compute Album Gains" {gainControl::folderAlbumGains $::TA(highlightPath)}
   bindings::addCommand $m.gain.folder "Reset Gain" {gainControl::resetSome [db::filesInFolder $::TA(highlightPath)]}
   if {[string length $::snackAmpSettings(mp3gainLoc)]} {
      $m.gain.folder add separator
      bindings::addCommand $m.gain.folder "mp3gain Tracks" {mp3gain::mp3gain 0 "/s s /r /d 3.0"}
      bindings::addCommand $m.gain.folder "mp3gain Albums" {mp3gain::mp3gain 0 "/s s /a /d 3.0"}
   }

   # Viewed Tracks
   menu $m.gain.viewed
   $m.gain add cascade -label "Viewed Tracks" -menu $m.gain.viewed -underline 0
   bindings::addCommand $m.gain.viewed "Compute Track Gain" {gainControl::computeSome [aplLB::getAllNames 1]}
   bindings::addCommand $m.gain.viewed "Compute Average Gain" {gainControl::averageSome [aplLB::getAllNames 1]}
   bindings::addCommand $m.gain.viewed "Reset Gain" {gainControl::resetSome [aplLB::getAllNames 1]}
   $m.gain.viewed add separator
   bindings::addCommand $m.gain.viewed "Increase Gain by +5%" {gainControl::adjustSome [aplLB::getAllNames 1] .05}
   bindings::addCommand $m.gain.viewed "Decrease Gain by -5%" {gainControl::adjustSome [aplLB::getAllNames 1] -.05}

   # Highlighted Folder and subfolders Tracks
   menu $m.gain.sub
   $m.gain add cascade -label "Selected Folder and Subfolders Tracks" -menu $m.gain.sub -underline 0
   bindings::addCommand $m.gain.sub "Compute Track Gain" {gainControl::computeSome [db::filesMatchingPattern [file join $::TA(highlightPath) *]]}
   bindings::addCommand $m.gain.sub "Compute Album Gain (Folders)" {gainControl::averageAlbums [aplFolder::foldersMatchingPattern [file join $::TA(highlightPath) *] 1]}
   bindings::addCommand $m.gain.sub "Compute Album Gains (Albums)" {gainControl::folderAlbumGains "$::TA(highlightPath)*"}
   bindings::addCommand $m.gain.sub "Re-check Gain Tags" {gainControl::setGains "$::TA(highlightPath)*"}
   bindings::addCommand $m.gain.sub "Reset Gain" {gainControl::resetSome [db::filesMatchingPattern [file join $::TA(highlightPath) *]]}
   if {[string length $::snackAmpSettings(mp3gainLoc)]} {
   $m.gain.sub add separator
      bindings::addCommand $m.gain.sub "mp3gain Tracks" {mp3gain::mp3gain 1 "/s s /r /d 3.0"}
      bindings::addCommand $m.gain.sub "mp3gain Album" {mp3gain::mp3gain 1 "/s s /a /d 3.0"}
   }

   ##
   # Misc
   #
   $m add separator
   bindings::addCommand $m "Properties..." {
      APL::treeProperties $::TA(highlightPath)
   }

}
#------------------------------------------------------------------------------
# Function: APL::cut
#
# Copy the files contained in the selected path to the clipboard
# for later moving.
#------------------------------------------------------------------------------
proc APL::cut {args} {
   set here [focus]
   if {$here == [winfo parent $::TREE]} {
      APL::popWatch
      clipboard append "[list $::TA(highlightPath)] "
   } elseif {[winfo parent $here] == $::PLB} {
      APL::popWatch
      clipboardCopy $::PLB 1
   } else {
      tk_messageBox -type ok -icon warning \
         -message "You must have a selection in either the tree of file list area"
   }
}
#------------------------------------------------------------------------------
# Function   : APL::treeProperties
# Description: Show the stats of the tree under the folder
# Author     : Tom Wilkason
# Date       : 6/22/2002
#------------------------------------------------------------------------------
proc APL::treeProperties {selectedFolder} {
   ##
   # Total up stats for all files within a tree
   #
   set txt [db::statsText "" $selectedFolder]
   tk_messageBox -type ok -icon info -title "Properties for $selectedFolder" \
      -message "Information for '$selectedFolder'\n\n$txt"
   return
}
#------------------------------------------------------------------------------
# Function   : APL::PasteFolderToTree
# Description: Move the files in the clipboard to the selected folder
# Author     : Tom Wilkason
# Date       : 6/18/2002
#------------------------------------------------------------------------------
proc APL::PasteFolderToTree {{mode auto}} {
   APL::popWatch 100
   set folderName $::TA(highlightPath)
   switch -- $mode {
      auto {
         if {[catalog::isMediaPlayer $folderName]} {
            if {[catalog::copyClipboardFiles $folderName] > 0} {
               aplLB::UpdateList $folderName [winfo ismapped $::PLB]
               APL::treeSelect $folderName
            }
         } else {
            if {[catalog::moveClipboardFiles $folderName] > 0} {
               aplLB::UpdateList $folderName [winfo ismapped $::PLB]
               APL::treeSelect $folderName
            }
         }
      }
      copy {
         if {[catalog::copyClipboardFiles $folderName] > 0} {
            aplLB::UpdateList $folderName [winfo ismapped $::PLB]
            APL::treeSelect $folderName
         }
      }
      move {
         if {[catalog::moveClipboardFiles $folderName] > 0} {
            aplLB::UpdateList $folderName [winfo ismapped $::PLB]
            APL::treeSelect $folderName
         }
      }
      default {}
   }
}
#------------------------------------------------------------------------------
# Function   : APL::createFolder
# Description: Prompt then create a subfolder under some level
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc APL::createFolder {folder} {
   set newName "                               "
   set res 0
   if {[getEntryDialog "Enter Folder Name" newName "Create Folder"]=="ok"} {
      if {[catch {
         set newName [file join $folder [string trim $newName]]
         file mkdir $newName
         aplFolder::Set $newName [aplFolder::Get [file dirname $newName]]
         newItem $newName
         APL::treeOpen [file dirname $newName]
         APL::treeSelect [file dirname $newName]
         update idletasks
         APL::viewHighlighted $::TREE
         set res 1
      } result] \
      } then {
         tk_messageBox -type ok -icon warning -message "Could not create folder $newName\n$result"
         set res 0
      }
   }
   return $res
}
#------------------------------------------------------------------------------
# Function   : APL::chooseIcon
# Description: Select the appropropiate icon based on file name
# Author     : Tom Wilkason
# TODO: can this be sped up? Call a function that returns the type
#------------------------------------------------------------------------------
proc APL::chooseIcon {Name} {
   variable icons
   return [lindex $icons [catalog::getIconType $Name]]
}
#------------------------------------------------------------------------------
# Function   : APL::treeSelect
# Description: macros::Highlight a node
# Author     : Tom Wilkason
# Date       : 3/4/2002
#------------------------------------------------------------------------------
proc APL::treeSelect {folder} {
   variable folderList
   Tree::setselection $::TREE [APL::pathToTreeName $folder]
}
#------------------------------------------------------------------------------
# Function   : APL::treeOpen
# Description: Open a node
# Author     : Tom Wilkason
# Date       : 3/4/2002
#------------------------------------------------------------------------------
proc APL::treeOpen {folder {recurse 0}} {
   Tree::open $::TREE [APL::pathToTreeName $folder] $recurse
}
#------------------------------------------------------------------------------
# Return whether a node is open or not
#------------------------------------------------------------------------------
proc APL::isOpen {folder} {
   return [Tree::isOpen $::TREE [APL::pathToTreeName $folder]]
}
#------------------------------------------------------------------------------
# Function   : APL::busyCursor
# Description: Turn on/off a busy cursor
# Author     : Tom Wilkason
# Date       : 3/9/2002
#------------------------------------------------------------------------------
proc APL::busyCursor {busy} {
   if {$busy} {
      .wFC config -cursor watch
      $::TREE config -cursor watch
   } else {
      .wFC config -cursor $::TA(arrow)
      $::TREE config -cursor $::TA(arrow)
   }

}
#------------------------------------------------------------------------------
# Expand or contract the children of the selected node, typically bound
# to a key
#------------------------------------------------------------------------------
proc APL::expandChildren {how {levels 10}} {
   if {$how} {
      Tree::close $::TREE [APL::pathToTreeName $::TA(highlightPath)] $levels
      Tree::open $::TREE [APL::pathToTreeName $::TA(highlightPath)] $levels
   } else {
      Tree::close $::TREE [APL::pathToTreeName $::TA(highlightPath)] $levels
   }
}
#------------------------------------------------------------------------------
# Function   : APL::indexFolderChildren
# Description: Catalog a folder tree, called after it has been modified.
#              Assign to hotkeys
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc APL::indexFolderChildren {folder} {
   global TA
   set folders [list]
   if {$folder eq "/Fixed Folders"} {
      foreach dir [catalog::fixedFolders] {
         lappend folders $dir
      }
   } elseif {$folder eq "/Attached Folders"} {
      foreach dir [catalog::attachedFolders] {
         lappend folders $dir
      }
   } else {
      lappend folders $folder
   }
   # Index each one
   foreach {fld} $folders {
      set opened [APL::isOpen $fld]
      catalog::Index $fld
   }
   if {$opened} {
      APL::treeOpen $folder
   }
   APL::SelectFile [file join $folder xx]  ;# need to get file name selected in listbox, not the folder name
}
# Needed for hotkey compatability
proc APL::catalogFolderChildren {folder} {
   APL::indexFolderChildren $folder
}
#------------------------------------------------------------------------------
# Function   : APL::catalogFolder
# Description: Catalog a single folder, called after it has been modified.
#              Assign to hotkeys
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc APL::catalogFolder {folder} {
   waitWindow::show "Cataloging\n\n$folder\n" 1
   Wait 1
   # Nice to refresh the levels that are showing
   db::cacheFolderDetails $folder $::snackAmpSettings(treeDepth)
   aplLB::UpdateList $folder [winfo ismapped $::PLB]
   waitWindow::close
   Wait 0
}
#------------------------------------------------------------------------------
# Function   : APL::GetTreeFolder
# Description: Return the folder name corresponding to the label
#              Throws an error if can find corresponding folder name
# Author     : Tom Wilkason
# Date       : 12/27/2001
#------------------------------------------------------------------------------
proc APL::GetTreeFolder {W lbl} {
   upvar #0 $W myTree
   regexp -- {^.+?\((.+?)\)$} $myTree($lbl:variable) x Var
   return $Var
}
#----------------------------------------------------------------------------
# Rebuild the entire tree from scratch
#----------------------------------------------------------------------------
proc APL::reBuildTree {args} {
   APL::buildAPL $::TREE
   APL::treeBind
   Tree::open $::TREE / 1
}
#------------------------------------------------------------------------------
# Function   : APL::buildAPL
# Description: Rebuild the Auto Play List tree with new data.
# Author     : Tom Wilkason
# Date       : 11/4/2001
#------------------------------------------------------------------------------
proc APL::buildAPL {tree} {
   variable icons
#   variable showFolderTypes
   global TA snackAmpSettings
   set base [join [lrange [split $tree .] 0 end-1] .]
   destroy $tree
   set ::Tree::font treeFont
   if {$snackAmpSettings(useAPLColors)} {
      set Tree::fg $snackAmpSettings(fg)
      set Tree::bg $snackAmpSettings(bg)
      settings::regColor $tree \
         -background ::snackAmpSettings(bg) \
         -foreground ::snackAmpSettings(fg) \
         -highlight  ::snackAmpSettings(highlight)
   } else {
      set Tree::fg black
      set Tree::bg white
   }
   set Tree::highlight $snackAmpSettings(highlight)
   # To work
   # 1. Text Tree selection outline box color can't be black (saTree ln 544)
   # 2. Text on tablelist for highlighted item can't be white
   Tree::create $tree -bg $Tree::bg -width 300 -height 400 -yscrollcommand [list $base.sb set]
   Tree::configure $tree -rippleSet $::snackAmpSettings(rippleSet)
   Tree::configure $tree -treeLine $::snackAmpSettings(treeLine)
   set i 0
   # TODO: Different icons for base folders (perhaps), hacked below
   if {[llength $snackAmpSettings(mp3Player)]} {
      foreach {item index} {"/Attached Folders" 2 "/Fixed Folders" 3} {
         # This links the variable to the dummy name which we can detect later
         Tree::newitem $::TREE $item  -variable ::teFolders($item) -image [lindex $icons $index]
      }
   }
   ##
   # Build up list of media paths and create node entries for them
   #
   foreach {Path type icon} $TA(mediaDirPair) {
      # Folders won't have a / on end, so we don't want our pattern match to either
      # Title for tree just has the leading slash, nothing else
      if {[llength $snackAmpSettings(mp3Player)]} {
         set TA(Title,$Path) "/$type/[string trim [string map {: {} / { }} $Path]]"
         switch -- $type {
            "Fixed Folders" {
               newItemIcon $Path [lindex $icons 3]
            }
            "Attached Folders" {
               newItemIcon $Path [lindex $icons $icon]
            }
            default {
               newItemIcon $Path [lindex $icons $icon]
            }
         }
      } else {
         set TA(Title,$Path) "/[string trim [string map {: {} / { }} $Path]]"
      }
      # Create each node in this folder
      foreach folder [aplFolder::foldersMatchingPattern ${Path}* 1] {
         newItemIcon $folder [lindex $icons $icon]
      }
   }
   Tree::setDirty $::TREE 0
   pack $tree -side left -fill both -expand 1 -padx 0 -pady 0
}
#----------------------------------------------------------------------------
# For tkdnd, during a drop outside here just return the name of the
# highlighted folder
#----------------------------------------------------------------------------
proc APL::dndSrcTreeDrag {args} {
   Trace
   if {$::tcl_platform(platform) == "unix"} {
      set path [file nativename $::TA(highlightPath)]
   } else {
      set path [list [file nativename $::TA(highlightPath)]]
   }
   # Note, maybe we want to glob files and return them?
   return $path
}
#----------------------------------------------------------------------------
# Dragging from the checked label
#----------------------------------------------------------------------------
proc APL::dndSrcCheckedTreeDrag {args} {
   Trace
   if {[aplFolder::isSet $::TA(highlightPath)]} {
      lappend folders [file nativename $::TA(highlightPath)]
   } else {
      set folders [list]
   }
   foreach folder [aplFolder::foldersMatchingPattern [file join $::TA(highlightPath) *]] {
      if {[aplFolder::isSet $folder]} {
         lappend folders [file nativename $folder]
      }
   }
   if {[llength $folders]>0} {
      return [dndFileList $folders]
   } else {
      return -code break
   }
   #############
#    set tracks [list]
#    foreach {track} [db::filesMatchingPattern [file join $::TA(highlightPath) *]] {
#       if {[aplFolder::isSet [file dirname $track]]} {
#          lappend tracks $track
#       }
#    }
#    if {[llength $tracks]>0} {
#       return $tracks
#    } else {
#       return -code break
#    }
}
#----------------------------------------------------------------------------
# Dragging from the total label
#----------------------------------------------------------------------------
proc APL::dndSrcTotalTreeDrag {args} {
   Trace
   lappend folders [file nativename $::TA(highlightPath)]
   foreach folder [aplFolder::foldersMatchingPattern [file join $::TA(highlightPath) *]] {
      lappend folders [file nativename $folder]
   }
   if {[llength $folders]>0} {
      return [dndFileList $folders]
   } else {
      return -code break
   }
   #############
#    set tracks [list]
#    foreach {track} [db::filesMatchingPattern [file join $::TA(highlightPath) *]] {
#       lappend tracks [file nativename $track]
#    }
#    if {[llength $tracks]>0} {
#       return $tracks
#    } else {
#       return -code break
#    }
}
#----------------------------------------------------------------------------
# After a drag operation, grab the selected files from the apl listbox and
# return them.
#----------------------------------------------------------------------------
proc APL::dndSrcLBDrag {src} {
   Trace
   set useFiles [list]
   set tryFiles [getSelectedTracks $::PLB]
   foreach {file} $tryFiles {
      if {[file exists $file]} {
         lappend useFiles [file nativename $file]
      }
   }
   return [dndFileList $useFiles]
}
#------------------------------------------------------------------------------
# Function   : APL::delItem
# Description: Remove a new node from the tree
# Author     : Tom Wilkason
# Date       : 5/8/2002
#------------------------------------------------------------------------------
proc APL::delItem {folder} {
   Tree::delitem $::TREE [APL::pathToTreeName $folder]
}
#------------------------------------------------------------------------------
# Function   : APL::newItemToRoot
# Description: Add a new node to the tree, and parents if necessary
# Author     : Tom Wilkason
# Date       : 12/18/2005
#------------------------------------------------------------------------------
proc APL::newItemToRoot {folder} {
   set treeFolder [APL::pathToTreeName $folder]
   lappend parents [list $folder $treeFolder]
   set parent [file dirname $folder]
   while {![Tree::hasParent $::TREE $treeFolder]} {
      set treeFolder [APL::pathToTreeName $parent]
      lappend parents [list $parent $treeFolder]
      set p1 [file dirname $parent]
      if {$p1 eq $parent} {
         break
      } else {
         set parent $p1
      }
   }
   # Build it up backwards
   foreach {pair} [lsort -index 0 $parents] {
      foreach {folder name} $pair {
         Tree::newitem $::TREE $name \
            -variable ::teFolders($folder) \
            -image [APL::chooseIcon $folder]
      }
   }
}
#------------------------------------------------------------------------------
# Function   : APL::newItem
# Description: Add a new node to the tree
# Author     : Tom Wilkason
# Date       : 5/8/2002
#------------------------------------------------------------------------------
proc APL::newItem {folder} {
   set name [APL::pathToTreeName $folder]
   # Make sure it is mounted
   if {$name ne ""} {
      set result [Tree::newitem $::TREE $name \
         -variable ::teFolders($folder) \
         -image [APL::chooseIcon $folder]]
      if {$result==0} {
         debug "[me] $folder can't be built\n $name"
         unset -nocomplain ::teFolders($folder)
      }
   }
}
#------------------------------------------------------------------------------
# Function   : APL::newItem
# Description: Add a new node to the tree, a bit faster method allowing icon
#              name to be passed in
# Author     : Tom Wilkason
# Date       : 9/8/2004
#------------------------------------------------------------------------------
proc APL::newItemIcon {folder icon} {
   set name [APL::pathToTreeName $folder]
   if {[string length $name] > 0} {
      # TFW: would be nice to find if it already exists
      if {[catch {
         Tree::newitem $::TREE $name \
            -variable ::teFolders($folder) \
            -image $icon
      } result] } then {
         saLog $::errorInfo
      }
   } else {
      #These folders may not be mounted
      #saLog "No translation for $folder"
   }
}
#----------------------------------------------------------------------------
# Add a new item and set it up all the way to its root
#----------------------------------------------------------------------------
proc APL::addToRoot {folder how} {
   global teFolders
   set hits 0
   set nodes [list]
   while {![info exists teFolders($folder)]} {
      incr hits
      # speed hack
      set teFolders($folder) $how
      lappend nodes $folder
      # Backup one
      set root [file dirname $folder]
      if {$folder eq $root} {
         break
      } else {
         set folder $root
      }
   }
   # Add nodes in the proper order
   foreach {node} [lsort -dictionary $nodes] {
      APL::newItem $node
   }
   return $hits
}
#------------------------------------------------------------------------------
# Build favorites menu
#------------------------------------------------------------------------------
proc APL::FavList {menu} {
   $menu delete 4 end
   $menu add separator
   foreach {favorite} $::snackAmpSettings(Favorites) {
      $menu add command -label $favorite \
      -command [list APL::goAndShow $favorite] \
      -compound left -image saImages.Tree.Folder.open
   }
}
#------------------------------------------------------------------------------
# Clear all favorites
#------------------------------------------------------------------------------
proc APL::clearFavorites {args} {
   if {"yes" eq [tk_messageBox -type yesno -icon question -message "Are you sure you want to remove all of your favorites?"]} {
      set ::snackAmpSettings(Favorites) [list]
   }
}
proc APL::addFavorite {args} {
   lappend ::snackAmpSettings(Favorites) $::TA(highlightPath)
   set  ::snackAmpSettings(Favorites) [lsort -dictionary $::snackAmpSettings(Favorites)]
}

#------------------------------------------------------------------------------
# Function   : APL::remFavorite
# Description: Remove a favorite from the favorite list
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc APL::remFavorite {menu} {
   global TA LB
   $menu delete 0 end
   foreach {fav} $::snackAmpSettings(Favorites) {
      $menu add command -label $fav \
         -command [list APL::remFavoriteEntry $fav] \
         -compound left -image saImages.Tree.Folder.open
   }
}
proc APL::remFavoriteEntry {entry} {
   set pos [lsearch -exact $::snackAmpSettings(Favorites) $entry]
   if {$pos>=0} {
      ldelete ::snackAmpSettings(Favorites) $pos
   }
}

#------------------------------------------------------------------------------
# Function   : APL::EditId3, NameToTag, TagToName
# Description: Operate on ID3 tags for multiple files
# Author     : Tom Wilkason
# Date       : 10/30/2001
#------------------------------------------------------------------------------
proc APL::EditId3 {args} {
   package require id3Editor
   Wait 1
   id3Editor::Edit [aplLB::nameSelected]
   Wait 0
}
proc APL::NameToTag {smart} {
   package require id3Table
   Wait 1
   id3Table::Edit [aplLB::nameSelected] $smart
   Wait 0
}
proc APL::treeNameToTag {smart} {
   package require id3Table
   # Limit to 200 files
   Wait 1
   set files [trackList::getFiles [aplFolder::getFolders $::TA(highlightPath) $::snackAmpSettings(treeDepth)]]
   if {[llength $files] > $::snackAmpSettings(ID3editLimit)} {
      tk_messageBox -type ok -icon warning \
         -message "Limiting operation to first $::snackAmpSettings(ID3editLimit) files"
   }
   id3Table::Edit [lrange $files 0 $::snackAmpSettings(ID3editLimit)]
   Wait 0
}

#------------------------------------------------------------------------------
# Return all files under the selected node for some dept
#------------------------------------------------------------------------------
proc APL::getDepthFiles {depth} {
   set filelist [list]
   set folders [aplFolder::getFolders $::TA(highlightPath) $depth]
   foreach {folder} $folders {
      # Hidden files are ignored
      set files [glob -nocomplain -directory $folder *]
      foreach {file} $files {
         if {![file isdirectory $file]} {
            lappend filelist $file
         }
      }
   }
   return [lsort $filelist]
}

proc APL::treeEditId3 {args} {
   package require id3Editor
   # Limit to 200 files
   Wait 1
   set files [trackList::getFiles [aplFolder::getFolders $::TA(highlightPath) $::snackAmpSettings(treeDepth)]]
   if {[llength $files]>$::snackAmpSettings(ID3editLimit)} {
      tk_messageBox -type ok -icon warning \
         -message "Limiting operation to first $::snackAmpSettings(ID3editLimit) files"
   }
   id3Editor::Edit [lrange $files 0 $::snackAmpSettings(ID3editLimit)]
   Wait 0
}
proc APL::TagToName {args} {
   package require id3Renamer
   Wait 1
   id3Renamer::Edit [aplLB::nameSelected]
   Wait 0
}
proc APL::treeTagToName {args} {
   package require id3Renamer
   # Limit to 200 files
   Wait 1
   set files [trackList::getFiles [aplFolder::getFolders $::TA(highlightPath) $::snackAmpSettings(treeDepth)]]
   if {[llength $files]>$::snackAmpSettings(ID3editLimit)} {
      tk_messageBox -type ok -icon warning \
         -message "Limiting operation to first $::snackAmpSettings(ID3editLimit) files"
   }
   id3Renamer::Edit [lrange $files 0 $::snackAmpSettings(ID3editLimit)]
   Wait 0
}
##
# Move files under the selected ID3 tag names
#
proc APL::insertTagName {what tags} {
   Trace
   global TA
   switch -- $what {
      "list" {
         set files [aplLB::nameSelected]
      }
      "tree" {
         set files [db::filesInFolder $TA(highlightPath) 0]
      }
      default {}
   }
   set fileList [list]
   set targetList [list]
   foreach {file} $files {
      set parts [list]
      set front [file dirname $file]
      set end [file tail $file]
      foreach {tag} $tags {
         set val [string map {: - * - ? - \" '} [db::getVal $file $tag]]   ;#"
         lappend parts $val
      }
      set middle [join $parts /]
      set newFile [file join $front $middle $end]
      lappend fileList $file
      lappend targetList $newFile
   }
   if {[llength $fileList]>0} {
      if {[catalog::moveTrackFolderPairs $fileList $targetList]} {
         tk_messageBox -icon warning -message "Could not rename all files, see console log"
      }
   }
   # Catalog this tree to see the new folders
   APL::indexFolderChildren $::TA(highlightPath)
}
#------------------------------------------------------------------------------
# Rename files by splitting name and moving into subfolders
# so each folder contains one album.
# Note: Only handles Artist - Album - xxx formatted files
#        Adding Album - xxx format
# No error checking for shorter formats.
#------------------------------------------------------------------------------
proc APL::SplitFile {args} {
   set dirName ""
   set Artist ""
   set Album ""
   set sep $::snackAmpSettings(splitChar)
   set files [aplLB::nameSelected]
   foreach file $files {
      if {[string length $dirName] == 0} {
         set dirName [file dirname $file]
      }
      set parts [split [file tail $file] $sep]
      # Skip short names
      if {[llength $parts] < 2} {
         # This doesn't skip everything
         continue
      }
      # Assume Artist - Album format
      foreach {Artist Album} $parts {
         set Artist [string trim $Artist]
         set Album [string trim $Album]
         set F(Artist,$Artist) 1
         set F(Album,$Album) 1
         break
      }
   }
   # Param 1 check for uniqueness
   if {[llength [array names F Artist,*]] > 1} {
      tk_messageBox -message  "Artist Names must be unique"
      return
   }
   ##
   # Create Artist folder and create from there
   #
   if {![string equal -nocase $Artist [file tail $dirName]]} {
      # If Artist Folder doesnt already exist, need to create it
      set base [file join [file dirname $file] $Artist]
   } else {
      # Artist folder already exists, so use it
      set base [file dirname $file]
   }
   # Param 2 check for uniqueness
   # Assume format of Artist/Album - xxx
   if {[llength [array names F Album,*]] > 1} {
      set index 1
      set message  "Do you want to split [llength $files] files into folder '$base'?"
   } else {
      set index 2
      set message "Do you want to split [llength $files] files for '$Artist' into folder '[file join $base $Album]'?"
   }
   ##
   # Move the files
   #
   set result [tk_messageBox -type yesno -icon question -message $message]
   if {$result == "yes"} {
      foreach file $files {
         switch -- $index {
            1 {set newFile [file join $base [string trim [join [lrange [split $file $sep] $index end] $sep]]]}
            2 {set newFile [file join $base $Album [string trim [join [lrange [split $file $sep] $index end] $sep]]]}
            default {}
         }
         if {[catalog::RenameFile $file $newFile]} {
            tk_messageBox -icon warning -message "Could not rename\n$file to \n$newFile\ncheck console log for errors"
         }
      }
      # Catalog this tree to see the new folders
      APL::indexFolderChildren $::TA(highlightPath)
   }
}
#------------------------------------------------------------------------------
# Rename the files based on the folder names beneath them
#------------------------------------------------------------------------------
proc APL::CombineFile {level} {
   set sep $::snackAmpSettings(splitChar)
   set files [aplLB::nameSelected]
   set fromList [list]
   set toList [list]
   foreach file $files {
      set folder [file dirname $file]
      set tail [file tail $file]
      set newTail [list]
      for {set i $level} {$i > 0} {incr i -1} {
         set newTail [linsert $newTail 0 [file tail $folder]]
         set folder [file dirname $folder]
      }
      set newTail [join $newTail " $sep "]
      set tail "$newTail $sep $tail"
      #set newFile [file join $folder $tail]
      set newFile [file join [file dirname $file] $tail] ;# leave in same folder
      lappend fromList $file
      lappend toList $newFile
   }
   if {[catalog::moveTrackFolderPairs $fromList $toList]} {
      tk_messageBox -icon warning -message "Could not rename all files, see console log"
   }
   # Catalog this tree to see the new folders
   #APL::indexFolderChildren $::TA(highlightPath)
}
#------------------------------------------------------------------------------
# Function   : APL::DeleteFiles
# Description: Remove some files from the auto playlist list box
# Author     : Tom Wilkason
# Date       : 10/30/2001
#------------------------------------------------------------------------------
proc APL::DeleteFiles {args} {
   set toDelete [aplLB::nameSelected]
   if {[llength $toDelete] > 0} {
      if {$::snackAmpSettings(dontMoveFiles)} {
         tk_messageBox -icon warning -type ok \
            -message "You don't have permission to move or remove files"
         return
      }
      set result [tk_messageBox -type yesno -title "Please Answer" \
         -icon question -message "Are you sure you want to delete:\n[join $toDelete \n]"]
      if {$result == "yes"} {
         foreach {File} $toDelete {
            catalog::DeleteFile $File
         }
         # Schedule a rescan when all is said and done
         # to reflect the new configuration
         trackList::populateHighighted 1 1
         gui::updateFileDisplay
      }
   }
}
#------------------------------------------------------------------------------
# Function   : APL::AddToPlayList
# Description: Add the entire set of songs in this folder to the playlist
# Author     : Tom Wilkason
# Date       : 10/28/2001
#------------------------------------------------------------------------------
proc APL::AddToPlayList {args} {
   ##
   # Now list the files out
   #
   set Cursor [LB::idxEnd]
   LB::AppendList [aplLB::getAllNames 1]
   LB::activate $Cursor
   gui::updateFileDisplay
}
#------------------------------------------------------------------------------
# Function   : APL::InsertToPlayList
# Description: Insert the entire set of songs in this folder to the playlist
# Author     : Tom Wilkason
# Date       : 10/28/2001
#------------------------------------------------------------------------------
proc APL::InsertToPlayList {args} {
   LB::userInsertList [aplLB::getAllNames 1]
}
#------------------------------------------------------------------------------
# Locate, highlight and expand a folder
# Show files for the folder
#------------------------------------------------------------------------------
proc APL::goAndShow {newName} {
   APL::treeOpen [file dirname $newName]
   APL::treeOpen $newName
   APL::treeSelect $newName
   set ::TA(highlightPath) $newName
   APL::viewHighlighted $::TREE
   if {[winfo ismapped $::PLB]} {
      APL::setFilter 0
      aplLB::UpdateList $newName
   }
}
#------------------------------------------------------------------------------
# Function   : APL::SelectFile
# Description: Given some file, select its folder in the tree if possible
# Author     : Tom Wilkason
# Date       : 6/6/2001
#------------------------------------------------------------------------------
proc APL::SelectFile {Path {pushPath 1}} {
   Trace
   variable folderList
   set W $::TREE
   if {[winfo exists $W] && ([wm state .wFC] == "normal") && [file exists $Path]} {
      if {[aplFolder::folderExists $Path 1]} {
         set file {}
         set Dir $Path
      } else {
         set file $Path
         set Dir [file dirname $Path]
      }
      # Save the path in the navagiation history
      if {$pushPath} {
         insertUnique $Dir
      }
      ##
      # Highlight the node for the file
      #
      APL::goAndShow $Dir
      ##
      # Update file listbox if it exists.
      # Note, must be drawn
      if {[winfo ismapped $::PLB]} {
         ##
         # highlight the item in the listbox
         #
         if {[string length $file] > 0} {
            set contents [aplLB::getAllNames]
            set index [lsearch -exact $contents [file tail $Path]]
            aplLB::setSelection $index
         }
      }
      # Give tree the focus to allow pasting to work
      focus [winfo parent $W]
   }
}
#------------------------------------------------------------------------------
# Function   : APL::viewHighlighted
# Description: Scroll the canvas so the highlighted item is visible
# Author     : Tom Wilkason
# Date       : 3/5/2002
#------------------------------------------------------------------------------
proc APL::viewHighlighted {W} {
   upvar #0 $W myTree
   update idletasks
   set OnScreenHeight [winfo height $W]
   set EntireHeight  [lindex [$W cget -scrollregion] 3]
   if {$EntireHeight > $OnScreenHeight} {
      set CursorY [lindex [$W bbox $myTree(selidx)] 3]
      if {[string length $CursorY] > 0} {
         set MoveTo [expr {double($CursorY-$OnScreenHeight+50)/$EntireHeight}]
         $W yview moveto $MoveTo
      }
   }
}

#------------------------------------------------------------------------------
# Function   : APL::pathToTreeName
# Description: Convert a File Path into a Tree Name
# Author     : Tom Wilkason
# Date       : 6/6/2001
#------------------------------------------------------------------------------
proc APL::pathToTreeName {folder} {
   global TA
   foreach {Path Type Icon} $TA(mediaDirPair) {
      # folder names won't end in a / (except drive vols) so our match pattern can't either
      # The titles are formed to not have path separators within them
      # Note: () or [] in media names need to be escaped for the regsub to work
      if {[string match ${Path}* $folder]} {
         set plen [string length $Path]
         return [file join $TA(Title,$Path) [string trimleft [string range $folder $plen end] /]]
      }
   }
   # Didn't find one mounted
   return {}
}
#------------------------------------------------------------------------------
# Function   : APL::setSomeFolders
# Description: Set/clear folders after base to some value
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc APL::setSomeFolders {base value} {
   aplFolder::SetSome "$base*" $value
   APL::refreshTree
}
#------------------------------------------------------------------------------
# Function   : APL::refreshTree
# Description: Refresh the tree such that any variables updated are relfected
#              in the checkboxes
# Author     : Tom Wilkason
# Date       : 1/11/2001
#------------------------------------------------------------------------------
proc APL::refreshTree {args} {
   global snackAmpSettings TA
   set W $::TREE
   # Save scroll % and restore it when done
   set scroll [lindex [$W yview] 0]
   foreach {Path} $TA(mediaDirs) {
      Tree::close $W $TA(Title,$Path)
      Tree::open $W $TA(Title,$Path)
   }
   $W yview moveto $scroll
}
#------------------------------------------------------------------------------
# Function   : APL::setFileListToShow
# Description: Show or hide the filelist
# Author     : Tom Wilkason
# Date       : 3/9/2001
#------------------------------------------------------------------------------
proc APL::setFileListToShow {what widget} {
   variable paneOrient
   if {$::snackAmpSettings(hideFileList)} {
      pane $what -orient $paneOrient -dynamic 0 \
         -size $::snackAmpSettings(FileListSize)
      $widget configure -relief sunken
      set ::snackAmpSettings(hideFileList) 0
   } else {
      $widget configure -relief flat   ;# button
      set ::snackAmpSettings(hideFileList) 1
      set curSize [pane $what -size]
      if {$curSize != {}} {
         set ::snackAmpSettings(FileListSize) $curSize
         pane forget [winfo parent $what] $what
      }
   }
}
#------------------------------------------------------------------------------
# Function   : RenameFile
# Description: Rename a file directly from the listbox
# Author     : Tom Wilkason
# Date       : 6/19/2002
#------------------------------------------------------------------------------
proc APL::RenameFile {args} {
   set W $::PLB
   if {[catch {aplLB::getActiveFullName} fullName]} {
      saLog "APL::RenameFile $fullName\n$::errorInfo"
      return
   }
   set oldName [file tail $fullName]
   set dirName [file dirname $fullName]
   set newName $oldName
   ##
   # Block on dialog
   #
   set name [catalog::renameInPlaceDialog $W 1 $oldName]
   ##
   # Back from window closure
   #
   if {[string length $name]>0 && $name != $oldName} {
      set newFullName [file join $dirName $name]
      if {[catalog::RenameFile $fullName $newFullName]} {
         tk_messageBox -icon warning -message "Could not rename\n$fullName to \n$newFullName\nCheck console log for errors"
      }
   }
}
#------------------------------------------------------------------------------
# Function   : renameFolder
# Description: Prompt then rename a whole folder
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc APL::renameFolder {folder} {
   set newFullName $folder
   set res 0
   if {[getEntryDialog $newFullName newFullName "Rename Path"]=="ok" && ($newFullName ne $folder)} {
      if {[catch {
         catalog::moveFolder $folder $newFullName
         if {[APLExists]} {
            APL::treeSelect $newFullName
            APL::viewHighlighted $::TREE
         }
         set res 1
      } result] \
      } then {
         tk_messageBox -type ok -icon warning -message "Could not rename folder $newFullName\n$result"
         set res 0
      }
   }
   return $res
}

namespace eval APLcover {
   variable B .aplCoverart
   variable lastSelectedFile ""
}
#------------------------------------------------------------------------------
# Toggle the display of album art
#------------------------------------------------------------------------------
proc APLcover::ToggleArt {args} {
   global TA
   if {[pane info $APL::coverPane] eq ""} {
      pane $APL::coverPane -size $::snackAmpSettings(aplArtSize)
      after idle [list APLcover::packCoverArt $::TA(highlightPath)]
      $TA(aplArtButton) configure -relief sunken
   } else {
      set ::snackAmpSettings(aplArtSize) [lindex [pane $APL::coverPane -size] 0]
      pane forget [winfo parent $APL::coverPane] $APL::coverPane
      $TA(aplArtButton) configure -relief flat
   }
}
#------------------------------------------------------------------------------
# Create a view window for coverart
# TODO: don't need
#------------------------------------------------------------------------------
proc APLcover::coverArtViewer {args} {
   variable B
   global snackAmpSettings
   if {![winfo exists $B]} {
      toplevel $B
      wm title $B "Folder Art"
      wm withdraw $B
      wm protocol $B WM_DELETE_WINDOW APLcover::ToggleArt
      bind $B <Double-ButtonPress-1> {
         APLcover::searchCoverArt
      }

      button $B.drop -text "Query Art" -relief flat -padx 0 -compound left -image saImages.web \
         -command {APLcover::searchCoverArt}
      balloon::button $B.drop "Press to bring up a dialog to query google for album art"
      label $B.ca -image "saImages.saLogoImage" -highlightthickness 0 -border 2 -relief sunken -bg black
      balloon::define $B.ca "Double click to query cover art for this album. Drag desired image from web browser to here to save it."
      pack $B.ca -expand true -fill both
      if {![catch {package require tkdnd} result] } then {
         dnd bindtarget $B.ca "UniformResourceLocator" <Drop> {APLcover::dndDropCoverArt %D}
         dnd bindtarget $B.ca "text/uri-list"          <Drop> {APLcover::dndDropCoverArt %D}
      }
      ontop::pinOnSystemMenu {} $B
      wm geometry $B {}
      after 100 [list myIcon $B]
   }
}
#----------------------------------------------------------------------------
# Search for cover art for the currently playing track
#----------------------------------------------------------------------------
proc APLcover::searchCoverArt {args} {
   set parts [file split $::TA(highlightPath)]
   set albums [db::fieldsInFolder $::TA(highlightPath) "Album"]
   set artists [db::fieldsInFolder $::TA(highlightPath) "Artist"]
   if {[llength $albums]==1} {
      set data(Album) [lindex $albums 0]
   } else {
      set data(Album) [lindex $parts end]
   }
   if {[llength $artists]==1} {
      set data(Artist) [lindex $artists 0]
   } else {
      set data(Artist) [lindex $parts end-1]
   }
   browser::coverart data
}

#------------------------------------------------------------------------------
# Insert the coverart for the currently playing track
# state is normal, iconic, or withdrawn
#------------------------------------------------------------------------------
proc APLcover::packCoverArt {folder args} {
   Trace
   if {[pane info $APL::coverPane] ne ""} {
      # If processing the same folder, then use existing cover image
      foreach {ca valid} [albumart::covername $folder] {break}
      if {$ca ne ""} {
         #
         if {[catch {image create photo "saImages.APLCoverArtS" -file $ca} img] } then {
            saLog $img
            set img "saImages.saLogoImage"
         } else {
            image create photo "saImages.APLCoverArt"
            foreach {height width} [albumart::getCoverDims saImages.APLCoverArtS $APL::coverLabel] {break}
            if {$width>0 && $height>0} {
               tkImageTools::resize "saImages.APLCoverArtS" "saImages.APLCoverArt" $width $height
               set img "saImages.APLCoverArt"
            } else {
               saLog "Source image $ca had height of $height and width of $width"
               set img "saImages.saLogoImage"
            }
         }
      } else {
         set img "saImages.saLogoImage"
      }
      $APL::coverLabel configure -image $img
   }
}
#------------------------------------------------------------------------------
# Insert the coverart for the currently playing track
# state is normal, iconic, or withdrawn
#------------------------------------------------------------------------------
proc APLcover::packFileCoverArt {file args} {
   variable lastSelectedFile
   if {$file eq $lastSelectedFile} {
      return
   } else {
      set lastSelectedFile $file
   }
   package require base64
   if {[pane info $APL::coverPane] ne ""} {
      # If processing the same folder, then use existing cover image
      if {[file exists $file]} {
         set image [id3Tag::getImage $file] ;# get raw jpeg info
         if {$image ne ""} {
            if {[catch {image create photo saImages.APLCoverArtS -data [base64::encode $image]} result] } then {
               saLog $result
            } else {
               image create photo "saImages.APLCoverArt"
               foreach {height width} [albumart::getCoverDims saImages.APLCoverArtS $APL::coverLabel] {break}
               if {$width>0 && $height>0} {
                  tkImageTools::resize "saImages.APLCoverArtS" "saImages.APLCoverArt" $width $height
                  set img "saImages.APLCoverArt"
                  $APL::coverLabel configure -image $img
               } else {
                  APLcover::packCoverArt [file dirname $file]
               }
            }

         } else {
            APLcover::packCoverArt [file dirname $file]
         }
      } else {
         # Try working on the folder version
         APLcover::packCoverArt [file dirname $file]
      }
   }
}

#------------------------------------------------------------------------------
#  Handler for dropping cover art on the album viewer
#------------------------------------------------------------------------------
proc APLcover::dndDropCoverArt {files} {
   Trace
   global TA
   set flag 0
   if {[llength $files]} {
      # Hack, if first file is a graphic, just replace the
      # folder image
      set ext [string tolower [file extension [lindex $files 0]]]
      switch -- $ext {
         .jpg -
         .png -
         .gif -
         .jpeg -
         .bmp {
            if {[macros::copyImageToFolder [lindex $files 0] $::TA(highlightPath)]} {
               APLcover::packCoverArt $::TA(highlightPath)
            }
         }
         default {
            tk_messageBox -type ok -icon warning \
               -message "Don't know how to handle file \n\n[lindex $files 0]"
         }
      }
   }
   return 1
}
package provide autoPlayList 1.0
