#------------------------------------------------------------------------------
# browser 1.0
#
# Functions to web scrape and control web browsers for snackAmp Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
# Function   : browser namespace
# Description: Used to invoke the http browser on the help docs
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/browser.tcl,v 1.3 2007/01/30 01:36:51 wilkason Exp $
}
namespace eval browser {
   variable playControl
   variable AMGSearch
   variable allmusicBase
   variable helpFile "user_manual/index.html"
   variable Jukebox ""
   variable cddbBase
   upvar #0 ::snackAmpSettings(cddbBase) cddbBase
   array set AMGSearch {
      Artist 1:
      Album 2:
      Title 3:
      Genre 4:
   }
   set allmusicBase {http://allmusic.com/cg/amg.dll?p=amg&sql=} ;#1Eva|Cassidy
   #  cddbBase {http://www.Gracenote.com/xm/search?q=} ;#1Eva+Cassidy&f=artist
   # http://images.google.com/images?sourceid=navclient&q=manaics+eat+for+two
   #set coverBase {http://images.google.com/images?sourceid=navclient&q=}
   # restricting to amazon.com
   #variable coverBase {http://images.google.com/images?svnum=100&imgsz=medium|large|xlarge|xxlarge}
   variable coverBase {http://images.google.com/images?svnum=N&imgsz=medium|large|xlarge|xxlarge}
   # filter=0 returns non-relavent images as well
   variable coverBase {http://images.google.com/images?svnum=N&&filter=1}
   #http://images.google.com/images?hl=en&q=snackamp&btnG=Google+Search&ie=UTF-8&oe=UTF-8&sa=N&tab=wi
   ##
   # The two below may change is Google changes the format of the image thumbnails
   # Extract the urls from the google return
   variable googleRegExp {q=tbn:.*?:(.*) width} ;# to strip the url
   # Extract the corresponding image sizes
   variable googleRegSize {&h=([0-9]+)&w=([0-9]+)&sz=}
   # Threshold for verifying images (link this into a snackamp setting)
   variable ratioThreshold
   # Extract the range of results (this could change at any moment!)
   variable fromToReg {Results <b>([0-9]+)</b> - <b>([0-9]+)</b> of.*?<b>([,0-9]+)</b>}
   variable maxFind
   variable maxShow
   upvar #0 ::snackAmpSettings(maxCoverShow) maxShow
   upvar #0 ::snackAmpSettings(maxCoverFind) maxFind
   variable maxCoverWidth
   variable minCoverWidth
   upvar #0 ::snackAmpSettings(maxCoverWidth) maxCoverWidth
   upvar #0 ::snackAmpSettings(minCoverWidth) minCoverWidth

   #variable fromToReg {id=lowerLimit>([0-9]+)</span>.*id=upperLimit>([0-9]+)</span>.*id=maxLimit>(.+?)</span>}

   upvar #0 ::snackAmpSettings(ratioThreshold) ratioThreshold

}
#------------------------------------------------------------------------------
# Function   : browser::help
# Description: Launch help using the browser
# Author     : Tom Wilkason
# Date       : 12/30/2001
#------------------------------------------------------------------------------
proc browser::init {args} {
   ##
   # Now check real folders
   #
   variable helpFile
   if {![info exists helpFile]} {
      foreach {location} {user_manual/index.html} {
         set name [file join $::TA(webRoot) $location]
         if {[file exists $name]} {
            set helpFile $location
            saLog "Found help file $location"
            break
         } else {
            saLog "cant find $name"
         }
      }
   }
}
proc browser::help {{tag {}}} {
   variable helpFile
   if {[checkEnable]==0} {
      return
   }

   if {[info exists helpFile]} {
      if {$tag=={}} {
         urlOpen $helpFile
      } else {
         urlOpen $helpFile#$tag
      }
   } else {
      tk_messageBox -type ok -icon warning \
         -message "No Help File available or web server is disabled"
   }

}
#------------------------------------------------------------------------------
# Function   : browser::findExecutable
# Description: Pulled from the wiki http://mini.net/cgi-bin/wikit/557.html
#------------------------------------------------------------------------------
proc browser::findExecutable {progname varname} {
   set progs [auto_execok $progname]
   if {[llength $progs]} {
       set pattern {([^[:space:]]+)\((.*)\)}
       set isArray [regexp $pattern $varname {} arr index]
       if {$isArray == 0} {
           upvar 1 $varname result
           set result [lindex $progs 0]
       } else {
           upvar $arr array
           set array($index) [lindex $progs 0]
       }
   }
   return [llength $progs]
}
#------------------------------------------------------------------------------
# Return a link to an allmusic search for some info
#------------------------------------------------------------------------------
proc browser::allMusicRef {text {type Artist}} {
   variable AMGSearch
   variable allmusicBase
   set parts [join [split $text] |]
   if {[catch {set url "$allmusicBase$AMGSearch($type)$parts"} result]} {
      return "$text"
   } else {
      return "<a href=\"$url\" title=\"Search AMG for $text\" target='amg'>$text</a>"
   }
}
#------------------------------------------------------------------------------
# Function   : browser::allMusic
# Description: Launch an allmusic query on passed in data
#------------------------------------------------------------------------------
proc browser::allMusic {_search {type Artist}} {
   variable AMGSearch
   variable allmusicBase
   upvar $_search search
   set redir {<meta http-equiv="refresh" content="0; URL=$url">}
   set parts [join [split $search] |]
   # Allmusic is no longer supporting queries, hardcode for lycos for now
#   set url "http://music.lycos.com/artist/discography.asp?QT=A&QW=$parts"
   #http://www.allmusic.com/cg/amg.dll?p=amg&sql=2:Jesus|Of|Cool
   if {[catch {set url "$allmusicBase$AMGSearch($type)$parts"} result]} {
      return
   }

   set redir [subst -nocommand $redir]
   set redirFile [file join $::HOME redirect.html]
   if {[catch {open $redirFile w} result]} {
      saLog "Error:$result"
      return
   } else {
      puts $result $redir
      close $result
   }
   urlOpen $redirFile
}
#------------------------------------------------------------------------------
# Bring up a little dialog to modify the query data for cover art images
#------------------------------------------------------------------------------
proc browser::coverart {_Array} {
   variable Artist
   variable Album
   upvar $_Array Array
   set Artist $Array(Artist)
   set Album $Array(Album)
   set tl .queryGUI
   if {![winfo exists $tl]} {
      toplevel $tl
      wm title $tl "Modify below, then google for images"
      wm protocol $tl WM_DELETE_WINDOW [list wm withdraw $tl]
      button $tl.arLab -relief flat -text "Artist"  -compound left -image saImages.web \
         -command [list browser::allMusic browser::Artist Artist]
      balloon::button $tl.arLab "Press to invoke a browser on the allmusic database for this artist"
      entry $tl.arEnt -textvariable browser::Artist -width 55
      button $tl.alLab -relief flat -text "Album"  -compound left -image saImages.web \
         -command [list browser::allMusic browser::Album Album]
      balloon::button $tl.alLab "Press to invoke a browser on the allmusic database for this album"
      entry $tl.alEnt -textvariable browser::Album -width 55

      grid $tl.arLab $tl.arEnt -sticky w
      grid $tl.alLab $tl.alEnt -sticky w
      grid [horzLine $tl.divide] -pady 3 -columnspan 2 -sticky nsew
      grid [frame $tl.buttons] -columnspan 2 -sticky nsew
      button $tl.buttons.qurBut -text "Query" -command [list browser::queryCoverartGui] \
         -default active -compound left -image saImages.web
      button $tl.buttons.canBut -text "Close" -command [list wm withdraw $tl] \
         -default normal -compound left -image saImages.Close


      checkbutton $tl.buttons.amazon -text "Amazon Only" -variable ::snackAmpSettings(amazonCoverArt)
      checkbutton $tl.buttons.walmart -text "Include Walmart" -variable ::snackAmpSettings(walmartCoverArt)
      checkbutton $tl.buttons.jpeg -text "Jpeg Only" -variable ::snackAmpSettings(jpegCoverArt)


      grid $tl.buttons.jpeg $tl.buttons.walmart $tl.buttons.amazon $tl.buttons.qurBut $tl.buttons.canBut -padx 6
      bind $tl <Escape> [list $tl.buttons.canBut invoke]
      bind $tl <Return> [list $tl.buttons.qurBut invoke]
      after 100 [list myIcon $tl]
   } else {
      wm deiconify $tl
   }
   # Automatically perform the query
   after 200 [list browser::queryCoverartGui]
}

#------------------------------------------------------------------------------
#  Query for cover art data on qoogle
#------------------------------------------------------------------------------

proc browser::queryCoverartGui {args} {
   global snackAmpSettings
   variable Artist
   variable Album
   variable coverBase
   variable googleRegExp
   variable googleRegSize
   variable ratioThreshold
   variable fromToReg
   variable maxFind
   variable maxShow
   variable minCoverWidth
   variable maxCoverWidth
   set HEAD {
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>

<head>
<title>Cover Art Search</title>
   <meta name="AUTHOR"      content="Tom Wilkason">
   <LINK REL=StyleSheet HREF="http://snackamp.sourceforge.net/snackamp.css" TYPE="text/css" MEDIA=screen>
</head>
<body>
   }
   set imageLinks [list]
   set imageLinks [browser::queryGoogleImages $Artist $Album]
   if {$snackAmpSettings(walmartCoverArt)} {
      foreach item [browser::queryWalmartImages $Artist $Album] {
         lappend imageLinks $item
      }
   }
   set limit [llength $imageLinks]
   set imageLinks [lrange [lsort -integer -index 3 -decreasing $imageLinks] 0 [expr {$maxShow-1}]]
   set totalLinks [llength $imageLinks]
   set rejected 0

   # reg exp to extrat image height and width
   # Create the redirection file
   set redirFile [file join $::HOME redirect.html]

   if {[catch {open $redirFile w} result]} {
      saLog "Error:$result"
      return
   } else {
      #"<img src=www.firstuniversal.clara.net/sade_files/sade-best.jpg>"
      if {[llength $imageLinks] > 0} {
         set count 0
         puts $result $HEAD
         puts $result "<table cellspacing='0' cellpadding='5'><tr>"
         puts $result {<td><img alt="Google" src="http://www.google.com/images/logo_sm.gif"></td>}
         puts $result [subst {<td><b>Top $totalLinks Unique Images out of $limit</b><br>}]
         puts $result "<b>Artist:'$Artist' Album:'$Album'</b><br>"
         puts $result "<b>Drag desired image to any SnackAmp cover art folder or display</b></td></tr></table>"
         puts $result "<table cellspacing='0' cellpadding='5'><tr>"
         foreach {image} $imageLinks {
            foreach {link h w pixels} $image {
#debug "$h $w $pixels $link"
               set tail [file tail $link]
               # Filter based on size
               if {$snackAmpSettings(sizeCoverArt) > 0} {
                  set row [subst -nocommand {<img src="$link" width=$snackAmpSettings(sizeCoverArt)>}]
               } else {
                  set row [subst -nocommand {<img src="$link" alt="$link">}]
               }
               if {$count%$snackAmpSettings(rowsCoverArt)==0} {
                  puts $result "</tr><tr>"
               }
               set lnkText "height=${h} x width=${w}"
               puts $result "<td>$row<br>[AREF $link $lnkText $link]</td>"
               incr count
            }
         }
         puts $result "</tr></table>"

      } else {

         # Just put up a dialog

         puts $result $HEAD
         puts $result {<img alt="Google" src="http://www.google.com/images/logo_sm.gif"><br>}
         if {$rejected > 0} {
            puts $result "<H2>Sorry...No images fit query (All $rejected images rejected)</H2>"
         } else {
            puts $result "<H2>Sorry...No images found,  try a different query</H2>"
         }
         puts $result "<H3>Artist:'$Artist'<br>Album:'$Album'</H3>"
      }
      puts $result </body>
      close $result
   }
   # Launch the browser on the new page so all the images are embedded
   urlOpen $redirFile
}
##
# Perform the image query in Google
#
proc browser::queryGoogleImages {Artist Album} {
   global snackAmpSettings
   variable coverBase
   variable googleRegExp
   variable googleRegSize
   variable ratioThreshold
   variable fromToReg
   variable maxFind
   variable maxShow
   variable minCoverWidth
   variable maxCoverWidth
   #set line {<img src=/images?q=tbn:U4jV6k9SWA8J:www.firstuniversal.clara.net/sade_files/sade-best.jpg width="110" height="110"></a></td><img src=/images?q=tbn:U4jV6k9SWA8J:www.firstuniversal.clara.net/sade_files/sade-best.jpg width="110" height="110"></a></td><img src=/images?q=tbn:dL0QYWFyNoQJ:www.artistdirect.com/Images/Sources/AMGCOVERS/music/cover200/dre600/e654/e65476ntdl1.jpg width="99" height="99"></a></td><img src=/images?q=tbn:OHlcbCTkOdkJ:images.amazon.com/images/P/B00007H044.01.LZZZZZZZ.jpg width="125" height="123"></a></td>}
   set url $coverBase
   set Tokens [list]
   foreach {field} {Artist Album} {
      # Improve on the cleaning up, use real http substituion
      set contents [string trim [HttpD::Encode [set $field]]]
      if {[slen $contents] > 0} {
         lappend Tokens $contents
      }
   }
   if {$snackAmpSettings(jpegCoverArt)} {
      append url "&as_filetype=jpg"
   }
   if {$snackAmpSettings(amazonCoverArt)} {
      append url "&as_sitesearch=images.amazon.com"
      append url "&as_sitesearch=amazon.com"
   }
   append url "&as_q="
   append url [join $Tokens +]
   set line ""
   set returned 0
   ##
   # Loop depending on number of images desired
   # Google hard codes 20 per return, but we can specify an offset
   #
   set offset "&start=0"
   set Returned 20
   set Found 0
   set oldFrom -1
   set imageLinks [list]
   set from 0
   set to 0
   set limit 0
   set rejected 0
   #<span id=lowerLimit>41</span></b> - <b><span id=upperLimit>51</span>
   #{Results <b>1</b> - <b>20</b> of about <b>73,000</b> for Results <b>1</b> - <b>20</b> of about <b>73,000</b> for}

   for {set offset 0} {1} {incr offset 20} {
      waitWindow::percent $Found $maxFind "Googling '$Artist' & '$Album' at offset $offset" 1
      if {[waitWindow::ifbreak]} {
         break
      }
      set query "${url}&start=${offset}"
      if {[catch {getUrlBinary $query} guts] } then {
         tk_messageBox -type ok -icon warning -message "Error retrieving $url\n$line\n$::errorInfo"
         return
      } else {
#puts "$guts"
         if {[regexp -nocase -- "did not match any documents" $guts -> xx]} {
            break
         } elseif {![regexp -nocase -- $fromToReg $guts -> from to limit]} {
            tk_messageBox -type ok -icon warning -message "Failed to parse image count returned!\n$query"
            clipboard clear
            clipboard append $query
            return
         }

#puts "$from $to $limit"
         # No more pages to page through
         if {$from == $oldFrom} {
            break
         } else {
            set oldFrom $from
         }

         set matches [regexp -all -inline -- $googleRegExp $guts]
         set sizes [regexp -all -inline -- $googleRegSize $guts]
         if {[llength $matches]>1} {
            ##
            # Filter here
            #
            foreach {match link} $matches {xxx h w} $sizes {
               if {![info exists hash($link)]} {
                  set ratio [expr {double($h)/double($w)}]
                  if {$ratio < $ratioThreshold && $ratio > 1/$ratioThreshold && $w >= $minCoverWidth && $w <= $maxCoverWidth} {
                     lappend imageLinks [list $link $h $w [expr {$h*$w}]]
                     incr Found
                     set hash($link) 1
                  } else {
                     incr rejected
#puts "rejecting $h $w size"
                  }
               }
            }
            ##
            # Look for the Next link at the bottom, if we don't have it there are no
            # more pages to search
            if {![regexp -nocase -- "Next</a>" $guts -> xx]} {
               #saLog "No next page"
               break
            } elseif {$Found >= $maxFind} {
               # If we have our quota, finish
               break
            }
         } else {
            # no matches found
            break
         }
      }
   }
   waitWindow::close
   # Sort largest to smallest image, limit to N max
   set imageLinks [lrange [lsort -integer -index 3 -decreasing $imageLinks] 0 [expr {$maxShow-1}]]
   set totalLinks [llength $imageLinks]
   return $imageLinks
}
#----------------------------------------------------------------------------
# Search walmart for images
#----------------------------------------------------------------------------
proc browser::queryWalmartImages {Artist Album} {
   global snackAmpSettings
   variable coverBase
   variable googleRegExp
   variable googleRegSize
   variable ratioThreshold
   variable fromToReg
   variable maxFind
   variable maxShow
   variable minCoverWidth
   variable maxCoverWidth
   set imageList [list]
   #set line {<img src=/images?q=tbn:U4jV6k9SWA8J:www.firstuniversal.clara.net/sade_files/sade-best.jpg width="110" height="110"></a></td><img src=/images?q=tbn:U4jV6k9SWA8J:www.firstuniversal.clara.net/sade_files/sade-best.jpg width="110" height="110"></a></td><img src=/images?q=tbn:dL0QYWFyNoQJ:www.artistdirect.com/Images/Sources/AMGCOVERS/music/cover200/dre600/e654/e65476ntdl1.jpg width="99" height="99"></a></td><img src=/images?q=tbn:OHlcbCTkOdkJ:images.amazon.com/images/P/B00007H044.01.LZZZZZZZ.jpg width="125" height="123"></a></td>}
   #http://www.walmart.com/search/search-ng.do?search_constraint=4104&search_query=little+feat&Continue.x=31&Continue.y=13&Continue=Find&ic=24_0
   set home "http://www.walmart.com"
   set url "$home/search/search-ng.do?search_constraint=4104&search_query="
   set Tokens [list]
   set linkPathRe {<a href='(/catalog/product\.do\?product_id=[0-9]+)'}
   set imagePathRe {<a href="javascript:photo_opener\('(.+?)&amp} ;#"
   foreach {field} {Artist Album} {
      # Improve on the cleaning up, use real http substituion
      set contents [string trim [HttpD::Encode [set $field]]]
      if {[slen $contents] > 0} {
         lappend Tokens $contents
      }
   }
   append url [join $Tokens +]
   append url "&search_constraint=4104&ic=48_0&ref="
#debug $url
   if {[catch {getUrlBinary $url} guts] } then {
      tk_messageBox -type ok -icon warning -message "Error retrieving\n$url\n$::errorInfo"
      return
   } else {
      # First check to see if a single page returned
      set fullImages [regexp -all -inline -- $imagePathRe $guts]
      if {[llength $fullImages]>0} {
         foreach {full img} $fullImages {
            lappend imageList [list $img 500 500 250000]
            break
         }
      } else {


         set productIDs [regexp -all -inline -- $linkPathRe $guts]

         # todo: If only one album returned, we will jump right to the details on it
         # Skip duplicates
         set uniqueid [list]
         foreach {full id} $productIDs {
            if {[info exists hash($id)]} {
               lappend uniqueid $id
            } else {
               set hash($id) 1
            }
         }

         set lpid [llength $uniqueid]
         set i 1
         foreach id $uniqueid {
            waitWindow::percent $i $lpid "Checking Walmart for '$Artist' & '$Album'" 1
            if {[waitWindow::ifbreak]} {
               break
            }
            if {$i >= $maxShow} {
               break
            }
            set url "$home$id"
            # Links to each product ID, grab each on and the large image
            update
#debug " $url"
            if {[catch {getUrlBinary $url} guts] } then {
               tk_messageBox -type ok -icon warning -message "Error retrieving\n$url\n$::errorInfo"
               return
            } else {
               # Should only be one image
               #<a href="javascript:photo_opener('http://i.walmart.com/i/p/00/77/79/66/15/0077796615602_500X500.jpg&amp;product_id=848554',
               set fullImages [regexp -all -inline -- $imagePathRe $guts]
               if {[llength $fullImages]>0} {
                  foreach {full img} $fullImages {
                     lappend imageList [list $img 500 500 250000]
#debug "   $img"
                     break
                  }
               }
            }
            incr i
         }
      }
      waitWindow::close
   }
   set imageList [lsort -index 0 -unique $imageList]
   return $imageList
}

#------------------------------------------------------------------------------
# Function   : browser::allMusic
# Description: Launch an cddb query on passed in data
#------------------------------------------------------------------------------
proc browser::cddb {_Array} {
   package require HttpD
   upvar $_Array Array
   variable cddbBase
   set redir {<meta http-equiv="refresh" content="0; URL=$url">}
   set url $cddbBase
   append url "freedb_search.php?words="
   set Tokens [list]
   set Fields [list]
   foreach {field} [array names Array] {
      if {$field == "Artist" || $field == "Album"} {
         # Improve on the cleaning up, use real http substituion
         set contents [string trim [HttpD::Encode $Array($field)]]
         # Title in freedb means Album name
         if {[slen $contents] > 0} {
            #lappend Fields "fields=[string tolower $field]"
            #eval lappend Tokens $contents
            lappend Tokens $contents
         }
      }
   }
   append url [join $Tokens +]
   #append url "&allfields=NO&allcats=YES&grouping=none&"
   append url "&allfields=NO&allcats=YES&fields=title&grouping=none&x=0&y=0"
   #append url [join $Fields &]
   # Create the redirection file
   set redir [subst -nocommand $redir]
   set redirFile [file join $::HOME redirect.html]
   if {[catch {open $redirFile w} result]} {
      saLog "Error:$result"
      return
   } else {
      puts $result $redir
      close $result
   }
   urlOpen $redirFile
}
##
# Search freedb, but just return the html contents
#
# TODO: Can this be replaced by proper database search?
proc browser::cddbHtml {_Array} {
   package require HttpD
   upvar $_Array Array
   variable cddbBase
   set url $cddbBase
   append url "freedb_search.php?words="
   array set lu {Artist artist Album title}
   set Tokens [list]
   set Fields [list]
   set tosearch ""
   foreach {field} [array names Array] {
      if {$field == "Artist" || $field == "Album"} {
         # Improve on the cleaning up, use real http substituion
         set contents [string trim [HttpD::Encode $Array($field)]]
         # Title in freedb means Album name
         if {[slen $contents] > 0} {
            lappend Tokens $contents
            append tosearch "&fields\[\]=$lu($field)"
         }
      }
   }
   append url [join $Tokens +]
   # Note, &page=2 append will grab second page of results (and so on)
   append url "&allfields=NO$tosearch&allcats=YES&grouping=none"
   return [macros::getUrlBinary $url]
}

proc browser::checkEnable {args} {
   if {$::snackAmpSettings(EnableWebServer)==0} {
      set ans [tk_messageBox -icon question -type yesno -title "Warning" -message "The internal web server is disabled\nwould you like to enable it?"]
      if {$ans=="yes"} {
         set ::snackAmpSettings(EnableWebServer) 1
         package require HttpD
         HttpD::startup
      } else {
         return 0
      }
   }
   return 1
}
#------------------------------------------------------------------------------
# Function   : browser::webHome
# Description: Open up the home page, check to enable browser first
#------------------------------------------------------------------------------
proc browser::webHome {args} {
   if {[checkEnable]==0} {
      return
   }
   urlOpen $::snackAmpSettings(internalWebBase)/
}
#------------------------------------------------------------------------------
# Function   : browser::urlOpen
# Description: Pulled from the wiki http://mini.net/cgi-bin/wikit/557.html
#------------------------------------------------------------------------------
proc browser::urlOpen {url {check 0}} {
   global env tcl_platform
   if {$check && [checkEnable]==0} {
      return
   }
   switch -- $tcl_platform(platform) {
      "unix" {
         if {[info exists env(NETSCAPE)]} {
            set envNetscape $env(NETSCAPE)
         } else {
            set envNetscape "netscape"
         }
         expr {
            [info exists env(BROWSER)] ||
            [findExecutable firefox        env(BROWSER)] ||
            [findExecutable opera          env(BROWSER)] ||
            [findExecutable mozilla        env(BROWSER)] ||
            [findExecutable $envNetscape   env(BROWSER)] ||
            [findExecutable iexplorer      env(BROWSER)] ||
            [findExecutable netscape       env(BROWSER)] ||
            [findExecutable lynx           env(BROWSER)]
         }
         # Support files from the ~ folder
         if {[file exists $url]} {
            set url [file nativename $url]
         }
         # lynx can also output formatted text to a variable
         # with the -dump option, as a last resort:
         # set formatted_text [ exec lynx -dump $url ] - PSE
         if {[catch {exec $env(BROWSER) $url > /dev/null < /dev/null &} emsg]} {
            if {[catch {exec $env(BROWSER) -remote openURL\(file:$url\) > /dev/null < /dev/null  &} emsg]} {
               # perhaps browser doesn't understand -remote flag
               if {[catch {exec $env(BROWSER) $url > /dev/null < /dev/null &} emsg]} {
                  saLog "Error displaying $url in browser\n$emsg"
                  tk_messageBox -type ok -icon warning \
                     -message "Error displaying $url in browser\n$emsg\n\nI can't find a web browser on your system, please specify one\nin under Preferences->Web Server->Help Browser."
               }
            }
         }
      }
      "windows" {
         # Always try to use the short name as a parameter
         if {[catch {file attributes $url -shortname} surl]} then {
            set surl $url
         }
         if {$tcl_platform(os) == "Windows NT"} {
            set rc [catch {exec $env(COMSPEC) /c start $surl &} emsg]
         } else {
            # Windows 95/98
            set rc [catch {exec start $surl &} emsg]
         }
         if {$rc} {
            saLog "Error displaying $url in browser\n$emsg"
         }
      }
      "macintosh" {
         if {0 == [info exists env(BROWSER)]} {
            set env(BROWSER) "Browse the Internet"
         }
         if {[catch {
               AppleScript execute\
                     "tell application \"$env(BROWSER)\"
               open url \"$url\"
               end tell
               "} emsg]
         } then {
            saLog "Error displaying $url in browser\n$emsg"
         }
      }
   } ;## end of switch
}
package provide browser 1.0
