#------------------------------------------------------------------------------
# snackAmp - Catalog package
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/catalog.tcl,v 1.114 2007/10/01 01:33:25 wilkason Exp $
}
namespace eval catalog {
   variable lastList [list]
   variable indexing 0
   variable flag 0
   variable indexUnits
   variable fcount 0
   variable break 0
   variable added 0
   variable mpdFormat
   variable Continue
   array set indexUnits {
      Minutes 60
      Hours 3600
      Days 86400
      Weeks 604800
   }
   variable defNames
   array set defNames {
      Artist "Unkown"
      Album  "Unkown"
      Title  "Unkown"
   }
   variable S
   upvar ::snackAmpSettings S
   variable alwaysCatalog
   upvar ::snackAmpSettings(alwaysCatalog) alwaysCatalog
   variable fastscan
   upvar ::snackAmpSettings(fastscan) fastscan
   variable extensions
   upvar ::TA(extensions) extensions
   set fastscan 1
}
#------------------------------------------------------------------------------
# Function   : catalog::saveAs
# Description: Save the current file as some format
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::saveAs {args} {
   set file [snack::getSaveFile]
   if {[string length $file]==0} {
      return
   }
   # Stop (should work even if playing)
   # Clone current sound
   set snd [snack::sound -file [soundControl::fileName]]
   # Get extension
   set ext [string tolower [string range [file extension $file] 1 end]]
   ##
   # Status Window
   #
   # Save the file
   if {[catch {
      switch -- $ext {
         "ogg" {
            #W: Prompt for -nominalbitrate, -maxbitrate, -minbitrate or use existing
            switch -- [$snd config -fileformat] {
               MP3 {set rate [$snd config -bitrate]}
               OGG {set rate [$snd config -nominalbitrate]}
               WAV {set rate [expr {[$snd config -rate]}]}
               default {set rate 0}
            }
            if {$rate>0} {
               waitWindow::show "Saving $file ... 0% complete"
               $snd write $file -fileformat $ext -nominalbitrate $rate -progress catalog::waiter
            } else {
               waitWindow::show "Saving ... 0% complete"
               $snd write $file -fileformat $ext -progress catalog::waiter
            }
         }
         default {
            $snd write $file -fileformat $ext -progress catalog::waiter
         }
      }
   } result]} {
      tk_messageBox -icon warning -type ok -title "Save Error" -message $result
   }
   # Make sure it is in the master list
   db::cacheFileName $file
   $snd destroy
   waitWindow::close
}
#------------------------------------------------------------------------------
# Function   : catalog::waiter
# Description: Update global used for wait box
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::waiter {cmd percent} {
   set percent [expr {round($percent*1000)/10.0}]
   waitWindow::show  "Saving ... $percent% complete"
}

#------------------------------------------------------------------------------
# Return true if this files has an extension that we car about
#------------------------------------------------------------------------------
proc catalog::isTrackFile {file} {
   variable extensions
   if {[lsearch -exact -sorted $extensions [string toupper [file extension $file]]]>=0} {
      return 1
   } else {
      return 0
   }
}
#------------------------------------------------------------------------------
# Function   : catalog::MonitorScanFolders
# Description: Monitor the base folders for incoming data, useful
#              for napaster adding files.
# Author     : Tom Wilkason
# Date       : 12/30/2000
#------------------------------------------------------------------------------
proc catalog::MonitorScanFolders {args} {
   global snackAmpSettings TA
   if {$snackAmpSettings(MonitorScanFolders)} {
      # TODO: update folders in APL also
      foreach {Folder} $snackAmpSettings(IncomingDirs) {
         foreach {file} [eval glob -nocomplain -directory {$Folder} [globAllPattern $snackAmpSettings(ScanFolderDepth)]] {
            if {[catalog::isTrackFile $file]} {
               if {[db::whereInDb $file]==-1} {
                  catalog::addNewEntry $file
               }
               Update
            }
         }
      }
   }
   after cancel catalog::MonitorScanFolders
   after 60000 catalog::MonitorScanFolders
}

#------------------------------------------------------------------------------
# Function   : catalog::fixMediaPath
# Description: Make sure the directories in the media path don't have trailing slashes
#              needed by tree matching functions
# Author     : Tom Wilkason
# Date       : 1/11/2001
# TBD: we can revalidate the path here
#------------------------------------------------------------------------------
proc catalog::fixMediaPath {args} {
   cleanPathList ::snackAmpSettings(IncomingDirs) 1
   cleanPathList ::snackAmpSettings(mediaDirs) 0 ;#1
   cleanPathList ::snackAmpSettings(mp3Player) 0 ;#1
}
#----------------------------------------------------------------------------
# Rescan all the media dirs after asking for permission.
#----------------------------------------------------------------------------
proc catalog::IndexAskFirst {args} {
   if {[tk_messageBox -icon question -type yesno -title "Please Respond" \
      -message "Are you sure you want to index the following folders?\n$::snackAmpSettings(mediaDirs)"]=="yes"} {
      catalog::Index
   }
}
#------------------------------------------------------------------------------
# Function   : catalog::Index
# Description: Reread the directories for files, include subdirectories
#              For each folder, add a menu entry
# Author     : Tom Wilkason
# Date       : 12/26/2000
#------------------------------------------------------------------------------
proc catalog::Index {{Folder ""}} {
   global TA snackAmpSettings
   variable fcount 0
   variable break
   set break 0
   set added 0
   Trace
   Wait 1
   set here [pwd]
   # If a folder is specifed, catalog all media dirs, otherwise just
   # catalog the specified folder retaining information on other folders.
   if {$Folder eq ""} {
      waitWindow::show "Indexing tracks\nPlease wait...\n" 1
      set saveNames [aplFolder::savePattern *]  ;# want folder also
      aplFolder::stripBranch *
      # Get a list of all the files intoList
      # Only scan the known mediaDirs
      foreach dir $TA(mediaDirs) {
         # Flag all as unmounted till recached
         db::mountFiles [file join $dir *] 0
#         indexRecurse $dir
         indexFolderTree $dir
         if {$break} {
            waitWindow::bump
            waitWindow::show "Restoring Tracks on\n$Folder\n"
            db::mountFiles [file join $dir *] 1
            aplFolder::restoreNames $saveNames 1
            break
         } else {
            # Any tracks not flagged as mounted here no longer exist, remove them
            db::removeUnmountedFiles [file join $dir *] 1
         }
      }
      waitWindow::show "Verifying/Reading Catalogs"
      ##
      # Update the APL settings
      # (should really restore the old memory settings)
      # Restore prior settings, note it indexing
      # is done prior to first time APL use, then
      # all apl values are 0. Need to fiX
      aplFolder::restoreNames $saveNames
      # Rebuild everything since we indexed everything
      waitWindow::show "Cleaning Up ... Almost Done"
      if {[APLExists]} {
         APL::reBuildTree
      }
   } else {
      # Set master list to all but those files in this branch
      waitWindow::show "Indexing tracks on\n$Folder\n"
      # Flag these files as unmounted
      # Then clear folders
      set saveNames [aplFolder::savePattern $Folder*]  ;# want folder also
      aplFolder::stripBranch [file join $Folder *]
      waitWindow::bump
      # Flag tracks in db as unmounted
      db::mountFiles [file join $Folder *] 0
      # Rescan selected folder (restore any apl folder settings)
      waitWindow::bump
#      indexRecurse $Folder
      indexFolderTree $Folder

      if {$break} {
         waitWindow::bump
         waitWindow::show "Restoring Tracks on\n$Folder\n"
         db::mountFiles [file join $Folder *] 1
         aplFolder::restoreNames $saveNames 1
      } else {
         waitWindow::bump
         # Any tracks not flagged as mounted here no longer exist, remove them
         db::removeUnmountedFiles [file join $Folder *] 1
         # Restore prior settings to folders that exist
         aplFolder::restoreNames $saveNames
      }
      #BUG: stale (i.e. no longer existing) teFolder settings can still hang around

      ## Add (or ignore) new items for each entry
      if {[APLExists]} {
         # remove branch
         APL::delItem $Folder
         # add it and the new children back
         if {[file exists $Folder]} {
            APL::newItem $Folder
            foreach folder [aplFolder::foldersMatchingPattern [file join $Folder *] 1] {
               APL::newItem $folder
            }
         }
         # TODO: Smart detect if any new or modified tracks happend before refreshing list
         aplLB::UpdateList $Folder [winfo ismapped $::PLB]
      }
   }
   ##
   # Build the folder list for the menu
   #
   waitWindow::close
   Wait 0
   # Save the files
   after 500 db::commit
}
#------------------------------------------------------------------------------
# Escape characters in file/folder names that cause problems
# with globbing and regexp.
#------------------------------------------------------------------------------
proc catalog::globEscape {dir} {
   return [string map {\\ \\\\ ( \\( ) \\) \{ \\{ \} \\} \] \\] \[ \\[} $dir]
}
#------------------------------------------------------------------------------
# Recursively scan a directory and build a list of all files
# No Longer Used, replaced by indexFolderTree
#------------------------------------------------------------------------------
proc catalog::indexRecurse {dir} {
   variable fcount
   variable fastscan
   variable break
   variable added
   variable alwaysCatalog
   global TA LB snackAmpSettings
   set dirList [list]
   set hasSome 0
   # as being removed after the index.
   if {$break} {
      return 1
   }
   # Show banner every once in a while
   if {$alwaysCatalog} {
      waitWindow::show "Indexing/Cataloging tracks on\n$dir\n[commify $fcount] Indexed..." 1
   } else {
      waitWindow::show "Indexing tracks on\n$dir\n[commify $fcount] Indexed..." 1
   }
   if {[waitWindow::ifbreak]} {
      set break 1
      return
   }
   # Get subfolders
   foreach file [glob -type d -nocomplain -directory $dir *] {
      lappend dirList $file
      incr hasSome
   }
   # Get regular files but only process track files
   waitWindow::bump
   foreach file [glob -type f -nocomplain -directory $dir *] {
      if {[isTrackFile $file]} {
         # Cache if needed
         incr hasSome
         if {$alwaysCatalog} {
            unset -nocomplain Data
            db::cacheFile $file Data 2
         } else {
            db::cacheFileName $file 1
         }
      }
   }
   # Only set the folder (as existing) for the tree if
   # it was non-empty
   if {$hasSome} {
      aplFolder::Set $dir 0      ;# Add this dir, if new will be turned of
      incr fcount $hasSome
   }
   # process each directory recursively
   foreach {folder} $dirList {
      indexRecurse $folder
   }
   return 0
}
#------------------------------------------------------------------------------
# Recursively scan a directory and build a list of all files
#------------------------------------------------------------------------------
proc catalog::indexFolderTree {dir} {
   Trace
   variable fcount
   variable fastscan
   variable break
   variable added
   variable alwaysCatalog
   variable dirList
   global TA LB snackAmpSettings
   set dirList [list]
   set masterDirList [list]
   set hasSome 0
   # as being removed after the index.
   if {$break} {
      return 1
   }
   # Show banner every once in a while
   # Build list of folders
   buildFolderList $dir masterDirList
   # Index each of the folders
   indexFolders [list $dir]
   indexFolders $masterDirList
   return 0
}
#------------------------------------------------------------------------------
# Recursively scan a directory and build a list of all files
#------------------------------------------------------------------------------
proc catalog::buildFolderList {dir {_masterDirList}} {
   Trace
   variable fcount
   variable fastscan
   variable break
   variable added
   variable alwaysCatalog
   variable dirList
   upvar $_masterDirList masterDirList
   global TA LB snackAmpSettings
   set dirList [list]
   set hasSome 0
   # as being removed after the index.
   if {$break} {
      return 1
   }
   # Get subfolders
   set children [glob -type d -nocomplain -directory $dir *]
   if {[llength $children]} {
      # Show banner every once in a while
      waitWindow::show "Scanning $dir\n[commify [llength $masterDirList]] folders total.." 1
      if {[waitWindow::ifbreak]} {
         set break 1
         return
      }
   }
   foreach file $children {
      lappend dirList $file
      lappend masterDirList $file
   }
   # process each directory recursively
   foreach {folder} $dirList {
      buildFolderList $folder masterDirList
   }
   return 0
}

#----------------------------------------------------------------------------
# Index all the folders and update the catalog accordingly
# pass in a list of folders to process
#----------------------------------------------------------------------------
proc catalog::indexFolders {folderList} {
   variable break
   variable added
   variable alwaysCatalog
   set fcount 0
   global TA LB snackAmpSettings
   set hasSome 0
   # as being removed after the index.
   if {$break} {
      return 1
   }
   # Show banner every once in a while
   if {$alwaysCatalog} {
      set note "Indexing/Cataloging"
   } else {
      set note "Indexing"
   }
   # Get regular files but only process track files
   set i 0
   set dirLength [llength $folderList]
   foreach {dir} [lsort $folderList] {
      waitWindow::percent [incr i] $dirLength "$note $dir\n[commify $fcount] tracks indexed..." 1
      if {[waitWindow::ifbreak]} {
         set break 1
         return
      }
      aplFolder::Set $dir 0      ;# Add this dir, if new will be turned of
      foreach file [glob -type f -nocomplain -directory $dir *] {
         if {[isTrackFile $file]} {
            incr fcount
            # Cache if needed
            if {$alwaysCatalog} {
               unset -nocomplain Data
               db::cacheFile $file Data 2
            } else {
               db::cacheFileName $file 1
            }
         }
      }
   }
   return 0
}


#----------------------------------------------------------------------------
# Background index a folder
#----------------------------------------------------------------------------
proc catalog::bgIndex {args} {
   global TA
   variable indexing
   variable indexUnits
   variable S
   if {$indexing} {
      saLog "Background index in progress"
   } else {
      saLog "Running a background index"
      set indexing 1
      foreach dir [catalog::fixedFolders] {
         set S(indexing) "Cleaning $dir"
         aplFolder::removeUnusedNodes [file join $dir *]
         bgIndexRecurse $dir
         db::removeUnmountedFiles [file join $dir *] 0
         # TODO: need to add any apl nodes that are new here
      }
      set indexing 0
      set S(indexing) "Idle"
      set S(lastIndex)              [clock seconds]
      set S(lastIndexFormatted)     [clock format [clock seconds]]
      set S(nextIndex)              [expr {[clock seconds]+$S(indexHours)*$indexUnits($S(indexUnits))}]
      set S(nextIndexFormatted)     [clock format $S(nextIndex)]
   }
}

#------------------------------------------------------------------------------
# Routinely run to see if a background index should be run
# TODO: Need to remove stale folders from tefolders
#------------------------------------------------------------------------------
proc catalog::bgIndexCheck {args} {
   variable indexing
   variable S
   after cancel catalog::bgIndexCheck
   set now [clock seconds]
   if {[string equal $S(indexStart) "0"]} {
      set secStart $now
   } elseif {[catch {clock scan $S(indexStart)} secStart] } then {
      saLog "Invalid time format: $snackAmpSettings(indexStart) for Index Start Time"
      set secStart $now
   }
   if {[string equal $S(indexStop) "0"]} {
      set secStop $now
   } elseif {[catch {clock scan $S(indexStop)} secStop] } then {
      saLog "Invalid time format: $snackAmpSettings(indexStop) for Index Stop Time"
      set secStop $now
   }
   if {$S(indexBackground) && $S(indexHours)>0 && $indexing==0} {
      # valid to check to see if we can run
      if {$now >= $S(nextIndex) && ($now >= $secStart) && ($now <= $secStop)} {
         after 0 catalog::bgIndex
      }
   }
   # Check again in five minutes
   after 300000 catalog::bgIndexCheck
}
#----------------------------------------------------------------------------
# Background index function
#----------------------------------------------------------------------------
proc catalog::bgIndexRecurse {dir} {
   global TA snackAmpSettings
   variable flag
   if {$snackAmpSettings(indexBackground)==0} {
      update
      return
   }
   set bgCatalog $snackAmpSettings(bgCatalog)
   set dirList [list]
   set hasSome 0
   # Flag folder as unmounted, recaching will remount files
   set snackAmpSettings(indexing) $dir
   db::mountFolder $dir 0
   # Get folders (for later)
   foreach file [glob -type d -nocomplain -directory $dir *] {
      update
      lappend dirList $file
      incr hasSome
   }
   # Get files
   foreach file [glob -type f -nocomplain -directory $dir *] {
      if {[catalog::isTrackFile $file]} {
         if {$bgCatalog} {
            unset -nocomplain Data
            db::cacheFile $file Data 2
            incr hasSome
         } else {
            if {[db::cacheFileName $file 1] == -1} {
               saLog "Cannot Catalog $file"
            } else {
               incr hasSome
            }
         }
      }
   }
   # Only set the folder (as existing) for the tree if
   # it was non-empty
   if {$hasSome} {
      aplFolder::SetIfBlank $dir 0      ;# Add this dir, if new will be turned of
   }
   # process each directory recursively
   foreach {folder} $dirList {
      # delay a bit in the background
      after 20 set catalog::flag 1
      vwait catalog::flag
      bgIndexRecurse $folder
   }
}
#----------------------------------------------------------------------------
# Check the media folders and rebuild tree if one as added or removed.
#----------------------------------------------------------------------------
proc catalog::rebuildIfNewMedia {args} {
   global TA
   set oldDirs $TA(mediaDirs)
   catalog::validateMediaPaths
   if {$oldDirs ne $TA(mediaDirs)} {
      if {[APLExists]} {
         APL::reBuildTree
      }
      tk_messageBox -type ok -icon info -title "Media Changed" \
         -message "Your available media folders are:\n[join $TA(mediaDirs) \n]"
   }
}
#----------------------------------------------------------------------------
# Verify all the mediadirs and set the mounted flag appropriatly
#----------------------------------------------------------------------------
proc catalog::validateMediaPaths {args} {
   Trace
   global TA snackAmpSettings
   set TA(mediaDirs) [list]
   set TA(mediaDirPair) [list]
   ##
   # invalidate everything
   db::mountFiles * 0
   # Check attached media
   foreach {Path} $snackAmpSettings(mp3Player) {
      # Try a force on the path
      if {[file isdirectory $Path]} {
         if {[lsearch -exact $TA(mediaDirs) $Path]==-1} {
            set patt [file join $Path *]
            lappend TA(mediaDirs) $Path
            lappend TA(mediaDirPair) $Path "Attached Folders" 2
            db::mountFiles $patt 1
         }
      } else {
         saLog "Attached Media folder '$Path' was not found"
      }
   }
   ##
   # Check hard folders
   foreach {Path} $snackAmpSettings(mediaDirs) {
      file exists $Path
      if {[file isdirectory $Path]} {
         if {[lsearch -exact $TA(mediaDirs) $Path]==-1} {
            set patt [file join $Path *]
            lappend TA(mediaDirs) $Path
            lappend TA(mediaDirPair) $Path "Fixed Folders" 0
            db::mountFiles $patt 1
         }
      } else {
         saLog "Fixed Media folder '$Path' was not found (is it mounted?)"
      }
   }
   return "" ;# always succeed
}
#------------------------------------------------------------------------------
# Function   : catalog::manageMediaFolders
# Description: Bring up dialog then update folder caches if needed
# Author     : Tom Wilkason
# Date       : 6/5/2002
#------------------------------------------------------------------------------
proc catalog::manageMediaFolders {args} {
   package require saSettings
   global snackAmpSettings
   set oldMedia $::TA(mediaDirs)
   folderManager::launch mediaDirs
   # Blocked here until GUI closes
   catalog::fixMediaPath
   catalog::validateMediaPaths
   if {[APLExists] && ($oldMedia ne $::TA(mediaDirs))} {
      APL::reBuildTree
   }
}
#------------------------------------------------------------------------------
# Function   : catalog::managePlayerFolders
# Description: Bring up dialog then update mp3 player caches if needed
# Author     : Tom Wilkason
# Date       : 6/5/2002
#------------------------------------------------------------------------------
proc catalog::managePlayerFolders {args} {
   package require saSettings
   global snackAmpSettings
   set oldMedia $::TA(mediaDirs)
   folderManager::launch mp3Player
   # Blocked here until GUI closes
   catalog::fixMediaPath
   catalog::validateMediaPaths
   if {[APLExists] && ($oldMedia ne $::TA(mediaDirs))} {
      APL::reBuildTree
   }
}
#------------------------------------------------------------------------------
# Function   : playlistStartup
# Description: Populate the playlist with the appropriate files
# Author     : Tom Wilkason
# Date       : 6/5/2002
#------------------------------------------------------------------------------
proc catalog::playlistStartup {args} {
   Trace
   global snackAmpSettings TA
   ##
   # Rescan if set true or the cache read failed
   #
   if {$snackAmpSettings(AutoRescan) || $TA(rescan)} {
      catalog::Index
   }
   ;# Validate paths
   catalog::validateMediaPaths
   # Read cache files, now restore playlist if required
   # otherwise just toss up music based on default AutoPlaylist file
   if {$snackAmpSettings(restorePlayList)} {
      waitWindow::show "Restoring Playlist"
      if {[catch {catalog::ReadPlaylistCache} result]} {
         waitWindow::show "Reading [file tail $snackAmpSettings(lastFolderFile)]"
         readAPLFile $snackAmpSettings(lastFolderFile) 1
         saLog "No play cache file found\n$result"
      }
   } else {
      waitWindow::show "Reading [file tail $snackAmpSettings(lastFolderFile)]"
      readAPLFile $snackAmpSettings(lastFolderFile) 1
   }
}
#------------------------------------------------------------------------------
# Add a new entry to the database and create nodes for it as required
#------------------------------------------------------------------------------
proc catalog::addNewEntry {newName {cloneName ""}} {
   Trace
   # database if it is a file
   if {[file isfile $newName]} {
      set dir [file dirname $newName]
      if {$cloneName ne ""} {
         unset -nocomplain Data
         if {[db::lookup $cloneName Data 1]>0} {
            set Data(File) $newName
            set Data(Folder) $dir
         }
         set Data(Depth) [llength [file split $newName]]
         db::cacheFile $newName Data 0
      } else {
         db::cacheFile $newName Data 1
      }
   } else {
      set dir $newName
   }
   # tefolders & node
   if {[APLExists]} {
      APL::addToRoot $dir 0
   } else {
      aplFolder::setToRoot $dir 0
   }
}
##
# Properties Dialog
#
proc catalog::properties {tracks} {
   set long [list]
   set short [list]
   foreach {track} $tracks {
      if {[file exists $track]} {
         lappend short [file tail $track]
         lappend long $track
      }
   }
   if {[llength $long] <= 0} {
      return
   }
   set i 0
   while {[info command .prop$i] != {}} {incr i}
   set top .prop$i
   toplevel $top
   wm title $top "Track Properties"
   if {[llength $long]==1} {
      set hull [labeledFrame $top.data -text [lindex $short 0]]
      set text [trackTextData [lindex $long 0] Details]
      label $hull.info -text $text -justify left -anchor nw
      pack $hull.info -expand true -fill both
      pack $top.data -expand true -fill both
   } else {
      set tof [tabbedOptionFrame::Create $top.tf -Width 40 -options $short -title "Track Properties"]
      wm protocol $top WM_DELETE_WINDOW "tabbedOptionFrame::close $top.tf;destroy $top"
      pack $tof -expand true -fill both
      set i 0
      foreach {track} $long {
         set text [trackTextData $track Details]
         set hull [tabbedOptionFrame::getHull $tof $i]
         label $hull.info -text $text -justify left -anchor nw
         pack $hull.info -expand true -fill both
         pack $hull -expand true -fill both
         incr i
      }
   }
   update idletasks
   myIcon $top
}
#------------------------------------------------------------------------------
# Function   : catalog::trackTextData
# Description: Return the details of a song in both an array and a text
#              string suitable for balloon help.
# Author     : Tom Wilkason
# Date       : 11/20/2001
#------------------------------------------------------------------------------
proc catalog::trackTextData {fullName _Details} {
   upvar $_Details Details
   set Details(ext) ""
   ##
   # Note: Grab this from a database instead of here if it exists
   #
   set hasTag [db::lookup $fullName Details 2]
   if {[catch {
      set Details(duration) [HMSfrac $Details(Duration) 1]

      set Details(ext) [string trimleft [string toupper [file extension $fullName]] .]
      if {$::snackAmpSettings(autoLevel) && [string is double $Details(sampRate)] && $Details(sampRate)>0 && $Details(Stop)} {
         append endData "Start\t: [format %.3f [expr {$Details(Start)/(1000*$Details(sampRate))}]] sec"
         append endData "\nStop\t: [format %.3f [expr { $Details(Duration)- ($Details(Stop)/(1000*$Details(sampRate))) }]] sec before end\n"
      }
      append endData "Rating\t: "  [mrating $Details(Rating)]
      append endData "\nVotes\t: $Details(Votes)"
      if {$Details(Played) > 0} {
         append endData "\nPlayed\t: "  $Details(Played) " Times"
      } else {
         append endData "\nPlayed\t: Not yet Played"
      }
      if {$Details(LastPlay) > 0} {
         append endData ", Last Played: "  [timeFormat $Details(LastPlay)]
      }
   } result]} {
      return "No Data for $fullName available"
   }
   switch -- $Details(ext) {
      MP3 {append Details(Rate) " kbps"}
      OGG {append Details(Rate) " kbps"}
      WAV {append Details(Rate) " kHz"}
      default {set Details(Rate) "Unknown"}
   }
   append Details(sampRate) " Khz"
   ##
   # File info
   #
   if {[catch {file stat $fullName Stat} result]} {
      return Unknown
   }
   set Details(fSize) "[commify $Stat(size)] Bytes"
   set Details(fDate) [timeFormat $Stat(mtime)]
   set Details(Cataloged) [timeFormat $Details(cataloged)]
   ##
   # Balloon help
   #
   set Info         "File\t: [file tail $fullName]\nFolder\t: [file dirname $fullName]\n"
   if {$hasTag} {
      append Info [id3Tag::formatID3Data Details]
      set Details(ID3) $Info
   } else {
      set Details(ID3) {}
   }
   id3Tag::appendIfNotBlank Info "Size\t:"     $Details(fSize)
   id3Tag::appendIfNotBlank Info "Date\t:"     $Details(fDate)
   id3Tag::appendIfNotBlank Info "Indexed\t:"  $Details(Cataloged)
   id3Tag::appendIfNotBlank Info "Duration\t:" $Details(duration)
   id3Tag::appendIfNotBlank Info "Bitrate\t:"  $Details(Rate)
   id3Tag::appendIfNotBlank Info "Freq\t:"     $Details(sampRate)
   if {$::snackAmpSettings(autoLevel)} {
      id3Tag::appendIfNotBlank Info "Gain\t:"  $Details(Gain)
      id3Tag::appendIfNotBlank Info "Alb Gain\t:"  $Details(AlbumGain)
   } else {
      set Details(levelFactor) "Not Set"
   }
   # Only reads the database
#    if {[info exists Data(track_gain)]} {
#       id3Tag::appendIfNotBlank Info "Track Gain\t:"     $Details(track_gain)
#    }
#    if {[info exists Data(album_gain)]} {
#       id3Tag::appendIfNotBlank Info "Album Gain\t:"     $Details(album_gain)
#    }
   append Info $endData
   return [string trim $Info]
}
#------------------------------------------------------------------------------
# Function   : trackFullData
# Description: Return the details of a song in an array for indexing
#              music files.
# Author     : Tom Wilkason
# Date       : 11/20/2001
#------------------------------------------------------------------------------
proc catalog::trackFullData {fullName _Details} {
   upvar $_Details Details
   ##
   # Retrieve the track sound info
   #
   set Details(size) [file size $fullName]
   if {[catalog::isTrackFile $fullName]} {
      soundControl::trackInfo Details $fullName
   }
   ##
   # Get additional file and ID3 info
   #
   set Details(File)   [file tail $fullName]
   ##
   # ID3 Data, if no tag then set Title = File root name
   #
   set result [id3Tag::id3Label $fullName Details $::snackAmpSettings(ID3ReadOrder)]
   if {$result==0} {
      # Return blanks for non-existent fields
      foreach {field} $id3Tag::v1Parts {
         set Details($field) ""
      }
      set Details(Title) [file rootname $Details(File)]
      set Details(Mounted) 1
      set Details(Tag) "--"
      # Clear remaining fields
   }
   ##
   # Read gain data from database
   # TODO: May want to read tag gain if dbgain is blank
   foreach {type} {Gain AlbumGain} {tag_type} {track_gain album_gain} {
      if {[catch {gainControl::get $fullName $type} gain]} {
         if {![info exists Details($type)]} {
            set Details($type) "--"
         }
      } else {
         set Details($type) $gain
      }
   }
   return $result
}
#------------------------------------------------------------------------------
# Function   : DeleteFile
# Description: Delete a file, if it is playing then handle it
# Author     : Tom Wilkason
# Date       : 3/24/2002
#------------------------------------------------------------------------------
proc catalog::DeleteFile {fileName} {
   Trace
   set currentSound [soundControl::fileName]
   set goNext 0
   if {[string equal $currentSound $fileName]} {
      # This is a trick to see if the song is actually playing
      if {[soundControl::isPlaying]} {
         soundControl::stop
         set goNext 1
      }
      soundControl::reset
   }
   # Now try to delete it
   if {[catch {
      cd $::HOME
      file delete -force $fileName
   } result] \
   } then {
      tk_messageBox -icon warning -type ok -message "Error removing $fileName\n$result"
   } else {
     catalog::removeFile $fileName
   }
   return $goNext
}

#------------------------------------------------------------------------------
# Function   :Remove
# Description: Remove a file from the master and playlists list if it exists
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::removeFile {File} {
   global PLB
   # Remove from the master list
   set index [db::whereInDb $File]
   if {$index >= 0} {
      ##
      # Remove from the master list
      #
      db::removeFile $index
      ##
      # Remove from the Playlist
      #
      LB::removeFile $File
      ##
      # Remove from the APL File List (in case it is there)
      #
      catch {aplLB::RemoveFile $File}
   }
}
#------------------------------------------------------------------------------
# Function   : renameInPlaceDialog
# Description: Bring up a in place editing window on a listbox
# Author     : Tom Wilkason
# Date       : 6/19/2002
#------------------------------------------------------------------------------
proc catalog::renameInPlaceDialog {W table {oldName "xxx"}} {
   Trace
   if {![winfo ismapped $W]} {
      return ""
   }
   ##
   # Need to compute location for dialog
   #
   set active [$W index active]
   set bba [$W bbox $active]
   if {[llength $bba]<4} {
      return ""
   }
   set w  .rndialog
   destroy $w
   toplevel $w -class SnackAmp
   foreach {x y - -} $bba {break}
   # If table, adjust for heading
   if {$table} {
      set y [expr {$y - [winfo y $W]}]
   }
   # Left Hand alignment
   set rx [winfo rootx $W]
   set ry [winfo rooty $W]
   set X [expr {$rx+$x}]
   set Y [expr {$ry+$y}]
   # Make it look like the text underneath
   set font [$W cget -font]
   set bg [$W cget -background]
   set fg [$W cget -foreground]
   set ::TA(rename) $oldName
   # Set size available width and linespacing+2
   set wid [expr {[string length $oldName]+20}]
   set Wwid [winfo width $W]
   set Wheight [expr {[font metrics $font -linespace]+2}]
   pack [entry $w.rn -bd 2 -relief groove -font $font -insertbackground $fg \
      -width $wid -bg $bg -fg $fg -textvariable ::TA(rename)] -expand true -fill x
   #$w.rn selection range 0 end
   wm withdraw $w
   wm overrideredirect $w 1
   set oldFocus [focus]
   wm transient $w [winfo toplevel $W]
   focus -force $w.rn
   wm geometry $w "${Wwid}x$Wheight+$X+$Y"
   bind $w <Escape> {
      destroy .rndialog
      set ::TA(rename) ""
   }
   bind $w <FocusOut> {
      destroy .rndialog
      set ::TA(rename) ""
   }
   bind $w <Return> [list destroy .rndialog]
   wm deiconify $w
   raise $w
   # Force on top if possible
   update idletasks
   ontop::ifpossible $w $w
   tkwait window $w
   ##
   # Back from block
   #
   if {[winfo exists $oldFocus]} {
      focus -force $oldFocus
   }
   # Waiting for window to go away
   return $::TA(rename)
}
#----------------------------------------------------------------------------
# Return a list of folders to process (remove or rename) and from/to pairs of file names
# used to get all files prior to a move or copy operation. Handles both a
# single rename or a folder based rename.
# the returned folders can be removed when done.
# Note: File normalize will change the case to match the file system on windows
#----------------------------------------------------------------------------
proc catalog::namePairs {from to} {
   Trace
   if {[file pathtype $from] ne "absolute"} {
      set from [file normalize $from]
   }
   if {[file pathtype $to] ne "absolute"} {
      set to [file normalize $to]
   }
   set subLen [expr {[string length $from]-1}]
   set pfiles [list]
   set pfolders [list]
   set fl [expr {[string length $from]+1}]
   if {[file isdirectory $from]} {
      foreach {files folders} [fileutils::rglob $from 1] {break}
      foreach {file} $files {
         set target [string replace $file 0 $subLen $to]
         lappend pfiles [list $file $target]
      }
      foreach {folder} $folders {
         set target [string replace $folder 0 $subLen $to]
         lappend pfolders [list $folder $target]
      }
      lappend pfolders [list $from $to]
   } elseif {[file isdirectory $to]} {
      set target [file join $to [file tail $from]]
      lappend pfiles [list $from $target]
      set pfolders [list]
   } elseif {[file exists $from]} {
      set target [string replace $from 0 $subLen $to]
      lappend pfiles [list $from $target]
      set pfolders [list]
   } else {
      saLog "[me] $from does not exist!"
   }
   return [list $pfolders $pfiles]
}
#------------------------------------------------------------------------------
# Function   : catalog::RenameFile
# Description: Try to rename/move a file, handle sound objects if they are playing
#              resume play if just renamed a playing file
# Author     : Tom Wilkason
# Date       : 3/22/2002
#------------------------------------------------------------------------------
proc catalog::RenameFile {oldName newName} {
   return [catalog::moveTrackFolderPairs [list $oldName] [list $newName]]
}
proc catalog::moveFolder {folder target} {
   return [catalog::moveTrackFolderPairs [list $folder] [list $target]]
}
#------------------------------------------------------------------------------
# Function   : catalog::moveFolder
# Description: Move a folder (and its files) to some other place
#              and adjust the APL tree accordingly.
#              folder is existing full name
#              target is new full name
#              only called by APL code
# Author     : Tom Wilkason
# Date       : 12/17/2005
#------------------------------------------------------------------------------
proc catalog::moveTrackFolderPairs {fromList toList} {
   Trace
   set failures 0
   set currentSound [soundControl::fileName]
   set wasPlaying 0
   set plchanged [list]
   set trchanged [list]
   set dpairs [list]
   set fpairs [list]

   if {$::snackAmpSettings(dontMoveFiles)} {
         tk_messageBox -icon warning -type ok \
            -message "You don't have permission to move or remove files"
      return 0
   }
   APL::busyCursor 1
   ##
   # Gather all the file and folder names
   # BUG: File Normalize will "correct" the case if it knows about the file on windows
   foreach {from} $fromList {target} $toList {
      foreach {_dpairs _fpairs} [catalog::namePairs $from $target] {break}
      foreach {item} $_dpairs {
         lappend dpairs $item
      }
      foreach {item} $_fpairs {
         lappend fpairs $item
      }
   }
   ##
   # Grab old indicies in database, used to update the APL track list
   # Do it before we rename any tracks
   #
   foreach {fpair} $fpairs {
      foreach {from to} $fpair {
         if {$from ne $to} {
            set row [db::whereInDb $from]
            if {$row > -1} {
               lappend trchanged $row $from $to
            }
         } else {
#            debug "1 Identical pairs $from"
         }
      }
   }
   ##
   # Process the files
   #
   set i 0
   set break 0
   set todo [llength $fpairs]
   foreach {fpair} [lsort -index 0 -decreasing $fpairs] {
      set cc 0
      foreach {from to} $fpair {
         if {$from eq $to} {
#debug "2 Identical pairs $from"
            continue
         }
         if {($currentSound eq $from)} {
            set wasPlaying [soundControl::isPlayingTrack $from location]
         }
         # Name collision, ask what to do
         # Handle file and folder names differently
         if {[file exists $to]} {
            set cc [expr {[string equal -nocase  $from $to] && \
               ($::tcl_platform(platform) eq "windows")}]
            if {$cc == 0} {
               set ans [catalog::checkOverwrite $from $to]
               if {$ans eq "yes"} {
                  db::removeFile $to
               } elseif {$ans eq "cancel"} {
                  debug "Skipping remaining renames"
                  incr break
                  break
               } else {
                  debug "$to Exists! Skiping \n $to with \n $from"
                  set idx [lsearch -exact $trchanged $from]
                  # remove file from trchanged list
                  if {$idx > -1} {
                     ldelete trchanged [expr {$idx-1}] [expr {$idx+1}]
                  }
                  continue
               }
            }
         }
         # make sure folder exists
         waitWindow::percent [incr i] $todo "Moving\n $from to\n $to" 1
         if {[waitWindow::ifbreak]} {
            incr break
            break
         }
         set dirpart [file dirname $to]
         if {![file isdirectory $dirpart]} {
            if {[catch {file mkdir $dirpart} result] } then {
               saLog $result
               continue
            }
         }
         ##
         # try and rename the file
         #
         if {[catch {catalog::rename $from $to $cc} result]} then {
            saLog $result
            incr failures
            continue
         } else {
            if {[catalog::isTrackFile $from]} {
               # If target doesn't exist, just catalog new file
               if {[db::renameFile $from $to]==-1} {
                  unset -nocomplain Data
                  db::cacheFile $to Data 1
               } else {
                  lappend plchanged $from $to
               }
            }
         }
      }
      if {$break} {
         break
      }
   }
   ##
   # Process the folders in reverse order, remove them if empty
   #
   set dirChanged [list]
   foreach {dpair} [lsort -index 0 -decreasing $dpairs] {
      foreach {from to} $dpair {
#puts " 1 $from\n 2 $to"
         aplFolder::rename $from $to
         APL::newItemToRoot $to
         lappend dirChanged $from $to
         # Don't attempt delete if new folder is ontop of older folder
         if {![string match [file join $from *] $to]} {
            # NOTE: If only the folder case changed, we will still see files
            if {[file exists $from] && [llength [glob -nocomplain -directory $from *]]==0} {
               if {[catch {file delete -force $from} result] } then {
                  saLog $result
               } else {
                  APL::delItem $from
               }
            } else {
               # This may be OK. User may have skiped a file move
               #saLog "[me] $from still contains files after a rename!"
            }
         }
      }
   }
   # Change all files that reference these files
   # TODO: Make this an option (one for each)
   waitWindow::show "Updating references.."
   catalog::RenamePLFiles $plchanged
   if {[llength $dirChanged]} {
      catalog::RenameAPLFiles $dirChanged
   }
   # Fixup references
   #
   # Change all visible references to these files
   LB::changePlaylistFiles $plchanged
   trackList::renameFiles $trchanged
   # resume playing track, do after playlist entry changes
   if {$wasPlaying} {
      soundControl::resumePlayingTrack [LB::nameActive] $location
   }
   APL::busyCursor 0
   waitWindow::close
   return $failures
}
#----------------------------------------------------------------------------
# Copy a file, typically to a media player
#----------------------------------------------------------------------------
proc catalog::copyFile {oldName newName} {
   Trace
   return [catalog::copyTrackFolderPairs [list $oldName] [list $newName]]
}
#------------------------------------------------------------------------------
# Copy a single folder
#------------------------------------------------------------------------------
proc catalog::copyFolder {folder target} {
   Trace
   return [catalog::copyTrackFolderPairs [list $folder] [list $target]]
}
#----------------------------------------------------------------------------
# Return a part of a file name that has valid characters
#----------------------------------------------------------------------------
proc catalog::fatname {root} {
   set root [string map {& and / " " [ ( ] ) \{ ( \} ) : "" ? "" \" '} $root]   ;#"
   return $root
}
#----------------------------------------------------------------------------
# Return a short name for small mp3 players
#----------------------------------------------------------------------------
proc catalog::shortname {file _data} {
   upvar $_data Data
   foreach {idx} {Artist Track Album Title} {
      if {[info exists Data($idx)]} {
         set Data($idx) [string trim [catalog::fatname $Data($idx)]]
      } else {
         set Data($idx) "Unknown"
      }
   }
   set newname [file join $Data(Artist) "$Data(Album) - $Data(Track) - $Data(Title)[file extension $file]"]
   return $newname
}
#----------------------------------------------------------------------------
# Copy sets of tracks
#----------------------------------------------------------------------------
proc catalog::copyTrackFolderPairs {fromList toList} {
   Trace
   set failures 0
   set currentSound [soundControl::fileName]
   set wasPlaying 0
   set plchanged [list]
   set trchanged [list]
   set dpairs [list]
   set fpairs [list]
   if {$::snackAmpSettings(useFcopy)} {
      set method bfcopy
   } else {
      set method bcopy
   }
   if {$::snackAmpSettings(dontMoveFiles)} {
         tk_messageBox -icon warning -type ok \
            -message "You don't have permission to move or remove files"
      return 0
   }
   APL::busyCursor 1
   ##
   # Gather all the file and folder names
   #
   foreach {from} $fromList {target} $toList {
      foreach {_dpairs _fpairs} [catalog::namePairs $from $target] {break}
      foreach {item} $_dpairs {
         lappend dpairs $item
      }
      foreach {item} $_fpairs {
         lappend fpairs $item
      }
   }
   ##
   # Process the files
   #
   set i 0
   set break 0
   set todo [llength $fpairs]
   foreach {fpair} [lsort -index 0 -decreasing $fpairs] {
      set cc 0
      foreach {from to} $fpair {
         if {$from eq $to} {
            continue
         }
#puts " 1 $from\n 2 $to"
         # Name collision, ask what to do
         # Handle file and folder names differently
         if {[file exists $to]} {
            set ans [catalog::checkOverwrite $from $to 1]
            if {$ans eq "cancel"} {
               debug "Skipping remaining renames"
               incr break
               break
            } elseif {$ans eq "no"} {
               debug "$to Exists! Skiping \n $to with \n $from"
               continue
            }
         }
         # make sure folder exists
         waitWindow::percent [incr i] $todo "Copying\n $from to\n $to" 1
         if {[waitWindow::ifbreak]} {
            incr break
            break
         }
         set dirpart [file dirname $to]
         if {![file isdirectory $dirpart]} {
            if {[catch {file mkdir $dirpart} result] } then {
               saLog $result
               continue
            }
         }
         # try and copy the file
         if {[catch {$method $from $to} result]} then {
            saLog $result
            incr failures
            continue
         } else {
            if {[catalog::isTrackFile $from]} {
               catalog::addNewEntry $to $from
            }
         }
      }
      if {$break} {
         break
      }
   }
   ##
   # Process the folders in reverse order, remove them if empty
   #
   set dirChanged [list]
   foreach {dpair} [lsort -index 0 -decreasing $dpairs] {
      foreach {from to} $dpair {
         APL::newItemToRoot $to
      }
   }
   APL::busyCursor 0
   waitWindow::close
   return $failures
}
#----------------------------------------------------------------------------
# Copy all or selected tracks to an attached media player
# TODO: Modify to pass in the names and add functions to return
# the appropriate names from the menu level
#----------------------------------------------------------------------------
proc catalog::copyToMediaPlayer {mp nameFunction} {
   Trace
   set names [eval $nameFunction]
   set mydirs [list]
   foreach {dir} $::snackAmpSettings(mediaDirs) {
      set pat [file join $dir *]
      # Need to handle M:/ vs //Bertha/M
      set len [expr {[string length $pat]-1}]
      lappend mydirs $dir $pat $len
   }
   set todo [llength $names]
   set counter 0
   set useFolders [catalog::getMediaMode $mp]
   set fromList [list]
   set toList [list]
   set errs 0
   set re [db::buildRatingRE]

   foreach {name} $names {
      if {![file exists $name]} {
         continue
      }
      # For each possible dest pattern, check
      foreach {dir patt len} $mydirs {
         if {[string match $patt $name]} {
            ##
            # Copy all files if not useFolders
            # otherwise copy only track files
            if {[catalog::isTrackFile $name]} {
               unset -nocomplain Data
               db::lookup $name Data 2
               if {$useFolders ne "Default" && $useFolders ne "yes"} {
                  set tail "[catalog::smartFileName Data $useFolders][file extension $name]"
                  set dest [file join $mp $tail]
               } else {
                  set dest [file join $mp [string range $name $len end]]
               }
            } elseif {$useFolders eq "yes"} {
               set Data(Rating) "--"
               set dest [file join $mp [string range $name $len end]]
            } else {
               continue
            }
            ##
            # If source and dest are different, make sure we pass the ratings test
            #
            if {$name ne $dest} {
               # TODO: check rating here (if setting set), do a lookup on all tracks
               if {$::snackAmpSettings(checkMediaFiles)==0 || [regexp $re $Data(Rating)]} {
                  #saLog "Using $name with rating $Data(Rating)"
                  lappend fromList $name
                  lappend toList $dest
               }  else {
                  saLog "Skipping $name with rating $Data(Rating)"
               }
            }
            continue
         }
      }
   }
   ##
   # Perform the copy (todo: Just add to a FIFO that is serviced in an after loop)
   #
   if {[llength $fromList]} {
      set errs [catalog::copyTrackFolderPairs $fromList $toList]
   }
   if {$errs} {
      tk_messageBox -type ok -icon warning -message "One or more copy errors occured\nplease check the console log"
   }
}


#------------------------------------------------------------------------------
# Rename a file, if just case chaning on windows do
# an intermediate rename
# Note: On windows (FAT32), very inconsisten on renaming case only
# Doesn't work at all if the folder part of the file is changing names
#------------------------------------------------------------------------------
proc catalog::rename {oldName newName {cc 0}} {
   Trace
   if {$cc} {
      set ccname "$newName.[clock seconds]"
      if {[catch {file rename -- $oldName $ccname} result]} {
         return -code error $result
      } else {
         if {[catch {file rename -- $ccname $newName} result]} {
            catch {file rename -force -- $ccname $oldName}
            return -code error $result
         }
#debug "[me] dual rename \n\t$oldName \n\t$newName\n\t$ccname"
      }
   } else {
      file rename -force -- $oldName $newName
   }
}
#------------------------------------------------------------------------------
# Check to two files conflict and ask the user what to do.
#------------------------------------------------------------------------------
proc catalog::checkOverwrite {oldName newName {skipSame 0}} {
   Trace
   waitWindow::close
   if {[catalog::isTrackFile $oldName]} {
      catalog::trackFullData $oldName oldInfo
      catalog::trackFullData $newName newInfo
      # oldName is really latest file, newName will be existing file
      set textInfo "
\tNew      ->     \tExisting
Len: \t$oldInfo(duration)      \t$newInfo(duration)
Rate:\t$oldInfo(Rate)          \t$newInfo(Rate)
Size:\t[commify $oldInfo(size)]\t[commify $newInfo(size)]
"
      if {($oldInfo(size) == $newInfo(size))} {
         if {$skipSame} {
            # bcopy will skip identical files
            set res "yes"
         } else {
            set res [tk_messageBox -icon question -type yesnocancel -title "File Exists" \
               -message "Warning:\n$newName\nexists and appears to be identical, do you still want to overwrite it?\n\n$textInfo"]
         }
      } else {
         set res [tk_messageBox -icon question -type yesnocancel -title "File Exists" \
            -message "Warning:\n$newName\nexists, do you want to overwrite it?\n\n$textInfo"]
      }
   } else {
      set res "yes"
   }
   return $res
}
#----------------------------------------------------------------------------
# Copy clipboard files to a target
#----------------------------------------------------------------------------
proc catalog::getClipboardFiles {target} {
   set files [list]
   set targets [list]
   if {[catch {selection get -selection CLIPBOARD} tryFiles]} {
      return [list $files $targets]
   }
   clipboard clear
   # Ignore clibboard entries that aren't files
   # Ignore clibboard entries are duplicates
   # Ignore errors due to funky clipboard data
   catch {
      foreach {file} [lsort -unique $tryFiles] {
         if {[file exists $file]} {
            lappend files $file
            if {[file isdirectory $file]} {
               lappend targets [file join $target [file tail $file]]
            } else {
               lappend targets $target
            }
         }
      }
   }
   return [list $files $targets]
}
#------------------------------------------------------------------------------
# Function   : catalog::copyClipboardFiles
# Description: Copy the files in the clipboard to some new folder
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::copyClipboardFiles {target} {
   Trace
   set errs 0
   set fromList [list]
   set toList [list]
   foreach {fromList toList} [catalog::getClipboardFiles $target] {break}
   if {[llength $fromList] > 0} {
      set result [tk_messageBox -type yesno -title "Please Answer" \
         -icon question -message "Are you sure you want to copy:\n[join $fromList \n]\n to\n$target"]
      if {$result == "yes"} {
         if {[llength $fromList]} {
            set errs [catalog::copyTrackFolderPairs $fromList $toList]
         }
      }
   }
   return $errs
}
#------------------------------------------------------------------------------
# Function   : catalog::moveClipboardFiles
# Description: Move the files in the clipboard to some new folder
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::moveClipboardFiles {target} {
   Trace
   set errs 0
   set fromList [list]
   set toList [list]
   foreach {fromList toList} [catalog::getClipboardFiles $target] {break}
   if {[llength $fromList] > 0} {
      set result [tk_messageBox -type yesno -title "Please Answer" \
         -icon question -message "Are you sure you want to move:\n[join $fromList \n]\n to\n$target"]
      if {$result == "yes"} {
         if {[llength $fromList]} {
            set errs [catalog::moveTrackFolderPairs $fromList $toList]
         }
      }
   }
   return $errs
}

proc catalog::attachedFolders {args} {
   set folders [list]
   set atf [lsort -unique $::snackAmpSettings(mp3Player)]
   foreach {folder} $::TA(mediaDirs) {
      if {[lsearch -exact -sorted $atf $folder]>-1} {
         lappend folders $folder
      }
   }
   return $folders
}
proc catalog::fixedFolders {{mounted 1}} {
   set folders [list]
   set atf [lsort -unique $::snackAmpSettings(mediaDirs)]
   foreach {folder} $::TA(mediaDirs) {
      if {[lsearch -exact -sorted $atf $folder]>-1} {
         lappend folders $folder
      }
   }
   return $folders
}
##
# Return mounted folders
#
proc catalog::mountedFolders {args} {
   return [lsort -unique $::TA(mediaDirs)]
}

#----------------------------------------------------------------------------
# Return 1 if the folder is one of the fixed ones
#----------------------------------------------------------------------------
proc catalog::isMountedFolder {dest} {
   foreach {folder} $::TA(mediaDirs) {
      if {[string match $folder* $dest]} {
         return 1
      }
   }
   return 0
}
#----------------------------------------------------------------------------
# Return the icon type for a path, the list numbers are assigned in catalog (HACK)
#----------------------------------------------------------------------------
proc catalog::getIconType {folder} {
   global TA
   foreach {Path type icon} $TA(mediaDirPair) {
      if {[string match ${Path}* $folder]} {
         return $icon
      }
   }
   return 0
}
##
# Return 1 if the destination is a media player and the source is not
#
proc catalog::isMediaPlayer {dest {source ""}} {
   Trace
   set idest 0
   foreach {folder} $::snackAmpSettings(mp3Player) {
      if {[string match [file join $folder *] $dest] || [string equal $folder $dest]} {
         set idest 1
      } else {
         set idest 0
      }
      if {[string length $source] && [string length $folder]} {
         set isource [string match [file join $folder *] $source]
         if {$idest==1 && $isource==0} {
            return 1
         } else {
            set $idest 0
         }
      } elseif {$idest} {
         return 1
      }
   }
   return $idest
}
#------------------------------------------------------------------------------
# Rename all the tracks in all autoplaylist and playlist files,
# called when renaming a folder or a track
#------------------------------------------------------------------------------
proc catalog::RenameAPLFiles {pairs} {
   Trace
   set Files [catalog::getAutoPlaylists]
   return [catalog::RenameFiles $Files $pairs]
}
proc catalog::RenamePLFiles {pairs} {
   Trace
   #TODO: doesn't work with non-native playlists
   set ntPairs [list]
   foreach {item} $pairs {
      lappend ntPairs [file nativename $item]
   }
   set Files [catalog::getPlaylists]
   return [catalog::RenameFiles $Files $ntPairs]
}
proc catalog::RenameFiles {Files pairs} {
   Trace
   #set newName [catalog::globEscape $newName]
   #puts "$oldName -> $newName"
   set changes 0
   foreach {file} $Files {
      if {[catch {open $file r} fid] } then {
         debug $fid
      } else {
         fconfigure $fid -buffersize 300000
         set contents [read $fid]
         close $fid
         set hits 0
         set i 0
         set tot [expr {[llength $pairs]/2}]
         foreach {oldName newName} $pairs {
            incr i
            set oldName [catalog::globEscape $oldName]
            set newName [string map {\\ \\\\} $newName] ;# for native names
            incr hits [regsub -all -- $oldName $contents $newName contents]
            if {$i%1000==0} {
               waitWindow::percent [incr i] $tot "Processing $file with $hits changes"
            }
         }
         if {$hits > 0} {
            if {[catch {open $file w} fod] } then {
               debug $fod
            } else {
#puts "$hits changes in $file\n  ('$oldName' to '$newName')"
               fconfigure $fod -buffersize 300000
               puts -nonewline $fod $contents
               close $fod
            }
            incr changes $hits
         }
      }
   }
   return $changes
}
#------------------------------------------------------------------------------
# Function   : removeFolder
# Description: Remove a folder and all of its children files
#              Only used by APL tree
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::removeFolder {folder} {
   set i 0
   if {$::snackAmpSettings(dontMoveFiles)} {
      tk_messageBox -icon warning -type ok \
         -message "You don't have permission to remove folders"
      return
   }

   # verify before removing the catalog file from disk
   foreach {files folders} [fileutils::rglob $folder 1] {break}
   set i [expr {[llength $files]+[llength $folders]}]  ;# don't just check music tracks
   set message "Are you sure you want to remove\n$folder\nIt contains [llength $files] files and [llength $folders] folders\n\nAll files and folders will be permanently deleted!"
   if {$i == 0 || [tk_messageBox -icon warning -title "Please Respond" -type yesno -message $message] == "yes"} {
      waitWindow::show "Removing $folder and contents"
      ##
      # Remove the files
      #
      foreach {file} $files {
         if {[catch {
            DeleteFile $file
         } result]} then {
            debug $result
         }
      }
      ##
      # Handle the folders, work from the longest folder back
      # to maximize the deleted folders
      lappend folders $folder
      foreach path [lsort -decreasing $folders] {
         if {[catch {
            file delete -force $path
            aplFolder::Unset $path
            APL::delItem $path
         } result]} then {
            debug $result
            return
         }
      }
      # move to parent
      APL::treeSelect [file dirname $folder]
      aplLB::UpdateList [file dirname $folder]
      gui::updateFileDisplay
      waitWindow::close
      db::commit
   }
   return
}
#------------------------------------------------------------------------------
# Function   : catalog::SavePlaylistCache
# Description: Cache the current playlist for later restoration
# Author     : Tom Wilkason
# TBD: Save/restore extended info if it is faster than a lookup
#------------------------------------------------------------------------------
proc catalog::SavePlaylistCache {args} {
   global TA
   set sList [LB::keyAndNames]  ;# key|file pairs
   set loc [LB::idxPlaying]
   if {$loc < 0} {
      set loc [LB::idxActive]
   }
   append sList "\n$loc"
   append sList "\n[soundControl::samplePosition]"
   append sList "\n$soundControl::playState"
   if {[catch {safeWrite $TA(playCacheFile) $sList} result]} {
      return -code error "Save Error: Unable to open '$TA(playCacheFile)' for writing\n"
   }
}
#------------------------------------------------------------------------------
# Function   : catalog::ReadPlaylistCache
# Description: Restore the playlist from a cached file (during initialization)
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::ReadPlaylistCache {args} {
   global TA
   ##
   # If new version (format) exists, use it, otherwise try old version
   #
   if {[file exists $TA(playCacheFile)]} {
      set pairs 1
      set playCacheFile $TA(playCacheFile)
   } else {
      set playCacheFile [string trimright $TA(playCacheFile) "3"] ;# try old style
      set pairs 0
   }
   if {![file exists $playCacheFile]} {
      return
   }
   if {[catch {open $playCacheFile r} InFile]} {
      return -code error "Read Error: Unable to open '$playCacheFile' for reading\n"
   }
   fconfigure $InFile -buffersize 300000
   set sList  [gets $InFile]
   set cIndex [gets $InFile]
   # Restore position within the file
   if {[eof $InFile]} {
      set cPosition -1
   } else {
      set cPosition [gets $InFile]
   }
   if {[eof $InFile]} {
      set playState "Stopped"
   } else {
      set playState [gets $InFile]
   }
   close $InFile
   # Replace current playlist very fast
   LB::deleteAll
   if {[llength $sList] > 0} {
      if {$pairs} {
         LB::replacePairList $sList
      } else {
         LB::insertList end $sList
      }
      if {$::snackAmpSettings(resumePlay)} {
         if {$playState=="Playing"} {
            after idle catalog::resumePlay $cIndex $cPosition
         } elseif {$playState=="Paused"} {
            after 500 catalog::resumePlay $cIndex $cPosition 1
         } else {
            macros::Highlight $cIndex
         }
      } else {
         macros::Highlight $cIndex
      }
      #saLog "Restored playlist cache '$pairs' '$playCacheFile' of [expr {[llength $sList]/2}] files"
   }
}
##
# Function to resume play
#
proc catalog::resumePlay {index offset {pause 0}} {
   macros::Highlight $index
   Play active $offset $pause
}
#------------------------------------------------------------------------------
# Return a list of all AutoPlaylist files
#------------------------------------------------------------------------------
proc catalog::getAutoPlaylists {args} {
   return [lsort -dictionary [glob -nocomplain -directory $::HOME *.tafs]]
}
#------------------------------------------------------------------------------
# Return a list of all playlist files
#------------------------------------------------------------------------------
proc catalog::getPlaylists {args} {
   Trace
   return [lsort -dictionary [glob -nocomplain -directory $::HOME *.taf *.m3u]]
}
#------------------------------------------------------------------------------
# Return the type of file spliting the user wans for media players
#------------------------------------------------------------------------------
proc catalog::dialogtMediaMode {mp} {
   variable mpdFormat
   variable Continue
   global snackAmpSettings
   set T .mpd
   destroy $T
   toplevel $T
   set instructions {
Default = Maintain File Structure
%A = Artist
%C = Album
%N = Track Number
%T = Track Title
You can use / between tags to create destination folder structure
   }
   set catalog::mpdFormat($mp) "Default"
   pack [frame $T.top -pady 5] -side top -fill x -expand true
   pack [label $T.top.lab -text "Enter how you would like to split the track names\n$instructions"]
   pack [entryHistory::create $T.top.filter catalog::mpdFormat($mp) ::snackAmpSettings(fileStructure)] \
      -side left -expand yes -fill x

   pack [frame $T.f] -side top -fill x -expand true
   pack [button $T.f.o -padx 4 -text OK -compound left -image saImages.Apply \
      -command [list set catalog::Continue ok] \
      -default active -relief raised] -side left
   pack [button $T.f.c -padx 4 -text Cancel -compound left -image saImages.Close \
      -command [list set catalog::Continue cancel] \
      -default normal -borderwidth 1 -relief raised] -side right
   bind $T <Return> [list $T.f.o invoke]
   bind $T <Escape> [list $T.f.c invoke]
   grab set $T
   set focus [focus]
   if {[winfo exists $focus]} {
      wm transient $T [winfo toplevel $focus]
   }
   centerWindow $T
   tkwait visibility $T
   focus $T
   set Continue 0
   myIcon $T

   #-------WAIT---------
   vwait catalog::Continue
   grab release $T
   destroy $T
   if {$Continue eq "ok"} {
      set snackAmpSettings(mediaMode,$mp) $mpdFormat($mp)
   }
}
#------------------------------------------------------------------------------
# Return a "yes" if using folder names, "no" if tag info
# ask user if not known.
#------------------------------------------------------------------------------
proc catalog::getMediaMode {mp} {
   global snackAmpSettings
   if {![info exists snackAmpSettings(mediaMode,$mp)]} {
#      set result [tk_messageBox -type yesno -title "Please respond" -icon question \
#         -message "Please answer yes to use folder structure to copy, \nno to use ID3 Artist/Album - Title for $mp"]
#      set snackAmpSettings(mediaMode,$mp) $result
      catalog::dialogtMediaMode $mp
   }
   return $snackAmpSettings(mediaMode,$mp)
}
#----------------------------------------------------------------------------
# Based on tags, return a proper file name path, used for
# file copying or renmaing.
#----------------------------------------------------------------------------
proc catalog::smartFileName {_Array fname} {
   variable defNames
   upvar $_Array Data
   set name ""
   # Use default field names if not passed in
   foreach {fld val} [array get defNames] {
      if {![info exists Data($fld)] || $Data($fld) eq ""} {
         set Data($fld) $val
      }
   }
   # Substitute the fields with appropriate data
   # %A - %C - %N - %T for example
   foreach {field} [list Artist Album Track Title] {code} [list %A %C %N %T] {
      if {[hasData Data($field)] && $Data($field) != "0"} {
         # Substitute tag with data if it exists
         if {$field=="Track"} {
            # Format tracks using two digits
            set Data($field) [id3Tag::cleanTrack $Data($field)]
            # Remove embedded / in track numbers
            regsub -all  -- {/} $Data($field) { of } Data($field)
         } else {
            # No embedded / separators in tag
            regsub -all -- {/} $Data($field) {-} Data($field)
         }
         regsub -all -- $code $fname $Data($field) fname
      }
   }
   # Remove unused or invalid or unused codes
   regsub -all  -- {%A|%C|%N|%T} $fname {} fname
   # Remove invalid characters that can't be used for file names
   set fname [string map {< " " > " " | " " * "" ? "" [ ( ] ) \{ ( \} ) \" '} $fname] ;#"
   # remove sequential separators (assumes - is used as separator)
   regsub -all -- {-[ ]+-} $fname {-} fname
   # Clean up around path separators
   regsub -all -- {/[ -.]+} $fname {/} fname
   regsub -all -- {[ -.]+/} $fname {/} fname
   # no sequential path separators
   while {[regsub -all -- {//} $fname {/} fname]} {}
   # Collapse remaining spaces and other bad juju to single space
   while {[regsub -all  -- {  |\\|\?|:|\*} $fname { } fname]} {}
   # Trim any whitespace or separators on ends
   set fname [string trim $fname " -"]
   return $fname
}
#----------------------------------------------------------------------------
# Utility routines
#----------------------------------------------------------------------------
proc tester {{num 5}} {
   set TRACKS [db::sortedList]
   puts [time {
      foreach {file} $TRACKS {
         set dir [file dirname $file]
      }
   } $num]
   puts [time {
      foreach {file} $TRACKS {
         set dir [file rootname $file]
      }
   } $num]
   puts [time {
      foreach {file} $TRACKS {
         set dir [file tail $file]
      }
   } $num]
}
#------------------------------------------------------------------------------
# Function   : catalog::ShowDups
# Description: Command line routine to show potential duplicate files
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc catalog::ShowDups {{pattern *}} {
   foreach {file} [db::filesMatchingPattern $pattern] {
      set dir [file dirname $file]
      set tail [file tail $file]
      set fileTail [lrange [file split $file] end-1 end]
      if {![info exists tailHash($tail)]} {
         set tailHash($tail) $file
      } else {
         if {[catch {
            set deltasize [expr {abs([file size $file]-[file size $tailHash($tail)])}]
            if {$deltasize<512} {
               saLog "Duplicate $tail\n   '$file'\n   '$tailHash($tail)'\n"
            }
         } result] } then {
            saLog $result
         }
         update
      }
   }
}
#------------------------------------------------------------------------------
# Function   : catalog::DupSizes
# Description: Command line routine to show potential duplicate files
#              based on like file sizes. Good for finding renamed files.
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc catalog::DupSizes {{pattern *} {threshold 2}} {
   foreach {file} [db::filesMatchingPattern $pattern] {
      if {[file extension $file] ne ".usk" && [db::lookup $file Data]} {
         set size $Data(size)
         lappend tailHash($size) [file tail $file]
      }
   }
   #Foreach file size
   foreach {entry files} [array get tailHash] {
      if {[llength $files] > 1} {
         set last [expr {[llength $files]-1}]
         set matches 0
         # Pass through each set of duplicate files
         for {set i 0} {$i < $last} {incr i} {
            set base [split [string map {" " ""} [string tolower [file tail [lindex $files $i]]]] -]
            set next [split [string map {" " ""} [string tolower [file tail [lindex $files [incr i]]]]] -]
            foreach part $base {
               if {[lsearch -exact $next $part] > -1} {
                  incr matches
               }
            }
         }
         if {$matches > $threshold} {
            puts "Duplicates found for size $entry ($matches)"
            foreach file $files {
               puts "   $file"
            }
         }
      }
   }
}
#------------------------------------------------------------------------------
# Read the cached files into memory
# Set up the default teFolders
# Returns a list of paths not found
#------------------------------------------------------------------------------
proc catalog::convertCacheFiles {args} {
   global TA snackAmpSettings
   ##
   # Restore the cached files from each media Dir
   # Support either new type (a list) or old type (Array)
   set result [list]
   # Move list into a hash table for faster access
   # and to allow replacing parts by single path
   #catalog::validateMediaPaths
   foreach {Path} [lsort -unique $TA(mediaDirs)] {
      waitWindow::show "Importing Directory $Path - Please Wait"
      # Read either v1 or v2 cache files
      if {![file exists $Path]} {
         waitWindow::close
         tk_messageBox -icon warning -type ok -title "Warning" \
            -message "Media Folder '$Path' does not exist,\nit may be named improperly or on \nan unavailable network computer."

      } else {
         if {[catch {set fid [open [file join $Path $TA(cacheFile)] r]}]} {
            lappend result $Path
         } else {
            # V2 files
            fconfigure $fid -buffersize 100000
            gets $fid tLlist
            close $fid
            saLog "Converting $Path (v2)"
            # add the new ones for this path back
            set i 0
            set total [llength $tLlist]
            foreach {File} $tLlist {
               set file [file join $Path $File]
               incr i
               if {$i%25==0} {
                  waitWindow::percent $i $total "Importing\n$Path" 1
                  if {[waitWindow::ifbreak]} {
                     db::commit
                     return
                  }
               }
               db::cacheFileName $file 1
            }
            db::commit
         }
      }
   }
   return
}
#------------------------------------------------------------------------------
# Function   : fixplus::correctCase
# Description: Replace + in file names (due to http download), also fix other
#              case problems and clean up spacing.
# Author     : Tom Wilkason
# Date       : 1/8/2001
#------------------------------------------------------------------------------
namespace eval fixplus  {}
proc fixplus::correctCase {folder} {
   APL::busyCursor 1
   update idletasks
   set changed 0
   # Use a folder or the selected names
   if {[file isdirectory $folder]} {
      set Files [glob -nocomplain -directory $folder -types {f l} *]
   } else {
      set Files $folder
   }
   set fromList [list]
   set toList [list]
   foreach File $Files {
      set folder [file dirname $File]
      set File   [file tail $File]
      ##
      # Clean up the file name
      #
      set ext [file extension $File]
      set root [file rootname $File]   ;# base part of name
      if {[catalog::isTrackFile $File]} {
         # Clean out funny stuff
         set newRoot [cleanName $root]
         set ofn [file join $folder $File]
         set fnf [file join $folder "$newRoot$ext"] ;#full new file
         ##
         # Have new file name
         #
         if {$newRoot ne $root} {
            ##
            # Try to rename the file here
            #
            lappend fromList $ofn
            lappend toList $fnf
         }
      }
   }
   if {[llength $fromList]} {
      set errs [catalog::moveTrackFolderPairs $fromList $toList]
   } else {
      set errs 0
   }
   APL::busyCursor 0
   return [llength $fromList]
}
#------------------------------------------------------------------------------
# Function   : cleanName
# Description: Clean up a file name to use - as separator, correct case and clean up
#              leading track numbers to use - as a separator
# Author     : Tom Wilkason
# Date       : 3/19/2002
#------------------------------------------------------------------------------
proc fixplus::cleanName {root} {
   # sp is user supplied split char, "-" is recommended
   set sp [string trim $::snackAmpSettings(splitChar)]
   set root [string map {| " " * "" ? "" [ ( ] ) \{ ( \} ) \" '} $root] ;#"
   regsub -all -- {[+_ ]+} $root { } root
   regsub -all -- {( *- *)|( *\+ *)} $root " $sp " root
   set root [string trim $root]
   # Fix first part if a number
   set reExp [subst -nocommand {^([0-9]+)[$sp .]*(.*?)}]
   if {[regexp -- $reExp $root -> num rest]} {
      set root "$num $sp $rest"
   }
   # Cap each token
   set root [stringTitle $root]
   return $root
}
#  ----------------------------------------------------------------------------
#  aplFolder namespace: Wrap access to the AutoPlaylist folder tree settings
#  ----------------------------------------------------------------------------

namespace eval aplFolder {
   variable mods 0
}

#------------------------------------------------------------------------------
# flag as dirty
#------------------------------------------------------------------------------
proc aplFolder::dirty {args} {
   variable mods
   incr mods
}
proc aplFolder::looper {time} {
   after cancel aplFolder::looper $time
   aplFolder::saveFolderNames
   after $time aplFolder::looper $time
}

#------------------------------------------------------------------------------
# Function   : getAPLFileList
# Description: Return a list of all APL files
# Author     : Tom Wilkason
# Date       : 12/24/2002
#------------------------------------------------------------------------------
proc aplFolder::getAPLFileList {args} {
   set result [list]
   # TODO: This would be faster to lsort -unique a list of folders
   # or to flag all files and query them
   foreach {File} [db::sortedList] {
      if {[isSet [file dirname $File]]} {
         lappend result $File
      }
   }
   return $result
}

#------------------------------------------------------------------------------
# Function   : foldersMatchingPattern
# Description: Return a list of folders in or under a path
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::foldersMatchingPattern {{pattern *} {sorted 0}} {
   if {$sorted} {
      return [lsort [array names ::teFolders $pattern]]
   } else {
      return [array names ::teFolders $pattern]
   }
}
#------------------------------------------------------------------------------
# Function   : aplFolder::getFolders
# Description: Return a list of some folders and subfolders of subfolders 'subfolders'
# Author     : Tom Wilkason
# Date       : 2/16/2003
#------------------------------------------------------------------------------
proc aplFolder::getFolders {selectedFolder {subfolders 0}} {
   lappend folders $selectedFolder
   # Find any remaining subfolders
   if {$subfolders>0} {
      set baseDepth [llength [file split $selectedFolder]]  ;#0 is just this folder
      foreach {index} [aplFolder::foldersMatchingPattern $selectedFolder/*] {
         set thisDepth [llength [file split $index]]
         if {$thisDepth <= ($baseDepth+$subfolders)} {
            lappend folders $index
         }
      }
   }
   # It would be nice to have -dictionary, but the sorted search will fail
   return [lsort $folders]
}
#------------------------------------------------------------------------------
# Function   : folderArray
# Description: Return an array (index element) of folders in or under a path
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::folderArray {{pattern *}} {
   return [array get ::teFolders $pattern]
}
#------------------------------------------------------------------------------
# Function   : isSet
# Description: Return true if the folder item is set (and exists)
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::isSet {dirName {value 1}} {
   if {[info exists ::teFolders($dirName)] && ($::teFolders($dirName) == $value)} {
      return 1
   } else {
      return 0
   }
}
#------------------------------------------------------------------------------
# Function   : aplFolder::info exists
# Description: Veriy the entry or the actual folder exists
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::folderExists {Path {check 0}} {
   if {$check} {
      return [expr {[info exists ::teFolders($Path)] || [file isdirectory $Path]}]
   } else {
      return [info exists ::teFolders($Path)]
   }
}
#------------------------------------------------------------------------------
# Function   : Unset
# Description: Remove one or moer entries
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::Unset {key} {
   unset -nocomplain ::teFolders($key)
}

#------------------------------------------------------------------------------
# Verify and remove all nodes that no longer exist
#------------------------------------------------------------------------------
proc aplFolder::removeUnusedNodes {pattern} {
   global teFolders
   foreach {key} [array names teFolders $pattern] {
      if {![file exists $key]} {
         unset teFolders($key)
         aplFolder::dirty
         if {[APLExists]} {
            APL::delItem $key
         }
      }
      update
   }
}
#------------------------------------------------------------------------------
# Function   : stripBranch
# Description: Remove one or moer entries
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::stripBranch {pattern} {
   array unset ::teFolders $pattern
   aplFolder::dirty
}
#------------------------------------------------------------------------------
# Function   : SetSome
# Description: Set some values to a value
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::SetSome {pattern value} {
   foreach {key} [array names ::teFolders $pattern] {
      set ::teFolders($key) $value
   }
   aplFolder::dirty
}
#------------------------------------------------------------------------------
# Function   : aplFolder::Set
# Description: Set an entry to some value
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::Set {key value} {
   set ::teFolders($key) $value
   aplFolder::dirty
}
#------------------------------------------------------------------------------
# Function   : aplFolder::SetIfBlank
# Description: Set an entry to some value only if it doesn't already exist
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::SetIfBlank {key value} {
   if {![info exists ::teFolders($key)]} {
      set ::teFolders($key) $value
      aplFolder::dirty
   }
}
#------------------------------------------------------------------------------
# Function   : aplFolder::SetIfExists
# Description: Set an entry to some value
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::SetIfExists {key value} {
   if {[info exists ::teFolders($key)]} {
      set ::teFolders($key) $value
      aplFolder::dirty
   }
}
#------------------------------------------------------------------------------
# Function   : aplFolder::Get
# Description: Set an entry to some value
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::Get {key} {
   return $::teFolders($key)
}
#------------------------------------------------------------------------------
# Function   : aplFolder::rename
# Description: Rename an entry retaining its old value
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::rename {path newName} {
   global teFolders
   Trace
   if {[info exists teFolders($path)]} {
      set value $teFolders($path)
      unset teFolders($path)
      set teFolders($newName) $value
   } else {
      set teFolders($newName) 0
   }
   aplFolder::dirty
}

#------------------------------------------------------------------------------
# Rename an entire branch
#------------------------------------------------------------------------------
proc aplFolder::renameBranch {oldBranch newBranch} {
   global teFolders
   if {[info exists teFolders($oldBranch)]} {
      set teFolders($newBranch) $teFolders($oldBranch)
   } else {
      set teFolders($newBranch) 0
   }
   set oldBranch "[string trimright $oldBranch /]/"
   set newBranch "[string trimright $newBranch /]/"
   foreach {From} [array names teFolders [file join $oldBranch *]] {
      set To [regsub -- ^$oldBranch $From $newBranch]
      set value $teFolders($From)
      unset teFolders($From)
      set teFolders($To) $value
   }
}
#----------------------------------------------------------------------------
# Return the number of tree entries
#----------------------------------------------------------------------------
proc aplFolder::entries {args} {
   return [array size ::teFolders]
}
proc aplFolder::savePattern {pattern} {
   return [array get ::teFolders $pattern]
}
proc aplFolder::restoreNames {saveNames {force 0}} {
   global teFolders
   if {$force} {
      foreach {folder value} $saveNames {
         set teFolders($folder) $value
      }
   } else {
      foreach {folder value} $saveNames {
         aplFolder::SetIfExists $folder $value
      }
   }
}
#------------------------------------------------------------------------------
# Function   : db::createTeFolders
# Description: Create all the teEntries and set them to 0 for the
#              database.
#              if a parent doesn't exist, create one and set it to 0
# Author     : Tom Wilkason
# Date       : 8/8/2003
#
#------------------------------------------------------------------------------
proc aplFolder::createTeFolders {{how 0}} {
   Trace
   global teFolders
   set mediaPatterns [list]
   foreach {dir} [lsort -unique $::TA(mediaDirs)] {
      lappend mediaPatterns [file join $dir *]
   }
   set hits 0
   set lastFolder {}
   $db::dbObj loop row {
      foreach {folder Mounted} [$db::dbObj get $row Folder Mounted] {break}
      ##
      # Build the folder tree entries
      #
      if {$Mounted==1 && ($folder ne $lastFolder)} {
         incr hits [aplFolder::setToRoot $folder $how]
      }
      set lastFolder $folder
   }
   aplFolder::dirty
   return $hits
}
##
# Check to see if a track is in a valid media folder
#
proc aplFolder::isInMediaDir {file mediaPatterns} {
   foreach {patt} $mediaPatterns {
      if {[string match $patt $file]} {
         return 1
      }
   }
   return 0
}
##
# Ensure we have a path to the root
#
proc aplFolder::setToRoot {folder how} {
   global teFolders
   set hits 0
   while {![info exists teFolders($folder)]} {
      incr hits
      set teFolders($folder) $how
      # Backup one
      set root [file dirname $folder]
      if {[string equal $folder $root]} {
         break
      } else {
         set folder $root
      }
   }
   return $hits
}
#------------------------------------------------------------------------------
# Function   : setTeDefault (recursive)
# Description: Fixup a branch of teFolders that may be undetermined
#              Work backwards and set the childrens value to that
#              of the parent all the way to the root. Default to 0
# Author     : Tom Wilkason
# Date       : 3/24/2001
#------------------------------------------------------------------------------
proc aplFolder::setTeDefault {Index} {
   global teFolders
   switch -- $teFolders($Index) {
      -1 {
         # Folder wasn't in new APL file,
         # Set it to it's parents value if possible
         set parent [file dirname $Index]
         if {$parent != $Index} {
            if {[info exists teFolders($parent)]} {
               set teFolders($Index) [aplFolder::setTeDefault $parent]
            } else {
              set teFolders($Index) 0
            }
         } else {
            set teFolders($Index) 0
         }
      }
      1 {
         # Folder is set on, if the parent was set off change
         # the parent to a partial state (we know it can't be off)
         set parent [file dirname $Index]
         if {($parent != $Index) && [aplFolder::isSet $parent 0]} {
            set teFolders($parent) 2
         }
      }
      2 {
         # If it was set partial before, just turn it off
         set teFolders($Index) 0
      }
      default {}
   }
   aplFolder::dirty
   return $teFolders($Index)
}
##
# Verify each tefolder entry (very slow)
#
proc aplFolder::check {args} {
   foreach {folder} [array names ::teFolders] {
      if {![file isdirectory $folder]} {
         unset ::teFolders($folder)
      }
   }
   aplFolder::dirty
}
##
# Save/restore the folder cache
# Only save known folders
#
proc aplFolder::saveFolderNames {args} {
   variable mods
   global teFolders
   global snackAmpSettings
   if {$mods>0} {
      foreach {folder} $snackAmpSettings(mediaDirs) {
         array set toCache [array get teFolders $folder]
         array set toCache [array get teFolders [file join $folder *]]
      }
      foreach {folder} $snackAmpSettings(mp3Player) {
         array set toCache [array get teFolders $folder]
         array set toCache [array get teFolders [file join $folder *]]
      }
      safeWrite $::TA(folderCacheFile) [array get toCache]
      set mods 0
   }
}
proc aplFolder::restoreFolderNames {args} {
   global TA
   if {[file exists $TA(folderCacheFile)]} {
      if {[catch {set fid [open $TA(folderCacheFile) r]} result]} {
         saLog $result
         return 0
      } else {
         if {[catch {
            fconfigure $fid -buffersize 300000
            set data [gets $fid]
            close $fid
            array set ::teFolders $data
            aplFolder::dirty
         } result] } then {
            saLog $result
            return 0
         }
      }
   }
   return 1
}
##
# Create an html page and launch the browser on it for the
# files on the clipboard
#
proc html {args} {
   if {[catch {selection get -selection CLIPBOARD} tryFiles]} {
      return ""
   }
   clipboard clear
   set files [list]
   # Ignore clibboard entries that aren't files
   # Ignore clibboard entries are duplicates
   # Ignore errors due to funky clipboard data
   set html ""
   catch {
      foreach {file} [lflip $tryFiles] {
         if {[db::lookup $file Data 1]>=3} {
            append html "<TR><TD>$Data(Artist)</TD>$Data(Title)<TD>[HMSfrac $Data(Duration)]</TD></TR>\n"
         }
      }
   }
   clipboard append $html
   return 1
}
aplFolder::looper 320000
package provide catalog 1.0
