#------------------------------------------------------------------------------
#                   C O L L A P S A B L E    F R A M E
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/cframe.tcl,v 1.17 2006/03/30 00:43:09 wilkason Exp $
}
#------------------------------------------------------------------------------
# Function   : cframe
# Description: Collapsable frame
# base:        name of outer collapsable frame, user is responsible for packing
# text:        title for the window frame, next to collapse button
# args:        additional user settings passed to inner frame
# returns hull: user can pack into this frame
# Author     : William J Giddings, modified by Tom Wilkason
#              Modified by Tom Wilkason for snackAmp such that the cframe
#              adjusts  the height but not the width of the application.
#------------------------------------------------------------------------------
proc cframe {base text args} {
   global cframe
   #-----------------------------------------------------------
   # Add the widget to the parent.
   #-----------------------------------------------------------
   # create button icons
   image create photo im_Open -data {
      R0lGODlhDAAKAOYAANnZ2Qln6w9r6wdo7wBPxQBe6QRk6wJl7wBNujV+6dvq
      /Lfm/iFr0AAEFgBZ4DSE7y2K/BF4/wAyfAFHwYy07vD//4TF/wU9mABP0Sl0
      4U+f/xmB/wNX0wAOJABBsAdMv6/P97Xj/0GO7wArgQxUxEaW+R6E/wVk7gAc
      SwAAAAA7qSVkybPa/nG5/xdl2SR35ieL/wht+gAvfgA2p0yE1pDL/zqU/yWH
      /wx0/wBAqwAEDAA0nWmh51Sn/wp2/wBPzQAOKwAsiw1CpGew+hRr6AAZUwAo
      hBhUuw44hAAAAgARPQAGGP//////////////////////////////////////
      ////////////////////////////////////////////////////////////
      ////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////yH5BAEAAAAA
      LAAAAAAMAAoAAAdogAECAwQAAAAABQYHCAkKCwwNAAAODxAREhMUFRYXABgZ
      GhscHR4fICEiIyQlJicoKQAqKywtLi8wMTIpAAAAMzQ1Njc4OTqAAIKCOzw9
      Pj9Ag4NBQkNERSmDgwBGR0hJg4ODSkuDgoEAOw==
   }

   image create photo im_Close -data {
      R0lGODlhCgAMAOYAANnZ2Q1x+Ch98wBZ6Bp28rba/lqW7QBR2BFv85rT/7/l
      /32s7ApRygBFwQAmYyVw2mu1/5jQ/5W/8SFbwQA4rwAIIwpKrDOJ9l+t/5jH
      +D5twQAwngAOLQAzkwxo6h+F/4rG/09+zAAoiQA9rCFauQps8hB+/yqF9SJQ
      ogAKJgBEuQVDsytu0yWF/xaD/x5y5AMregAADwFDthlr3xWA/xJ9/xRdzQAb
      VQAAAgJGuAl1/xF4+wxIrwAONAAAAAFEtw1m5gU0jAAGGQAbTAAbUgACCf//
      ////////////////////////////////////////////////////////////
      ////////////////////////////////////////////////////////////
      ////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////yH5BAEAAAAA
      LAAAAAAKAAwAAAdbgAABAgOAAIKDAAQFBgeEgggJCgsMDYQODxAREhMUhBUW
      FxgZGhuEHB0eHyAhIoMjJCUmJygpACorLC0uLzAxgjIzNDU2NziDOTo7PD0+
      hD9AQUKEgkNERYSCgQA7
   }

   frame $base
   # let the user pack the outside base frame
   frame $base.ohull  -borderwidth 1 ;#-height 180 ;#-relief ridge

   button $base.button  -borderwidth 1 \
      -image im_Open -padx 2 \
      -relief flat \
      -activebackground [[winfo toplevel $base] cget -background]
   label $base.label  -anchor w -relief flat -borderwidth 1 -text $text
   bind $base.button <Button-1> {
      [%W cget -image] configure -gamma 0.5
      cframe.toggle [winfo parent %W].ohull.ihull
   }
   bind $base.label <Button-1> {
      cframe.toggle [winfo parent %W].ohull.ihull
   }
   bind $base.button <Any-Enter> {
      #%W configure -relief raised
      [%W cget -image] configure -gamma 2
   }
   bind $base.button <Any-Leave> {
     # %W configure -relief flat
      [%W cget -image] configure -gamma 1
   }

   bind $base.label <Any-Enter> {
      [[winfo parent %W].button cget -image] configure -gamma 2
   }
   bind $base.label <Any-Leave> {
      [[winfo parent %W].button cget -image] configure -gamma 1
   }
   #
   # pack special frame/button
   #
   pack $base.ohull -anchor nw -expand 1 -fill both -side left  -ipadx 2 ;#-ipady 10
   place $base.button  -x 0 -y -1 -width 18 -height 18 -anchor nw -bordermode ignore
   place $base.label  -x 19 -y 1  -height 14 -anchor nw -bordermode ignore -width 500

   #
   # Create a label underneath the "title" to prevent overlap
   # then pack a frame in below that which will be the hull i.e. childsite for users
   #
   label $base.ohull.padder -text ""
   pack $base.ohull.padder -side top
   set hull $base.ohull.ihull
   # Create the inner frame using any additional arguments passed in
   eval frame {$hull} $args
   pack $hull -expand true -fill both
   # geometry will be "0x0" initially
   foreach {width height} [split [wmGeom $base] x] {break}
   set cframe($base,openSize) $height
   set cframe($base,hidden) 0

   return $hull
}
#-----------------------------------------------------------
# Toggle the widget, toggle the 'can be opened' icon.
# can also called as cframe.toggle $cframeWidget
#-----------------------------------------------------------
proc cframe.hidemode {ihull} {
   global cframe
   set base [winfo parent [winfo parent $ihull]]
   place forget $base.button
   place forget $base.label
   pack forget $base.ohull.padder
   set cframe($base,hidden) 1
}
#----------------------------------------------------------------------------
# External interface to toggle the frame, allow a controlling
# button to be passed in
#----------------------------------------------------------------------------
proc cframe.toggle {ihull {but ""}} {
   set button "[winfo parent [winfo parent $ihull]].button"
   set a [$button cget -image]
   if {$a == "im_Open"} {
      cframe.close $ihull
      if {[winfo exists $but]} {
         $but configure -relief flat
      }
   } else {
      cframe.open $ihull
      if {[winfo exists $but]} {
         $but configure -relief sunken
      }
   }
}
proc cframe.tryOpen {ihull} {
   set button "[winfo parent [winfo parent $ihull]].button"
   set a [$button cget -image]
   if {$a != "im_Open"} {
      cframe.open $ihull
   }
}
proc cframe.tryClose {ihull} {
   set button "[winfo parent [winfo parent $ihull]].button"
   set a [$button cget -image]
   if {$a == "im_Open"} {
      cframe.close $ihull
   }
}

#-----------------------------------------------------------
# Collapse the widget, display the 'can be opened' icon.
# can also called as cframe.close $cframeWidget
#-----------------------------------------------------------
proc cframe.close {ihull} {
   global cframe
   set base [winfo parent [winfo parent $ihull]]
   foreach {width height} [split [wmGeom $base] x] {break}
   $base.button configure -image im_Close
   pack forget $base.ohull
   # Make outer frame just small enough for button and title
   if {$cframe($base,hidden)} {
      $base configure -height 0
   } else {
      $base configure -height 16 -width 50
   }
   # Restore the size last saved unless it is first time, in that
   # case make the height the default requested height.
   set cframe($base,openSize) $height
   set myTopLevel [winfo toplevel $base]
   if {[info exists cframe($base,closedSize)] && $cframe($base,closedSize) > 0} {
      set height $cframe($base,closedSize)
   } else {
      update idletasks       ;# needed for reqheight to work
      set height [winfo reqheight [winfo toplevel $base]]
   }
   wm geometry $myTopLevel "${width}x$height"
}

#-----------------------------------------------------------
# Open the widget, display the 'can be closed' icon.
# cal also called as cframe.open $cframeWidget
#-----------------------------------------------------------
proc cframe.open {ihull} {
   global cframe
   set base [winfo parent [winfo parent $ihull]]
   foreach {width height} [split [wmGeom $base] x] {break}
   set cframe($base,closedSize) $height
   $base.button configure -image im_Open
   pack $base.ohull -anchor nw -expand 1 -fill both -side left -ipadx 2 -ipady 10
   # Restore the size last saved unless it is first time, in that
   # case make the height the default requested height.
   set myTopLevel [winfo toplevel $base]
   if {[info exists cframe($base,openSize)] && $cframe($base,openSize) > 0} {
      set height $cframe($base,openSize)
   } else {
      update idletasks       ;# needed for reqheight to work
      set height [winfo reqheight [winfo toplevel $base]]
   }
   wm geometry $myTopLevel "${width}x$height"
}
#-----------------------------------------------------------
# Return the geometry
#-----------------------------------------------------------
proc wmGeom {base} {
   set Geom [wm geometry [winfo toplevel $base]]
   set size [lindex [split $Geom "+"] 0]
   # Override initial 1x1 to requested geometry -> {}
   # It hasn't been packed so we don't know at this point
   if {$size == "1x1"} {
      set size {0x0}
   }
   return $size
}

package provide cframe 1.0
