#----------------------------------------------------------------------------
# Show all tail folders without cover art
#----------------------------------------------------------------------------
namespace eval clt {
   namespace export *
}
proc clt::missingArt {pattern} {
   albumart::clear
   set folders [db::foldersMatchingPattern $pattern 1]
   set todo [llength $folders]
   set i 1
   foreach {folder} $folders {
      # Check child dirs
      set dirs [glob -type d -nocomplain -directory $folder *]
      if {[llength $dirs]==0} {
         waitWindow::percent [incr i] $todo "Checking $folder" 1
         if {[waitWindow::ifbreak]} {
            waitWindow::close
            return
         }
         foreach {image valid} [albumart::covername $folder] {break}
         if {$valid==0} {
            set files [glob -type f -nocomplain -directory $folder *]
            # Check each file for embedded image
            foreach {file} $files {
               if {[catch {id3Tag::getImage $file} Image] } then {
                  saLog "Error on $file\n$::errorInfo"
               } elseif {$Image ne ""} {
                  puts "   $file has imbedded image"
                  set fname [file join $folder folder.jpg]
                  set apicfod [open $fname w]
                  fconfigure $apicfod -translation binary -encoding binary
                  puts -nonewline $apicfod $Image
                  close $apicfod
                  puts "   wrote image to $fname"
                  unset -nocomplain albumart::cache($folder)
                  set valid 1
                  break
               }
            }
         }
         if {$valid==0} {
            puts "Missing Art'$folder'"
         }
      }
   }
   waitWindow::close
}
#------------------------------------------------------------------------------
# Test to pull an APIC image from one file and insert into another file
#------------------------------------------------------------------------------
proc imtest {args} {
   set f1 "C:/Incoming/Gregg Allman/1997 One More Try/01 - test 1.mp3"
   set f2 "C:/Incoming/Gregg Allman/1997 One More Try/01 - test 1 (mod).mp3"

   file copy -force $f1 $f2
   set image [id3Tag::getImage $f1] ;# get raw jpeg info

   if {$image ne ""} {
      if {0} {
         # Wrte to temp file
         set tfile [file join $::env(TEMP) folder.jpg]
         set fod [open $tfile w]
         fconfigure $fod -translation binary -encoding binary
         puts -nonewline $fod $image
         close $fod

         id3Tag::putImageFromFile $f2 $tfile
      } else {
         id3Tag::putImageFromData $f2 $image
      }

      set image [id3Tag::getImage $f2]
      if {$image ne ""} {
         image create photo macros::myImage -data [base64::encode $image]
         set size [image height macros::myImage]
         puts "Output Size is $size"
      } else {
         debug "No output image found in $f2"
      }

   } else {
      debug "No input image found in $f1"
   }
}
proc clt::APICArt {pattern} {
   variable apicTitle
   albumart::clear
   package require base64
   set folders [db::foldersMatchingPattern $pattern 1]
   set todo [llength $folders]
   set i 1
   destroy .apic
   toplevel .apic
   set apicTitle ""
   image create photo apic
   pack [label .apic.lb -image apic]
   pack [label .apic.tx -textvariable macros::apicTitle]
   foreach {folder} $folders {
      # Check child dirs
      set dirs [glob -type d -nocomplain -directory $folder *]
      if {[llength $dirs]==0} {
         waitWindow::percent [incr i] $todo "Checking $folder" 1
         if {[waitWindow::ifbreak]} {
            waitWindow::close
            return
         }
         set files [glob -type f -nocomplain -directory $folder *]
         # Check each file for embedded image
         foreach {file} $files {
            set Image [id3Tag::getImage $file]
            if {$Image ne ""} {
               set idata [base64::encode $Image]
               set apicTitle [file tail $file]
               if {[catch {
                  image create photo rawApic -data $idata
                  image create photo apic
                  tkImageTools::resize rawApic apic 200 200
                  puts "Image->[image height rawApic]x[image width rawApic]:$file"
               } result] } then {
                  debug $result
               }
               update
            }
         }
      }
   }
   waitWindow::close
}
proc clt::fixFolderNames {args} {
   set folders [aplFolder::foldersMatchingPattern * 1]
   set tlen [llength $folders]
   set i -1
   foreach {Index} $folders {
      set subs [aplFolder::getFolders $Index 1]
      if {[llength $subs]==1} {
         set front [lindex [file tail $Index] 0]
         if {![string is integer -strict $front]} {
            waitWindow::percent [incr i] $tlen "Checking $Index for year" 1
            set newName [macros::prependFolderYear $Index]
            puts "Renamed $Index to $newName"
            if {[waitWindow::ifbreak]} {
               return
            }
         }
      }
   }
   waitWindow::close
}

#------------------------------------------------------------------------------
#
#------------------------------------------------------------------------------
proc clt::updateTimeTags {{pattern *}} {
   set DB  $db::DB
   set result [list]
   set rows [mk::select $DB -exact Mounted 1 -glob File $pattern]
   set i 0
   set len [llength $rows]
   foreach {row} $rows {
      set fname  [mk::get $DB!$row File]
      if {[file exists $fname]} {
         db::updateTimeTag $fname
      }
      waitWindow::percent [incr i] $len "Updating Database Time\n[splitFile $fname]" 1
      if {[waitWindow::ifbreak]} {
         break
      }
   }
   waitWindow::close
}
##
# Ensure tracks have both a V2 tag as well as a replay_gain tag for rockbox
#
proc clt::dumpGains {pattern} {
   set i 0
   set DB  $db::DB
   set result [list]
#   set 2weekago [expr {[clock seconds]-2*7*24*60*60}]
#   set rows [mk::select $DB -exact Mounted 1 -glob File $pattern -min date $2weekago]
   set rows [mk::select $DB -exact Mounted 1 -glob File $pattern]
   set i 0
   set len [llength $rows]
   set gained 0
   set tagged 0
   foreach {row} $rows {
      set fname  [mk::get $DB!$row File]
      waitWindow::percent [incr i] $len "Check Gain Tags\n[splitFile $fname]" 1
      if {[waitWindow::ifbreak]} {
         break
      }
      if {[file exists $fname] && [id3Tag::ismp3 $fname]} {
         set fix 0
         set nogain 1
         unset -nocomplain id3Data
         if {[catch {id3Tag::id3V2Get $fname id3Data} result] } then {
            puts "$fname\n$result"
            continue
         }
         unset -nocomplain uval
         foreach {gain} {track_gain album_gain} {
            # data is in dB
            if {[info exists id3Data($gain)]} {
               set nogain 0
               set val $id3Data($gain)
               set uval($gain) [gainControl::limit $val]
               # clamp values
               if {$val != $uval($gain)} {
                  set fix 1
                  puts "> $fname\n   $gain = $id3Data($gain) ($val limited to $uval($gain))"
               }
            }
         }
         # check for excessive values
         if {[info exists uval(track_gain)] && [info exists uval(album_gain)]} {
            set ratio [expr {abs($uval(track_gain)-$uval(album_gain))}]
            if {$ratio > 3.0} {
               set fix 1
            }
         }
         if {$fix || $nogain} {
            set track_gain [gainControl::limit [toDb [gainControl::create $fname]]]
            if {![info exists uval(album_gain)]} {
               set album_gain $track_gain
               puts " album_gain=$track_gain (from track_gain)"
            } else {
               set album_gain $uval(album_gain)
            }
            set ratio [expr {abs($track_gain-$album_gain)}]
            if {$ratio > 3.0} {
               puts " Excessive ratio=$ratio tr=$track_gain al=$album_gain"
               set album_gain $track_gain
            }
            puts " >>> Fix tr=$track_gain al=$album_gain\n\t$fname"
            # values should NOT be in dB
            if {[catch {gainControl::addTag $fname [fromDb $track_gain] [fromDb $album_gain]} result] } then {
               saLog $result
            }
         }
      }
   }
   waitWindow::close
}
##
# Make sure each track as a gain embedded in it, upgrade tag if necessary
# todo: Attached to auto-level menu on tree and file
#
proc clt::setGains {pattern} {
   set i 0
   set DB  $db::DB
   set result [list]
#   set 2weekago [expr {[clock seconds]-2*7*24*60*60}]
#   set rows [mk::select $DB -exact Mounted 1 -glob File $pattern -min date $2weekago]
   set rows [mk::select $DB -exact Mounted 1 -glob File $pattern]
   set i 0
   set len [llength $rows]
   set gained 0
   set tagged 0
   foreach {row} $rows {
      set fname  [mk::get $DB!$row File]
      if {[file exists $fname]} {
         unset -nocomplain id3Data
         set result [id3Tag::id3Label $fname id3Data {V2 V1}]
         if {[info exists id3Data(Tag)] && [string match "V2*" $id3Data(Tag)]} {
            if {![info exists id3Data(track_gain)]} {
               waitWindow::percent [incr i] $len "Computing Gain For\n[splitFile $fname]" 1
               set val [gainControl::create $fname]
               gainControl::addTag $fname $val "xxx"
               incr gained
            } else {
               waitWindow::percent [incr i] $len "Skipping\n[splitFile $fname]" 1
            }
         } elseif {[array size id3Data] > 2 && [id3Tag::ismp3 $fname]} {
            waitWindow::percent [incr i] $len "Upgrading Tags and Computing Gain For\n[splitFile $fname]" 1
            id3Tag::id3V2Modify $fname id3Data

            set val [gainControl::create $fname]
            gainControl::addTag $fname $val "xxx"
            incr gained
            incr tagged
            if {[catch {file stat $fname Stat} result] } then {
               saLog $result
            } else {
               db::setVal $fname Size $Stat(size)
               db::setVal $fname Date $Stat($db::checkTime)
            }
         } else {
            waitWindow::percent [incr i] $len "No Tag Found\n[splitFile $fname]" 1
            saLog "$fname has no tag"
         }
         if {[waitWindow::ifbreak]} {
            break
         }
      } else {
         saLog "$fname does not exist"
      }
   }
   saLog "Upgraded $tagged V1->V2 tags, Added $gained Gain settings"
   waitWindow::close
   return  1
}

##
# Correct incorrect ID3V2 tag lengths at the beginning of a file
#
proc clt::fixTag {pattern {fix 0}} {
   set DB  $db::DB
   set i 0
   set result [list]
   #set 2weekago [expr {[clock seconds]-2*7*24*60*60}]
   #set rows [mk::select $DB -exact Mounted 1 -glob File $pattern -max date $2weekago]
   set rows [mk::select $DB -exact Mounted 1 -glob File $pattern]
   set i 0
   foreach {row} $rows {
      set file  [mk::get $DB!$row File]
      waitWindow::percent [incr i] [llength $rows] "Checking\n[splitFile $file]" 1
      if {[waitWindow::ifbreak]} {
         break
      }
      if {[id3Tag::checkTagLength $file $fix data]} {
         if {[array size data]>0} {
            #array unset data APIC
            #puts "[array size data]-----$file"
            #puts "------$file"
            #parray data
         }
      }
   }
   waitWindow::close
   return
}
##
# Find the start/ending point of the track
#
proc clt::fixStart {pattern} {
   set i 0
   set DB  $db::DB
   set result [list]
#   set 2weekago [expr {[clock seconds]-2*7*24*60*60}]
#   set rows [mk::select $DB -exact Mounted 1 -glob File $pattern -min date $2weekago]
   set rows [mk::select $DB -exact Mounted 1 -glob File $pattern] ;# Don't get tracks without autolevel data
   set i 0
   foreach {row} $rows {
      set fname  [mk::get $DB!$row File]
      waitWindow::percent [incr i] [llength $rows] "Finding Start/Stop locations for gapless playback\n[splitFile $fname]" 1
      if {[waitWindow::ifbreak]} {
         break
      }
      if {[catch {soundControl::infoObj $fname} snd] } then {
         saLog "$fname: $snd"
         continue
      }
      ##
      # Save the start and stop positions
      #
      set Start [soundControl::findStart $snd]
      if {[string is integer -strict $Start]} {
         db::setVal $fname Start $Start
      } else {
         saLog "gainControl::create Got '$Start' for $fname Start Position"
         set Start 10000
      }
      set Stop [soundControl::findStop $snd]
      if {[string is integer -strict $Stop]} {
         db::setVal $fname Stop $Stop
      } else {
         saLog "gainControl::create Got '$Stop' for $fname Stop Position"
         set Stop 10000
      }
      db::setVal $fname Duration [$snd length -units seconds]
   }
   waitWindow::close
   return  1
}
proc clt::tagtest {{pattern *}} {
   package require id3dll
   set rows [mk::select $db::DB -glob File $pattern -exact Mounted 1 -exact Tag "V1.2"]
   foreach {row} $rows {
      unset -nocomplain dbData
      array set dbData [mk::get $db::DB!$row]
      set fullName $dbData(File)
      if {[catch {array set v2Data [id3::id3Label $fullName V2]} result] } then {
         continue
      }
      if {[catch {array set v1Data [id3::id3Label $fullName V1]} result] } then {
         continue
      }
      array set dbData [mk::get $db::DB!$row]
      if {[catch {
         if {[string length $v2Data(Title)] > [string length $dbData(Title)]} {
            puts "v2='$v2Data(Title)' db='$dbData(Title)'";update
   #          foreach {field} [array names id3Data] {
   #             catch {set id3Data($field) $dbData($field)}
   #          }
   #          if {[catch {id3Tag::id3V1Modify $fullName id3Data} result] } then {
   #             saLog $result
   #          }
         }
      } result] } then {
         saLog "$fullName ->$result"
         return
      }
      if {$row%100==0} {puts -nonewline .;update}
   }
}

#----------------------------------------------------------------------------
# Create a batch file to launch mp3gain manually
#----------------------------------------------------------------------------
proc clt::mp3gainbat {{pattern *}} {
   global teFolders
   set teKeys [lsort [array names teFolders $pattern]]
   set tail [list]
   set fod [open c:/mp3gain.bat w]
   foreach {key} $teKeys {
      if {[lsearch -sorted -glob $teKeys $key/*]<0} {
         puts $fod "cd \"[file nativename $key]\""
         puts $fod "mp3gain /a /d 3.0 /k /q *.mp3\""
      }
   }
   close $fod
}

#------------------------------------------------------------------------------
# Update the tag data from the database in case the file tags were
# corrupted.
#------------------------------------------------------------------------------
proc clt::fixTags {{pattern *}} {
   waitWindow::show "Fixing ... please wait"
   set rows [mk::select $db::DB -glob File $pattern -exact Mounted 1 -exact Tag "V1.2"]
   set m [llength $rows]
   set n 0
   foreach {row} $rows {
      unset -nocomplain dbData
      array set dbData [mk::get $db::DB!$row]
      set fullName $dbData(File)
      set result [id3Tag::id3Label $fullName id3Data V1]
      if {[string length $dbData(Title)] > [string length $id3Data(Title)]} {
         puts "db='$dbData(Title)' tag='$id3Data(Title)'"
         foreach {field} [array names id3Data] {
            catch {set id3Data($field) $dbData($field)}
         }
      }
      if {$row%10==0} {
         waitWindow::percent [incr n] $m "Fixing $n of $m please wait..."
      }
   }
   waitWindow::close
}
proc clt::fixTags1 {{pattern *}} {
   waitWindow::show "Fixing ... please wait"
   set rows [mk::select $db::DB -glob File $pattern -exact Mounted 1]
   set m [llength $rows]
   set n 0
   foreach {row} $rows {
      unset -nocomplain dbData
      unset -nocomplain id3V1Data
      unset -nocomplain id3V2Data
      array set dbData [mk::get $db::DB!$row]
      set fullName $dbData(File)
      id3Tag::id3Label $fullName id3V1Data V1
      id3Tag::id3Label $fullName id3V2Data V2
      # Note: Need to strip .. at end of V1 tag first
      catch {set id3V1Data(Title) [string trimright $id3V1Data(Title) .]}
      if {
          [info exists id3V2Data(Title)] && [info exists id3V1Data(Title)] && \
          [string length $id3V2Data(Title)] > [string length $id3V1Data(Title)] && \
          [string match $id3V1Data(Title)* $id3V2Data(Title)]
      } {
         puts "v2='$id3V2Data(Title)'\nv1='$id3V1Data(Title)'\n";update
         if {[catch {id3Tag::id3V1Modify $fullName id3V2Data} result] } then {
            saLog $result
         }
      }
      if {$row%10==0} {
         waitWindow::show "Fixing $n of $m please wait..."
      }
      incr n
   }
   waitWindow::close
}

#------------------------------------------------------------------------------
# Hand routine to show folders with excessive volume
#------------------------------------------------------------------------------
proc clt::showGains {{pattern *}} {
   set result [list]
   foreach {folder} [lsort [array names ::teFolders $pattern]] {
      foreach {folder gain} [gainControl::folderGain $folder] {break}
      if {$gain < -5.0} {
         waitWindow::show "Checking Gain\n\n$folder\nGain=$gain" 1
         lappend result [list $folder $gain]
         if {[waitWindow::ifbreak]} {
            break
         }
      }
   }
   set result [lsort -increasing -index 1 -real $result]
   waitWindow::close
   destroy .gainFix
   toplevel .gainFix
   set g .gainFix
   set i 0
   foreach {pair} $result {
      foreach {folder gain} $pair {
         set cb [button $g.[incr i] -text "mp3gain" -command [list mp3gain::mp3gainFolders [list $folder] "/s s /r /d 3.0"]]
         set lb [label $g.[incr i] -text $folder -anchor w -relief sunken]
         set gn [label $g.[incr i] -text $gain -anchor w -relief sunken]
         grid $cb $lb $gn -sticky ew
         if {$i>30} {
            break
         }
      }
   }
}
##
# Utility to rename cover art to a standard name
#
proc clt::renameCoverArt {pattern} {
   set FJ "Folder.jpg"
   # AlbumArtSmall.jpg is for small images
   foreach {folder} [aplFolder::foldersMatchingPattern $pattern 1] {
      update
      if {[catch {cd $folder} result] } then {
         saLog $result
         continue
      }
      set hasFJ [file exists $FJ]
      set images [glob -nocomplain *.jpg]
      set numImages [llength $images]
      if {$numImages==1} {
         set useImage [lindex $images 0]
         if {$hasFJ==0} {
            puts "$folder Renaming '$useImage' to '$FJ'"
            if {[catch {file rename -force $useImage $FJ} result] } then {
               saLog $result
            }
         }
      } elseif {$numImages>1} {
         puts "$folder has [llength $images] images"
         set maxDate 0
         foreach {image} $images {
            # Use the latest image date
            set mtime [file mtime $image]
            if {$mtime > $maxDate} {
               set useImage $image
               set maxDate $mtime
            }
         }
         if {[catch {file rename -force $useImage $FJ} result] } then {
            saLog $result
         }
      }
   }
   cd $::HOME
}
proc clt::showWithoutCoverArt {pattern} {
   foreach {folder} [aplFolder::foldersMatchingPattern $pattern 1] {
      if {[file isdirectory $folder]} {
         set subdirs [glob -type d -nocomplain -directory $folder *]
         set images  [glob -nocomplain -directory $folder *.jpg *.png]
         if {[llength $images]==0} {
            set images  [glob -type hidden -nocomplain -directory $folder *.jpg *.png]
         }
         if {[llength $images]==0 && [llength $subdirs] == 0} {
            puts "Empty:$folder"
            update
         }
      }
   }
}

proc clt::speed1 {{pattern *}} {
   set DB $db::DB
   set rows [mk::select $DB]
   foreach {row} $rows {
      foreach {artist album folder} [mk::get $DB!$row Artist Album Folder] {break}
      lappend Artist $artist
      lappend Album  $album
      lappend Folder $folder
   }
   puts "Artists [llength [lsort -unique $Artist]]"
   puts "Albums [llength [lsort -unique $Album]]"
   puts "Folders [llength [lsort -unique $Folder]]"
}

#------------------------------------------------------------------------------
#  Dump all of the images
#------------------------------------------------------------------------------
proc clt::dumpImages {args} {
   set folder "C:/CVS/snackAmp/icons"
   foreach image [image names] {
      set suffix [$image cget -format]
      set root [string map {:: - . _} $image]
      set file [file join $folder $root]
      if {$suffix eq ""} {
         foreach {format suffix} {GIF gif PNG png} {
            if {[catch {$image write $file.$suffix -format $format} result] } then {
               puts $result
               file delete -force $file.$suffix
            } else {
               puts "Wrote $file.$suffix"
            }
         }
      } else {
         if {[catch {$image write $file.$suffix -format [string toupper $suffix]} result] } then {
            saLog $result
            file delete -force $file.$suffix
         } else {
            puts "Wrote $file.$suffix"
         }
      }
   }
}
proc clt::showImages {args} {
   foreach image [lsort [image names]] {
      puts "$image -> [$image cget -format]"
   }
}


#------------------------------------------------------------------------------
# sandbox code
#------------------------------------------------------------------------------
proc demo {args} {
   set W .demo
   destroy $W
   toplevel $W
   set TA(demovar) dummy
   menubutton $W.en -indicatoron 1 -pady 0 -font listbox \
      -textvariable ::TA(demovar) -menu $W.en.menu -direction flush
#          button $W.en$row -pady 0 -font listbox \
#             -textvariable ::${W}($Index)  \
#             -image saImages.menubutton -compound right
   balloon::define $W.en "Click to select a genre"
   widgetWatch::setWatch $W.en TA(demovar)

   menu $W.en.menu \
      -postcommand [list id3Editor::_postGenreNames $W.en.menu ::TA(demovar)]

   label $W.en -indicatoron 1 -pady 0 -font listbox \
      -textvariable ::TA(demovar) -menu $W.en.menu -direction flush

   pack $W.en $W.nmb -fill x -expand true

}
#------------------------------------------------------------------------------
# Rename files to Artist/Album - track - title
#------------------------------------------------------------------------------
proc clt::artRename {args} {
   set from [pwd]
   foreach {files folders} [fileutils::rglob $from 1] {break}
   foreach {file} [lsort $files] {
      unset -nocomplain Data
      if {[db::lookup $file Data 1]} {
         set newname [catalog::shortname $file Data]
         set folder $Data(Artist)
         file mkdir $folder
         if {$newname ne $file} {
            if {[catch {file rename -force $file $newname} result] } then {
               puts $result
               puts "   $file"
               puts "   $newname"
               update
            } else {
               db::lookup $newname Data 2
            }
         }
      }
   }
}

package provide commandline 1.0
