#------------------------------------------------------------------------------
# fileAssociation Package 1.0
#
# Functions to associate file types for SnackAmp
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#
#------------------------------------------------------------------------------
package require registry
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/fileAssociation.tcl,v 1.10 2004/02/05 05:05:26 wilkason Exp $
}
namespace eval fileAssociation {
# These currently get temp saved to a file first, so no good
   variable mimeTypes {
      .m3u audio/x-mpegurl
      .pls audio/x-scpls
      .mp3 audio/x-mpeg
      .mp3 audio/mpeg
      .ogg audio/x-ogg
      .ogg audio/ogg
      .wav audio/x-wav
   }

}
#------------------------------------------------------------------------------
# Function   : isRegisteredMF
# Description: Determine if a file is registered to SnackAmp
#------------------------------------------------------------------------------
proc fileAssociation::isRegisteredMF {{ftype .mp3} _var} {
   Trace
   upvar $_var var
   if {[catch {registry get [join [list HKEY_CLASSES_ROOT $ftype] \\] ""} class]} {
      set var 0
      return 0
   }
   set resKey "HKEY_CURRENT_USER\\Software\\SnackAmp\\Restore\\$class\\DefaultIcon"
   if {[catch {registry get $resKey ""} defIcon]} {
      set var 0
      return 0
   }
   set var 1
   return 1
}

#------------------------------------------------------------------------------
# Function   : registerMF
# Description: Register some file type on windows for launching snackamp
# need to a setting for mime types for m3u/pls file
# HKEY_CLASSES_ROOT\MIME\Database\Content Type\audio/x-scpls -> Extension=.pls
# Will have to rename and CSLID items to CSLID.bak to allow the file association
# to be used.
#------------------------------------------------------------------------------
proc fileAssociation::registerMF {{ftype .mp3}} {
   variable mimeTypes
   if {[tk_messageBox -icon question -type yesno -title "Please Respond" \
      -message "Are you sure you want to associate $ftype files with SnackAmp"]!="yes"} {
      return
   }

   package require registry
   set conf($ftype,default) 1
   set conf($ftype,dde) 1
   set conf($ftype,assoc) $ftype
   set resKey "HKEY_CURRENT_USER\\Software\\SnackAmp\\Restore"
   foreach key [array names conf *,assoc] {
      set ext [lindex [split $key ,] 0]
      # Set the class name of the file type
      if {[catch {registry get [join [list HKEY_CLASSES_ROOT $ftype] \\] ""} class]} {
         set class "[string trimleft $ext .]file"
         registry set [join [list HKEY_CLASSES_ROOT $ext] \\] "" $class
         registry set [join [list HKEY_CLASSES_ROOT $class] \\] "" "[string toupper [string trimleft $ext .]] sound format"
      }
      # Save the existing icon
      if {![catch {registry get HKEY_CLASSES_ROOT\\$class\\DefaultIcon ""} file]} {
         registry set $resKey\\$class\\DefaultIcon "" $file
      }
      if {![catch {registry get HKEY_CLASSES_ROOT\\$class\\shell ""} value]} {
         registry set $resKey\\$class\\shell "" $value
      }
      # Register the default icon for this class to SnackAmp (only gets icon if wrapped)
      registry set HKEY_CLASSES_ROOT\\$class\\DefaultIcon "" "$::TA(appname)"
      set base HKEY_CLASSES_ROOT\\$class\\shell
      ##
      # Add menu handlers (These can just be removed later)
      #
      # Add son(s) to playlist
      if {$conf($ext,default)} {
         registry set $base "" snAmp.append
      }
      registry set $base\\snAmp.append "" "&Queue to SnackAmp"
      registry set $base\\snAmp.append\\command "" "$::TA(appname)"
      if {$conf($ext,dde)} {
         registry set $base\\snAmp.append\\ddeexec "" "PlayFile {%1} 1"
         registry set $base\\snAmp.append\\ddeexec\\Application "" TclEval
         registry set $base\\snAmp.append\\ddeexec\\Topic "" snackAmp
      }
      # Play the file
      if {$conf($ext,default)} {
         registry set $base "" snAmp.play
      }
      registry set $base\\snAmp.play "" "&Play using SnackAmp"
      registry set $base\\snAmp.play\\command "" "$::TA(appname)"
      if {$conf($ext,dde)} {
         registry set $base\\snAmp.play\\ddeexec "" "PlayFile {%1}"
         registry set $base\\snAmp.play\\ddeexec\\Application "" TclEval
         registry set $base\\snAmp.play\\ddeexec\\Topic "" snackAmp
      }
      foreach {Ext type} $mimeTypes {
         if {$ext==$Ext} {
            if {[catch {registerMIME $ext $type} result]} {
               saLog "registerMIME:$result"
            }
         }
      }
   }
}
#------------------------------------------------------------------------------
# registerMIME: Register a mime type for IE handling
#------------------------------------------------------------------------------
proc fileAssociation::registerMIME {ext type} {
   set loc "HKEY_CLASSES_ROOT\\MIME\\Database\\Content Type\\$type"
   # Rename any CLSID handlers to allow extensions to be used
   if {[catch {registry set $loc "Extension" $ext} result]} {
      saLog "registerMIME: $result"
   } else {
      saLog "Registered '$ext' as type '$type'"
   }
   if {![catch {registry get $loc "CLSID"} value]} {
      registry set $loc "CLSID.sa" $value
      registry delete $loc "CLSID"
   }

}
#------------------------------------------------------------------------------
# Function   : unRegisterMF
# Description: UnRegister some file type on windows for launching snackamp
#------------------------------------------------------------------------------
proc fileAssociation::unRegisterMF {{ftype .mp3}} {
   Trace
   if {[catch {registry get [join [list HKEY_CLASSES_ROOT $ftype] \\] ""} class]} {
      puts "Not registered to snackAmp"
      return
   }
   if {[tk_messageBox -icon question -type yesno -title "Please Respond" \
      -message "Are you sure you want to restpre $ftype associations to their previous settings?"]!="yes"} {
      return
   }
   set resKey "HKEY_CURRENT_USER\\Software\\SnackAmp\\Restore"
   # Restore single entries for
   #HKEY_CLASSES_ROOT .mp3
   #HKEY_CLASSES_ROOT mp3file
   # Remove snAmp.append handlers
   registry delete HKEY_CLASSES_ROOT\\$class\\shell\\snAmp.append
   registry delete HKEY_CLASSES_ROOT\\$class\\shell\\snAmp.play
   # Restore old Icon
   if {![catch {registry get $resKey\\$class\\DefaultIcon ""} file]} {
      registry set HKEY_CLASSES_ROOT\\$class\\DefaultIcon "" $file
      registry delete $resKey\\$class\\DefaultIcon
   }
   if {![catch {registry get $resKey\\$class\\shell ""} value]} {
      registry set HKEY_CLASSES_ROOT\\$class\\shell "" $value
      registry delete $resKey\\$class\\shell
   }
   foreach {ext type} $mimeTypes {
      if {$ext==$ftype} {
         if {[catch {unregisterMIME $ext $type} result]} {
            saLog "unregisterMIME:$result"
         }
      }
   }
}
#------------------------------------------------------------------------------
# unregisterMIME: Restore any previous CLSID values
#------------------------------------------------------------------------------
proc fileAssociation::unregisterMIME {ext type} {
   set loc "HKEY_CLASSES_ROOT\\MIME\\Database\\Content Type\\$type"
   # Rename any CLSID handlers to allow extensions to be used
   if {![catch {registry get $loc "CLSID.sa"} value]} {
      registry set $loc "CLSID" $value
      registry delete $loc "CLSID.sa"
   }

}
# Example: Let's define a file type, .acd, that's an 'address card file.' We'll use
# a class name of AddrCard and use CardFile.tcl to open it:
#
#     registerFileType .acd AddrCard "Address Card File" \
#         [file join $scriptDir CardFile.tcl]
#
#----------------------------------------------------------------------
#
# RegisterFileType::RegisterFileType --
#
#       Register a file type on Windows
#
# Author:
#       Kevin Kenny <kennykb@acm.org>.
#       Last revised: 27 Nov 2000, 22:35 UTC
#
# Parameters:
#       extension -- Extension (e.g., .tcl) of the new type
#                    being registered.
#       className -- Class name (e.g., "tclfile") of the new type
#       textName  -- Textual name (e.g. "Tcl Script") of the
#                    new type.
#       script    -- Name of the file containing a Tcl script
#                    to run when a file of the given type is
#                    opened.  The script will receive the name
#                    of the file in [lindex $argv 0].
#
# Options:
#       -icon FILENAME,NUMBER
#               Set the icon for files of the new type
#               to be the NUMBER'th icon in the given file.
#               The file must be a full path name.
#       -mimetype TYPE
#               Set the MIME type corresponding to the new
#               file type to the specified string.
#       -new BOOLEAN
#               If BOOLEAN is true, set things up so that
#               the new file type appears in the "New" menu
#               in the Explorer and the system tray.
#       -text BOOLEAN
#               If BOOLEAN is true, the new file type contains
#               plain ASCII text of some sort.  Set the
#               Edit and Print actions to open and print
#               ASCII files.
#
# Results:
#       None.
#
# Side effects:
#       Adds the following keys to the system registry:
#
#       HKEY_CLASSES_ROOT
#         (Extension)           (Default value)         ClassName
#                               "Content Type"          MimeType        [1]
#           ShellNew            "NullFile"              ""              [2]
#         (ClassName)           (Default value)         TextName
#           DefaultIcon         (Default value)         IconName,#      [3]
#           Shell
#             Open
#               command         (Default value)         -SEE BELOW-
#             Edit
#               command         (Default value)         -SEE BELOW-     [4]
#             Print
#               command         (Default value)         -SEE BELOW-     [4]
#         MIME
#           Database
#             Content Type
#               (MimeType)      (Default value)         Extension       [1]
#
#       [1] These values are added only if the -mimetype option is used.
#       [2] This value is added only if the -new option is true.
#       [3] This value is added only if the -icon option is used.
#       [4] These values are added only if the -text option is true.
#
#       The command to open the file consists of three arguments.
#       The first is the name of the current Tcl executable.  The
#       second is the script name, and the third is "%1", which causes
#       the target file to be passed as a command-line argument.
#       The edit command is the command that opens text files, and the
#       print command is the command that prints text files.
#
#----------------------------------------------------------------------
namespace eval RegisterFileType {
}
proc RegisterFileType::RegisterFileType {
   extension className textName script args
} {

   package require registry

   # extPath is the class path for the file's extension

   set extPath HKEY_CLASSES_ROOT\\$extension
   registry set $extPath {} $className sz

   # classPath is the class path for the file's class

   set classPath HKEY_CLASSES_ROOT\\$className
   registry set $classPath {} $textName sz

   # shellPath is the shell key within classPath

   set shellPath $classPath\\Shell

   # Set up the 'Open' action

   set openCommand {}
   append openCommand \" \
         [file nativename [info nameofexecutable]] \
         \" { } \" [file nativename $script] \" { } \"%1\"
   registry set $shellPath\\open\\command {} $openCommand sz

   # Process optional args

   foreach {key val} $args {
      switch -exact -- $key {

         -mimetype {

            # Set up the handler for the MIME content type,
            # and add the content type item to the database

            registry set $extPath "Content Type" $val sz
            set mimeDbPath \
                  "HKEY_CLASSES_ROOT\\MIME\\Database"
            append mimeDbPath "\\Content Type\\" $val
            registry set $mimeDbPath Extension \
                  $extension sz
         }

         -icon {

            # Add the file icon to the shell database

            if {![regexp {^(.*),([^,]*)} $val \
                     junk file icon]} {
               error "-icon option requires\
                     fileName,iconNumber"
            }
            registry set $classPath\\DefaultIcon {} \
                  [file nativename $file],$icon sz
         }

         -text {
            if {$val} {

               # Copy the Print action for text files
               # into the Print action for the new type

               set textPath \
                     HKEY_CLASSES_ROOT\\txtfile\\Shell
               if {![catch {
                     registry get \
                           $textPath\\print\\command {}
                  } pCmd]} {
                  registry set \
                        $shellPath\\print\\command \
                        {} $pCmd sz
                  registry set \
                        $shellPath\\print {} \
                        &Print sz

               }

               # Copy the Open action for text files
               # into the Edit action for the new type.

               if {![catch {
                     registry get \
                           $textPath\\open\\command {}
                  } eCmd]} {
                  registry set \
                        $shellPath\\edit\\command \
                        {} $eCmd sz
                  registry set \
                        $shellPath\\edit {} \
                        &Edit sz
               }
            }
         }

         -new {
            if {$val} {

               # Add the 'NullFile' action to the
               # shell's New menu

               registry set $extPath\\ShellNew NullFile \
                     {} sz
            }
         }

         default {
            error "unknown option $key, must be -icon,\
                  -mimetype, -new or -text"
         }
      }
   }
}


package provide fileAssociation 1.0
