#------------------------------------------------------------------------------
# snackAmp - freedb package 1.0 for snackAmp
# provides dynamic http applications for SnackAmp
# web server.
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2005 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
namespace eval freedb {
   variable server
   upvar #0 snackAmpSettings(freedb) server
   # link into here from snackAmpSettings
#  http://at.freedb.org:80/~cddb/cddb.cgi (Vienna, Austria)
#  http://au.freedb.org:80/~cddb/cddb.cgi (Sydney, Australia)
#  http://ca.freedb.org:80/~cddb/cddb.cgi (Winnipeg, Manitoba, Canada)
#  http://ca2.freedb.org:80/~cddb/cddb.cgi (Montreal, Quebec, Canada)
#  http://de2.freedb.org:80/~cddb/cddb.cgi (Berlin, Germany)
#  http://http://es.freedb.org:80/~cddb/cddb.cgi (Madrid, Spain)
#  http://fi.freedb.org:80/~cddb/cddb.cgi (Tampere, Finland)
#  http://ru.freedb.org:80/~cddb/cddb.cgi (Saint-Petersburg, Russia)
#  http://uk.freedb.org:80/~cddb/cddb.cgi (London, UK)
#  http://us.freedb.org:80/~cddb/cddb.cgi (San Jose, California, USA)
   # Note: the 150 and 75 frame offsets are from mp3split, not sure
   # if they are perfect for ripped data
   variable initFrameOffset 150
   variable Frameoffset 75 ;# assume 75 frames per second
   variable Application "SnackAmp"
   variable ApplicationVersion 3.x
}
#----------------------------------------------------------------------------
# Return a number of albums for a given set of tracks
#----------------------------------------------------------------------------
proc freedb::albumList {tracks} {
   variable server
   variable Application
   variable ApplicationVersion
   set hassome 0
   set found [list]

   # My discid doesn't match the one from freedb in the checksum, however returns are good.
   set cddb_discid [freedb::discid $tracks]
   set url "$server?cmd=cddb+query+[ENCODE $cddb_discid]&hello=noname+localhost+$Application+$ApplicationVersion&proto=6 HTTP/1.1\r\nHost: freedb.freedb.org:80\r\nConnection: close"
   set contents [freedb::getUrlBinary $url]

   # Parse the contents
   set contents [split $contents \n]
   # State machine to read in the valid rows
   foreach {row} $contents {
      switch -glob -- $row {
         "." {
            set hassome 0
         }
         "202 *" {
            # not in database
            set hassome 0
            break
         }
         "211 *" {
            # An inexact match, return following rows
            set hassome 1
         }
         "200 *" {
            # An exact match, return this row
            set hassome 0
            lappend found [string trimleft $row "200 "]
         }
      	"401 *" {
            saLog "Specified CDDB entry not found."
            break
         }
      	"403 *" {
            saLog "Database entry is corrupt."
            break
         }
      	"409 *" {
            saLog "No handshake."
            break
         }
         default {
            if {$hassome} {
               lappend found $row
            }
         }
      }
   }
   return $found
}
#----------------------------------------------------------------------------
# Query all album/artist data from a freedb search
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# !!!!!!! This is all very tenuous and fragle and prone to change !!!!!!!
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#----------------------------------------------------------------------------
proc freedb::albumListHTML {Artist Album} {
   Trace
   if {[string length $Artist]} {
      set Data(Artist) $Artist
   }
   if {[string length $Album]} {
      set Data(Album) $Album
   }
   set contents [browser::cddbHtml Data]
   ##
   # For each record returned, query the details on the record
   #
   set records [list]

#set fod [open ~/freedb.txt w]
#puts $fod $contents
#close $fod
   # The titles look like this
   # title="Various Artists / A Knight's Tale Movie Soundtrack"
   # The details for each album are below that and look like this
   # url pulls raw page
   #                cat            url   discid
   # <b>Disc-ID:</b>(.+?)<a href="(.+?)">(.+?)</a>
   set _titles [regexp -line -all -inline -- {title=\"(.+?)\"} $contents]   ;#"
   set titles [list]
   # kludge, to ignore the no cd message in between titles
   foreach {tit} $_titles {
      if {![string match -nocase "*CD available*" $tit]} {
         lappend titles $tit
      }
   }
   # translate quotes
   set titles [string map {&quot; \" &acute; ` &amp; &} $titles] ;#"
   set discinfo [regexp -line -all -inline -- {<b>Disc-ID:</b>(.+?)<a href=\"(.+?)\">(.+?)</a>} $contents]   ;#"
   set lasttitle ""
   set i 0
#puts "titles([llength $titles])=$titles"
#puts "\ndiscinfo([llength $discinfo])=$discinfo"
   foreach {xxx title} $titles {xx cat url id} $discinfo {
#puts "->$xxx $title"
      #country ba0e080d Williams, Lucinda / World Without Tears
      if {[regexp -- {font} $title]} {
         set utitle "$lasttitle ([incr i])"
      } else {
         set i 0
         set utitle $title
         set lasttitle $title
      }
      ;# Only want first part of genre (cat)
      set cat [lindex $cat 0]
      # Note we now have the exact url to seach for
#puts "'$cat' '$id' '$utitle'"
                     ;# rock 7908110a The Shins / Chutes too Narrow
      lappend records "$cat $id $utitle"
      # todo: Could improve by parsing and putting into a proper list here
   }
   return $records
}

#----------------------------------------------------------------------------
# Return the details of a particular album
#----------------------------------------------------------------------------
proc freedb::albumDetails {albumTag _Array} {
   upvar $_Array Array
   variable server
   set category [lindex $albumTag 0]
   set discid [lindex $albumTag 1]
   set url "$server?cmd=cddb+read+$category+$discid+&hello=noname+localhost+SnackAmp+3.1&proto=6 HTTP/1.1\nHost: freedb.freedb.org:80\nConnection: close"
   set contents [freedb::getUrlBinary $url]
   set contents [split $contents \n]
   set Album ""
   set hassome 0
   set offsets 0
   set dlen 3
   set offsetList [list]
   # State machine to read in the valid tracks
   foreach {row} $contents {
#puts $row
      switch -glob -- $row {
         "*Track frame offsets*" {
            set offsets 1
         }
         "DTITLE*" {
            foreach {tag Title} [split $row =] {break}
            set Array(Title) $Title
            set Album $Title
            set hassome 1
         }
         "DYEAR*" {
            foreach {tag Year} [split $row =] {break}
            set Array(Year) $Year
         }
         "DGENRE*" {
            foreach {tag Genre} [split $row =] {break}
            set Array(Genre) $Genre
         }
         "TTITLE*" {
            set row [string trimleft $row "TTITLE"]
            regexp {(.+?)=(.*$)} $row -> Track Title
            incr Track
            append Array(Track,$Track) $Title
            set hassome 1
         }
         "*Disc length*" {
            set dlen [lindex $row 3]
         }
         "#*" {
            if {$offsets} {
               # Grab offsets to compute track lengths
               set oft [lindex $row 1]
               if {[string is integer -strict $oft]} {
                  lappend offsetList [lindex $row 1]
               }  else {
                  set offsets 0
               }

            }
         }
      }
   }
   # figure out the track lengths, 75 frames per second is fixed, start with disk length
   #todo bulletproof, handle errors
   lappend offsetList [expr {$dlen*75}]
   set items [llength $offsetList]
   for {set i 0} {$i < ($items-1)} {incr i} {
      set Track [expr {$i+1}]
      set one [lindex $offsetList $i]
      set two [lindex $offsetList [expr {$i+1}]]
      set length [expr {($two-$one)/75}]
      set Array(Length,$Track) $length
   }
   return $Album
}
#----------------------------------------------------------------------------
# Compute the freedb discid from the tracks
#----------------------------------------------------------------------------
proc freedb::discid {tracks} {
   variable initFrameOffset
   variable Frameoffset ;#75 frames = 1 sec
   set disc_length 0
   set checksum 0
   set num_tracks 0
   set total_seconds $initFrameOffset
   set query_string ""
   set secstring ""
   foreach {fname} [lsort -dictionary $tracks] {
      incr num_tracks
      if {[db::lookup $fname Data 1] > 0} {
         set secs [expr {round($Data(Duration))}]
         append secstring $secs
         append query_string " $total_seconds"
         incr disc_length $secs
         incr total_seconds [expr {$secs*$Frameoffset}]
      }
   }
   # Checksum is total of all digits in seconds
   foreach {c} [split $secstring {}] {incr checksum $c}
   set cddb_discid [expr {($checksum % 0xFF) << 24 | ($disc_length << 8) | $num_tracks}]
   set cddb_discid [format %08x $cddb_discid]
   return "$cddb_discid $num_tracks$query_string $disc_length"
}
#----------------------------------------------------------------------------
# Query the web for some url and return the entire page
# works automatically if a proxy has already been established
#----------------------------------------------------------------------------
proc freedb::getUrlBinary {url} {
   package require http 2.4
   if {[catch {
      set token [::http::geturl $url]
      set data [::http::data $token]
      ::http::cleanup $token
   } result] } then {
      tk_messageBox -type ok -icon warning -message "$result\nHttp error, do you have a proxy setup?"
      set data "Http Error"
   }
   return $data
}

package provide freedb 1.0
