#------------------------------------------------------------------------------
# gpUtils 1.0
#
# General Purpose functions for snackAmp Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/gpUtils.tcl,v 1.88 2007/10/01 01:35:28 wilkason Exp $
}
#------------------------------------------------------------------------------
# Function   : K
# Description: K Combinator to work around tcl internal copy on lrepace/linsert
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc K {a b} {set a}
interp alias {} slen {}  string length
##
# Pop an item off a list, left or right
#
proc lpop {how listName} {
   upvar $listName list
   switch -- $how {
      "right" {
         K [lindex $list end] [set list [lrange $list 0 end-1]]
      }
      "left" {
         K [lindex $list 0] [set list [lrange $list 1 end]]
      }
      default {
         return -code error "lpop right|left listName"
      }
   }
}
##
# Pop an item onto a list, left or right
#
proc lpush {how listName item} {
   upvar $listName list
   switch -- $how {
      "right" {
         lappend list $item
      }
      "left" {
         set list [linsert [K $list [set list ""]] 0 $item]
      }
      default {
         return -code error "lpush right|left listName item"
      }
   }
}

#------------------------------------------------------------------------------
# Remove some list entries inplace
#------------------------------------------------------------------------------
proc lipreplace {listName ind1 ind2 v1 v2} {
   upvar $listName list
   set list [lreplace [K $list [set list {}]] $ind1 $ind2 $v1 $v2]
}
#------------------------------------------------------------------------------
# Function   : ldelete
# Description: Remove an entry from a list by index
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc ldelete {_list start {end ""}} {
   upvar $_list list
   if {$end eq ""} {
      set list [lreplace [K $list [set list {}]] $start $start]
   } else {
      set list [lreplace [K $list [set list {}]] $start $end]
   }
}
#----------------------------------------------------------------------------
# Extract the matching items from a list, using globbing
#----------------------------------------------------------------------------
proc lextract {_list value} {
   upvar $_list list
   set result [list]
   foreach {loc} [lsort -decreasing [lsearch -glob -all $list $value]] {
      lappend result [lindex $list $loc]
      set list [lreplace [K $list [set list {}]] $loc $loc]
   }
   return $result
}

#------------------------------------------------------------------------------
# In place insert, limited to one insert value
#------------------------------------------------------------------------------
proc lipinsert {_list index value} {
   upvar $_list list
   set list [linsert [K $list [set list ""]] $index $value]
}
proc lipinsert1 {_list index value} {
   upvar $_list list
   set list [linsert [lindex [list $list [set list ""]] 0] $index $value]
}
proc ttt {} {
   set LIST [list]; time {lappend LIST A} 5000
   puts "1 [time {lipinsert LIST 40 B} 5000] [llength $LIST]"
   set LIST [list]; time {lappend LIST A} 5000
   puts "2 [time {lipinsert1 LIST 40 B} 5000] [llength $LIST]"
}
#------------------------------------------------------------------------------
# Flip a list around
#------------------------------------------------------------------------------
proc lflip {L} {
   set res [list]
   set i [llength $L]
   while {$i} {lappend res [lindex $L [incr i -1]]} ;# rmax
   return $res
 } ;# RS,

#------------------------------------------------------------------------------
#
#------------------------------------------------------------------------------
proc diag {{how on}} {
  switch -- $how {
     "on" {set $::snackAmpSettings(Trace) 1}
     "off" {set $::snackAmpSettings(Trace) 0}
     default {puts "diag on|off"}
  }
}
#------------------------------------------------------------------------------
# Reverse a list (very fast with new lset command)
#------------------------------------------------------------------------------
# proc lipreverse {listName} {
#    upvar $listName list
#    set len  [llength $list]
#    set tot [expr {$len/2}]
#    for {set i 0} {$i < $tot} {incr i} {
#       set end [expr {$len-$i-1}]
#       set head [lindex $list $i]
#       lset list $i [lindex $list $end]
#       lset list $end $head
#    }
#    return $list
# }
#------------------------------------------------------------------------------
# Function   : shuffle
# Description: Shuffle a list, used to randomize play list
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc shuffle {list} {
   #expr {srand([clock clicks])}
   set len [llength $list]
   set len2 $len
   incr len -1
   for {set i 0} {$i < $len} {incr i} {
      set n [expr {int($i + $len2 * rand())}]
      incr len2 -1

      # Swap elements at i & n
      set temp [lindex $list $i]
      lset list $i [lindex $list $n]
      lset list $n $temp
   }
   return $list
}
##
# Morph a list (i.e. slightly shuffle it)
#
proc morph {list {perc 100}} {
   if {$perc <= 0 || $perc > 100} {
      return $list
   }
   set modulo [expr {int(100/$perc)}]
   #expr {srand([clock clicks])}
   set len [llength $list]
   set len2 $len
   incr len -1
   for {set i 0} {$i < $len} {incr i} {
      incr len2 -1
      if {$i%$modulo==0} {
         set n [expr {int($i + $len2 * rand())}]
         # Swap elements at i & n
         set temp [lindex $list $i]
         lset list $i [lindex $list $n]
         lset list $n $temp
      }
   }
   return $list
}
proc lipmorph {_list {perc 100}} {
   upvar $_list list
   if {$perc <= 0 || $perc > 100} {
      return
   }
   set modulo [expr {int(100/$perc)}]
   #expr {srand([clock clicks])}
   set len [llength $list]
   set len2 $len
   incr len -1
   for {set i 0} {$i < $len} {incr i} {
      incr len2 -1
      if {$i%$modulo==0} {
         set n [expr {int($i + $len2 * rand())}]
         # Swap elements at i & n
         set temp [lindex $list $i]
         lset list $i [lindex $list $n]
         lset list $n $temp
      }
   }
   return
}

#------------------------------------------------------------------------------
# Function   : orderList
# Description: Return a list ordered by the sub list as the second argument
# Author     : Tom Wilkason
# Date       : 8/5/2003
#------------------------------------------------------------------------------
proc orderList {bigList subList} {
   set results [list]
   set eat 0
   foreach {element} $bigList {
      set subPos [lsearch -exact $subList $element]
      if {$subPos>=0} {
         lappend results [lindex $subList $eat]
         incr eat
      } else {
         lappend results $element
      }
   }
   return $results
}
#------------------------------------------------------------------------------
# Function   : shuffleByFolder
# Description: Shuffle a list by folder, keep folder contents sorted
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc shuffleByFolder {list {cmd lsort}} {
   set tmp [list]
   # Create array of folders containing the files within
   foreach item $list {
      lappend fcontents([file dirname $item]) $item
   }
   # Shuffle the folder names
   foreach item [array names fcontents] {
      lappend tmp [list $item [expr {rand()}]]
   }
   set res [list]
   # For each folder, append the contents sorted
   foreach item [lsort -real -index 1 $tmp] {
      foreach {orig num} $item {
         foreach {file} [$cmd $fcontents($orig)] {
            lappend res $file
         }
      }
   }
   return $res
}
#------------------------------------------------------------------------------
# Function   : ldsort
# Description: Lsort by dictionary
# Author     : Tom Wilkason
# Date       : 12/28/2002
#------------------------------------------------------------------------------
proc ldsort {list} {
   return [lsort -dictionary $list]
}
proc sortedSearch {list item} {
   return [lsearch -exact -sorted $list $item]
}
proc Unset {_var} {
   upvar $_var var
   unset -nocomplain var
}
proc lsearchAll {List Item} {
   return [lsearch -all -exact $List $Item]
}

#------------------------------------------------------------------------------
# Append an item to the list if it is not already there
#------------------------------------------------------------------------------
proc lappendUnqiue {_listVar item} {
   upvar $_listVar listVar
   if {![info exists listVar] || [lsearch -exact $listVar $item]==-1} {
      lappend listVar $item
   }
}
#------------------------------------------------------------------------------
# Function   : commify
# Description: Insert commas into an integer for display
# Author     : Jeff Hobbs
#------------------------------------------------------------------------------
# proc commify {num {sep ,}} {
#     while {[regsub {^([-+]?\d+)(\d\d\d)} $num "\\1$sep\\2" num]} {}
#     return $num
# }
# This is a wee bit faster
#proc commify {number} {
#   regsub -all {\d(?=(\d{3})+($|\.))} $number {\0,}
#}

#------------------------------------------------------------------------------
# This is actually 2x faster, but limted to integers
# Why? The regexp engine in Tcl is dog slow
#------------------------------------------------------------------------------
proc commify {num} {
   set rem ""
   if {[string is integer -strict $num]} {
      if {$num == 0} {
         return 0
      }
      while {$num>0} {
         set rem ",[format %3.3d [expr {$num%1000}]]$rem"
         set num [expr {$num/1000}]
      }
   }
   return [string trimleft [string trimleft $rem ,] "0"]
}
#------------------------------------------------------------------------------
# Function   : filesSize
# Description: Return a short version of the file size
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc fileSize {val} {
   # val assumed in MB
   if {$val<1024} {
      return "[format %.2f $val] MB"
   } elseif {$val < 1048576} {
      return "[format %.2f [expr {$val/1024.}]] GB"
   } else {
      return "[format %.2f [expr {$val/1048576.}]] TB"
   }
}

#------------------------------------------------------------------------------
# Windows doesn't support the "active" button border, so add some more
#------------------------------------------------------------------------------
proc extraPad {base} {
   if {$::tcl_platform(platform)=="windows"} {
      set extPad 5
   } else {
      set extPad 0
   }
   return [expr {$extPad+$base}]
}
# ----------------------------------------------------------------------------
# Return a glob pattern of subs depth, must use eval glob.
# ----------------------------------------------------------------------------
proc globAllPattern {{subs 0}} {
   set pattern *
   incr subs
   for {set i 1} {$i < $subs} {incr i} {
      append pattern " [join [string repeat {* } [expr {$i+1}] ] /]"
   }
   return $pattern
}
#------------------------------------------------------------------------------
# Properly case a sentence, not perferfect
# ignore specific words in TA(lcwords)
#------------------------------------------------------------------------------
proc stringTitle {String} {
   global TA
   set consider 0
   upvar #0 lcwords TA(lcwords)
   set result ""
   foreach {word} [split $String] {
      # skip embedded blanks
      if {$word eq ""} {continue}
      # Find first alpha char of word to titleize
      set i 0
      foreach {c} [split $word ""] {
         if {[string is alpha $c]} {
            break
         }
         incr i
      }
      # TODO: Handle embedded puncuation in words e.g. John O'Conner
      # Check for roman numerial
      if {[regexp -inline -- {[IVXLivxl.]+} $word] eq $word} {
         set word [string toupper $word]
      # Skip words we wan't kept lower (after first real word)
      } elseif {$consider && ([lsearch -exact -sorted $TA(lcwords) $word]>-1)} {
         set word [string tolower $word]
      } else {
         set word [string totitle $word $i end]
         #set word [wordTitle $word]
      }
      # If this is an alpha, flag next for consideration to case
      if {[string is alpha $word]} {
         set consider 1
      } else {
         set consider 0
      }
      append result $word " "
   }
   return [string trim $result]
}
#----------------------------------------------------------------------------
# Improved title, in work
# won't separate out 1-line
# Caps everything after ', not just when one letter is before
#----------------------------------------------------------------------------
proc stringTitlexx {line} {
   global TA
   upvar #0 lcwords TA(lcwords)
   set pchars {[ \.']}
   set nchars {[^ \.']}
   set res [regexp -inline -all -indices -- {[ \.'-][^ \.'-]+} " $line"]
   foreach pair $res {
      foreach {s e} $pair {
         incr e -1
         set word [string range $line $s $e]
         if {[regexp -inline -- {[IVXLivxl.]+} $word] eq $word} {
            # Uppercase Roman Numerals
            set word [string toupper $word]
         } elseif {[lsearch -exact -sorted $TA(lcwords) $word]>-1} {
            # Skip words we wan't kept lower (after first real word)
            set word [string tolower $word]
         } else {
            # make this word a proper title
            set word [string totitle $word]
         }
         # insert back into original line
         set line [string replace $line $s $e $word]
      }
   }
   return [string trim $line]
}
#------------------------------------------------------------------------------
# Function   : chooseFile
# Description: Bring up file selector to select a file. Assign it to the var
# Author     : Laurent Duperval
# Date       : 1/12/2002
#------------------------------------------------------------------------------
proc chooseFile {{start .} {parent .}} {
   if {$start == "."} {
      set start [pwd]
   }
   set choice [tk_getOpenFile -initialdir $start -parent $parent]
   if {$choice == {}} {
      return {}
   } else {
      return [casedDirName $choice 1]
   }
}
#------------------------------------------------------------------------------
# Function   : cleanPathList
# Description: Correct normal goofs in a path name and clean it up
#              so it looks like a unix path name. Doesn't verify the
#              file actually exists.
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc cleanPathList {_Var {check 0}} {
   upvar $_Var Path
   if {$Path != {}} {
      set result [list]
      foreach {token} $Path {
         if {$token != {}} {
            lappend result [casedDirName $token $check]
         }
      }
      set Path $result
   }
   return ""
}
#------------------------------------------------------------------------------
# Function   : casedDirName
# Description: Correct the file case to match what is actually on the hard
#              disk rather than what was input. Windows only issue.
#              Note: Will not work with UNC because glob wont
# Author     : Tom Wilkason
# Date       : 3/12/2001
#------------------------------------------------------------------------------
proc casedDirName {choice {verify 0}} {
   if {$choice != {}} {
      set choice [eval file join [file split [string trim $choice]]]
      if {$verify && ($::tcl_platform(platform) == "windows")} {
         while {1} {
            if {[catch {set choice [file attributes $choice -longname]} message]} {
               set answer [getEntryDialog "'$choice' cannot be found, please correct it or 'cancel'" choice "Please Correct Entry"]
               if {$answer == "cancel"} {
                  break
               }
            } else {
               break
            }
         }
      }
   }
   return $choice
}
#------------------------------------------------------------------------------
# Function   : chooseDirectory
# Description: Wrapper for tk_chooseDirectory so it doesn't return
#              uppercase names on Windows (yuck!)
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc chooseDirectory {{start .} {parent .} {title "Please Choose a Directory"}} {
   if {$start == "."} {
      set start [pwd]
   }
   set choice [tk_chooseDirectory -initialdir $start -parent $parent -title $title]
   if {$choice == {}} {
      return {}
   } else {
      return [casedDirName $choice 1]
   }
}
#------------------------------------------------------------------------------
# Function   : cleanKeyName
# Description: Clean up a tk key name to a shortened version
#------------------------------------------------------------------------------
proc cleanKeyName {name} {
   set name [string map {< "" > "" Control Ctrl Key- ""} $name]
}
#------------------------------------------------------------------------------
# Return the short name of a file
#------------------------------------------------------------------------------
proc nameToShort {File} {
   global snackAmpSettings TA
   set fcar /
   if {$snackAmpSettings(ListDepth) > 0} {
      # File joing doesnt work due to single list
      set File [string map "// $fcar" $File]      ;# remove UNC stuff
      set name [join [lrange [file split $File] end-$snackAmpSettings(ListDepth) end] $TA(spcar)]
   } else {
      set name [file tail $File]
   }
   if {$::TA(noTail)} {
      set name [file rootname $name]
   }
   return $name
}
#------------------------------------------------------------------------------
# Namespace to watch textvariables linked to an entry box and
# color them based on the data changed
# Binding for ctrl-z to undo changes
#------------------------------------------------------------------------------
namespace eval widgetWatch {
   variable Widgets
   variable wwMenu .wwMenu
   if {![winfo exists $wwMenu]} {
      menu $wwMenu -tearoff 0
      event add <<wwUndo>> <Control-Key-z>
      event add <<wwRedo>> <Control-Key-y>
      # TFW:Build on this to cut/copy/paste, should be restricted to widgets like entries
      foreach {Text Event} {
         Cut   <<Cut>>
         Copy  <<Copy>>
         Paste <<Paste>>
         Undo  <<wwUndo>>
         Redo  <<wwRedo>>
      } {
         $wwMenu add command -label $Text \
            -command [list widgetWatch::menuCallback $Event] \
            -accelerator [lindex [cleanKeyName [event info $Event]] 0]
      }
   }
}
#------------------------------------------------------------------------------
# Menu callback to generate the appropriate event
#------------------------------------------------------------------------------
proc widgetWatch::menuCallback {Event} {
   event generate [focus] $Event
}
#------------------------------------------------------------------------------
# User callable routine to set a watch on an
# entry widget text variable, sets the
# text background pink when the data changes
#------------------------------------------------------------------------------
proc widgetWatch::setWatch {widget _Element} {
   if {[catch {
      upvar #0 $_Element element
      set cleanElement [string trimleft $_Element :]
      variable Widgets
      set Widgets($widget) $widget
      set Widgets($widget,Orig) $element
      set Widgets($widget,Redo) $element
      set Widgets($widget,bg) [$widget cget -background]
      trace variable $_Element w [list widgetWatch::setCallback $widget]
      trace variable $_Element u [list widgetWatch::resetCallback $widget]
      # Undo/Redo keys
      bind $widget <<wwUndo>> [list widgetWatch::undo $widget $_Element]
      bind $widget <<wwRedo>> [list widgetWatch::redo $widget $_Element]
      bind $widget <3> [list tk_popup $widgetWatch::wwMenu %X %Y]
   } result] } then {
      puts "setWatch $result\n$::errorInfo"
   }
}
#------------------------------------------------------------------------------
# Undo a setting back to the original value
#------------------------------------------------------------------------------
proc widgetWatch::undo {widget _Glob} {
   variable Widgets
   set Widgets($widget,Redo) [set $_Glob]
   set $_Glob $Widgets($widget,Orig)
}
#------------------------------------------------------------------------------
# Redo a setting just prior to the last undo
#------------------------------------------------------------------------------
proc widgetWatch::redo {widget _Glob} {
   variable Widgets
   set $_Glob $Widgets($widget,Redo)
}
#------------------------------------------------------------------------------
# Define the callback to trap modifications to variables
#------------------------------------------------------------------------------
proc widgetWatch::setCallback {widget _Glob item args} {
   variable Widgets
   if {[string length $item]} {
      if {[catch {
         upvar $_Glob GlobArray
         set value $GlobArray($item)
         set ovalue $Widgets($widget,Orig)
         if {[string equal $value $ovalue]} {
            $widget configure -background $Widgets($widget,bg)
         } else {
            $widget configure -background pink
         }
      } result] } then {
         puts "setCallback $result\n$::errorInfo"
      }
   }
}
#------------------------------------------------------------------------------
# Callback when the variable is unset, this
# removes local copies data
#------------------------------------------------------------------------------
proc widgetWatch::resetCallback {widget _Glob item args} {
   variable Widgets
   if {[catch {
      unset Widgets($widget)
      unset Widgets($widget,bg)
      unset Widgets($widget,Orig)
      unset Widgets($widget,Redo)
   } result] } then {
      puts "resetCallback $result\n$::errorInfo"
   }
}
#------------------------------------------------------------------------------
# Hand type in to get memory usage of globals
#------------------------------------------------------------------------------
proc usage {args} {
   foreach var [lsort -dictionary [info globals]] {
      upvar $var Var
      if {[array exists ::$var]} {
         set names [llength [array names Var]]
         puts "Array   : $var has [slen [array get Var]] bytes, $names elements,"
      } else {
         puts "Variable: $var has [slen $Var] bytes"
      }
   }
   foreach ns [namespace children ::] {
      foreach var [lsort [info vars ${ns}::*]] {
         upvar $var Var
         if {[array exists ::$var]} {
            set names [llength [array names Var]]
            catch {puts "Array   : $var has [slen [array get Var]] bytes, $names elements,"}
         } else {
            catch {puts "Variable: $var has [slen $Var] bytes"}
         }
      }
   }
}
#------------------------------------------------------------------------------
# Function   : safeWrite
# Description: Write a file out safely
# Author     : Tom Wilkason
# Date       : 5/5/2003
#------------------------------------------------------------------------------
proc safeWrite {file contents} {
   # Write to temp file
   set tfile [file join [file dirname $file] [clock clicks].dat]
   if {[catch {open $tfile w} OutFile]} {
      return -code error "Save Error: Unable to open '$file' for writing\n$OutFile"
   }
   if {[catch {
      fconfigure $OutFile -buffersize 200000
      puts -nonewline $OutFile $contents
      close $OutFile
   } result] } then {
      catch {close $OutFile}
      return -code error $result
   }

   # Swap Names
   if {[catch {
      # Now swap names
      if {[file exists $file]} {
         file delete -force $file
      }
      file rename $tfile $file
   } result] } then {
      file delete -force $tfile
      return -code error $result
   } else {
      return 1
   }
}
#------------------------------------------------------------------------------
# Some tools to deal with ip address, can use to mask incoming connections
#------------------------------------------------------------------------------
namespace eval iptools {
   proc myIP {args} {
   	# find out localhost's IP address
   	# courtesy David Gravereaux, Heribert Dahms
        for {set i 0} {$i < 20} {incr i} {
           if {![catch {
             set TheServer [socket -server none -myaddr [info hostname] 0]
             set MyIP [lindex [fconfigure $TheServer -sockname] 0]
             close $TheServer
           } result] } then {
             return $MyIP
           }
        }
        puts "Using default IP address for [info hostname]"
        return 127.0.0.1
   }
   proc subNetMask {ip} {
      foreach {a b c d} [split $ip .] {break}
      if {$a < 128} {
        # class A
        return 255.0.0.0
      } elseif {$a < 192} {
        # class B
        return 255.255.0.0
      } else {
        # class C
        return 255.255.255.0
      }
   }
   proc meetsMask {test ip mask} {
      if {[string match "127.0.0.*" $test]} {
         return 1
      }
      foreach {t} [split $test .] {a} [split $ip .] {A} [split $mask .] {
        if {($t&$A) != ($a&$A)} {
          return 0
        }
      }
      return 1
   }
   ##
   # Determine if a url matches a pattern
   #
   proc localUrl {url} {
      regexp -- {(http://)*(.+)\?.*} $url -> -> url
      foreach patt [split $::snackAmpSettings(localurl) "\; "] {
         if {[string match -nocase $patt $url]} {
            return 1
         }
      }
      return 0
   }
   variable myIP [myIP]
   variable defMask [subNetMask $myIP]
}
namespace eval fileutils {
   variable fileList [list]
}
#------------------------------------------------------------------------------
# Recursively scan a directory and build a list of all files
#------------------------------------------------------------------------------
proc fileutils::rglob {dir {retDir 0}} {
   variable fileList
   variable dirList
   set dirList [list]
   set fileList [list]
   set dir [file normalize $dir]
   _rglob $dir
   if {$retDir} {
      return [list [lsort $fileList] [lsort $dirList]]
   } else {
      return [lsort $fileList]
   }
}
proc fileutils::_rglob {dir} {
   variable fileList
   variable dirList
   set hasSome 0
   set dList [list]
   # get normal files
   foreach file [glob -type f -nocomplain -directory $dir -nocomplain *] {
      lappend fileList $file
   }
   # get hidden files
   foreach file [glob -nocomplain -type {f hidden} -directory $dir -nocomplain *] {
      lappend fileList $file
   }
   # get folders
   foreach file [glob -type d -nocomplain -directory $dir -nocomplain *] {
      lappend dirList $file
      lappend dList $file
   }
   # process each directory recursively
   foreach {folder} $dList {
      update
      fileutils::_rglob $folder
   }
}

#------------------------------------------------------------------------------
# Function   : hasData
# Description: Return true if a array field has data
# Author     : Tom Wilkason
# Date       : 2/7/2001
#------------------------------------------------------------------------------
proc hasData {_item} {
   upvar $_item item
   return [expr {[info exists item] && ($item ne "")}]
}

#------------------------------------------------------------------------------
#
#------------------------------------------------------------------------------
proc allcommands {{pattern *}} {
   foreach {ns} [lsort [namespace children ::]] {
      set children [info commands ${ns}::${pattern}]
      if {[llength $children]} {
         puts $ns
      }
      foreach {cmd} [lsort $children] {
         puts "   ${cmd}"
      }
   }
}
#----------------------------------------------------------------------------
# Copy over data from an input array for fields with data
#----------------------------------------------------------------------------
proc copyArrayData {_array _input} {
   upvar $_array array
   upvar $_input input
   foreach {item} [array names input] {
      if {[slen $input($item)] > 0} {
         set array($item) $input($item)
      }
   }
}
#----------------------------------------------------------------------------
# Merge in data into an array if the array has blank or non existent fields
# don't overwrite existing data.
#----------------------------------------------------------------------------
proc mergeArrayData {_array _input} {
   upvar $_array array
   upvar $_input input
   foreach {item} [array names input] {
      if {(![info exists array($item)] || [slen $array($item)]==0) && ([slen $input($item)] > 0)} {
         set array($item) $input($item)
      }
   }
}

#------------------------------------------------------------------------------
#
#------------------------------------------------------------------------------
proc ratingToNum {rating} {
   if {$rating == "--"} {
      return 0
   } else {
      return [string length $rating]
   }
}
#------------------------------------------------------------------------------
# Function   : debug
# Description: Debug output
#------------------------------------------------------------------------------
proc debug1 {Message} {
   set Stack [list]
   set depth [info level]
   set lastArgs [list]
   for {set i 1} {$i < $depth} {incr i} {
      lappend Stack [lindex [info level $i] 0]
      set lastArgs [lrange [info level $i] 1 end]
   }
   saLog "Debug:[join $Stack =>] \{$lastArgs\} : $Message"
}
proc debug {args} {
   puts "[lindex [info level 1] 0]:$args"
}

#----------------------------------------------------------------------------
# Utility function to show call stack frame
#----------------------------------------------------------------------------
proc Trace {{force 0}} {
   if {$force || $::snackAmpSettings(Trace)} {
      set Stack [list]
      set depth [info level]
      set lastArgs [list]
      for {set i 1} {$i < $depth} {incr i} {
         lappend Stack [lindex [info level $i] 0]
         set lastArgs [lrange [info level $i] 1 end]
      }
      saLog "[join $Stack =>] \{$lastArgs\}"
   }
}
#----------------------------------------------------------------------------
# Return proc name of calling function
#----------------------------------------------------------------------------
proc me {args} {
   return [lindex [info level 1] 0]
}

#----------------------------------------------------------------------------
# Debugging to log all errorInfo hits
#----------------------------------------------------------------------------
proc le {{start 1}} {
   trace vdelete ::errorInfo w showlogerrors
   if {$start} {
      trace variable ::errorInfo w showlogerrors
   } else {
      trace vdelete ::errorInfo w showlogerrors
   }
}
proc showlogerrors {var index operand} {
   upvar $var ER
   puts "-------------\nError:$ER"
}
#------------------------------------------------------------------------------
# Function   : toDb/fromDb
# Description: Convert a gain factor from and to Db
# Author     : Tom Wilkason
# Date       : 7/29/2002
#------------------------------------------------------------------------------
proc toDb {value} {
   return [expr {log10($value)*$::snackAmpSettings(gainFactor)}]
}
proc fromDb {value} {
   return [expr {pow(10.,$value/$::snackAmpSettings(gainFactor))}]
}
proc tologDb {value} {
   if {$value <= 0.0} {
      return $::snackAmpSettings(minDB)
   }  else {
      if {[catch {expr {log10($value)*20.0}} val] } then {
         saLog "[me] $value: $val"
         return $::snackAmpSettings(minDB)
      } else {
         return $val
      }
   }
}
proc fromlogDb {value} {
   if {[catch {expr {pow(10.,$value/20.0)}} val] } then {
      saLog "[me] $value: $val"
   } else {
      return $val
   }
}

#------------------------------------------------------------------------------
# Take a scale input value, and scale to 0-1
# Used for line mixers that are inheriently logarithmic
#------------------------------------------------------------------------------
proc fromlinDb  {value} {
   set range [expr {double(-$::snackAmpSettings(minDB))}]
   set scvalue [expr {($value-$::snackAmpSettings(minDB))/$range}]
   if {$scvalue < 0} {
      return 0.0
   } else {
      return $scvalue
   }
}
#------------------------------------------------------------------------------
# Function   : saLog
# Description: Log and event, either to a file or the screen
# Author     : Tom Wilkason
# Date       : 2/8/2001
#------------------------------------------------------------------------------
proc saLog {Message} {
   set now  [timestamp]
   if {$::snackAmpSettings(logFile)} {
      if {[catch {set fid [open $::TA(logFile) a]} result]} {
         puts $result
         puts "$now $Message"
         return
      }
      puts $fid "$now $Message"
      flush $fid
      close $fid
   } else {
      puts "$now $Message"
      #puts stdout "$now $Message"
   }
}

#------------------------------------------------------------------------------
# Return the root folder name a file is on
#------------------------------------------------------------------------------
proc fileRoot {path} {
   while {1} {
      set parent [file dirname $path]
      if {[string equal $parent $path]} {
         return $path
      } else {
         set path $parent
      }
   }
}
#------------------------------------------------------------------------------
# Check if a file exists, handle proper casing on windows
#------------------------------------------------------------------------------
if {[string equal $::tcl_platform(platform) "windows"]} {
   proc fileExists {file} {
      if {[catch {file attributes $file -longname} name] } then {
         return 0
      } else {
         if {![string match $name $file]} {
            return 0
         }
      }
      return 1
   }
} else {
   proc fileExists {file} {
      if {![file exists $file]} {
         return 0
      }
      return 1
   }
}
##
# IFDEF: Use for conditional "compilation"
#
proc IFDEF {_var code {else ""}} {
   upvar $_var var
   if {[info exists var]} {
      if {[slen $code]} {
         uplevel $code
      }
   } else {
      if {[slen $else]} {
         uplevel $else
      }
   }
}
#----------------------------------------------------------------------------
# Encode a gif file for use in saImages
#----------------------------------------------------------------------------
proc encodeFile {filename} {
   package require base64
   set file [ open  $filename ]
   fconfigure $file -translation binary
   set text [read $file]
   set result [base64::encode $text]
   close $file
   return "saImages.photo xxxx \{\n$result\n\}"
}

#------------------------------------------------------------------------------
# Return a properly formatted list of files
#------------------------------------------------------------------------------
proc dndFileList {folders} {
   if {$::tcl_platform(platform)=="unix"} {
      set folders [join $folders \r\n]
   }
   return $folders
}
#----------------------------------------------------------------------------
# Watch a variable R.S.
#----------------------------------------------------------------------------
proc watch {args} {
   foreach arg $args {
      uplevel 1 "trace var $arg w {puts $arg:\[set $arg\] ;#}"
   }
 }
#------------------------------------------------------------------------------
# Show hex values for a string
#------------------------------------------------------------------------------
proc rawdump {String} {
   set raw ""
   for {set idx 0} {$idx < [slen $String]} {incr idx} {
      set byte [string index $String $idx]
      binary scan $byte h v
      append raw "0x$v:$byte "
   }
   return $raw
}
proc wrap {string maxlen} {
   set splitstring {}
   foreach line [split $string "\n"] {
      lappend splitstring [wrapline $line $maxlen]
   }
   return [join $splitstring "\n"]
}
proc wrapline {line maxlen} {
   set string [split $line " "]
   set newline [list [lindex $string 0]]
   foreach word [lrange $string 1 end] {
      if {[slen $newline]+[slen $word] > $maxlen} {
         lappend lines [join $newline " "]
         set newline {}
      }
      lappend newline $word
   }
   lappend lines [join $newline " "]
   return [join $lines "\n"]
}
##
# File copy replacement that doesn't block sound playing
# Adaptive block size depending on transport speed
#
proc bcopy {from to {force 1}} {
   Trace
   set maxSize  1000000
   set chunk $maxSize
   if {[file isdirectory $from]} {
      saLog "Copying whole folder '$from' to '$to'"
      # todo: recurse and copy all files/sub-folders
      file copy -force -- $from $to
      return
   }
   set stime [file mtime $from]
   set dtime [expr {$stime-1}]
   # Destination
   # only dest folder specified
   if {[file exists $to]} {
      if {[file isdirectory $to]} {
         set to [file join $to [file tail $from]]
      } else {
         set dtime [file mtime $to]
      }
   }
   ##
   # Only copy tracks with a later timestamp (or not existing)
   #
   if {$stime > $dtime} {
      set fid [open $from r]
      fconfigure $fid -translation binary -encoding binary -buffersize $chunk -buffering full
      # Make sure parent of dest exists (place to put the file)
      set toDir [file dirname $to]
      if {![file exists $toDir]} {
         file mkdir $toDir
      }
      if {$force==0 && [file exists $to]} {
         error "Can't overwrite $to without force option"
      }
      ##
      # Ready to create target
      # Open target file
      if {[catch {
         set ftd [open $to w]
         #fconfigure $ftd -translation binary -encoding binary -buffersize $chunk -buffering full
         fconfigure $ftd -translation binary -encoding binary -buffersize $maxSize -buffering full
         set csize $::snackAmpSettings(chunk)
         set minSize     [expr {8192*3}]
         set targetTime 80000.0   ;#80 ms, to keep music playing
         set i 0
         set totBytes 0
         set totTime 0.0
         # Use an adaptive chunk size to keep music playing regardless
         # of how fast/slow the target media is.
         while {![eof $fid]} {
            set t [lindex [time {
               set bytes [fcopy $fid $ftd -size $csize]
            } 1] 0]
            incr i
            incr totBytes $bytes
            set totTime [expr {$totTime+$t}]
            # Use running values to be more accurate
            set mbps [expr {double($totBytes)/$totTime}]
#debug $csize $mbps $from
            set cs [expr {int($targetTime*$mbps)}]
            # Clamp unruly values
            if {$cs > $maxSize} {
               set csize $maxSize
            } elseif {$cs < $minSize} {
               set csize $minSize
            } else {
               set csize $cs
            }
            update
         }
         set ::snackAmpSettings(chunk) $csize
         # saLog "$::snackAmpSettings(chunk) average block size @ $mbps MBytes/s\n $from ->\n $to"
      } result] } then {
         saLog $result
         catch {close $ftd}
         catch {close $fid}
         file delete -force $to
         return -code error $result
      } else {
         close $fid
         close $ftd
         file mtime $to $stime
         return 1
      }
   }
}
##
# File copy, faster than bcopy but stops music
#
proc bfcopy {from to {force 1}} {
   Trace
   update
   if {[file isdirectory $from]} {
      saLog "Copying whole folder '$from' to '$to'"
      # todo: recurse and copy all files/sub-folders
      file copy -force -- $from $to
      return
   }
   set stime [file mtime $from]
   set dtime [expr {$stime-1}]
   # Destination
   # only dest folder specified
   if {[file exists $to]} {
      if {[file isdirectory $to]} {
         set to [file join $to [file tail $from]]
      } else {
         set dtime [file mtime $to]
      }
   }
   ##
   # Only copy tracks with a later timestamp (or not existing)
   #
   if {$stime > $dtime} {
      if {$force==0 && [file exists $to]} {
         error "Can't overwrite $to without force option"
      }
      ##
      # Ready to create target
      # Open target file
      if {[catch {
         if {$force} {
            file copy -force $from $to
         } else {
            file copy $from $to
         }
      } result] } then {
         saLog $result
         return -code error $result
      } else {
         file mtime $to $stime
         return 1
      }
   }
}


#------------------------------------------------------------------------------
# Fast routines for formatting time
#------------------------------------------------------------------------------
proc timeFormat {value} {
  return [clock format $value -format $::snackAmpSettings(timeString)]
}
#------------------------------------------------------------------------------
# Function   : timestamp
# Description: Return a timestamp for NOW
# Author     : Tom Wilkason
# Date       : 2/1/2003
#------------------------------------------------------------------------------
proc timestamp {args} {
   set secs [clock seconds]
   set ms [clock clicks -milliseconds]
   set base [expr { $secs * 1000 }]
   set fract [expr { $ms - $base }]
   if { $fract > 1000 } {
      set diff [expr { $fract / 1000 }]
      incr secs $diff
      incr fract [expr { -1000 * $diff }]
   }
   return "[clock format $secs -format %T].[format %03ld $fract]"
}
##
# Return just the minutes seconds and optionally the .frac
#
proc HMSfrac {len {frac 0}} {
   set value [expr {round($len)}]
   set s [expr {$value%60}]
   set m [expr {($value%3600)/60}]
   set h [expr {$value/3600}]
   if {$h > 0} {
      set str [format "%d:%02d:%02d" $h $m $s]
   } else {
      set str [format "%02d:%02d" $m $s]
   }
   if {$frac} {
      append str ".[expr {int(10*($len-int($len)))}]"
   }
   return $str
}
interp alias {} timify {} HMSfrac
#------------------------------------------------------------------------------
# Return the rating in a nice unicode font and viceversa
# \u25CF
#------------------------------------------------------------------------------
proc mrating {rating} {
   return [string map "* \u25C6" $rating]
}
proc brating {rating} {
   return [string map "\u25C6 *" $rating]
}

#------------------------------------------------------------------------------
# clamp a value between limits
#------------------------------------------------------------------------------
proc clamp {value min max} {
   if {$value < $min} {
      set value $min
   } elseif {$value > $max} {
      set value $max
   }
   return $value
}

#------------------------------------------------------------------------------
# return the max value
#------------------------------------------------------------------------------
proc max {args} {
   set val -1e9
   foreach {arg} $args {
      if {$arg > $val} {
         set val $arg
      }
   }
   return $val
}
#------------------------------------------------------------------------------
# Separate a file folder from its tail
#------------------------------------------------------------------------------
proc splitFile {file} {
   return "\n[file dirname $file]\n[file tail $file]"
}

#------------------------------------------------------------------------------
# Copy array 1 into array 2 for non blank array 1 fields
#------------------------------------------------------------------------------
proc arrayCopyNoBlank {_a1 _a2} {
   upvar $_a1 a1
   upvar $_a2 a2
   foreach {element value} [array get a1] {
      if {$value ne ""} {
         set a2($element) $value
      }
   }
}
package provide gpUtils 1.0

