#------------------------------------------------------------------------------
# id3Renamer 1.0
#
# ID3 file renamer/editor windows for snackAmp Player in Tcl/Tk
# Uses id3Tag package
# Copyright (C) 2003 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
# Function   : id3TagToName
# Description: Window to edit the proposed file names based on tag data
# Author     : Tom Wilkason
# Date       : 2/8/2001
#------------------------------------------------------------------------------
namespace eval id3Renamer {
}
proc id3Renamer::Edit {fnames} {
   #option add *ID3Editors*Label.font entry interactive ;# high priority
   package require id3Editor
   set i 0
   #global fnameData ;# Note, due to upvar #0, TFW Fix later
   if {[llength $fnames] == 0} {
      tk_messageBox -type ok -icon warning -message "No files specified"
      return
   }
   foreach fname $fnames {
      if {![file exists $fname]} {
         saLog "'$fname' Does not exist"
      }
   }
   set here [focus]
   # Get a unique window, note $W is also a global variable
   # holding stuff for this instance of the GUI
   while {[winfo exists .tnIdent$i]} {incr i}
   set W .tnIdent$i
   toplevel $W -cursor $::TA(arrow)  -class ID3Editors
   wm title $W "Tag To Name Editor"
   wm protocol $W WM_DELETE_WINDOW [list id3Editor::closeWindow $W [me]]
   settings::Hide $W
   upvar #0 $W myData
   global $W.renamer
   upvar #0 $W.renamer myData
   set linkData $W.renamer

   pack [id3Renamer::build $W.renamer $fnames] -fill both -expand true

   settings::restoreWidth $W [me]
   ontop::pinOnSystemMenu {} $W
   myIcon $W
   focus $W
   settings::Restore $W
}
#------------------------------------------------------------------------------
#  Build the guts of the renamer gui, put it into a toplevel or hull
#------------------------------------------------------------------------------
proc id3Renamer::build {parent fnames} {
   set base [frame $parent]
   set W [winfo toplevel $base]
   upvar $W myData
   global $W.renamer
   upvar $W.renamer myData
   set linkData $W.renamer
   ##
   # Build the GUI
   #
   set i 0
   set span 3
   set header [frame $base.header]
   ##
   # Frame to hold the Name guessing
   #
   set hull [labeledFrame $header.rep -text "Name Template (for Guessing)"]
   #set ::TA(smartName) $::snackAmpSettings(smartName)
   grid [button $hull.snl -highlightthickness 0 -pady 0 -text "Guess" \
    -command [list id3Renamer::_Guess ::$linkData] -image saImages.Guess -compound left] \
        [entryHistory::create $hull.sne ::snackAmpSettings(smartName) ::snackAmpSettings(id3Structure) -width 20] -sticky nsew

   #widgetWatch::setWatch $hull.sne ::snackAmpSettings(smartName)
   balloon::define $hull.sne \
      "Enter naming scheme for converting tags to file names:\n%%A - Artist, %%C - Album, %%N - Track Number, %%T - Title"
   balloon::define $hull.snl "Push this to guess the correct file names based on tags"
   ##
   # Frame to hold the replacement text
   #
   set hull [labeledFrame $header.entry -text "Replace Text"]
   # header
   grid [button $hull.rlbh -pady 0 -text "Replace" -command [list id3Renamer::_Substitute ::$linkData]] \
        [entry $hull.rlbf -textvariable ::${linkData}(fromText) -width 30] \
        [entry $hull.rlbt -textvariable ::${linkData}(toText) -width 30] \
        -sticky nsew
   # Expand entry boxes
   grid columnconfigure $hull 1 -weight 1
   grid columnconfigure $hull 2 -weight 1

   balloon::define $hull.rlbh "Push this to replace text in names (Change field must be checked)"
   balloon::define $hull.rlbf "Enter Text to Replace here"
   balloon::define $hull.rlbt "Enter Replacement Text here"
   pack $header.rep -side left
   pack $header.entry -side left -fill x -expand true
   pack $header -side top -anchor w -fill x -expand false

   # Line
   pack [horzLine $W.l[incr i]] \
      -expand false -fill x -side top

   # Decide when to use a scroll bar and how big to make the screen,
   # don't want to make any bigger than necessary.
   set needHeight [expr {round(([llength $fnames]+1) * [font metrics entry -linespace]*1.3)}]
   set maxHeight [winfo pixels . $::snackAmpSettings(scrolledCanvasHeight)]
   if {$maxHeight > $needHeight} {
      set maxHeight $needHeight
   }
   set bg [$base cget -background]
   set sw [ScrolledWindow $base.data -relief flat -borderwidth 1 -background $bg]
   set sf [ScrollableFrame $sw.f -constrainedwidth true -height $maxHeight -background $bg]
   $sw setwidget $sf
   set hull [$sf getframe]
   # header
   grid [label $hull.lb[incr i] -text "Change"] \
        [label $hull.lb[incr i] -text "Old Name"] \
        [label $hull.lb[incr i] -text "New Name"] \
        -sticky ew
   ##
   # Build one row for each file
   #
   set tag 0
   foreach fname $fnames {
      # Get existing tag data
      # Propose new file name
      set myData($tag,oldFullName) $fname
      set fn [file rootname [file tail $fname]]
      set myData($tag,oldName) $fn
      set myData($tag,newName) $fn
      set myData($tag,ok) 1
      # Note using a very small font allows the checkbuttons to pack tighter
      grid [checkbutton $hull.cb$i -variable ::${linkData}($tag,ok) -font {Helvetica 5}] \
         [label $hull.ln$i -width 50 -textvariable ::${linkData}($tag,oldName) -relief sunken -anchor w] \
         [entry $hull.en$i -width 50 -textvariable ::${linkData}($tag,newName)] \
         -sticky nsew
         widgetWatch::setWatch $hull.en$i ::${linkData}($tag,newName)
         ##
         # Bind up/down keys to go up down in column
         #
         bind $hull.en$i <Up> [list Focus $hull.en[expr {$i-1}]]
         bind $hull.en$i <Down> [list Focus $hull.en[expr {$i+1}]]

      incr tag
      incr i
   }
   grid columnconfigure $hull 1 -weight 1
   grid columnconfigure $hull 2 -weight 1
   ##
   # Frame for OK/Cancel/Apply buttons
   #
   pack [frame $base.but] -fill x -expand false -side bottom
   # Line
   pack [horzLine $base.l[incr i]] \
      -expand false -fill x -side bottom
   # Data
   pack $base.data -side top -expand true -fill both
   # Check All
   set checkAll [button $base.but.[incr i] -default normal -text "Check All" \
      -command [list id3Editor::cbSetOK $linkData 1]]
   balloon::define $checkAll "Push this to check all checkboxes"

   # Check None
   set checkNone [button $base.but.[incr i] -default normal -text "Check None" \
      -command [list id3Editor::cbSetOK $linkData 0]]
   balloon::define $checkNone "Push this to un-check all checkboxes"

   #fix
   set fix [button $base.but.[incr i] -width 8 -default normal -text "Fix Case" \
      -command [list id3Renamer::_Fix $linkData]]
   balloon::define $fix "Push this to automatically correct file name spacing and capitalization"

   # Apply
   set app [button $base.but.[incr i] -default normal -text "Apply"  -compound left -image saImages.Save \
      -command [list id3Renamer::_callBack $linkData [me] 0]]
   balloon::define $app "Push this rename the changed files that are checked"

   # Apply and close
   set ok [button $base.but.[incr i] -default active -text "OK"  -padx 4 \
      -compound left -image saImages.Apply \
      -command [list id3Renamer::_callBack $linkData [me] 1]]
   balloon::define $ok "Push this to rename the changed files that are checked, then close this dialog"

   # Cancel
   set can [button $base.but.[incr i] -default normal -text "Cancel" -compound left -image saImages.Close \
      -command [list id3Editor::closeWindow $W [me]]]
   balloon::define $can "Push this to ignore changes made and restore the previous settings"

   grid $checkAll $checkNone $fix $app $ok $can  -padx [extraPad 1] -pady 2 -sticky ns
   ##
   # Key bindings
   #
   $ok configure -default active
   #bind $base <Return> [list $ok invoke]
   bind [winfo toplevel $base] <Escape> [list $can invoke]
   return $base
}


#------------------------------------------------------------------------------
# Function   : id3Renamer::_Substitute
# Description: Replace some text in enabled file names
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Renamer::_Substitute {_W} {
   upvar $_W W
   foreach {element} [lsort [array names W *,ok]] {
      foreach {key xx} [split $element ,] {break}
      if {$W($element)} {
         # _W is traced variable, use full name when setting
         #regsub  -- $W(fromText) $W($key,oldName) $W(toText) ${_W}($key,newName)
         #regsub -- $W(fromText) $W($key,newName) $W(toText) ${_W}($key,newName)
         set W($key,newName) [string map [list $W(fromText) $W(toText)] [set ${_W}($key,newName)]]
      }
   }
}
#------------------------------------------------------------------------------
# Function   : id3Renamer::_Guess
# Description: Guess the appropriate name based on a template
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Renamer::_Guess {_W} {
   upvar $_W W
   foreach {element} [lsort [array names W *,ok]] {
      foreach {tag xx} [split $element ,] {break}
      set fname $W($tag,oldFullName)
      id3Tag::id3Label $fname fnameData $::snackAmpSettings(ID3ReadOrder)
      set fn [file rootname [file tail $fname]]
      set sfn [catalog::smartFileName fnameData $::snackAmpSettings(smartName)]
      if {($sfn ne "") && $W($tag,ok) == 1} {
         # Use orig name, traced variable
         set W($tag,newName) $sfn
      }
   }
}
#------------------------------------------------------------------------------
# Function   : id3Renamer::_Fix
# Description: Clean the file name
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Renamer::_Fix {_W {patt *,ok}} {
   upvar $_W W
   #global fnameData ;# needs to be global
   foreach {element} [lsort [array names W $patt]] {
      foreach {tag xx} [split $element ,] {break}
      set fname $W($tag,newName)
      set fn [fixplus::cleanName $fname]
      if {($fn ne "") && ($fn ne $fname)} {
         # Use orig name, ${_W} is a traced variable
         set W($tag,newName) $fn
         set W($tag,ok) 1
      }
   }
}
#------------------------------------------------------------------------------
# Function   : id3Renamer::_callBack
# Description: Rename a set of files
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Renamer::_callBack {_W winTag {close 0}} {
   upvar $_W W
   ##
   # Link in table editing data (will ignore if it doesn't exist)
   regexp {(\..+?)\.} $_W -> parent
   global $parent.table
   upvar $parent.table tableData

   #
   set fromList [list]
   set toList [list]
   set elements [lsort [array names W *,ok]]
   set tot [llength $elements]
   set i 0
   foreach {element} $elements {
      # If checkbox checked, try to rename it
      if {$W($element)} {
         set tag [lindex [split $element ,] 0]
         set oldFullName $W($tag,oldFullName)
         set newBase     $W($tag,newName)
         set newFullName [file join [file dirname $oldFullName] "$W($tag,newName)[file extension $oldFullName]"]
         if {$oldFullName ne $newFullName} {
            # Bug: When renaming a file, the index in the database changes so the
            # apl listbox index is invalid
            lappend fromList $oldFullName
            lappend toList $newFullName
            # Update the fields
            set W($tag,oldFullName) $newFullName
            set W($tag,oldName)  [file rootname [file tail $newFullName]]
            # If associated table editor, rename the base file names for tags
            if {[info exists tableData($tag,oldFullName)]} {
               set tableData($tag,oldFullName) $newFullName
            }
         }
      }
   }
   ##
   # Rename the files
   #
   if {[llength $fromList] > 0} {
       if {[catalog::moveTrackFolderPairs $fromList $toList]} then {
         tk_messageBox -icon warning -message "Could not rename one or more files\nsee console log for errors"
      }
   }
   # If close widget when done indicated.
   if {$close} {
     id3Editor::closeWindow $_W $winTag
   }
   waitWindow::close
}
package provide id3Renamer 1.0

