#----------------------------------------------------------------------------
# Establish a client interface for an IRC remote control server
# This currently works with the ATI remote
# Requires WinLIRC ATI Remote plug-in http://winlirc.sourceforge.net/
# If other LIRC remotes are used, may require mods to switch statement below
#----------------------------------------------------------------------------
##
# IR Handler for remote controls
#
namespace eval irc {
   variable socket 0
   variable enabled
   variable ircCheck
   variable ircEnable
   variable port
   variable host
   variable useIRC
   upvar #0 snackAmpSettings(useIRC)     useIRC
   upvar #0 snackAmpSettings(ircEnabled) enabled
   upvar #0 snackAmpSettings(ircCheck)   ircCheck
   upvar #0 snackAmpSettings(ircEnable)  ircEnable
   upvar #0 snackAmpSettings(ircPort)    port
   upvar #0 snackAmpSettings(ircHost)    host
}

#----------------------------------------------------------------------------
# Install the LIRC socket interface
#----------------------------------------------------------------------------
proc irc::Server {args} {
   variable socket
   variable port
   variable host
   if {[string length [string trim $host]]==0} {
      set host [info hostname]
   }
   catch {close $socket} result
   if [catch {set socket [socket $host $port]} result] {
      saLog "Can't establish IRC socket for $host on port $port\n$result"
      set socket 0
   } else {
      fconfigure $socket -translation auto
      fileevent $socket readable [list irc::Process $socket]
      saLog "IRC Server Established on port $port"
   }
}
proc irc::Close {args} {
   variable socket
   catch {close $socket}
   set socket 0
}
#########################################################################
# Function    : irc::Process
# Description : Called from fileevent on a readable input on my port
# Author      : Tom Wilkason
# Date        : 6/2/1999
#########################################################################
proc irc::Process {sock} {
   global snackAmpSettings
   variable socket
   variable host
   variable port
   variable enabled
   variable ircCheck
   variable ircEnable
   ;# If eof or partial line the close up the channel now
   if {[catch {gets $sock line}]||[eof $sock]} {
      # end of file or abnormal connection drop
      catch {close $sock}
      set socket 0
      saLog "IRC socket terminated for $host on port $port, restarting"
      after 200 irc::Reopen
   } else {
      ## line example
      #                    cmd  type
      #000000000001002c 00 pb_c atiremotewonder
      #000000000001002c 00 pb_c atiremotewonder
      #0000000000010025 00 play atiremotewonder
      #0000000000010027 00 stop atiremotewonder
      # Process the received command line
      # The forth param is the remote type e.g. "atiremotewonder"
      set remote [lindex $line 3]
      # The third parameter is the ascii command
      set command [lindex $line 2]
      set ::snackAmpSettings(ircLastKey) $command
      ##
      # Check for an enable key, otherwise disable
      #
      if {[lsearch -exact $ircCheck $command]>=0} {
         if {[string equal $command $ircEnable]} {
            set enabled 1
         } else {
            set enabled 0
         }
      }
      if {$enabled} {
         switch -- [string tolower $command] {
            vol_up {bumpLevel 1}
            vol_down {bumpLevel -1}
            fastforward {Next}
            rewind {Prev}
            1 -
            2 -
            3 -
            4 -
            5 {
               gui::setTrackRating $::TA(currentMF) [string repeat * $command]
            }
            pause {Pause}
            stop {Stop}
            play {Play}
            channel_up {bumpScaling .1}
            channel_down {bumpScaling -.1}
            mute {Pause}
            right {macros::skipPlay 10}
            left {macros::skipPlay -10}
            record {taExit}
            fullscreen {
               if {[wm state .saTop]=="iconic"} {
                  wm deiconify .saTop
               } else {
                  foreach {child} [wm stackorder .] {
                     catch {wm iconify $child}
                  }
               }
            }
            up {wm deiconify .saTop}
            down {
               if {[wm overrideredirect .saTop]} {
                  wm withdraw .saTop
               } else {
                  wm iconify .saTop
               }
            }
            ok {::browser::urlOpen $::snackAmpSettings(internalWebBase)/remote/index.html}
            dirpad_up {macros::MoveTo up}
            dirpad_down {macros::MoveTo down}
            default {
            }
         }
      }
   }
}
#------------------------------------------------------------------------------
# Run a loop to open the servers
#------------------------------------------------------------------------------
proc irc::Reopen {args} {
   variable socket
   variable useIRC
   after cancel irc::Reopen
   if {$useIRC && ($socket==0 ||[lsearch -exact [file channels] $socket]<0)} {
      irc::Server
   } elseif {$socket!=0 && $useIRC==0} {
      irc::Close
   }
   after 10000 irc::Reopen
}
#----------------------------------------------------------------------------
# Fire up the server
#----------------------------------------------------------------------------
irc::Reopen
package provide ircserver 1.0




