#------------------------------------------------------------------------------
# macros 1.0
#
# High level commands for SnackAmp
#
# Copyright (C) 2003 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/macros.tcl,v 1.59 2007/07/03 00:46:49 wilkason Exp $
}
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# Macros Namespace
#/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
namespace eval macros {
   namespace export *
}
#------------------------------------------------------------------------------
# Function:  macros::PlayOffset
#
# Play a music file at some offset, likely bound to slider control
#------------------------------------------------------------------------------
proc macros::PlayOffset {{manual 0}} {
   Trace
   global TA LB snackAmpSettings
   # Make sure we have an open file
   set soundLength [soundControl::length]
   if {$soundLength > 0} {
      # Some files are fast and some are slow to respond to -start = 0
      if {$manual == 0} {
         set offset [expr {round(double($gui::trackPosition)/$gui::songDuration*$soundLength)}]
      } else {
         set offset $manual
      }
      if {[catch {
         soundControl::stop
         ;# Inverse active track
         LB::inverseVideoFile $TA(currentMF) 1
         # Workaround for bug on linux/ogg where file plays from last pos, not from -start
         if {$::tcl_platform(platform) == "unix"} {
            soundControl::sound [soundControl::fileName]
         }
         soundControl::play -start $offset -command [soundControl::getNextCommand $::snackAmpSettings(trackGap)]
      } result]} then {
         saLog "PlayOffset: $result"
         return
      }
      soundControl::Timer
   }
}
#------------------------------------------------------------------------------
# Function   : bumpScaling
# Description: Adjust the current scaling for the playing song for some amount
# Author     : Tom Wilkason
# Date       : 1/9/2002
#------------------------------------------------------------------------------
proc macros::bumpScaling {percent} {
   global TA
   set gain [soundControl::getTrackGain]
   # TODO: Need to adjust album amount perhaps?
   gainControl::addTag $TA(currentMF) [expr {$gain*(1.0+$percent)}] "xxx"
   soundControl::setTrackGain [expr {$gain*(1.0+$percent)}]
}
#------------------------------------------------------------------------------
# Function   : Assign
# Description: Routine to assign global hotkeys, called from hotkeys file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::Assign {key Command} {
   if {[catch {bind .saTop $key $Command} result]} {
      saLog "FilterAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
   }
}
#------------------------------------------------------------------------------
# Function   : FilterAssign
# Description: Routine to assign filter box hotkeys, called from hotkeys file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::FilterAssign {key Command} {
   if {[catch {bind $::FI $key $Command} result]} {
      saLog "FilterAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
   }
}
#------------------------------------------------------------------------------
# Function   : PlaylistAssign
# Description: Routine to assign play list hotkeys, called from hotkeys file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::PlaylistAssign {key Command} {
   if {[catch {bind [LB::body] $key $Command} result]} {
      saLog "PlaylistAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
   }
}
#------------------------------------------------------------------------------
# Function   : APL_ListAssign
# Description: Routine to assign APL play list hotkeys, called from hotkeys file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::APL_ListAssign {key Command} {
   if {[APLExists]} {
      set plb [$::PLB bodypath]
      if {[catch {bind $plb $key $Command} result]} {
         saLog "APL_ListAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
      }
   }
}
#------------------------------------------------------------------------------
# Function   : APL_TreeAssign
# Description: Routine to assign APL tree hotkeys, called from hotkeys file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::APL_TreeAssign {key Command} {
   if {[APLExists]} {
      if {[catch {bind [winfo parent $::TREE] $key $Command} result]} {
         saLog "APL_TreeAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
      }
   }
}
#------------------------------------------------------------------------------
# Function   : APL_TreeAssign
# Description: Routine to assign APL tree hotkeys, called from hotkeys file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::APL_TreeNodeAssign {key Command} {
   if {[APLExists]} {
      if {[catch {$::TREE bind x $key $Command} result]} {
         saLog "APL_TreeNodeAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
      }
   }
}
#------------------------------------------------------------------------------
# Function   : APLFilterAssign
# Description: Routine to assign APL filter box hotkeys, called from hotkeys file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::APLFilterAssign {key Command} {
   if {[APLExists]} {
      if {[catch {bind $::SAFind $key $Command} result]} {
         saLog "APLFilterAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
      }
   }
}

#------------------------------------------------------------------------------
# Function   : SelectFile
# Description: Highlight the folder the selected file is in
# Author     : Tom Wilkason
# Date       : 3/18/2002
#------------------------------------------------------------------------------
proc macros::SelectFile {{playing 0}} {
   if {!([winfo exists .wFC] && [wm state .wFC] == "normal")} {
      AutoPlaylist
   }
   set TL [winfo toplevel [focus]]
   if {$playing} {
      APL::SelectFile [soundControl::fileName]
   } else {
      if {$TL eq $::SA} {
         APL::SelectFile [LB::nameActive]
      } else {
         APL::SelectFile [aplLB::getActiveFullName]
      }
   }
}

#------------------------------------------------------------------------------
# Function: macros::EditPLID3
#
# Invoke the id3 editor (in group mode) on the selected playlist file
#------------------------------------------------------------------------------
proc macros::EditPLID3 {args} {
   package require id3Editor
   set files [LB::nameSelected]
   if {[llength $files] > $::snackAmpSettings(ID3editLimit)} {
      tk_messageBox -type ok -icon warning \
         -message "Limiting operation to first $::snackAmpSettings(ID3editLimit) files"
   }
   id3Editor::Edit [lrange $files 0 $::snackAmpSettings(ID3editLimit)]
}

#------------------------------------------------------------------------------
# Function: macros::PLNameToTag
#
# Invoke the name to tag editor on the playlist selected files
#------------------------------------------------------------------------------
proc macros::PLNameToTag {smart} {
   package require id3Table
   id3Table::Edit [LB::nameSelected] $smart
}

#------------------------------------------------------------------------------
# Function: macros::PLTagToName
#
# Invoke the tag to name editor on the playlist selected files
#------------------------------------------------------------------------------
proc macros::PLTagToName {args} {
   package require id3Renamer
   id3Renamer::Edit [LB::nameSelected]
}

#------------------------------------------------------------------------------
# Function: macros::removeTag
#
# Remove an ID3 tag (not currently implemented)
#------------------------------------------------------------------------------
proc macros::removeTag {files tags} {
   set todo [llength $files]
   set counter 0
   if {[tk_messageBox -type yesno -title "Please Confirm" \
      -message "Are you sure you want to remove the $tags tags?"]=="yes"} {
      set todo [llength $files]
      set counter 0
      foreach {file} $files {
         incr counter
         foreach {startOffset endOffset} [id3Tag::findTagOffsets $file $tags] {break}
         if {$startOffset != 0 || $endOffset != 0} {
            set wasPlaying [soundControl::isPlayingTrack $file playlocation]
            waitWindow::percent $counter $todo "Removing Tag(s) for\n[splitFile $file]" 1
            id3Tag::trimFile $file $startOffset $endOffset
            # Want to re-cache file here
            unset -nocomplain Details
            db::cacheFile $file Details 1
            ##
            # Snackamp Unique, resume track playing and resume
            #
            if {$wasPlaying} {
               soundControl::resumePlayingTrack $file $playlocation
            }
         }
      }
      trackList::updateFiles $files
      waitWindow::close
   }
   return 1
}

#------------------------------------------------------------------------------
# Function: macros::EditPlaying
#
# Edit the tag info on the currently playing track
#------------------------------------------------------------------------------
proc macros::EditPlaying {args} {
   package require id3Editor
   id3Editor::Edit [list $::TA(currentMF)]
}
#------------------------------------------------------------------------------
# Function   : macros::Highlight
# Description: macros::Highlight one and only one selection
# Author     : Tom Wilkason
# Date       : 12/27/2000
#------------------------------------------------------------------------------
proc macros::Highlight {i {inverse 0}} {
   global TA
   LB::highlightIndex $i $inverse
   if {[APLExists] && $::snackAmpSettings(syncAPL)} {
      set TA(entryName) [LB::nameFromIdx $i]
      APL::SelectFile $TA(entryName)
   }
}
##
# Move the highlight up or down
#
proc macros::MoveTo {how} {
   global TA
   set i [lindex [LB::idxSelected] 0]
   if {![string is integer -strict $i]} {
      set i 0
   }
   switch -- $how {
      "down" {
         incr i
      }
      "up" {
         incr i -1
      }
      default {}
   }
   macros::Highlight $i
}
#------------------------------------------------------------------------------
# Function   : Swap
# Description: Move a selection up/down the list (bind keys to this)
# Author     : Tom Wilkason
# Date       : 10/29/2001
#------------------------------------------------------------------------------
proc macros::Swap {how} {
   global LB
   if {[llength [LB::idxSelected]] > 0} {
      if {$how=="up"} {
         foreach moveIndex [lsort -integer [LB::idxSelected]] {
            set above [expr {$moveIndex - 1}]
            LB::moveByIndex $above $moveIndex
         }
      } else {
         foreach moveIndex [lsort -integer -decreasing [LB::idxSelected]] {
            set above [expr {$moveIndex + 1}]
            LB::moveByIndex $above $moveIndex
         }
      }
   }
}
#----------------------------------------------------------------------------
# Toggle the playlist
#----------------------------------------------------------------------------
proc macros::PlaylistToggle {args} {
   cframe.toggle $::TA(CF) $::TA(playlistButton)
}
#------------------------------------------------------------------------------
# Function   : PlayGain/getPlayGain
# Description: Provide x2 gain control to better match audio experience
# Author     : Tom Wilkason
# Date       : 10/19/2001
#------------------------------------------------------------------------------
proc macros::PlayGain {db} {
   #Trace
   soundControl::setGlobalGain $db
   if {$db > -79} {
      #set db "[format "%.1f" [expr {log10($value/100.0)*20}]] dB"
      set dbt "[format %.0f $db] dB"
   } else {
      set dbt "Volume = off"
   }
   set ::songArray(Title) "Volume $dbt"
   # restore it after a second
   after cancel soundControl::restoreTitle
   after 1500 soundControl::restoreTitle
   return $db
}
#------------------------------------------------------------------------------
# Function   : bumpLevel
# Description: Adjust the current play level by some amount
# Author     : Tom Wilkason
# Date       : 11/9/2001
#------------------------------------------------------------------------------
proc macros::bumpLevel {factor} {
   set db [tologDb [expr {[soundControl::getGlobalGain]/100.0}]]
   PlayGain [expr {$db + $factor}]
}
#------------------------------------------------------------------------------
# Function   : Console
# Description: Toggle the console for debugging purposes (WIN32 only)
# Author     : Tom Wilkason
# Date       : 12/27/2000
#------------------------------------------------------------------------------
proc macros::Console {args} {
   global TA LB
   set state [console eval {wm state .}]
   package require commandline   ;# command line tools
   switch -- $state {
      normal {
         set TA(Console) 0
         console hide
      }
      withdrawn {
         set TA(Console) 1
         console title "SnackAmp console"
         console show
         console eval {bind .console <Control-w> {wm withdraw .}}
      }
      default {
         set TA(Console) 1
         console title "SnackAmp console"
         console show
         console eval {bind .console <Control-w> {wm withdraw .}}
      }
   }
}
#------------------------------------------------------------------------------
# bring up a tkcon window if possible
#------------------------------------------------------------------------------
proc macros::tcon {args} {
   if {[catch {package require tkcon} result] } then {
      saLog $result
   } else {
      set ::tkcon::OPT(cols) 120
      set ::tkcon::OPT(slaveeval) {
         tkcon attach Main
         puts "Attached..."
   	}
      tkcon show
      tkcon buffer 2048
      #pack [button .tkcon.fstatus.tabs.new -text New] -side left
      tkcon title "SnackAmp $::SAVersion Console"
   }
}

#------------------------------------------------------------------------------
#
#------------------------------------------------------------------------------
proc db {args} {
   package require twDebugInspector
   twDebug::Inspector .newToplevel
}
#----------------------------------------------------------------------------
# Snapshot all APL files into playlists
#----------------------------------------------------------------------------
proc macros::snapshotAllAPLs {patt {fpatt *}} {
   # save current teFolders
   array set saveTeFolders [array get ::teFolders]

   # process each playlist file
   foreach fName [catalog::getAutoPlaylists] {
      if {[string match $patt $fName]} {
         if {[macros::snapshotAPL $fName $fpatt]==-1} {
            return
         }
      }
   }

   # Restore older settings
   array set ::teFolders [array get saveTeFolders]
   db::syncToAPL
}
#------------------------------------------------------------------------------
# Create a playlist for tracks within an APL
#------------------------------------------------------------------------------
proc macros::snapshotAPL {fName {fpatt *}} {
   if {[file exists $fName]} {
      if {[aplFolder::entries]==0} {
         aplFolder::createTeFolders 0
      }
      waitWindow::show "Setting up APL Export for\n$fName" 1
      if {[waitWindow::ifbreak]} {
         return -1
      }

      # Read APL entries
      if {[catch {open $fName r} fid] } then {
         saLog "Error $fid"
         return 0
      }
      fconfigure $fid -buffersize 100000
      set full [split [read $fid] \n]
      close $fid

      # Invalidate all current settings
      aplFolder::SetSome * -1
      # Only update values if an entry already exists from prior scan
      # This prevents garbage if directories were moved
      foreach line $full {
         foreach {Index Value} $line {
            aplFolder::SetIfExists $Index $Value
         }
      }
      ##
      # Some folders will still be tagged with -1 from
      # above invalidation, meaning they weren't in the new autoplaylist
      # file. If so, then set them to the value of their parent if possible.
      # Also, set any parents to a 2 if they have children with checks.
      #
      foreach {Index} [aplFolder::foldersMatchingPattern * 1] {
         aplFolder::setTeDefault $Index
      }
      # Flag all tracks that match the teFolders
      db::syncToAPL

      # Get a list of all APL tracks, sorted appropriately
      set mediaPatterns [list]
      foreach {dir} [catalog::mountedFolders] {
         # handle m:/ as well as /folder
         set dir "[string trimright $dir / ]/"
         if {[string match $fpatt $dir]} {
            set patt [file join $dir *]
            lappend mediaPatterns $patt
            set m3u($patt) [file join $dir [file rootname [file tail $fName]].m3u]
            set tracks($patt) [list]
            set ml($patt) [string length $dir]
         }
      }
      waitWindow::show "Filtering/Sorting Tracks for\n$fName" 1
      if {[waitWindow::ifbreak]} {
         return -1
      }

      # Split tracks in to buckets depending on which media folder they reside
      # only deal with tracks on media folders of interest
      foreach track [filtering::sortAPLlist] {
         if {[string match $fpatt $track]} {
            foreach {patt} $mediaPatterns {
               if {[string match $patt $track]} {
                  # Get relative to the dir and place in that dir
                  lappend tracks($patt) [string range $track $ml($patt) end]
                  break
               }
            }
         }
      }

      ## TODO: Only create playlist file if it doesn't exist or APL timestamp is newer.


      # Save playlist files for each media folder if possible
      foreach {patt plf} [array get m3u] {
         waitWindow::show "Creating playlist file\n$plf" 1
         if {[waitWindow::ifbreak]} {
            return -1
         }
         if {[llength $tracks($patt)]>0} {
            if {[catch {open $plf w} fid] } then {
               saLog $fid
            } else {
               puts $fid [join $tracks($patt) \n]
               saLog "Wrote [llength $tracks($patt)] to $plf"
               close $fid
               update
            }
         } else {
            #if {[file exists $plf]} {
            #   file delete -force $plf
            #   saLog "Removed empty $plf"
            #}
         }
      }
   } else {
      saLog "Could not find APL file $fName"
   }
   return 0
}
#------------------------------------------------------------------------------
# Function   : applyCurrentAPL
# Description: Apply the latest Auto Playlist settings, ignoring any filtering
#              in effect.
# Author     : Tom Wilkason
# Date       : 11/8/2001
#------------------------------------------------------------------------------
proc macros::applyCurrentAPL {args} {
   Trace
   if {$::snackAmpSettings(askApplyAPL)} {
      set result [tk_messageBox -type yesno -title "Please Answer" \
         -icon question -message "Do you want to replace the Playlist with a new generated AutoPlaylist?"]
      if {$result eq "no"} {
         return
      }

   }
   soundControl::latency long
   Wait 1
   waitWindow::show "Building AutoPlaylist..."
   # Update db with files flagged
   db::syncToAPL
   waitWindow::show "Sorting/Shuffling AutoPlaylist..."
   # Get a list of all APL tracks, sorted appropriately
   set insertList [filtering::sortAPLlist]
   waitWindow::show "Inserting AutoPlaylist..."
   LB::deleteAll
   if {[llength $insertList] > 0} {
      LB::insertList 0 $insertList
      LB::tryRestorePlaying
   }
   Update
   macros::saveSettings
   catalog::SavePlaylistCache
   waitWindow::close
   Wait 0
   soundControl::latency
}
#------------------------------------------------------------------------------
# Function   : RenameFileIP
# Description: Rename a file directly from the listbox
# Author     : Tom Wilkason
# Date       : 6/19/2002
#------------------------------------------------------------------------------
proc macros::RenameFile {{fullName {}}} {
   set W $::LB
   if {[string length $fullName] == 0} {
      set fullName [LB::nameActive]    ;# lookup long name
   }
   if {[string length $fullName]==0} {
      return
   } elseif {![file exists $fullName]} {
      tk_messageBox -icon warning -message "'$fullName' does not exist"
   }
   set oldName [file tail $fullName]
   set dirName [file dirname $fullName]
   ##
   # Block on dialog
   #
   set name [catalog::renameInPlaceDialog $W 0 $oldName]
   ##
   # Back from window closure
   #
   if {($::TA(rename) ne "") && ($::TA(rename) ne $oldName)} {
      set newFullName [file join $dirName $::TA(rename)]
      if {[catalog::RenameFile $fullName $newFullName]} {
         tk_messageBox -icon warning -message "Could not rename\n$fullName to \n$newFullName\n$::errorInfo"
      }
   }
}
#------------------------------------------------------------------------------
# Function   : macros::RenameFile
# Description: Rename a file using a dialog, called from key binding or menu
# Author     : Tom Wilkason
# Date       : 10/30/2001
#------------------------------------------------------------------------------
proc macros::RenameFileDialog {{fullName {}}} {
   if {[string length $fullName] == 0} {
      set fullName [LB::nameActive]    ;# lookup long name
   }
   set oldName [file tail $fullName]
   set dirName [file dirname $fullName]
   set newName $oldName

   set answer [getEntryDialog "$oldName" newName "Rename File"]
   if {($answer eq "ok") && ($newName ne $oldName)} {
      set newFullName [file join $dirName $newName]
      if {[catalog::RenameFile $fullName $newFullName]} {
         tk_messageBox -icon warning -message "Could not rename\n$fullName to \n$newFullName\n$::errorInfo"
      }
   }
}

#------------------------------------------------------------------------------
# Function   : SaveFolderFile
# Description: Save the current folder settings to a file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::SaveFolderFile {{fName ""}} {
   global TA LB snackAmpSettings
   ##
   # If a list for GlobData is passed in, the second item should be the name to save as (if different)
   #
   if {$fName == ""} {
      set types [list [list "AutoPlaylists Files" .tafs] [list "All Files" "*"]]
      set fName [tk_getSaveFile -title "Save AutoPlaylists File" -filetypes $types \
         -initialdir $::HOME -defaultextension .tafs]
      if { $fName == "" } {
         return  ""
      }
   }
   # Force extension in case dialog lost it (as on linux)
   set fName "[file rootname $fName].tafs"
   if {[catch {open $fName w} InFile]} {
      set result [tk_messageBox -type ok -title "SnackAmp - Please Respond" -icon warning \
         -message "Save Error: Unable to open '$fName' for writing"]
      return ""
   }
   fconfigure $InFile -buffersize 100000

   set snackAmpSettings(lastFolderFile) $fName
   set TA(AutoPlayListSelection) [file rootname [file tail $fName]]
   if {$InFile!=0} {
      foreach {File} [aplFolder::foldersMatchingPattern * 1] {
         puts $InFile [list $File [aplFolder::Get $File]]
      }
      close $InFile
      if {[APLExists]} {
         Tree::setDirty $::TREE 0
      }
      set snackAmpSettings(lastFolderFile) $fName

      return $fName
   } else {
      return -1
   }
}

#------------------------------------------------------------------------------
# Function   : removeAPLFile
# Description: Remove an APL file, called from menu callback
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::removeAPLFile {fName} {
   set result [tk_messageBox -type yesno -title "Please Answer" \
      -icon question -message "Are you sure you want to remove APL setting '[file rootname [file tail $fName]]'?"]
   if {[string equal $result "yes"]} {
      if {[catch {file delete -force $fName} result] } then {
         tk_messageBox -type ok -title "Error" \
            -icon warning -message "Failed to remove '[file rootname [file tail $fName]]'\n$result"
      }

   }
}
#------------------------------------------------------------------------------
# Function   : removePLFile
# Description: Remove an playlist file, called from menu callback
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::removePLFile {fName {skip 0}} {
   if {$skip} {
      set result "yes"
   } else {
      set result [tk_messageBox -type yesno -title "Please Answer" \
         -icon question -message "Are you sure you want to delete playlist '[file rootname [file tail $fName]]'?"]
   }
   if {$result eq "yes"} {
      if {[catch {file delete -force $fName} result] } then {
         tk_messageBox -type ok -title "Error" \
            -icon warning -message "Failed to delete '[file rootname [file tail $fName]]'\n$result"
      }

   }
}
#------------------------------------------------------------------------------
# Function   : readAPLFile
# Description: Read a file and change the favorate folders
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::readAPLFile {fName {apply 0}} {
   Trace
   global TA snackAmpSettings
   Wait 1
   ;# Build default APL folders if not yet done
   # Save a corresponding playlist file
   set lf $::snackAmpSettings(lastFolderFile)
   if {[file exists $fName]} {
      if {[aplFolder::entries]==0} {
         aplFolder::createTeFolders 0
      }
      ;# Can we remove the option below?
      ##
      # TBD: Move this code below to aplFolder namespace
      #
      if {$::snackAmpSettings(savePlonApply) && [file exists $lf] && $lf!=$fName} {
         set plfile "[file rootname $lf].m3u"
         file delete -force $plfile
         SavePlaylistFile $::LB replace all $plfile
         Update
      }
      if {[catch {
         set fid [open $fName r]
      } result] \
      } then {
         tk_messageBox -icon warning -type ok -title "Open Error" -message $result
         return
      }
      fconfigure $fid -buffersize 100000
      set full [split [read $fid] \n]
      close $fid
      # Invalidate all current settings
      aplFolder::SetSome * -1
      # Only update values if an entry already exists from prior scan
      # This prevents garbage if directories were moved
      foreach line $full {
         foreach {Index Value} $line {
            aplFolder::SetIfExists $Index $Value
         }
      }
      Update
      ##
      # Some folders will still be tagged with -1 from
      # above invalidation, meaning they weren't in the new autoplaylist
      # file. If so, then set them to the value of their parent if possible.
      # Also, set any parents to a 2 if they have children with checks.
      #
      foreach {Index} [aplFolder::foldersMatchingPattern * 1] {
         aplFolder::setTeDefault $Index
      }
      if {[winfo exists .wFC]} {
         APL::refreshTree
      }
      ;# See if we need to apply it also
      if {$apply} {
         applyCurrentAPL
         Wait 1
      }
   } else {
      # If no default file create one with everything one
      ##
      # Create the default Folder list based on files
      # New ones are turned off by default.
      # We should already have the teFolders set to some value
      #
      aplFolder::SetSome * 1
      SaveFolderFile $fName
      applyCurrentAPL
      Wait 1
      waitWindow::close
#       set result [tk_messageBox -type yesno -title "SnackAmp - Please Respond" -icon question \
#          -message "All of your AutoPlaylist folders containing music are set true \nby default, would you like to change them now?"]
#       if {$result == "yes"} {
#          AutoPlaylist
#       }

   }
   set snackAmpSettings(lastFolderFile) $fName
   if {[APLExists]} {
      Tree::setDirty $::TREE 0
   }
   set TA(AutoPlayListSelection) [file rootname [file tail $fName]]
   Wait 0
}
#------------------------------------------------------------------------------
# Function   : saveCurrentAPL
# Description: Save and then apply the current AutoPlaylist
# Author     : Tom Wilkason
# Date       : 11/8/2001
#------------------------------------------------------------------------------
proc macros::saveCurrentAPL {{apply 1}} {
   global snackAmpSettings
   SaveFolderFile $snackAmpSettings(lastFolderFile)

   if {$apply} {
      applyCurrentAPL
   }
   if {[APLExists]} {
      Tree::setDirty $::TREE 0
   }
}
#------------------------------------------------------------------------------
# Function   : Clear
# Description: Clear the playlist
# Author     : Tom Wilkason
# Date       : 12/27/2000
#------------------------------------------------------------------------------
proc macros::Clear {args} {
   if {[tk_messageBox -icon question -type yesno -title "Please Respond" \
      -message "Are you sure you want to clear the playlist"]=="yes"} {
      global TA LB
      LB::deleteAll
   }
}
#------------------------------------------------------------------------------
# Function   : OpenFile
# Description: Open a single file for playing, add it to play list and play it
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::OpenFile {args} {
   global TA LB
   if {[catch {
      # This isn't in earlier snack versions. Also needs
      # Tcl 8.43+ to work, therefore the catch.
      set files [snack::getOpenFile -format MP3 -multiple 1]
   } result] } then {
      set files [snack::getOpenFile -format MP3]
   }
   # Here is a funky way to tell if a list of files was returned
   # Look at the string rep.
   if {![string match "\{*" $files]} {
      LB::InsertPlay $files
   } elseif {[llength $files] > 0} {
      LB::userInsertList $files
   }
}
#----------------------------------------------------------------------------
# Open a folder for insertion into the playlist
#----------------------------------------------------------------------------
proc macros::OpenFolder {args} {
   global TA LB
   set folder [chooseDirectory [pwd] $::SA]
   set files [list]
   Wait 1
   if {[file isdirectory $folder]} {
      foreach {track} [fileutils::rglob $folder] {
         Update
         if {[catalog::isTrackFile $track]} {
            lappend files $track
         }
      }
   }
   LB::userInsertList $files
   Wait 0
}
#------------------------------------------------------------------------------
# Open up a URL, save a shortcut and play it
# TODO: Handle all http files using TranslateFiles
#------------------------------------------------------------------------------
proc macros::OpenURL {args} {
   if {[playurl::checkuskfolder]} {
      set shortcut [playurl::addPromptUrl $::snackAmpSettings(uskDir)]
      if {$::playurl::ok} {
         LB::InsertPlay $shortcut
      }
   }
}
#------------------------------------------------------------------------------
# Function   : macros::Append
# Description: Wrapper for LB::AppendFile for macros & registry
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::Append {file} {
   LB::AppendFile $file
}
#------------------------------------------------------------------------------
# Function   : macros::PlayResume
# Description: Play button binding, resume play if paused otherwise
#              play active
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::PlayResume {args} {
   switch -- [soundControl::playState] {
      "Playing" {
         # Restart
         Play
      }
      "Paused" {
         # Resume play
         Pause
      }
      "Stopped" {
         # Play
         Play
      }
   }
}
#------------------------------------------------------------------------------
# Function   : macros::skipPlay
# Description: Skip some number of seconds in the playing track
# Author     : Tom Wilkason
# Date       : 8/9/2003
#------------------------------------------------------------------------------
proc macros::skipPlay {amount} {
   set elapsed [soundControl::position]
   set ::TA(setdrag) 1
   set newTime [expr {$elapsed+$amount}]
   if {$newTime < 0} {
      set newTime 0
   }
   gui::setTime $newTime
   PlayOffset
}
#------------------------------------------------------------------------------
# Function   : Play
# Description: Play a list item. Bound to Play key but used elsewhere.
#              Updates the GUI with the song info. Allows an offset within
#              the song to be passed in.
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::Play {{What active} {offset -1} {pause 0}} {
   Trace
   global TA LB snackAmpSettings
   ##
   # Determine if the current active index
   # is a valid song. If so play it. Otherwise, flag it and jump to the next one.
   #
   if {[LB::idxEnd]==0} {
      Stop
      set TA(currentMF) ""
      return
   }
   if {$snackAmpSettings(NextAfterCurrnt)} {
      set i [LB::idxPlaying]
      if {$i <= 0} {
         set i [LB::idxActive]
      }
   } else {
      set i [LB::idxActive]
   }
   ##
   # Based on prev, current or next position cursor to proper
   # file
   #
   switch -- $What {
      current {
         Stop
         if {$i == "" || $i<=0} {
            set i 0
         }
      }
      active {
         # Always use active (e.g double click)
         Stop
         set i [LB::idxActive]
         if {$i == "" || $i<=0} {
            set i 0
         }
      }
      next {
         # Only called, on automatic transition to next song
         # Can be called just before other song complete.
         # Don't issue a stop to keep the snack queue flowing.
         set soundControl::playState "Stopped"
         killAfter soundControl::Timer
         gui::setTime 0
         if {$i == "" || $i<0} {
            set i 0
         } else {
            incr i
         }
         # check for continuous play
         if {$i >= [LB::idxEnd]} {
            if {$snackAmpSettings(ContinuousPlay)} {
               set i 0
               # Resort on a wrap if some form of random play
               filtering::sortPlaylist
            } else {
               return
            }
         }
      }
      prev {
         if {$i == ""} {
            set i 0
         } elseif {$i==0} {
            set i [expr {[LB::idxEnd]-1}]
         } else {
            if {$i > 0} {
               incr i -1
            }
         }
      }
      default {
         saLog "Invalid subcommand :Play $What"
      }
   }
   ##
   # Get the track name at this index
   #
   set SelectionName [LB::nameFromIdx $i]
   set TA(currentMF) $SelectionName

   ##
   # If file does not exist then assume it needs to be mounted
   #
   if {[string length $SelectionName] == 0} {
      # macros::Highlight this file then skip to the next one
      # if we haven't been commanded to stop
      saLog "Play: Could not validate '[LB::nameFromIdx $i]' as valid file"
      set ::songArray(Title) "'[LB::nameFromIdx $i]' invalid"
      set ::songArray(Artist) "File not verifiable"
      set ::songArray(Album) "Error"
      LB::activate $i
      set LB::indexPlaying $i
      LB::invalidateIndex $i
      if {$::TA(STOP) == 0} {
         after 1000 Next
         # Next will clear background so delay
         after 1500 [list LB::invalidateIndex $i]
      }
      return
   }
   # File exists, inverse video it showing it is playing
   # large playlists are slow to inverse, so delay the inverse
   # till after track is playing
   set LB::indexPlaying $i
   after idle [list LB::inverseVideoIndex $i]
   ##
   # If this is a url to play, then play it directly
   #
   if {[playurl::isusk $SelectionName]} {
      soundControl::reset url
      foreach {index} [array names ::songArray] {
         set ::songArray($index) ""
      }
      if {[playurl::checkuskfolder]} {
         set ::TA(dummyName) [file join $::snackAmpSettings(uskDir) dummy.mp3]
         # Cover art
         if {$::snackAmpSettings(showCoverArt)} {
            after 20 {gui::packCoverArt $::TA(dummyName) $::snackAmpSettings(fadeImages)}
            set delay 3000
         } else {
            set delay 2000
         }
         # Small cover art (only if good resizer available)
         if {$::snackAmpSettings(smallArt) && $gui::resize} {
            after 50 {gui::smallCoverArt $::TA(dummyName)}
         }
         # Miniplay cover art
         if {[winfo exists .miniPlayer] && [winfo ismapped .miniPlayer]} {
            after 75 {miniPlayer::smallCoverArt $::TA(dummyName)}
         }
         # Fading pop-up
         if {$::snackAmpSettings(fadePopup)} {
            trackTransition::cancel
            after $delay trackTransition::show $::snackAmpSettings(sliderSpeed)
         }
      }

      # TODO: should we read from the database instead?
      id3Tag::id3Label $SelectionName ::songArray $::snackAmpSettings(ID3ReadOrder)
      set ::songArray(Album) "$SelectionName"
      set max [expr {3600*24}]      ;# 24 hours
      set gui::songDuration 0;#$max
      $gui::tsScale configure -to $max
      catch {$miniPlayer::mpSS configure -to $max}
      catch {$gui::mpSS configure -to $max}
      set songTitle [string map {// /} [file rootname [join [lrange [file split $TA(currentMF)] end-1 end] /]]]
      wm title $::SA "$songTitle"
      set TA(songTitle) $songTitle
      soundControl::Timer
      soundControl::pauseRelief 0

      if {[catch {playurl::playusk [soundControl::sound] $SelectionName [soundControl::getNextCommand] ::songArray} result] } then {
         saLog $::errorInfo
         set ::songArray(Title) "Error: $result"
         after 200 [list eval [soundControl::getNextCommand]]
         return
      }
      return
   }
   ##
   # Process a file
   #
   ##
   # It seems faster on play -start to have the short file name and cd to that directory
   # Also, UNC names are really slow (on Win2K anyway)!
   # Strangly, network drive mappings are also fast (if not using UNC)
   #
   if {![file exists $SelectionName]} {
      saLog "$SelectionName does not exist"
      after 500 [list LB::invalidateIndex $i]
      after 100 Next
      return
   }
   soundControl::reset [string tolower [file extension $SelectionName]]
#   soundControl::reset file
   cd $::HOME

   if {[catch {soundControl::fileName $SelectionName} result]} {
      tk_messageBox -icon warning -type ok -title "sound error" \
         -message "Error creating sound object using $SelectionName\n$result\n$::errorInfo"
      return
   }
   ##
   # Check for an invalid header
   #
   foreach {sndSamples sndRate x x x x sndExt} [soundControl::sinfo] {break}
   if {$sndSamples == -1 || $sndRate==0 || [catalog::isTrackFile ".$sndExt"] < 0} {
      soundControl::fileName ""
      saLog "Invalid file: $SelectionName"
      set ::songArray(Title) "Invalid header:$SelectionName"
      update
      if {$snackAmpSettings(AutoDelete)} {
         Delete
      }
      after 1000 [list eval [soundControl::getNextCommand]]
      return
   }
   # #
   # Get ID3 and detailed song info and display it, if ID3 not available
   # Use file name as song title

   catalog::trackTextData $SelectionName ::songArray
   set ::songArray(Rating) [mrating $::songArray(Rating)]
   if {$::songArray(ID3)=={}} {
      set ::songArray(Title) [stringTitle [file rootname [file tail $TA(currentMF)]]]
   }
   # This should be use when applying an APL
   if {$What == "next"} {
      # Check for minimum rating before play
      set rating [ratingToNum [db::getVal $SelectionName Rating]]
      if {$rating > 0 && ($rating < [ratingToNum $::snackAmpSettings(minTxtRating)])} {
         set ::songArray(Title) "(Rating) Skipping $::songArray(Title)"
         saLog "(Rating) Skipping $SelectionName"
         if {$::TA(STOP)==0} {
            after idle Next
         }
         return
      }
      ##
      # Verify minimum bit rate
      #
      set rate [lindex $::songArray(Rate) 0]
      if {$rate < $::TA(minRate)} {
         set ::songArray(Title) "(Bitrate) Skipping $::songArray(Title)"
         saLog "(Bitrate) Skipping $SelectionName"
         if {$::TA(STOP)==0} {
            after idle Next
         }
         return
      }
   }
   ##
   # Adjust gain if the song has an auto level setting (and it is enabled)
   #
   if {$snackAmpSettings(autoLevel)} {

      set value [gainControl::createIfNeeded $SelectionName]
      ##
      # Pre compute the next track in the playlist if possible
      # Wait about 20 seconds to do so
      if {$snackAmpSettings(autoLevelPreComute)} {
         set nextSelection [LB::nameFromIdx [expr {$i+1}]]
         if {[string length $nextSelection]} {
            gainControl::bgCreate $nextSelection
         }
      }
      # Clamp db Gain at runtime
      set value [fromDb [gainControl::limit [toDb $value]]]
      soundControl::setTrackGain $value
   } else {
      soundControl::setTrackGain 1.0
   }
   # Fires Next command at end of this file
   if {$offset < 0} {
      set offset [gainControl::getStart [soundControl::fileName]]
   }
   # Cover art
   if {$::snackAmpSettings(showCoverArt)} {
      after 20 {gui::packCoverArt [soundControl::fileName] $::snackAmpSettings(fadeImages)}
      set delay 1000
   } else {
      set delay 10
   }
   # Small cover art (only if good resizer available)
   if {$::snackAmpSettings(smallArt) && $gui::resize} {
      after 50 {gui::smallCoverArt [soundControl::fileName]}
   }
   # Miniplay cover art
   if {[winfo exists .miniPlayer] && [winfo ismapped .miniPlayer]} {
      after 75 {miniPlayer::smallCoverArt [soundControl::fileName]}
   }
   # Fading pop-up
   if {$::snackAmpSettings(fadePopup)} {
      trackTransition::cancel
      after $delay trackTransition::show $::snackAmpSettings(sliderSpeed)
   }
   # Enough already, play the darn track!
   soundControl::play \
      -start $offset \
      -command [soundControl::getNextCommandBump [soundControl::fileName] $::snackAmpSettings(trackGap)]
   if {$pause} {
      Pause
   }
   # Only log this if going to a file
   if {$::snackAmpSettings(logFile)} {
      saLog "Playing $TA(currentMF)"
   }
   ##
   # What ~10 ticks on 5 or multiple sec intervals
   #
   set trackTime [soundControl::duration]
   #$gui::tsScale configure -to $trackTime
   if {[string is double -strict $trackTime]} {
      $gui::tsScale configure -to $trackTime
      catch {$gui::mpSS configure -to $trackTime}
      # Miniplayer (catch if not up)
      if {[info exists miniPlayer::mpSS] && [winfo exists $miniPlayer::mpSS]} {
         $miniPlayer::mpSS configure -to $trackTime
      }
   }
   ##
   # Song Title is shown in the title bar
   #
   set songTitle [string map {// /} [file rootname [join [lrange [file split $TA(currentMF)] end-1 end] /]]]
   # If the folder name is the same as the first part of file name,
   # Then we can ignore the folder name
   if {[string match -nocase "[file dirname $songTitle]*" [file tail $songTitle]]} {
      set songTitle [file tail $songTitle]
   }
   wm title $::SA "$songTitle"
   set TA(songTitle) $songTitle
   set TA(playlistEntry) [LB::nameToKey $TA(currentMF)]

   set gui::songDuration $trackTime
   if {$snackAmpSettings(autoSavePlaylist)} {
      catalog::SavePlaylistCache
   }
   LB::playingKeyUpdate
   soundControl::Timer
}
#----------------------------------------------------------------------------
# Play next sound file in list
#----------------------------------------------------------------------------
proc macros::Next {} {
   Trace
   Stop
   Play next
}
#----------------------------------------------------------------------------
# Play the next track right now
#----------------------------------------------------------------------------
proc macros::NextNow {} {
   Trace
   Stop
   Play next
}
#----------------------------------------------------------------------------
# Play previous sound file in list
#----------------------------------------------------------------------------
proc macros::Prev {} {
   Trace
   Stop
   Play prev
}
#----------------------------------------------------------------------------
# Shuffle the selected files in the playlist
#----------------------------------------------------------------------------
proc macros::shuffleSelection {args} {
   LB::shuffleSelection ::LB
}
#----------------------------------------------------------------------------
# Sort the selected files in the playlist
#----------------------------------------------------------------------------
proc macros::sortSelection {args} {
   LB::sortSelection ::LB
}
#----------------------------------------------------------------------------
# Stop playback of a file and cancle the update timer
#----------------------------------------------------------------------------
proc macros::Stop {args} {
   Trace
   global TA LB
   killAfter soundControl::Timer
   after cancel Next
   soundControl::stop
   gui::setTime 0
   soundControl::pauseRelief 2
   playurl::cleanupLast
}
##
# Stop and clear any open files
#
proc macros::resetStop {args} {
   Stop
   soundControl::resetObj
}

#------------------------------------------------------------------------------
# Move the cursor to the currently playing file,
# assign to <Home> key
#------------------------------------------------------------------------------
proc macros::locatePlaying {args} {
   LB::seePlaying
}

#------------------------------------------------------------------------------
# Pause/resume playback
#------------------------------------------------------------------------------
proc macros::Pause {args} {
   Trace
   global TA LB
   if {[snack::audio active]} {
      if {$soundControl::paused==1} {
         soundControl::pause 0
         soundControl::Timer
      } else {
         killAfter soundControl::Timer
         soundControl::pause 1
      }
   }
}
#------------------------------------------------------------------------------
# Callback to remove the current selections from the display
#------------------------------------------------------------------------------
proc macros::Cut {{playNext 0}} {
   Trace
   global TA LB
   set playActive 0
   set playing [LB::idxPlaying]
   if {[llength [LB::idxSelected]] > 0} {
      set Index [LB::idxAnchor]
      foreach {ind} [lsort -decreasing -integer [LB::idxSelected]] {
         if {$playing==$ind} {
            set playActive [soundControl::isPlaying]
            if {$playActive} {
               Stop
            }
         }
         LB::deleteIndex $ind
      }
      # Stay on 0 if that is what we were on
      ##macros::Highlight [expr {$Index-0}]
      gui::updateFileDisplay
      # Set selection to current active
      LB::setSelection [LB::idxActive]
      # This will play the next song
      if {$playNext && $playActive} {
         Play
      }
   }
}
#------------------------------------------------------------------------------
# Cut the listbox to the clibboard
# (only used for main playlist)
#------------------------------------------------------------------------------
proc macros::clipboardCut {lb {append 1}} {
   Trace
   global TA
   if {[llength [$lb curselection]] > 0} {
      clipboardCopy $lb $append
      Cut 1
      gui::updateFileDisplay
   }
}
#------------------------------------------------------------------------------
# Copy the listbox files to the clipboard
#------------------------------------------------------------------------------
proc macros::clipboardCopy {lb {append 1}} {
   Trace
   global TA LB
   if {[llength [$lb curselection]] > 0} {
      if {$append == 0} {
         clipboard clear
      } else {
         clipboard append " "
      }
      set data [getSelectedTracks $lb]
      clipboard append $data
   }
}
#------------------------------------------------------------------------------
# Paste the listbox files to the playlist
# Note: Only works with LB playlist
#------------------------------------------------------------------------------
proc macros::clipboardPaste {lb {clear 1}} {
   Trace
   global TA
   if {[catch {selection get -selection CLIPBOARD} files]} {
      return
   }
   set current [$lb index anchor]
   set processed [list]
   foreach {file} $files {
      # Only insert files that are in the master list
      # And not already processed (i.e. no duplicates)
      #W: move to shared
      if {[lsearch -exact $processed $file] < 0 && [db::whereInDb $file] >= 0} {
         lappend processed $file
      }
   }
   if {[llength $processed]>0} {
      LB::insertList [LB::idxActive] $processed
   }
   if {$clear} {
      clipboard clear
   }
}
#------------------------------------------------------------------------------
# Callback to remove the current selections from the disk
#------------------------------------------------------------------------------
proc macros::Delete {{Next 1}} {
   Trace
   global TA LB
   set goNext 0
   set toDelete [list]
   if {[llength [LB::idxSelected]] > 0} {
      if {$::snackAmpSettings(dontMoveFiles)} {
         tk_messageBox -icon warning -type ok \
            -message "You don't have permission to remove files"
         return $goNext
      }
      #set Index [LB::idxAnchor]

      set selection [LB::idxSelected]
      foreach {idx} $selection {
         lappend toDelete [LB::nameFromIdx $idx]
      }

      set result [tk_messageBox -type yesno -title "Please Answer" \
         -icon question \
         -message "Are you sure you want to delete the selected files?\n[join $toDelete \n]"]
      if {$result == "yes"} {
         foreach {File} [lsort -unique $toDelete] {
            LB::removeFile $File
            catalog::DeleteFile $File
         }
         ##macros::Highlight $Index
      }
   }
   ##
   # Play the next song if we deleted the current one and we want to
   #
   if {$Next && $goNext} {
      Play
   }
   gui::updateFileDisplay
   return $goNext
}
#----------------------------------------------------------------------------
# Bring up or down the Auto Playlist window, depending on its state
#----------------------------------------------------------------------------
proc macros::ToggleArt {args} {
   gui::ToggleArt
}
proc macros::AutoPlaylist {args} {
   if {![winfo exists .wFC]} {
      package require saTree
      package require autoPlayList
      $::TA(treeButton) configure -relief sunken
      # First make sure we have the proper roots for the tree viewer
      APL::FolderTree
      update  ;# let it draw
      # May not have a selection
      catch {APL::SelectFile $::TA(currentMF)}
   } elseif {[wm state .wFC] != "normal"} {
      #overrelief setting forces flat when mouse over
      $::TA(treeButton) configure -relief sunken
      APL::FolderTree
      update  ;# let it draw
      catch {APL::SelectFile $::TA(currentMF)}
   } else {
      wm withdraw .wFC
      $::TA(treeButton) configure -relief flat
   }
}
#------------------------------------------------------------------------------
# Function   : macros::findNextSong
# Description: Locate the next occurance of a song in the playlist based on the
#              current filter value
# Author     : Tom Wilkason
# Date       : 11/08/2001
#------------------------------------------------------------------------------
proc macros::findNextSong {{top 0}} {
   global TA snackAmpSettings
   # Get the current playlist
   set regPattern [list]                                ;#retain wildcards
   set pattern [split [string map {" or " "|" " OR " "|" "*" ".*" "?" "."} [catalog::globEscape $TA(filter)]] |]
   foreach part $pattern {
      lappend regPattern "($part)+"
   }
   set regPattern [join $regPattern |]
   if {$top} {
      set i -1
   } else {
      set i [lindex [LB::idxSelected] 0]
      if {$i=={}} {
         set i -1
      }
   }
   set ok 2
   ##
   # Allow one wrap around if at first not found
   #
   if {[catch {
      while {$ok} {
         incr i
         foreach {Entry} [LB::nameFromIdx $i end] {
            if {[regexp -nocase -- $regPattern $Entry]} {
               macros::Highlight $i
               return
            }
            incr i
         }
         set i -1
         incr ok -1
      }
   } result]} {
      #saLog "findNextSong $result"
   }
}
#------------------------------------------------------------------------------
# Function   : macros::playSpecificSong
# Description: Locate and Play a specific song in the Playlist, ignore if not in playlist
# Author     : Tom Wilkason
# Date       : 4/28/2002
#------------------------------------------------------------------------------
proc macros::playSpecificSong {song} {
   Trace
   set i 0
   set i [lsearch -exact [LB::keyOrNamesAll 1] $song]
   if {$i >= 0} {
      LB::activate $i
      Play active
   }
}
#------------------------------------------------------------------------------
# Function   : macros::clearFilter
# Description: Clear the current filter entry box
# Author     : Tom Wilkason
# Date       : 11/9/2001
#------------------------------------------------------------------------------
proc macros::clearFilter {args} {
   set ::TA(filter) {}
}

#------------------------------------------------------------------------------
# Function: macros::selectAll
#
# Select all the files in the playlist
#------------------------------------------------------------------------------
proc macros::selectAll {args} {
   LB::setSelection 0 end
}
#------------------------------------------------------------------------------
# Function   : macros::filterMusic
# Description: Restrict files shown based on the filter & folder settings
# Author     : Tom Wilkason
# Date       : 12/26/2000
#------------------------------------------------------------------------------
proc macros::filterMusic {how {what def}} {
   filtering::filterMusic $how $what
}
#------------------------------------------------------------------------------
# Return selected tracks
#------------------------------------------------------------------------------
proc macros::selectedTracks {ListBox {what "selected"}} {
   set names [list]
   if {$ListBox == $::LB} {
      if {$what == "selected"} {
         set Indicies [LB::idxSelected]
      } else {
         set Indicies [LB::idxAll]
      }
      foreach {Index} $Indicies {
         lappend names [LB::nameFromIdx $Index]
      }
   } elseif {$ListBox == $::PLB} {
      if {$what == "selected"} {
         set names [aplLB::nameSelected]
      } else {
         set names [aplLB::getAllNames 1]
      }
   } else {
      saLog "unknown listbox '$ListBox'"
   }
   return $names
}
#------------------------------------------------------------------------------
# Function   : macros::SavePlaylistFile
# Description: Save current files in list
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::SavePlaylistFile {ListBox {how "replace"} {what "selected"} {fName {}}} {
   Trace
   global TA LB snackAmpSettings
   ##
   # If a list for GlobData is passed in, the second item should be the name to save as (if different)
   #
   if {$how=="replace"} {
      set access "w"
   } else {
      set access "a+"
   }
   # Use different dialog depending on what you want to do
   array set tk_dialog {
      replace tk_getSaveFile
      append tk_getOpenFile
   }
   if {$what == "selected"} {
      set Indicies [LB::idxSelected]
   } else {
      set Indicies [LB::idxAll]
   }
   if {[string length $fName]==0} {
      set types [list [list "Playlist files" {.m3u .taf .pls}] [list "All Files" "*"]]

      set fName [$tk_dialog($how) -title "Save [llength $Indicies] tracks to Playlist File" -filetypes $types \
         -initialdir $::HOME -defaultextension .m3u]
      if { $fName == "" } {
         return  ""
      }
      # Set default extension to m3u
      if {[file extension $fName]==""} {
         set fName "$fName.m3u"
      }
   } elseif {[file exists $fName]} {
      if {$how=="replace" && [tk_messageBox -icon question -type yesno -title "Overwrite File" -message "Is it ok to $how $fName with [llength $Indicies] tracks?"]!="yes"} {
         return {}
      }
   }
   if {[catch {open $fName $access} InFile]} {
      tk_messageBox -icon warning -type ok -title "File Error" -message "Save Error: Unable to open '$fName' for writing"
      return {}
   }
   set TA(PlayListSelection) [file rootname [file tail $fName]]
   set snackAmpSettings(PlayListFile) $fName
   set ext [file extension $fName]
   if {[string equal $ext .m3u]} {
      puts $InFile "#\n# Playlist file: [file nativename $fName]"
      puts $InFile "# Saved on [clock format [clock seconds]] by user $::tcl_platform(user) on host [info hostname]\n#"
   }
   if {$InFile!=0} {
      ##
      # Process either selected or entire list
      #
      set names [list]
      set fldPatt [file join [file dirname $fName] *]
      set ml [string length [string trimright [file dirname $fName] / ]]
      incr ml
      if {$ListBox == $::LB} {
         foreach {Index} $Indicies {
            set File [LB::nameFromIdx $Index]
            # Todo: Strip front part if it can be made relative to pwd
            if {[string equal $ext .m3u]} {
               # Make file relative if possible
               if {[string match $fldPatt $File]} {
                  set File [string range $File $ml end]
               }
               set File [file nativename $File]
            }
            puts $InFile $File
         }
      } elseif {$ListBox == $::PLB} {
         if {$what == "selected"} {
            set names [aplLB::nameSelected]
         } else {
            set names [aplLB::getAllNames 1]
         }
         foreach {File} $names {
            if {[string equal $ext .m3u]} {
               if {[string match $fldPatt $File]} {
                  set File [string range $File $ml end]
               }
               set File [file nativename $File]
            }
            puts $InFile $File
         }
      } else {
         saLog "unknown listbox '$ListBox'"
      }
      #W: Only save if playing is in list
      if {[lsearch -exact $names $::TA(currentMF)]>= 0} {
         puts $InFile "#Playing=$::TA(currentMF)"
      }
      close $InFile
      return $fName
   } else {
      return -1
   }
}
#------------------------------------------------------------------------------
# Function   : macros::ReadPlaylistFile
# Description: Verify playlist file and notify user of any problems
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::ReadPlaylistFile {fName {how "replace"}} {
   # If file doesn't exist, then prompt for it
   if {$fName == {} || ![file exists $fName]} {
      set types [list [list "Playlist files" {.m3u .taf .pls}] [list "All Files" "*"]]
      set fName [tk_getOpenFile -title "Open Playlist File" -filetypes $types \
         -initialdir $::HOME]
      if { $fName == "" } {
         return  ""
      }
   }
   set failed [ProcessPlaylistFile $fName $how 1]
   set numFailed [llength $failed]
   if {$numFailed > 0} {
      ##
      # Tell user how many failed (if any)
      #
      if {$numFailed > 10} {
         set failed [lrange $failed 0 9]
         lappend failed "...."
      }
      tk_messageBox -icon warning -type ok -title "Invalid Files" \
         -message "A total of $numFailed files do no exist, and could not be added to the playlist\n[join $failed \n]"
   }
   return $numFailed
}
#------------------------------------------------------------------------------
# Function   : macros::ProcessPlaylistFile
# Description: Read a file and append/replace current list
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::ProcessPlaylistFile {fName {how "replace"} {resume 0}} {
   Trace
   global TA LB snackAmpSettings
   if {[catch {
      set fid [open $fName r]
   } result] \
   } then {
      tk_messageBox -icon warning -type ok -title "Open Error" -message $result
      return
   }
   Wait 1
   set full [split [string trim [read $fid]] \n]
   close $fid
   # Insert the files if they are not already in the list
   # Also support winamp format playlists, this means no files can
   # start with # or else they will be ignored (that is a comment in m3u)
   set here [file dirname $fName]
   set push [pwd]
   cd $here
   set comment "\#*"
   set failed [list]
   set i 0
   set cp [LB::idxActive]
   set toinsert [list]
   set processLines [llength $full]
   ##
   # Process each row
   #
   foreach file $full {
      incr i
      # Based on file extension
      switch -- [string tolower [file extension $fName]] {
         .taf -
         .m3u {
            # standard playlist handling
            if {! [string match $comment $file]} {
               # If a url, create a usk and process the usk file
               if {[string match -nocase "http://*" $file]} {
                  # Create a usk (url) shortcut, then append it
                  # Note: trying tail of file name, works if a file if a stream
                  # may not work as well.
                  set file [playurl::addUrl $::snackAmpSettings(uskDir) [file tail $file] $file]
               } else {
                  # This will be ignored if abs path names
                  set file [file join $here $file]
                  # Verify existence (can be slow on a net)
                  if {[file exists $file]} {
                     lappend toinsert $file
                  }
               }
               if {$i%50==0} {
                  waitWindow::percent $i $processLines "Playlist Files Verified..." 1
                  if {[waitWindow::ifbreak]} {
                     break
                  }
               }
            } else {
               if {$resume && [string match "#Playing*" $file]} {
                  set resumeFile [lindex [split $file =] 1]
                  if {$::snackAmpSettings(restorePlaying) && [file exists $resumeFile]} {
                     set ans [tk_messageBox -type yesno -icon question -message "Do you want to resume playing of \n$resumeFile"]
                     if {$ans=="yes"} {
                        after 1000 [list PlayFile $resumeFile]
                     }
                  }
               }
            }
         }
         .pls {
            # pls handling
            # valid entries are Filen=...
            # skip the others
            if {[regexp -- {File([0-9]+)=(.*)} $file -> Fnum url]} {
               if {[string match -nocase "http://*" $url]} {
                  lappend toinsert [playurl::addUrl $::snackAmpSettings(uskDir) [file tail $url] $url]
               } else {
                  # This will be ignored if abs path names
                  set file [file join $here $url]
                  # Verify existence (can be slow on a net)
                  if {[file exists $url]} {
                     lappend toinsert $url
                  }
               }
            }
         }
      }
   }
   ;# Insert the tracks
   if {[llength $toinsert] > 0} {
      if {$how=="replace"} {
         LB::deleteAll
         LB::insertList 0 $toinsert
      } else {
         LB::insertList $cp $toinsert
      }
   }
   Wait 0
   waitWindow::close
   set snackAmpSettings(PlayListFile) $fName
   set TA(PlayListSelection) [file rootname [file tail $fName]]
   cd $push
   gui::updateFileDisplay
   return $failed
}
#------------------------------------------------------------------------------
# Function   : macros::Title
# Description: Change the title to the last file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::Title {args} {
   Trace
   global snackAmpSettings
   if {[APLExists]} {
      wm title .wFC "SnackAmp AutoPlaylist Manager - [file rootname [file tail $::snackAmpSettings(lastFolderFile)]]"
   }
   set snackAmpSettings(PlayListFile) "[file rootname $snackAmpSettings(lastFolderFile)].m3u"
}

#------------------------------------------------------------------------------
# Function   : macros::Insert
# Description: Wrapper for LB::userInsertList for macros & registry
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::Insert {file} {
   LB::userInsertList [list $file]
}
#------------------------------------------------------------------------------
# Try and make a file writable
#------------------------------------------------------------------------------
proc macros::makeWritable {file} {
   if {[file exists $file]} {
      foreach {attrib} {
         -hidden
         -system
         -readonly
      } {
         set val [file attributes $file $attrib]
         lappend attribs $val
         if {$val} {
            file attribute $file $attrib 0
         }
      }
   } else {
      set attribs [list]
   }
   return $attribs
}
#----------------------------------------------------------------------------
# Restore a set of previous attributes
#----------------------------------------------------------------------------
proc macros::restoreAttribs {file attribs} {
   if {[file exists $file]} {
      if {[llength $attribs] == 3} {
         foreach {attrib} {-hidden -system -readonly} saved $attribs {
            set new [file attributes $file $attrib]
            if {$new != $saved} {
               file attribute $file $attrib $saved
            }
         }
      } else {
         foreach {attrib} {
            -hidden
            -system
            -readonly
         } {
            lappend attribs [file attributes $file $attrib]
         }
      }
   }  else {
      set attribs [list]
   }
   return $attribs
}
#------------------------------------------------------------------------------
# Function   : PlayFile
# Description: Play a file at the top of the list (linked to DDE command on Windows)
#              Set this up early so a file association doesn't time out when launching
#              SnackAmp on a slow computer.
#              TODO: Use TranslateFiles to translate, this to insert and play
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::PlayFile {file {justInsert 0}} {
   set toappend [TranslateFiles [list $file]]
   ;# If we inserted some, and not just inserting, play the first one
   if {[llength $toappend]} {
      set end [LB::userInsert]
      LB::userInsertList $toappend
      if {$justInsert==0} {
         LB::activate $end
         Play active
      }
   }
   return
}

#------------------------------------------------------------------------------
# Save all the settings, don't exit
#------------------------------------------------------------------------------
proc macros::saveSettings {args} {
   global TA snackAmpSettings
   catalog::SavePlaylistCache
   # Save last width if not zoomed
   if {[wm state $::SA] ne "zoomed"} {
      set snackAmpSettings(minWidth) [winfo width $::SA]
   }
   scan [wm geometry $::SA] "%dx%d+%d+%d" w h x1 x2
   set snackAmpSettings(playerPos) [list $x1 $x2]
   set snackAmpSettings(mediaDirs) [lsort -unique $snackAmpSettings(mediaDirs)]
   if {[winfo exists .albumart]} {
      scan [wm geometry .albumart] "%dx%d+%d+%d" w h x1 x2
      set snackAmpSettings(albumartPos) "+$x1+$x2"
   }
   settings::saveSettings
   aplFolder::saveFolderNames
   db::commit
   saLog "Settings Saved"
}
#------------------------------------------------------------------------------
# Function   : Exit
# Description: Exit gracefully
# Author     : Tom Wilkason
# Date       : 1/9/2001
#------------------------------------------------------------------------------
proc macros::taExit {{exit 1}} {
   global TA snackAmpSettings
   catch {
      winico_delall
   }
   catch {APL::close}
   macros::saveSettings
   soundControl::stop
   HttpD::Shutdown
   db::closeDB
   # close any remaining channels
   foreach {channel} [file channels] {
      catch {close $channel}
   }
   if {$exit} {
      exit
   }
}
#------------------------------------------------------------------------------
# Copy the selection to a staging file for burning to a CD
#------------------------------------------------------------------------------
proc macros::linkToStaging {lb {append 1}} {
   Trace
   global TA LB
   set data [list]
   if {[llength [$lb curselection]] > 0} {
      foreach {ind} [lsort -decreasing -integer [$lb curselection]] {
         switch -glob -- $lb {
            ".saTop*"  {lappend data [LB::nameFromIdx $ind]}
            ".wFC*"    {lappend data [aplLB::rowIndex2file $ind]}
         }
      }
   }
   ;# Create N lists of files, for each root drive
   foreach file [lflip $data] {
      lappend toStage([fileRoot $file]) $file
   }
   ;# For each root drive, process the files
   foreach folder [array names toStage] {
      set staging [file join $folder "staging"]
      if {[file exists $staging]} {
         ;# First, find the maximum number to start with
         set i 0
         set existing [lsort -dictionary [glob -nocomplain [file join $staging *]]]
         foreach track $existing {
            set front [string trimleft [lindex [split [file tail $track] -] 0] "0"]
            if {[string is integer -strict $front] && $front > $i} {
               set i $front
            }
         }
         incr i
         saLog "Starting with $i"
         ##
         # Now link the files in
         #
         foreach file [lflip $toStage($folder)] {
            set num [format "%04d - " $i]
            set newName [file join $staging $num[file tail $file]]
            if {![file exists $newName]} {
               saLog "Creating $newName"
               if {[catch {file link $newName $file} result]} {
                  saLog $result
               } else {
                  incr i
               }
            }
         }
      } else {
         puts "$staging doesn't exist"
      }
   }
}
#----------------------------------------------------------------------------
# mp3gain workspace
#----------------------------------------------------------------------------
namespace eval mp3gain {
   variable STOP 0
   variable DONE 0
   variable QUEUE [list]
   variable mp3gainLoc
   upvar #0  snackAmpSettings(mp3gainLoc) mp3gainLoc
}
#----------------------------------------------------------------------------
# When a file is readable, dump output to the console
#----------------------------------------------------------------------------
proc mp3gain::isReadable {f} {
   variable STOP
   variable DONE
   # The channel is readable; try to read it.
   # Note: The initial eof check isn't really needed
   if {[eof $f]} {
      # End of file on the channel
      catch {fileevent $f readable {}}
      catch {close $f}
      saLog "end of file"
      set DONE 1
   } else {
      set status [catch {gets $f line} result]
      if { $status != 0 } {
         # Error on the channel
         saLog "error reading $f: $result"
         set DONE 2
      } elseif {$result >= 0} {
         # Successfully read the channel
         saLog "> $line"
         set line [string trim $line]
         if {[string length $line]>5} {
            waitWindow::show "mp3gain in progress\n\n$line" 1
         }
      } elseif {[fblocked $f]} {
         # Read blocked.  Just return
      } elseif {[eof $f]} {
         catch {fileevent $f readable {}}
         catch {close $f}
         saLog "end of file on read"
         set DONE 1
      } else {
         # Something else
         saLog "can't happen"
         set DONE 3
      }
   }
}

#------------------------------------------------------------------------------
# Launch and block on execution of a program
#------------------------------------------------------------------------------
proc mp3gain::launch {folder cmd} {
   Trace
   variable STOP
   variable DONE
   set DONE 0
   # Open a pipe
   cd $folder
   set fid [open "|$cmd"]
   saLog "Opening $fid";update
   # Set up to deliver file events on the pipe
   fconfigure $fid -buffering line -blocking 0
   fileevent $fid readable [list mp3gain::isReadable $fid]
   # Launch the event loop and wait for the file events to finish
   vwait mp3gain::DONE
   # Close the pipe
   catch {close $fid}
   saLog "Closing $fid";update
}
#------------------------------------------------------------------------------
# Create an mp3gain command file
#------------------------------------------------------------------------------
proc mp3gain::mp3gain {recurse {opts "/a /d 3.0"}} {
   Trace
   variable STOP
   variable QUEUE
   variable mp3gainLoc
   lappend folders $::TA(highlightPath)
   if {$recurse} {
      foreach {folder} [aplFolder::foldersMatchingPattern [file join $::TA(highlightPath) *] 1] {
         lappend folders $folder
      }
      saLog "Processing [llength $folders]"
   }
   foreach key $folders {
      if {[waitWindow::ifbreak]} {
         break
      } else {
         if {[file exists $key]} {
            waitWindow::show "mp3gain in progress\n\nProcessing '$key'" 1
            saLog "Processing '$key'";update
            mp3gain::launch $key "$mp3gainLoc $opts /k /q *.mp3"
            if {[waitWindow::ifbreak]} {
               break
            }
            gainControl::resetSome [db::filesInFolder $key]
            if {[string match "*/a*" $opts]} {
               saLog "Computing folder average";update
               gainControl::averageSome [db::filesInFolder $key]
            } else {
               saLog "Computing track averages";update
               gainControl::computeSome [db::filesInFolder $key]
            }
         }
      }
   }
   waitWindow::close
}
#----------------------------------------------------------------------------
# Gain on some passed in folders
#----------------------------------------------------------------------------
proc mp3gain::mp3gainFolders {folders {opts "/a /d 3.0"}} {
   variable STOP
   variable QUEUE
   variable mp3gainLoc
   foreach key $folders {
      if {[waitWindow::ifbreak]} {
         break
      } else {
         if {[file exists $key]} {
            waitWindow::show "mp3gain in progress\n\nProcessing '$key'" 1
            saLog "Processing '$key'";update
            mp3gain::launch $key "$mp3gainLoc $opts /k /q *.mp3"
            if {[waitWindow::ifbreak]} {
               break
            }
            gainControl::resetSome [db::filesInFolder $key]
            if {[string match "*/a*" $opts]} {
               saLog "Computing folder average";update
               gainControl::averageSome [db::filesInFolder $key]
            } else {
               saLog "Computing track averages";update
               gainControl::computeSome [db::filesInFolder $key]
            }
         }
      }
   }
   waitWindow::close
}

#------------------------------------------------------------------------------
# Macros to prepend the year to the folder name
# Just for Tom
#------------------------------------------------------------------------------
proc macros::fixFolderChildren {folder} {
   set i -1
   set subs [aplFolder::getFolders $folder 1]
   if {[llength $subs]>1} {
      foreach {Index} [lrange $subs 1 end] {
         set tlen [llength [lrange $subs 1 end]]
         set front [lindex [file tail $Index] 0]
         if {![string is integer -strict $front]} {
            waitWindow::percent [incr i] $tlen "Checking $Index for year" 1
            set newName [macros::prependFolderYear $Index]
            #puts "Renamed $Index to $newName"
            if {[waitWindow::ifbreak]} {
               return
            }
         }
      }
   }
   waitWindow::close
}
#----------------------------------------------------------------------------
# Prepend the folder year to some folder
#----------------------------------------------------------------------------
proc macros::prependFolderYear {folder} {
   APL::busyCursor 1
   set name ""
   set tail [file tail $folder]
   set years [db::fieldsInFolder $folder "Year"]
   # Check freedb discid
   if {[lindex $years 0] eq "" || [llength $years]>1} {
      package require freedb
      set found [freedb::albumList [db::filesInFolder $folder]]
      if {[llength $found]>0} {
         foreach {record} $found {
            if {[regexp -- {(.+? .+?) (.*)$} $record -> tag artalbum]} {
               set found [freedb::albumDetails $tag Data]
               if {[llength $found]>0} {
                  if {[info exists Data(Year)]} {
                     set year $Data(Year)
                     if {[string is integer -strict $year]} {
                        set years [list $year]
                        puts "$tail discid Found year $artalbum $year"
                        break
                     }
                  }
               }
            }
         }
      }
   }

   # Check freedb text search if prior failed
   if {[lindex $years 0] eq ""} {
      set Artist [db::fieldsInFolder $folder "Artist"]
      set Album $tail
      set found [freedb::albumListHTML [lindex $Artist 0] $Album]
      if {[llength $found]>0} {
         foreach {record} $found {
            if {[regexp -- {(.+? .+?) (.*)$} $record -> tag artalbum]} {
               set found [freedb::albumDetails $tag Data]
               if {[llength $found]>0} {
                  if {[info exists Data(Year)]} {
                     set year $Data(Year)
                     if {[string is integer -strict $year]} {
                        set years [list $year]
                        puts "$tail search Found year $artalbum $year"
                        break
                     }
                  }
               }
            }
         }
      }
   }

   if {[llength $years]==1} {
      set year [string trim [lindex $years 0]]
      if {$year ne ""} {
         set year [lindex $years 0]
         set tail [file tail $folder]
         set trunk [file dirname $folder]
         set newname [file join $trunk "$year $tail"]
         puts "$tail Using $year in folder"
         APL::busyCursor 0
         catalog::moveFolder $folder $newname
         set name $newname
      }
   } else {
      APL::busyCursor 0
      puts "$tail No year found..."
      # display dialog with checkbox choices
   }
   return $name
}

#------------------------------------------------------------------------------
# Command line tools
#------------------------------------------------------------------------------
proc macros::cl {args} {
   package require commandline
}



#----------------------------------------------------------------------------
# Make the following a separate namespace in a separate file
#----------------------------------------------------------------------------




#------------------------------------------------------------------------------
# Function   : dndPlayFileList
# Description: Play a list of files, inserting all but the last one
#              Bound to the snackamp GUI
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::dndPlayFileList {files what} {
   Trace
   global TA
   set flag 0
   if {[llength $files]} {
      set tracks [macros::TranslateFiles $files]
      if {[llength $tracks]} {
         set here [LB::idxPlaying]
         LB::insertList [incr here] $tracks
         LB::activate $here
         Play active
      }
   }
   return ""
}
#----------------------------------------------------------------------------
# dnd drop target, insert incoming files at cursor
# Bound to the listbox
#----------------------------------------------------------------------------
proc macros::dndDropFileList {x y files} {
   Trace
   global LB
   Trace
   if {$x>0} {
      set new [$LB nearest $y]
   } else {
      set new end
   }
   if {[llength $files]} {
      set tracks [macros::TranslateFiles $files]
      if {[llength $tracks]} {
         LB::insertList $new $tracks
      }
   }
   return 1
}
#------------------------------------------------------------------------------
# Translate a list of files/folders into tracks, support playlists also
# if link is to a http site, read the url if it is a playlist and process it.
# TFW: Need support recursive calls for nested playlists/folders.
# TODO: Handle recursive folders
#------------------------------------------------------------------------------
proc macros::TranslateFiles {files {base ""}} {
   global TA
   Trace
   set toappend [list]
   foreach {file} $files {
      set file [string map {\\ /} $file]
      set ext [string tolower [file extension $file]]
      ##
      # Override the extension if http file info in header
      # NOTE: http mp3 files may not support a HEAD and the getUrlType will hang
      if {[string match -nocase "http://*" $file]} {
         foreach {httpType file} [macros::getUrlType $file] {break}
         switch -exact -- $httpType {
            application/xml -
            text/xml {
               set ext .xml
            }
            audio/mpeg {
               set ext .mp3
               # an mp3 file link
            }
            image/jpeg {
               set ext .jpeg
            }
            audio/x-scpls {
               set ext .pls
            }
            audio/x-mpegurl {
               set ext .m3u
            }
            default {
            }
         }
      } elseif {[regexp -- {file:(.*)} $file -> url]} {
         # handle file::xxxx format
         set file [file nativename [HttpD::Decode $url]]
      }
      ##
      # TODO: Handle HTTP and local files difference,
      # for HTTP files download non mpeg into memory
      # for regular files read into memory
      #

      switch -- $ext {
         .xml {
            # TODO:Need to append here then add later, mod handlXML to create the usk and return list of files
            foreach {toadd} [macros::handleXML $file] {
               lappend toappend $toadd
            }
         }
         .jpg -
         .png -
         .gif -
         .bmp  -
         .jpeg {
            macros::copyImageToFolder $file [file dirname $TA(currentMF)]
            gui::packCoverArt $TA(currentMF)
            gui::smallCoverArt $TA(currentMF)
            # Don't append to play since we just replaced the image
         }
         .opml {
            foreach {toadd} [macros::handleOPML $file] {
               lappend toappend $toadd
            }
         }
         .m3u
         -
         .pls {
            # If coming from a web site, create a usk file,
            # otherwise, read in the playlist and process each line
            if {[string match -nocase "http://*" $file]} {
               playurl::checkuskfolder
               set contents [split [macros::getUrlBinary $file] \n]
            } else {
               if {[file readable $file]} {
                  set fid [open $file r]
                  set contents [split [string trim [read $fid]] \n]
                  close $fid
               } else {
                  tk_messageBox -icon warning -type ok -message "Playlist File '$file' cannot be read"
                  return $toappend
               }
            }
            ;# Handle playlist types uniquely
            set fileTail [file rootname [file tail $file]]
            switch -- $ext {
               .m3u {
                  set i 0
                  foreach {row} $contents {
                     if {$row eq ""} {
                        continue
                     }
                     array unset Data
                     set drow [HttpD::unSub $row]  ;# not sure if this is needed
                     if {[string match "\#*" $drow]} {
                        # ignore
                     } elseif {[string match -nocase "number*" $drow]} {
                        # ignore
                     } elseif {[string match -nocase "\[playlist*" $drow]} {
                        # ignore
                     } else {
                        set fullFile [file join [file dirname $file] $drow]
                        if {[string match -nocase "http://*" $drow]} {
                           # Create a usk (url) shortcut, then append it
                           set name [file tail $drow]
                           # Note: trying tail of file name, works if a file if a stream
                           # may not work as well.
                           set Data(Title) $name
                           set Data(Genre) "Stream"
                           set fname [playurl::addUrl $::snackAmpSettings(uskDir) "$name" $row]
                           lappend toappend $fname
                           incr i
                        } elseif {[file exists $fullFile]} {
                           lappend toappend $fullFile
                        } else {
                           saLog "Playlist File:'$row' does not exist"
                        }
                     }
                  }
               }
               .pls {
                  set title $fileTail
                  set i 0
                  set Fnum 0
                  set Tnum 0
                  foreach {row} $contents {
                     if {[regexp -- {File([0-9]+)=(.*)} $row -> Fnum url]} {
                        set pls($Fnum,url) $url
                     }
                     if {[regexp -- {Title([0-9]+)=(.*)} $row -> Tnum title]} {
                        set title [string trim [file rootname $title] .]
                        set pls($Tnum,title) $title
                     }
                  }
                  foreach {index} [lsort -dictionary [array names pls *,url]] {
                     array unset Data
                     foreach {num type} [split $index ,] {break}
                     if {[info exists pls($num,title)]} {
                        set title $pls($num,title)
                     } else {
                        set title $fileTail.$num
                     }
                     # TODO: Should handle both local and http files
                     set Data(Title) $title
                     set Data(Album) $url
                     set Data(Genre) "Stream"
                     lappend toappend [playurl::addUrl $::snackAmpSettings(uskDir) $title $pls($index) Data]
                  }
               }
            }
         }
         default {
            ##
            # TODO: Handle specific file types
            #
            if {[file isdirectory $file]} {
               foreach {track} [fileutils::rglob $file] {
                  if {[catalog::isTrackFile $track]} {
                     lappend toappend $track
                  }
               }
            } elseif {[string match -nocase "http://*" $file]} {
               # Dragging a file to the player, try to add a usk for it
               # stream from the url
               set chFile [lindex [split $file ?] 0]
               if {[catalog::isTrackFile $chFile]} {
                  set fname [playurl::addUrl $::snackAmpSettings(uskDir) "$chFile" $file]
                  lappend toappend $fname
               } else {
                  tk_messageBox -icon warning -type ok -message "File $file is currently unsupported"
               }
               ##
               # Todo: Can parse xml here also
               #
            } elseif {[catalog::isTrackFile $file]} {
               # usk shortcut files also handled here
               set ff [file join $base $file]
               if {[file readable $ff]} {
                  lappend toappend [file join $base $file]
               }
            } else {
               tk_messageBox -icon warning -type ok -message "Default File '$file' cannot be read"
            }
         }
      }
   }
   return $toappend
}
#------------------------------------------------------------------------------
# Get the info on some url and close it
# This only works on http urls that handle HEAD requests since
# we use the -validate 1 option. Some streams don't handle this well.
#------------------------------------------------------------------------------
proc macros::getUrlType {url} {
   Trace
   package require uri
   package require http 2.4
   set urlHistory [list]
   while {1} {
      if {[lsearch -exact $urlHistory] >= 0} {
         # Have already been here, an endless loop
         return $url
      }
      lappend urlHistory $url
      # BUG: This will spin the CPU on a local slimserver
      # because HEAD requests are not supported
      set token [::http::geturl $url -validate 1]
      set code [::http::ncode $token]
      upvar $token state
      set type $state(type)
      # type of content
      set tpe [lindex [split $type \;] 0]
      ##
      # Handle redirects
      #
      if {$code == 200} {
         ::http::cleanup $token
         return [list $tpe $url]
      } elseif {$code >= 300 && $code <= 307} {
         # Look for redirect within the metadata
         array set meta $state(meta)
         if {[info exists meta(Location)]} {
            unset -nocomplain URI
            array set URI [::uri::split $url] ;# Need host info from here
            # Look for redirect embedded in Location
            array set uri [::uri::split $meta(Location)]
            unset meta
            # replace host with one from orig url (if specified) incase not fully qualified
            if {$uri(host) == ""} {
               set uri(host) $URI(host)
            }
            set nurl [eval ::uri::join [array get uri]]
            ::http::cleanup $token
            set url $nurl
         } elseif {[info exists meta(url)] && ($meta(url) ne $url)} {
            ::http::cleanup $token
            set url $meta(url)
         } else {
            # WTF
            ::http::cleanup $token
            saLog "[me] Redirect Code $code ->Couldn't get redirection of $url"
            return [list $code $url]
         }
      } else {
         ::http::cleanup $token
         saLog "[me] Return Code $code ->Couldn't get redirection of $url"
         return [list "unknown" $url]
      }
   }
}
#------------------------------------------------------------------------------
# Grab the contents of some URL and return it
# Used to grab binary data such as a jpeg
# Handle various redirects
#------------------------------------------------------------------------------
proc macros::getUrlBinary {url} {
   Trace
   package require http 2.4
   package require uri
   set urlHistory [list]
   while {1} {
      if {[lsearch -exact $urlHistory] >= 0} {
         # Have already been here, an endless loop
         return $url
      }
      lappend urlHistory $url
      # Read the data from the url, if redirected handle below
      set token [::http::geturl $url -binary 1]
      set contents [::http::data $token]
      set code [::http::ncode $token]
      # Check if Redirected, if can handle do so recursively
      # <a HREF="http://991.com/newgallery/Elton-John-Friends-357292.jpg">
      if {$code == 200} {
         # All is well
         ::http::cleanup $token
         return $contents
      } elseif {$code >= 300 && $code <= 307} {
         # Look for redirect within the metadata
         upvar $token state
         array set meta $state(meta)
         if {[info exists meta(Location)]} {
            # Look for redirect embedded in Location
            unset -nocomplain URI
            array set URI [::uri::split $url] ;# Need host info from here
            array set uri [::uri::split $meta(Location)]
            unset meta
            if {$uri(host) == ""} {
               set uri(host) $URI(host)
            }
            set nurl [eval ::uri::join [array get uri]]
            ::http::cleanup $token
            set url $nurl
            lappend urlHistory $url
         } elseif {[info exists meta(url)] && ($meta(url) ne $url)} {
            # Look for redirect on URL directly
            ::http::cleanup $token
            set url $meta(url)
            lappend urlHistory $url
         } elseif {[regexp -nocase -- {<a HREF="(.*)">} $contents -> nurl]} {
            # Look for redirects within the page contents
            ::http::cleanup $token
            set url $nurl
            lappend urlHistory $url
         } else {
            # WTF
            ::http::cleanup $token
            saLog "Return Code $code ->Couldn't handle redirection of $url"
            return ""
         }
      } else {
         ::http::cleanup $token
         saLog "[me] Error code $code: Can't handle $url"
         return ""
      }
   }
}

#------------------------------------------------------------------------------
# Process XML files and extract the enclosed podcast url(s) if possible
# and append them to the playlist
#------------------------------------------------------------------------------
proc macros::handleOPML {file} {
   Trace
   ##
   # opml has the following format:
   #
   #<outline text="A Brit Abroad" type="link" url="http://www.steve-lacey.com/podcast.xml"/>
   #<outline text="ApocXP's Blog O'Stuff Radio" type="link" url="http://jarvelo.brinkster.net/podcast.aspx"/>
   #<outline text="atmaspheric | radio" type="link" url="http://feeds.feedburner.com/AtmasphericRadio"/>
   #<outline text="Auto Audio" type="link" url="http://www.raxhaven.com/podcast.php"/>
   #<outline text="BFBS Games Gadgets Gizmos - G3" type="link" url="http://www.bfbs2.com/g3.xml"/>
   set tracks [list]
   if {[string match -nocase "http://*" $file]} {
      playurl::checkuskfolder
      set contents [macros::getUrlBinary $file]
      # Parse each item and build usk
      set items [regexp -all -inline -- {<outline (.+?)>} $contents]
      foreach {raw item} $items {
         # Extract the enclosure (tricky)
         if {[regexp -- {\stext=[\'|\"](.+?)[\'|\"]\s} $item -> text]} {
            puts "found text:$text"
         } else {
            set url "none"
         }
         if {[regexp -- {\url=[\'|\"](.+?)[\'|\"]\s} $item -> url]} {
            puts "found outline:$url"
         } else {
            set desc "none"
         }
         if {[regexp -- {\type=[\'|\"](.+?)[\'|\"]\s} $item -> type]} {
            puts "found outline:$type"
         } else {
            set desc "none"
         }

         set desc $text

         if {$url ne "none"} {
            lappend tracks [playurl::addUrl $::snackAmpSettings(uskDir) $desc $url]
         } else {
            saLog "Can't add $url in $item, not enough info"
         }
      }
   }
   return $tracks
}
#------------------------------------------------------------------------------
# Process XML files and extract the enclosed podcast url(s) if possible
# and append them to the playlist
# This could be MUCH improved by using tclXML
#------------------------------------------------------------------------------
# xml tag         channel     item     where content appears in iTunes
# <title>              Y  Y  Name column
# <link>               Y     website link and arrow in Name column
# <copyright>          Y     not visible
# <pubDate>               Y  Release Date column
# <itunes:author>      Y  Y  Artist column                         (Artist)
# <itunes:block>       Y  Y  prevent an episode or podcast from appearing
# <itunes:category>    Y     Category column and in iTunes Music Store Browse
# <itunes:image>       Y     same location as album art
# <itunes:duration>       Y  Time column
# <itunes:explicit>    Y  Y  parental advisory graphic in Name column
# <itunes:keywords>    Y  Y  not visible but can be searched
# <itunes:new-feed-url> Y     not visible, used to inform iTunes of new feed URL location
# <itunes:owner>       Y     not visible, used for contact only
# <itunes:subtitle>    Y  Y  Description column
# <itunes:summary>     Y  Y  when the "circled i" in Description column is clicked
#------------------------------------------------------------------------------
proc macros::handleXML {file} {
   Trace
   if {[string match -nocase "http://*" $file]} {
      playurl::checkuskfolder
      set contents [macros::getUrlBinary $file]
#      set contents [playurl::getUrlBody $file]
      set tracks [list]
      # Parse each item and build usk
      set items [regexp -all -inline -- {<item>(.+?)</item>} $contents]
      set track 0
      foreach {raw item} $items {
         # Extract the enclosure (tricky)
         # Title Artist Comp Album Year Note Track SubGenre Genre
# <GUID ISPERMALINK="false">http://podcasts.cnn.net/cnn/services/podcasting/digital.life/audio/2007/01/digital0112.mp3</GUID>

         if {[regexp -nocase -- {<enclosure\s+?url=[\'|\"](http:/.+?)[\'|\"]\s} $item -> url]} {
            puts "[me] found enclosure:$url"
         } elseif {[regexp -nocase -- {<guid.*?>(http:/.+?)\?.*?</guid>} $item -> url]} {
            puts "[me] found guiid:$url (skip param)"
         } elseif {[regexp -nocase -- {<guid.*?>(http:/.+?)</guid>} $item -> url]} {
            puts "[me] found guiid:$url"
         } else {
            set url "none"
         }
         set desc "Unknown"
         # Title
         if {[regexp -- {<title>(.+?)</title>} $item -> desc]} {
            #puts "[me] found Title:$desc"
            set Data(Title) [HttpD::unSub $desc]
         }
         # Artist
         if {[regexp -- {<itunes:author>(.+?)</itunes:author>} $item -> author]} {
            #puts "[me] found Artist:$author"
            set Data(Artist) [HttpD::unSub $author]
         } elseif {[regexp -- {<author>(.+?)</author>} $item -> author]} {
            set Data(Artist) [HttpD::unSub $author]
         }
         # Comment
         if {[regexp -- {<itunes:subtitle>(.+?)</itunes:subtitle>} $item -> note]} {
            set Data(Note) [HttpD::unSub $note]
         } elseif {[regexp -- {<description>(.+?)</description>} $item -> note]} {
            set Data(Note) [HttpD::unSub $note]
         }
         # Genre
         if {[regexp -- {<category>(.+?)</category>} $item -> genre]} {
            set Data(Genre) [HttpD::unSub $genre]
         } else {
            set Data(Genre) "PodCast"
         }
         if {[regexp -- {<pubDate>(.+?)</pubDate>} $item -> year]} {
            set Data(Year) [HttpD::unSub $year]
         }
         set Data(Track) [incr track]
         # Check to add
         if {$url ne "none"} {
            lappend tracks [playurl::addUrl $::snackAmpSettings(uskDir) $desc $url Data]
         } else {
            saLog "Can't add $url in $item, not enough info"
         }
      }
   }
   return $tracks
}
#------------------------------------------------------------------------------
# Copy a file to a different folder, used for images in files or url links
#------------------------------------------------------------------------------
proc macros::copyImageToFolder {url folderName} {
   Trace
   set fname [file join $folderName "folder.jpg"]
   # Wait some time to allow the file to be transferred
   # needed for firefox
   set result 0
   if {![string match "http://*" $url]} {
      # "text/uri-list"
      # Make sure we found one
      if {[file exists $url] && [file size $url]>0} {
         if {[file exists $fname]} {
            set attribs [macros::restoreAttribs $fname {}]
            set resp [tk_messageBox -type yesno -icon warning -message "Is it ok to overwrite $fname\nwith $url"]
         } else {
            set attribs [list]
            set resp "yes"
         }
         if {$resp eq "yes"} {
            albumart::clear
            saLog "Copying file $url to $folderName"
            macros::makeWritable $fname
            file copy -force $url $fname
            macros::restoreAttribs $fname $attribs
            incr result
         }
      } else {
         tk_messageBox -title "OK to Continue" -icon warning -message "$url does not exist or is zero length"
      }
   } else {
      # "UniformResourceLocator"
      # httpget (works more reliablay with firefox)
      if {[file exists $fname]} {
         set attribs [macros::restoreAttribs $fname {}]
         set resp [tk_messageBox -type yesno -icon warning -message "Is it ok to overwrite $fname\nwith $url"]
      } else {
         set attribs [list]
         set resp yes
      }
      if {$resp eq "yes"} {
         if {[catch {getUrlBinary $url} contents] } then {
            tk_messageBox -type ok -icon warning -message "Error retrieving $url\n$contents"
         }
         if {[string length $contents]} {
            macros::makeWritable $fname
            if {[catch {open $fname w} fid] } then {
               saLog $fid
               tk_messageBox -type ok -icon warning -message "$fid \nfor $fname"
            } else {
               albumart::clear
               saLog "Copying url $fname to $folderName"
               fconfigure $fid -translation binary -encoding binary
               puts -nonewline $fid $contents
               close $fid
               incr result
            }
            macros::restoreAttribs $fname $attribs
         } else {
            tk_messageBox -type ok -icon warning -message "Nothing downloaded or error for\n$url"
         }
      }
   }
   return $result
}


package provide macros 1.0
