##################################################################################
# snackAmp - metakit package 1.0 for snackAmp
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
##################################################################################
set [file rootname [file tail [info script]]]_Metakit {
   $Header: /cvsroot/snackamp/snackamp/lib/metakit.tcl,v 1.57 2007/10/01 01:36:28 wilkason Exp $
}
;##
;# Metakit Interface
;#
package require Mk4tcl
interp alias {} isnumber {} string is double -strict
namespace eval db {
   variable isSorted 0
   variable Cols
   variable dbObj
   ;# Metakit but on linux, won't work with ~ as part of name, need full path
   variable dbLocation [file nativename [file join $::HOME snackDB].mk]
   variable csvLocation [file nativename [file join $::HOME snackDB].mke]
   variable dbLock     [file nativename [file join $::HOME snackDB].lck]
   #variable fieldSpec [list File Folder Mounted:I FullData:I Depth:I Title Artist Album Track Media Year Genre Note Comp Enc Desc Rate:S sampRate:D Duration:D size:I date:I Gain Tag Played LastPlay:I Rating Start:I Stop:I Votes:I cataloged:I]
   variable fieldSpec [list File Folder Mounted:I FullData:I Depth:I Title Artist Album Track Media Year Genre Note Comp Enc Desc Rate:S sampRate:D Duration:D size:I date:I Gain AlbumGain Tag Played LastPlay:I Rating Start:I Stop:I Votes:I cataloged:I APL:B]
   variable fields
   variable changesList
   variable checkTime "mtime"
   variable state "close"
   variable type "metakit"
   variable DB "db.catalog"
   variable mods 0
   variable ratingOptions [list -- * ** *** **** *****]
   variable commitRate 1000
   variable commitAfter 0
   variable SORT 0
   # Build up variables for substitution
   lappend changesList "\[list"
   foreach {fld} $fieldSpec {
      set f [lindex [split $fld :] 0]
      if {$f=="File"} {
         lappend changesList $f \$$f
      } else {
         lappend fields $f
         lappend changesList $f \$Details($f)
      }
   }
   lappend changesList "\]"
   # Make it a nice list for substitution
   set changesList [join $changesList]
}
#----------------------------------------------------------------------------
# Create a lock file
#----------------------------------------------------------------------------
proc db::lock {args} {
   Trace
   variable dbLock
   if {![file exists $dbLock]} {
      set fod [open $dbLock w]
      puts $fod "[list $::tcl_platform(user) [clock seconds] [info hostname]]"
      close $fod
      return 1
   } else {
      return 0
   }
}
#----------------------------------------------------------------------------
# Release a lock file
#----------------------------------------------------------------------------
proc db::unlock {args} {
   Trace
   variable state
   variable dbLock
   if {[file exists $dbLock]} {
      if {[catch {file delete -force $dbLock} result] } then {
         saLog $result
      }
   }
}
#----------------------------------------------------------------------------
# Return info on a lock filed (if one exists)
#----------------------------------------------------------------------------
proc db::lockInfo {args} {
   Trace
   variable dbLock
   if {[file exists $dbLock]} {
      foreach {var} {user time host} {set $var {}}
      set fid [open $dbLock r]
      foreach {user time host} [read $fid] {break}
      close $fid
      return [list $user $time $host]
   } else {
      return [list]
   }
}
proc db::exists {args} {
   variable dbLocation
   return [file exists $dbLocation]
}
#----------------------------------------------------------------------------
# Function   : db::open
# Description: Create the user database if needed
# Author     : Tom Wilkason
# Date       : 9/21/2002
#----------------------------------------------------------------------------
proc db::openDB {args} {
   variable dbLocation
   variable dbLock
   variable dbObj
   variable fieldSpec
   variable state
   variable DB
   variable SORT
   # Ignore if db already open
   if {[string equal $state "close"]} {
      # Check for lock file
      set linfo [db::lockInfo]
      if {[llength $linfo]>0} {
         foreach {user time host} $linfo {break}
         waitWindow::close
         set res [tk_messageBox -title "Database Locked" -type yesno -icon warning \
            -message "The SnackAmp database is locked by user $user on host $host,\ndo you want to override the lock?"]
         if {$res=="yes"} {
            db::unlock
         } else {
            exit
         }
      }
      # First time, must create $dbLocation
      if {[catch {mk::file open db $dbLocation -nocommit} result]} {
         waitWindow::close
         set res [tk_messageBox -title "Database Open Error" -type ok -icon warning \
            -message "The SnackAmp database '$dbLocation' could not be opened\n\n$::errorInfo"]
         exit
      }

      set state "open"
      ##
      # main db to hold data, uses blocking
      #
      mk::view layout db.catalog $fieldSpec
      mk::view open db.catalog __db_data ;# Don't touch __db_data command
      ##
      # db.catalog_map is secondary view for hashing
      #
      mk::view layout db.catalog_map {_H:I _R:I}
      ##
      # Create a hash map view for direct access on find
      # combines the base and hash map views
      mk::view open db.catalog_map __db_map
      # __db_data is raw command, __db_map is just map view, dbObj is hashed view
      # Create an ordered/hash view, all udpates are made in an ordered fashion
      # so we don't have to do any sorting during the query
      if {$SORT} {
         # use sorted view
         set dbObj [[__db_data view hash __db_map 1] view ordered 1]
      } else {
         # use unsorted view
         ### Needs sorting to return something from select
         set dbObj [__db_data view hash __db_map 1]
      }

      db::lock
      ##
      # To create a mapped view on Mounted, create a new view (v2) and
      # populate it with a single integer value containing all the
      # rows returned from a select Mounted 1 from the base view (v1)
      # set v3 [$v1 view map $v2]. Then use the v3 view to access all
      # of the data from $v1. Not sure if adds/deletes will work correctly?
   }
}
#----------------------------------------------------------------------------
# Create a music listing of all the tracks
#----------------------------------------------------------------------------
proc db::document {expfile} {
   variable dbObj
   variable DB
   set catFiles [list]
   if {[catch {open $expfile w} fid]} {
      return
   }
   fconfigure $fid -buffersize 500000
   set rows [mk::select $DB]
   puts $fid {"Artist","Album","Track","Title","File"}
   foreach {row} $rows {
      foreach {File Artist Album Track Title} [$dbObj get $row File Artist Album Track Title] {break}
      puts $fid [subst -nocommand {"$Artist","$Album","$Track","$Title","$File"}]
   }
   close $fid
   tk_messageBox -icon info -type ok -message "csv list of files created as $expfile"
   saLog "[llength $rows] records exported"
}
##
# Export all of the data to a flat file for later import
#
proc db::exportOnePath {expfile {path /}} {
   global TA snackAmpSettings
   variable dbObj
   variable DB
   variable SORT
   set pattern [file join $path *]
   set catFiles [list]
   if {[catch {open $expfile w} fid]} {
      saLog "$fid"
      set folder [tk_chooseDirectory -initialdir [file dirname $expfile] -title "Select alternate folder to save list"]
      if {[file exists $folder]} {
         set expfile [file join $folder [file tail $expfile]]
         saLog "[me] Using alternate location $expfile"
         if {[catch {open $expfile w} fid]} {
            saLog "[me] Failed to open $expfile"
            return 0
         }
      } else {
         saLog "[me] User selected no or invalid path"
         return 0
      }
   }
   fconfigure $fid -buffersize 500000 -buffering full -translation lf
   set trim [string length $path]
   if {$SORT} {
      set rows [mk::select $DB -glob File $pattern]
   } else {
      set rows [mk::select $DB -sort File -glob File $pattern]
   }
   set i 0
   set tr [llength $rows]
   foreach {row} $rows {
      set Row [$dbObj get $row]
      set File [lindex $Row 1]
      set Folder [lindex $Row 3]
      # Make it relative to base folder
      lset Row 1 [string trimleft [string range $File $trim end] /]
      lset Row 3 [string trimleft [string range $Folder $trim end] /]
      set Row [string map {\n " "} $Row] ;# no embedded cr or lf (0xA=LF is used in ID3 tags)
      puts $fid $Row
      incr i
      if {($i%50)==0} {
         waitWindow::percent $i $tr "Exporting\n$path to\n$expfile" 1
         if {[waitWindow::ifbreak]} {
            break
         }
      }
   }
   close $fid
   saLog "$i records exported to $expfile"
   return $i
}
##
# Export all of the data to a flat file for later import
#
proc db::exportRAW {args} {
   variable dbObj
   variable DB
   variable csvLocation
   set catFiles [list]
   set im 0
   db::fixFolders
   if {[catch {open $csvLocation w} fid]} {
      saLog "$fid"
      return $im
   }
   set size [$dbObj size]
   fconfigure $fid -buffersize 500000 -buffering full -translation lf
   set rows [mk::select $DB -sort File]  ;# sort just in case not kept sorted
   # BUG: Some Note fields have CRLF in them causing a problem, how to deal
   foreach {row} $rows {
      set ROW [string map {\n " "} [$dbObj get $row]]
      puts $fid $ROW
      incr im
      if {($im%100)==0} {
         waitWindow::percent $im $size "Exporting Database"
      }
   }
   close $fid
   saLog "[llength $rows] records exported"
   return $im
}
##
# Import a previously exported file, append to the database
#
# set f C:/Incoming/snackDB.mke

proc db::importRAW {size} {
   global TA snackAmpSettings
   variable dbObj
   variable csvLocation
   ##
   # Restore the cached files from each media Dir
   #
   set catFiles [list]
   if {[catch {open $csvLocation r} fid]} {
      return
   }
   fconfigure $fid -buffersize 100000 -buffering full -translation lf
   set im 0
   while {[gets $fid row]>0} {
      incr im
      if {[catch {eval $dbObj insert end $row} result]} then {
         saLog "$result:\n\t$row"
         continue
      }
      if {($im%100)==0} {
         waitWindow::percent $im $size "Importing Database"
      }
   }
   close $fid
   saLog "$im records imported"
   db::Commit
}
#----------------------------------------------------------------------------
# Import a single group of files within a path
# Fixup the relative path names properly
#----------------------------------------------------------------------------
proc db::importOnePath {file path} {
   global TA snackAmpSettings
   variable dbObj
   ##
   # Restore the cached files from each media Dir
   #
   set catFiles [list]
   set expfile [file join $path $file]
   if {[catch {open $expfile r} fid]} {
      saLog $fid
      return [list 0 0]
   }
   fconfigure $fid -buffersize 100000 -buffering full -translation lf
   set im 0
   set rp 0
   set i 0
   # Flag all in this branch as unmounted
   db::mountFiles [file join $path *] 0
   while {[gets $fid row]>0} {
      if {[catch {file join $path [lindex $row 1]} File] } then {
         saLog "$File -> $row"
         continue
      }
      set Folder [file dirname $File]
      unset -nocomplain Info
      array set Info $row
      set Info(File) $File
      set Info(Folder) $Folder
      set Info(Depth) [llength [file split $File]]
      ##
      # Append or replace, depending if entry exists
      #
      if {[catch {$dbObj find File $File} index] } then {
         incr im
         set row [array get Info]
         eval $dbObj insert end $row
      } else {
         incr rp
         # Only import certain fields if they are replacing default fields
         #
         array set OldData [$dbObj get $index]
         set Info(Mounted) 1
         # Gain, only if mine is --, otherwise gain, start & stop
         if {$OldData(Gain)!="--"} {
            unset -nocomplain Info(Gain)
            unset -nocomplain Info(Start)
            unset -nocomplain Info(Stop)
            unset -nocomplain Info(AlbumGain)
         }
         # Rating, only if mine is --
         if {$OldData(Rating)!="--"} {
            unset -nocomplain Info(Rating)
         }
         # Played, import larger
         if {$OldData(Played) > $Info(Played)} {
            unset -nocomplain Info(Played)
         }
         # LastPlay, import larger
         if {$OldData(LastPlay) > $Info(LastPlay)} {
            unset -nocomplain Info(LastPlay)
         }
         # Some fields may now be gone, so we won't set them
         set row [array get Info]
         eval $dbObj set $index $row
      }
      if {([incr i]%25)==0} {
         waitWindow::show "Added [commify $im], and replaced [commify $rp] records\n\n$path"
      }
   }
   close $fid
   db::dirty
   set removed [db::removeUnmountedFiles [file join $path *]]
   saLog "$im records added, $rp records replaced, $removed records removed"
   return [list $im $rp $removed]
}
##
# Return a list of folders in a tree
#
proc db::folderList {{pattern *}} {
   variable DB
   variable dbObj
   foreach {row} [mk::select $DB -glob Folder $pattern -exact Mounted 1] {
      set hash([$dbObj get $row Folder]) {}
   }
   return [lsort [array names hash]]
}
#----------------------------------------------------------------------------
# Function   : restart
# Description: Close and delete the databases
# Author     : Tom Wilkason
# Date       : 12/26/2002
#----------------------------------------------------------------------------
proc db::restart {args} {
   Trace
   variable dbLocation
   variable csvLocation
   waitWindow::show "Exporting Database..."
   Wait 1
   set exported [db::exportRAW]
   if {$exported > 0} {
      waitWindow::show "Recreating Database..."
      db::closeDB
      # TBD: Rename file, then delete it when import succeeds
      file delete -force -- ${db::dbLocation}.bak
      file rename -force -- $db::dbLocation ${db::dbLocation}.bak
      db::openDB
      waitWindow::show "Importing Database..."
      db::importRAW $exported
      waitWindow::release
      waitWindow::close
      Wait 0
   } else {
      saLog "No files exported"
   }
}
#----------------------------------------------------------------------------
# Function   : db::lookup
# Description: Return details for a file, if not in db the optionally add it
# Author     : Tom Wilkason
# Date       : 12/25/2002
# TODO: Can just use cacheFile with appropriate parameters
#----------------------------------------------------------------------------
proc db::lookup {File _Data {cache 1}} {
   upvar $_Data Data
   variable dbObj
   set data [list]
   if {[catch {$dbObj find File $File} row] } then {
      ;# Didn't exist, add the data if needed
      if {$cache} {
         set cr [db::cacheFile $File Data 1]
      } else {
         set cr [db::cacheFile $File Data 0]
      }
   } else {
      # Existed
      array set Data [$dbObj get $row]
      if {($Data(FullData)==0 && $cache==1) || $cache==2} {
         set cr [db::cacheFile $row Data $cache]
      } else {
         set cr 1
      }
   }
   # If cache failed, return 0
   if {$cr==0} {
      return 0
   }
   # This happens on file not existing
   if {[info exists Data(Gain)]} {
      set Data(Gain) [gainControl::levelText $Data(Gain)]
   }
   if {[info exists Data(AlbumGain)]} {
      set Data(AlbumGain) [gainControl::levelText $Data(AlbumGain)]
   }
   return [array size Data]
}
#----------------------------------------------------------------------------
# Update the file mtime and file size in the database, call after a file has been modified
#----------------------------------------------------------------------------
proc db::updateTimeTag {File} {
   variable dbObj
   variable checkTime
   if {[catch {$dbObj find File $File} row] } then {
      #debug "$File not found"
   } else {
      if {[catch {file stat $File Stat} result] } then {
         saLog $result
      } else {
         $dbObj set $row size $Stat(size)
         $dbObj set $row date $Stat($checkTime)
      }
   }
}
#----------------------------------------------------------------------------
;# Cache a single file name looking up info for it
;# If it exists, do nothing
#----------------------------------------------------------------------------
proc db::cacheFileName {File {chkMount 1}} {
   Trace
   variable checkTime
   variable dbObj
   ;# If file not in db, then cache just basic part
   if {[catch {$dbObj find File $File} row] } then {
      if {[catch {file stat $File Stat} result]} {
         debug $result
         return -1
      } else {
         set row [$dbObj size]
         $dbObj insert end \
            File     $File        \
            Folder   [file dirname $File] \
            Depth    [llength [file split $File]]   \
            cataloged [clock seconds] \
            FullData  0            \
            size     $Stat(size)  \
            date     $Stat($checkTime) \
            Played   0  \
            LastPlay 0  \
            Rating   "--"  \
            Votes    0  \
            Mounted  1  \
            Gain     "--" \
            AlbumGain "--"
         db::dirty
      }
   } else {
      # Found, flag as mounted
      if {$chkMount} {
         db::dirty
         $dbObj set $row Mounted 1
      }
   }
   return $row
}


#----------------------------------------------------------------------------
# Increment the play counter for some file
#----------------------------------------------------------------------------
proc db::bumpPlayed {File {num 1}} {
   Trace
   variable dbObj
   if {[catch {$dbObj find File $File} row] } then {
      debug $row
      return
   }
   foreach {played last} [$dbObj get $row Played LastPlay] {break}
   if {![string is integer -strict $played]} {
      set played 1
   } else {
      incr played $num
   }
   # Update play count and last played (round to hour)
   $dbObj set $row Played $played LastPlay [clock seconds]
   db::dirty
}
#----------------------------------------------------------------------------
# Increment the rating for some file by some amount
#----------------------------------------------------------------------------
proc db::bumpRating {File {num 1}} {
   variable dbObj
   if {[catch {$dbObj find File $File} row] } then {
      debug $row
      return
   }
   set rating [$dbObj get $row Rating]
   if {$rating=="--"} {
      if {$num == -1} {
         set num 2
      } elseif {$num==1} {
         set num 4
      } elseif {$num < 1} {
         set num 1
      } elseif {$num > 5} {
         set num 5
      }
      set rating [string repeat * $num]
   } else {
      set cr [string length $rating]
      incr cr $num
      if {$cr < 1} {set cr 1} elseif {$cr > 5} {set cr 5}
      set rating [string repeat * $cr]
   }
   $dbObj set $row Rating $rating
   db::dirty
}
#------------------------------------------------------------------------------
#  Set a single element to some value in the database
#------------------------------------------------------------------------------
proc db::setVal {File Field val} {
   variable dbObj
   variable mods
   if {[catch {$dbObj find File $File} row] } then {
      debug $row
      return ""
   }
   db::dirty
   $dbObj set $row $Field $val
   return $val
}
#------------------------------------------------------------------------------
#  Set multiple elements to some value in the database
#------------------------------------------------------------------------------
proc db::setVals {File args} {
   variable dbObj
   variable mods
   if {[catch {$dbObj find File $File} row] } then {
      debug $row
      return ""
   }
   db::dirty
   eval [list $dbObj set $row $args]
   return $val
}
#------------------------------------------------------------------------------
#  Get a single element to some value in the database
#------------------------------------------------------------------------------
proc db::getVal {File Field} {
   variable dbObj
   if {[catch {$dbObj find File $File} row] } then {
      return ""
   }
   set val [$dbObj get $row $Field]
   return $val
}
#------------------------------------------------------------------------------
#  Get a single element to some value in the database
#------------------------------------------------------------------------------
proc db::getValues {File Fields} {
   variable dbObj
   if {[catch {$dbObj find File $File} row] } then {
      return [list] ;# TBD: return blank
   }
   set val [eval $dbObj get $row $Fields]
   return $val
}

#----------------------------------------------------------------------------
# Increment the rating for a viewed folder (or folders)
#----------------------------------------------------------------------------
proc db::bumpFolderRating {pattern depth {num 1}} {
   set fileNames [db::filesInFolder $pattern $depth]
   foreach {file} $fileNames {
      db::bumpRating $file $num
   }
   trackList::updateFiles $fileNames

}
#----------------------------------------------------------------------------
# Set the rating to some value
#----------------------------------------------------------------------------
proc db::setFolderRating {pattern depth rating} {
   set fileNames [db::filesInFolder $pattern $depth]
   variable dbObj
   foreach {file} $fileNames {
      if {[catch {$dbObj find File $file} row] } then {
         debug $row
         return
      }
      $dbObj set $row Rating $rating
   }
   db::dirty
   trackList::updateFiles $fileNames

}
proc db::bumpTrackRating {tracks {num 1}} {
   foreach {file} $tracks {
      db::bumpRating $file $num
   }
   trackList::updateFiles $tracks
}
#----------------------------------------------------------------------------
# Cache a single file looking up info for it
# If row passed in, assume it already exists
# mode 0: Modify with passed in data
# mode 1: Full Catalog
# mode 2: Full Catalog only if required
#----------------------------------------------------------------------------
proc db::cacheFile {File _Details {mode 1}} {
   Trace
   variable dbObj
   variable changesList
   variable checkTime
   variable fields
   upvar $_Details userDetails
   # Start building command for eval
   set cmd [list $dbObj]
   ##
   # Was row specified? If so it must exist
   # Pull known data into Details
   set existing 1
   if {[string is integer -strict $File]} then {
      array set Details [$dbObj get $File]
      lappend cmd set $File File $Details(File)
      set File $Details(File)
   } else {
      if {[catch {$dbObj find File $File} row] } then {
         lappend cmd insert end File $File
         set Details(cataloged) [clock seconds]
         set existing 0
      } else {
         array set Details [$dbObj get $row]
         if {$Details(cataloged)==0} {
            set Details(cataloged) [clock seconds]
         }
         lappend cmd set $row
      }
   }
   ;##
   ;# File info
   ;#
   if {[catch {file stat $File Stat} result]} {
      debug $result
      return "Stat Error"
   }
   ##
   # Read the track/id3 info on the file
   # if no data was passed in, otherwise
   # use the user passed in data to override
   # any existing data
   set  result ""
   switch -- $mode {
      0 {
         #----------------------------------------------------------------------------
         # Build command to update only passed in fields
         #----------------------------------------------------------------------------
         foreach Index $fields {
            # Allow blank fields
            if {[info exists userDetails($Index)]} {
               lappend cmd $Index $userDetails($Index)
            }
         }
         set Details(Mounted) 0
         set result "Modifying"
         set skip 1
      }
      1 {
         set result "Cataloging"
         set skip 0
      }
      2 {
         # Only read if required
         if {$existing} {
            if {$Details(FullData)==1 && ($Stat(size)==$Details(size)) && ($Stat($checkTime) <= $Details(date))} {
               set result "Skipping"
               set skip 1
            } else {
               set result "Updating"
               set skip 0
            }
         } else {
            set result "Cataloging"
            set skip 0
         }
      }
      default {
         set skip 0
         set result "Cataloging"
      }
   }
   #----------------------------------------------------------------------------
   # Need to read the data from the file, first fill array with any user data
   #----------------------------------------------------------------------------
   if {$skip==0} {
      # Assume adding the file
      # replace with split calls for track/id3 capture
      if {[catch {catalog::trackFullData $File newDetails} res] } then {
         puts "[me] $res\n$::errorInfo"
         return "Error"
      }
      set Details(FullData) 1
      # merge in newly read details
      #BUG: If existing fields exist, but new fields are blank (i.e. no tag)
      # Then then existing fields will be maintained
      array set Details [array get newDetails]
   } else {
      # Flag it again as mounted and return now
      if {$Details(Mounted)==0} {
         lappend cmd Mounted 1
         eval $cmd
      }
      array set Details [array get userDetails]
      # return now, we wrote user data
      return $result
   }
   #----------------------------------------------------------------------------
   # If play info doesn't exist, then fill it in, these are not found
   # in an ID3 tag or snack info.
   #----------------------------------------------------------------------------
   foreach {field type default} {
      AlbumGain double  "--"
      Gain      double  "--"
      FullData  integer 0
      LastPlay  integer 0
      Played    integer 0
      Rating    ascii   "--"
      Start     integer 0
      Stop      integer 0
      Votes     integer 0
   } {
      if {![info exists Details($field)] || ![string is $type -strict $Details($field)]} {
         set Details($field) $default
      }
   }
   #----------------------------------------------------------------------------
   # These fields always get updated
   #----------------------------------------------------------------------------
   set Details(Depth)  [llength [file split $File]]
   set Details(File)   $File
   set Details(Folder) [file dirname $File]
   set Details(size)   $Stat(size)
   set Details(date)   $Stat($checkTime)
   set Details(Mounted) 1
   #----------------------------------------------------------------------------
   # Build command to update only known fields
   #----------------------------------------------------------------------------
   foreach Index $fields {
      if {[info exists Details($Index)]} {
         lappend cmd $Index $Details($Index)
      }
   }
#if {$Details(cataloged)==0} {
#   puts $cmd
#}
   eval $cmd
   db::dirty
   # Return the updated array to the user
   array set userDetails [array get Details]
   return $result
}
#----------------------------------------------------------------------------
# Function   : db::removeFile
# Description: Remove a file from the database
# Author     : Tom Wilkason
# Date       : 1/19/2003
#----------------------------------------------------------------------------
proc db::removeFile {File} {
   variable dbObj
   variable mods
   Trace
   if {[string is integer -strict $File]} {
       $dbObj delete $File ;# Make sure row indicies work here
       incr mods
   } else {
      if {[catch {$dbObj find File $File} row] } then {
         return 0
      } else {

         # todo: swap with last row and delete last row
         # this will keep row indicies for following tracks the same
#         set lastRow [expr {[$dbObj size]-1}]
#         set lastRowData [$dbObj get $lastRow]
#         eval $dbObj set $row $lastRowData
#         $dbObj delete $lastRow


         $dbObj delete $row ;# Make sure row indicies work here
         incr mods
      }
   }
   return 1
}

#------------------------------------------------------------------------------
# Make the database match the current APL settings
# Keeps a record of APL settings in the tables
#------------------------------------------------------------------------------
proc db::syncToAPL {args} {
   variable dbObj
   $dbObj loop row {
      $dbObj set $row APL [aplFolder::isSet [$dbObj get $row Folder]]
   }
   db::dirty
   return 1
}
#------------------------------------------------------------------------------
# Sync the APL tracks to some arbitray folder array
#------------------------------------------------------------------------------
proc db::syncToAPLVar {_folderVar} {
   variable dbObj
   upvar $_folderVar folderVar
   $dbObj loop row {
      set F [$dbObj get $row Folder]
      if {[info exists folderVar($F)] && $folderVar($F)} {
         set val 1
      } else {
         set val 0
      }
      $dbObj set $row APL $val
   }
   return 1
}

#------------------------------------------------------------------------------
# Build a RE expression that includes ratings
# regexp {^\*$|^\*\*$} will match one or two *
# [mk::select $db::DB -regexp Rating [db::buildRE [list * **]]]
#------------------------------------------------------------------------------
proc db::buildRatingRE {args} {
   variable ratingOptions
   set toRate [list]
   foreach rating $ratingOptions {
      if {$::snackAmpSettings(cbRating,$rating)} {
         lappend toRate $rating
      }
   }
   return [buildRE $toRate]
}
proc db::buildRE {items} {
   set patt [list]
   foreach {item} $items {
      lappend patt "^[string map {* \\*} $item]$"
   }
   return [join $patt "|"]
}
#------------------------------------------------------------------------------
# Return a master list of files with APL flag set
# syncToAPL should have been called sometime earlier to make sure
# the database is updated
#------------------------------------------------------------------------------
proc db::aplFiles {{max 0}} {
   Trace
   variable DB
   variable dbObj
   ##
   # Create the folder tree entries if needed
   # (whey is code below necessary here?)
   if {[aplFolder::entries]==0} {
      aplFolder::createTeFolders 0
   }
   # Search folders that have the APL flag set
   set insertList [list]
   set re [db::buildRatingRE]
   foreach {row} [mk::select $DB -count $max -exact Mounted 1 -exact APL 1 -regexp Rating $re] {
      set name [$dbObj get $row File]
      lappend insertList $name
   }
   return $insertList
}

#------------------------------------------------------------------------------
# Return the row number of APL files, no ratings
#------------------------------------------------------------------------------
proc db::allAplFiles {args} {
   variable DB
   variable dbObj
   set sortList [list]
   foreach {row} [mk::select $DB -exact Mounted 1 -exact APL 1] {
      lappend sortList [$dbObj get $row File]
   }
   return $sortList
}
#------------------------------------------------------------------------------
# Run some routines to clear out the database for cruft

#------------------------------------------------------------------------------
proc db::cleanDatabase {args} {
   set ans [tk_messageBox -icon question -type yesno -title "Please Respond" \
   -message "Dou want want to remove all database entries for unmounted fixed tracks?"]
   if {$ans eq "yes"} {
      catalog::validateMediaPaths
      set removed [db::removeUnmountedFixedFiles *]
      waitWindow::close
      tk_messageBox -icon info -type ok -message "A total of $removed unmounted tracks were removed from the database"
   }
   set ans [tk_messageBox -icon question -type yesno -title "Please Respond" \
   -message "Do you want to verify each non-portable database entry exists?"]
   if {$ans eq "yes"} {
      set removed [db::chkFixed *]
      tk_messageBox -icon info -type ok -message "A total of $removed missing tracks were removed from the database"
      waitWindow::close
   }
}
#------------------------------------------------------------------------------
# Remove all unmounted entries
#------------------------------------------------------------------------------
proc db::removeUnmountedFiles {pattern {dialog 1}} {
   variable dbObj
   variable mods
   variable DB
   set rows [list]
   set removed 0
   # Since rows are just integers mix and match db calls for speed
   set rows [mk::select $DB -glob File $pattern -exact Mounted 0]
   set tot [llength $rows]
   foreach row [lsort -integer -decreasing $rows] {
      $dbObj delete $row
      if {$dialog && [incr removed]%10==0} {
         waitWindow::percent $removed $tot "Removing Stale Entries under\n$pattern\n"
      }
   }
   if {$removed > 0} {
      saLog "Removed $removed unmounted tracks under $pattern"
      db::dirty
   }
   return $removed
}
#----------------------------------------------------------------------------
# Remove unmounted entries that are not in portable media path
#----------------------------------------------------------------------------
proc db::removeUnmountedFixedFiles {pattern} {
   variable dbObj
   variable mods
   variable DB
   set pth [list]
   foreach {p} [lsort -unique $::snackAmpSettings(mp3Player)] {
      lappend pth [file join $p *]
   }
   set removed 0
   set i 0
   set rows [mk::select $DB -glob File $pattern -exact Mounted 0]
   set len [llength $rows]
   foreach row [lsort -integer -decreasing $rows] {
      set isPP 0
      incr i
      set File [$dbObj get $row File]
      foreach {p} $pth {
         if {[string match $p $File]} {
            set isPP 1
            break
         }
      }
      # Only processed fixed folders
      if {$i%300==0} {
         waitWindow::percent $i $len "Removed $removed unmounted fixed tracks" 1
         if {[waitWindow::ifbreak]} {
            break
         }
      }
      if {$isPP==0} {
         $dbObj delete $row
         if {[incr removed]%100==0} {
            waitWindow::bump
         }
      }
   }
   if {$removed > 0} {
      saLog "Removed $removed unmounted tracks under $pattern"
      db::dirty
   }
   return $removed
}
#----------------------------------------------------------------------------
# Verify that each Mounted file exists, if not then remove from the database.
#----------------------------------------------------------------------------
proc db::chkMounted {pattern} {
   variable dbObj
   variable DB
   set result [list]
   set rows [lsort -integer -decreasing [mk::select $DB -exact Mounted 1 -glob File $pattern]]
   set len [llength $rows]
   set i 0
   set removed 0
   foreach {row} $rows {
      incr i
      set isPP 0
      set File [$dbObj get $row File]
      if {![file exists $File]} {
         saLog "Missing:$File"
         db::removeFile $File
         incr removed
      }
      if {$i%30==0} {
         waitWindow::percent $i $len "($row) Checking for Missing Files ($removed Removed)..." 1
         if {[waitWindow::ifbreak]} {
            break
         }
      }
   }
   return  $removed
}
#----------------------------------------------------------------------------
# Check all of the fixed folders for existence
#----------------------------------------------------------------------------
proc db::chkFixed {pattern} {
   variable dbObj
   variable DB
   set result [list]
   set rows [lsort -integer -decreasing [mk::select $DB -glob File $pattern]]
   set len [llength $rows]
   set i 0
   set removed 0
   set pth [list]
   foreach {p} [lsort -unique $::snackAmpSettings(mp3Player)] {
      lappend pth [file join $p *]
   }
   foreach {row} $rows {
      incr i
      set isPP 0
      set File [$dbObj get $row File]
      # Check if a portable file
      foreach {p} $pth {
         if {[string match $p $File]} {
            set isPP 1
            break
         }
      }
      # Only processed fixed folders
      if {$isPP==0} {
         if {![file exists $File]} {
            saLog "Missing:$File"
            db::removeFile $File
            incr removed
         }
         if {$i%30==0} {
            waitWindow::percent $i $len "($row) Checking for Missing Files ($removed Removed)..." 1
            if {[waitWindow::ifbreak]} {
               break
            }
         }
      }
   }
   return $removed
}

#----------------------------------------------------------------------------
# Remove a branch from the database, work backwards
#----------------------------------------------------------------------------
proc db::removeBranch {pattern} {
   variable dbObj
   variable DB
   Trace
   set rows [lsort -integer -decreasing [mk::select $DB -glob File $pattern]]
   foreach row $rows {
      $dbObj delete $row
      db::dirty
   }
   return [llength $rows]
}
#----------------------------------------------------------------------------
# Function   : db::renameFile
# Description: renameFile a file
# Author     : Tom Wilkason
# Date       : 1/19/2003
# Note: Because this is a sorted view, we must change the filename last
#       because it is moved within the view. (bug)
#----------------------------------------------------------------------------
proc db::renameFile {File newName} {
   Trace
   variable dbObj
   variable DB
   if {[string is integer -strict $File]} {
      $dbObj set $File Folder [file dirname $newName] Depth [llength [file split $newName]] File $newName
      db::dirty
   } else {
      if {[catch {$dbObj find File $File} row] } then {
         saLog "[me] $File not found"
         return -1
      } else {
         $dbObj set $row Folder [file dirname $newName] Depth [llength [file split $newName]] File $newName
         db::dirty
      }
   }
   return 1
}
#----------------------------------------------------------------------------
# Close the database
#----------------------------------------------------------------------------
proc db::closeDB {args} {
   Trace
   variable state
   variable mods
   db::commit
   set mods 0
   db::unlock
   if {$state ne "close"} {
      mk::file close db
      set state "close"
   }
}
#----------------------------------------------------------------------------
# If the database has been flaged as modified, commit it and clear flag
#----------------------------------------------------------------------------
proc db::commit {args} {
   variable state
   variable mods
   set m $mods
   if {$mods > 0 && $state ne "close"} {
      Trace
      db::Commit
#saLog "Commiting $m metakit Changes"
      return $m
   } else {
      return 0
   }
}
#----------------------------------------------------------------------------
# Function   : db::Commit
# Description: Flush the data after cataloging
# Author     : Tom Wilkason
# Date       : 1/7/2001
#----------------------------------------------------------------------------
proc db::Commit {} {
   Trace
   variable mods
   mk::file commit db
   set mods 0
}
#----------------------------------------------------------------------------
# Flag database as dirty, use for commiting changes
#----------------------------------------------------------------------------
proc db::dirty {args} {
   variable state
   variable mods
   variable commitRate
   variable commitAfter
   incr mods
   return
}
#----------------------------------------------------------------------------
# Return the number of tracks indexed
# Note: Should limit to media folders
#----------------------------------------------------------------------------
proc db::numTracks {args} {
   variable dbObj
   return [$dbObj size]
}
##
# Relative to above: This is faster near the root, slower at the tail
# //Bertha/Music/Rock/Album Rock/Moody Blues/Time Traveller
#----------------------------------------------------------------------------
proc db::filesInFolder {folder {depth 0}} {
   Trace
   variable DB
   variable dbObj
   set result [list]
   set folder [string trimright $folder "/*"]
   if {$depth>0} {
      set pattern [file join $folder *]
      set maxDepth [expr {[llength [file split $folder]]+$depth+1}]
      set rows [mk::select $DB -glob File $pattern -max Depth $maxDepth -exact Mounted 1]
   } else {
      # Used mostly by http stuff, group by Album name
      set rows [mk::select $DB -sort {Album File}  -exact Folder $folder -exact Mounted 1]
   }
   foreach {row} $rows {
      lappend result [$dbObj get $row File]
   }
   return $result
}
#----------------------------------------------------------------------------
# Return a list of albums or other fields in a folder
#----------------------------------------------------------------------------
proc db::fieldsInFolder {folder field} {
   Trace
   variable DB
   variable dbObj
   set rows [mk::select $DB -exact Folder $folder -exact Mounted 1]
   foreach {row} $rows {
      set fld [$dbObj get $row $field]
      if {[info exists result($fld)]} {
         incr result($fld)
      } else {
         set result($fld) 1
      }
   }
   return [array names result]
}
#----------------------------------------------------------------------------
# Return a list of files in an array indexed by passed in field name
# e.g. array of artists with tracks containing each artist
#----------------------------------------------------------------------------
proc db::dataInFolder {folder field _var} {
   Trace
   upvar $_var result
   variable DB
   variable dbObj
   set hits 0
   set rows [mk::select $DB -exact Folder $folder -exact Mounted 1]
   foreach {row} $rows {
      foreach {File fld} [$dbObj get $row File $field] {break}
      lappend result($fld) $File
      incr hits
   }
   return $hits
}
#----------------------------------------------------------------------------
# Return a list of files not matching a pattern
# Needs metakit select -not option
#----------------------------------------------------------------------------
proc db::filesNotMatchingPattern {pattern {ignoreCase 0}} {
   Trace
   variable DB
   variable dbObj
   set result [list]
   if {$ignoreCase} {
      foreach {row} [mk::select $DB -exact Mounted 1] {
         set name [$dbObj get $row File]
         if {![string match -nocase $pattern $name]} {
            lappend result $name
         }
      }
   } else {
      foreach {row} [mk::select $DB -exact Mounted 1] {
         set name [$dbObj get $row File]
         if {![string match $pattern $name]} {
            lappend result $name
         }
      }
   }
   return $result
}
#----------------------------------------------------------------------------
# Return files matching some pattern
#----------------------------------------------------------------------------
proc db::filesMatchingPattern {pattern {ignoreCase 0}} {
   Trace
   variable DB
   variable dbObj
   set result [list]
   if {$ignoreCase} {
      set match -globnc
   } else {
      set match -glob
   }
   foreach {row} [mk::select $DB $match File $pattern -exact Mounted 1] {
      lappend result [$dbObj get $row File]
   }
   return $result
}
# ----------------------------------------------------------------------------
# Return number of tracks matching a pattern
# ----------------------------------------------------------------------------
proc db::countMatchingPattern {pattern {ignoreCase 0}} {
   Trace
   variable DB
   variable dbObj
   set result [list]
   if {$ignoreCase} {
      set match -globnc
   } else {
      set match -glob
   }
   return [llength [mk::select $DB $match File $pattern -exact Mounted 1]]
}
#----------------------------------------------------------------------------
# Return a list of size and duration for tracks matching a pattern
# unused
#----------------------------------------------------------------------------
proc db::statsMatchingPattern {pattern {ignoreCase 0}} {
   Trace
   variable DB
   variable dbObj
   set result [list]
   if {$ignoreCase} {
      set match -globnc
   } else {
      set match -glob
   }
   set Size 0
   set Duration 0
   set rows  [mk::select $DB $match File $pattern -exact Mounted 1]
   foreach {row} $rows {
      foreach {size duration} [$dbObj get $row size Duration] {break}
      incr Size $size
      set Duration [expr {$Duration+$duration}]
   }
   return [list [llength $rows] $Size $Duration]
}
#----------------------------------------------------------------------------
# test routine to reset all the cataloged times to non-zero value
#----------------------------------------------------------------------------
proc db::checkCatalogDate {args} {
   variable DB
   variable dbObj
   variable checkTime
   set fixed 0
   foreach {row} [mk::select $DB  -exact Mounted 1] {
      foreach {cataloged File} [$dbObj get $row cataloged File] {break}
      if {$cataloged==0} {
         if {![catch {file stat $File Stat} result]} {
            $dbObj set $row cataloged $Stat($checkTime)
            incr fixed
         }
      }
   }
   db::dirty
   saLog "Corrected $fixed missing catalog time records"
}
#----------------------------------------------------------------------------
# mk::select w/o sort is 80% faster than dbObj loop or mk::select with sort
#----------------------------------------------------------------------------
proc db::sortedList {args} {
   Trace
   variable dbObj
   variable DB
   variable SORT
   set result [list]
   foreach {row} [mk::select $DB -exact Mounted 1] {
      lappend result [$dbObj get $row File]
   }
   if {$SORT} {
      return $result
   } else {
      # sort manually
      return [lsort $result]
   }
}
#----------------------------------------------------------------------------
# Make sure all of the tracks Folder property matches the dirname of the file name
#----------------------------------------------------------------------------
proc db::fixFolders {args} {
   Trace
   variable dbObj
   variable DB
   set result [list]
   # Looping is very fast
   # unsorted object selects don't work with metakit
   $dbObj loop row {
      foreach {File Folder} [$dbObj get $row File Folder] {break}
      set dname [file dirname $File]
      if {$Folder ne $dname} {
         puts "$File not in $Folder"
         update
         $dbObj set $row Folder $dname
      }
   }
   return
}
#----------------------------------------------------------------------------
# Return a list of tracks shuffled then sub-sorted by user specified field
# assumes integer sub-sorting
# modes are a list of textual modes to sort on, either 1,2 or 3 are passed in.
# This uses a lot of memory the first time.
#----------------------------------------------------------------------------
proc db::subSorting {modes {tracks ""}} {
   variable DB
   variable dbObj
   variable ratingOptions
   global snackAmpSettings
   set db::isSorted 1 ;# not needed here
   set result [list]
   set pairs [list]
   # each mode contains {field -type -sortmode}
   set index 1
   ##
   # Append the fields to be queried
   #
   lappend fields "File"
   set i 1
   foreach {subsort} [lflip $modes] {
      foreach {field type sort} $subsort {break}
      lappend fields $field
      incr i
   }
   ##
   # Build the master sorted list
   set v [subSort::getWindow]
   # Build search list to restrict to valid ratings
   set ro [list]
   foreach rating [lsort $ratingOptions] {
      if {$::snackAmpSettings(cbRating,$rating)} {
         lappend ro $rating
      }
   }
   ##
   # APL Files, do everything in one pass rather than call db::aplFiles
   #
   if {[llength $tracks]==0} {
      ##
      # Create the folder tree entries if needed
      #
      if {[aplFolder::entries]==0} {
         aplFolder::createTeFolders 0
      }
      set insertList [list]
      set re [db::buildRatingRE]
      foreach {row} [mk::select $DB -count [settings::maxAPL] -exact Mounted 1 -exact APL 1 -regexp Rating $re] {
         array set Data [$dbObj get $row]
         # Round the LastPlay appropriately
         set Data(LastPlay)  [expr {($Data(LastPlay)/$v)*$v}]
         set Data(date)      [expr {($Data(date)/$v)*$v}]
         set Data(cataloged) [expr {($Data(cataloged)/$v)*$v}]
         set row [list]
         foreach {field} $fields {
            lappend row $Data($field)
         }
         lappend pairs $row
      }
   } else {
      ##
      # Incoming Playlist
      #
      foreach {track} $tracks {
         unset -nocomplain Data
         # We will NOT catalog track if needed.
         # NOTE: This is redundant to applying an APL
         if {[db::lookup $track Data 0] > 0} {
            # Only allow flagged codes
            if {($Data(Rating) eq "") || ([lsearch -exact -sorted $ro $Data(Rating)] >=0)} {
               # Round the LastPlay/Timestamp appropriately
               set Data(LastPlay)  [expr {($Data(LastPlay)/$v)*$v}]
               set Data(date)      [expr {($Data(date)/$v)*$v}]
               set Data(cataloged) [expr {($Data(cataloged)/$v)*$v}]
               set row [list]
               foreach {field} $fields {
                  lappend row $Data($field)
               }
               lappend pairs $row
            }
         } else {
            saLog "[me] No data found for '$track'";update
         }
      }
   }
   ##
   # Determine if a shuffle is in the inner
   #
   if {$snackAmpSettings(sortShuffle)} {
      Update
      set pairs [shuffle $pairs]
   }
   ##
   # Then subsort, in reverse order
   #
   foreach {subsort} [lflip $modes] {
      foreach {field type sort} $subsort {break}
      Update
      set pairs [lsort $sort $type -index $index $pairs]
      incr index
   }
   # Return just the file names
   # Bug: Below fails if not sorting is done at all
   # because only File is returned as a single element and
   # braces in the file name cause an error for lindex
   if {[catch {
      foreach item $pairs {
         lappend result [lindex $item 0]
      }
   } res] } then {
      saLog "$res: '$item'"
   }
   ##
   # Morph the list a bit
   #
   if {$snackAmpSettings(postmorph)} {
      Update
      lipmorph result $snackAmpSettings(postmorphamount)
   }
   return $result
}
#----------------------------------------------------------------------------
;# Position in db of a file, -1 if not in db
#----------------------------------------------------------------------------
proc db::whereInDb {File} {
   Trace
   variable dbObj
   if {[catch {$dbObj find File $File} row] } then {
      return -1
   } else {
      return $row
   }
}
#----------------------------------------------------------------------------
# First position after some offset, allow glob matching
#----------------------------------------------------------------------------
proc db::matchInDb {pattern {offset 0}} {
   Trace
   variable dbObj
   set result [list]
   set view $dbObj
   set end [$view size]
   # dbObj is already sorted (not really)
   for {set row $offset} {$row < $end} {incr row} {
      foreach {File Mounted} [$view get $row File Mounted] {break}
      if {$Mounted && [string match -nocase $pattern $File]} {
         return $row
      }
   }
   return -1
}
#------------------------------------------------------------------------------
# Return a track name at some row
#------------------------------------------------------------------------------
proc db::TrackFromDb {row} {
   Trace
   variable dbObj
   if {[catch {$dbObj get $row File} File] } then {
      return ""
   } else {
      return $File
   }
}
#----------------------------------------------------------------------------
# Rename a branch of files at once
# Usefull if the underlying folder gets relocated but the database data
# is still correct.
# NOTE: Only works correctly if the database is kept unsorted!!
# Note: Nevery use mk::set File since hashing will break, use dbObj
#----------------------------------------------------------------------------
proc db::renameBranch {branch newBranch} {
   Trace
   variable DB
   variable dbObj
   set pairs [list]
   set pattern [file join $branch *]
   # Consistent trailing / on both
   set branch "[string trimright $branch /]/"
   set newBranch "[string trimright $newBranch /]/"
   if {$branch ne $newBranch} {
      set rows [mk::select $DB -glob File $pattern]
      foreach {row} $rows {
         set file [$dbObj get $row File]
         lappend pairs $file
         # replace front part
         set file [regsub -- ^$branch $file $newBranch]
         set folder [file dirname $file]
         set depth [llength [file split $file]]
         $dbObj set $row File $file Folder $folder Depth $depth
         db::dirty
         lappend pairs $file
      }
   }
   return $pairs
}
#----------------------------------------------------------------------------
# Flag the files in pattern as mounted/unmounted
# Can I do this? Is DB via a hash map view?
#----------------------------------------------------------------------------
proc db::mountBranch {pattern how} {
   variable DB
   variable dbObj
   foreach {row} [mk::select $DB -glob Folder $pattern -exact Mounted [expr {!$how}]] {
      $dbObj set $row Mounted $how
   }
   db::dirty
   return 1
}
#----------------------------------------------------------------------------
# Similar to mountBranch, just a bit slower but gets files in base folder
#----------------------------------------------------------------------------
proc db::mountFiles {pattern how} {
   variable DB
   variable dbObj
   foreach {row} [mk::select $DB -glob File $pattern -exact Mounted [expr {!$how}]] {
      $dbObj set $row Mounted $how
   }
   db::dirty
   return 1
}
#----------------------------------------------------------------------------
# Flag a single folder as mounted or unmounted
#----------------------------------------------------------------------------
proc db::mountFolder {folder how} {
   variable DB
   variable dbObj
   set rows [mk::select $DB -exact Folder $folder -exact Mounted [expr {!$how}]]
   foreach {row} $rows {
      $dbObj set $row Mounted $how
   }
   db::dirty
   return 1
}
#----------------------------------------------------------------------------
# Return a complete tree, including paths to the root.
# TODO: Make a database of the folders, hashed by folder name
# Have only a mounted field
#----------------------------------------------------------------------------
proc db::buildFolderTree {{pattern *}} {
   variable DB
   variable dbObj
   aplFolder::stripBranch $pattern
   set hits 0
   set how 0
   set rows [mk::select $DB -glob Folder $pattern -exact Mounted 1]
   foreach {row} $rows {
      set folder [$dbObj get $row Folder]
      incr hits [aplFolder::setToRoot $folder $how]
   }
   return $hits
}
##
# Return raw database statistics
#
proc db::statsTree {_var {pattern ""}} {
   Trace
   variable DB
   variable dbObj
   upvar $_var stats
   set rows [list]
   set i 0
   set stats(Cataloged) 0
   set stats(UnMounted) 0
   set stats(Size) 0
   set stats(Dur) 0
   set fields [list Folder Mounted FullData Artist Album Genre size Duration]

   if {[string length $pattern]==0} {
      set rows [mk::select $DB]
   } elseif {$pattern eq "/Fixed Folders/*"} {
      foreach dir $::snackAmpSettings(mediaDirs) {
         set slice [mk::select $DB -glob File $dir*]
         foreach {row} $slice {
            lappend rows $row
         }
      }
   } elseif {$pattern eq "/Attached Folders/*"} {
      foreach dir $::snackAmpSettings(mp3Player) {
         set slice [mk::select $DB -glob File $dir*]
         foreach {row} $slice {
            lappend rows $row
         }
      }
   } else {
      set rows [mk::select $DB -glob File $pattern]
   }
   foreach row $rows {
      foreach val [$dbObj get $row Folder Mounted FullData Artist Album Genre size Duration] index $fields {
         switch -- $index {
            "FullData" {
               if {$val} {
                  incr stats(Cataloged)
               }
            }
            "Mounted" {
               if {$val==0} {
                  incr stats(UnMounted)
               }
            }
            "size" {
               incr stats(Size) [expr {$val/1024}]
            }
            "Duration" {
               set stats(Dur) [expr {$stats(Dur)+$val}]
            }
            default {
               set hash($index,$val) {}
            }
         }
      }
      incr i
      if {$i%500==0} {
         update
      }
   }
   foreach field $fields {
      set stats($field) [llength [array names hash $field,*]]
   }
   set stats(Files) $i
   set stats(Size) [expr {$stats(Size)/1024}]
   return $i
}
##
# Return text of database statistics
# TODO: If "Fixed Folders" or "Attached Folders" passed in then show all data
proc db::statsText {fill {pattern ""}} {
   Wait 1
   db::statsTree Data [file join $pattern *]
   foreach {index tit} {
      Files     "Indexed Tracks"
      Cataloged "Cataloged Tracks"
      UnMounted "UnMounted Tracks"
      Size      "Size (MB)"
      Folder    Folders
      Artist    Artists
      Album     Albums
      Genre     Genres
   } {
      append result [format "%-9s\t%s\n" [commify $Data($index)] $tit]
   }
   set secs [expr {round($Data(Dur))}]
   set days "[expr {$secs/86400}] days "
   if {$secs == 0} {
      set days ""
   }
   append result "Duration\t$days[clock format $secs -gmt 1 -format %T]"

   Wait 0
   return [string trimright $result]
}
#----------------------------------------------------------------------------
# Return a list of distinct items in the database
#----------------------------------------------------------------------------
proc db::unique {item} {
   variable dbObj
   $dbObj loop row {
      set hash([$dbObj get $row $item]) {}
   }
   return [lsort [array names hash]]
}

proc db::unique1 {item} {
   variable dbObj
   $dbObj loop row {
      lappend hash [$dbObj get $row $item]
   }
   return [lsort -unique $hash]
}
#------------------------------------------------------------------------------
# Not sure how to use these grouped views yet, maybe for a media
# viewer like winamp5 has.
# view view groupby subview ?prop ...?
#
# Returns a new view derived from view as described below.  This is almost
# an inverse of [view view flatten subview].  The argument subview must be
# the name of an existing property in view that is itself a view.  The
# properties of the returned view will be the properties prop followed by
# the property subview.  The properties of property subview in the returned
# view will be all the properties of view not named among the prop
# arguments.  Any properties of subview as defined in view do not appear in
# the returned view.  There is one row in the returned view for every unique
# combination of values of the prop properties in view.  For each row in
# view with that combination, a row will be added to the subview view
# containing the values of the properties not named as prop arguments.
#------------------------------------------------------------------------------
proc db::subViews {args} {
   variable dbObj
   set res [list]
   # group by folders
   set fv [$dbObj view groupby ffsv:V Folder]
   # loop on folders
   $fv loop c {
puts $c
      set sv [$fv open $c ffsv]
      # loop on files in folder
      $sv loop csv {
         lappend res [$sv get $csv File]
      }
      $sv close
      unset sv
$fv close
unset fv
return $res
   }
}


#------------------------------------------------------------------------------
# Return a subset of passed in files that match the pattern
# Used for searching the playlist or AutoPlaylist files
#------------------------------------------------------------------------------
proc db::filteredView {files fields pattern re} {
   variable dbObj
   set result [list]
   # This seems very fast now that we keep $dbObj sorted
   # -regexp Rating $re doesn't seem to work on select like it does on mk::select
   set view [$dbObj select -sort File -regexp Rating $re -globnc $fields *$pattern*]
   foreach {File} $files {
      if {![catch {$view find File $File} row] } then {
         lappend result [$view get $row File Album Artist Title Folder Rating]
      }
   }
   $view close
   return $result
}
#----------------------------------------------------------------------------
# db::tableTreeData
#
# Loop over some depth of folders and return a nested list of data specified
# in the col3Use in tableData as well as the total size and time of the tracks.
# Update locate cache files of the data for faster access next time.
# todo:: Relies on a couple TA varibles and tableList::variables, need
# to decouple them
# TBD: Is filtering on all tracks optimum here? Not for file names only, but
#      for also filtering tag data it is the only way.
#----------------------------------------------------------------------------
proc db::tableTreeData {baseIndex col2Use checkStat depth filter filterHow _totalTime _totalSize _tableData _Index2File} {
   Trace
   variable dbObj
   variable DB
   variable checkTime
   # These are returned data
   upvar $_totalTime totalTime
   upvar $_totalSize totalSize
   upvar $_tableData tableData
   upvar $_Index2File Index2File
   ##
   # Grab all the rows for this depth
   #
   set indLen [string length $baseIndex]
   incr indLen
   set maxDepth [expr {[llength [file split $baseIndex]]+$depth+1}]
   # Tweak select for speed, filtering is against file names
   if {[string length $filter]} {
      Wait 1
      set indLen 0 ;# show entire folder
      set rows [mk::select $DB -count [settings::maxAPL] -globnc $filterHow *$filter* -exact Mounted 1]
   } elseif {$baseIndex eq ""} {
      # Select everything
      Wait 1
      set rows [mk::select $DB -count [settings::maxAPL] -exact Mounted 1]
   } elseif {$depth==0} {
      # Select just one folder
      set rows [mk::select $DB -exact Folder $baseIndex -exact Mounted 1]
   } else {
      # Select multiple folders
      set rows [mk::select $DB -glob File [file join $baseIndex *] -max Depth $maxDepth -exact Mounted 1]
   }
   ##
   # Build the command line to append each row for later eval
   #
   lappend cmdLine lappend rowData
   foreach {col} $col2Use {
      lappend cmdLine "\$Data($col)"
   }
   set cmdLine [join $cmdLine]
   set records [llength $rows]
   set i 0
   # Iterate over each row
   # bug if only one record returned (row 0), gets a failure
   set flen [llength $baseIndex]
   foreach {row} $rows {
      incr i
      ##
      # Grab the entire row into an array
      #
      unset -nocomplain Data
      array set Data [$dbObj get $row]
      set fullFile $Data(File)
      # Only display tail
      set Data(File) [file tail $fullFile]
      # Cache if necessary
      # Todo: Can be optimized for the case where checkStat is set
      #       Can we use 'cataloged' instead?
      if {$checkStat} {
         if {[catch {file stat $fullFile Stat} result] } then {
            saLog $result
            continue
         }
         set changed [expr {($Stat(size)!=$Data(size)) || ($Stat($checkTime) > $Data(date))}]

         # Formatted fields are shown in tablelist
         set Data(Size) $Stat(size)
         set Data(Date) [timeFormat $Stat($checkTime)]
      } else {
         set Data(Size) $Data(size)
         set Data(Date) [timeFormat $Data(date)]
         set changed 0
      }
      # This LastPlay is not rounded
      if {$Data(LastPlay)>0} {
         set Data(LastPlay) [timeFormat $Data(LastPlay)]
      } else {
         set Data(LastPlay) "Not Played"
      }
      set Data(Sample) $Data(sampRate)
      if {$Data(FullData)==0 ||($changed)} {
         set res [db::cacheFile $fullFile Data 2]
         set Data(File) [file tail $Data(File)]
         waitWindow::percent $i $records "$res\n${baseIndex}..." 1
         if {[waitWindow::ifbreak]} {
            break
         }
      } elseif {($i%200)==0} {
         Update
      }
      set Data(Cataloged) [timeFormat $Data(cataloged)]

      # Get relative folder info
      set Data(Folder) [string range $Data(Folder) $indLen end]
      if {$Data(Folder) eq ""} {
         set Data(Folder) .
      }
      set Data(Gain) [gainControl::levelText $Data(Gain)]
      set Data(AlbumGain) [gainControl::levelText $Data(AlbumGain)]

      catch {
      ;# TFW: Note size vs Size, need to standardize
         set totalSize [expr {$totalSize+$Data(size)/1024}]
         set totalTime [expr {$totalTime+$Data(Duration)}]
      }
#puts "$fullFile: $Data(Date) $Data(Size) $Data(Sample)"
      # Build row of desired data
      set Data(Key) $row
      set rowData [list]

      # Build row and append row to result
      eval $cmdLine
      lappend tableData $rowData
   }
   Wait 0
   return 0
}
#----------------------------------------------------------------------------
# Remove then rescan a folders worth of tracks
# Doesn't remove old entries
# Todo: Add option of only cacheFile 2 (if needed)
#----------------------------------------------------------------------------
proc db::cacheFolderDetails {folder {depth 0}} {
   # Remove any stale database entries
   set files [db::filesInFolder $folder $depth]
   foreach {file} $files  {
      if {![file exists $file]} {
         db::removeFile $file
      }
   }
   update
   set tryFiles [eval glob -nocomplain -directory {$folder} [globAllPattern $depth]]
   set total [llength $tryFiles]
   # Add any files that may exist and force recache all data
   set i 0
   foreach {file} $tryFiles {
      if {[catalog::isTrackFile $file]} {
         unset -nocomplain Data
         set how [db::cacheFile $file Data 1]
         waitWindow::percent [incr i] $total "Cataloging\n[splitFile $file]" 1
         if {[waitWindow::ifbreak]} {
            return
         }
      }
   }
}

#----------------------------------------------------------------------------
# Catalog some path to some depth
#----------------------------------------------------------------------------
proc db::catalog {baseIndex checkStat depth} {
   Trace
   package require snID3
   variable dbObj
   variable DB
   variable checkTime
   ##
   # Grab all the rows for this depth
   #
   set indLen [string length $baseIndex]
   incr indLen
   set maxDepth [expr {[llength [file split $baseIndex]]+$depth+1}]
   # Tweak select for speed, filtering is against file names
   if {[string length $baseIndex]==0} {
      # Select everything
      set rows [mk::select $DB -exact Mounted 1]
   } elseif {$depth==0} {
      # Select just one folder
      set rows [mk::select $DB -exact Folder $baseIndex -exact Mounted 1]
   } else {
      # Select multiple folders
      set rows [mk::select $DB -glob File [file join $baseIndex *] -max Depth $maxDepth -exact Mounted 1]
   }
   set i 0
   # Iterate over each row
   # bug if only one record returned (row 0), gets a failure
   set flen [llength $baseIndex]
   set records [llength $rows]
   foreach {row} $rows {
      ##
      # Grab the entire row into an array
      #
      incr i
      unset -nocomplain Data
      array set Data [$dbObj get $row]
      set how [db::cacheFile $Data(File) Data $checkStat]
      if {[string length $how]} {
         waitWindow::percent $i $records "$how Catalog Records\n[splitFile $Data(File)]" 1
         if {[waitWindow::ifbreak]} {
            break
         }
      } elseif {($i%25)==0}  {
         waitWindow::percent $i $records "Checking database for needed updates\n\nPlease Wait\n" 1
         if {[waitWindow::ifbreak]} {
            break
         }
      }
   }
   return $i
}
#----------------------------------------------------------------------------
# Function   : looper
# Description: Commit any changes on a 10 sec clock
# Author     : Tom Wilkason
# Date       : 12/26/2002
#----------------------------------------------------------------------------
proc db::looper {time} {
   after cancel db::looper $time
   db::commit
   after $time db::looper  $time
}
##
# Reset play counts to 0
#
proc db::resetPlayCount {{pattern *}} {
   set files [db::filesInFolder $pattern 100]
   set num [llength $files]
   if {$num > 0} {
         if {[tk_messageBox -icon question -type yesno \
         -title "Please Verify" -message "Are you sure you want to reset $num playcount entries"]=="yes"} {
         foreach {file} $files {
            db::setVal $file Played 0
         }
      }
      trackList::updateFiles $files
   }
   return 1
}
##
# Reset play counts to 0
#
proc db::resetPlayTime {{pattern *}} {
   set files [db::filesInFolder $pattern 100]
   set num [llength $files]
   if {$num > 0} {
         if {[tk_messageBox -icon question -type yesno \
         -title "Please Verify" -message "Are you sure you want to reset $num playcount entries"]=="yes"} {
         foreach {file} $files {
            db::setVal $file LastPlay 0
         }
      }
      trackList::updateFiles $files
   }
   return 1
}
#----------------------------------------------------------------------------
# Return tracks on an album for cover art
#----------------------------------------------------------------------------
proc db::albumTracks {Artist Album Folder} {
   Trace
   variable DB
   variable dbObj
   set result [list]             ;# Artist $Artist
   foreach {row} [mk::select $DB Album $Album Folder $Folder -exact Mounted 1] {
      lappend result [$dbObj get $row Track Title]
   }
   return $result
}
##
# List of mounted folders matching some pattern
#
proc db::foldersMatchingPattern {pattern {sorted 0}} {
   variable DB
   variable dbObj
   # Return all rows that have gain entries
   set rows [mk::select $DB -exact Mounted 1 -glob Folder $pattern]
   foreach {row} $rows {
      set cache([$dbObj get $row Folder]) 1
   }
   if {$sorted} {
      return [lsort [array names cache]]
   }  else {
      return [array names cache]
   }
}

#------------------------------------------------------------------------------
# Function   : gainControl Namespace
# Description: Functions to access the auto-level gain control
# Author     : Tom Wilkason
# Date       : 2/27/2003
#------------------------------------------------------------------------------
namespace eval gainControl {
   variable lastAfter 0
   variable maxGain
   variable minGain
   variable autoLevel
   variable autoLevelSeconds
   variable autoTruncate
   variable tagGainValues
   variable gainFactor
   upvar #0 ::snackAmpSettings(maxGain)          maxGain
   upvar #0 ::snackAmpSettings(minGain)          minGain
   upvar #0 ::snackAmpSettings(autoLevel)        autoLevel
   upvar #0 ::snackAmpSettings(autoLevelSeconds) autoLevelSeconds
   upvar #0 ::snackAmpSettings(autoTruncate)     autoTruncate
   upvar #0 ::snackAmpSettings(tagGainValues)    tagGainValues
   upvar #0 ::snackAmpSettings(upgradeTags)      upgradeTags
   upvar #0 ::snackAmpSettings(gainFactor)       gainFactor
}
proc gainControl::levelText {value} {
   variable autoLevel
   variable gainFactor
   if {$autoLevel && [string is double -strict $value]} {
      if {[catch {expr {log10($value)*$gainFactor}} db] } then {
         set db 0.0
      } else {
         set db [limit $db]
      }
      #set result "[string trim [format %-5.2f $db]] dB"
      set result [format "%.2f %s" $db "dB"]
   } else {
      set result "--"
   }
   return $result
}
#------------------------------------------------------------------------------
# Remove all stale equalization settings
#------------------------------------------------------------------------------
proc gainControl::bgCreate {fname} {
   variable lastAfter
   after cancel $lastAfter
   set lastAfter [after 20000 [list gainControl::createIfNeeded $fname]]
}
#------------------------------------------------------------------------------
# Compute a gain & start/stop setting if needed
# account for prior gain but no start/stop
#------------------------------------------------------------------------------
proc gainControl::createIfNeeded {fname {snd ""}} {
   Trace
   if {![file exists $fname]} {
      saLog "[me]:$fname no longer exists"
      return 0
   }
   # Create the gain
   if {![gainControl::exists $fname]} {
      set gain [gainControl::create $fname $snd]
      addTag $fname $gain "xxx"
   } else {
      # gain exists, make sure start/stop does as well
      set Stop [db::getVal $fname Stop]
      if {[isnumber $Stop] &&  $Stop == 0} {
         createStartStop $fname $snd
      }
   }
   return [gainControl::getModeGain $fname]
}
#------------------------------------------------------------------------------
# Get the gain amount for playing
# Default to regular gain if needed
#------------------------------------------------------------------------------
proc gainControl::getModeGain {fname} {
   global snackAmpSettings
   switch -- $snackAmpSettings(gainType) {
      "Track" {
         set gain [db::getVal $fname "Gain"]
      }
      "Album" {
         set gain [db::getVal $fname "AlbumGain"]
         if {![isnumber $gain]} {
            set gain [db::getVal $fname "Gain"]
         }
      }
      "Album, Track if Shuffled" {
         if {$snackAmpSettings(sortShuffle) || $snackAmpSettings(postmorph)} {
            set gain [db::getVal $fname "Gain"]
         } else {
            set gain [db::getVal $fname "AlbumGain"]
            if {![isnumber $gain]} {
               set gain [db::getVal $fname "Gain"]
            }
         }
      }
      default {
         set gain [db::getVal $fname "Gain"]
      }
   }
   if {[isnumber $gain]} {
      return $gain
   } else {
      saLog "Invalid gain of '$gain' for $fname"
      return 0
   }
}
##
# Auto Level the track, overwrite any existing values
#
proc gainControl::create {fname {snd ""}} {
   variable autoLevelSeconds
   Trace
   if {[catch {
      if {$snd eq ""} {
         set snd [soundControl::infoObj $fname]
      } else {
         $snd configure -file $fname -fileformat [soundControl::getType $fname]
      }
   } result] } then {
      saLog "[me]:$result for $fname"
      return 0
   }
   db::setVal $fname Duration [$snd length -units seconds]
   # Start/Stop
   foreach {Start Stop} [createStartStop $fname $snd] {break}
   Update
   # Track Gain
   set gain [soundControl::logPower $snd $autoLevelSeconds $Start [expr {[$snd length]-$Stop}]]
   db::setVal $fname Gain $gain
   return $gain
}
##
# Find the starting and ending point of the track
#
proc gainControl::createStartStop {fname snd} {
   ##
   # Save the start and stop positions
   #
   # Sound object
   if {[catch {
      if {$snd eq ""} {
         set snd [soundControl::infoObj $fname]
      } else {
         $snd configure -file $fname -fileformat [soundControl::getType $fname]
      }
   } result] } then {
      saLog "[me]:$result for $fname"
      return 0
   }
   ##
   # Compute the start/stop positions and save them
   #
   set Start [soundControl::findStart $snd]
   if {[string is integer -strict $Start]} {
      db::setVal $fname Start $Start
   } else {
      set Start 0
      saLog "[me] Got '$Start' for $fname Start Position"
   }
   Update
   set Stop [soundControl::findStop $snd]
   if {[string is integer -strict $Stop]} {
      db::setVal $fname Stop $Stop
   } else {
      set Stop 0
      saLog "[me] Got '$Stop' for $fname Stop Position"
   }
   return [list $Start $Stop]
}
##
# Return the value for the gain control of a track
#
proc gainControl::get {fname {what "Gain"}} {
   Trace
   set gain [db::getVal $fname $what]
   if {[isnumber $gain]} {
      return $gain
   } else {
      return -code error -1
   }
}

#------------------------------------------------------------------------------
# Compute the slope of a series of data
# 0=no dynamics
# higher numbers have greater dynamics
#------------------------------------------------------------------------------
proc gainControl::slope {data} {
   set sumx 0.0
   set sumy 0.0
   set sum2x 0.0
   set sum2y 0.0
   set sumxy 0.0
   set num 0
   set num [llength $data]
   set i 0
   foreach {y} $data {
      set x $i
      set sumx [expr {$x + $sumx}]
      set sum2x [expr {$x * $x + $sum2x}]
      set sumy [expr {$y + $sumy}]
      set sum2y [expr {$y * $y + $sum2y}]
      set sumxy [expr {$x * $y + $sumxy}]
      incr i
   }

   set denx [expr {$num * $sum2x - $sumx * $sumx}]
   set deny [expr {$num * $sum2y - $sumy * $sumy}]
   set top  [expr {$num * $sumxy - $sumx * $sumy}]
   if {$denx*$deny!=0} {
      set corr [expr {$top / sqrt($denx * $deny)}]
      set slp [expr {$top / $denx}]
   } else {
      set slp 0
      set corr 0
   }
   #set inter [expr {($sumy - $sumx * $slp) / $num}]
   return $slp
}
#----------------------------------------------------------------------------
# Add replay gain value to the ID3V2 tag if required
# Will upgrade to V2 tags if indicated
#----------------------------------------------------------------------------
proc gainControl::addTag {fname trackAmount albumAmount} {
   variable tagGainValues
   variable upgradeTags
   set upgraded 0
   set mod 0
   if {$tagGainValues && [id3Tag::ismp3 $fname]} {
      unset -nocomplain data
      id3Tag::id3Label $fname data {V2 V1}
      if {[array size data] > 0} {
         if {[string match "V1*" $data(Tag)]} {
            if {$upgradeTags == 0} {
               return
            } else {
               incr upgraded
            }
         }
         if {[isnumber $trackAmount]} {
            set gv [format "%+3.3f" [limit [toDb $trackAmount]]]
            if {![info exists data(track_gain)] || ([info exists data(track_gain)] && ($gv ne $data(track_gain)))} {
               set data(track_gain) $gv
               incr mod
            }
         }
         if {[isnumber $albumAmount]} {
            set gv [format "%+3.3f" [limit [toDb $albumAmount]]]
            if {![info exists data(album_gain)] || ([info exists data(album_gain)] && ($gv ne $data(album_gain)))} {
               set data(album_gain) $gv
               incr mod
            }
         }
         # Only write if we need to, set tag type in case it was upgraded
         if {$mod} {
            id3Tag::id3V2Modify $fname data
            db::setVal $fname Tag $data(Tag)
         }
      }
   }
}

#------------------------------------------------------------------------------
# Limit a gain control value
# dB is in dB
#------------------------------------------------------------------------------
proc gainControl::limit {dB} {
   variable maxGain
   variable minGain
   if {$dB < $minGain} {
      set dB $minGain
   } elseif {$dB > $maxGain} {
      set dB $maxGain
   }
   return $dB
}
##
# Reset the gain control values for a track
#
proc gainControl::remove {fname} {
   foreach {type} {Gain AlbumGain} {
      db::setVal $fname $type "--"
   }
   return 0
}
##
# Check if the gain control values exist for a track
#
proc gainControl::exists {fname {type Gain}} {
   # TODO: use define method Gain, AlbumGain (shuffled)
   #Trace 1
   variable upgradeTags
   variable tagGainValues
   set gain [db::getVal $fname $type]
   if {[isnumber $gain]} {
      # Double check the tag exists in the file as well
      if {$tagGainValues} {
         id3Tag::id3Label $fname id3Data V2
         if {[info exists id3Data(track_gain)]} {
            return 1
         } else {
            return 0
         }
      } else {
         return 1
      }
   } else {
      return 0
   }
}
##
# Reset all of the tags
#
proc gainControl::resetAll {{pattern *}} {
   variable DB
   variable dbObj
   set files [list]
   # Return all rows that have gain entries
   set rows [mk::select $DB -exact Mounted 1 -regexp Gain {[0-9]+}]
   set num [llength $rows]
   if {$num > 0} {
         if {[tk_messageBox -icon question -type yesno \
         -title "Please Verify" -message "Are you sure you want to remove $num gain entries"]=="yes"} {
         foreach {row} $rows {
            $dbObj set $row Gain "--" AlbumGain "--"
         }
      }
      db::commit
   }
   return 1
}
##
# Remove some gain control settings
#
proc gainControl::adjustSome {fileNames {percent .1}} {
   global TA LB

   if {[llength $fileNames] == 0} {
      set fileNames [LB::nameSelected]
   }
   foreach {file} $fileNames {
      foreach {type} {Gain AlbumGain} {
         if {![catch {gainControl::get $file $type} gain]} {
            set gain [expr {$gain*(1.0+$percent)}]
            db::setVal $file $type $gain
         } else {
            saLog $gain
         }
      }
   }
   # Update any data that is currently shown
   trackList::updateFiles $fileNames
}

proc gainControl::resetSome {{fileNames {}}} {
   global TA LB

   if {[llength $fileNames] == 0} {
      set fileNames [LB::nameSelected]
   }
   foreach {file} $fileNames {
      gainControl::remove $file
   }
   # Update any data that is currently shown
   trackList::updateFiles $fileNames
}
##
# Make sure each track as a gain embedded in it, upgrade tag if necessary
# todo: Attached to auto-level menu on tree and file
#
proc gainControl::setGains {pattern} {
   set i 0
   set DB  $db::DB
   set result [list]
#   set 2weekago [expr {[clock seconds]-2*7*24*60*60}]
#   set rows [mk::select $DB -exact Mounted 1 -glob File $pattern -min date $2weekago]
   set rows [mk::select $DB -exact Mounted 1 -glob File $pattern]
   set i 0
   set len [llength $rows]
   set gained 0
   set tagged 0
   foreach {row} $rows {
      set fname  [mk::get $DB!$row File]
      if {[file exists $fname]} {
         unset -nocomplain id3Data
         set result [id3Tag::id3Label $fname id3Data {V2 V1}]
         if {[info exists id3Data(Tag)] && [string match "V2*" $id3Data(Tag)]} {
            if {![info exists id3Data(track_gain)]} {
               waitWindow::percent [incr i] $len "Computing Gain For\n[splitFile $fname]" 1
               set val [gainControl::create $fname]
               gainControl::addTag $fname $val "xxx"
               incr gained
            } else {
               waitWindow::percent [incr i] $len "Skipping\n[splitFile $fname]" 1
            }
         } elseif {[array size id3Data] > 2 && [id3Tag::ismp3 $fname]} {
            waitWindow::percent [incr i] $len "Upgrading Tags and Computing Gain For\n[splitFile $fname]" 1
            id3Tag::id3V2Modify $fname id3Data
            set val [gainControl::create $fname]
            gainControl::addTag $fname $val "xxx"
            incr gained
            incr tagged
         } else {
            waitWindow::percent [incr i] $len "No Tag Found\n[splitFile $fname]" 1
            saLog "$fname has no tag"
         }
         if {[waitWindow::ifbreak]} {
            break
         }
      } else {
         saLog "$fname does not exist"
      }
   }
   saLog "Upgraded $tagged V1->V2 tags, Added $gained Gain settings"
   waitWindow::close
   return  1
}

#------------------------------------------------------------------------------
# Perform an album gain on a number of albums where pattern represents
# the file.
#------------------------------------------------------------------------------
proc gainControl::folderAlbumGains {{pattern *}} {
   global teFolders
   set folders [lsort [array names teFolders $pattern]]
   set tail [list]
   set totfolders [llength $folders]
   set donefolders 0
   foreach {folder} $folders {
      unset -nocomplain Data
      if {[db::dataInFolder $folder "Album" Data]} {
         foreach {album} [array names Data] {
            set files $Data($album)
            set totfiles [llength $files]
            if {[llength $files]>5 && $album ne ""} {
               waitWindow::percent [incr donefolders] $totfolders "Computing Gain for\n$album\n$totfiles Tracks" 1
               if {[waitWindow::ifbreak]} {
                  waitWindow::close
                  return 0
               } else {
                  set reset 1
                  set process 0
                  set lastAlbGain ""
                  set lastGain ""
                  #
                  # Determine if we need to reset first (if gains are averaged)
                  #
                  foreach {f} $files {
                     # Determine if album gains are different
                     set albgain [db::getVal $f "AlbumGain"]
                     if {![isnumber $albgain] || ([isnumber $lastAlbGain] && $lastAlbGain != $albgain)} {
                        set process 1
                        break
                     } else {
                        set lastAlbGain $albgain
                     }
                  }
                  ##
                  # OK to process the files
                  #
                  if {$process} {
                     foreach {f} $files {
                        # Determine if Gains are different
                        set gain [db::getVal $f "Gain"]
                        if {[isnumber $gain] && [isnumber $lastGain] && $lastGain != $gain} {
                           set reset 0
                           break
                        } else {
                           set lastGain $gain
                        }
                     }
                     # If reset first, then do so now
                     if {$reset} {
                        foreach {f} $files {
                           gainControl::remove $f
                        }
                     }
                     if {[catch {averageSome $files} result] } then {
                        waitWindow::close
                        return 0
                     }
                  }
               }
            }
         }
      }
   }
   waitWindow::close
   return 1
}
##
# Take a number of tracks and compute then apply the
# average gain. Used to equalize a complete album
# with the same gain.
#
proc gainControl::averageAlbums {folders} {
   foreach {folder} $folders {
      averageSome [db::filesInFolder $folder]
   }
}
proc gainControl::folderGain {folder} {
   set total 0.0
   set count 0
   set result [list]
   foreach {track}  [db::filesInFolder $folder] {
      set gain [db::getVal $track Gain]
      if {[isnumber $gain]} {
         set total [expr {$total+$gain}]
         incr count
      }
   }
   # need five tracks in the album
   if {$count > 5} {
      return [list $folder [toDb [expr {$total/$count}]]]
   } else {
      return [list $folder 0.0]
   }
}

##
# Take a number of tracks and compute then apply the
# average gain. Used to equalize a complete album
# with the same gain.
#
proc gainControl::averageSome {{fileNames {}}} {
   global TA LB

   if {[llength $fileNames] == 0} {
      set fileNames [LB::nameSelected]
   }
   set total 0
   set done [list]
   # Get tracks with gains
   foreach {file} $fileNames {
      if {[catch {get $file} val]} {
         waitWindow::show "Auto-Leveling [file tail $file]" 1
         set val [gainControl::create $file]
      }
      if {[isnumber $val]} {
         set lookup($file) $val
         set total [expr {$val+$total}]
         lappend done $file
      }
      if {[waitWindow::ifbreak]} {
         waitWindow::close
         return -code break
      }
   }
   ##
   # Compute the average and apply that to each of the selected files
   #
   set tot [llength $done]
   if {$tot > 1} {
      set i 0
      set gain [expr {$total/double($tot)}]
      foreach {file} $done {
         waitWindow::percent [incr i] $tot "Averaging Gain for\n[splitFile $file]" 1
         if {[waitWindow::ifbreak]} {
            break
         }
         db::setVal $file AlbumGain $gain
         # Set the track and album gain for the track
         addTag $file $lookup($file) $gain
      }
   }
   waitWindow::close
   # Update any data that is currently shown
   trackList::updateFiles $fileNames
}
##
# Take a number of tracks and compute then apply the
# average gain. Used to equalize a complete album
# with the same gain.
#
proc gainControl::computeSome {fileNames {recompute "yes"}} {
   global TA LB
   if {[llength $fileNames] == 0} {
      set fileNames [LB::nameSelected]
   }
   set total 0
   set found [llength $fileNames]

   foreach {file} $fileNames {
      incr total
      if {$recompute == "no" || ![gainControl::exists $file]} {
         waitWindow::percent $total $found "Auto-Leveling Gain For\n[splitFile $file]" 1
         if {[waitWindow::ifbreak]} {
            break
         }
         set val [gainControl::create $file]
         addTag $file $val "xxx"
         if {($total%20)==0} {
            db::commit
         }
         if {[waitWindow::ifbreak]} {
            break
         }
      }
   }
   # Update any data that is currently shown
   trackList::updateFiles $fileNames
   waitWindow::close
}
proc gainControl::getkey {fname} {
   # Use key containing folder, prevents duplicates
   return [file join [file tail [file dirname $fname]] [file tail $fname]]
}
##
# Return the first point in the track that contains music
#
proc gainControl::getStart {fname} {
   variable autoTruncate
   set st [db::getVal $fname Start]
   if {$autoTruncate && [string is integer -strict $st] && $st > 0} {
      return $st
   } else {
      return 0
   }
}
##
# Return the last point in the track that contains music
#
proc gainControl::getStop {fname {snd ""}} {
   variable autoTruncate
   set endp [db::getVal $fname Stop]
   if {$autoTruncate && [string is integer -strict $endp] && $endp > 0} {
      return $endp
   } else {
      if {[string length $snd]==0} {
         set snd [soundControl::infoObj $fname]
      }
      return [$snd length]
   }
}

package provide snackMetakit 1.0
