#------------------------------------------------------------------------------
# snackAmp - minPlayer package 1.0 for snackAmp
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/miniPlayer.tcl,v 1.35 2007/02/03 02:09:10 wilkason Exp $
}
#----------------------------------------------------------------------------
# Namespace for miniPlayer
#----------------------------------------------------------------------------
namespace eval miniPlayer {
   variable mpSS ""
   variable splash
   variable Hidden 0
   variable Pinned 0
   variable lastFolderName ""
   variable imageLabel
   variable coverArt
   variable activeMiniPlayer
   variable fg
   variable bg
   variable efg
   variable ebg
   variable lastTrack ""
   upvar #0 snackAmpSettings(miniArt) coverArt
   upvar #0 snackAmpSettings(activeMiniPlayer) activeMiniPlayer
   upvar #0 snackAmpSettings(bg) bg
   upvar #0 snackAmpSettings(fg) fg
   upvar #0 snackAmpSettings(efg) efg
   upvar #0 snackAmpSettings(ebg) ebg
}
#------------------------------------------------------------------------------
# Function   : launch
# Description: Mini-player that pops up from the taskbar
# Author     : Tom Wilkason
# Date       : 2/1/2001
#------------------------------------------------------------------------------
proc miniPlayer::launch {X Y} {
   variable coverArt
   variable imageLabel
   variable activeMiniPlayer 1
   variable fg
   variable bg
   variable efg
   variable ebg
   variable imageSize
   variable lastTrack
   set w .miniPlayer
   if {[winfo exists $w]} {
      if {$::TA(currentMF) ne $lastTrack} {
         miniPlayer::smallCoverArt $::TA(currentMF)
      }
      focus -force $w.tb.tt.title
      wm deiconify $w
      fade::in $w
      return
   }
   ##
   # Create miniPlayer window with no border, we will handle dragging
   #
   toplevel $w -relief raised -borderwidth 3 -class SnackAmp
   wm focusmodel $w active  ;# Does this force a focus on pop-up?
   # Bottom Frame
   # Song Title
   pack [frame $w.tb] -expand false -fill x -side top -pady 2
   pack [frame $w.tb.tt] -expand true -fill x -side left

   label $w.tb.tt.title -textvariable ::TA(playlistEntry) \
      -bg $bg -bd 0 -fg $fg -font songInfo -width 45 \
      -anchor w -relief sunken -cursor fleur
   settings::regColor $w.tt.title -bg bg -fg fg
   pack $w.tb.tt.title -expand true -fill x -side top -padx 4

   # Time Slider
   set trackTime  [soundControl::duration]
   set miniPlayer::mpSS [scale $w.tb.tt.ss -show no \
      -orient horiz \
      -width $::snackAmpSettings(timeSliderHeight) \
      -len 80 \
      -from 0. -to $trackTime \
      -resolution 0.1 \
      -borderwidth 0 \
      -var gui::trackPosition \
      -sliderlength $::snackAmpSettings(timeSliderWidth) \
      -activebackground blue \
      -highlightcolor $bg \
      -troughcolor $bg \
      -command gui::updateTime]

   settings::regColor $miniPlayer::mpSS -troughcolor bg \
      -highlightcolor bg
   pack $w.tb.tt.ss -expand true -fill x -side top -padx 4

   # Close miniPlayer
   button $w.tb.close -command {miniPlayer::close} -image saImages.Close -padx 0 -pady 0 -borderwidth 1
   balloon::button $w.tb.close "Close mini-player"
   pack $w.tb.close -side right -padx 0 -pady 0

   frame $w.frame
   # Volume Slider
   scale $w.frame.volume -show no -orient horizontal \
         -width 10 \
         -sliderlength 20 \
         -len 80 -var soundControl::acGaindB \
         -from $::snackAmpSettings(minDB) -to 0 \
         -activebackground blue -troughcolor $bg \
         -command {PlayGain}
   settings::regColor $w.frame.volume -troughcolor bg

   balloon::define $w.frame.volume "Adjust the volume"


   #balloon::define $miniPlayer::mpSS "Show/Adjust playback location within file"
   bind $miniPlayer::mpSS <ButtonPress-1> {
      set TA(setdrag) 0
   }
   # Show time when scale is dragged
   bind $miniPlayer::mpSS <B1-Motion> {
      miniPlayer::mpTime %W %X %Y
   }
   bind $miniPlayer::mpSS <ButtonRelease-1> {
      set TA(setdrag) 1
      PlayOffset
      destroy %W.balloon
   }
   bind $miniPlayer::mpSS <ButtonPress-2> {set TA(setdrag) 0}
   bind $miniPlayer::mpSS <ButtonRelease-2> {set TA(setdrag) 1; PlayOffset}
   bind $miniPlayer::mpSS <ButtonPress-3> {set TA(setdrag) 0}
   bind $miniPlayer::mpSS <ButtonRelease-3> {set TA(setdrag) 1; PlayOffset}

   pack $w.frame -padx 4 -pady 2 -fill x -expand true
   pack $w.frame.volume -expand true -fill x -side left
   balloon::widgetDefine .miniPlayer.tb.tt.title

   # Since this is override direct, we have to handle drags
   bind  .miniPlayer <Enter> {
      focus -force .miniPlayer.tb.tt.title
      wm deiconify .miniPlayer
   }
   # Pop-up info on the song
   if {$::tcl_platform(platform) != "windows"} {
      bind .miniPlayer <Button-5> [list bumpLevel -1]
      bind .miniPlayer <Button-4> [list bumpLevel 1]
   } else {
      bind .miniPlayer <MouseWheel> {
         set amount [expr {%D/abs(%D)*1}] ;# get +/- 2
         bumpLevel $amount
      }
   }
   bind  .miniPlayer.tb.tt.title <ButtonPress-1> {
      # Save offset from corner to mouse
      set W [winfo toplevel %W]
      set miniPlayer::splash(X) [expr {%X - [winfo x $W]}]
      set miniPlayer::splash(Y) [expr {%Y - [winfo y $W]}]
      # Kill the little balloon
      destroy .miniPlayer.tb.tt.title.balloon
   }
   bind  .miniPlayer.tb.tt.title <B1-Motion> {
      set x [expr {%X - $miniPlayer::splash(X)}]
      set y [expr {%Y - $miniPlayer::splash(Y)}]
      wm geometry [winfo toplevel %W] "+$x+$y"
      set ::snackAmpSettings(positionMiniPlayer) [list $x $y]
      destroy .miniPlayer.tb.tt.title.balloon
   }
   # System Menu on RMB
   if {$::snackAmpSettings(fancyMenu)==0} {
      bind  .miniPlayer.tb.tt.title <ButtonPress-3> {
         destroy .miniPlayer.tb.tt.title.balloon
         tk_popup $::SA.bRmenubar %X %Y
      }
   }
   # Buttons for audio control
   set buttonData {
      image mini_prev   Prev       "Play previous in list"
      image mini_play   PlayResume "Play Selection"
      image mini_pause  Pause      "Pause/Resume playback"
      image mini_stop   Stop       "Stop playing"
      image mini_next   Next       "Play next in list"
   }
   # build each one
   set tf $w.frame
   foreach {bmType File Command Help} $buttonData {
      button  $tf.$File -$bmType "saImages.$File" -command $Command -borderwidth 1
      pack $tf.$File -side left -padx 0 -pady 0
      balloon::button $tf.$File $Help
   }

   # Rating
   set pm $tf.rating.popup
   pack [menubutton $tf.rating -relief flat -image saImages.iconthumbsup -menu $pm]\
    -side left -padx 0 -pady 0
   set m [menu $pm \
   -postcommand [list gui::ratingPopup $pm]]
   balloon::button $tf.rating "Adjust rating of the playing track, the more stars the better."

   # Support hiding of main window
   pack [checkbutton $tf.hide \
      -selectimage saImages.arrowhide -image saImages.arrowshow -indicatoron 0 -bd 0 \
      -selectcolor $bg -variable miniPlayer::Hidden -command {
      switch -- [wm state $::SA] {
         normal {wm withdraw $::SA; set miniPlayer::Hidden 1}
         withdrawn {wm deiconify $::SA; set miniPlayer::Hidden 0}
         iconic {wm withdraw $::SA; set miniPlayer::Hidden 1}
         default {wm deiconify $::SA; set miniPlayer::Hidden 0}
      }
   }] -side left
   balloon::define $tf.hide "Hide/Show the main snackAmp window."
   # Icon to pin it if can pin
   if {$ontop::enabled} {
      pack [checkbutton $tf.pin -image saImages.unpinit -selectimage saImages.pinit -indicatoron 0 -bd 0 \
         -selectcolor $bg \
         -variable miniPlayer::Pinned \
         -command {ontop::control .miniPlayer $miniPlayer::Pinned}] -side left
      balloon::define $tf.pin "Press to keep the mini player on top."
      #----------------------------------------------------------------------------
      # Transparency button
      #----------------------------------------------------------------------------
      fade::setDefTrans .miniPlayer
      pack [menubutton $tf.trb -image saImages.transparent -menu $tf.trb.tr \
         -relief flat -bd 0] -side left
      balloon::button $tf.trb "Select transparency level for mini player"
      menu $tf.trb.tr
      for {set i 0.3} {$i <= 1.01} {set i [expr {$i+0.05}]} {
         if {$i>0.951} {
            set i 1.0
            set lab "Normal"
         } else {
            set lab "[format %0.0f [expr {$i*100}]]%"
         }
         $tf.trb.tr add radiobutton -label $lab -value $i\
            -variable snackAmpSettings(.miniPlayer,transparency) \
            -command [list fade::setTrans .miniPlayer $i]
      }
   }
   #----------------------------------------------------------------------------
   # Search frame
   #----------------------------------------------------------------------------
   set sf [frame $w.search]

   set sf0 [label $sf.lab -text "Search:"]
   set sf1 [entryHistory::create $sf.search ::TA(filter) ::snackAmpSettings(History)]
   set sf1h [entryHistory::hull $sf.search]
   $sf1h config -background $ebg -foreground $efg
   balloon::define $sf1h "Enter text to search for, use buttons or Return for playlist search and Control-Return for Media Library search"

   set sf2 [button $sf.b1 -image saImages.findImage -command {
      if {[wm state $::SA]!="normal"} {
         wm deiconify $::SA
         set miniPlayer::Hidden 0
      }
      cframe.tryOpen $::TA(CF)
      findNextSong
   }]
   balloon::button $sf.b1 "Search Playlist"

   set sf3 [button $sf.b2 -image saImages.filterOff -command {
      filtering::searchFilter
   }]
   balloon::button $sf.b2 "Search Media Library"

   bind $sf1h <Return> [list $sf.b1 invoke]
   bind $sf1h <Control-Return> [list $sf.b2 invoke]

   grid $sf0 $sf1 $sf2 $sf3 -sticky ew
   grid columnconfigure $sf 1 -weight 1
   pack $sf -side top -fill x -expand true
   #----------------------------------------------------------------------------
   # Cover art frame
   #----------------------------------------------------------------------------
   if {$coverArt} {
      set CF [cframe $w.ca "Cover Art"]
      # Cover art above window
      set imageLabel [label $CF.cover -bd 2 -relief groove]
      pack $imageLabel -expand true -fill both
      pack $w.ca -side bottom -expand true -fill both
   }
   ##
   # Figure out where to put the window, TL corner is window 0,0
   #
   # Hide it and take away window decoration
   wm withdraw $w
   wm overrideredirect $w 1
   focus -force $w
   # with it hidden, let the layout take place so we know its geometry
   update idletasks

   # Get miniplayer size  and position at bottom right of usable area
   set rX [winfo reqwidth $w]
   # Save the required image size and keep it constant
   if {![info exists imageSize]} {
      set imageSize $rX
   }
   if {$coverArt} {
      miniPlayer::smallCoverArt [soundControl::fileName] $imageSize
      update idletasks
   }
   set rY [winfo reqheight $w]
   # Don't place over taskbar
   # Locate the player in the right hand bottom corner then pop it up.
   if {($X+$Y)==0} {
      set x [expr {$::TA(screen_right)-$rX}]
      set y [expr {$::TA(screen_bottom)-$rY}]
   } else {
      set x $X
      set y $Y
   }
   wm geometry $w "+$x+$y"
   set ::snackAmpSettings(positionMiniPlayer) [list $x $y]
   fade::off $w
   wm deiconify $w
   if {$::tcl_platform(os) eq "Linux"} {
      # yuck: linux bug workaround
      wm geometry $w "+$x+$y"
   }
   fade::in $w
}
#----------------------------------------------------------------------------
# Cover art for the mini player
#----------------------------------------------------------------------------
proc miniPlayer::smallCoverArt {track {size 0}} {
   Trace
   variable imageLabel
   variable coverArt
   variable lastFolderName
   variable imageSize
   variable lastTrack
   if {$coverArt} {
      set sourceImage [albumart::setCurrentImage $track]
      set folder [file dirname $track]
      if {$folder ne $lastFolderName} {
         if {$sourceImage ne "" && [albumart::valid]} {
            image create photo "saImages.miniCoverArt" ;# this replaces old version
            if {$size==0} {
               set height $imageSize
            } else {
               set height $size
            }
            foreach {height width} \
               [albumart::getCoverDims $sourceImage $height] {break}
            if {$width>0 && $height>0} {
               set lastTrack $track

               tkImageTools::resize $sourceImage "saImages.miniCoverArt" $width $height
               $imageLabel configure -image "saImages.miniCoverArt"
            } else  {
               $imageLabel configure -image ""
            }
         } else {
            $imageLabel configure -image ""
         }
         set lastFolderName $folder
      }
   }
}
#------------------------------------------------------------------------------
# Function   : miniPlayer::mpTime
# Description: Show time on miniplayer when dragging mouse
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc miniPlayer::mpTime {W X Y} {
   set time [HMSfrac $gui::trackPosition]
   balloon::show $W $X $Y "Time: $time"
}
#------------------------------------------------------------------------------
# Function   : miniPlayer::toggle
# Description: Toggle the mini player visibility
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc miniPlayer::toggle {{X 0} {Y 0}} {
   variable activeMiniPlayer
   if {$activeMiniPlayer == 0} {
      miniPlayer::close
   } else {
      miniPlayer::launch $X $Y
   }
}
#----------------------------------------------------------------------------
# Version that doesn't rely on the activeMiniPlayer bit being set first
#----------------------------------------------------------------------------
proc miniPlayer::toggler {{X 0} {Y 0}} {
   variable activeMiniPlayer
   if {$activeMiniPlayer == 1} {
      miniPlayer::close
      set activeMiniPlayer 0
   } else {
      miniPlayer::launch $X $Y
      set activeMiniPlayer 1
   }
}

#------------------------------------------------------------------------------
# Function   : miniPlayer::close
# Description: Close the miniplayer and make sure main window is not also
#              withdrawn on non-windows systems (i.e. no task bar)
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc miniPlayer::close {args} {
   variable activeMiniPlayer 0

   set w .miniPlayer
   # todo: need blocking fadeout (have code)
   fade::out $w
   # If no taskbar icon, then make sure both windows
   # not closed.
   if {$::TA(iconsVisible)==0} {
      if {[wm state $::SA] == "withdrawn"} {
         wm deiconify $::SA
         set miniPlayer::Hidden 0
      }
   }
}
package provide miniPlayer 1.0
