#------------------------------------------------------------------------------
# snackAmp - playList package 1.0 for snackAmp
# provides access to the main playlist
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2003 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/playList.tcl,v 1.27 2006/04/29 15:05:51 wilkason Exp $
}
#----------------------------------------------------------------------------
# Kludge to prevent a M1-Motion from extending the scroll area in a scrolled
# listbox. This is from 8.4 library and overwrites existing code!
#----------------------------------------------------------------------------
proc ::tk::ListboxAutoScan {w} {
   variable ::tk::Priv
   after cancel Priv(afterId)
   if {![winfo exists $w]} return
   if {$w eq $::LB} {
      set x $Priv(x)
      set y $Priv(y)
      if {$y >= [winfo height $w]} {
         $w yview scroll 1 units
         LB::Drag $y   ;# Just updates the old position after scroll
      } elseif {$y < 0} {
         $w yview scroll -1 units
         LB::Drag $y   ;# Just updates the old position after scroll
      } elseif {$x >= [winfo width $w]} {
         #$w xview scroll 2 units
      } elseif {$x < 0} {
         #$w xview scroll -2 units
      } else {
         return
      }
      #ListboxMotion $w [$w index @$x,$y]
      set Priv(afterId) [after 50 [list tk::ListboxAutoScan $w]]
   }
}

#------------------------------------------------------------------------------
# LB Namespace wrapper for the ListBox
#------------------------------------------------------------------------------
namespace eval LB {
   global snackAmpSettings
   variable indexPlaying -1
   variable webInsert -1
   variable oldPosition 0
   variable fields [list Duration Artist Title]
   variable LBdata
   variable Kdata
   variable cvTi 0
   variable cvBi 0
   if {![info exists Kdata]} {
      set Kdata [list]
   }
}
proc LB::updateFields {} {
   global snackAmpSettings
   variable fields
   ##
   # Listbox fields
   #
   set LB::fields [list]
   foreach num {1 2 3 4} {
      if {[info exists snackAmpSettings(pl$num)] && $snackAmpSettings(pl$num)!="None"} {
         lappend LB::fields $snackAmpSettings(pl$num)
      }
   }

}
#----------------------------------------------------------------------------
# Create the listbox
#----------------------------------------------------------------------------
proc LB::create {W} {
   variable data
   updateFields
   set lb [Scrolled_Listbox $W \
   -selectmode       "extended" \
   -listvariable     LB::LBdata                         \
   -exportselection  no                                 \
   -activestyle      dotbox                             \
   -height           $::snackAmpSettings(playListLines) \
   -foreground       $::snackAmpSettings(fg)            \
   -selectforeground $::snackAmpSettings(fg)            \
   -selectbackground $::snackAmpSettings(highlight)     \
   -background       $::snackAmpSettings(bg)]

   settings::regColor $lb  \
      -foreground       ::snackAmpSettings(fg) \
      -selectforeground ::snackAmpSettings(fg) \
      -selectbackground ::snackAmpSettings(highlight) \
      -background       ::snackAmpSettings(bg)

   # Background validate the playlist tracks
      #after 5000 [list LB::catalogBackground 0 end]

   return $lb
}
proc LB::body {} {
   global LB
   return $LB
}
proc LB::Y {y} {
   global LB
   return $y
}

#------------------------------------------------------------------------------
# Grab the row attributes
#------------------------------------------------------------------------------
proc LB::getRowAttribs {pos} {
   global LB
   foreach att {-foreground -background -selectforeground -selectbackground} {
      lappend result [$LB itemcget $pos $att]
   }
   if {[lindex $result 0] ne ""} {
         $LB itemconfigure $pos \
            -background "" \
            -foreground "" \
            -selectbackground "" \
            -selectforeground ""
   }
   return $result
}
#------------------------------------------------------------------------------
# Set the row attributes from a prior grab
#------------------------------------------------------------------------------
proc LB::setRowAttribs {pos att} {
   global LB
   foreach {fg bg sfg sbg} $att {break}
   if {$fg ne ""} {
      $LB itemconfigure $pos \
         -background $bg \
         -foreground $fg \
         -selectbackground $sbg \
         -selectforeground $sfg
   }
}

#------------------------------------------------------------------------------
# Function   : LB::moveByIndex
# Description: Move an entry restore any highlighting it may have
# Author     : Tom Wilkason
# Date       : 6/3/2002
#------------------------------------------------------------------------------
proc LB::moveByIndex {to_pos from_pos} {
   variable indexPlaying
   variable Kdata
   variable LBdata
   global LB
   set end [$LB index end]
   if {$to_pos == $from_pos || $to_pos < 0 || $to_pos >= $end} {
      return 0
   }
   set tatt [getRowAttribs $to_pos]
   set fatt [getRowAttribs $from_pos]
   # Swap Index data
   set fitem [lindex $Kdata $to_pos]
   lset Kdata $to_pos [lindex $Kdata $from_pos]
   lset Kdata $from_pos $fitem

   # Swap listbox data
   set item [lindex $LBdata $to_pos]
   lset LBdata $to_pos [lindex $LBdata $from_pos]
   lset LBdata $from_pos $item

   # Move selection to new pos
   $LB selection clear $from_pos
   $LB selection set $to_pos
   $LB activate $to_pos

   # Swap the attributes
   setRowAttribs $from_pos $tatt
   setRowAttribs $to_pos $fatt
   # Move playing index if needed
   if {$indexPlaying == $from_pos} {
      set indexPlaying $to_pos
   }
   fixWebInsert
}
##
# Make sure the queue insert position is after the playing index
#
proc LB::fixWebInsert {args} {
   variable indexPlaying
   variable webInsert
   if {$indexPlaying >= $webInsert} {
      set webInsert [expr {$indexPlaying+1}]
   }
}
##
# Reset the queue insert position to just after the playing position
#
proc LB::resetWebInsert {args} {
   variable indexPlaying
   variable webInsert
   set webInsert [expr {$indexPlaying+1}]
}
#------------------------------------------------------------------------------
# Make sure some file is viewable
#------------------------------------------------------------------------------
proc LB::seeFile {File} {
   global LB
   if {$File eq ""} {
      return 0
   }
   set index [LB::idxFile $File]
   if {$index > -1} {
      $LB see $index
      return 1
   } else {
      return 0
   }
}
#------------------------------------------------------------------------------
# Make sure the playing file is viewable
#------------------------------------------------------------------------------
proc LB::seePlaying {args} {
   global LB
   variable indexPlaying
   if {$indexPlaying > -1} {
      $LB see $indexPlaying
      return 1
   } else {
      $LB see [idxActive]
      return 0
   }
}
#----------------------------------------------------------------------------
# Return the playlist values of the indicies "start ?end?"
#----------------------------------------------------------------------------
proc LB::keyFromIndex {args} {
   global LB
   return [eval $LB get $args]
}
#----------------------------------------------------------------------------
# Return the index at the index specified
#----------------------------------------------------------------------------
proc LB::index {index} {
   global LB
   return [$LB index $index]
}
#----------------------------------------------------------------------------
# Active the entry specified by index
#----------------------------------------------------------------------------
proc LB::activate {index} {
   global LB
   if {[catch {$LB activate $index} result] } then {
      debug $result
   } else {
      $LB see $index
   }
}
#----------------------------------------------------------------------------
# Select the entries specified by the parameters
#----------------------------------------------------------------------------
proc LB::setSelection {args} {
   global LB
   eval $LB selection set $args
}
#------------------------------------------------------------------------------
# Function   : sortSelection
# Description: Sort the selection (only used for playlist)
# Author     : Tom Wilkason
# Date       : 12/4/2001
#------------------------------------------------------------------------------
proc LB::sortSelection {{_listBox ::LB}} {
   variable Kdata
   upvar $_listBox lb
   global TA
   set selection [lsort -decreasing -integer [$lb curselection]]
   set current [$lb index anchor]
   if {[llength $selection] > 1} {
      ##
      # Cut the data out and sort it
      #
      foreach {ind} $selection {
         lappend data [list [$lb get $ind] [lindex $Kdata $ind]]
      }
      $lb delete [lindex $selection end] [lindex $selection 0]
      ldelete Kdata [lindex $selection end] [lindex $selection 0]
      set data [lsort -index 1 -decreasing -dictionary $data]
      ##
      # Now paste it back in
      #
      foreach {pair} $data {
         foreach {key file} $pair {
            $lb insert $current $key
            lipinsert Kdata $current $file

         }
      }
      $lb selection anchor $current
      $lb selection set [lindex $selection end] [lindex $selection 0]
   }
}
#------------------------------------------------------------------------------
# Function   : shuffleSelection
# Description: Shuffle the selection (only used for playlist)
# Author     : Tom Wilkason
# Date       : 12/4/2001
#------------------------------------------------------------------------------
proc LB::shuffleSelection {{_listBox ::LB}} {
   variable Kdata
   upvar $_listBox lb
   set selection [lsort -decreasing -integer [$lb curselection]]
   set current [$lb index anchor]
   if {[llength $selection] > 1} {
      ##
      # Cut the data out and sort it
      #
      foreach {ind} $selection {
         lappend data [list [$lb get $ind] [lindex $Kdata $ind]]
      }
      $lb delete [lindex $selection end] [lindex $selection 0]
      ldelete Kdata [lindex $selection end] [lindex $selection 0]
      set data [shuffle $data]
      ##
      # Now paste it back in
      #
      foreach {pair} $data {
         foreach {key file} $pair {
            $lb insert $current $key
            lipinsert Kdata $current $file
         }
      }
      $lb selection anchor $current
      $lb selection set [lindex $selection end] [lindex $selection 0]
   }
}
#------------------------------------------------------------------------------
# Function   : LB::updateFiles
# Description: Update the playlist data to reflect any new changes
#              to a file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc LB::updateFiles {fileList} {
   variable Kdata
   variable LBdata
   # Change just the lookup, the keys stay the same
   # Should work for all key types, allow old
   # tags to remain at the end
   set idx -1
   foreach value $Kdata {
      incr idx
      foreach {oldFile} $fileList {
         if {$value eq $oldFile} {
            lset LBdata $idx [LB::nameToKey $oldFile]
         }
      }
   }
}
#------------------------------------------------------------------------------
# Function   : LB::changePlaylistFiles
# Description: Change all of the instances of a file lookup in the main
#              playlist with another. Called when renaming a file.
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc LB::changePlaylistFiles {fileList} {
   variable Kdata
   variable LBdata
   # Change just the lookup, the keys stay the same
   # Should work for all key types, allow old
   # tags to remain at the end
   set idx -1
   foreach value $Kdata {
      incr idx
      foreach {oldFile newFile} $fileList {
         if {$value eq $oldFile} {
            lset Kdata $idx $newFile
            lset LBdata $idx [LB::nameToKey $newFile]
         }
      }
   }
}
#----------------------------------------------------------------------------
# Change a single entry
#----------------------------------------------------------------------------
proc LB::changePlaylistFile {from to} {
   variable Kdata
   variable LBdata
   set idx [lsearch -exact $Kdata $from]
   if {$idx > -1} {
      lset Kdata $idx $to
      lset LBdata $idx [LB::nameToKey $to]
   }
}

#------------------------------------------------------------------------------
# Set the rating for the selected tracks to some value
#------------------------------------------------------------------------------
proc LB::setSelectedRating {value} {
   set files [LB::nameSelected]
   foreach {file} $files {
      db::setVal $file Rating $value
   }
   trackList::updateFiles $files
}
#------------------------------------------------------------------------------
# After the playlist has been altered, try and restore
# the playing song with highlighting.
# This assumes the current highlighted file
# was removed first.
# TFW: If duplicates exist in the playlist
#      this will find only the first one.
#------------------------------------------------------------------------------
proc LB::tryRestorePlaying {args} {
   global TA snackAmpSettings
   variable indexPlaying
   if {![LB::seeFile $TA(currentMF)]} {
      LB::activate 0
   } else {
      if {[soundControl::playState]!="Stopped"} {
         LB::inverseVideoFile $TA(currentMF) 1
         if {$snackAmpSettings(moveToTop)} {
            LB::moveByIndex 0 [LB::idxPlaying]
            LB::seePlaying
            LB::resetWebInsert
         }
      } else {
         LB::activate 0 ;# will be last item after insert
      }
   }
}
#------------------------------------------------------------------------------
# Select/Drag, set on mouse click
#------------------------------------------------------------------------------
proc LB::Select {y} {
   global TA LB
   variable oldPosition
   variable Kdata
   destroy [body].balloon
   set oldPosition [$LB nearest $y]
   set Value [$LB get $oldPosition]
   set File [lindex $Kdata $oldPosition]
   if {$Value ne ""} {
      set TA(entryName) $File
      if {$::snackAmpSettings(updateAPL)} {
         catch {APL::SelectFile $File}
      }
   }
}
#------------------------------------------------------------------------------
# Drag an entry up/down
# TODO: Only works with offset of +/1
#       because swap is done rather than delete insert because swap is much faster
#------------------------------------------------------------------------------
proc LB::Drag {y} {
   global TA LB
   variable oldPosition
   set new [$LB nearest $y]
   if {$new == -1 || $new==$oldPosition} {
      return
   } elseif {$new < $oldPosition} {
      foreach moveIndex [lsort -integer [LB::idxSelected]] {
         set above [expr {$moveIndex - 1}]
         LB::moveByIndex $above $moveIndex
      }
   } else {
      foreach moveIndex [lsort -integer -decreasing [LB::idxSelected]] {
         set below [expr {$moveIndex + 1}]
         LB::moveByIndex $below $moveIndex
      }
   }
   set oldPosition $new
   resetWebInsert  ;# just in case we changed playing position
}
#------------------------------------------------------------------------------
# Function   : LB::highlightIndex
# Description: Highlight entry $i in the playlist
# Author     : Tom Wilkason
# Date       : 8/8/2003
#------------------------------------------------------------------------------
proc LB::highlightIndex {i {inverse 0}} {
   global LB
   $LB selection clear 0 end
   set i [expr {$i < 0 ? 0 : $i}]
   set i [expr {$i >= [LB::idxEnd] ? [expr {[LB::idxEnd] - 1}] : $i}]
   LB::setSelection $i
   $LB activate $i
   $LB see $i
   $LB selection anchor $i
   if {$inverse} {
      LB::inverseVideoIndex $i
   }
}
#----------------------------------------------------------------------------
# Mark an entry as invalid (e.g. if the file doesn't exist)
#----------------------------------------------------------------------------
proc LB::invalidateIndex {i} {
   global LB
   if {$i >= 0 && $i < [$LB size]} {
      $LB itemconfigure $i -background darkgray
   }
}
#------------------------------------------------------------------------------
# Inverse video a file
# note Can get rid of all the LB::inverseVideoFile xx 0 calls, no longer needed
#------------------------------------------------------------------------------
proc LB::inverseVideoFile {File {On 1}} {
   Trace
   variable indexPlaying
   global LB
   if {$On} {
      set index [LB::idxFile $File]
      if {$index >= 0} {
         LB::inverseVideoIndex $index
      }
   } else {
      LB::resetIinverseVideo
   }
}
#------------------------------------------------------------------------------
# Inverse video a selection
# TBD: Can remove the off option
#------------------------------------------------------------------------------
proc LB::inverseVideoIndex {index {On 1}} {
   #Trace 1
   variable indexPlaying
   LB::resetIinverseVideo
   global LB TA
   if {$index > -1} {
      set bg  [$LB cget -background]
      set fg  [$LB cget -foreground]
      set sbg [$LB cget -selectbackground]
      set sfg [$LB cget -selectforeground]
      ##
      # Color or uncolor the file
      #
      if {$On} {
         $LB itemconfigure $index -background $fg -foreground $bg \
            -selectforeground $sbg -selectbackground $sfg
         set indexPlaying $index
      } else {
         $LB itemconfigure $index -background "" -foreground "" \
            -selectforeground "" -selectbackground ""
      }
   }
   LB::activate $index
   fixWebInsert
   if {[APLExists] && $::snackAmpSettings(syncAPL)} {
      set TA(entryName) [LB::nameFromIdx $index]
      APL::SelectFile $TA(entryName)
   }
}
##
# Reset all of the inverse video (very fast)
# Better than trying to keep track of what is inversed
proc LB::resetIinverseVideo {args} {
   global LB
   set end [$LB size]
   for {set idx 0} {$idx < $end} {incr idx} {
      if {[$LB itemcget $idx -selectforeground] ne ""} {
         $LB itemconfigure $idx -background "" -foreground "" \
            -selectbackground "" -selectforeground ""
      }
   }
}
#----------------------------------------------------------------------------
# Returan all names in the playlist, either the short or full names
#----------------------------------------------------------------------------
proc LB::keyOrNamesAll {{long 0}} {
   global LB
   variable LBdata
   variable Kdata
   set files [list]
   if {$long} {
      return $Kdata
   } else {
      return $LBdata
   }
}
#----------------------------------------------------------------------------
# Return pairs of key|name for saving cache
#----------------------------------------------------------------------------
proc LB::keyAndNames {args} {
   global LB
   variable LBdata
   variable Kdata
   set files [list]
   foreach {item} $LBdata {file} $Kdata {
      lappend files $item $file
   }
   return $files
}

#----------------------------------------------------------------------------
# Return the ending index of the listbox
#----------------------------------------------------------------------------
proc LB::idxEnd {args} {
   global LB
   return [$LB size]
}
#------------------------------------------------------------------------------
# Locate a file in the playlist from the anchor position
# TODO: This can be ambiguous if multiple copies if the same
# file are in the playlist
#------------------------------------------------------------------------------
proc LB::idxFile {File} {
   Trace
   variable LBdata
   variable Kdata
   global LB
   if {$File eq ""} {
      return -1
   }
   set anchor [$LB index active]
   set index [lsearch -exact -start $anchor $Kdata $File]
   if {$index == -1} {
      set index [lsearch -exact [lrange $Kdata 0 $anchor] $File]
      if {$index == -1} {
         return -1
      }
   }
   return $index
}
#----------------------------------------------------------------------------
# Return the indicies of the selected tracks in the playlist
#----------------------------------------------------------------------------
proc LB::idxSelected {args} {
   global LB
   return [$LB curselection]
}
#----------------------------------------------------------------------------
# Return the index of the current queue point
#----------------------------------------------------------------------------
proc LB::idxQueued {args} {
   variable webInsert
   return $webInsert

}
#----------------------------------------------------------------------------
# Return an array of all the indices in the playlist
#----------------------------------------------------------------------------
proc LB::idxAll {args} {
   global LB
   set result [list]
   set end [$LB size]
   for {set i 0} {$i < $end } {incr i} {
      lappend result $i
   }
   return $result
}
#----------------------------------------------------------------------------
# Return the index of the anchor within the current selection
#----------------------------------------------------------------------------
proc LB::idxAnchor {args} {
   global LB
   return [$LB index anchor]
}
#----------------------------------------------------------------------------
# Return the index of the currently active track
#----------------------------------------------------------------------------
proc LB::idxActive {args} {
   global LB
   return [$LB index active]
}
#----------------------------------------------------------------------------
# Return the index of the currently playing track
#----------------------------------------------------------------------------
proc LB::idxPlaying {args} {
   variable indexPlaying
   return $indexPlaying
}
#------------------------------------------------------------------------------
# Function   : LB::nameToKey
# Description: Return a song name with the appropriate folders prepended
#              Remove the file extension if needed
# Author     : Tom Wilkason
# Date       : 1/3/2001
#------------------------------------------------------------------------------
proc LB::nameToKey {File} {
   Trace
   variable fields
   global snackAmpSettings TA
   set name [list]
   set result [db::getValues $File $fields]
   # If file not found
   if {[llength $result]==0} {
      return [nameToShort $File]
   }
   ##
   # Duration is always present
   # (below line is for real)
   foreach $fields $result {break}
   set row ""
   ##
   # var fields
   #
   set f 0
   foreach {field} $fields {
      switch -- $field {
         None {continue}
         Duration {
            if {$Duration>0} {
               append row "[HMSfrac $Duration]   "
            }
         }
         ShortName {
            append row "[nameToShort $File]$TA(spcar)"
            incr f
         }
         default {
            set val [string trim [set $field]]
            if {$val ne ""} {
               append row "$val$TA(spcar)"
               incr f
            }
         }
      }
   }
   # If no tag fields, return from file name
   if {$f} {
      set row [string trimright $row $TA(spcar)]
   } else {
      append row [nameToShort $File]
   }
   return $row
}

#----------------------------------------------------------------------------
# Return the full name of an index
#----------------------------------------------------------------------------
proc LB::nameFromIdx {Index {end ""}} {
   global LB
   variable Kdata
   if {$end eq ""} {
      return [lindex $Kdata $Index]
   }  else {
      return [lrange $Kdata $Index $end]
   }
}

#----------------------------------------------------------------------------
# Return the long file name from the track entry
#----------------------------------------------------------------------------
proc LB::nameFromKey {hash} {
   # Just return the file name passed in
   return $res
}
#----------------------------------------------------------------------------
# Return the full file name of the the currently active index
#----------------------------------------------------------------------------
proc LB::nameActive {args} {
   global LB
   variable Kdata
   if {[catch {
      set active [$LB index active]
      set res [lindex $Kdata $active]
   }]} then {
      set res ""
   }
   return $res
}
#----------------------------------------------------------------------------
# Return the full file names of the selected items in the playlist,
#----------------------------------------------------------------------------
proc LB::nameSelected {args} {
   global LB
   variable Kdata
   set fullNames [list]
   foreach {Index} [LB::idxSelected] {
      lappend fullNames [lindex $Kdata $Index]
   }
   return $fullNames
}
#----------------------------------------------------------------------------
# Remove all of the entries in the playlist
#----------------------------------------------------------------------------
proc LB::deleteAll {args} {
   global LB
   variable indexPlaying
   variable webInsert
   variable Kdata

   $LB delete 0 end
   ldelete Kdata 0 end
   set indexPlaying -1
   set webInsert -1
   gui::updateFileDisplay
}
#----------------------------------------------------------------------------
# Remove an entry at some position
#----------------------------------------------------------------------------
proc LB::deleteIndex {position} {
   global LB
   variable indexPlaying
   variable webInsert
   variable Kdata
   $LB delete $position
   ldelete Kdata $position

   if {[string is double $position]} {
      if {$indexPlaying >= $position} {
         incr indexPlaying -1
      }
      if {$webInsert >= $position} {
         incr webInsert -1
      }
   }
   return $position
}
#----------------------------------------------------------------------------
# Replace a track entry at position i
# unused
#----------------------------------------------------------------------------
proc LB::replaceKeyAtIndex {i newKey} {
   ;# Fast using lset
   global LB
   variable LBdata
   lset LBdata $i $newKey
}
#------------------------------------------------------------------------------
# Function   : LB::removeFile
# Description: Remove all instances of an entry in the playlist
# Author     : Tom Wilkason
# Date       : 8/7/2003
#------------------------------------------------------------------------------
proc LB::removeFile {File} {
   variable Kdata
   ##
   # Since the playlist can have multiple entries for the same file, we have
   # to take some care and remove each of them from the bottom up.
   set lnFile [LB::nameToKey $File]
   set end [expr {[LB::idxEnd]-1}]
   for {set i $end} {$i >= 0} {incr i -1} {
      set Entry [LB::keyFromIndex $i]
      if {[string equal $Entry $lnFile]}  {
         LB::deleteIndex $i
      }
   }
}
#------------------------------------------------------------------------------
# Allow a user to insert a file for playing, uses an internal
# pointer to keep track of where the insert should take place.
# It is placed after the last user insert file, once those
# files are played, it is placed just after the currently selected
# file.
#------------------------------------------------------------------------------
proc LB::userInsert {{track ""}} {
   # webInsert should always be at least one more than indexPlaying
   # If not, then make it so.
   variable indexPlaying
   variable webInsert
   if {$track!=""} {
      if {$webInsert > $indexPlaying} {
         # Append to web insert (webInsert is incremented automatically)
         LB::InsertFile $webInsert $track
      } else {
         # Insert just after playing position
         LB::InsertFile [expr {$indexPlaying+1}]  $track
      }
   }
   return $webInsert
}
#----------------------------------------------------------------------------
# Insert a list of items in the playlist following the current user play
# queue position.
#----------------------------------------------------------------------------
proc LB::userInsertList {items} {
   Trace
   # webInsert should always be at least one more than indexPlaying
   # If not, then make it so.
   flashMouse
   variable indexPlaying
   variable webInsert
   if {$webInsert > $indexPlaying} {
      # Append to web insert (webInsert is incremented automatically)
      set pos $webInsert
   } else {
      # Insert just after playing position
      set pos [expr {$indexPlaying+1}]
   }
   LB::insertList $pos $items
   gui::updateFileDisplay
   return [llength $items]
}
#----------------------------------------------------------------------------
# Insert a list of items in the playlist following the playing track
#----------------------------------------------------------------------------
proc LB::nextInsertList {items} {
   Trace
   # webInsert should always be at least one more than indexPlaying
   # If not, then make it so.
   flashMouse
   variable indexPlaying
   variable webInsert
   set pos [expr {$indexPlaying+1}]
   set newend [expr {$pos+[llength $items]}]
   if {$webInsert < $newend} {
      set webInsert $newend
   }
   LB::insertList $pos $items
   gui::updateFileDisplay
   return [llength $items]
}
#------------------------------------------------------------------------------
# Insert a song at current position of the playlist and then play it
#------------------------------------------------------------------------------
proc LB::InsertPlay {file} {
   Trace
   global TA LB
   variable indexPlaying
   if {$file != {}} {
      flashMouse
      ##
      # If file exists in playlist, jump to it instead
      # of inserting it
      #
      set curPost [LB::idxFile $file]
      if {$curPost < 0} {
         set curPost [LB::idxActive]
         incr curPost
         # Insert into playlist adjusting the que position
         LB::insertList $curPost [list $file]
         # If a new incoming file, make sure it is also in the master list
         db::cacheFileName $file
      } else {
         # This is a hack, adjust the playing index then make sure the queue is adjusted
         set indexPlaying $curPost
         LB::resetWebInsert
      }
      LB::activate $curPost
      Play active
   }
}
#------------------------------------------------------------------------------
# Insert a file into the playlist
#------------------------------------------------------------------------------
proc LB::InsertFile {curPost file} {
   if {$file != {}} {
      # Make sure it is in the master list
      LB::insertList $curPost [list $file]
   }
}

#------------------------------------------------------------------------------
# Append a file to the bottom of the play list
#------------------------------------------------------------------------------
proc LB::AppendFile {file} {
   if {$file ne ""} {
      LB::insertList end [list $file]
      # If a new incoming file, make sure it is also in the master list
      db::cacheFileName $file
   }
}
#------------------------------------------------------------------------------
# Append a file to the bottom of the play list
#------------------------------------------------------------------------------
proc LB::AppendList {files} {
   flashMouse
   set apList [list]
   foreach {item} $files {
      if {$item ne ""} {
         db::cacheFileName $item
         lappend apList $item
      }
   }
   if {[llength $apList] > 0} {
      LB::insertList end $apList
   }
}
#------------------------------------------------------------------------------
# Insert a list of entries into the listbox
#------------------------------------------------------------------------------
proc LB::insertList {position files} {
   global LB
   variable indexPlaying
   variable LBdata
   variable webInsert
   variable Kdata
   set items [list]
   # Translate to short name and make sure
   # they are in the lookup table
   if {$position eq "end"} {
      set position [$LB size]
   }
   foreach {item} $files {
      set key [LB::nameToKey $item]
      lappend items $key
   }
   # This is a faster insert
   if {[llength $LBdata]} {
      eval $LB insert $position $items
      set pos $position
      foreach {f} $files {
         lipinsert Kdata $pos $f
         incr pos
      }
   }  else {
      set LBdata $items
      set Kdata $files
   }
   # Adjust play position here if needed
   if {[string is double $position]} {
      if {$indexPlaying >= $position} {
         incr indexPlaying [llength $items]
      }
      if {$webInsert >= $position} {
         incr webInsert [llength $items]
      }
   }
   gui::updateFileDisplay
}
#----------------------------------------------------------------------------
# Very fast restore of playlist, keys and files
# called from restore playlist cache
#----------------------------------------------------------------------------
proc LB::replacePairList {files} {
   global LB
   variable indexPlaying
   variable webInsert
   variable LBdata
   variable Kdata
   set items [list]
   # Translate to short name and make sure
   # they are in the lookup table
   foreach {key file} $files {
      lappend items $key
      lappend Kdata $file
   }
   # This is a faster insert
   set LBdata $items
   gui::updateFileDisplay
}
#----------------------------------------------------------------------------
# Compute the total time for the display
#----------------------------------------------------------------------------
proc LB::totTime {args} {
   variable LBdata
   variable Kdata
   set totalTime 0
   set idx 0
   set catOpt $::snackAmpSettings(catPlaylist)
   foreach row $LBdata file $Kdata {
      set ret [scan $row "%d:%d:%d" h m s]
      switch -- $ret {
         3 {
            set totalTime [expr {$s+$m*60+$h*3600+$totalTime}]
         }
         2 {
            set totalTime [expr {$m+$h*60+$totalTime}]
         }
         default {
# TODO: Maybe only do it if index is row is visible and have a
#different event to update when the scroll changes
            if {[db::lookup $file Data $catOpt]} {
               set totalTime [expr {$totalTime+$Data(Duration)}]
               # Update the display properly
               lset LBdata $idx [LB::nameToKey $file]
               update
            }
         }
      }
      incr idx
   }
   set totalTime [expr {round($totalTime)}]
   set days [expr {$totalTime/86400}]
   if {$days == 0} {
      return "([HMSfrac $totalTime])"
   } else {
      set totalTime [expr {$totalTime%86400}]
      return "($days days [HMSfrac $totalTime])"
   }
}


#------------------------------------------------------------------------------
# Make sure all of the tracks in the viewed portion of the playlist are
# cataloged.
#------------------------------------------------------------------------------
proc LB::catalogViewed {args} {
   Trace
   variable LBdata
   variable cvTi
   variable cvBi
   global LB
   foreach {top bot} [$LB yview] {
      set len [$LB size]
      set topIndex [expr {int($len*$top)}]
      set botIndex [expr {int($len*$bot)}]
   }
   if {$cvTi != $topIndex || $cvBi != $botIndex} {
      LB::catalogBackground $topIndex $botIndex
   }
   set cvTi $topIndex
   set cvBi $botIndex
   after 1000 LB::catalogViewed
}
##
# TBD: Add break code and ability to restart if
# when the playlist changes (due to sorting or apl apply)
#
proc LB::catalogBackground {topIndex botIndex} {
   # TBD: Tie this to a configure or scroll event
   variable LBdata
   ##
   # Scan through each entry and schedule an update
   #
   set sch 0
   for {set i $topIndex} {$i <= $botIndex } {incr i} {
      LB::catalogIndex $i
      update
   }
   return
}
#------------------------------------------------------------------------------
# Update the entry the the currently playing track, if file
# has changed and db updated, this may update the entry
#------------------------------------------------------------------------------
proc LB::playingKeyUpdate {args} {
   variable indexPlaying
   LB::catalogIndex $indexPlaying
}
#------------------------------------------------------------------------------
# Catalog a single index, and update the entry
# Bug; if the same key is reused for multiple files
# can get one missing.
#------------------------------------------------------------------------------
proc LB::catalogIndex {index} {
   Trace
   variable LBdata
   variable Kdata
   global LB
   if {[catch {
      set file [lindex $Kdata $index]
      unset -nocomplain Data
      if {[db::lookup $file Data 1] > 0} {
         set newKey [LB::nameToKey $file]
         lset LBdata $index $newKey
         update
      }
   } result] } then {
      saLog $result
   }
}
# No longer needed
proc LB::relayout {args} {
   global LB
   variable indexPlaying
   variable LBdata
   variable Kdata
   ;# Save current tracks
   set active [LB::idxActive]
   set data [LB::idxSelected]
   set files [list]
   #replace them
   $LB delete 0 end
   foreach {item} $Kdata {
      set key [LB::nameToKey $item]
      lappend LBdata $key
   }
   foreach {position} $data {
      LB::setSelection $position
   }
   LB::activate $active
   LB::inverseVideoIndex $indexPlaying
}
#------------------------------------------------------------------------------
# Callback when dragging the entire playlist
#------------------------------------------------------------------------------
proc LB::dndSrcLBDrag {src} {
   variable Kdata
   Trace
   set Indicies [LB::idxSelected]
   set tracks [list]
   foreach {Index} $Indicies {
      lappend tracks [file nativename [lindex $Kdata $Index]]
   }
   return [dndFileList $tracks]
}

#------------------------------------------------------------------------------
# Return name of current track (for dropping)
#------------------------------------------------------------------------------
proc LB::currentTrack {args} {
   return [list $::TA(currentMF)]
}

#------------------------------------------------------------------------------
# Return number of lines in playist
#------------------------------------------------------------------------------
proc LB::lines {args} {
   global LB
   foreach {st fin} [$LB yview] {break}
   set sz [$LB size]
   set lines [expr {round(($fin-$st)*$sz)}]
   return $lines
}
#------------------------------------------------------------------------------
#
#------------------------------------------------------------------------------
#trace variable LB::LBdata w tLBdata
#trace variable LB::Kdata w tKdata
proc tKdata {a args} {
   puts "Kdata=[llength $LB::Kdata] $a"
}
proc tLBdata {a args} {
   puts "LBdata=[llength $LB::LBdata] $a"
}

#------------------------------------------------------------------------------
#
#------------------------------------------------------------------------------
proc t1 {args} {
   puts "Kdata=[llength $LB::Kdata] LBdata=[llength $LB::LBdata]"
}
package provide playList 1.0



