#------------------------------------------------------------------------------
# snackAmp - playurl package 1.0 for snackAmp
# provides music filtering utilities
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2003 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
namespace eval playurl {
   package require snack
   package require http 2.4
   variable s
   variable lastSocket {}
   variable lastToken {}
   variable icydata
   variable linkedInfo
   variable desc ""
   variable url ""

}
##
# Play a shortcut file, grab info from it for display
#
proc playurl::playusk {snd file _nextCommand _linkedInfo} {
   variable nextCommand
   variable lastSocket
   variable lastToken
   variable linkedInfo $_linkedInfo

   set nextCommand $_nextCommand
   Trace
   set url [playurl::readurl $file]
   set Data(Title) ""
   set Data(Album) ""
   id3Tag::id3Label $file Data V1
   if {[regexp {http://(.+?):(.+)/} $url -> host port]} {
      if {[string equal -nocase [info hostname] $host] && $port==$::snackAmpSettings(usedPort)} {
         saLog "Can't play $url from same SnackAmp Instance"
         return -code error "Local $url"
      }
   }

   if {[catch {playurl::play $snd $url} er] } then {
      return -code error $er
   } else {
      return [list $Data(Title) $Data(Album)]
   }
}
##
# Play a url, after the headers the playcb will file
# where we connect the socket to snack
# Note: -command xxx is used so the geturl
# doesn't block.
proc playurl::play {snd url} {
   Trace
   variable s $snd
   variable icydata
   unset -nocomplain icydata
   if {[catch {
      # -headers [list Icy-MetaData: 1]
      http::geturl $url -handler {playurl::playcb} \
                        -binary true \
                        -command playurl::done \
                        -blocksize 50000
   } result] } then {
      return -code error $result
   }
   return 1
}
##
# Background callback, should only be called once
# since we disable fileevents. We highjack the socket
# from http and feed it directly into snack.
# This routine is first called right after the header
# is read and before any data has been read.
#
# Transfer to snack based on a nifty trick by Kristofer Lawson
#
proc playurl::playcb {sock token} {
   Trace
   variable linkedInfo
   variable nextCommand
   variable s
   variable lastToken
   variable lastSocket
   variable icydata
   upvar $linkedInfo mylink

   set keys 0
   set lastToken $token
   set lastSocket $sock
   ##
   # Steal & Configure the socket as binary, give it a big buffer
   #
    fconfigure $sock -translation {binary binary} -encoding binary \
       -blocking true -buffering full -buffersize 200000
   ##
   # We only want to be callback once, just hook up the sockets
   # and let snack take over.
   #
   fileevent $sock readable {}
   upvar $token stateVar
    switch -glob -- $stateVar(type) {
       *ogg* {set type OGG}
       *mpeg* {set type MP3}
       default {set type ""}
    }
   if {[catch {
      if {[string length $type]} {
         $s configure -channel $sock -debug $::snackAmpSettings(Trace)  -fileformat $type
      } else {
         $s configure -channel $sock -debug $::snackAmpSettings(Trace)
      }
      update
      # Buffer data
      for {set i 0} {$i <= 10} {incr i} {
         waitWindow::percent $i 10 "Buffering $mylink(Title) $type [expr {$i*100}]ms..."
         after 100
         update
      }
      if {[string equal $type "MP3"]} {
         # Eat any leading ID3V2 tags if possible
         waitWindow::show "Reading Leading Tag $mylink(Title)..."
         set keys [id3Tag::id3V2Get $sock mylink]
      }
      waitWindow::show "Playing $mylink(Title)..."
      set soundControl::startOffset 0.0
      $s play -command [list playurl::cleanup $sock $token]
   } result] } then {
      close $sock
      catch {http::cleanup $token}
      after idle [list eval $nextCommand]
      saLog "playurl::playcb: $result"
      return 0
   } else {
      # remove "Connecting ..." from front of string
      waitWindow::close
      set mylink(fDate) [clock format [clock seconds] -format "%x %X"]
      if {$keys==0} {
         set mylink(Track) 1
      }
      # Save any icycast data for use displaying on the front GUI
      if {[info exists stateVar(meta)]} {
         array set icydata $stateVar(meta)
#parray icydata
         foreach {icy ar} {
            icy-br          rate
            icy-genre       Genre
            icy-name        Title
            ice-description Album
            ice-genre       Genre
            ice-name        Title
            icy-url         Album
            icy-artist      Artist
            Content-Length  Size
            Last-Modified   Date
         } {
            if {[info exists icydata($icy)]} {
               set mylink($ar) $icydata($icy)
            }
         }
      }
      # TODO Save the ID3tag data we have here for the usk file
      # hardcode songInfo (for now)
      # Need to upate the use file with the "correct" incoming icy data if possible
      # what is the usk file name?
      #   id3Tag::id3V1Modify $fname localArray
      #
      waitWindow::close
      return 1
   }
}
##
# Validate a URL (streams don't typically have HEAD features, so not useful)
#
proc playurl::validate {url} {
   Trace
   if {[catch {
      http::geturl $url -validate 1
   } token]} then {
      return -code error $token
   }
   upvar $token stateVar
   set code [lindex [http::code $token] 1]
   if {[string equal $code 200]} {
      return 1
   } else {
      return 0
   }
}
##
# Proxy required callback, called before every get,
# Determines if the proxy is needed (local or remote).
#
proc playurl::proxycb {url} {
   global snackAmpSettings
   if {[iptools::localUrl $url]==0 && $snackAmpSettings(useproxy)} {
      return [list $snackAmpSettings(proxyhost) $snackAmpSettings(proxyport)]
   } else {
      return [list]
   }
}
##
# Called when http get (stream complete) is done, either normally or if
# the stream didn't play
#
proc playurl::cleanup {sock token} {
   Trace
   variable nextCommand
   variable s
   upvar $token state
   #set code [lindex $state(http) 1]
   ##
   # 503 is when the url didn't exist
   # 302 is
   #
   #soundControl::stop
   catch {close $sock}
   http::cleanup $token
   # Fire the next sound command
   after 200 [list eval $nextCommand]
}
##
# Callback when stream is done, not sure what to do here since snack
# is handling the stream. Don't think this is ever called
# since we cancel the file events
# Will be called on ::http::9 $token
proc playurl::done {token} {
   Trace
   variable linkedInfo
   variable nextCommand
   Trace
   upvar $linkedInfo mylink
   set mylink(Title) "Done"
   soundControl::stop
   # Fire the next sound command
   after idle [list eval $nextCommand]
}
##
# Cleanup routine, called from Stop button
#
proc playurl::cleanupLast {args} {
   Trace
   variable lastToken
   variable lastSocket
   if {$lastToken != ""} {
      catch {close $lastSocket}
      catch {http::cleanup $lastToken}
      set lastSocket ""
   }
}
##
# Create a new url as a file using the
# ID3 tag info to hold
#
proc playurl::addPromptUrl {folder} {
   set tl .apurl
   destroy $tl
   toplevel $tl
   wm title $tl "URL Shortcut in $folder"
   variable desc
   variable url
   variable ok
   set ok 0
   set i 0
   set fr [frame $tl.[incr i]]
   pack $fr -side top -fill x
   grid [label $fr.[incr i] -text "Enter Description"] \
        [entry $fr.[incr i] -textvariable playurl::desc -width 45]
   grid [label $fr.[incr i] -text "Enter http url"] \
        [entry $fr.[incr i] -textvariable playurl::url  -width 45]

   # Line
   pack [frame $tl.[incr i] -height 2 -borderwidth 2 -relief sunken] -expand true -fill x

   # Button frame
   set bf [frame $tl.[incr i]]
   pack $bf -side bottom -fill none -expand false ;# This will center it
   #  ----------------------------------------------------------------------------
   #  Buttons at the bottom
   #  ----------------------------------------------------------------------------
   # Apply and Close
   set okBut [button $bf.[incr i] -text "OK"  -padx 4 \
      -width 80 -default active -borderwidth 1 -command {
         set ::playurl::ok 1
         destroy .apurl
   } -relief raised -compound left -image saImages.Apply]
   bind $tl <Return> [list $okBut invoke]

   # Cancel, restore settings
   set canBut [button $bf.[incr i] -text "Cancel" -width 80 -default normal -borderwidth 1 -command {
      set ::playurl::ok 0
      destroy .apurl
   } -relief raised -compound left -image saImages.Close]
   bind $tl <Escape> [list $canBut invoke]
   pack $okBut $canBut -side left -padx 20 -pady 5
   focus $fr.3
   centerWindow $tl
   myIcon $tl
   # Wait for user to close window
   grab $tl
   tkwait visibility $tl
   raise $tl
#  ----------------------------------------------------------------------------
#  Blocking
#  ----------------------------------------------------------------------------
   vwait ::playurl::ok
   if {$playurl::ok} {
      # Create the file
      if {[string length $playurl::url]} {
         if {[string length $playurl::desc]==0} {
            set playurl::desc $playurl::url
         }
         playurl::addUrl $folder $playurl::desc $playurl::url
      }
   }
}
##
# Create a new url shortcut file, store info within
# the file itself.
#
proc playurl::addUrl {folder description url} {
   set file [string trim [string map {/ - ? " " * " " : " " . _} $description]]
   set file [file join $folder $file.usk]
   if {[file exists $file]} {
      #saLog "playurl::addUrl: Overwriting $file"
   }
   if {[catch {open $file w} fid] } then {
      saLog "playurl::addUrl: $fid"
      return "Error with $file"
   } else {
      writeurl $url $fid
      close $fid
   }
   # Also put the url into the Album/Note fields to see in the APL table
   set url [HttpD::Decode $url]
   set parts [split $url ?]
   # TFW: Can clean up the Album/Artist part a little better
   #set Data(Album) [lindex $parts 0]
   foreach {Index} $id3Tag::v1Parts {
      set Data($Index) {}
   }
   if {[llength $parts]>1} {
      set Data(Artist)  [lindex $parts 1]
   }
   set Data(Title) $description
   id3Tag::id3V1Modify $file Data
   return $file
}
##
# Grab the contents of some URL and return it
# Used to grab a pls/m3u file for use
# in decoding.
#
proc playurl::getUrlBody {url} {
   #direct to contents
   package require http 2.4
   #set url "http://www.shoutcast.com/sbin/shoutcast-playlist.pls?m=1457&file=filename.pls"
   set token [::http::geturl $url]
   upvar #0 $token state
   return [string trim $state(body)]
}
##
# Determine if a file is a url shortcut file
#
proc playurl::isusk {file} {
   if {[string equal -nocase [file extension $file] ".usk"]} {
      return 1
   } else {
      return 0
   }
}
##
# Read a url from a url shortcut
#
proc playurl::readurl {file} {
   set fid [open $file r]
   set offset [id3Tag::id3V2Offset $fid]
   if {$offset > 0} {
      if {[catch {seek $fid $offset start} result] } then {
         debug $result
      }
   }
   set url [string trim [read $fid 256]]
   close $fid
   return $url
}
##
# Write a url to a url shortcut
#
proc playurl::writeurl {url fid} {
   set ulen [string length $url]
   if {[string length $url]} {
      set blank "$url[string repeat " " [expr {512-$ulen}]]"
      seek $fid 0 start
      puts -nonewline $fid $blank
   }

}
##
# Verify we have a location to save the http shortcuts in
#
proc playurl::checkuskfolder {args} {
   if {![file exists $::snackAmpSettings(uskDir)]} {
      set deffolder [file join [lindex [catalog::mountedFolders] 0] http]
      if {![file exists $deffolder] && [catch {file mkdir $deffolder} result]} {
         set deffolder [lindex [catalog::mountedFolders] 0]
      }
      tk_messageBox -title "Choose Folder" -type ok \
         -message "Please select a folder to save your URL shortcuts in.\n$deffolder is the default"
      set deffolder [chooseDirectory $deffolder [focus]]
      if {![file exists $deffolder]} {
         return 0
      } else {
         set ::snackAmpSettings(uskDir) $deffolder
         return [file exists $::snackAmpSettings(uskDir)]
      }
   } else {
      return 1
   }
}
package provide playurl 1.0
# Set up proxy callbacks
http::config -proxyfilter playurl::proxycb  -useragent "SnackAmp"
