#------------------------------------------------------------------------------
# snackAmp - playurl_mc package 1.0 for snackAmp
# uses memchan channel driver for buffer and instream detection
# of metadata
# See
# http://wiki.tcl.tk/13305
# http://wiki.tcl.tk/14032
# for inspiration
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2003 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
namespace eval playurl {
   package require snack
   package require Memchan
   package require http 2.4
   package require uri
   if {![info exists lastToken]} {
      variable s
      variable lastToken {}
      variable icydata
      variable linkedInfo
      variable desc ""
      variable url ""
      variable fd ""
      variable readSize 0
      variable header
      variable total 0
      variable lastusk ""
      variable metaint 0
      variable bcar "\u25B7"
      variable urlHistory [list]
      # 250000KB ~ 15 seconds at 128Kbps + the internal socket buffer
      variable minBuffer
      upvar ::snackAmpSettings(minBuffer) minBuffer
      variable fixedSize 8192
   }
}
##
# Play a shortcut file, grab info from it for display
#
proc playurl::playusk {snd file _nextCommand _linkedInfo} {
   Trace
   variable nextCommand
   variable lastToken
   variable lastusk
   variable linkedInfo $_linkedInfo
   variable urlHistory [list]
   set nextCommand $_nextCommand
   set lastusk $file
   set url [playurl::readurl $file]
   set Data(Title) "$url"
   soundControl::pushTitle $url
   set Data(Album) "Please Wait..."
   id3Tag::id3Label $file Data
   update
   if {[catch {playurl::play $snd $url} er] } then {
      return -code error $er
   } else {
      return [list $Data(Title) $Data(Album)]
   }
}
#----------------------------------------------------------------------------
# Play a url, after the headers the playcb will file
# where we connect the socket to snack
# Note: -command xxx is used so the geturl
# doesn't block.
#----------------------------------------------------------------------------
proc playurl::play {snd url} {
   Trace
   variable urlHistory
   variable lastToken
   if {$lastToken ne ""} {
      saLog "Still have a stream in process"
      return 1
   }
   variable s $snd
   variable icydata
   unset -nocomplain icydata
   if {[lsearch -exact $urlHistory $url] < 0} {
      if {[catch {
         lappend urlHistory $url
         http::geturl $url -handler [list playurl::playcb $snd] \
                           -binary true \
                           -headers {icy-metadata 1} \
                           -command playurl::done
      } result] } then {
         return -code error $result
      }
   } else {
      saLog "Recursive url $url: Skipping"
   }
   return 1
}
#----------------------------------------------------------------------------
# Background callback, should only be called once
# since we disable fileevents. We highjack the socket
# from http and feed it info a memchan fifo.
# This routine is first called right after the header
# is read and before any data has been read. So at this
# point the url is valid.
#
# Note, this has a reentrant problem
#----------------------------------------------------------------------------
proc playurl::playcb {snd sock token} {
   Trace
   variable linkedInfo
   variable nextCommand
   variable s
   variable icydata
   variable fd
   variable metaint
   variable readSize
   variable total 0
   variable minBuffer
   variable lastusk
   variable bcar
   variable lastToken
   variable urlHistory
   variable fixedSize
   if {$minBuffer < $fixedSize} {
      set minBuffer $fixedSize
   }
   set bcar "\u25B7"
   upvar $linkedInfo mylink
   upvar $token stateVar
   set code [::http::ncode $token]
   ##
   # If this is a redirect, then find it and reschedule the player on the new link
   #
   if {$code >= 300 && $code <= 307} {
      array set meta $stateVar(meta)
      if {[info exists meta(Location)]} {
         array set uri [::uri::split $meta(Location)]
         unset meta
         set url [eval ::uri::join [array get uri]]
saLog "Redirect\n $stateVar(url) to \n $url"
         catch {close $sock}
         ::http::cleanup $token
         set lastToken ""
         after idle [list playurl::play $snd $url]
         return
      }
   }

   set type ""
   set keys 0
   set lastToken $token
   #----------------------------------------------------------------------------
   # Create the buffer, fd buffersize indicates how much can be read off at once by snack
   # sock[et] -> fifo fd -> snack
   #----------------------------------------------------------------------------
   set fd [fifo]
   # The $fd buffersize tells snack how mych t read off the fifo, if it
   # is too large then the entire buffer will be read at once (which is OK).
   # BUG: if buffersize is set less than amount in fifo, we get squawks in the music
   #      or so it seems.
   fconfigure $fd -translation {binary binary} -encoding binary \
         -blocking true -buffering none -buffersize $minBuffer
   ##
   # Configure the stream socket as binary, give it a smaller buffer
   # pass it off to the stream handling code
   fconfigure $sock -translation {binary binary} -encoding binary \
       -blocking false -buffering full -buffersize $minBuffer

   # Save any icycast data for use displaying on the front GUI
   #set stateVar(status) 0
   set Artist "Various"
   unset -nocomplain icydata
   if {[info exists stateVar(meta)]} {
      array set icydata $stateVar(meta)
      foreach {icy ar} {
         icy-br          Rate
         icy-metaint     Metadata
         icy-genre       Genre
         icy-name        Title
         ice-description Album
         ice-genre       Genre
         ice-name        Title
         icy-url         Album
         icy-artist      Artist
         Content-Length  Size
         Last-Modified   Date
      } {
         if {[info exists icydata($icy)]} {
            switch -- $ar {
               Rate {
                  if {[string is integer -strict $icydata($icy)]} {
                     set mylink(Rate) "$icydata($icy) kbps"
                  } else {
                     set mylink($ar) $icydata($icy)
                  }
               }
               default {set mylink($ar) $icydata($icy)}
            }
         }
      }
   }
   ##
   # First thing read the header and process from there
   # http://www.atm.ox.ac.uk/user/iwi/charmap.html
   #
   if {[info exists icydata(icy-metaint)] && $icydata(icy-metaint)>0} {
      set metaint $icydata(icy-metaint)
      set readSize $metaint
      fileevent $sock readable [list playurl::readStreamMetaInt $sock]
#      set bschar "\u2587" ;# big block (no gaps)
      set bschar "\u25B6" ;# Arrow
#      set bschar "\u266A" ;# Musical note
   } else {
      set metaint 0
      fileevent $sock readable [list playurl::readStream $sock]
      set bschar "\u25A0" ;# small block
   }
   ##
   # Update the usk tags with latest info, update playlist as well
   #
   if {[info exists stateVar(meta)] && [file exists $lastusk]} {
      arrayCopyNoBlank mylink Data
      id3Tag::id3V1Modify $lastusk Data
      id3Tag::id3V2Modify $lastusk Data
      # update last played and playcount also
      db::cacheFile $lastusk Data 1
      db::bumpPlayed $lastusk
      LB::changePlaylistFiles [list $lastusk $lastusk]
   }
   ##
   # Buffer the data in the fifo for a bit, make sure we have some before we continue
   # During the update readStream[MetaInt] is running
   set bs 0
   # 100 x 50 = 5 seconds
   for {set i 0} {$i <= 100} {incr i} {
      update
      after 50
      set bs [fifoSize $fd]
      if {$bs > $minBuffer} {
         break
      }
   }
   ##
   # If no or minimal data if fifo, we must have timeout
   #
   if {$bs < $fixedSize} {
      saLog "Insufficient data for stream playing. Only have $bs bytes"
      set mylink(Artist) "Insufficient data: $bs bytes"
      playurl::done $token
      return 0
   }
   ##
   # Try and help the codec detection
   #
   if {[info exists stateVar(type)]} {
      switch -glob -- $stateVar(type) {
         *ogg* {set type OGG}
         *mpeg* {set type MP3}
         default {set type ""}
      }
   }
   ##
   # Connect the snack channel to the in-memory channel
   #
   if {[catch {
      if {[string length $type]} {
         # Eat any leading ID3V2 tags for non-shoutcast buffer
         # and update the tag if it has some
         if {$type eq "MP3" && $metaint==0} {
            set keys [id3Tag::id3V2Get $fd junk]
            if {$keys > 0} {
               # TODO: Bring over only  certain keys??
               # TODO: Tag can contain an image (like CNN), use it
               if {[file exists $lastusk]} {
                  arrayCopyNoBlank junk Data
                  arrayCopyNoBlank junk mylink
                  id3Tag::id3V1Modify $lastusk Data
                  id3Tag::id3V2Modify $lastusk Data
                  LB::changePlaylistFiles [list $lastusk $lastusk]
               }
            }
         }
         $s configure -channel $fd -debug $::snackAmpSettings(Trace)  -fileformat $type
      } else {
         $s configure -channel $fd -debug $::snackAmpSettings(Trace)
      }
      set soundControl::startOffset 0.0
      # Start snack now that we have some buffered in the fifo fd
      set bcar $bschar
      #set bcar "\u25A0"
      $s play -command [list playurl::done $token]
   } result] } then {
      # Failed to configure
      playurl::done $token
      saLog "[me] Error: $result\n$::errorInfo"
      return 0
   } else {
      # remove "Connecting ..." from front of string
      set mylink(fDate) [clock format [clock seconds] -format "%x %X"]
      if {$keys==0} {
         set mylink(Track) 1
      }
      # Everything proceeds from this point....
      return 1
   }
}
##
# Proxy required callback, called before every get,
# Determines if the proxy is needed (local or remote).
#
proc playurl::proxycb {url} {
   global snackAmpSettings
   if {[iptools::localUrl $url]==0 && $snackAmpSettings(useproxy)} {
      return [list $snackAmpSettings(proxyhost) $snackAmpSettings(proxyport)]
   } else {
      return [list]
   }
}

#------------------------------------------------------------------------------
# Try and return the fifo size
#------------------------------------------------------------------------------
proc playurl::fifoSize {fd} {
   if {$fd eq "" || [catch {fconfigure $fd -length} bs] } then {
      return 0
   } else {
      return $bs
   }
}
##
# Http done handler, cleanup and launch the next command
# Called with stream is finshed or doesn't exist
proc playurl::done {token} {
   Trace
   variable lastToken
   variable linkedInfo
   variable nextCommand
   variable fd
   if {[info exists linkedInfo]} {
      upvar $linkedInfo mylink
      set mylink(Artist) "Insufficient Data Available"
   }
   # If naturally closing out, play next command
   if {[cleanupLast $token]} {
      #TODO: Want to allow stop command to interceded here
      after 1000 $nextCommand
      soundControl::stop
   }
   # Fire the next sound command
}
##
# Cleanup routine, called from Stop button or when stream is finished
#
proc playurl::cleanupLast {{token ""}} {
   Trace
   variable lastToken
   variable linkedInfo
   variable fd
   if {$token ne ""} {
      set lastToken $token
   }
   # Notify the user
   if {[info exists linkedInfo]} {
      upvar $linkedInfo mylink
      set mylink(Artist) "Complete"
   }
   # Clean up channels and memory
   upvar $lastToken state
   if {$lastToken != ""} {
      tryCloseChannel state(sock)
      unset -nocomplain state
      tryCloseChannel fd
      set lastToken ""
      return 1
   } else {
      return 0
   }
}
##
# Close a channel if we can find it
#
proc playurl::tryCloseChannel {_channel} {
   upvar $_channel channel
   if {[info exists channel] && [lsearch -exact [file channels] $channel]>-1} {
      close $channel
   }
}

##
# Create a new url as a file using the
# ID3 tag info to hold
#
proc playurl::addPromptUrl {folder} {
   set tl .apurl
   destroy $tl
   toplevel $tl
   wm title $tl "URL Shortcut in $folder"
   variable desc
   variable url
   variable ok
   set ok 0
   set i 0
   set fr [frame $tl.[incr i]]
   pack $fr -side top -fill x
   set desc ""
   set url ""
   grid [label $fr.[incr i] -text "Enter http url"] \
        [entry $fr.[incr i] -textvariable playurl::url  -width 45]
   grid [label $fr.[incr i] -text "Optional Description"] \
        [entry $fr.[incr i] -textvariable playurl::desc -width 45]

   # Line
   pack [frame $tl.[incr i] -height 2 -borderwidth 2 -relief sunken] -expand true -fill x

   # Button frame
   set bf [frame $tl.[incr i]]
   pack $bf -side bottom -fill none -expand false ;# This will center it
   #  ----------------------------------------------------------------------------
   #  Buttons at the bottom
   #  ----------------------------------------------------------------------------
   # Apply and Close
   set okBut [button $bf.[incr i] -text "OK"  -padx 4 \
      -width 80 -default active -borderwidth 1 -command {
         set ::playurl::ok 1
         destroy .apurl
   } -relief raised -compound left -image saImages.Apply]
   bind $tl <Return> [list $okBut invoke]

   # Cancel, restore settings
   set canBut [button $bf.[incr i] -text "Cancel" -width 80 -default normal -borderwidth 1 -command {
      set ::playurl::ok 0
      destroy .apurl
   } -relief raised -compound left -image saImages.Close]
   bind $tl <Escape> [list $canBut invoke]
   pack $okBut $canBut -side left -padx 20 -pady 5
   focus $fr.3
   centerWindow $tl
   myIcon $tl
   # Wait for user to close window
   grab $tl
   tkwait visibility $tl
   raise $tl
#  ----------------------------------------------------------------------------
#  Blocking
#  ----------------------------------------------------------------------------
   vwait ::playurl::ok
   if {$playurl::ok} {
      # Create the file
      if {[string length $playurl::url]} {
         if {[string length $playurl::desc]==0} {
            set playurl::desc $playurl::url
         }
         playurl::addUrl $folder $playurl::desc $playurl::url
      }
   }
}
##
# Create a new url shortcut file, store info within
# the file itself.
# TODO: Optional Arra to Pass in
#
proc playurl::addUrl {folder description url {_Data ""}} {
   Trace
   set file [string trim [string map {\[ ( \] ) / - ? " " * " " : " " . _} $description]]
   set file [file join $folder $file.usk]
   if {[file exists $file]} {
      #saLog "playurl::addUrl: Overwriting $file"
   }
   if {[catch {open $file w} fid] } then {
      saLog "playurl::addUrl: $fid"
      return "Error with $file"
   } else {
      writeurl $url $fid
      close $fid
   }
   # Also put the url into the Album/Note fields to see in the APL table
   set url [HttpD::Decode $url]
   set parts [split $url ?]
   # TFW: TODO: Can clean up the Album/Artist part a little better
   soundControl::pushTitle $description
   set Data(Title) $description
   foreach {Index} $id3Tag::v1Parts {
      set Data($Index) {}
   }
   # If user passed in array, use that data instead
   if {$_Data ne ""} {
      upvar $_Data argData
      array set Data [array get argData]
   } else {
      if {[llength $parts]>1} {
         set Data(Album) [file tail [lindex $parts 0]]
         set Data(Artist)  [lindex $parts 1]
      } else {
         set Data(Artist)  [file tail [lindex $parts 0]]
      }
   }
   # Update both tags with the info
   id3Tag::id3V1Modify $file Data
   id3Tag::id3V2Modify $file Data
   db::cacheFile $file Data 1
   LB::changePlaylistFiles [list $file $file]
   return $file
}
##
# Grab the contents of some URL and return it
# Used to grab a pls/m3u file for use
# in decoding.
#
proc playurl::getUrlBody {url} {
   #direct to contents
   package require http 2.4
   #set url "http://www.shoutcast.com/sbin/shoutcast-playlist.pls?m=1457&file=filename.pls"
   set token [::http::geturl $url]
   set contents [::http::data $token]
   ::http::cleanup $token
   return [string trim $contents]
}
##
# Determine if a file is a url shortcut file
#
proc playurl::isusk {file} {
   if {[string equal -nocase [file extension $file] ".usk"]} {
      return 1
   } else {
      return 0
   }
}
##
# Read a url from a url shortcut
#
proc playurl::readurl {file} {
   set fid [open $file r]
   set offset [id3Tag::id3V2Offset $fid]
   if {$offset > 0} {
      if {[catch {seek $fid $offset start} result] } then {
         debug $result
      }
   }
   set url [string trim [read $fid 256]]
   close $fid
   return $url
}
##
# Write a url to a url shortcut
#
proc playurl::writeurl {url fid} {
   set ulen [string length $url]
   if {[string length $url]} {
      set blank "$url[string repeat " " [expr {512-$ulen}]]"
      seek $fid 0 start
      puts -nonewline $fid $blank
   }

}
##
# Verify we have a location to save the http shortcuts in
#
proc playurl::checkuskfolder {args} {
   if {![file exists $::snackAmpSettings(uskDir)]} {
      set deffolder [file join [lindex [catalog::mountedFolders] 0] http]
      if {![file exists $deffolder] && [catch {file mkdir $deffolder} result]} {
         set deffolder [lindex [catalog::mountedFolders] 0]
      }
      tk_messageBox -title "Choose Folder" -type ok \
         -message "Please select a folder to save your URL shortcuts in.\n$deffolder is the default"
      set deffolder [chooseDirectory $deffolder [focus]]
      if {![file exists $deffolder]} {
         return 0
      } else {
         set ::snackAmpSettings(uskDir) $deffolder
         return [file exists $::snackAmpSettings(uskDir)]
      }
   } else {
      return 1
   }
}
#----------------------------------------------------------------------------
# Code lifted from wiki and modified here
# Thanks to DZ and JT for ideas
# I corrected the pops-clicks by changing the fconfigure options
#----------------------------------------------------------------------------
proc playurl::readStream {sock} {
   variable total
   variable fd
   variable linkedInfo
   variable minBuffer
   variable fixedSize
   # stream has just music data, no music information
   set bs [fifoSize $fd]
   # Can handle desired buffer size here
   if {$bs < $minBuffer} {
      upvar $linkedInfo mylink
      set mylink(Artist) [playurl::bufferString $bs]
      if {[catch {read $sock $fixedSize} data] } then {
         saLog $data
         return
      }
      set slen [string length $data]
      if {$slen==0} {
         fileevent $sock readable {}
         after 100 [list playurl::regReadEvent playurl::readStream $sock]
      } else {
         incr total $slen
         if {[catch {puts -nonewline $fd $data} result] } then {
            fileevent $sock readable {}
            saLog "[me] $result"
         }
      }
   } else {
      # flow control
      # slows down repeated calls to this function when buffer is full
      fileevent $sock readable {}
      after 300 [list playurl::regReadEvent playurl::readStream $sock]
      upvar $linkedInfo mylink
      set mylink(Artist) "Stream Buffer Full ([expr {$bs/1000}]KB)"
   }
}
##
# Return a string representing the buffer length
#
proc playurl::bufferString {bs} {
   variable minBuffer
   variable bcar
   set maxChar 15  ;# twice this
   return "([expr {$bs/1000}]KB) [string repeat $bcar [expr {($bs*$maxChar)/$minBuffer}]]"
}
#----------------------------------------------------------------------------
# Read a block of stream data and send it to snack.
# If we have exausted the entire block, setup to read the first
# byte of the following block which contains the embedded metadata
# length/16.
#----------------------------------------------------------------------------
proc playurl::readStreamMetaInt {sock} {
   variable readSize
   variable total
   variable fd
   variable metaint
   variable linkedInfo
   variable minBuffer
   set bs [fifoSize $fd]
   # If use paused, fifo fill continue to fill
   if {$bs < $minBuffer} {
      set data [read $sock $readSize]
      set slen [string length $data]
      if {$slen==0} {
         # hold off a bit
         fileevent $sock readable {}
         after 100 [list playurl::regReadEvent playurl::readStreamMetaInt $sock]
      } else {
         incr total $slen
         if {[catch {puts -nonewline $fd $data} result] } then {
            fileevent $sock readable {}
            saLog "[me] $result"
         }
      }
      upvar $linkedInfo mylink
      set mylink(Artist) [playurl::bufferString $bs]
      # See if we are ready for the title
      if {$total != $metaint} {
         set readSize [expr {$metaint - $total}]
      } else {
         # and end of block, next read is the metadata size (typically 0)
         set readSize $metaint
         set total 0
         fileevent $sock readable [list playurl::readTitleLength $sock]
      }
   } else {
      # flow control
      # slows down repeated calls to this function when buffer is full
      set mylink(Artist) "Stream Buffer Full ([expr {$bs/1000}]KB)"
      fileevent $sock readable {}
         # hold off a bit
      after 200 [list playurl::regReadEvent playurl::readStreamMetaInt $sock]
   }
}

#------------------------------------------------------------------------------
# Register a file event, catch errors incase the user hit Stop in the mean time
# Also gets an error when stream completes
#------------------------------------------------------------------------------
proc playurl::regReadEvent {event sock} {
   if {[catch {fileevent $sock readable [list $event $sock]} result] } then {
      saLog "[me] $result"
   }
}
#----------------------------------------------------------------------------
# Read the meta data lengths, first byte at the start of a block
# It determines the number of bytes of the metadata
#----------------------------------------------------------------------------
proc playurl::readTitleLength {sock} {
   Trace
   variable fd
   set c 0
   set titleSize [read $sock 1]
   scan $titleSize %c c
   set titleSize [expr {$c * 16}]
   if {$c == 0} {
      # no meta data follows, just stream
      fileevent $sock readable [list playurl::readStreamMetaInt $sock]
   } else {
      # title follows
      fileevent $sock readable [list playurl::readTitle $sock $titleSize]
   }
}
#----------------------------------------------------------------------------
# Read the text of the title and update the display accordinly
#----------------------------------------------------------------------------
proc playurl::readTitle {sock size} {
   Trace
   variable linkedInfo
   upvar $linkedInfo mylink
   #Shoutcast song information looks like this:
   #  StreamTitle='<title>';StreamUrl='<url>';

   set t ""
   while {$size > 0} {
      set data [read $sock $size]
      append t $data
      set size [expr {$size - [string length $data]}]
   }
   set t [string trim $t]
   if {[regexp -nocase -- {streamtitle='(.*?)';} $t foo _title] && $_title != ""} {
      set mylink(Title) $_title
      soundControl::pushTitle $_title
   }
   if {[regexp -nocase -- {streamurl='(.*?)';} $t foo url]} {
      if {$url ne ""} {
         set mylink(Album) $url
      }
   }
   # Can start the read of the next block here
   fileevent $sock readable [list playurl::readStreamMetaInt $sock]
}

package provide playurl_mc 1.0
# Set up proxy callbacks
http::config -proxyfilter playurl::proxycb  -useragent "SnackAmp"
