#------------------------------------------------------------------------------
# snackAmp - saFiltering package 1.0 for snackAmp
# provides music filtering utilities
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/saFiltering.tcl,v 1.36 2006/02/16 01:22:13 wilkason Exp $
}
namespace eval filtering {
   variable sortModes [list "Sort by all Tracks" "Shuffle by all Tracks" "Shuffle by Folder, Sort by Track" "Shuffle by Folder, Shuffle by Track"]
}
#------------------------------------------------------------------------------
# Function   : filtering::filterHistory
# Description: postcommand callback to Display the history filter, called from
#              button next to filter box.
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc filtering::filterHistory {menu} {
   global snackAmpSettings
   # Only keep up to last to filters
   set snackAmpSettings(History) [lrange $snackAmpSettings(History) 0 20]
   $menu delete 0 end
   destroy $menu.replace
   destroy $menu.append
   destroy $menu.select
   destroy $menu.insert
   $menu add cascade -label "Append To Playlist..." -menu [menu $menu.append] -compound left -image saImages.appendImage
   $menu add cascade -label "Queue To Playlist..." -menu [menu $menu.insert] -compound left -image saImages.queue
   $menu add cascade -label "Replace Playlist..." -menu [menu $menu.replace]  -compound left -image saImages.replaceImage
   $menu add cascade -label "Select In Playlist..." -menu [menu $menu.select]
   # Append
   $menu.append add command -label "Filtering '$::TA(filter)' on all known files" -command {
      filtering::pushFilter
      macros::filterMusic append all
   }
   $menu.append add command -label "Filtering '$::TA(filter)' on AutoPlaylist files" -command {
      filtering::pushFilter
      macros::filterMusic append apl
   }

   $menu.select add command -label "Filtering '$::TA(filter)' on current Playlist" -command {
      filtering::pushFilter
      macros::filterMusic select playList
   }

   # Replace
   $menu.replace add command -label "Filtering '$::TA(filter)' on all known files" -command {
      filtering::pushFilter
      macros::filterMusic replace all
   }
   $menu.replace add command -label "Filtering '$::TA(filter)' on AutoPlaylist files" -command {
      filtering::pushFilter
      macros::filterMusic replace apl
   }
   $menu.replace add command -label "Filtering '$::TA(filter)' on current Playlist" -command {
      filtering::pushFilter
      macros::filterMusic replace playList
   }

   # Insert
   $menu.insert add command -label "Filtering '$::TA(filter)' on all known files" -command {
      filtering::pushFilter
      macros::filterMusic queue all
   }
   $menu.insert add command -label "Filtering '$::TA(filter)' on AutoPlaylist files" -command {
      filtering::pushFilter
      macros::filterMusic queue apl
   }
   $menu.insert add command -label "Filtering '$::TA(filter)' on current Playlist" -command {
      filtering::pushFilter
      macros::filterMusic queue playList
   }

   $menu add separator
   $menu add command -label "Reset Filter to AutoPlaylist" -command {
      set ::TA(filter) {}
      macros::applyCurrentAPL
   }
   $menu add separator
   foreach {item} $snackAmpSettings(History) {
      $menu add command -label "$item" -command "[list set ::TA(filter) $item]"
   }
}
#------------------------------------------------------------------------------
# Function   : filtering::pushFilter
# Description: Save the latest filter entry if it is not already in the filter history
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc filtering::pushFilter {args} {
   if {$::TA(filter) != {} && [lsearch -exact $::snackAmpSettings(History) $::TA(filter)] < 0} {
      lpush left ::snackAmpSettings(History) $::TA(filter)
   }
}
#------------------------------------------------------------------------------#
# Function   : filtering::filterMusic
# Description: This is a common routine to filter the playlist, either by
#              text or matching against the AutoPlaylist settings (or both)
#              The end result is placed into the playlist, either replacing
#              or appending tracks already there.
# Author     : Tom Wilkason
# Date       : 7/1/2002
#------------------------------------------------------------------------------
proc filtering::filterMusic {how what} {
   Trace
   global TA snackAmpSettings
   ##
   # Filter on entire list or just the playlist
   #
   set useList [list]
   switch -- $what {
      "playList" {
         # Use only playlist
         set useList [LB::keyOrNamesAll 1]
      }
      1 -
      "all" {
         # Use all files (last known list)
         set useList [db::sortedList]  ;#This is why I want it kept ordered, maybe write to a file!
      }
      0 -
      "apl" {
         set useList [db::aplFiles]
      }
      default {
         # use Match against All flag
         set useList [db::sortedList]
      }
   }
   ##
   # Where to put the filtered data
   #
   switch -- $how {
      "replace" {
         LB::deleteAll
         set select 0
         set from "end"
      }
      "insert" {
         set from [expr {[LB::idxActive]+1}]
         set select 0
      }
      "queue" {
         set from [expr {[LB::idxQueued]+1}]
         set select 0
      }
      "append" {
         set from "end"
         set select 0
      }
      "select" {
         set from "end"
         set select 1
      }
      default {
         set from "end"
         set select 0
      }
   }
   set insertList [list]
   ##
   # Support boolean searches with 'or' clause
   #
   if {[string length $::TA(filter)]==0} {
      set fast 1
      set regPattern ""
   } else {
      set fast 0
      set regPattern [list]                                ;#retain wildcards
      set pattern [split [string map {" or " "|" " OR " "|" "*" ".*" "?" "."} [catalog::globEscape $TA(filter)]] |]
      foreach part $pattern {
         lappend regPattern "($part)+"
      }
      set regPattern [join $regPattern |]
   }
   ##
   # Now scan through each file in the list and filter it
   #
   if {[catch {
      set i -1
      foreach {File} $useList {
         incr i
         # Pattern match (of if no pattern ignore regexp for speed)
         if {$fast || [regexp -nocase -- $regPattern $File]} {
            # Either create an insert list or just select the track
            if {$select} {
               LB::setSelection $i
            } else {
               lappend insertList $File
            }
         }
      }
   } result]} {
      saLog "filtering::filterMusic: '$regPattern' $result"
   }
   if {[llength $insertList] > 0} {
      # Insert per the current sorting/shuffling
      LB::insertList $from [sortAList $insertList]
   }
   LB::tryRestorePlaying
}

#------------------------------------------------------------------------------
# Search the media library and display results in AutoPlaylist
#------------------------------------------------------------------------------
proc filtering::searchFilter {args} {
   if {![APLExists]} {
      package require saTree
      package require autoPlayList
      $::TA(treeButton) configure -relief sunken
      # First make sure we have the proper roots for the tree viewer
      APL::FolderTree
      update  ;# let it draw
   } elseif {[wm state .wFC] != "normal"} {
      $::TA(treeButton) configure -relief sunken
      APL::FolderTree
      update  ;# let it draw
   }
   set ::snackAmpSettings(FilterWhat) "Any"
   set ::TA(SAFilter) $::TA(filter)  ;# copy over search data
   if {[string length $::TA(SAFilter)]} {
      APL::setFilter 1
      trackList::populate "" $::PLB 1 1
   }
   raise .wFC
}
#------------------------------------------------------------------------------
# Function   : filtering::sortPlaylist
# Description: Sort or shuffle the playlist files
# Author     : Tom Wilkason
# Date       : 12/26/2000
#------------------------------------------------------------------------------
proc filtering::sortPlaylist {args} {
   Trace
   global TA snackAmpSettings
   soundControl::latency long
   Wait 1
   set playList [LB::keyOrNamesAll 1] ;# Get full file names in playlist
   if {[llength $playList] > 0} {
      LB::deleteAll
      LB::insertList end [filtering::sortAList $playList]
   } else {
      gui::updateFileDisplay
   }
   catalog::SavePlaylistCache
   LB::tryRestorePlaying
   Wait 0
   soundControl::latency
}
#------------------------------------------------------------------------------
# Function   : filtering::sortAList
# Description: Sort a list (of files) based on the current sorting mode
#              Used by the playlist sorter
# Author     : Tom Wilkason
# Date       : 12/26/2000
#------------------------------------------------------------------------------
proc filtering::sortAList {Items} {
   variable sortModes
   global snackAmpSettings
   set index [lsearch -exact $sortModes $snackAmpSettings(sortModeTxt)]
   switch -- $index {
      0  {return [lsort -dictionary $Items]}
      1  {return [shuffle           $Items]}
      2  {return [shuffleByFolder   $Items ldsort]}
      3  {return [shuffleByFolder   $Items shuffle]}
      default {
         package require subSorting
         set tosort [subSort::getModes]
         return [db::subSorting $tosort $Items]
      }
   }
}
#----------------------------------------------------------------------------
# Sort the APL prior to applying it, reduces multiple db passes
#----------------------------------------------------------------------------
proc filtering::sortAPLlist {args} {
   variable sortModes
   global snackAmpSettings
   set index [lsearch -exact $sortModes $snackAmpSettings(sortModeTxt)]
   set max  [expr {[settings::maxAPL]-1}]
   switch -- $index {
      0  {return [lsort -dictionary [db::aplFiles $max]]}
      1  {return [shuffle           [db::aplFiles $max]]}
      2  {return [shuffleByFolder   [db::aplFiles $max] ldsort]}
      3  {return [shuffleByFolder   [db::aplFiles $max] shuffle]}
      default {
         package require subSorting
         set tosort [subSort::getModes]
         # apl list is built in db::subSorting for speed
         return [db::subSorting $tosort]
      }
   }
}

package provide saFiltering 1.0

