#------------------------------------------------------------------------------
# snackAmp - saSettings package 1.0 for snackAmp
# provides user preferences and GUI
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/saSettings.tcl,v 1.128 2007/10/01 01:33:49 wilkason Exp $
}
namespace eval settings {
   variable reg
   variable FONTS
   variable Colors
   variable taskBarIconList {Prev Play Pause Stop Next ToggleArt AutoPlaylist}
   variable taskBarIconImages
   # image to icon translation
   array set taskBarIconImages {
      Prev         saImages.player_start
      Play         saImages.player_play
      Pause        saImages.player_pause
      Stop         saImages.player_stop
      Next         saImages.player_next
      ToggleArt    saImages.coverart
      AutoPlaylist saImages.treeImage
   }
   # On XP, the icons are drawn backwards!
   if {$::tcl_platform(osVersion)>=5.1} {
      set taskBarIconList [lflip $taskBarIconList]
   }
   ##
   # Create the initial fonts we will need, they may be scaled or changed later
   #  macintosh or unix
   switch -- $::tcl_platform(platform) {
      "windows" {
         set FONTS {
            entry      {"Trebuchet MS" 8  normal roman} "Entry Fields"
            listbox    {Tahoma  8  normal roman}        "List Boxes"
            smallLabel {Tahoma  8  normal italic}       "Small Track Label"
            songInfo   {Verdana 8  normal roman}        "Small Track Info"
            songLabel  {"Trebuchet MS" 11 bold italic}  "Track Title"
            songArtist {"Trebuchet MS" 9 bold roman}    "Track Artist"
            timeStatus {Verdana 13 bold   roman}        "Time Counters"
            timeScale  {Tahoma  6  normal roman}        "Time Scale"
            treeFont   {Tahoma  8  normal roman}        "AutoPlaylist"
            balloon    {Tahoma  8  normal roman}        "Balloon Help"
         }
      }
      default {
         set FONTS {
            entry      {Helvetica 10 bold   roman}  "Entry Boxes"
            listbox    {Helvetica 10 normal roman}  "List Boxes"
            smallLabel {Helvetica 10 normal italic} "Small Track Label"
            songInfo   {Helvetica 10 bold   roman}  "Small Track Info"
            songArtist {Helvetica 11 bold   roman}  "Track Artist"
            songLabel  {Helvetica 13 bold   roman}  "Track Title"
            timeStatus {Helvetica 14 bold   roman}  "Time Counters"
            timeScale  {Helvetica 8  normal roman}  "Time Scale"
            treeFont   {Helvetica 10 normal roman}  "AutoPlaylist"
            balloon    {Helvetica 9  normal roman}  "Balloon Help"
         }
      }
   }
   ##
   # Set some defaults for colors, overridden later
   #
   set ::snackAmpSettings(ebg)  "#ffffe6"    ;# entry box background
   set ::snackAmpSettings(efg)  "#584943"    ;# entry text color
   set ::snackAmpSettings(bg)   "#443833"    ;# text background
   set ::snackAmpSettings(tfg)  "LightCyan"  ;# Title Text Color
   set ::snackAmpSettings(afg)  "#8dc9f5"    ;# Artist Text Color
   set ::snackAmpSettings(fg)   "#ffffe6"    ;# text foreground
   set ::snackAmpSettings(sfg)  "#50acf1"    ;# Time Counter forground
   set ::snackAmpSettings(lfg)  "#ffe44a"    ;# song extra info label forground
   set ::snackAmpSettings(tsbg) "#64524a"    ;# table strip background
   set ::snackAmpSettings(highlight) "#50acf1"  ;# highlight color
   # playlist fields
   set ::snackAmpSettings(pl1)  Duration
   set ::snackAmpSettings(pl2)  Artist
   set ::snackAmpSettings(pl3)  Title
   set ::snackAmpSettings(pl4)  None
   #   default    {Helvetica 10 normal roman} "Default"
   foreach {Font Specs Description} $FONTS {
      foreach {family size weight slant}  $Specs {break}
      font create $Font -family $family -weight $weight -size $size -slant $slant
   }
}
#------------------------------------------------------------------------------
# Function   : settings::tabbedSettingsGui
# Description: Main Settings GUI
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc settings::tabbedSettingsGui {args} {
   variable reg
   package require snID3
   global TA snackAmpSettings
   set B .tsg
   destroy $B
   toplevel $B  -cursor $::TA(arrow)  -class SnackAmp
   wm withdraw $B
   wm protocol $B WM_DELETE_WINDOW {settings::closeWindow 0}
   wm title $B "SnackAmp Preferences"
   set options {
      "Play Settings"      "playSettings"      "- Settings that are common or affect "
      "User Interface"     "uiSettings"        "- Settings that are common or affect "
      "Cover Art"          "caSettings"        "- Cover Art and Pop-up Track Info "
      "Playlist"           "plSettings"        "- Adjust the fields and sortin in the playlist "
      "AutoPlaylist"       "APLSettings"       "- (Applies to multi-column table only)"
      "Auto Level"         "AutoLevelSettings" ""
      "ID3 Tags"           "id3Settings"       ""
      "Folder Settings"    "folderSettings"    ""
      "Database"           "dbSettings"        ""
      "Colors/Fonts"       "schemeSettings"    ""
      "Web Server"         "webServer"         ""
      "Web User Interface" "webUI"             ""
      "Streaming Audio"    "streamingSettings" ""
      "Miscellaneous"      "AdvancedSettings"  ""
   }
   # Windows only
   if {$::tcl_platform(platform) == "windows"} {
      lappend options "File Associations" "associationSettings" ""
   }

   foreach {item callback help} $options {
      lappend entries $item
   }
   # Save current settings for possible restore
   array set ::oldSnackAmpSettings [array get snackAmpSettings]
   ##
   # Create a tabbed option frame
   #
   set tof [tabbedOptionFrame::Create $B.tf -options $entries -title "Select settings"]
   pack $tof -expand true -fill both
   set i 0
   foreach {item callback help} $options {
 #     update
      set hull [tabbedOptionFrame::getHull $tof $i]
      # Create metawidget to pack into hull for this tab
      set tab [$callback $hull]
      pack $hull
      incr i
   }

   # Line
   pack [frame $B.[incr i] -height 2 -borderwidth 2 -relief sunken] -expand true -fill x
   # Button frame
   set bf [frame $B.[incr i]]
   pack $bf -side bottom -fill none -expand false
   #  ----------------------------------------------------------------------------
   #  Buttons at the bottom
   #  ----------------------------------------------------------------------------

   # Apply
   set app [button $bf.[incr i] -text "Apply" -padx 4\
      -font listbox -default normal -borderwidth 1 \
      -command {settings::applySettings} -relief raised \
      -compound left  -image saImages.Save]
   balloon::define $app "Push this apply and save the current settings"

   # Apply and Close
   set okBut [button $bf.[incr i] -text "OK"  -padx 4\
      -font listbox -default active -borderwidth 1 \
      -command {settings::closeWindow 0} -relief raised \
      -compound left -image saImages.Apply]
   bind $B <Return> [list $bf.$i invoke]
   balloon::define $okBut "Push this to apply and save the current settings then close this dialog"

   # Cancel, restore settings
   set canBut [button $bf.[incr i] -text "Cancel" -default normal -font listbox -borderwidth 1 \
      -command {settings::closeWindow 1} -relief raised -compound left -image saImages.Close]
   bind $B <Escape> [list $bf.$i invoke]
   pack $app $okBut $canBut -side left -padx 20 -pady 5
   balloon::define $canBut "Push this to ignore the changes made and restore the prior settings."
   catch {wm transient $B .saTop}
   centerWindow $B
   # Lock down the minimum size of each tab (do after center to prevent flash)
   tabbedOptionFrame::lock $tof
   myIcon $B
   # Lock the overall minimum size
   lockSize $B
}
#------------------------------------------------------------------------------
# Playlist settings
#------------------------------------------------------------------------------
proc settings::plSettings {W} {
   ##
   #
   #
   set optionData {
      spinner2 ListDepth                {"Short Name Depth" {0 5 1}}
       "Adjusts the number of folder names that are prepended to the track names in the playlist when displaying short names. Useful if your files are named with only track title, or possibly track title and album name and you organize them by artist. Set to 0 if you only want the file name to show up. Ignored if showing Tag data."
      spinner2 playListLines            {"Playlist Lines" {5 50 1}}
       "Adjust this to the number of lines you want to initially show in the playlist window when SnackAmp starts."
      checkbutton restorePlayList         "Restore Prior Playlist"
       "Check this if you want to automatically restore the entire playlist and location from the previous session."
      checkbutton restorePlaying              "Restore Play Location on Playlists"
       "Check this to be prompted to resume playing at the track when the playlist was last saved"

      checkbutton moveToTop              "Move Track to Top after Sort"
       "Check this to move the playing track to the top of the playlist after a filtering or sort operation"
      checkbutton noTail              "Hide File Extension in Playlist"
       "Check this to hide (or uncheck to show) the file extension in the main Playlist (requires a restart)"
      checkbutton catPlaylist         "Catalog Tracks as needed"
       "Check this to Automatically catalog tracks in the playlist when needed. Useful to show track time."
   }
   pack [subFrame $optionData "Playlist Settings" $W.plsett] -side top -anchor w -fill x


   set hull [labeledFrame $W.sp -text "Playlist Fields" -bd 2 -relief groove -font listbox]
   checkbutton $hull.uas -text "Pad Separator" -variable ::snackAmpSettings(spcharSpace) \
         -pady 0 -highlightthickness 0 \
         -variable ::snackAmpSettings(spcharSpace) \
         -anchor w -font listbox  -font listbox
   listOptionMenu $hull.ps ::snackAmpSettings(spcar) [list "" "/" "-" "\\" "" ""] 2

   # hardcode hack
   label $hull.lps -text "Field Separator" -anchor e
   set fields [list None Duration Artist Title Album Track Genre ShortName File Rating]
   foreach num {1 2 3 4} {
      listOptionMenu $hull.mb$num ::snackAmpSettings(pl$num) $fields
   }
   button $hull.layout -text "Apply" -font listbox -command {
      if {$snackAmpSettings(spcharSpace)} {
         set TA(spcar) " $snackAmpSettings(spcar) "
      } else {
         set TA(spcar) "$snackAmpSettings(spcar)"
      }
      LB::updateFields
      LB::relayout
   } -relief raised -padx 4 -pady 0
   balloon::define $hull.layout "Push this change the playlist layout (Apply settings below first)"

   grid $hull.lps $hull.ps $hull.uas $hull.layout
   grid $hull.mb1 $hull.mb2 $hull.mb3 $hull.mb4
   pack $W.sp -side top -anchor w -fill x
   package require subSorting
   pack [subSort::create $W] -fill x -expand true -anchor nw

}
#------------------------------------------------------------------------------
# Function   : settings::playSettings
# Description: Settings page Misc functions
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc settings::playSettings {W} {
   set optionData {
      checkbutton ContinuousPlay          "Default of Continuous Play"
       "Check if you want to continuously play the playlist, uncheck if you want to stop after last track in playlist"
      checkbutton NextAfterCurrnt         "Next Track after Playing Track"
       "Check this to play the next song after the playing song rather than after the current selection."

      checkbutton resumePlay              "Resume Play on Restart"
       "Check this to have SnackAmp resume play at the point the application exited (if a track was playing and restore prior playlist is set)."

      checkbutton AutoDelete              "Auto Delete Invalid Files"
       "Check this to automatically remove a file if it detected as invalid or corrupt (be careful with this setting, the file will be permanently removed!)"

      checkbutton autoTruncate               "Remove Silence"
       "Check this to remove silence at the beginning and ending of your tracks. Requires Auto Leveling enabled."
   }

   pack [subFrame $optionData "Track Settings" $W.settings] -side top -anchor w -fill x
   #stuffSettingData $optionData $W
   ##
   # Play Mode Settings (note the order)
   #
   set optionData {
      scale2 normalPlayLatency    {"Buffer Time (ms)" {500 5000 200}}
       "Adjusts the snack audio buffer time, increase if you find audio choppy. 2000 ms is typical."
      scale2 trackGap             {"Track Gap Time (ms)" {0 5000 100}}
       "Additional time to delay between tracks. Typically used with gapless playback feature to insert silence between tracks."
   }
   pack [subFrame $optionData "Time Settings" $W.sc] -side top -anchor w -fill x


   set hull [labeledFrame $W.pm -text "Play Mode" -bd 2 -relief groove -font listbox]
   set i 0
   foreach {opt text} {
      "Next Track Plays"     "Play the next track in the playlist (normal play)"
      "Play Track Once"      "Play current track once and then stop"
      "Repeat Play of Track" "Repeat play of current track"
   } {
      pack [radiobutton $hull.$i -text $opt -value $i -font listbox -variable ::snackAmpSettings(SongPlayMode)] \
         -side left
      balloon::define $hull.$i $text
      incr i
   }
   #grid $W.pm -columnspan 4 -sticky ew
   pack $W.pm -side top -anchor w -fill x


}
#----------------------------------------------------------------------------
# Adjustments for cover art and pop-up info
#----------------------------------------------------------------------------
proc settings::caSettings {W} {
   variable taskBarIconList
#      checkbutton showCoverArt        "Show Cover Art"
#       "Check this to show cover art for the playing track (if it exists in the folder)."
   set optionData {
      spinner2 desArtSize             {"Desired Cover Art Height" {50 1000 50}}
      "Enter the desired cover art size for the cover art/album track window and APL cover art"
      spinner2 sliderArtSize             {"Pop-up Cover Art Height" {25 1000 25}}
      "Enter the minimum cover art height that will be displayed, larger images will be scaled down"

      spinner2 minArtSize            {"Minimum Cover Art Height" {50 600 50}}
      "Enter the minimum cover art height that will be displayed, smaller images will be scaled up.\nOnly used if image resizer library not available."
      spinner2 maxArtSize             {"Maximum Cover Art Height" {100 2000 50}}
      "Enter the minimum cover art height that will be displayed, larger images will be scaled down.\nOnly used if image resizer library not available."

      checkbutton smallArt           "Cover art in player window"
       "Check this to show a thumbnail cover art in the player window. (restart required)"
      checkbutton miniArt           "Cover art in mini-player window"
       "Check this to show cover art in the mini-player window."
      checkbutton leftSmallArt       "Cover art on left of player window."
       "Check to show cover art larger on left of main player. (restart required)"
   }

   pack [subFrame $optionData "Cover Art Display" $W.visual] -side top -anchor w -fill x

   set optionData {
      checkbutton amazonCoverArt       "Search Amazon Only"
       "Restrict the album cover search to Amazon only."
      checkbutton jpegCoverArt          "Restrict to jpeg"
       "Restrict the album cover search to jpeg images only."

      checkbutton walmartCoverArt       "Include Walmart"
       "Include large Walmart images in search."
      x x x x
      x x x x

      spinner2 rowsCoverArt             {"Show Columns" {1 10 1}}
       "Set the number of image columns to show per row in the web browser."
      spinner2 sizeCoverArt             {"Display size" {0 1000 50}}
       "Set the display size of the returned cover art, set to 0 for native resolution."

      spinner2 maxCoverFind             {"Images Searched" {20 100 10}}
       "Set the number of images to search within the Google query, less is faster."
      spinner2 maxCoverShow             {"Images Shown" {1 100 1}}
       "Set the number of images to show from the Google query, less is faster."

      spinner2 maxCoverWidth             {"Maximum Width" {0 2000 50}}
       "Reject all images with pixel width greater than this."
      spinner2 minCoverWidth             {"Minumum Width" {0 2000 50}}
       "Reject all images with pixel width Less than this."

      spinner2 ratioThreshold           {"Image Squarness" {1.0 1.5 .1}}
       "Select the ratio of height to width or width to height to restrict images that don't look like cover art."
      x x x x
      x x x x

      entry coverart                "Names for Cover Art Images"
       "Enter a list of names, in preference order, that can be used to locate cover art images located in each folder, wildcards may be used."
   }
   pack [subFrame $optionData "Cover Art Searching" $W.search] -side top -anchor w -fill x

   set optionData {
      checkbutton fadePopup           "Pop-up Track Info"
       "Check this to show a small window in lower right hand corner of desktop momentarily showing track information when a track starts playing."

      checkbutton slideCoverArt       "Cover Art with Pop-up"
       "Check this to include cover art (if available) on the Pop-up Track Info window."

      option2 sliderPosition
      {{Pop-up Initial Position} {{Left Top} {Left Bottom} {Right Top} {Right Bottom} {Top Left} {Top Right} {Bottom Left} {Bottom Right}}}
      "Select where you want the Pop-up Track Info window to appear each track"

      spinner2 sliderDelay             {"Pop-up Delay (ms)" {1000 20000 500}}
       "Set the amount of time the Pop-up Track Info will be displayed (in milli-seconds)."
      spinner2 sliderAlpha             {"Pop-up Opacity" {0.1 1.0 0.05}}
       "Set opacity of the Pop-up Track Info. 0.0 = fully transparent (Windows only)."
      spinner2 sliderSpeed             {"Pop-up Speed" {1 40 1}}
       "Set the speed of the sliding Pop-up Track Info, higher number is faster"
   }
   pack [subFrame $optionData "Track Info Pop-up" $W.adj] -side top -anchor w -fill x
}
#------------------------------------------------------------------------------
# Function   : settings::uiSettings
# Description: User Interface Settings
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
#      checkbutton showTicks           "Show Ticks on Volume"
#       "Check this to show volume level tickmarks."
proc settings::uiSettings {W} {
   variable taskBarIconList
   set optionData {
      checkbutton ID3ballon          "Balloon Help Track Details"
       "Check this to have detailed ID3 and file information pop-up when hovering the mouse cursor over a track in a playlist."
      checkbutton showLabels          "Show Labels"
       "Check this to display the title next to each label on the main interface."

      checkbutton globalMixer           "Control System Volume"
       "Check this to have have the volumn control the system volume, uncheck for just SnackAmp."
       checkbutton linscale              "Log Mixer"
        "Check this if your systems mixer receives logarithmic inputs (some Linux systems need this on). Affects how the volume sliders scale the levels. Only applys if using global mixer."

      checkbutton restoreMiniPlayer   "Restore Mini-player Position"
       "Check if you want restore the Mini-player position and state upon startup."
      checkbutton levelMeter          "Display Sound Level Meter"
       "Check if you want to show a sound level meter on the front (requires restart)."

      checkbutton fadeImages          "Fade in/out Images"
       "Check this to fade in/out balloon help, miniplayer and album art (Platform Specific, Needs Tcl 8.4.8 or greater)."
      checkbutton restorePlayerPos    "Restore Player Position"
       "Check if you want restore the Main Player position and state upon startup."

      checkbutton showMS    "Show Milliseconds"
       "Check if you want to show 1/10s of a second on the time counter display."
      checkbutton minSysTray    "Minimize to System Tray"
       "Check if you want to minimize SnackAmp to the system tray (Needs system tray support)."
      checkbutton fancyMenu    "No Window Decoration"
       "Check if you want hide the windows decoration (restart required)."
   }

   pack [subFrame $optionData "Visual Settings" $W.visual] -side top -anchor w -fill x

   set optionData {
      spinner2 meterUpdateRate             {"Meter Update Rate (ms)" {10 500 10}}
       "Set the time the audio meter is updated."
      spinner2 balloonDelay                {"Balloon Help Delay (ms)" {100 5000 100}}
       "Adjust this to set the amount of delay before a pop-up balloon is shown."
      spinner2 balloonDuration             {"Balloon Duration (ms)" {500 10000 200}}
       "Adjust this to set the amount time a balloon is shown before it disappears."
      spinner2 FontSizeFactor              {"Font Size Factor" {.5 1.5 .05}}
       "Adjust the overall font size scaling to your system. Recommend you leave at 1 and adjust the fonts individually for your system."
      spinner2 timeSliderHeight            {"Time Slider Height" {1 50 1}}
       "Adjust the height of the time slider."
      spinner2 timeSliderWidth             {"Time Slider Width" {1 50 1}}
       "Adjust the width of the time slider."
      spinner2 volSliderHeight             {"Volume Slider Height" {1 50 1}}
       "Adjust the height of the volume slider."
      spinner2 volSliderWidth              {"Volume Slider Width" {1 50 1}}
      "Adjust the width of the volume slider."
   }
   pack [subFrame $optionData "Visual Adjustments" $W.adj] -side top -anchor w -fill x

   ##
   # Display Info (note the order)
   #
   set hull [labeledFrame $W.info -text "Track Display Details" -bd 2 -relief groove -font listbox]
   set i 1  ;# starts at 1
   foreach {opt text} {
     "Title Only"              "Show only the Track Title on the main display, RMB on title to show menus (compact form)"
     "+ Art., Alb. & Rate"     "Show only the Track Title, Artist, Album & Rate"
     "+ Year, Track & Genre"   "Show only the Track Title, Artist, Album, Rate, Year, Track and Genre"
     "All Info"                "Show all of the information on the main display (full form)"
   } {
      pack [radiobutton $hull.$i -text $opt -value $i -font listbox -variable ::snackAmpSettings(showDetail)] \
         -side left
      balloon::define $hull.$i $text
      incr i
   }
   set e 0
   pack $W.info -side top -anchor w -fill x
   if {$::tcl_platform(platform) eq "windows" || $::tcl_platform(os) eq "Linux"} {
      set hull [labeledFrame $W.pm -text "Taskbar Icons to Use" -bd 2 -relief groove -font listbox]
      foreach {item} $taskBarIconList {
         lappend gr [checkbutton $hull.[incr e] \
            -pady 0 -highlightthickness 0 \
            -variable ::snackAmpSettings(icon,$item) \
            -anchor w -text $item -font listbox]
         balloon::define $hull.$e "Check to show $item taskbar"
         if {[llength $gr]>=8} {
            eval grid $gr -sticky w
            set gr [list]
         }
      }
      if {[llength $gr]>=1} {
         eval grid $gr -sticky w
      }
      pack $W.pm -side top -anchor w -fill x
   }

}
#------------------------------------------------------------------------------
# Function   : settings::folderSettings
# Description: Settings for the Folders
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
#      entrydir uskDir                     "Folder to hold default Streaming Shortcuts"
#       "Select a folder to place Streaming Shortcuts when playing streams from the web (e.g. shoutcast). You should place it within your current media folders to manage it from the AutoPlaylist Manager"
#      checkbutton MonitorCacheFolders     "Monitor Catalogs for Changes"
#       "Check this to continuously monitor for changes to the catalog files (i.e. from other users) and read them in if they change."
#       entryList mediaDirs                     "Media Folders"
#        "Enter a list of music library folders separated by spaces to catalog. Each folder and its sub folders will be cataloged. Each base folder will contain a catalog file.\ne.g. d:/music \"//george/music library\""
#       entryList IncomingDirs                  "Monitor Folders"
#        "Enter a list of music library folders separated by spaces to monitor for changes (e.g. CD ripping location). New files will be automatically appended to the playlist. Sub-folders are not monitored."
proc settings::folderSettings {W} {
   set i 0

   set optionData {
      button      {folderManager::launch IncomingDirs}  "Manage Folders" "Press to manage the folders to monitor for changes"
      spinner2    ScanFolderDepth   {"Monitor Depth" {0 10 1}}
      x x x x
      "Enter the number of folder levels to check for changed files beneath the monitor folder(s)"
      checkbutton MonitorScanFolders      "Monitor Folders for New Files"
       "Check if you want to combine the indexing and cataloging steps. Simplifies things but indexing will be slower or large collections."
   }
   pack [subFrame $optionData "Monitor Folders" $W.[incr i]] -side top -anchor w -fill x

   set optionData {
      entry fileExtensions                "Supported File Extensions"
       "Enter a list of file extenstions (with the . before) to support when cataloging files. Upper or lowercase is acceptable."
      entrybrowse mp3gainLoc                "mp3gain program"
       "Locate your mp3gain program if you want to be able to launch it from within the Autoplaylist manager."
      entrybrowse  hotkeyFile             "Hotkey Assignment File"
       "Enter the name of file containing hotkey assignments (you must restart after changing)."
   }
   pack [subFrame $optionData "Other Settings" $W.[incr i]] -side top -anchor w -fill x
   ##
   # List of short words
   #
   set hull [labeledFrame $W.words -text "Short Words" -bd 2 -relief groove -font listbox]
   pack [entry $hull.sw -textvariable ::snackAmpSettings(shortWords)] -fill both -expand true
   balloon::define $hull.sw "These words will be keep as lower case when automatically correcting track/file names"
   pack $W.words -side top -anchor w -fill x


}
#------------------------------------------------------------------------------
# Function   : settings::id3Settings
# Description: Settings for the Folders
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc settings::id3Settings {W} {
   set i 0
   set optionData {
      entry2 smartName                 "Format for Smart Naming"
      "Enter a naming scheme for converting tags to file names:\n%%A - Artist, %%C - Album, %%N - Track Number, %%T - Title"
      spinner2 ID3editLimit                {"Group Edit Track Limit" {1 1000 25}}
      "Enter the maximum number of tracks that be edited with the ID3 Tag editors."
      entry2 splitChar                 "Name Separator"
      "Enter character used when cleaning up or building a file name from tags or cleaning up file names, recommend using -"
      entry2 scrolledCanvasHeight      "Editor Window Initial Height"
      "Enter a number (in inches) followed by an i e.g. 4i to denote how high the tag editor window should be when editing a large number of files. Any additional space needed will be scrollable."
   }
   pack [subFrame $optionData "ID3 Editor" $W.[incr i]] -side top -anchor w -fill x
   set optionData {
      checkbutton ID3OKCreate         "Group Editing will Create Tags"
       "Check this to allow the ID3 Group Editor to create tags when editing multiple files. With this off, files without tags are ignored."
      checkbutton preserveTime         "Preserve Time Stamp on Edit"
       "Check this to not change the file time stamp when changing the ID3 tag. Recommended you left this un-checked."
      checkbutton ID3ReadOnly          "Disallow ID3 Tag Modification"
       "Check this inhibit modification of ID3 Tag information."
      checkbutton ID3V12               "Enable ID3 V1.2 Extension"
       "Check this to allow V1.2 Tags to be written if V1.X writing is enabled (Compatible extension to V1 Tags)."
      checkbutton preserveV2data       "Preserve Existing V2 Data"
       "Check this to maintain existing ID3V2.x tags not being modified by SnackAmp."
      checkbutton zeroPadTrack      "Zero Pad Tracks"
       "Put a zero in front of track numbers when editing ID3 tags"
      option2 ID3V1                   {"Write ID3 V1.X Tags" {"Never" "Always" "If Existing" "If New"}}
       "Check this to allow V1.X Tags to be written along with the V2.X tags."
      option2 ID3ReadOrder            {"Tag Read Order" {"V1 V2 OGG" "V2 V1 OGG"}}
       "Specify the order in which tags are read to populate the database"
      option2 ID3V2                   {"Write ID3 V2.3 Tags" {"Never" "Always" "If Existing" "If New"}}
       "Specify how V2.3 Tags are to be written along with the V1.X tags."
      option2 defaultEncoding         {"ID3 V2.3 Tag write Unicode" {"Never" "Always" "Only If Needed"}}
       "Select when to use unicode with writing ID3 V2 tags, (not all players support unicode yet)."
   }
   pack [subFrame $optionData "ID3 Tags" $W.[incr i]] -side top -anchor w -fill x
   set optionData {
      option freedb         {"Query freedb server" {
          http://at.freedb.org:80/~cddb/cddb.cgi
          http://au.freedb.org:80/~cddb/cddb.cgi
          http://ca.freedb.org:80/~cddb/cddb.cgi
          http://ca2.freedb.org:80/~cddb/cddb.cgi
          http://de2.freedb.org:80/~cddb/cddb.cgi
          http://es.freedb.org:80/~cddb/cddb.cgi
          http://fi.freedb.org:80/~cddb/cddb.cgi
          http://ru.freedb.org:80/~cddb/cddb.cgi
          http://uk.freedb.org:80/~cddb/cddb.cgi
          http://us.freedb.org:80/~cddb/cddb.cgi
      }}
       "Select the freedb server to use for database queries."
      entry cddbBase "HTML freedb server"
       "Select the freedb server to use for html queries."
   }
   pack [subFrame $optionData "Freedb Options" $W.[incr i]] -side top -anchor w -fill x
   ##
   # Custom genres
   #
   set hull [labeledFrame $W.genre -text "Custom Genres" -bd 2 -relief groove -font listbox]
   set ::TA(text,genre) $hull.genres
   pack [text $hull.genres -height 4] -fill both -expand true
   $hull.genres insert end $::snackAmpSettings(customGenres)
   pack $W.genre  -side top -anchor w -fill x
   bind $hull.genres <Return> {
      $::TA(text,genre) insert insert \n
      break
   }
   balloon::define $hull.genres "Enter custom genres here, use quotes for multi-word entries"
   settings::regColor $hull.genres -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
}
#------------------------------------------------------------------------------
# Apply the new settings
#------------------------------------------------------------------------------
proc settings::applySettings {args} {
   global snackAmpSettings TA
   settings::saFontSize $snackAmpSettings(FontSizeFactor)
   settings::updateShortWords
   settings::saveSettings
   array set ::oldSnackAmpSettings [array get ::snackAmpSettings]
   catch {set snackAmpSettings(HttpLocal) [file normalize $snackAmpSettings(HttpLocal)]}   ;# TFW: 8.4 unique
   set snackAmpSettings(HttpLocal) [string trim $snackAmpSettings(HttpLocal)]

   if {$snackAmpSettings(spcharSpace)} {
      set TA(spcar) " $snackAmpSettings(spcar) "
   } else {
      set TA(spcar) "$snackAmpSettings(spcar)"
   }
   # pull from text box
   set snackAmpSettings(customGenres) [string trim [$::TA(text,genre) get 1.0 end]]
   ##
   # Listbox fields
   #
   LB::updateFields
   catch {set snackAmpSettings(HttpLocal) [file normalize $snackAmpSettings(HttpLocal)]}  ;# TFW: 8.4 unique
   settings::updateShortWords
   settings::saveSettings
}
#------------------------------------------------------------------------------
# Function   : settings::closeWindow
# Description: When closing the settings gui either restore the old
#              settings or save the new ones.
# Author     : Tom Wilkason
# Date       : 7/6/2003
#------------------------------------------------------------------------------
proc settings::closeWindow {cancel} {
   global snackAmpSettings oldSnackAmpSettings TA
   if {[winfo exists .tsg]} {
      wm withdraw .tsg
   }
   if {!$cancel} {
      set snackAmpSettings(customGenres) [string trim [$::TA(text,genre) get 1.0 end]]
   }
   foreach {key} [array names snackAmpSettings] {
      # Only process the settings if something changed.
      if {![info exists oldSnackAmpSettings($key)] || ($snackAmpSettings($key) ne $oldSnackAmpSettings($key))} {
         if {!$cancel} {
            ##
            # Save
            #
# Need an LB function to just replace each line
#             if {[winfo exists .saTop] && ($::oldSnackAmpSettings(ListDepth) != $snackAmpSettings(ListDepth))} {
#                macros::filterMusic replace "all"
#             }
            if {$key=="FontSizeFactor"} {
               settings::saFontSize $snackAmpSettings(FontSizeFactor)
            }
            settings::applySettings
            set snackAmpSettings(currentColorScheme) $schemeSettings::schemeSelection
            break
         } else {
            ##
            # Cancel (only bring over the changed values to reduce unneeded traces)
            #
            foreach {index value} [array get oldSnackAmpSettings] {
               if {$snackAmpSettings($index) != $value} {
                  set snackAmpSettings($index) $value
                  switch -- $index {
                     FontSizeFactor {
                        settings::saFontSize $snackAmpSettings(FontSizeFactor)
                     }
                     shortWords {
                        settings::updateShortWords
                     }
                     HttpLocal {
                        catch {
                           set snackAmpSettings(HttpLocal) [file normalize $snackAmpSettings(HttpLocal)]  ;# TFW: 8.4 unique
                        }
                     }
                  }
               }
            }
            Update
            # Restore the old settings
            settings::installFontCallbacks
            settings::applyColor
            break
         }
      }
   }
   destroy .tsg
}
##
# Configure the metakit settings
#
proc settings::dbSettings {W} {
   ##
   # Sub-sorting radio buttons
   #
   set i 0
   ;# Metakit
#      checkbutton fastscan        "Fast Folder Index"
#       "Press this to index files 2X faster, but won't see folders ending in same file extensions as sound files."
   set optionData {
      button      {folderManager::launch mediaDirs}  "Manage Media" "Press to manage the media folders"
      button      {folderManager::launch mp3Player}  "Manage Players" "Press to manage the folders on mountable media players"
      x x x x
      x x x x
      checkbutton AutoRescan              "Index Media Folders on Startup"
       "Check if you want to automatically re-index your folders on startup. Not recommended for large media sets."
      checkbutton dontMoveFiles              "Inhibit File Modifications"
       "Check if you want to inhibit moving, removing or renaming files or folders."
      checkbutton alwaysCatalog              "Index & Catalog tracks together"
       "Check if you want to combine the indexing and cataloging steps. Simplifies things but indexing will be slower or large collections."
   }
   pack [subFrame $optionData "Media Folders" $W.[incr i]] -side top -anchor w -fill x

   set optionData {
     label indexing "Index Status"
        "Current background index state"
     label lastIndexFormatted "Last Index Completed"
        "Date/Time for the last index completed"
     label nextIndexFormatted "Next Index Scheduled"
        "Date/Time for the next index"
     checkbutton indexBackground "Background Index"
        "Current background index state"
     checkbutton bgCatalog "Background Catalog"
        "In addition to Indexing, update the full catalog data for each track in the background"
     spinner2 indexHours {"Index Every" {1 240 1}}
        "Enter a time in hours between background indexing to update the database automatically."
     option2 indexUnits {"Units" {"Minutes" "Hours" "Days" "Weeks"}}
        "Every units of time."
     entry2   indexStart "Start Index After"
        "Enter a time (03:00) in 24 hour time format to allow background indexing to start. Enter 0 to allow any time"
     entry2   indexStop "Stop Index Before"
        "Enter a time (03:00) in 24 hour time format to prevent indexing from starting. Enter 0 to allow any time"
     button {after 0 catalog::bgIndex}  "Index Now"
        "Background Index the database immediately"
   }
   pack [subFrame $optionData "Background Indexing" $W.[incr i]] -side top -anchor w -fill x
   # Other
   set optionData {
     button {db::cleanDatabase}  "Check Database"
        "Verify the entries in the database."
     button {db::restart}          "Compact Database"
        "Resort and pack the database."
   }
   pack [subFrame $optionData "Maintenance" $W.[incr i]] -side top -anchor w -fill x

}
# ----------------------------------------------------------------------------
# Various color settings
# ----------------------------------------------------------------------------
proc settings::colorSettings {W} {
   set changeColors {
      "Text Background"          bg
      "Text Foreground"          fg
      "Selected Text Background" highlight
      "Table Stripe Background"  tsbg
      "Entry Background"         ebg
      "Entry Foreground"         efg
      "Title Text Foreground"    tfg
      "Artist Text Foreground"   afg
      "Info Label Foreground"    lfg
      "Time Counter Foreground"  sfg
   }
   set i 0
   set tg [list]
   foreach {Text Index} $changeColors {
      button $W.$Index -text $Text -font listbox \
         -relief raised -borderwidth 1 -command [list settings::colorPicker $Index]
      balloon::define $W.$Index "Push this to set the $Text Color"
      button $W.lbl$Index -width 5 -bd 1 -relief raised \
         -bg $::snackAmpSettings($Index) -command [list settings::colorMenu $W.lbl$Index $Index]
      regColor $W.lbl$Index -background ::snackAmpSettings($Index)
      lappend tg $W.$Index $W.lbl$Index
      if {[llength $tg]>=4} {
         eval grid $tg -sticky ew -padx 2 -pady 0
         set tg [list]
      }
   }
   if {[llength $tg]} {
      eval grid $tg -sticky ew -padx 2 -pady 0
   }
}
#------------------------------------------------------------------------------
# Function   : settings::fontSettings
# Description: Settings page for Fonts, uses a font dialog
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc settings::fontSettings {W} {
   variable FONTS
   global snackAmpSettings
   set i 0
   set span 5
   set togrid [list]
   foreach {Name Specs Description} $FONTS {
      button $W.bt$Name -text "$Description" -relief raised -font listbox\
         -command [list settings::setfont $Name $W.bt$Name]
      balloon::define $W.bt$Name "$Specs"

      label $W.lb$Name -text "Sample Text" -relief flat -font $Name -anchor w
      lappend togrid $W.bt$Name $W.lb$Name
      if {[llength $togrid]==4} {
         eval grid $togrid -sticky ew  -padx 2 -pady 0
         set togrid [list]
      }
   }
   if {[llength $togrid]} {
      eval grid $togrid -sticky ew  -padx 2 -pady 0
   }
   return
}
# ----------------------------------------------------------------------------
# Bring up a dialog to modify the font for some fontname
# ----------------------------------------------------------------------------
proc settings::setfont {fontname widget} {
   global snackAmpSettings
   set tl .tsg
   if {[lsearch -exact [font names] $fontname] >= 0} {
      array set fa [font configure $fontname]
      set initfont [list $fa(-family) $fa(-size) $fa(-weight) $fa(-slant)]
      set font [SelectFont .fontdlg \
         -type dialog -parent $tl -font $fontname -title "Setting $fontname"]
      if {[string length $font]} {
         array set fa [font actual $font]
         foreach item {family size weight slant} {
            set snackAmpSettings($fontname,$item) $fa(-$item)
         }
         set initfont [list $fa(-family) $fa(-size) $fa(-weight) $fa(-slant)]
         balloon::define $widget "$initfont"
         settings::saFontSize $snackAmpSettings(FontSizeFactor)
      }
   } else {
      saLog "$fontname is an invalid font name"
   }
}

#------------------------------------------------------------------------------
# Function   : settings::schemeSettings
# Description: Settings page for Colors
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc settings::schemeSettings {W} {
   # Colors
   set fr [labeledFrame $W.colors -text "Colors" -bd 2 -relief groove -font listbox]
   grid $W.colors -sticky ew
   settings::colorSettings $fr
   # Fonts
   set fr [labeledFrame $W.fonts -text "Fonts" -bd 2 -relief groove -font listbox]
   grid $W.fonts -sticky ew
   settings::fontSettings $fr
   # Divider Line
   grid [frame $W.line -height 2 -borderwidth 2 -relief sunken] \
      -pady 2 -sticky ew

   # Button frame
   set bf [frame $W.bf]
   grid $bf -sticky ew -pady 0
   schemeSettings::fetch
   set schemeSettings::schemeSelection $::snackAmpSettings(currentColorScheme)
   button $bf.save   -relief raised -text "Save Scheme"  \
      -command {schemeSettings::save}   -font listbox  -compound left -image saImages.Save
   button $bf.saveas -relief raised -text "Save as..."  \
      -command {schemeSettings::saveas}  -font listbox -compound left -image saImages.SaveAs
   button $bf.remove -relief raised -text "Remove"  \
      -command {schemeSettings::remove}  -font listbox -compound left -image saImages.delete
   menubutton $bf.opt -anchor w -direction flush -width 18 -font listbox\
      -indicatoron 1 -menu $bf.opt.menu -textvariable schemeSettings::schemeSelection
   menu $bf.opt.menu \
      -postcommand [list schemeSettings::populate $bf.opt.menu]
   grid $bf.save $bf.saveas $bf.remove $bf.opt -padx 4
}
#------------------------------------------------------------------------------
# Function   : settings::webServer
# Description: Settings page Misc functions
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
#      entry2 httpMaxused             "Max Number of Requests"
#       "Specify the maximum number of simultaneous requests per client (~200 works well)."
proc settings::webServer {W} {
   set i 0
   set optionData {
      entry2 port                     "Initial Http Server Port Number"
       "Enter a port number to use for the integrated web server. This should be a free port on your system."
      label2 usedPort                  "Used Http Server Port Number"
       "Actual port used by internal web server, will be next free port at or after above port."
      label2 HttpHits                 "Number of HTTP connects this session"
       "Enter the maximum number of simultaneous clients."
      spinner2 bufSize             {"Buffer Size" {1024 32768 1024}}
       "Enter the buffer size for each client write for web server (in bytes)."
      checkbutton EnableWebServer "Enable Internal Web Server"
       "Check this to allow a web browser to remote control this player or show internal help (requires restart)."
      checkbutton EnableRemoteControl "Enable Remote Control via Browser"
       "Check this to allow a web browser to remote control this player or show internal help (requires restart)."
      entry HttpUsers                 "Users allowed to access webserver"
       "If you want to restrict access to your www administration panel, enter user:password pairs here."
   }
   pack [subFrame $optionData "Server Settings" $W.[incr i]] -side top -anchor w -fill x
   set optionData {
      entry subNetMask               "Subnet Mask for Local Addresses"
       "Enter your Subnet Mask to validate local vs. remote clients, used for checking local only access."
      label ipAddress               "Your detected IP Address"
       "IP Address of this node (if 0 no ports were available to check)."
      checkbutton localOnly           "Local Web Access Only"
       "Check this to only allow access from clients on your local network (those within your subnet). Requires Internal Web Server to be enabled."
   }
   pack [subFrame $optionData "Local Access Settings" $W.[incr i]] -side top -anchor w -fill x
#      entry HttpDefault               "Browser Control File Name"
#       "File name to use for main play control page. Do not specify a folder, it must be located in a 'docs' folder under the location of SnackAmp."
   set optionData {
      entry HttpLocal                 "Local Folder to map /local to"
       "If desired, you can enter a folder here on your lan that the url /local will map to. This allows you to extend SnackAmp with your own HTML pages."
      entry cssName                   "Cascading Style Sheet File"
       "Location of Cascading Style Sheet used for dynamic web content."
      label internalWebBase           "URL of Internal Web Server"
       "URL of the Internal Web Server. Enter this value in your browser for remote control."
      label webRoot                   "Root Folder of Internal Web Server"
       "Root Folder of the Internal Web Server."
   }
   if {$::tcl_platform(platform)=="unix"} {
      lappend optionData \
      entrybrowse HelpBrowser        "Help Browser" \
       "Program to use for browsing help pages. Leave empty to use system default"
   }
   pack [subFrame $optionData "URL Locations" $W.[incr i]] -side top -anchor w -fill x
#   set furtureOptionData {
#      entry allmusicBase             "Internet Query Location"
#       "Web URL to use for querying web data"
#   }
#      entry helpFile                 "Help File Name"
#       "File name to use for help."
}
proc settings::webUI {W} {
   set i 0
   set optionData {
      spinner2 httpNextLength           {"Number Of Queued Songs" {0 100 10}}
       "Specify the number of future playlist songs to show in the remote control web page."
      spinner2 refreshRate              {"Browser Refresh Rate" {5 60 5}}
       "Enter time in seconds the play control browser should refresh itself."
      spinner2 minNextTrack            {"Minimum Track Separation" {0 50 5}}
      "Minimum separation between the same track track when appending tracks from the web server (to prevent users from playing the same track(s) over and over)"
      x x x x
      x x x x
      checkbutton listbytag "List by Artist Track"
       "Check to list tracks by thier tag information. Uncheck to list by folder, file names."
      checkbutton cleanListings "Clean Up Browser Listings"
       "Check to show folder and file names in browser in proper case and remove underscores."
   }
   pack [subFrame $optionData "Track Listings" $W.[incr i]] -side top -anchor w -fill x
   # jukebox
   set optionData {
      checkbutton jukebox,rettofolder "Return to Same Browsing Folder"
       "Enable if you want to return to the same folder after queueing a track, otherwise will return to top level folders."
      checkbutton jukebox,linkartist "Add Browser Links to Artist/Folder"
       "Enable if you want to have hyperlinks added to the Artist/Folder names to browse directly to that track folder."
      checkbutton jukebox,popup "Show Pop-up Info"
       "Enable if you want to have an icon next to each track to pop-up additional info on the track."
   }
   pack [subFrame $optionData "JukeBox" $W.[incr i]] -side top -anchor w -fill x
   # Remote Control
   set optionData {
      checkbutton remote,rettofolder "Return to Same Browsing Folder"
       "Enable if you want to return to the same folder after queueing a track, otherwise will return to top level folders."
      checkbutton browseNowPlaying "Now Playing while Browsing"
       "Check to show the now playing information while browsing folders."
      checkbutton remote,linkartist "Add Browser Links to Artist/Folder"
       "Enable if you want to have hyperlinks added to the Artist/Folder names to browse directly to that track folder."
      checkbutton remote,popup "Show Pop-up Info"
       "Enable if you want to have an icon next to each track to pop-up additional info on the track."
   }
   pack [subFrame $optionData "Remote Control" $W.[incr i]] -side top -anchor w -fill x
   set optionData {
      spinner2 coversize         {"Playing Art Image Size" {25 600 25}}
       "Enter the size of the cover art, in pixels, that is displayed on a web browser for the playing track."
      spinner2 coverbrowsersize  {"Browser Art Image Size" {25 600 25}}
       "Enter the size of the cover art, in pixels, that is displayed on a web browser for folder being browsed."
      spinner2 artrowcount        {"Cover Art per Row" {1 8 1}}
      "Number of album covers to show in a row when browsing folders on web page"
      x x x x
      x x x x
      checkbutton remote,brcoverart "Show Coverart For Jukebox Browser"
       "Show Coverart for each folder listed while browsing, if disabled only show coverart for current folder."
      checkbutton jukebox,nocoverart "'No Cover Art' Image for no Cover Art"
       "When no cover art is available, show a stand image that says 'No Cover Art' as a placeholder, if unchecked show nothing. Requires Show Coverart to be enabled"
   }
   pack [subFrame $optionData "Cover Art" $W.[incr i]] -side top -anchor w -fill x
}
#------------------------------------------------------------------------------
# Function   : settings::AdvancedSettings
# Description: Settings page Advanced functions
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc settings::AutoLevelSettings {W} {
   set ::snackAmpSettings(xxx) ""

   set optionData {
      button {soundControl::calibrateAudio}        "Compute\nAuto Levels"
       "Iterate over each track and compute the level setting if needed."
      button {settings::resetGainData}        "Reset\nAuto Levels"
       "Press this to reset all auto level values."
   }

   lappend optionData button {db::resetPlayCount}  \
   "Reset\nPlay Counts" \
   "Press this to reset the play counter to 0 for each track."
   lappend optionData button {db::resetPlayTime}  \
   "Reset\nLast Play Time" \
   "Press this to reset the last played time to 0 for each track."
   pack [subFrame $optionData "Auto Level Functions" $W.f1] -side top -anchor w -fill x

   set optionData {
      scale2 maxGain                 {"Max Gain (in dB)" {0 6 .1}}
       "Adjust to limit the amount of gain applied during playback to quiet songs. 3 dB is typical. Applied at playback."
      scale2 minGain                 {"Max Atten (in dB)" {-15 0 .1}}
       "Adjust to limit the amount of attenuation applied to loud songs during playback. -10 dB is typical. Applied at playback."

      scale2 autoLevelReference      {"Reference Offset (dB)" {-6 6 1}}
       "Adjust the internal reference level used during computation. If number is too large, clipping may occur. If unsure leave at 0. Applied during leveling, if you change this setting, you should Auto Level your entire collection."
      scale2 autoLevelSeconds        {"Sample Period (sec)" {5 50 1}}
       "Adjusts period between samples for the Auto Level setting, smaller is slower but more accurate. 25 is typical, try 10 if you have a fast CPU. Used during leveling."

      scale2 autoLevelMaxSamples     {"Max Samples" {5 100 5}}
       "Enter a number to limit the maximum number of samples taken, larger numbers are more accurate but slower to compute. For example if Sample Period is 10 sec, and the track is 10 minutes long, 60 samples or this setting, whichever is less, will be taken. Used during leveling."
      scale2 autoLevelWindow         {"Window Size" {512 18000 1024}}
       "Adjusts the Window Size of the FFT for computing the Auto Level setting, larger is slower but more accurate. 4096 is typical, try 8192 or greater if you have a fast CPU. Used during leveling."

      scale2 gainFactor              {"Level Control Factor" {15 50 1}}
       "Adjusts the amount of level control applied during computation, 20-40 is typical, larger numbers produces a smaller effect (i.e. loud tracks stay somewhat loud, quiet tracks stay somewhat quiet). 20 is Theory but may not work as well as 33. Applied during leveling and used during playback."
      scale2 autoLevelInc            {"Gap Window Length" {0.01 .1 .01}}
       "Adjusts the sample window length, in seconds, to look for gaps. Smaller is more accurate but slower. 0.05 should work well. Used during gap detection."

      checkbutton autoLevel           "Enable Auto Leveling"
       "Check this to have SnackAmp adjust the audio level based on how loud the track is. Useful for mp3/ogg files encoded using different encoders. Recommend you enable this if different rippers encoded your tracks. Used during leveling and playback."
      checkbutton autoLevelPreComute  "Level queued tracks"
       "Check this to auto level the next track in the queue while the current track is playing. You may have to disable on slow computers to aviod an audible pause. Used during playback."

      checkbutton tagGainValues       "Add Gain to ID3 V2 tag"
       "Check this to insert the gain value into the replay_gain ID3V2 tag (used by some players)."
      checkbutton upgradeTags         "Upgrade V1 tags to V2 if needed"
       "Check this to insert the gain value into the replay_gain ID3V2 tag (used by some players)."

      option2 gainType                   {"Use Gain From" {"Track" "Album" "Album, Track if Shuffled"}}
       "Select which type of gain to use."
       x x x x

   }
   pack [subFrame $optionData "Auto Level Settings" $W.f2] -side top -anchor w -fill x
}
##
# Diagnostic Settings
#
proc settings::AdvancedSettings {W} {
   # Auto Level
   # Diagnostics
   set optionData {
      checkbutton Trace               "Enable Diagnostic Outputs"
       "Check this to print out diagnostic output, not normally needed."
      checkbutton logFile             "Log Activity to [file tail $TA(logFile)]"
       "Check this to direct all console output to a file"

      checkbutton autoSavePlaylist    "Save Playlist on Track Change"
       "If checked, the play list will be saved after every track, only useful if you are experiencing problems. The playlist is always saved on exit. Set the Restore Prior Playlist option to restore the playlist at startup"
      checkbutton enableConsole       "Enable use of Console"
       "If not checked, the console option will not be shown in the settings menu (requires restart)"

      checkbutton Version2Settings     "Use Clear Text Settings File"
       "Check this to save your settings in a user editable file format"

      checkbutton webLog              "Log Web Access"
       "Check this to log all web access to YYMMDD.log files in your home folder (requires restart)."
   }
   pack [subFrame $optionData "Diagnostic Settings" $W.tr] -side top -anchor w -fill x
   set optionData {
      checkbutton useIRC         "Enable Infrared Remote Control"
       "Check this to create a socket to enable Infrared remote controls to control SnackAmp. Requires a restart"
      checkbutton useAPI         "Enable Socket Interface"
       "Check this to create a socket to listen for remote socket control, similar to xmms_remote commands. Requires a restart"
      entry2 ircPort             "LIRC Port number"
       "TCP/IP Port number to use for Remote Control"
      entry2 apiPort            "Port for Socket API"
      "The last remote key received by SnackAmp"
      entry2 ircHost             "Host Name"
       "Enter host name of IRC controller, leave blank if this computer"
      entry2 ircEnable             "LIRC Enable Key"
       "Enter the remote keystroke to enable this SnackAmp instance (leave blank to allow all keys)"
      label2 ircLastKey            "LIRC Last Key"
      "The last remote key received by SnackAmp"
      label2 ircEnabled         "Listening for Commands"
       "This is set to 1 if this SnackAmp instance is listening for LIRC commands"
      entry ircCheck             "LIRC Check Buttons"
       "Enter a list of remote buttons to check to enable/disable this SnackAmp instance. Make sure your Enable key is one of them. All non matching keys will disable this instance."
   }
   pack [subFrame $optionData "Infrared/API Remote Control" $W.irc] -side top -anchor w -fill x
   set optionData {
      checkbutton checkMediaFiles     "Check Track Ratings"
       "Check to skip track copy if ratings do not match current sorting options."
      checkbutton useFcopy     "Use Faster Copy"
       "Use faster file copy method, may introduce skips in music."
   }
   pack [subFrame $optionData "Attached Media" $W.am] -side top -anchor w -fill x

}
#------------------------------------------------------------------------------
# Function   : settings::APLSettings
# Description: Settings page Misc functions
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc settings::APLSettings {W} {
   ##
   # Other options
   #
   set e 0
   set gr [list]
   set optionData {
     checkbutton updateAPL        "Sync on Selection"  "Sync AutoPlaylist Tree on track selection"
     checkbutton syncAPL          "Sync on Change"     "Sync AutoPlaylist Tree on track change"

     checkbutton treeLine         "Node Lines"         "Show Node-Lines on AutoPlaylist Tree View"
     checkbutton useTableStripes  "Row Stripes"        "Add Stripe to Alternate Table Rows"

     checkbutton useTableBars     "Column Bars"        "Add Vertical Bar Between Columns"
     checkbutton autoTableStretch "Auto Expand"        "Allow Table Columns to Expand to Fill Window"

     checkbutton useAPLColors     "Use Colors"         "Use SnackAmp Colors for Text and Background (needs restart)"
     checkbutton dndExpand        "Expand on Drop"     "Expand Target Folder after Moving A Folder to it"

     checkbutton savePlonApply    "Auto-Save Playlist" "Save current Playlist before changing to a new AutoPlaylist"
     checkbutton askApplyAPL      "Confirm Playlist Change" "Check to confirm replacement of Playlist with new AutoPlaylist"

     checkbutton rippleSet        "Ripple Tree Selections" "Check to have children of nodes selected when a parent is selected."
     checkbutton incrKeyFilter    "Incremental Search" "Check to have track table updated after keystrokes entered into filter entry."

     checkbutton checkStat        "Check File Stats" "Check the timestamp on each file and update catalog if needed during track listing (will slow down folder listings slightly)"
     spinner2    maxAPLApply   {"Max APL Tracks to Apply" {0 50000 1000}} "Enter a maximum number of tracks to be inserted to the playlist when applying an AutoPlaylist. Enter 0 to not limit the number of tracks."

   }
#     checkbutton showFolderTypes  "Group Media by Type" "If you have  a large number of media folders mounted, check this to group them. (requires restart)"
   pack [subFrame $optionData "General Settings" $W.apl] -side top -anchor w -fill x
   foreach {xxxx Setting Text Help} $optionData {
      trace variable ::snackAmpSettings($Setting) w settings::setTableOption
   }
   # Columns to Show
   set hull [labeledFrame $W.pm -text "Columns To Show in AutoPlaylist Table" -bd 2 -relief groove -font listbox]
   foreach {item} [lrange $::TA(cols) 1 end] {
      lappend gr [checkbutton $hull.[incr e] \
         -pady 0 -highlightthickness 0 \
         -variable ::snackAmpSettings(col,$item,show) \
         -anchor w -text $item -font listbox]
      balloon::define $hull.$e "Check to show $item in playlist table"
      if {[llength $gr]>=8} {
         eval grid $gr -sticky w
         set gr [list]
      }
   }
   pack $W.pm -side top -anchor w -fill x
}
##
# Streaming Media options
#
proc settings::streamingSettings {W} {
   set optionData {
      entrydir uskDir            "Streaming Shortcut Folder"
       "Select a folder to place Streaming Shortcuts when playing streams from the web (e.g. shoutcast). You should place it within your current media folders to manage it from the AutoPlaylist Manager"
      entry minBuffer                  "Desired Buffer Size"
       "Desired buffer size for streaming audio in KBytes."
   }
   pack [subFrame $optionData "Play Settings" $W.play] -side top -anchor w -fill x

   set optionData {
      entry proxyhost                 "Host Name of Proxy"
       "If you need a proxy to connect to the internet, enter the host name here. Don't forget to enable it with checkbox below."
      entry proxyport                 "Port Number for Proxy"
       "Enter port number of proxy. Don't forget to enable it with checkbox below."
      entry localurl                 "Exclude patterns for local hosts"
       "Enter matching patterns to indicate local (non--proxy) machines separated by semi-colons. E.g. *.mydomain.com;*george*"
      checkbutton useproxy            "Enable Use of Proxy"
       "Check this if your internet connection requires a proxy."
   }
   pack [subFrame $optionData "Proxy Settings" $W.pxy] -side top -anchor w -fill x

   set optionData {
      checkbutton EnableStreaming "Enable Internal Streaming Server"
       "Check this to allow a web browser to remote control this player or show internal help (requires restart)."
      checkbutton useIP "Use raw IP Address for playlists."
       "Check this to embed this host's raw IP address rather than host name for use by streaming clients to access this server. Raw addresses may be needed if hostname not available in dns."
      entry altStSrc                 "Alternate Stream Source"
       "Enter host name or IP Address to use instead of this hosts host name or IP address for by streaming clients to access this server (may need if behind a firewall or router). Raw IP must be disabled for this entry to be used."
   }
   pack [subFrame $optionData "Streaming Server" $W.adj] -side top -anchor w -fill x
}
##
# Helper for subframes
#
proc settings::subFrame {optionData Title W} {
   set hull [labeledFrame $W -text $Title -bd 2 -relief groove -font listbox]
   stuffSettingData $optionData $hull
   return $W
}
##
# Web helper to return settings state
#
proc settings::webStatus {index} {
   global snackAmpSettings
   if {[info exists snackAmpSettings($index)]} {
      if {$snackAmpSettings($index)} {
         return "Enabled"
      } else {
         return "Disabled"
      }
   } else {
      return "Unknown"
   }
}
#------------------------------------------------------------------------------
# Function   : settings::resetGainData
# Description: Reset all gain data to default values
# Author     : Tom Wilkason
# Date       : 7/7/2002
#------------------------------------------------------------------------------
proc settings::resetGainData {args} {
   gainControl::resetAll
}

#------------------------------------------------------------------------------
# Return the maxAPL value, don't allow 0
#------------------------------------------------------------------------------
proc settings::maxAPL {args} {
   set max $::snackAmpSettings(maxAPLApply)
   if {$max <= 0} {
      set max 1000000
   }
   return $max
}
#------------------------------------------------------------------------------
# Function   : settings::setTableOption
# Description: Callback with table options get modified
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc settings::setTableOption {_glob option args} {
   upvar $_glob glob
   set Var $glob($option)
   if {[APLExists]} {
      switch -- $option {
         useTableStripes {
            if {$Var} {
               $::PLB config -stripebackground $::snackAmpSettings(tsbg)
            } else {
               $::PLB config -stripebackground ""
            }
         }
         useTableBars {
            if {$Var} {
               $::PLB config -showseparators yes
            } else {
               $::PLB config -showseparators no
            }
         }
         autoTableStretch {
            if {$Var} {
               $::PLB config -stretch all
            } else {
               $::PLB config -stretch [list]
            }
         }
         rippleSet {
            set curVar [Tree::cget $::TREE -rippleSet]
            if {$curVar!=$Var} {
               Tree::configure $::TREE -rippleSet $Var
            }
         }
         treeLine {
            set curVar [Tree::cget $::TREE -treeLine]
            if {$curVar!=$Var} {
               Tree::configure $::TREE -treeLine $Var
               Tree::buildwhenidle $::TREE
            }
         }
      }
   }
}
#------------------------------------------------------------------------------
# Function   : settings::schemeSettings
# Description: Settings page for File Associations
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc settings::associationSettings {W} {
   variable reg
   variable FONTS
   global snackAmpSettings
   package require fileAssociation
   set exts $snackAmpSettings(fileExtensions)
   # Support playlists
   lappend exts pls m3u
   foreach type [lsort $exts] {
      fileAssociation::isRegisteredMF .$type settings::reg($type)
      radiobutton $W.cb$type -text "Associate $type with SnackAmp" \
         -command "[list fileAssociation::registerMF .$type];[list fileAssociation::isRegisteredMF .$type settings::reg($type)]" \
         -variable settings::reg($type) -value 1 -font listbox
      radiobutton $W.res$type -text "Restore $type to original" \
         -command "[list fileAssociation::unRegisterMF .$type];[list fileAssociation::isRegisteredMF .$type settings::reg($type)]" \
         -variable settings::reg($type) -value 0 -font listbox
      grid $W.cb$type $W.res$type -sticky w
   }

}
#------------------------------------------------------------------------------
# Function   : settings::postMenuNames
# Description: On pop-up, post the fonts, if already built then post menu
#              Menu is shared for all buttons to save memory
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc settings::postMenuNames {Menu Name} {
   # If menu has data then use it,
   # otherwise build it up with the
   # font information.
   if {[$Menu index end] == 0 || [$Menu index end]=="none"} {
      set families [lsort -dictionary [font families]]
      foreach {family} $families {
         set first [string index $family 0]
         lappend fontNames($first) $family
      }
      $Menu delete 0 end
      foreach {Value} [lsort [array names fontNames]] {
         $Menu add cascade -label $Value -menu $Menu.f$Value
         menu $Menu.f$Value
         foreach {item} $fontNames($Value) {
           $Menu.f$Value add command -label "$item" \
            -command [list settings::setFontFamily $item $Name] -font [list $item]
         }
      }
   }
}
#------------------------------------------------------------------------------
# Function   : settings::installFontCallbacks
# Description: Install callbacks to immediately change the visible fonts when
#              the user changes them
# Author     : Tom Wilkason
# Date       : 2/6/2002
#------------------------------------------------------------------------------
proc settings::installFontCallbacks {args} {
   ##
   # Callbacks to immediatly set the fonts when the user changes them
   #
   trace vdelete ::fontMetrics w settings::fontCallback
   array set ::fontMetrics [array get ::snackAmpSettings *,size]
   array set ::fontMetrics [array get ::snackAmpSettings *,weight]
   array set ::fontMetrics [array get ::snackAmpSettings *,slant]
   trace variable ::fontMetrics w settings::fontCallback
}
#------------------------------------------------------------------------------
# Function   : settings::setFontFamily
# Description: On callback, set the appropriate widget type to a font value passed in
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc settings::setFontFamily {family Name} {
   set ::snackAmpSettings($Name,family) $family
   # This will make the effect immediate
   settings::saFontSize $::snackAmpSettings(FontSizeFactor)
}
#------------------------------------------------------------------------------
# Function   : settings::fontCallback
# Description: When radio buttons change, change the fonts immediatly
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc settings::fontCallback {_glob index mode} {
   upvar $_glob Settings
   set ::snackAmpSettings($index) $Settings($index)
   settings::saFontSize $::snackAmpSettings(FontSizeFactor)
}
#------------------------------------------------------------------------------
# Function   : settings::colorPicker
# Description: Bring up color picker to change colors
# Author     : Tom Wilkason
# Date       : 10/30/2001
#------------------------------------------------------------------------------
proc settings::colorPicker {index} {
   global snackAmpSettings
   if {$::tcl_platform(platform) == "unix"} {
      set tl .tsg
      set i 0; while {[winfo exists .cp[incr i]]} {continue}
      set result [SelectColor .cp$i \
         -parent $tl \
         -title "Select $index color"\
         -color $snackAmpSettings($index)]
   } else {
      set result [tk_chooseColor -initialcolor $snackAmpSettings($index)]
   }
   if {[string length $result]} {
      set snackAmpSettings($index) $result
      settings::applyColor
   }
}
##
# Pop-down color menu for the colored rectangless
#
proc settings::colorMenu {w index} {
   global snackAmpSettings
   set result [SelectColor::menu $w.color [list below $w] -color [$w cget -background]]
   if {[string length $result]} {
     set snackAmpSettings($index) $result
     settings::applyColor
   }
}
#------------------------------------------------------------------------------
# Function   : settings::fileChooser
# Description: Bring up file selector to select a file. Assign it to the var
# Author     : Tom Wilkason
# Date       : 10/30/2001
#------------------------------------------------------------------------------
proc settings::fileChooser {varname {parentWin .}} {
   upvar $varname filename
   set file [chooseFile [file dirname $filename] $parentWin]
   if {[string length $file] > 0} {
      set filename $file
   }
}
proc settings::folderChooser {varname {parentWin .}} {
   upvar $varname filename
   set file [chooseDirectory [file dirname $filename] $parentWin]
   if {[string length $file] > 0} {
      set filename $file
   }
}
#------------------------------------------------------------------------------
# Function   : settings::regColor
# Description: register a color for a widget
# Author     : Tom Wilkason
# Date       : 10/30/2001
#------------------------------------------------------------------------------
proc settings::regColor {widget args} {
   variable Colors
   foreach {option color} $args {
      set Colors($widget,$option) $color
   }
}
#------------------------------------------------------------------------------
# Function   : settings::applyColor
# Description: Go through each option and apply the latest values if they
#              are different.
# Author     : Tom Wilkason
# Date       : 10/30/2001
#------------------------------------------------------------------------------
proc settings::applyColor {args} {
   variable Colors
   foreach {index} [array names Colors] {
      foreach {widget option} [split $index ,] {break}
      catch {
         set val [$widget cget $option]
         set newVal [set $Colors($index)]
         if {$val ne $newVal} {
            $widget config $option $newVal
         }
      }
   }
}
#------------------------------------------------------------------------------
# Function   : settings::stuffSettingData
# Description: General purpose routine to grid a number of items
#              for use in settings GUIs.
# Author     : Tom Wilkason
# Date       : 5/15/2002
#------------------------------------------------------------------------------
proc settings::stuffSettingData {optionData B {i 100}} {
   global TA
   set e [expr {$i+100}]
   set ew [expr {$e+100}]
#   set cbData [list]
# Grid is assumed to be 4 across
   set enData [list]
   foreach {Widget LinkedVariable Label Help} $optionData {
      switch -- $Widget {
         label {
            # Uses four slots
            grid [label $B.[incr i] -text [subst $Label] -anchor e -font listbox] \
               [$Widget $B.[incr e] -textvariable snackAmpSettings($LinkedVariable) -width 25 \
               -font listbox \
               -relief sunken \
               -anchor w \
               -foreground $::snackAmpSettings(efg)] \
               - - \
               -sticky ew -padx 2
            balloon::define $B.$e $Help
            settings::regColor $B.$e -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
         }
         label2 {
            # Uses two slots
            if {[string is double -strict $::snackAmpSettings($LinkedVariable)]} {
               set anc e
            } else {
               set anc w
            }
            lappend enData [label $B.[incr i] -text [subst $Label] -anchor e -font listbox]
            lappend enData [label $B.[incr e] \
               -textvariable snackAmpSettings($LinkedVariable) \
               -font listbox \
               -width 6 \
               -relief sunken \
               -anchor $anc \
               -foreground $::snackAmpSettings(efg)]
            balloon::define $B.$e $Help
            settings::regColor $B.$e -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
         }
         scale2 {
            # Uses one slots
            foreach {Label choices} $Label {break}
            foreach {min max res} $choices {break}
            set interval [expr {($max-$min)/10.0}]
            # -tickinterval $interval
            lappend enData [scale $B.[incr e] -label [subst $Label] \
               -from $min -to $max  -orient horiz \
               -font listbox -variable snackAmpSettings($LinkedVariable) -width 7 -bd 2  \
               -sliderlength 15 -resolution $res -font balloon -relief groove \
               -activebackground blue -troughcolor $::snackAmpSettings(bg)]
            balloon::define $B.$e $Help
            settings::regColor $B.$e -troughcolor ::snackAmpSettings(bg)
         }
         option2 {
            # Uses two slots
            foreach {Label choices} $Label {break}
            lappend enData [label $B.[incr i] -text [subst $Label] -anchor e -font listbox]
            lappend enData [listOptionMenu $B.[incr e] ::snackAmpSettings($LinkedVariable) $choices]
            balloon::define $B.$e $Help
            settings::regColor $B.$e -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
         }
         spinner2 {
            # Uses two slots
            foreach {Label choices} $Label {break}
            lappend enData [label $B.[incr i] -text [subst $Label] -anchor e -font listbox]
            foreach {from to increment} $choices {break}
            lappend enData [spinbox $B.[incr e] -width 4 -from $from -to $to -increment $increment\
                -textvariable ::snackAmpSettings($LinkedVariable) \
                -background $::snackAmpSettings(ebg) \
                -foreground $::snackAmpSettings(efg)]
            balloon::define $B.$e $Help
            settings::regColor $B.$e -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
         }
         entry2 {
            # Uses two slots
            if {[string is double -strict $::snackAmpSettings($LinkedVariable)]} {
               set anc right
            } else {
               set anc left
            }
            lappend enData [label $B.[incr i] -text [subst $Label] -anchor e -font listbox]
            lappend enData [entry $B.[incr e] -textvariable snackAmpSettings($LinkedVariable) -width 6 \
               -background $::snackAmpSettings(ebg) \
               -foreground $::snackAmpSettings(efg) \
               -font entry \
               -justify $anc]
            balloon::define $B.$e $Help
            settings::regColor $B.$e -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
         }
         entry {
            # Uses four slots
            set anc left
            grid [label $B.[incr i] -text [subst $Label] -anchor e -font listbox] \
                 [entry $B.[incr e] -textvariable snackAmpSettings($LinkedVariable) -width 25 \
               -background $::snackAmpSettings(ebg) -font entry -justify $anc \
               -foreground $::snackAmpSettings(efg)]  \
               - - \
               -sticky ew -padx 2
            balloon::define $B.$e $Help
            settings::regColor $B.$e -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
         }
         option {
            # Uses four slots
            foreach {Label choices} $Label {break}
            grid [label $B.[incr i] -text [subst $Label] -anchor e -font listbox] \
                 [listOptionMenu $B.[incr e] ::snackAmpSettings($LinkedVariable) $choices] \
                 - - \
                 -sticky ew -padx 2

            balloon::define $B.$e $Help
            settings::regColor $B.$e -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
         }
         entrybrowse {
            # Uses four slots
            # Only want to set the browser for systems that use env(BROWSER), e.g. *nix
            # TFW: Check this logic
            grid [label $B.[incr i] -text [subst $Label] -anchor e -font listbox] \
                 [entry $B.[incr e] -textvariable snackAmpSettings($LinkedVariable) -width 25\
                   -font listbox -background $::snackAmpSettings(ebg) -foreground $::snackAmpSettings(efg)] \
                 - -\
                 [button $B.[incr ew] -text "Browse..." -width 9 -command [list settings::fileChooser snackAmpSettings($LinkedVariable) $B] \
                 -font listbox -bd 1 -pady 0 -relief raised] -sticky ew -padx 2
            balloon::define $B.$e $Help
            settings::regColor $B.$e -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
         }
         entrydir {
            # Uses four slots
            grid [label $B.[incr i] -text [subst $Label] -anchor e -font listbox] \
                 [entry $B.[incr e] -textvariable snackAmpSettings($LinkedVariable) \
                 -background $::snackAmpSettings(ebg) -foreground $::snackAmpSettings(efg)] \
                 - -\
                 [button $B.[incr ew] -text "Browse..."  -width 9 -command [list settings::folderChooser snackAmpSettings($LinkedVariable) $B] \
                 -font listbox -bd 1 -pady 0 -relief raised] -sticky ew -padx 2
            balloon::define $B.$e $Help
            settings::regColor $B.$e -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
         }
         entryList {
            # Uses four slots
            grid [label $B.[incr i] -text [subst $Label] -anchor e -font listbox] \
                 [entry $B.[incr e] -textvariable snackAmpSettings($LinkedVariable) \
                   -background $::snackAmpSettings(ebg) -foreground $::snackAmpSettings(efg)] \
                   - -\
                 [button $B.[incr ew] -text "Manage..."  -width 9 -command [list folderManager::launch $LinkedVariable] \
                 -font listbox -bd 1 -pady 0 -relief raised] -sticky ew -padx 2
            balloon::define $B.$e $Help
            settings::regColor $B.$e -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
         }
         checkbutton {
            # Uses two slot s
            lappend enData [checkbutton $B.[incr e] -pady 0 \
               -highlightthickness 0 -width 10 -variable snackAmpSettings($LinkedVariable) \
               -anchor w -text [subst $Label] -font listbox] -
            balloon::define $B.$e $Help
         }
         button {
            # Uses one slot
            lappend enData [button $B.[incr e] -relief raised -anchor center -command $LinkedVariable \
            -text [subst $Label] -font listbox]
            balloon::define $B.$e $Help
         }
         x { lappend enData x}
         \- { lappend enData -}
         ^ { lappend enData ^}
         default {}
      }
      # Grid in small entry boxes if we are four across
      if {[llength $enData] == 4} {
         eval grid $enData -sticky ew -padx 2 -pady 0
         set enData [list]
      }
   }
   # Grid remaining entries
   if {[llength $enData] > 0} {
      eval grid $enData -sticky ew -padx 2 -pady 0
      set enData [list]
   }
   # Allow entrys and other RH stuff to expand
   for {set i 0} {$i < 5} {incr i} {
      grid columnconfigure $B $i -weight 1
   }
}
#------------------------------------------------------------------------------
# Function   : settings::saveSettings
# Description: Save the settings from the settings::settingsGui
# Author     : Tom Wilkason
# Date       : 12/30/2000
#------------------------------------------------------------------------------
proc settings::saveSettings {args} {
   global snackAmpSettings TA
   catalog::fixMediaPath
   settings::buildFileExtenstions
   # Save playlist lines
   upvar snackAmpSettings(playListLines) pll
   set pll [LB::lines]
   if {$pll < 2 || $pll > 50} {
      set pll 20
   }
   if {[catch {
      set fid [open $TA(settingsFile) w]
   } result] \
   } then {
      tk_messageBox -icon warning -type ok -title "Open Error" -message $result
   } else {
      # Strip off early portion of folderList to prevent huge memory usage
      while {$snackAmpSettings(folderPos) > 20} {
         lpop left snackAmpSettings(folderList)
         incr snackAmpSettings(folderPos) -1
      }

      fconfigure $fid -buffersize 20000
      if {$snackAmpSettings(Version2Settings)} {
         set result "#######\n"
         append result "# SnackAmp Settings File Saved [clock format [clock seconds]]\n"
         append result "#\n"
         append result "array set ::snackAmpSettings \{\n"
         foreach {index} [lsort -dictionary [array names ::snackAmpSettings *]] {
            set value [string trim $::snackAmpSettings($index)]
            # Format the line real pretty
            set line [format "   %-25s %s\n" [list $index] [list $value]]
            append result $line
         }
         append result "\}\n"
      } else {
         set result [array get ::snackAmpSettings]
      }
      puts $fid $result
      close $fid
   }
}
##
# Set the default fonts
#
proc settings::setDefaultFont {args} {
   variable FONTS
   global snackAmpSettings
   foreach {Font Specs Description} $FONTS {
      foreach {family size weight slant}  $Specs {break}
      foreach item {family size weight slant} {
         set ::snackAmpSettings($Font,$item) [set $item]
      }
   }
}
##
# Scale all of the fonts by some factor
#
proc settings::saFontSize {factor} {
   variable FONTS
   global snackAmpSettings
   set families [font families]
   lappend families Helvetica Times
   set families [lsort $families]
   foreach {Name Specs Description} $FONTS {
      set fontTry $::snackAmpSettings($Name,family)
      foreach {try} [list $fontTry "arial" "Arial" "helvetica" "Helvetica"] {
         if {[lsearch -sorted -exact $families $try] < 0} {
            saLog "The font family '$Description' you have selected, '$::snackAmpSettings($Name,family)' is invalid "
         } else {
            set ::snackAmpSettings($Name,family) $try
            break
         }
      }
      set size  $::snackAmpSettings($Name,size)
      if {[catch {
         font configure $Name -size [expr {round($size*$factor)}] \
            -family $::snackAmpSettings($Name,family) \
            -slant $::snackAmpSettings($Name,slant) \
            -weight $::snackAmpSettings($Name,weight)
      } result]} {
         saLog $result
      }
   }
   return
}
##
# Update the short word tokens
# Words that are ignored with changing the case in a track title or file name
proc settings::updateShortWords {args} {
   global TA
   foreach token $::snackAmpSettings(shortWords) {
      lappend TA(lcwords) [string tolower $token] [string totitle $token] [string toupper $token]
   }
#    foreach {letter} {a b c d e f g h j k l m n o p q r s t u v w x y z} {
#       lappend TA(lcwords) [string toupper $letter] [string tolower $letter]
#    }
   # can bsearch this
   set TA(lcwords) [lsort -unique $TA(lcwords)]
}

#------------------------------------------------------------------------------
# Function   : settings::readSettings
# Description: Read the default settings from the user home dir, first assign
#              defaults then read the user file to change them.
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc settings::readSettings {args} {
   global snackAmpSettings TA
   variable taskBarIconList
   # Defaults below
   settings::setDefaultFont
   set snackAmpSettings(activeMiniPlayer) 0
   set snackAmpSettings(allmusicBase)           {http://allmusic.com/cg/amg.dll?p=amg&sql=} ;#1Eva|Cassidy
   set snackAmpSettings(altStSrc) ""
   set snackAmpSettings(AutoDelete)             0
   set snackAmpSettings(customGenres)           ""
   set snackAmpSettings(chunk)                  [expr {int(pow(2,18))}]
   set snackAmpSettings(autoLevel)              1
   set snackAmpSettings(autoLevelPreComute)     0
   set snackAmpSettings(tagGainValues)          0
   set snackAmpSettings(gainType)              "Track"
   set snackAmpSettings(useFcopy)               0
   set snackAmpSettings(checkMediaFiles)        0
   set snackAmpSettings(autoTruncate)           1
   set snackAmpSettings(autoLevelMaxSamples)   50
   set snackAmpSettings(autoLevelReference)     2
   set snackAmpSettings(autoLevelSeconds)       15
   set snackAmpSettings(autoLevelWindow)       8192
   set snackAmpSettings(autoLevelInc)          0.05 ;# 50ms
   set snackAmpSettings(AutoRescan)             0
   set snackAmpSettings(autoSavePlaylist)       0
   set snackAmpSettings(balloonDelay)           1000
   set snackAmpSettings(balloonDuration)        3000
   set snackAmpSettings(bufSize)                16384
   set snackAmpSettings(cleanListings) 0
   set snackAmpSettings(browseNowPlaying) 0
   set snackAmpSettings(ContinuousPlay)         1
   set snackAmpSettings(coverart)              [list folder.jpg folder.png folder.gif .folder.jpg *.jpg *.png]
   set snackAmpSettings(coversize)              150
   set snackAmpSettings(coverbrowsersize)       150
   set snackAmpSettings(cssName) "/snackamp.css"
   set snackAmpSettings(currentColorScheme)     "Default"
   set snackAmpSettings(dontMoveFiles)          0
   set snackAmpSettings(enableConsole)          1
   set snackAmpSettings(EnableRemoteControl)    0
   set snackAmpSettings(EnableStreaming)        0
   set snackAmpSettings(EnableWebServer)        0
   set snackAmpSettings(fadeFactor)             0.6
   set snackAmpSettings(Favorites) [list]
   set snackAmpSettings(fadePause)              0
   set snackAmpSettings(fastscan) 1
   set snackAmpSettings(filterHistory) [list]
   set snackAmpSettings(findHistory) [list]
   set snackAmpSettings(fileExtensions)         [list mp3 ogg wav au aif usk]
   set snackAmpSettings(FileListSize)           0.75
   set snackAmpSettings(FontSizeFactor)         1
   set snackAmpSettings(gainFactor)             25.0
   set snackAmpSettings(minDB)                 -50
   #set snackAmpSettings(lastGlobalGain)         50
   set snackAmpSettings(lastGlobalGaindB)       -30
   set snackAmpSettings(HelpBrowser)            {}
   set snackAmpSettings(hideFileList)           0
   set snackAmpSettings(History)                [list]
   set snackAmpSettings(hotkeyFile)             $::HOME/snackAmpHotKeys.tcl
   set snackAmpSettings(HttpHits)               0
   set snackAmpSettings(HttpLocal)              ""
   set snackAmpSettings(HttpUsers)              ""
   set snackAmpSettings(httpMaxused)            5000
   set snackAmpSettings(httpNextLength)         50
   #bg indexing stuff
   set snackAmpSettings(indexUnits)             "Hours"
   set snackAmpSettings(indexHours)             24
   set snackAmpSettings(indexStart)             0
   set snackAmpSettings(indexStop)              0
   set snackAmpSettings(bgCatalog)              0
   set snackAmpSettings(alwaysCatalog)          0
   set snackAmpSettings(nextIndex)              [clock seconds]
   set snackAmpSettings(lastIndex)              0
   set snackAmpSettings(lastIndexFormatted)    "Not Run"

   set snackAmpSettings(nextIndexFormatted)     [clock format $snackAmpSettings(nextIndex)]
   set snackAmpSettings(indexBackground)        0

   set snackAmpSettings(ID3ballon)              1
   set snackAmpSettings(showCoverArt)           0
   set snackAmpSettings(slideCoverArt)          1
   set snackAmpSettings(amazonCoverArt)         0
   set snackAmpSettings(walmartCoverArt)        1
   set snackAmpSettings(jpegCoverArt)           1
   set snackAmpSettings(rowsCoverArt)           3
   set snackAmpSettings(sizeCoverArt)           300
   set snackAmpSettings(maxCoverFind)           20
   set snackAmpSettings(maxCoverShow)           9
   set snackAmpSettings(minCoverWidth)          100
   set snackAmpSettings(maxCoverWidth)          1000
   set snackAmpSettings(ratioThreshold)         1.2
   set snackAmpSettings(showLabels)             0
   set snackAmpSettings(ID3editLimit)           200
   set snackAmpSettings(ID3OKCreate)            0
   set snackAmpSettings(ID3ReadOnly)            0
   set snackAmpSettings(ID3V12)                 1
   set snackAmpSettings(defaultEncoding)        "Only If Needed"
   set snackAmpSettings(ID3V2)                  "Always"
   set snackAmpSettings(ID3V1)                  "If Existing"
   set snackAmpSettings(ID3ReadOrder)           "V2 V1 OGG"
   set snackAmpSettings(IncomingDirs)           {}
   set snackAmpSettings(incrKeyFilter)          1
   set snackAmpSettings(checkStat)              0
   set snackAmpSettings(ipAddress)              $iptools::myIP
   set snackAmpSettings(jukebox,nocoverart) 1
   set snackAmpSettings(jukebox,brcoverart) 0
   set snackAmpSettings(jukebox,linkartist) 1
   set snackAmpSettings(jukebox,popup)      1
   set snackAmpSettings(jukebox,rettofolder) 1
   set snackAmpSettings(lastFolderFile)         $TA(defaultFolderFile)
   set snackAmpSettings(levelMeter) 0
   set snackAmpSettings(PlayListFile)           {}
   set snackAmpSettings(lineSelect)             [lindex [snack::mixer lines]  0]
   set snackAmpSettings(listbytag) 1
   set snackAmpSettings(ListDepth)              1                           ;# Will show all folder
   set snackAmpSettings(localOnly) 0
   set snackAmpSettings(localurl) ""
   set snackAmpSettings(logFile)                0
   set snackAmpSettings(maxAPLApply)            0
   set snackAmpSettings(maxGain)                4
   set snackAmpSettings(minGain)               -9
   set snackAmpSettings(minNextTrack)          10
   set snackAmpSettings(artrowcount)            4
   set snackAmpSettings(minTxtRating) "--"
   set snackAmpSettings(minWidth)             350                         ;# Initial width of the track info area
   set snackAmpSettings(mp3gainLoc)            ""
   set snackAmpSettings(MonitorCacheFolders)    0
   set snackAmpSettings(MonitorScanFolders)     0
   set snackAmpSettings(ScanFolderDepth)        1
   set snackAmpSettings(moveToTop)              0
   set snackAmpSettings(NextAfterCurrnt)        1
   set snackAmpSettings(noTail)                 1                           ;# Will show all folder
   set snackAmpSettings(catPlaylist)            0
#   set snackAmpSettings(playlistID3)            1
   set snackAmpSettings(playListLines)          15
   set snackAmpSettings(playResolution) 12
   set snackAmpSettings(port)                   8127
   set snackAmpSettings(positionMiniPlayer) [list 0 0]
   set snackAmpSettings(preserveTime)           0
   set snackAmpSettings(preserveV2data)         1
   set snackAmpSettings(zeroPadTrack)           0
   set snackAmpSettings(xxxxxx)         0
   set snackAmpSettings(priority) normal
   set snackAmpSettings(proxyhost) ""
   set snackAmpSettings(proxyport) ""
   set snackAmpSettings(refreshRate)            20
   set snackAmpSettings(remote,brcoverart)      1
   set snackAmpSettings(remote,linkartist)      1
   set snackAmpSettings(remote,popup)           1
   set snackAmpSettings(remote,rettofolder)     1
   set snackAmpSettings(restoreMiniPlayer)      1
   set snackAmpSettings(restorePlayerPos)       1
   if {[catch {wm attributes . -alpha}]} then {
      set snackAmpSettings(fadeImages)          0
   } else {
      set snackAmpSettings(fadeImages)          1
   }
   set snackAmpSettings(fadePopup)              1
   set snackAmpSettings(smallArt)               1
   set snackAmpSettings(miniArt)                0
   set snackAmpSettings(leftSmallArt)           1
   set snackAmpSettings(minSysTray)             0
   set snackAmpSettings(minArtSize)            150
   set snackAmpSettings(aplArtSize)             .35
   set snackAmpSettings(sliderPosition)         "Right Top"
   set snackAmpSettings(sliderDelay)            6000
   set snackAmpSettings(sliderAlpha)            0.8
   set snackAmpSettings(sliderSpeed)            7
   set snackAmpSettings(maxArtSize)             300
   set snackAmpSettings(desArtSize)             250
   set snackAmpSettings(sliderArtSize)          75
   set snackAmpSettings(timeSliderHeight)       5
   set snackAmpSettings(timeSliderWidth)        25
   set snackAmpSettings(volSliderHeight)       5
   set snackAmpSettings(volSliderWidth)        25
   set snackAmpSettings(restorePlaying)         1
   set snackAmpSettings(restorePlayList)        1
   set snackAmpSettings(resumePlay)             1
   set snackAmpSettings(rippleSet)              1
   set snackAmpSettings(FilterWhat)          "Any"
   set snackAmpSettings(savePlonApply)          0
   set snackAmpSettings(askApplyAPL)            0
   set snackAmpSettings(scrolledCanvasHeight)   4i
   set snackAmpSettings(shortWords) [list is it am by a the in of by on for we you they with and to has do]
   set snackAmpSettings(showDetail) 2
   set snackAmpSettings(fancyMenu)              0
   set snackAmpSettings(showMS)                 0
   set snackAmpSettings(showTicks)              1
   set snackAmpSettings(smartName)              "%A - %C - %N - %T"
   set snackAmpSettings(freedb)                 "http://uk.freedb.org:80/~cddb/cddb.cgi"
   set snackAmpSettings(cddbBase)               "http://www.freedb.org/"

   set snackAmpSettings(SongPlayMode)           0
   set snackAmpSettings(sortModeTxt)            {Sort by all Tracks}
 #  set snackAmpSettings(sortPlay)               0
   set snackAmpSettings(splitChar)              "-"
   set snackAmpSettings(subNetMask)             $iptools::defMask
   set snackAmpSettings(syncAPL)                0
   set snackAmpSettings(timeString)             "%Y-%m-%d %H:%M:%S"
   set snackAmpSettings(Trace)                  0
   set snackAmpSettings(treeDepth) 0
   set snackAmpSettings(treeLine) 0
   set snackAmpSettings(treeSize)               "640x480"
   set snackAmpSettings(updateAPL)              0
   set snackAmpSettings(usedPort) "disabled"
   set snackAmpSettings(useIP) 0
   set snackAmpSettings(useproxy) 0
   set snackAmpSettings(uskDir)                 ""
   set snackAmpSettings(minBuffer)              250000
   set snackAmpSettings(version)               2.0
   set snackAmpSettings(Version2Settings)       1
   set snackAmpSettings(ircPort)                8765
   set snackAmpSettings(ircHost)                ""
   set snackAmpSettings(apiPort)                8775
   set snackAmpSettings(ircEnable)              ""
   set snackAmpSettings(ircEnabled)             0
   set snackAmpSettings(ircCheck)              ""
   set snackAmpSettings(ircLastKey)              ""
   set snackAmpSettings(useIRC)                 0
   set snackAmpSettings(useAPI)                 0
   set snackAmpSettings(webLog)                 0
   set snackAmpSettings(spcar)                  ""
   set snackAmpSettings(windowScale)            "Days"
   set snackAmpSettings(spcharSpace)            1
   set snackAmpSettings(sortShuffle) 0
   set snackAmpSettings(postmorph) 0
   set snackAmpSettings(postmorphamount) 25
   set snackAmpSettings(mp3Player) ""
   set snackAmpSettings(folderList) [list]
   set snackAmpSettings(id3Structure) {
      "%N - %T"
      "%A/%C/%N - %T"
      "%A/%C - %N - %T"
      "%A - %C - %N - %T"
      "%C/%N - %T"
      "%C - %N - %T"
   }
   set snackAmpSettings(fileStructure) {
      "Default"
      "%A/%C/%N - %T"
      "%A/%C - %N - %T"
      "%A - %C - %N - %T"
      "%C/%N - %T"
      "%C - %N - %T"
   }
   if {$::tcl_platform(os) eq "Linux"} {
      set snackAmpSettings(globalMixer)            0
      set snackAmpSettings(linscale) 1
   } else {
      set snackAmpSettings(linscale) 0
      set snackAmpSettings(globalMixer)            1
   }
#   set snackAmpSettings(showFolderTypes) 1
   set snackAmpSettings(folderPos) -1
   if {[info exists ::env(PLAYLATENCY)]} {
      set snackAmpSettings(normalPlayLatency) $::env(PLAYLATENCY)
   } else {
      set snackAmpSettings(normalPlayLatency) 2000.0
   }
   set snackAmpSettings(meterUpdateRate) 150
   set snackAmpSettings(trackGap) 0
   set snackAmpSettings(longPlayLatency) 5000.0

   # Tablelist columns to use initially
   set snackAmpSettings(table,sortcol) 1
   set snackAmpSettings(table,sortorder) "increasing"
   foreach {icon} $taskBarIconList {
      set snackAmpSettings(icon,$icon) 0
   }
   set snackAmpSettings(icon,Pause) 1
   ##
   # Default Column Info, will be changed by user
   # Col: Name width alignment sorting editable show
   set colInfo {
      Key       0  right dictionary 0 1
      Track     0  right dictionary 1 1
      Title     28 left  dictionary 1 1
      Artist    20 left  dictionary 1 1
      Album     20 left  dictionary 1 1
      Media     0  left  dictionary 1 0
      Year      0  right dictionary 1 0
      Genre     0  left  dictionary 1 0
      Note      25 left  dictionary 1 0
      Comp      0  left  dictionary 1 0
      Enc       0  left  dictionary 0 0
      Desc      52 left  dictionary 0 0
      Duration  0  right timify     0 1
      Rate      0  right integer    0 1
      Sample    0  right real       0 0
      Folder    25 left  dictionary 0 0
      File      40 left  dictionary 1 0
      Size      0  right commify    0 0
      Date      0  right ascii      0 0
      Gain      0  right dictionary 0 0
      AlbumGain 0  right dictionary 0 0
      Tag       0  right ascii      0 0
      Rating    0  left  mrating    0 0
      Played    0  right integer    1 0
      Votes     0  right integer    1 0
      Cataloged 0  right ascii      0 0
      LastPlay  0  right ascii      0 0
   }
#      Start    0  right integer    1 0
#      Stop     0  right integer    1 0

   # Save in an array for easy access
   foreach {col width align type edit show} $colInfo {
      lappend ::TA(cols) $col
      set snackAmpSettings(col,$col,width) $width
      set snackAmpSettings(col,$col,align) $align
      set snackAmpSettings(col,$col,type) $type
      set snackAmpSettings(col,$col,edit) $edit
      set snackAmpSettings(col,$col,show) $show
   }
   foreach {n} [list -- * ** *** **** *****] {
      set snackAmpSettings(cbRating,$n) 1
   }

   set snackAmpSettings(colorder) $::TA(cols)
   set snackAmpSettings(useTableStripes) 1
   set snackAmpSettings(useTableBars) 1
   set snackAmpSettings(autoTableStretch) 1
   set snackAmpSettings(dndExpand)    1
   set snackAmpSettings(useAPLColors) 1
   set snackAmpSettings(outputDevice) [lindex [snack::audio outputDevices] 0]
   foreach {index} {0 1 2} {
      set snackAmpSettings(sortPlay,$index) "None"
      set snackAmpSettings(sortHow,$index) "-increasing"
   }
   ##
   # Try to read a user settings file, then a common settings file.
   # if neither exists, use these defaults.
   set ::TA(rescan) 0
   if {![file exists $::TA(settingsFile)]||[catch {
      # Try User settings file
      settings::readSettingsFile $::TA(settingsFile)
      settings::saFontSize $snackAmpSettings(FontSizeFactor)
   } result]} then {
      saLog $::errorInfo

      set msg {

Please read below if this is your first time using SnackAmp.

The next dialog will allow you to add media folders to your library before
launching SnackAmp for the first time.  You may add as many folders as you
would like.

You may press Close at any time to jump directly to SnackAmp.

This dialog is found on the SnackAmp File Menu under Manage Media Library.

Answering Yes will bring up the Media Manager dialog.
Answering No will proceed directly to SnackAmp.

      }
      set ::snackAmpSettings(mediaDirs) ""
      set ::TA(mediaDirs) ""
      settings::saFontSize $snackAmpSettings(FontSizeFactor)
      package require saSettings
      settings::buildFileExtenstions
      if {[catch {
         # Try a common file in same folder as script/executable
         settings::readSettingsFile $::TA(defaultSettingsFile)
      } result]} then {
         # no settings file, assume new user
         waitWindow::close
         set result [tk_messageBox -type yesno -title "SnackAmp - Please Respond" \
            -icon info -message [subst $msg]]
         if {$result == "yes"} {
            #catch {file delete $::HOME}
            db::openDB
            catalog::manageMediaFolders
         }
      }
   }
   if {$snackAmpSettings(spcharSpace)} {
      set ::TA(spcar) " $snackAmpSettings(spcar) "
   } else {
      set ::TA(spcar) "$snackAmpSettings(spcar)"
   }
   # Restart Settings
   set ::TA(noTail)       $snackAmpSettings(noTail)
   #set ::TA(showTicks)    $snackAmpSettings(showTicks)
   set ::TA(database) "snackMetakit"
   set ::HASDB 1
   updateShortWords
   settings::buildFileExtenstions
   # Create the columns in the order last saved
   # This reorders all of the columns based on the
   # order of the displayed columns.
   set ::TA(cols) [orderList $::TA(cols) $snackAmpSettings(colorder)]
   # Remove unneeded settings (from old version)
   foreach key [array names snackAmpSettings column*] {
      unset snackAmpSettings($key)
   }
   # Do special stuff for help browser
   # If help entry changes, update internal settings
   if {[string length $snackAmpSettings(HelpBrowser)] > 0} {
      set ::env(BROWSER) $snackAmpSettings(HelpBrowser)
   }
   set TA(mediaDirs) [lsort -unique $snackAmpSettings(mediaDirs)]
   set TA(AutoPlayListSelection) [file rootname [file tail $snackAmpSettings(lastFolderFile)]]
   set snackAmpSettings(indexing) "Idle"
   # force override, new value
   set snackAmpSettings(col,Rating,type) "mrating"
   set snackAmpSettings(mp3gainLoc) [file normalize $snackAmpSettings(mp3gainLoc)]

   trace variable snackAmpSettings(HelpBrowser) wu settings::updateBrowser
   trace variable snackAmpSettings(lastFolderFile) w macros::Title
}
#------------------------------------------------------------------------------
# Function   : updateBrowser
# Description: Changes the env(BROWSER) entry when the help browser is
# changed. Deletes it if the Value is set to an empty string.
#------------------------------------------------------------------------------
proc settings::updateBrowser {arrName index op} {
   # Normally we should worry about the array name but since we *know* its
   # snackAmpSettings, we won't.
   global snackAmpSettings

   if {[string length $snackAmpSettings(HelpBrowser)] == 0} {
      # do nothing
   } elseif {[string equal $op u]} {
      if {[info exists ::env(BROWSER)]} {
         unset ::env(BROWSER)
      }
   } else {
      set ::env(BROWSER) $snackAmpSettings(HelpBrowser)
   }
}
##
##
# Read the snackAmp settings file, handling both old
# and new formats.
#
proc settings::readSettingsFile {fileName} {
   global TA snackAmpSettings
   set fid [open $fileName r]
   set data [read $fid]
   close $fid
   # Handle new or old formats
   if {[string range $data 0 2] == "###"} {
      # New style
      if {[catch {eval $data} result]} {
         tk_messageBox -icon warning -type ok -title "Open Error" \
         -message "Error interpreting $fileName\n$::errorInfo"
         return -code error $result
      }
   } else {
      # Old style
      array set snackAmpSettings $data
   }
}
#------------------------------------------------------------------------------
# Function   : settings::buildFileExtenstions
# Description: Create the lists of file extensions needed by snackAmp
# Author     : Tom Wilkason
# Date       : 11/15/2001
#------------------------------------------------------------------------------
proc settings::buildFileExtenstions {args} {
   global TA
   unset -nocomplain TA(extensions)
   foreach {extension} $::snackAmpSettings(fileExtensions) {
      # force an extension letter
      set ext ".[string toupper [string trimleft $extension .]]"
      lappend TA(extensions) $ext
   }
   # Make sure usk is there
   lappend TA(extensions) ".USK"
   # sorted lists search faster
   set TA(extensions) [lsort -unique $TA(extensions)]
}

#------------------------------------------------------------------------------
# Save the width of a window in snackAmpSettings
# Used by ID3 editors so retain the use specified width between uses
#------------------------------------------------------------------------------
proc settings::pushWidth {window tag} {
   global snackAmpSettings
   scan [wm geometry $window] "%dx%d+%d+%d" w h x1 x2
   set snackAmpSettings(size,$tag) $w
}

#------------------------------------------------------------------------------
# Restore the width of a window from snackAmpSettings
#------------------------------------------------------------------------------
proc settings::restoreWidth {window tag} {
   global snackAmpSettings
   wm geometry $window {}
   update
   if {[info exists snackAmpSettings(size,$tag)]} {
      scan [wm geometry $window] "%dx%d+%d+%d" w h x1 x2
      set w $snackAmpSettings(size,$tag)
      wm geometry $window "${w}x${h}"
   }
}
proc settings::pushSize {window tag} {
   global snackAmpSettings
   scan [wm geometry $window] "%dx%d+%d+%d" w h x1 x2
   set snackAmpSettings(size,$tag) "${w}x${h}"
}
#------------------------------------------------------------------------------
# Restore the width of a window from snackAmpSettings
#------------------------------------------------------------------------------
proc settings::restoreSize {window tag} {
   global snackAmpSettings
   wm geometry $window {}
   update
   if {[info exists snackAmpSettings(size,$tag)]} {
      set size $snackAmpSettings(size,$tag)
      wm geometry $window $size
   }
}
#------------------------------------------------------------------------------
#  Slide the window off screen
#------------------------------------------------------------------------------
proc settings::Hide {W} {
   global TA
   scan [wm geometry $W] "%dx%d+%d+%d" w h x1 x2
   set TA($W,x1) $x1
   set TA($W,x2) $x2
   wm geometry $W +-5000+-5000
   return 1
}

#------------------------------------------------------------------------------
#  Slide the window back on screen
#------------------------------------------------------------------------------
proc settings::Restore {W} {
   global TA
   wm geometry $W "+$TA($W,x1)+$TA($W,x2)"
}


#------------------------------------------------------------------------------
# If no defaults present copy some to user settings folder
#------------------------------------------------------------------------------
proc settings::createUserDefaults {args} {
   global TA
   set source $TA(snackRoot)
   set dest $::HOME
   foreach {type} {ss scs} {
      set dests [glob -nocomplain -directory $dest *.$type]
      # If less  2 on target, assume new or < 3.1
      # snackamp install, then copy files
      if {[llength $dests] < 2} {
         foreach {sfile} [glob -nocomplain -directory $source *.$type] {
            set target [file join $dest [file tail $sfile]]
            if {![file exists $target]} {
               file copy -force $sfile $target
               saLog "Created default config file: $target"
            }
         }
      }
   }
}
#------------------------------------------------------------------------------
# Function   : bindings namespace
# Description: Handle key binding functions for displaying shortcuts
# Author     : Tom Wilkason
# Date       : 6/15/2002
#------------------------------------------------------------------------------
namespace eval bindings {
   ##
   # save the shortcut key bindings for all the widgets
   #
   proc showBindings {args} {
      foreach window [list .saTop .wFC $::FI $::LB $::TREE [winfo parent $::TREE]] {
         if {[winfo exists $window]} {
            foreach key [lsort [bind $window]] {
               if {[string match <<*>> $key]} {
                  set key  [cleanKeyName [lindex [event info $key]  0]]
               }
               set command [string trim [bind $window $key]]
               if {[llength $command] < 3} {
                  puts "  [cleanKeyName $key] = '$command'"
                  set ::TA(binding,$window,$command) [cleanKeyName $key]
               }
            }
         }
      }
   }
   ##
   # Internal function to lookup a binding
   #
   proc Accelerator {W command} {
      variable Data
      set command [string trim $command]
      if {[info exists Data($command)]} {
         return $Data($command)
      } else {
         return ""
      }
   }
   ##
   # Init the lookup table for some menu
   #
   proc init {win _Data} {
      variable Data
      variable W $win
      foreach key [bind $W] {
         set boundCommand [string trim [bind $W $key]]
         if {[string match <<*>> $key]} {
            set key  [cleanKeyName [lindex [event info $key]  0]]
            set Data($boundCommand) $key
         } elseif {[llength $boundCommand] < 5} {
            set Data($boundCommand) [cleanKeyName $key]
         }
      }
      return ""
   }
   ##
   # Add a command menu with the appropriate accelerator shortcut appended
   #
   proc addCommand {Menu Label Command {underline ""}} {
      # TODO: Have array of images to use, create a lookup based on a switch statement
      # and use that image, if image is blank don't use one
      variable W
      set image [getImage $Label]
      if {$underline==""} {
         if {[string length $image]} {
            $Menu add command -label $Label \
            -accelerator [bindings::Accelerator $W [string trim $Command]] \
            -command $Command -compound left -image $image
         } else {
            $Menu add command -label $Label \
            -accelerator [bindings::Accelerator $W [string trim $Command]] \
            -command $Command
         }
      } else {
         if {[string length $image]} {
            $Menu add command -label $Label \
               -accelerator [bindings::Accelerator $W [string trim $Command]] \
               -command $Command -underline $underline -compound left -image $image
         } else {
            $Menu add command -label $Label \
               -accelerator [bindings::Accelerator $W [string trim $Command]] \
               -command $Command -underline $underline
         }
      }
   }
}

#------------------------------------------------------------------------------
# Return an image based on the label
#------------------------------------------------------------------------------
proc bindings::getImage {Label} {
   switch -glob -- $Label {
      "Save*as*"       {return saImages.SaveAs}
      "Save*to*"       {return saImages.SaveAs}
      "*Home Page*"    {return saImages.web}
      "+*"             {return saImages.iconthumbsup}
      "-*"             {return saImages.iconthumbsdown}
      "Apply*"         {return saImages.Apply}
      "*AutoPlaylist*" {return saImages.smtreeImage}
      "About*"         {return saImages.info}
      "Append*to*"     {return saImages.SaveAs}
      "Append*from*"   {return saImages.appendImage}
      "Append*tracks*"   {return saImages.appendImage}
      "Close*"         {return saImages.Close}
      "Copy*"          {return saImages.copy}
      "Clear*"         {return saImages.Close}
      "Cut*"           {return saImages.cut}
      "Decrease*"      {return saImages.iconthumbsdown}
      "Delete*"        {return saImages.delete}
      "Edit*"          {return saImages.edit-16}
      "Exit*"          {return saImages.Close}
      "Guess*"         {return saImages.Guess}
      "ID3*"           {return saImages.edit-16}
      "Increase*"      {return saImages.iconthumbsup}
      "Internal*"      {return saImages.web}
      "Manage*"        {return saImages.preferences}
      "Open*"          {return saImages.openFile}
      "Paste*"         {return saImages.paste}
      "Play*"          {return saImages.player_play}
      "Preferences*"   {return saImages.preferences}
      "Properties*"    {return saImages.info}
      "Queue*"         {return saImages.queue}
      "Remove Selection" {return saImages.Close}
      "Remove*"        {return saImages.delete}
      "Rename*"        {return saImages.edit-16}
      "Save*"          {return saImages.Save}
      "Shuffle*"       {return saImages.shuffleImage}
      "Sort*"          {return saImages.sortImage}
      "Track Info*"    {return saImages.info}
      "User*"          {return saImages.web}
      "Versions*"      {return saImages.info}
      "Check*"      {return saImages.info}
      default {return ""}
   }
}
#------------------------------------------------------------------------------
# Function   : settings::assignHotKeys
# Description: Init routine to process the user hot keys file.
# Author     : Tom Wilkason
# Date       : 2/23/2001
#------------------------------------------------------------------------------
proc bindings::assignHotKeys {args} {
   global snackAmpSettings TA
   ##
   # If this is a wrapped file, create the hotkey file if it doesn't already exist
   #
   if {$TA(isWrapped) && ![file exists $snackAmpSettings(hotkeyFile)]} {
      set vfsHotkeyFile [file join $TA(snackRoot) snackAmphotKeys.tcl]
      if {![file exists [file dirname $snackAmpSettings(hotkeyFile)]]} {
         set snackAmpSettings(hotkeyFile) [file join $::HOME snackAmphotKeys.tcl]
      }
      if {[catch {
         set fin [open $vfsHotkeyFile r]
         set fout [open $snackAmpSettings(hotkeyFile) w]
         puts -nonewline $fout [read $fin]
         close $fin
         close $fout
         waitWindow::close
         set result [tk_messageBox -title "Create Hotkey File" \
            -icon info -message "I Created a new hotkey file [file nativename $snackAmpSettings(hotkeyFile)]\n you can edit it in a text editor if so desired."]
      } result] } then {
         saLog $::errorInfo
      }
   }
   if {([winfo exists .saTop]) && ([file exists $snackAmpSettings(hotkeyFile)])} {
      if {[catch {source $snackAmpSettings(hotkeyFile)} result]} {
         waitWindow::close
         set result [tk_messageBox -title "Hotkey Error" \
            -icon info -message "$::errorInfo"]
      } else {
         saLog "Using hotkey file of $snackAmpSettings(hotkeyFile)"
      }
   }
}
#------------------------------------------------------------------------------
# Function   : settings::defaultHotKeys
# Description: Default hot keys, user can override
# Author     : Tom Wilkason
# Date       : 2/23/2001
#------------------------------------------------------------------------------
proc bindings::defaultHotKeys {args} {
#------------------------------------------------------------------------------
#  Playlist Assignments
#------------------------------------------------------------------------------
   PlaylistAssign <BackSpace> {Cut 1}
   PlaylistAssign <Delete> {Cut 1}
   PlaylistAssign <Control-Key-Up> {
      Swap up
      break
   }
   PlaylistAssign <Control-Key-Down> {
      Swap down
      break
   }
   PlaylistAssign <Control-Key-a> {selectAll}
   PlaylistAssign <Return> {Play}
   PlaylistAssign <<Copy>> {clipboardCopy $::LB 0}
   PlaylistAssign <<Cut>> {clipboardCut $::LB 0}
   PlaylistAssign <<Paste>> {clipboardPaste $::LB 1}
   PlaylistAssign <Control-Key-o> {sortSelection}
   PlaylistAssign <Control-Key-s> {shuffleSelection}
   ##
   # Default Assignments for Playlist
   # Note: Add break as as last command so
   # prevent rolling up to top level.
#------------------------------------------------------------------------------
#  TOP Level Assignments
#------------------------------------------------------------------------------
   #Assign <<Copy>> {clipboardCopy $::LB 0}
   #Assign <<Cut>> {clipboardCut $::LB 0}
   #Assign <<Paste>> {clipboardPaste $::LB 1}
   Assign <F1> {::browser::help}
   Assign <Control-Key-f> {findNextSong 1}
   Assign <Control-Key-g> {findNextSong 0}
   Assign <Alt-Key-Up>    {bumpLevel 1;break}
   Assign <Alt-Key-Down>  {bumpLevel -1;break}
   Assign <Key-Home>   locatePlaying
   Assign <F2>         {RenameFile [LB::nameActive]}
   Assign <Control-F2> {RenameFile [soundControl::fileName]}
   # Keys for editing ID3
   Assign <F3>         {EditPLID3}
   Assign <Control-F3> {EditPlaying}
   Assign <F4>         {PLNameToTag 0}
   Assign <Alt-F4>     {taExit}   ;# If not added the F4 binding fires, windows unique?
   Assign <F5>         {SelectFile 0}
   Assign <Control-F5> {SelectFile 1}
   Assign <F6>         {PLTagToName}
#------------------------------------------------------------------------------
#  File Box Assignments
#------------------------------------------------------------------------------
#  FilterAssign <Double-1> macros::clearFilter
   FilterAssign <Escape> clearFilter
   FilterAssign <Return> {filterMusic replace}
   FilterAssign <Double-1> {
      clearFilter
      filterMusic replace
   }
   FilterAssign <Control-Return> {
      filterMusic replace
   }
   FilterAssign <Alt-Return> {
      filterMusic append
   }
}
#------------------------------------------------------------------------------
# Function   : folderManager
# Description: Pop-up box to manage the folders
# Author     : Tom Wilkason
# Date       : 12/17/2001
#------------------------------------------------------------------------------
namespace eval folderManager {
   variable fmLB
   variable Index
   variable allFolder "<All Folders>"
}
#------------------------------------------------------------------------------
# Function   : folderManager::launch
# Description: Launch the folder
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc folderManager::launch {_Index} {
   variable fmLB
   variable Index
   variable allFolder
   set Index $_Index
   catch {destroy .fm}
   toplevel .fm -cursor $::TA(arrow)  -class SnackAmp
   wm withdraw .fm
   grid [frame .fm.list] [frame .fm.button] -sticky ns
   wm title .fm "Media Library Manager"
   set fmLB [Scrolled_Listbox .fm.list.lb -exportselection no -height 8 -width 40 \
      -foreground $::snackAmpSettings(fg) \
      -background $::snackAmpSettings(bg) \
      -selectforeground $::snackAmpSettings(bg) \
      -selectbackground $::snackAmpSettings(fg)]
   # Listbox popup menu
   settings::regColor $fmLB  \
      -foreground       ::snackAmpSettings(fg) \
      -background       ::snackAmpSettings(bg) \
      -selectforeground ::snackAmpSettings(bg) \
      -selectbackground ::snackAmpSettings(fg)
   set m $fmLB.popup
   menu $m
   $m add command -label "Discard Folder"  -command {}
   $m add command -label "Add Folder"  -command {}
   pack .fm.list.lb -fill y -expand true
   lappend buttons add    "Add Folder"     folderManager::Add    "Add a new folder to the list on the left"
   if {$Index == "mediaDirs" || $Index == "mp3Player"} {
      $fmLB insert 0 $allFolder
      lappend buttons ind "Index Folder"      folderManager::Index         "Index selected folder and save track file names (fast)"
      lappend buttons cat "Catalog Folder"    folderManager::Catalog       "Catalog selected folder and save track data. SnackAmp will automatically catalog tracks when it needs to, however performing this function up front will speed up SnackAmp operation."
      lappend buttons atl "Auto-Level Folder" folderManager::Auto-Level    "Auto-Level the selected media folder. SnackAmp will automatically auto-level tracks when it needs to, however performing this function up front will speed up SnackAmp operation."
      lappend buttons epc "Export Catalog"    folderManager::exportCatalog "Export a catalog for this folder"
      lappend buttons ipc "Import Catalog"    folderManager::importCatalog "Import a previously exported catalog for this folder"
      if {$Index == "mp3Player"} {
         lappend buttons apl "APL->Playlists"  folderManager::snapshotAllAPLs "Build m3u playlist files here based on all Autoplaylist settings"
      }
   }
   lappend buttons \
    change "Change Folder"  folderManager::Edit   "Modify the selected folder on the left"                                         \
    remove "Discard Folder" folderManager::Remove "Remove the selected folder from the list (does not modify the folder contents)" \
    ok     "Close"          folderManager::OK     "Push this to apply and save the current settings then close this dialog"
#    cancel "Cancel"         {destroy .fm}         "Push this to ignore changes made and restore the previous settings"
   # Buttons
   set i 0
   foreach {but Text Command Help} $buttons {
      button .fm.button.$but -text $Text -command $Command -relief raised -pady 0 -default normal
      balloon::define .fm.button.$but $Help
      lappend toPack .fm.button.$but
   }
   # Insert current data into box
   eval pack $toPack -fill x
   if {[llength $::snackAmpSettings($Index)] > 0} {
      eval $fmLB insert end $::snackAmpSettings($Index)
   }
   $fmLB selection clear 0 end
   $fmLB selection set 0
   $fmLB activate 0
   ##
   # Default bindings (the -default active doesn't seem to work on windows)
   #
   bind .fm <Return> [list .fm.button.ok invoke]
   .fm.button.ok configure -default active
   bind .fm <Escape> {destroy .fm}

   catch {wm transient .fm .saTop}

   centerWindow .fm
   myIcon .fm
   lockSize .fm
   # Block on window closure
   ###grab set .fm
   tkwait window .fm
   ###grab release .fm
}
#------------------------------------------------------------------------------
# Function   : folderManager::OK
# Description: OK Button
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc folderManager::OK {args} {
   variable fmLB
   variable Index
   variable allFolder
   # TFW Note: This causes blanks in elements to be surrounded by {}
   # Which previously they may have been surrounded by blanks.
   set ::snackAmpSettings($Index) [list]
   foreach {row} [$fmLB get 0 end] {
      if {$row ne $allFolder} {
         lappend ::snackAmpSettings($Index) $row
      }
   }
   destroy .fm
}
#------------------------------------------------------------------------------
# Function   : folderManager::Add
# Description: Add a folder
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc folderManager::Add {args} {
   variable fmLB
   variable Index
   set result [chooseDirectory . .]
   if {[file exists $result]} {
      $fmLB insert end $result
      $fmLB selection clear 0 end
      $fmLB selection set end
      $fmLB activate end
      set dbFile [file join $result "snackDB.mke"]
      if {[file exists $dbFile]} {
         set res [tk_messageBox -type yesno -icon question -message \
            "Import file detected\n\nWould you like to import existing track information on '$result' now?"]
         if {$res eq "yes"} {
             if {[importCatalog $result]} {
               return
             }
         }
      }

      ;# Ask user if they want to index it
      set res [tk_messageBox -type yesno -icon question -message \
         "Would you like to index the tracks on '$result' now?"]
      if {$res=="yes"} {
         folderManager::Index
         if {$::snackAmpSettings(alwaysCatalog)==0} {
            set res [tk_messageBox -type yesno -icon question -message \
               "Cataloging can be done now or an as needed basis by SnackAmp\nWould you like to catalog the tracks on '$result' now?"]
            if {$res=="yes"} {
               folderManager::Catalog 1
            }
         }
      }
   }
}
#------------------------------------------------------------------------------
# Function   : folderManager::Edit
# Description: Change an existing folder along with database entries
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc folderManager::Edit {args} {
   variable fmLB
   variable Index
   upvar ::snackAmpSettings($Index) PATHS
   set pos [$fmLB index active]
   set oldBranch [$fmLB get active]
   set newBranch [chooseDirectory $oldBranch .]
   if {[file exists $newBranch]} {
      $fmLB delete $pos
      $fmLB insert $pos $newBranch
      $fmLB selection clear 0 end
      $fmLB selection set $pos
      $fmLB activate $pos
      # rename database entries
      waitWindow::show "Renaming Database Entries..."
      set plchanged [db::renameBranch $oldBranch $newBranch]
      # change entries in the media dirs
      set index [lsearch -exact $PATHS $oldBranch]
      if {$index >= 0} {
         lset PATHS $index $newBranch
      }
      cleanPathList PATHS 0
      catalog::validateMediaPaths
      # rename playlists and autoplaylists
      # Change all files that reference these files
      if {[llength $plchanged]} {
         ##
         # Create a matching list of folders for APL changing
         #
         set oldBranch "[string trimright $oldBranch /]/"
         set newBranch "[string trimright $newBranch /]/"
         waitWindow::show "Updating Playlist Files..."
         # Playlist Files
         set Files [catalog::getPlaylists]
         catalog::RenameFiles $Files [list $oldBranch $newBranch]
         # Change all visible references to these files
         waitWindow::show "Updating Playlist References..."
         # need full file list                         .
         LB::changePlaylistFiles $plchanged

         waitWindow::show "Updating AutoPlaylist Files..."
         set Files [catalog::getAutoPlaylists]
         catalog::RenameFiles $Files [list $oldBranch $newBranch]

         # Rename the current folder settings as well
         aplFolder::renameBranch $oldBranch $newBranch
      }
      # Rebuild the entire tree
      if {[APLExists]} {
         APL::reBuildTree
      }
      waitWindow::close
   }
}
#------------------------------------------------------------------------------
# Function   : folderManager::Index
# Description: Catalog a folder
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc folderManager::Index {args} {
   variable fmLB
   variable Index
   variable allFolder
   set result [$fmLB get [$fmLB index active]]
   if {[string equal $result $allFolder]} {
      catalog::IndexAskFirst
   } else {
      if {[file exists $result]} {
         catalog::Index $result
      } else {
         tk_messageBox -icon warning -type ok -message "$result does not exist!"
      }
   }
}
#------------------------------------------------------------------------------
# Function   : folderManager::Catalog
# Description: Catalog a folder
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc folderManager::Catalog {{bypass 0}} {
   variable fmLB
   variable Index
   variable allFolder
   set result [$fmLB get [$fmLB index active]]
   if {$bypass} {
      set res "yes"
   } else {
      set res [tk_messageBox -type yesno -icon question -message \
         "Cataloging reads the track and ID3 data from each track and saves it for later use.\nSnackAmp can also do this on an as needed basis.\n\nWould you like to catalog the tracks in '$result' anyway?"]
   }
   if {$res=="yes"} {
      if {[string equal $result $allFolder]} {
         waitWindow::show "Cataloging all media folders" 1
         db::catalog "" 2 100
      } else {
         if {[file exists $result]} {
            waitWindow::show "Cataloging\n$result" 1
            db::catalog $result 2 100
         } else {
            tk_messageBox -icon warning -type ok -message "$result does not exist!"
         }
      }
      waitWindow::close
   }
}
#----------------------------------------------------------------------------
# Export playlists for all APL settings
#----------------------------------------------------------------------------
proc folderManager::snapshotAllAPLs {{bypass 0}} {
   variable fmLB
   variable Index
   variable allFolder
   set result [$fmLB get [$fmLB index active]]
   if {$bypass} {
      set res "yes"
   } else {
      set res [tk_messageBox -type yesno -icon question -message \
         "Would you like to create m3u playlist files for each track contained in an Autoplaylist for '$result'"]
   }
   if {$res=="yes"} {
      if {[string equal $result $allFolder]} {
         foreach {path} $::snackAmpSettings(mp3Player) {
            waitWindow::show "Cataloging all media folders" 1
            macros::snapshotAllAPLs * "$result*"
         }
      } else {
         waitWindow::show "Cataloging\n$result" 1
         macros::snapshotAllAPLs * "$result*"
      }
      waitWindow::close
   }
}
#----------------------------------------------------------------------------
# Build playlist from APL Entries
#----------------------------------------------------------------------------
proc folderManager::buildPlaylist {{folder {}}} {
   set tracks [db::filesInFolder $folder 20]
   set pltracks [list]
   set len [string length $folder]
   foreach {track} $tracks {
      set fd [file dirname $track]
      if {[aplFolder::isSet $fd]} {
         lappend pltracks [string range $track $len end]
      }
   }
   if {[llength $pltracks] > 0} {
      set plf [file join $folder [file rootname [file tail $::snackAmpSettings(lastFolderFile)]].m3u]
      if {[catch {open $plf w} fid] } then {
         saLog $fid
         tk_messageBox -icon warning -type ok -title "Notice" -message "Error '$fid'\nwriting [llength $pltracks] tracks to $plf"
      } else {
         puts $fid [join $pltracks \n]
         tk_messageBox -icon info -type ok -title "Notice" -message "[llength $pltracks] tracks written to $plf"
         close $fid
      }
   }
}

#------------------------------------------------------------------------------
#  Import a catalog folder
#------------------------------------------------------------------------------
proc folderManager::importCatalog {{folder {}}} {
   global snackAmpSettings
   variable fmLB
   variable Index
   variable allFolder
   if {[string length $folder]==0} {
      set result [$fmLB get [$fmLB index active]]
   }  else {
      set result $folder
   }
   set ans [tk_messageBox -type yesno -title "SnackAmp - Please Respond" \
      -icon info -message \
      "Are you sure you want to import $result? into the database, some existing entries will be replaced?"]
   if {$ans != "yes"} {
      return 0
   }
   Wait 1
   set setAPL [aplFolder::entries] ;# flag if we have some entries (the APL has been launched)
   set imported ""
   set csvLocation [file tail $db::csvLocation]
   # If process all folders, do in a loop
   if {[string equal $result $allFolder]} {
      foreach {path} [catalog::mountedFolders] {
         if {[file exists [file join $path $csvLocation]]} {
            foreach {im rp rm} [db::importOnePath $csvLocation $path] {break}
            append imported "\n$path - [commify $im] Tracks Added, [commify $rp] Replaced, [commify $rm] Removed"
            if {$setAPL} {
               set saveNames [aplFolder::savePattern $path*]  ;# want folder also
               aplFolder::stripBranch $path*
               waitWindow::show "Restoring Folders ..."
               foreach {ind} [db::folderList [file join $path *]] {
                  aplFolder::setToRoot $ind 0
               }
               aplFolder::restoreNames $saveNames
            }
            Update
         } else {
            append imported "\nNo import file for $path"
         }
      }
   } else {
      # Otherwise, just process this one folder
      if {[file exists [file join $result $csvLocation]]} {
         foreach {im rp rm} [db::importOnePath $csvLocation $result] {break}
         if {$setAPL} {
            Update
            set saveNames [aplFolder::savePattern $result*]  ;# want folder also
            aplFolder::stripBranch [file join $result *]
            waitWindow::show "Restoring Folders ..."
            foreach {ind} [db::folderList [file join $result *]] {
               aplFolder::setToRoot $ind 0
            }
            aplFolder::restoreNames $saveNames
         }
         append imported "\n$result - [commify $im] Tracks Added, [commify $rp] Replaced, [commify $rm] Removed"
      } else {
         append imported "\nNo import file for $result"
      }
   }
   # rebuild APL folders
   # TODO: This is setting new ones in the APL tree
   # the teFolder settings aren't set for the new imported folders
   Update
   # Rebuild the entire tree
   if {[APLExists]} {
      APL::reBuildTree
   }
   Wait 0
   waitWindow::close
   tk_messageBox -title "Import Complete" -message "Imported$imported" -type ok
   return 1
}
#------------------------------------------------------------------------------
#  Expor a catalog folder
#------------------------------------------------------------------------------
proc folderManager::exportCatalog {{folder {}}} {
   variable fmLB
   variable Index
   variable allFolder
   Wait 1
   if {[string length $folder]==0} {
      set result [$fmLB get [$fmLB index active]]
   }  else {
      set result $folder
   }
   set exported ""
   set csvLocation [file tail $db::csvLocation]
   if {[string equal $result $allFolder]} {
      foreach {path} [catalog::mountedFolders] {
         set recs [db::exportOnePath [file join $path $csvLocation] $path]
         append exported "\n$recs tracks for $path to [file join $path $csvLocation]"
         Update
      }
   } else {
      set recs [db::exportOnePath [file join $result $csvLocation] $result]
      append exported "\n$recs tracks for $result to [file join $result $csvLocation]"
   }
   Wait 0
   waitWindow::close
   tk_messageBox -title "Export Complete" -message "Exported$exported\n\nYou may import this data into another SnackAmp database." -type ok
}
# ----------------------------------------------------------------------------
# Auto-Level the folder and its contents
# ----------------------------------------------------------------------------
proc folderManager::Auto-Level {args} {
   variable fmLB
   variable Index
   variable allFolder
   set result [$fmLB get [$fmLB index active]]
   set res [tk_messageBox -type yesno -icon question -message \
      "Auto-Leveling computes gain factors for each track, no files are modified.\nSnackAmp can also do this on an as needed basis.\n\nWould you like to precompute the values for '$result' anyway?"]
   if {$res=="yes"} {
      set all [tk_messageBox -type yesno -icon question -message \
         "Enter Yes to compute only needed tracks, No to recompute all tracks?"]
      if {[string equal $result $allFolder]} {
         soundControl::calibrateAudio $all
      } else {
         if {[file exists $result]} {
            gainControl::computeSome [db::filesMatchingPattern [file join $result *]] $all
         } else {
            tk_messageBox -icon warning -type ok -message "$result does not exist!"
         }
      }
   }
}
#------------------------------------------------------------------------------
# Function   : folderManager::Remove
# Description: Remove an existing folder
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc folderManager::Remove {args} {
   variable fmLB
   variable Index
   variable allFolder
   set pos [$fmLB index active]
   set result [$fmLB get $pos]
   set res [tk_messageBox -type yesno -icon question -message \
      "Would you like to remove the track information for '$result' from the database?"]
   if {$res=="yes"} {
      Wait 1
      if {$result==$allFolder} {
         set res [tk_messageBox -type yesno -icon warning -message \
            "Are you sure you want to clear the entire database?"]
         if {$res=="yes"} {
            set num [db::removeBranch "*"]
            aplFolder::stripBranch "*"
            tk_messageBox -type ok -icon info -message "A total of $num tracks were removed."
         }
      } else {
         set num [db::removeBranch [file join $result *]]
         aplFolder::stripBranch [file join $result *]
         tk_messageBox -type ok -icon info -message "A total of $num tracks were removed."
      }
      unset -nocomplain ::snackAmpSettings(mediaMode,$result)
      # Always remove the entries
      if {$result==$allFolder} {
         $fmLB delete 1 end
         #
         if {$Index eq "mediaDirs"} {
            set ::TA(mediaDirs) [catalog::attachedFolders 1]
         } else {
            set ::TA(mediaDirs) [catalog::fixedFolders 1]
         }
         set mediaDirs [list]
         set ::snackAmpSettings($Index) $mediaDirs
      } else {
         $fmLB delete $pos
         set mediaDirs [list]
         # Remove this entry
         lextract ::snackAmpSettings($Index) $result
         catalog::validateMediaPaths
      }
   }
   Wait 0
}
# ----------------------------------------------------------------------------
# Namespace to handle saving and restoring color schemes
# ----------------------------------------------------------------------------
namespace eval schemeSettings {
   variable schemeSelection ""
   variable colors {bg fg highlight tsbg ebg efg afg tfg lfg sfg}
   variable fontkeys {family weight slant size}

}
proc schemeSettings::saveas {args} {
   set name ""
   if {[getEntryDialog "Enter Name" name "Color Scheme"]=="ok"} {
      saveFile $name
   }
}
##
# Write a settings file
#
proc schemeSettings::saveFile {{name ""}} {
   variable schemeSelection
   variable settings
   global snackAmpSettings
   fetch
   if {$name==""} {
      if {[getEntryDialog "Enter Name" name "Color Scheme"]!="ok"} {
         return 0
      }
   }
   set fname [file join $::HOME $name.scs]
   # Check for file exists
   if {[catch {
      set fid [open $fname w]
   } result] \
   } then {
      tk_messageBox -icon warning -type ok -title "Open Error" -message $result
      return 0
   } else {
      puts $fid [array get settings]
      close $fid
      set schemeSelection $name
      set snackAmpSettings(currentColorScheme) $name
      apply $name
      return 1
   }
}
##
# Copy current settings to snackAmpDefaults
#
proc schemeSettings::apply {args} {
   global snackAmpSettings
   variable settings
   variable colors
   variable fontkeys
   foreach color $colors {
      set snackAmpSettings($color) $settings($color)
   }
   foreach key $fontkeys {
      array set snackAmpSettings [array get settings *,$key]
   }
   # Apply the colors and settings
   settings::applyColor
   settings::saFontSize $snackAmpSettings(FontSizeFactor)

}
proc schemeSettings::save {args} {
   variable schemeSelection
   saveFile $schemeSelection
}
proc schemeSettings::remove {args} {
   global snackAmpSettings
   variable schemeSelection
   if {[removeFile $schemeSelection]} {
      set schemeSelection ""
      set snackAmpSettings(currentColorScheme) ""
   }
}
proc schemeSettings::removeFile {{name ""}} {
   variable schemeSelection
   variable settings
   set fname [file join $::HOME $name.scs]
   if {[file exists $fname]} {
      file delete -force $fname
   }
   if {[file exists $fname]} {
      return 0
   } else {
      return 1
   }
}
##
# Read a settings file into settings
#
proc schemeSettings::readFile {name} {
   global snackAmpSettings
   variable settings
   variable schemeSelection
   set fname [file join $::HOME $name.scs]
   if {[catch {
      set fid [open $fname r]
   } result] \
   } then {
      tk_messageBox -icon warning -type ok -title "Open Error" -message $result
      return 0
   } else {
      set len [gets $fid row]
      close $fid
      if {$len > 0} {
         array set settings $row
         set schemeSelection $name
         set snackAmpSettings(currentColorScheme) $name
         apply $name
         return 1
      }  else {
         return 0
      }
   }
}
##
# Get the current snackAmpSettings (during init)
#
proc schemeSettings::fetch {args} {
   global snackAmpSettings
   variable settings
   variable colors
   variable fontkeys
   foreach color $colors {
      set settings($color) $snackAmpSettings($color)
   } ;# end foreach
   foreach key $fontkeys {
      array set settings [array get snackAmpSettings *,$key]
   } ;# end foreach
}
proc schemeSettings::comboPopulate {menu} {
   variable schemeSelection
   set choices [schemeSettings::getFiles]
   $menu configure -values $choices
}
proc schemeSettings::populate {menu} {
   variable schemeSelection
   $menu delete 0 end
   ##
   #  Extended Options
   #
   set choices [schemeSettings::getFiles]
   foreach {choice} $choices {
      $menu add radiobutton -label $choice -variable schemeSettings::schemeSelection -command {
         schemeSettings::readFile $schemeSettings::schemeSelection
      }
   }
}
##
# Return a list of all subsorting file names
#
proc schemeSettings::getFiles {args} {
   variable settings
   set result [list]
   foreach name [lsort [glob -nocomplain -directory $::HOME *.scs]] {
      lappend result [file tail [file root $name]]
   }
   return $result
}
package provide saSettings 1.0

