#------------------------------------------------------------------------------
# trackList Package 1.0
#
# trackList functions for snackAmp Player in Tcl/Tk
# used by the autoplaylist manager
#
# Copyright (C) 2006 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
# trackList namespace: Used for table widget
#------------------------------------------------------------------------------
namespace eval trackList {
   variable col2Use [list]
   variable lastFolder ""
   variable editCellButton
   variable editCellState 0
   variable editCellCheckState 0
   variable TABLEDATA [list]
   variable butrel
   variable activeFile ""
   trace variable trackList::activeFile w {after 10 [list APLcover::packFileCoverArt $trackList::activeFile]}

   array set butrel {0 sunken 1 flat}
   package require Tablelist
   # work on using combo boxes for some colums
   #tablelist::addBWidgetComboBox combo
}
#------------------------------------------------------------------------------
# Schedule a rescan when all is said and done
# to reflect the new configuration
#------------------------------------------------------------------------------
proc trackList::populateHighighted {sort insert {delay "idle"}} {
   afterCancel trackList::populate
   after $delay [list trackList::populate $::TA(highlightPath) $::PLB $sort $insert]
}
#------------------------------------------------------------------------------
# Return a file name based on row index
# Col 0 has a hidden column with original key values
# which are maintained with the table is sorted.
# use the value to lookup the correct file name
#------------------------------------------------------------------------------
proc trackList::rowIndex2file {index} {
   variable TABLEDATA
   return [db::TrackFromDb [lindex $TABLEDATA [list $index 0]]]
}
#------------------------------------------------------------------------------
# Update the track list for all incoming file names, they should
# have already been updated in the datbase
#------------------------------------------------------------------------------
proc trackList::renameFiles {indexFiles} {
   Trace
   variable TABLEDATA
   ##
   # Update APL oldFile List (if visible in list)
   #
   if {[APLExists] && [info exists TABLEDATA]} {
      ;# Build list of indicies for fast search
      set indicies [list]
      foreach row $TABLEDATA {
         lappend indicies [lindex $row 0]
      }
      set indicies [lsort $indicies]

      # Only do if "file" column is visible
      set fileIndex [expr {[lsearch -exact [$::PLB cget -columns] "File"]/3}]
      if {$fileIndex > 0} {
         set ri 0
         foreach {oldIndex oldFile newFile} $indexFiles {
            # Make sure this file is displayed in the APL
            # A slight speed improvement for longer lists
            if {[lsearch -exact -sorted $indicies $oldIndex]<0} {
               continue
            }
            # Get the new index, it may have changed due to auto sorting
            set newindex [db::whereInDb $newFile]
            if {$newindex < 0} {
               saLog "Could not find $newFile in database"
               continue
            }
            set newTail [file tail $newFile]
            set oldTail [file tail $oldFile]
            set ri 0
            foreach row $TABLEDATA {
               set key [lindex $row  0]
               set tail [lindex $row $fileIndex]
               # if this key and file tail match, change this entry
               if {$key==$oldIndex && [string match $tail $oldTail]} {
                  setVal $ri 0          $newindex
                  setVal $ri $fileIndex $newTail
                  break ;# can only have one
               }
               incr ri
            }
         }
      }
   }
}
#----------------------------------------------------------------------------
# Update APL data for each track passed in
# They should have already been updated in the datbase
#----------------------------------------------------------------------------
proc trackList::updateFiles {indexFiles} {
   variable TABLEDATA
   ##
   # Update APL data for each track passed in
   #
   if {[APLExists] && [info exists TABLEDATA]} {
      # Only do if "file" column is visible
      foreach {xx col xx} [$::PLB cget -columns] {lappend cols $col}
      foreach {newFile} $indexFiles {
         # Get the new index, it may have changed due to auto sorting
         set index [db::whereInDb $newFile]
         if {$index < 0} {
            saLog "Could not find $newFile in database"
            continue
         }
         unset -nocomplain Data
         set ri 0
         # Iterate through displayed row and update as necessary
         # Only lookup data for know files
         foreach row $TABLEDATA {
            set key [lindex $row 0]
            # If this file is being displayed
            if {$key==$index} {
               if {![info exists Data]} {
                  db::lookup $newFile Data
               }
               set ci 0
               foreach {col} $cols {
                  if {[info exists Data($col)]} {
                     # Some fields from database need to be formatted
                     # date, File, (lookup formats gain)...
                     switch -- $col {
                        "File" {
                           setVal $ri $ci [file tail $Data($col)]
                        }
                        "LastPlay" {
                           if {$Data(LastPlay)==0} {
                              setVal $ri $ci "Not Played"
                           }  else {
                              setVal $ri $ci [timeFormat $Data(LastPlay)]
                           }
                        }
                        default {
                           setVal $ri $ci $Data($col)
                        }
                     }
                  }
                  incr ci
               }
               continue
            }
            incr ri
         }
      }
   }
}

#------------------------------------------------------------------------------
# Map selected row to full file name
#------------------------------------------------------------------------------
proc trackList::row2file {W y} {
   set File {}
   set y [expr {$y + [winfo y $W]}]
   set index [$::PLB nearest $y]
   return [rowIndex2file $index]
}
#----------------------------------------------------------------------------
# Speical handling of certain columns
#----------------------------------------------------------------------------
proc trackList::colClick {W x y X Y} {
   variable activeFile
   variable ROW
   variable COL
   trackList::pushClick $W $x $y
   set activeFile [row2file $W $y]
   set col [$::PLB columncget [$::PLB nearestcolumn [expr {$x + [winfo x $W]}]] -title]
   switch -- $col {
      "Rating" {
         # call the pop-up which uses the ratingPopup postcommand callback below
         tk_popup $::PLB.plbRating $X $Y
      }
      "Genre" {
         # call the pop-up which uses the ratingPopup postcommand callback below
         tk_popup $::PLB.plbGenre $X $Y
      }
      default {}
   }
}
#----------------------------------------------------------------------------
# Pop-up menu to adjust track rating
# This is called from a post command
#----------------------------------------------------------------------------
proc trackList::ratingPopup {menu} {
   variable activeFile
   variable trackRating
   variable TABLEDATA
   variable ROW
   variable COL
   $menu delete 0 end
   #  activeFile set during click
   set me $activeFile
   set trackRating [mrating [db::getValues $me Rating]]
   foreach {n} $db::ratingOptions {
      $menu add radiobutton -label [mrating $n] -value [mrating $n] \
      -variable trackList::trackRating \
      -command "[list db::setVal $me Rating $n]\;[list trackList::setVal $trackList::ROW $trackList::COL [mrating $n]]"
   }
}
#----------------------------------------------------------------------------
# Pop-up menu to adjust genre
# This is called from a post command
#----------------------------------------------------------------------------
proc trackList::genrePopup {menu} {
   variable activeFile
   variable trackRating
   variable TABLEDATA
   variable ROW
   variable COL
   set me $activeFile
   # remove children from last call
   foreach {child} [winfo children $menu] {
      destroy $child
   }
   $menu delete 0 end
   #  activeFile set during click
   set genres [concat $id3Tag::v1Genres $::snackAmpSettings(customGenres)]
   foreach {genre} $genres {
      set first [string index $genre 0]
      lappend genreNames($first) $genre
   }
   foreach {Value} [lsort [array names genreNames]] {
      $menu add cascade -label $Value -menu $menu.f$Value
      menu $menu.f$Value
      foreach {item} $genreNames($Value) {
        $menu.f$Value add command -label "$item" \
         -command [list trackList::setGenre $me $item]
      }
   }
}

#------------------------------------------------------------------------------
# Update the genre from a callback in the tracklist
#------------------------------------------------------------------------------
proc trackList::setGenre {me item} {
   # Only modify if tag already has data
   if {[id3Tag::id3Label $me Data $::snackAmpSettings(ID3ReadOrder)]} {
      set Data(Genre) $item
      id3Tag::id3Modify $me Data
   }
   # Update database either way
   db::setVal $me Genre $item
   trackList::setVal $trackList::ROW $trackList::COL $item
}
#----------------------------------------------------------------------------
# trackList::pushClick
# Save the row/cell index where the last click was
#----------------------------------------------------------------------------
proc trackList::pushClick {W x y} {
   variable ROW
   variable COL
   # W is PLB.body
   foreach {ROW COL} [split [$::PLB nearestcell [expr {$x + [winfo x $W]}] [expr {$y + [winfo y $W]}]] ,] {break}
}
#----------------------------------------------------------------------------
#  trackList::setVal
#
#  Set some row/col value in the displayed data, faster than refreshing
#----------------------------------------------------------------------------
proc trackList::setVal {row col newVal} {
   variable TABLEDATA
   lset TABLEDATA $row $col $newVal
}
#------------------------------------------------------------------------------
# trackList::editCellStart
#
# When a cell is first edited, this callback fires
#
#------------------------------------------------------------------------------
proc trackList::editCellStart {W row col newVal} {
   Trace
   variable oCellValue
   set oCellValue $newVal
   return $newVal
}
#------------------------------------------------------------------------------
# trackList::editCellEnd
#
# When a cell is edited, this callback fires
#------------------------------------------------------------------------------
proc trackList::editCellEnd {W row col newVal} {
   Trace
   variable TABLEDATA
   variable oCellValue
   set col [$W columncget $col -title]
   set file [trackList::rowIndex2file $row]
   switch -exact -- $col {
      Folder {
         # This can be dangerous, disable Folder from editing
         set newFullName [file join $newVal [file tail $file]]
         if {[catalog::RenameFile $file $newFullName]} {
            tk_messageBox -icon warning -message "Could not rename\n$file to \n$newFullName\nCheck console log for error"
            return $oCellValue
         } else {
            # Need to change the tracklist key data, renamng file changed its row number
            # in the database making the database different than the TABLEDATA row info.
            # find the database row of this new track and replace the TABLEDATA instance
            setVal $row 0 [db::whereInDb $newFullName]
         }
      }
      File {
         set newFullName [file join [file dirname $file] $newVal]
         if {[catalog::RenameFile $file $newFullName]} {
            tk_messageBox -icon warning -message "Could not rename\n$file to \n$newFullName\nCheck console log for error"
            return $oCellValue
         } else {
            # Need to change the tracklist key data
            trackList::setVal $row 0 [db::whereInDb $newFullName]
         }
      }
      Title -
      Artist -
      Album -
      Track -
      Media -
      Year -
      Genre -
      Note -
      Genre -
      Comp -
      Desc -
      Comp {
         if {$::snackAmpSettings(ID3ReadOnly)==0} {
            if {[catch {
               set fnameData($col) $newVal
               id3Tag::id3Modify $file fnameData
            } result] } then {
               debug $result
               return $oCellValue
            }
         } else {
            return $oCellValue
         }
      }
      Gain {
         # This needs a conversion first
         gainControl::add $file [lindex $newVal 0]
      }
      Rating {
         db::setVal $file Rating $newVal
      }
      Played {
         db::setVal $file Played $newVal
      }
      default {saLog "Unknown option '$col'"}
   }
   return $newVal
}

#------------------------------------------------------------------------------
# trackList::changeEdit
#
# Change the editing mode of a table
#------------------------------------------------------------------------------
proc trackList::changeEdit {args} {
   variable editCellButton
   variable editCellCheckState
   variable editCellState
   variable butrel
   catch {tablelist::canceleditingSubCmd $::PLB}
   ##
   # set or clear the editable columns
   #
   set i 0
   $editCellButton configure -relief $butrel($editCellState)
   foreach {xx key xx} [$::PLB cget -columns] {
      if {$editCellState} {
         $::PLB columnconfigure $i -editable 0
      } else {
         $::PLB columnconfigure $i -editable $::snackAmpSettings(col,$key,edit)
      }
      incr i
   }
   set editCellState [expr {!$editCellState}]
   set editCellCheckState $editCellState
}

#------------------------------------------------------------------------------
# trackList::buildTable
#
# Builds the widget w as a tablelist widget
# contained in a newly created top-level widget.  Returns the name of the
# tablelist widget.
#------------------------------------------------------------------------------
proc trackList::buildTable {top _config} {
   variable col2Use
   variable TABLEDATA
   upvar $_config config
   # List of possible columns
   set col2Use [list]
   set i 0
   set j 0

   set sortMode [list]
   # Build up col info
   # TFW: Need to do this in saved order
   foreach key $::TA(cols) {
      if {$config(col,$key,show) || $key=="Key"} {
         lappend col2Use $config(col,$key,width) $key $config(col,$key,align)
         lappend sortMode $j $config(col,$key,type)
         incr j
      }
   }
   #
   # Create a scrolled tablelist widget with 5 dynamic-width
   # columns and interactive sort capability within the top-level
   #
   set tf [frame $top]     ;# User should pack in frame
   set tbl $tf.tbl
   set vsb $tf.vsb
   set hsb $tf.hsb
   tablelist::tablelist $tbl \
         -columns $col2Use \
         -labelcommand tablelist::sortByColumn \
         -xscrollcommand [list scrollCmd $hsb  [list grid $hsb -row 1 -column 0 -sticky ew]] \
         -yscrollcommand [list scrollCmd $vsb  [list place $vsb -relx 1.0 -rely 0.0 -anchor ne -relheight 1.0]] \
         -foreground       $Tree::fg \
         -background       $Tree::bg   \
         -selectforeground $::snackAmpSettings(fg) \
         -selectbackground $::snackAmpSettings(highlight) \
         -showseparators   no \
         -borderwidth      1 \
         -selectmode       "extended" \
         -setgrid          no \
         -stripeheight     1\
         -listvariable   trackList::TABLEDATA \
         -editstartcommand {trackList::editCellStart} \
         -editendcommand {trackList::editCellEnd} \
         -movablecolumns 1
   # $PLB cget -columns will return current cols in order (width tag alignment)
   # Define editiable columns
   ##
   # Set the options
   #
   settings::regColor $tbl \
      -foreground ::snackAmpSettings(fg) \
      -background ::snackAmpSettings(bg) \
      -selectforeground ::snackAmpSettings(fg) \
      -selectbackground ::snackAmpSettings(highlight)

   if {$::snackAmpSettings(useTableStripes)} {
      settings::regColor $tbl -stripebackground ::snackAmpSettings(tsbg)
      $tbl config -stripebackground $::snackAmpSettings(tsbg)
   }
   if {$::snackAmpSettings(useTableBars)} {
      $tbl config -showseparators yes
   }
   if {$::snackAmpSettings(autoTableStretch)} {
      $tbl config -stretch all
   }
   # Define how each col is sorted.
         # -yscrollcommand [list scrollCmd $vsb  [list grid $vsb -row 0 -column 1 -sticky ns]]
   if {[catch {
      foreach {col mode} $sortMode {
         switch -- $mode {
            "commify" {
               $tbl columnconfigure $col -sortmode integer -formatcommand commify
            }
            "timify" {
               $tbl columnconfigure $col -sortmode real -formatcommand HMSfrac
            }
            "mrating" {
               $tbl columnconfigure $col -sortmode ascii -formatcommand mrating
            }
            default {
               $tbl columnconfigure $col -sortmode $mode
            }
         }
      }
   } result]} {
      debug $result
   }
   # Hide column used for the key
   $tbl columnconfigure 0 -hide 1
   scrollbar $vsb -orient vertical   -command [list $tbl yview]
   scrollbar $hsb -orient horizontal -command [list $tbl xview]
   #
   # Manage the widgets
   #
   grid $tbl -row 0 -column 0 -sticky news
   place $vsb -relx 1.0 -rely 0.0 -anchor ne -relheight 1.0
   grid $hsb -row 1 -column 0 -sticky ew
   grid rowconfigure    $tf 0 -weight 1
   grid columnconfigure $tf 0 -weight 1
   # Below has hard reference to table .body
   mscroll $tbl.body
   return $tbl
}
#------------------------------------------------------------------------------
# trackList::saveColumnWidths
#
# Save the current column widths for later reuse
#
#
#------------------------------------------------------------------------------
proc trackList::saveColumnWidths {} {
   variable lastFolder
   variable TABLEDATA
   global snackAmpSettings
   Trace
   # Save displayed col widths and order
   set colOrder [list]
   foreach {width name just} [$::PLB cget -columns] {
      lappend colOrder $name
      set snackAmpSettings(col,$name,width) $width
   }
   set ::snackAmpSettings(colorder) $colOrder
   set ::snackAmpSettings(table,sortcol) [$::PLB sortcolumn]
   set ::snackAmpSettings(table,sortorder) [$::PLB sortorder]
}
#------------------------------------------------------------------------------
# trackList::populate
#
# Populate the table widget with track information
# if 'insert' is not true, the cache files are built
# but no attempt is made to populate the listbox
#------------------------------------------------------------------------------
proc trackList::populate {baseIndex tbl sort insert} {
#   variable Index2File
   variable lastFolder
   variable TABLEDATA
   set llen [llength $TABLEDATA]
   set col2Use [list]
   set isDirty 0
   ##
   # Only use the columns that are currently used
   #
   foreach {xx col xx} [$tbl cget -columns] {
      lappend col2Use $col
   }
   catch {tablelist::canceleditingSubCmd $::PLB}
   # Init some values
   set totalSize 0
   set totalTime 0
   set tableData [list]
   ##
   # Return a nested list of table data for this folder and treeDepth subfolders
   # Cache any data as necessary
   set filterHow $::snackAmpSettings(FilterWhat)
   if {$::TA(useSAFilter)} {
      set filter $::TA(SAFilter)
      set depth 100
      # Special case, match multiple fields
      if {$filterHow == "Any"} {
         set filterHow [list Title Artist Album File Comp]
      }
   } else {
      set filter ""
      set depth $::snackAmpSettings(treeDepth)
   }
   set isDirty [db::tableTreeData $baseIndex $col2Use $::snackAmpSettings(checkStat) $depth $filter $filterHow totalTime totalSize tableData _Index2File]
   ##
   # Insert the list configSet into the tablelist widget
   # Only if data changed or files added/removed  && ($isDirty || $llen != [llength $tableData] || ![string equal $baseIndex $lastFolder])
   if {$insert} {
      if {$baseIndex ne $lastFolder} {
         $::PLB selection clear 0 end
      }
      # TABLEDATA is linked such that changing it changes
      # the displayed data.
      set trackList::TABLEDATA $tableData
      #array unset Index2File
      #array set Index2File [array get _Index2File]
      set ::TA(ck,tableSize) "Size: [commify $totalSize] KB"
      set ::TA(ck,tableTracks) "Tracks: [commify [llength $tableData]]"
      set days [expr {int($totalTime)/86400}]
      if {$days == 0} {
         set ::TA(ck,tableTime) "Time: [HMSfrac $totalTime]"
      } else {
         set totalTime [expr {int($totalTime)%86400}]
         set ::TA(ck,tableTime) "Time:${days}d [HMSfrac $totalTime]"
      }
      if {$sort} {
         ##
         # Sort by last used or by col 1 by default (col 0 is hidden)
         #
         set lastSort [$::PLB sortcolumn]
         set lastSortOrder [$::PLB sortorder]
         if {$lastSort >= 0} {
            $tbl sortbycolumn $lastSort -$lastSortOrder
         } elseif {$::snackAmpSettings(table,sortcol) >= 0} {
            if {[catch {
               $tbl sortbycolumn $::snackAmpSettings(table,sortcol) \
                  -$::snackAmpSettings(table,sortorder)
            } result] } then {
               debug $result
            }
         }
      }
   }
   waitWindow::close
   set lastFolder $baseIndex
}

#------------------------------------------------------------------------------
# Function: trackList::getFiles
#
# Return a list of files within some folders
#------------------------------------------------------------------------------
proc trackList::getFiles {folders} {
   set result [list]
   foreach folder $folders {
      foreach candidate [db::filesInFolder $folder] {
         lappend result $candidate
      }
   }
   return [lsort -dictionary $result]
}
#------------------------------------------------------------------------------
# Function   : aplLB Namespace
# Description: Create bindings for the APL tree
# Author     : Tom Wilkason
# Date       : 3/9/2001
#------------------------------------------------------------------------------
namespace eval aplLB {
   package require saPane
}
proc aplLB::commonBindings {args} {
   # Toplevel bindings
   bind .wFC <F5>       {SelectFile 0}
   bind .wFC <Control-F5> {SelectFile 1}
   bind .wFC <Alt-F4> {APL::close}
   # Track List
   APL_ListAssign <F2> {::APL::RenameFile}
   APL_ListAssign <Control-F2> ::APL::RenameFile
   APL_ListAssign <F3> ::APL::EditId3
   APL_ListAssign <F4> {::APL::NameToTag 0}
#      APL_TreeAssign <F4> {::APL::treeNameToTag 0}
   APL_TreeAssign <F3> {::APL::treeEditId3}
   APL_ListAssign <F6> {::APL::TagToName}
   APL_TreeAssign <F6> {::APL::treeTagToName}
   # todo: replace with tkdnd
   APL_ListAssign <Delete> ::APL::DeleteFiles
   APL_ListAssign <B1-Motion> {
      # Prevent the select-drag action
      # It would mess up the DND
      $::PLB config -cursor hand2
      break
   }
   APL_ListAssign <ButtonPress-3> {
      tk_popup $::PLB.popup %X %Y
   }
   APL_ListAssign <ButtonPress-1> {
      # Allow key bindings to work
      focus %W
      # Process any special columns
      trackList::colClick %W %x %y %X %Y
   }
   APL_ListAssign <<Copy>> {APL::popWatch;clipboardCopy $::PLB 0;break}
   ##
   # DND leave, grap to the clipboard
   #
   # todo: replace with tkdnd
   APL_ListAssign <B1-Leave> {
      # Leaving the listbox, snapshot the files
      clipboardCopy $::PLB 0
      break   ;# Prevent the drag select
   }
   ##
   # Drop action coming from the APL list
   #
   APL_ListAssign <ButtonRelease-1> {
      # See if we are over a node on the tree
      set target [winfo containing %X %Y]
      if {[winfo exists $target]} {
         if {$target eq $::TREE} {
            APL::cbDndPLBToTree %X %Y
         } elseif {$target eq [LB::body]} {
            APL::cbDndPLBToLB %X %Y
         } elseif {[winfo toplevel $target] == [winfo toplevel $::LB]} {
            APL::cbDndPLBToLB 0 0
         }
      }
      $::PLB config -cursor $::TA(arrow)
      unset -nocomplain target
   }
}
proc aplLB::appendSelectedFiles {args} {
   #if {[array size ::trackList::Index2File] > 0} {
      LB::AppendList [aplLB::nameSelected]
   #}
}
proc aplLB::insertSelectedFiles {args} {
   if {[winfo parent $::TREE] == [focus]} {
      LB::userInsertList [aplLB::getAllNames 1]
   } else {
      #if {[array size ::trackList::Index2File] > 0} {
         LB::userInsertList [aplLB::nameSelected]
      #}
   }
}
##
# Adjust the user rating of the track(s)
#
proc aplLB::bumpRating {path depth amount} {
   if {[winfo parent $::TREE] == [focus]} {
      # Folder
      db::bumpFolderRating $path $depth $amount
   } else {
      # Selected Files
      set selected [aplLB::nameSelected]
      db::bumpTrackRating $selected $amount
   }
}
##
# Play the highlighted files or folder
#
proc aplLB::playSelectedFiles {args} {
   if {[winfo parent $::TREE] == [focus]} {
      # Folder
      set end [LB::userInsert]
      LB::userInsertList [aplLB::getAllNames 1]
      LB::activate $end
      Play active
   } else {
      # Selected Files
      set selected [aplLB::nameSelected]
      if {[llength $selected] == 1} {
         LB::InsertPlay [lindex $selected 0]
      } else {
         set end [LB::userInsert]
         if {[LB::userInsertList $selected] > 0} {
            LB::inverseVideoFile xx 0
            LB::activate $end
            Play active
         }
      }
   }
}
##
# Menus bound to listbox
#
proc aplLB::menus {args} {
   ##
   # Rating popup
   #
   set m $::PLB.plbRating
   menu $m -postcommand [list trackList::ratingPopup $m]
   ##
   # Genre popup
   #
   set m $::PLB.plbGenre
   menu $m -postcommand [list trackList::genrePopup $m]
   ##
   # File list box bindings
   #
   set plb [$::PLB bodypath]
   bindings::init $plb menuBindings
   set m $::PLB.popup
   menu $m
   ##
   # Copy the files to the playlist
   #
   bindings::addCommand $m "Play Track(s)" {aplLB::playSelectedFiles} 0
   bindings::addCommand $m "Queue Track(s) to Play Next" {aplLB::insertSelectedFiles} 0
   bindings::addCommand $m "Append Track(s) to Playlist" {aplLB::appendSelectedFiles}
   $m add separator
   bindings::addCommand $m "Increase Rating" {aplLB::bumpRating $::TA(highlightPath) 100 1}
   bindings::addCommand $m "Decrease Rating" {aplLB::bumpRating $::TA(highlightPath) 100 -1}
   $m add separator
   ##
   # Rename the active selection
   #
   menu $m.edit
   $m add cascade -label "Edit" -menu $m.edit -underline 0 -compound left -image saImages.edit-16
   bindings::addCommand $m.edit "Delete Tracks(s)"  {::APL::DeleteFiles} 0
   $m.edit add separator
   bindings::addCommand $m.edit "Copy Track(s) to Clipboard"  {APL::popWatch;clipboardCopy $::PLB 0} 0
   bindings::addCommand $m.edit "Append Track(s) to Clipboard" {APL::popWatch;clipboardCopy $::PLB 1} 0
   $m.edit add separator
   # clipboardPaste doesn't work with APL table
   bindings::addCommand $m.edit "Copy Clipboard Tracks Here" {APL::popWatch;APL::PasteFolderToTree copy} 0
   bindings::addCommand $m.edit "Move Clipboard Tracks Here" {APL::popWatch;APL::PasteFolderToTree move} 0
#      bindings::addCommand $m.edit "Autolevel Tracks" {} 0

   menu $m.tn
   $m add cascade -label "Track Names" -menu $m.tn -underline 0
   bindings::addCommand $m.tn "Rename Track"  {::APL::RenameFile} 0
   bindings::addCommand $m.tn "Combine Name with Folder (1)" {::APL::CombineFile 1}
   bindings::addCommand $m.tn "Combine Name with Folders (2)" {::APL::CombineFile 2}
   bindings::addCommand $m.tn "Split Name Into Folders" {::APL::SplitFile}
   bindings::addCommand $m.tn "Correct Track Name(s)" {fixplus::correctCase [aplLB::nameSelected]}
   $m.tn add separator
   bindings::addCommand $m.tn "Move Selected Tracks under Artist Tag" {::APL::insertTagName list {Artist}}
   bindings::addCommand $m.tn "Move Selected Tracks under Artist/Album Tags" {::APL::insertTagName list {Artist Album}}
   bindings::addCommand $m.tn "Move Selected Tracks under Album Tag" {::APL::insertTagName list {Album}}

   menu $m.plf
   $m add cascade -label "Playlist File" -menu $m.plf -underline 0 -compound left -image saImages.playlist
   bindings::addCommand $m.plf "Save Selected to File"  {macros::SavePlaylistFile $::PLB replace selected}
   bindings::addCommand $m.plf "Save All to File"  {macros::SavePlaylistFile $::PLB replace all}
   ##
   # Media Players
   #
   set pm [menu $m.medplay \
      -postcommand [list gui::mediaPlayerPopup $m.medplay {selectedTracks $::PLB}]]
   $m add cascade -label "Copy Selected to Media Player" -menu $pm \
      -underline 0 -compound left -image saImages.copy
   $m.plf add separator
   bindings::addCommand $m.plf "Append Selected to File"  {macros::SavePlaylistFile $::PLB append selected}
   bindings::addCommand $m.plf "Append All to File"  {macros::SavePlaylistFile $::PLB append all}

   menu $m.id3
   $m add cascade -label "ID3 Tags" -menu $m.id3 -underline 0 -compound left -image saImages.edit-16
   bindings::addCommand $m.id3 "ID3 Tag Editor"  {::APL::EditId3} 0
   bindings::addCommand $m.id3 "ID3 Table Editor"  {::APL::NameToTag 0}
   bindings::addCommand $m.id3 "ID3 Tag To Name Editor" {::APL::TagToName}
   $m.id3 add separator
   bindings::addCommand $m.id3 "Remove ID3V1 Tag for selected tracks" {macros::removeTag [aplLB::nameSelected] "V1"}
   bindings::addCommand $m.id3 "Remove ID3V2 Tag for selected tracks" {macros::removeTag [aplLB::nameSelected] "V2"}
   bindings::addCommand $m.id3 "Remove ID3V1/V2 Tags for selected tracks" {macros::removeTag [aplLB::nameSelected] {"V1" "V2"}}

   menu $m.gain
   $m add cascade -label "Auto-Level" -menu $m.gain -underline 0 -compound left -image saImages.autoLevel
   bindings::addCommand $m.gain "Compute Track(s) Gain" {gainControl::computeSome [aplLB::nameSelected]}
   bindings::addCommand $m.gain "Average Track(s) Gain" {gainControl::averageSome [aplLB::nameSelected]}
   bindings::addCommand $m.gain "Reset Track(s) Gain" {gainControl::resetSome [aplLB::nameSelected]}
   $m add separator
   bindings::addCommand $m "Track Info..." {catalog::properties [aplLB::nameSelected]}

#      $m add separator
#      bindings::addCommand $m "Create Stream Shortcut Here" {
#         if {[playurl::checkuskfolder]} {
#            playurl::addPromptUrl $TA(highlightPath)
#            aplLB::UpdateList $::TA(highlightPath) 1
#         }
#      }
}

proc aplLB::selected {args} {
   return [$::PLB curselection]
}
proc aplLB::setSelection {index} {
   global PLB
   $PLB selection clear 0 end
   $PLB selection set $index
   $PLB activate $index
   $PLB see $index
}
##
# Table Stuff
#
#----------------------------------------------------------------------------
# Called from button, catalog and rescan the folder and appropriate subfolders.
#----------------------------------------------------------------------------
proc aplLB::RefreshFolderTree {path} {
   APL::indexFolderChildren $path
   update
   UpdateList $path 1
}
##
# Update the listbox based on incoming folder
# Generally called when a new folder is selected
proc aplLB::UpdateList {index {insert 1}} {
   Trace
   ##
   # New Table
   #
   if {$insert} {
      if {[catch {
         trackList::populate $index $::PLB 1 1
      } result] } then {
         saLog $::errorInfo
      }
   }
   ##
   # Update the selected counts
   #
   after idle aplLB::statusBar
   # Signal done
}

#------------------------------------------------------------------------------
# Update the status bar with current selection info
#------------------------------------------------------------------------------
proc aplLB::statusBar {args} {
   global TA
   after cancel aplLB::statusBar
   set index $TA(highlightPath) ;# always use the selected path
   set checked 0
   set unchecked 0
   set pattern "[string trimright $index /]/*"
   # Show folders checked instead of files (faster)
   foreach {dirName} [aplFolder::foldersMatchingPattern $pattern] {
      if {[aplFolder::isSet $dirName 1]} {
         incr checked
      } else {
         incr unchecked
      }
   }
   if {($checked+$unchecked) >= 0} {
      set TA(ck,selected)   "Checked: $checked"
      set TA(ck,unselected) "Unchecked: $unchecked"
      set TA(ck,total)      "Folders: [expr {$checked+$unchecked}]"
      if {[info command ntFreeSpace] ne ""} {
         if {[catch {
            set fr [ntFreeSpace [file nativename [fileRoot $index]]]
            if {[llength $fr]==2} {
               foreach {tot free} $fr {break}
               set ::TA(ck,free)   "[fileSize $free] Free"
            }
         } result] } then {
            set ::TA(ck,free)   "UnKnown Free"
            saLog $result
         }
      }
      # Total tracks in tree
      set TA(ck,totalTree) "Tracks: [commify [db::countMatchingPattern $pattern]]"
   }
}
proc aplLB::rowIndex2file {index} {
   return [trackList::rowIndex2file $index]
}
proc aplLB::getAllNames {{long 0}} {
   set files [list]
   # Get file name for each row
   set end [$::PLB index end]
   for {set i 0} {$i < $end } {incr i} {
      if {$long} {
         lappend files [trackList::rowIndex2file $i]
      }  else {
         lappend files [file tail [trackList::rowIndex2file $i]]
      }
   }
   return $files
}
proc aplLB::getActiveFullName {args} {
   set active  [$::PLB index active]
   return [trackList::rowIndex2file $active]
}
proc aplLB::nameSelected {args} {
   set fullNames [list]
   foreach {index} [selected] {
      lappend fullNames [trackList::rowIndex2file $index]
   }
   return $fullNames
}
proc aplLB::RemoveFile {File} {
   if {[winfo exists $::PLB]} {
      set lnFile [file tail $File]
      set index [lsearch -exact [getAllNames] $lnFile]
      if {$index > -1} {
         $::PLB delete $index
      }
   }
}
##
# Key bindings for table
#
proc aplLB::bindings {args} {
   # Common Keys
   commonBindings
   # Couple hot keys

   APL_ListAssign <Double-ButtonPress-1> {
      #if {[array size ::trackList::Index2File] > 0} {
         if {[catch {trackList::row2file %W %y} File] } then {
            saLog $File
         } else {
            LB::InsertPlay $File
         }
      #}
   }
}


package provide trackList 1.0
