#------------------------------------------------------------------------------
# snackAmp - urlBrowser package 1.0 for snackAmp
# provides dynamic http applications for SnackAmp
# web server.
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
##
# url to show a directory listing to allow browsing
# of files by folder. Used by both the jukebox and remote control functions.
# todo:
# button to play folder
# button in insert file/folder into playlist
# todo: Add form controls to define how remote/tracks/sorting is to be used.
namespace eval dirBrowser {
   variable locateForm {
          <form method="GET" action="" TITLE="Press to browse by folder name">
            <input type="submit" value="Search Folders" class="goButton" ALIGN="right">
            <input type="text" name="db_locate_folder" size="12" class="formText"  TITLE="Enter text then Search to locate the first folder that contains the text. Wildcards * and ? are supported.">
          </form>
   }
   variable lastFolder
}
#------------------------------------------------------------------------------
# Function   : dirBrowser::browsePlay
# Description: Called from jukebox, remote and display urls to show a folder of files
# Author     : Tom Wilkason
# Date       : 6/19/2003
#------------------------------------------------------------------------------
proc dirBrowser::doit {sock {params ""}} {
   Trace
   variable locateForm
   variable lastFolder
   set myns [uplevel {string trim [namespace current] :}]
   switch -glob -- $myns {
      jukebox* {set control 2}
      remote* {set control 1}
      display* {set control 0}
      default {set control 0}
   }
   upvar #0 Httpd$sock data
   set title "SnackAmp Jukebox"
   set result [HTH::leader $title $::snackAmpSettings(cssName)]
   append result "<body>"
   append result [js::scripts popupInfo toggleAll_jb_add toggleAll_jb_search_pl textWindow]
   set useFolder ""
   set Tracks [list]
   set sortMethod "shuffle"
   if {[string length $params]} {
      foreach param [split $params &]  {
         foreach {CmdType Cmd} [split $param =] {break}
         set dCmd [HttpD::Decode $Cmd]
         switch -- $CmdType {
            db_add_start {
               # When processing a playlist add, this is the first param
               set playing [LB::idxActive]
               set old_end [LB::idxEnd]
               # The back reference uses the db_add_start param (which is the params from the
               # request that originated the search) which was passed in here. It is saved
               # In a hidden field from the prior form.
               append result [AREF "?[NOAMP $dCmd]" Back]
               append result [H2 "Added the following tracks to the playlist"]
               append result "<table class='songPlaylist'>\n"
            }
            db_add {
               # When processing a playlist add, these are the middle params, one per item to add
               set track [NOAMP [ROOTTAIL $dCmd]]
               set folder [NOAMP [HTH::displayFolder $dCmd]]
               append result [HTH::tableRow [list $folder $track] headerRow]
               lappend Tracks $dCmd
            }
            db_append_process {
               # When processing a playlist add, this is the final param, signals end of list
               append result "</table>\n"
               if {[llength $Tracks] > 0} {
                  foreach {track} $Tracks {
                     set next [LB::idxFile $track]
                     if {$control==1 || $next==-1 || $next <= [LB::idxPlaying] || ($next-[LB::idxActive])>$::snackAmpSettings(minNextTrack)} {
                        LB::userInsert $track
                     }
                  }
                  append result [H2 "Your requested tracks were queued to play"]
                  gui::updateFileDisplay
               } else {
                  append result [H2 "No tracks selected or tracks already in playlist"]
               }
            }
            db_insert_process {
               ;# When processing a playlist add, this is the final param, signals end of list
               append result "</table>\n"
               append result [H2 "Your requested tracks are now queued to play next"]
               set cp [LB::idxPlaying]
               LB::insertList $cp $Tracks
            }
            db_locate_folder {
               # Search the folder list from the last folder position
               # searched for this socket. Allows rotating through
               # all matching folders.
               set pos 0
               set ip $data(ip)
               set folderList [aplFolder::foldersMatchingPattern *$dCmd* 1]
               if {[info exists lastFolder($ip)]} {
                  set pos [lsearch -exact $folderList $lastFolder($ip)]
                  if {$pos > -1} {
                     incr pos
                     if {$pos >= [llength $folderList]} {
                        set pos 0
                     }
                  } else {
                     set pos 0
                  }
               }
               set firstFolder [lindex $folderList $pos]
               if {$firstFolder eq ""} {
                  append result [H2 "No Matches found for ' *$dCmd*'"]
               } else {
                  set useFolder [ENCODE $firstFolder]
                  set lastFolder($ip) $firstFolder
               }
            }
            bt_track_name {
               # step 1
               set bt_track_name $dCmd
            }
            bt_track_rating {
               # step 2
               # Set the track rating for some track
               HTH::clearCache $bt_track_name
               db::setVal $bt_track_name Rating $Cmd
               trackList::updateFiles [list $bt_track_name]

               set useFolder [ENCODE [file dirname $bt_track_name]]
               # Note, rating menu doesn't add the folder option to the url so the screen
               # is blank
            }
            bt_track_voting {
               # step 2
               # Set the track voting for some track
               # needs function to increment vote
               HTH::clearCache $bt_track_name
               #db::setVal $bt_track_name Rating $Cmd
               set useFolder [ENCODE [file dirname $bt_track_name]]
               # Note, rating menu doesn't add the folder option to the url so the screen
               # is blank
            }
            play {
               # Play a specific song
               # /control?play=some+name.mp3
               set track $dCmd
               if {[catch {LB::InsertPlay $track} error] } then {
                  debug $result
                  append result $result
               } else {
                  append result [H2 "Playing '$track'"]
               }
            }
            append {
               set track $dCmd
               if {[HTH::isMedia $track]} {
                  set next [LB::idxFile $track]
                  if {$control==1 || $next==-1 || $next <= [LB::idxPlaying] || abs($next-[LB::idxActive])>$::snackAmpSettings(minNextTrack)} {
                     # Insert after last web insert (instead of end) if possible
                     LB::userInsert $track
   #                  append result [H2 "Added $track to Playlist"]
                     gui::updateFileDisplay
                  } else {
                     append result [H2 "Skipping $track, it is already in playlist"]
                  }
               } else {
                  append result [H2 "Skipping $track, it is invalid"]
               }
            }
            delete {
               LB::deleteIndex $dCmd
               gui::updateFileDisplay
               # Remove from playlist
            }
            folder {
               # No refresh when browsing
               set useFolder $Cmd
            }
         }
      }
      # If a folder to browse to was selected
      if {$useFolder ne ""} {
         append result [browser::playlistHeader $control 10]
         append result [H2 "Select or search for a folder and pick your favorite track"]
         append result $locateForm
         append result [browser::folderList $useFolder $control $params]
      }
   } else {
      append result [browser::playlistHeader $control 15]
      append result [H2 "Select or search for a folder and pick Your favorite track"]
      append result $locateForm
      append result [browser::folderList [info hostname] $control $params]
   }
   append result "</body>\n</html>\n"
   return [list dhtml html [string trim $result]]
}

#------------------------------------------------------------------------------
#  Return the playlist and head portion if selected
#------------------------------------------------------------------------------
proc browser::playlistHeader {control toshow} {
   # Only inhibit if control is remote
   if {$::snackAmpSettings(browseNowPlaying)||$control!=1} {
      #append result [HTH::getRefreshScript [HTH::nextRefresh] "?folder=[info hostname]"]
      append result [H2 "SnackAmp currently [string tolower [soundControl::playState]]"]
      append result [HTH::trackInfo [soundControl::fileName] 2]
      append result {<table><tr><td valign="top">}
      append result [HTH::nextToPlay $control $toshow]
      append result {</td><td valign="top">}
      append result [HTH::queuedToPlay $control $toshow]
      append result {</td></tr></table>}
      return $result
   } else {
      return ""
   }
}
##
# Return the folderList of the browser window
#
proc browser::folderList {Cmd control params} {
   Trace
   # If selected is the toplevel, assume parent is the hostname
   set folder [HttpD::Decode $Cmd]
   array set sides {1 left 2 left}
   set sideNum 1
   set files [list]
   set me  [info hostname]
   append result "<div class='folderDiv'>"
   append result [subst {$HTH::folderopen [AREF "?folder=$me" $me]}]
   if {$me eq $folder} {
      set atroot 1
      set parent $me
      set folders $::TA(mediaDirs)
      append result "<br>\n"
   } else {
      set atroot 0
      set folders [list]
      set validFolder 0
      ##
      # For security, verify this folder is actual one off
      # one of the media folders.
      #
      foreach media $::TA(mediaDirs) {
         if {[string match $media* $folder ]} {
            set validFolder 1
            break
         }
      }
      # default to the root
      if {$validFolder==0} {
         set parent $me
         set folders $::TA(mediaDirs)
         append result "<br>\n"
      }  else {
         ##
         # Display this and child folders
         #
         # Create back track in proper order
         set tlevel $folder
         set dirPath {}
         while {1} {
            lpush left dirPath $tlevel
            # set dirPath [linsert $dirPath 0 $tlevel]
            # Can't back up beyond a media folder
            if {[lsearch -exact $::TA(mediaDirs) $tlevel] >= 0} {
               break
            }
            set under [file dirname $tlevel]
            if {$tlevel eq $under} {
               # Can't backup any more
               break
            } else {
               set tlevel $under
            }
         }
         ##
         # Create a back path to the root folder from selected folder
         # Last folder had no hyperlink
         set i 0
         foreach {tlevel} $dirPath {
            incr i
            if {$i < [llength $dirPath]} {
               append result [subst {<b></b>[SP][AREF "?folder=[ENCODE $tlevel]" [FTAILIFCAN $tlevel] [file tail $tlevel]] }]
            }  else {
               append result "<b></b>[SP][FTAILIFCAN $tlevel]<br>\n"
            }
         }
         ##
         # Build up a list of all the folders and files in the two tables, folders then tracks
         # TFW: cache the files and folders, fire after time to unset them later
         set files [db::filesInFolder $folder 0]
         # First folder returned is this one, skip it
         set folders [lrange [aplFolder::getFolders $folder 1] 1 end]
         # Show Prev/Next indicators for sibling folders
         set parent [lindex $dirPath end-1]
         #set siblings [lsort -dictionary [glob -nocomplain -types d -directory $parent *]]
         set siblings [lrange [aplFolder::getFolders $parent 1] 1 end]
         set pos [lsearch -exact $siblings $folder]
         if {$pos >= 0} {
            set prev [lindex $siblings [expr {$pos-1}]]
            set next [lindex $siblings [expr {$pos+1}]]
            if {[string length $prev]} {
               append result [subst {<b>Prev</b>[SP][AREF "?folder=[ENCODE $prev]" [FTAILIFCAN $prev] [file tail $prev]] }]
            }
            if {[string length $next]} {
               append result [subst {<b>Next</b>[SP][AREF "?folder=[ENCODE $next]" [FTAILIFCAN $next] [file tail $next]] }]
            }
         }
      }
   }
   ##
   # Show the sub-folders in a table along with coverart if possible
   #
   if {[llength $folders]} {
      incr sideNum
      set tail [file tail $folder]
      # Find minimum folder length for indent offsets ($control==1 && )
      if {($::snackAmpSettings(remote,brcoverart)) || ($control==2 && $::snackAmpSettings(jukebox,brcoverart))} {
         set sca 1
      } else {
         set sca 0
      }
      ##
      # Todo: Put these in a grid within a table
      #
      set totFolders [llength $folders]
      set folderCount 0
      append result "<table>"
      append result "<tr>"
      foreach item $folders {
         if {$sca} {
            set cov [HTH::coverArt $item $::snackAmpSettings(coverbrowsersize)]
            set ca [HTH::coverArtShadow [AREF "?folder=[ENCODE $item]" $cov [file tail $item]]]
         } else {
            set ca ""
         }
         if {$atroot} {
            set link [subst {$HTH::folderclosed [AREF "?folder=[ENCODE $item]" $item [file tail $item]]<br>\n}]
         } else {
            set link [subst {$HTH::folderclosed [AREF "?folder=[ENCODE $item]" [FTAILIFCAN $item] [file tail $item]]<br>\n}]
         }

         if {($folderCount % $::snackAmpSettings(artrowcount))==0} {
            append result {</tr><tr>}
         }
         incr folderCount
         append result [subst {<td valign='top'>$ca$link</td>}]
      }
      append result "</tr>"
      append result "</table>"

   }
   append result "</div>"
   ##
   # Show the files in a table
   # allow case for with/without remote control
   if {[llength $files]} {
      if {($control==1 && $::snackAmpSettings(remote,rettofolder)) || ($control==2 && $::snackAmpSettings(jukebox,rettofolder))} {
         set rtf 1
      } else {
         set rtf 0
      }
      append result "<div class='fileDiv'>"
      if {$control==1} {
         append result [subst {<FORM METHOD="GET" ACTION="" NAME="jb_add_form">}]
         append result "<input type=hidden name=db_add_start value=\"$params\">"
      }
      # Album/Folder Title
      append result "<font size='5'>[file tail $folder]</font><br>"
      # Table two across, listing on left album cover on right
      append result {<table><tr><td valign="top">}
      append result {<table class='songPlaylist'>}
      foreach file $files {
         set bgc [HTH::rowStyle i]
         # This is for dag to not show the same folder over
         # if not rtf then return to root
         if {$rtf} {
            set command $Cmd
         } else {
            # return to the top of the browser
            set command [info hostname]
         }
         ##
         # Is remote control enabled or not
         #
         set inImage [AREF "?append=[ENCODE $file]&folder=$command" $HTH::insertbutton "Queue Track to Play"]

         set buttonLink [HTH::popupInfo $file]
         if {$control==1} {
            set cb "<INPUT TYPE=checkbox NAME='db_add' VALUE=\"$file\">"
            set plImage [HTH::popUpProcess play "" $file]
            append result [HTH::tableRow [list $cb $buttonLink $inImage $plImage [DBINFO $file]] $bgc]
         } else {
            append result [HTH::tableRow [list $buttonLink $inImage [DBINFO $file]] $bgc]
         }
      }
      append result "</table>\n"
      append result {</td><td valign="top">}
      append result [HTH::coverArtShadow [HTH::coverArt $folder $::snackAmpSettings(coversize)]]
      append result "</td></tr></table>"
      # end table

      if {$control==1} {
         append result "<INPUT TYPE=submit VALUE='Queue Selected' NAME=db_append_process>\n"
         append result "<INPUT TYPE=submit VALUE='Insert Selected' NAME=db_insert_process>\n"
         append result "</FORM>\n"
         append result [subst {[AREF "javascript:toggleAll_jb_add(true)" "Select All"] | [AREF "javascript:toggleAll_jb_add(false)" "Unselect All"]<br>}]
      }
      append result "\n</div>"
   } else {
   #                  append result "<hr>\n"
   #                  append result [H2 "No tracks within [file tail $folder]"]
   }
   return $result
}
package provide urlBrowser 1.0


