#------------------------------------------------------------------------------
# snackAmp - urlJukebox package 1.0 for snackAmp
# provides dynamic http applications for SnackAmp
# web server.
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
##
# Wrapper for jukebox function, makes the browser namespace react based
# on calling function
#
#------------------------------------------------------------------------------
# Function   : display
# Description: Main function that is called by web server when /display
#              is entered.
#              e.g. /display
# Author     : Tom Wilkason
# Date       : 12/15/2002
#------------------------------------------------------------------------------
proc jukebox/header::process {sock {params ""}} {
   Trace
   append Result [HTH::leader "SnackAmp Playlist" $::snackAmpSettings(cssName)]
   append Result "<body>\n"
   append Result [HTH::getRefreshScript [HTH::nextRefresh] "/jukebox/header"]
   append result [H2 "Now [string tolower [soundControl::playState]]:"]
   append Result [HTH::trackInfo [soundControl::fileName] 2]
   append Result "</body>\n</html>"
   return [list dhtml html $Result]
}
proc jukebox/next::process {sock {params ""}} {
   Trace
   append Result [HTH::leader "SnackAmp Playlist" $::snackAmpSettings(cssName)]
   append Result "<body>\n"
   append Result [HTH::getRefreshScript [HTH::nextRefresh] "/jukebox/next"]
   # Need to launch hypertext to browse into a different target, how?
   append Result [HTH::nextToPlay 2 10]
   append Result "</body>\n</html>"
   return [list dhtml html $Result]
}
proc jukebox/display::process {sock {params ""}} {
   Trace
   append Result [HTH::leader "SnackAmp Playlist" $::snackAmpSettings(cssName)]
   append Result "<body>\n"
   append Result [HTH::getRefreshScript [HTH::nextRefresh] "/jukebox/display"]
   # Wait to let player take hold
   #append Result ${search::close_window}
   append Result [HTH::playStatus 0]
   append Result [HTH::trackInfo [soundControl::fileName] 2]
   append Result [HTH::nextToPlay 2 $::snackAmpSettings(httpNextLength)]
   append Result "</body>\n</html>"
   return [list dhtml html $Result]
}
proc jukebox/last::process {sock {params ""}} {
   Trace
   append Result [HTH::leader "SnackAmp Playlist" $::snackAmpSettings(cssName)]
   append Result "<body>\n"
   append Result [HTH::getRefreshScript [HTH::nextRefresh] "/jukebox/last"]
   append Result [HTH::queuedToPlay 2 20]
   append Result "</body>\n</html>"
   return [list dhtml html $Result]
}
proc jukebox/search::process {sock {params ""}} {
   Trace
   search::doit $sock $params
}
##
# Init the directory browser, from jukebox or remote
#
proc jukebox/browser::process {sock {params ""}} {
   Trace
   package require urlBrowser
   dirBrowser::doit $sock $params
}
##
# Jukebox folder browser (for running in an object)
#
proc jukebox/folder::process {sock {params ""}} {
   Trace
   set control 2
   append result [HTH::leader "SnackAmp Playlist" $::snackAmpSettings(cssName)]
   append result "<body>\n"
   if {[string length $params]} {
      foreach param [split $params &]  {
         foreach {CmdType Cmd} [split $param =] {break}
         set dCmd [HttpD::Decode $Cmd]
         switch -- $CmdType {
            folder {
               append result [H2 "Select a folder and pick your favorite track"]
               append result [browser::folderList $Cmd $control $params]
            }
            play {
               # Play a specific song
               # /control?play=some+name.mp3
               set track $dCmd
               if {[catch {LB::InsertPlay $track} error] } then {
                  debug $result
                  append result $result
               } else {
                  append result [H2 "Playing '$track'"]
               }
            }
            append {
               set track $dCmd
               if {[HTH::isMedia $track]} {
                  set next [LB::idxFile $track]
                  if {$control==1 || $next==-1 || $next <= [LB::idxPlaying] || ($next-[LB::idxActive])>$::snackAmpSettings(minNextTrack)} {
                     # Insert after last web insert (instead of end) if possible
                     LB::userInsert $track
                     append result [H2 "Queue $track to Playlist"]
                     gui::updateFileDisplay
                  } else {
                     append result [H2 "Skipping $track, it is already in playlist"]
                  }
               } else {
                  append result [H2 "Skipping $track, it is invalid"]
               }
            }
         }
      }
   } else {
      append result [H2 "Select a folder and pick your favorite track"]
      append result [browser::folderList [info hostname] $control $params]
   }
   append result "</body>\n</html>\n"
   return [list dhtml html [string trim $result]]
}
package provide urlJukebox 1.0
