#------------------------------------------------------------------------------
# snackAmp - urlRedirect package 1.0 for snackAmp
# provides dynamic http applications for SnackAmp
# web server.
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
#
# To create a new dynamic url create a namespace with the name of the url
# e.g. namespace eval playlist -> /playlist
# to enable it, create an init routine that returns a 1 (for security)
# you can have the url return either an existing file (e.g. an mp3 file), in
# that case create a namespace command ::process, that returns the name
# of the file.
#
# If you want to create dynamic html with the namespace, create a namespace
# command ::process, that returns content. You will need an extra parameter
# passed to it that you set to the file extension required.
#
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/urlRedirect.tcl,v 1.39 2007/03/10 16:58:52 wilkason Exp $
}
##
# Misc routines not in a namespace
#
proc AREF {ref txt {title ""}} {
   if {[string length $title]} {
      return "<a href='$ref' title='$title'>$txt</a>"
   } else {
      return "<a href='$ref' title='$txt'>$txt</a>"
   }
}
proc H2 {ref} {
   return "<h2>$ref</h2>\n"
}
proc H1 {ref} {
   return "<h1>$ref</h1>\n"
}
proc SP {args} {
   return {&nbsp;}
}
proc NOAMP {data} {
   return [string map {& &amp\;} $data]
#   return [string map {& %26} $data]
}
proc ENCODE {data} {
   return [HttpD::Encode $data]
}
proc ESCAPE {data} {
   return [string map {' \\' & %26} $data]
}
proc ROOTTAIL {file} {
   return [file rootname [file tail $file]]
}
##
# Show the tail of a file without the folder
# or file extension. This is implied to be
# embedded in a table elsewhere
#
proc DBINFO {file {method 1}} {
   if {$method} {
      set result [db::getValues $file {Title Duration Track Rating Folder}]
      # If file not found
      if {[llength $result]!=5} {
         set Title [nameToShort $file]
         return "</td><td>$Title</td><td>"
      } else {
         foreach {Title duration Track rating Folder} $result {break}
         set Duration [HMSfrac $duration]
         set rating [HTH::browserRatingMenu $rating $file $Folder]
         return "$rating</td><td>$Track[SP]</td><td>$Title </td><td>$Duration"
      }
   }  else {
      return [TAILIFCAN $file]
   }
}
proc TAILIFCAN {folder} {
   set tail [file tail $folder]
   if {$::snackAmpSettings(cleanListings)} {
      if {[string length $tail]} {
         return [fixplus::cleanName [file rootname $tail]]
      } else {
         return [fixplus::cleanName $folder]
      }
   }  else {
      if {[string length $tail]} {
         return [file rootname $tail]
      } else {
         return $folder
      }
   }
}
##
# Like TAILIFCAN but intended for folders
# where the file extension is not stripped
# off
#
proc FTAILIFCAN {folder} {
   set tail [file tail $folder]
   if {$::snackAmpSettings(cleanListings)} {
      if {[string length $tail]} {
         return [fixplus::cleanName $tail]
      } else {
         return [fixplus::cleanName $folder]
      }
   } else {
      if {[string length $tail]} {
         return $tail
      } else {
         return $folder
      }
   }
}
##
# Javascript routines, embed js::start, the variable then js::end
#
namespace eval js {
   proc start {args} {
     return {
     <script language="JavaScript">
     //<!--
     }
   }
   proc end {args} {
      return  {
     //-->
     </script>
     }
   }
   proc scripts {args} {
      set result [start]
      foreach {script} $args {
         variable $script
         append result [set [set script]]
      }
      append result [end]
      return $result
   }
   variable popupInfo {
        function popupInfo(url,height,width,tag) {
           var prefs = "scrollbars=yes,menubars=no,toolbars=no,"
           + "resizable=yes,left=0,top=0,height="+height+",width="+width;
           window.open(url, tag, prefs);
        }
   }
   variable textWindow {
        function textWindow(url) {
           var prefs = "scrollbars=no,menubars=no,toolbars=no,"
           + "resizable=yes,left=100,top=100,height=20,width=500;"
           tWin=window.open(url, "Text", prefs);
           setTimeout("tWin.close()",1000)
        }
   }
   variable toggleAll_jb_add {
        function toggleAll_jb_add(checked) {
           len = document.jb_add_form.elements.length;
           var i;
           for (i = 0; i < len; i++) {
              document.jb_add_form.elements[i].checked = checked;
           }
        }
   }
   variable toggleAll_jb_search_pl {
        function toggleAll_jb_search_pl(checked) {
           len = document.jb_search_pl_form.elements.length;
           var i;
           for (i = 0; i < len; i++) {
              document.jb_search_pl_form.elements[i].checked = checked;
           }
        }
   }
   variable countDown {
     <script language="JavaScript">
     //<!--
        //configure refresh interval (in seconds)
         var countDownInterval=$period;
         var countDownTime=countDownInterval+1;
         function countDown(){
            countDownTime--;
            if (countDownTime <=0) {
               countDownTime=countDownInterval;
               clearTimeout(counter);
               window.location.href = '$url'
            } else {
               // Check again in one second
               counter=setTimeout("countDown()", 1000);
            }
         }
         // Start if off
         if (document.all||document.getElementById)
            countDown();
         else
            window.onload=countDown;
     //-->
     </script>
   }
   variable foldersel {
      function foldersel(url,folder) {
        var useurl='none';
        var selsort='no sort';
        var selhow='no stream';
        var selsub='no';
        var i=0;
        //stream
        //selhow = getselected(document.option_form.cmb_How,2);
        for (i==0;i<2;i++) {
           if (document.option_form.cmb_How[i].selected) {
              selhow=document.option_form.cmb_How[i].text;
              break;
           }
        }

        //subfolders
        for (i==0;i<2;i++) {
           if (document.option_form.cmb_Sub[i].selected) {
              selsub=document.option_form.cmb_Sub[i].text;
              break;
           }
        }
        //sort
        for (i==0;i<4;i++) {
           if (document.option_form.cmb_Sort[i].selected) {
              selsort=document.option_form.cmb_Sort[i].text;
              break;
           }
        }
        useurl = url+
          '?sort='  + selsort +
          '&tree='   + selsub  +
          '&how='    + selhow  +
          '&folder=' + folder;
        window.open(useurl,"tinyWindow", 'width=100,height=100');
      }
   }
   variable getselected {
      // getselected: Return the selected item from a combo box
      function getselected(element,len) {
         var i;
         for (i==0; i<len; i++) {
            alert(element[i].selected + ' ' + element[i].text);
            if (element[i].selected) {
               return element[i].text;
            }
         }
         return 'none';
      }
   }
}
#------------------------------------------------------------------------------
# Http helper routines
#------------------------------------------------------------------------------
namespace eval HTH {
# TFW: Can these references be computed from root as relative?
   variable appendbutton   {<img border=0 src="/images/append.png">}
   variable folderclosed   {<img border=0 src="/images/folderclosed.gif">}
   variable folderopen     {<img border=0 src="/images/folderopen.gif">}
   variable infobutton     {<img border=0 src="/images/info.png">}
   variable insertbutton   {<img border=0 src="/images/queue.gif">}
   variable playbutton     {<img border=0 src="/images/play.gif">}
   variable trackimage     {<img border=0 src="/images/track.gif">}
   variable trackcropimage {<img border=0 src="/images/trackcrop.gif">}
   variable deleteimage    {<img border=0 src="/images/delete.png">}
   # Contains details on tracks in playlist
   variable id3Cache
   # Contains details on now-playing track
   variable InfoCache
   variable initDetails
   array set initDetails {
       Active        "&nbsp;"
       "File Name"   "&nbsp;"
       Folder        "&nbsp;"
       Artist        "&nbsp;"
       Title         "&nbsp;"
       Album         "&nbsp;"
       Track         "&nbsp;"
       Media         "&nbsp;"
       Year          "&nbsp;"
       Genre         "&nbsp;"
       Note          "&nbsp;"
       Comp          "&nbsp;"
       Encoding      "&nbsp;"
       Desc          "&nbsp;"
       Size          "&nbsp;"
       Date          "&nbsp;"
       Rating        "--"
       Duration      "&nbsp;"
       Position      "&nbsp;"
       "Bit Rate"    "&nbsp;"
       Gain          "&nbsp;"
   }
   array set row_styles {
      0 {lightRows}
      1 {darkRows}
   }
   variable ratingStar
   array set ratingStar {
   --   {<div class = "img-shadow"><img border=0 src="/images/star0.png" alt="No Rating"></div>}
   *   {<div class = "img-shadow"><img border=0 src="/images/star1.png" alt="*"></div>}
   **   {<div class = "img-shadow"><img border=0 src="/images/star2.png" alt="**"></div>}
   ***   {<div class = "img-shadow"><img border=0 src="/images/star3.png" alt="***"></div>}
   ****   {<div class = "img-shadow"><img border=0 src="/images/star4.png" alt="****"></div>}
   *****   {<div class = "img-shadow"><img border=0 src="/images/star5.png" alt="*****"></div>}
   }
   #       <td class="songTitle" colspan="5">[browser::allMusicRef $Details(Title) Title]</td>
   # TODO: For Rating, add a radio button to change it
   variable trackDetailsTemplate {
      <table class="detailsTable">
        <tr>
          <td rowspan="8" valign="top">$Details(Cover)</td>
        </tr>
        <tr>
          <td class="songTitle" colspan="5">$Details(Title)</td>
        </tr>
        <tr>
          <td class="songArtist" colspan="5">$Details(Artist)</td>
        </tr>
        <tr>
          <td class="songAlbum"  colspan="5">$Details(Album) $Details(Year)</td>
        </tr>
        <tr>
          <td class="songTemplateLabel">Rating</td>
          <td class="songTemplateInfo">$HTH::ratingStar($Details(Rating))</td>
        </tr>
        <tr>
          <td class="songTemplateLabel">Track</td>
          <td class="songTemplateInfo">$Details(Track)</td>
        </tr>
        <tr>
          <td class="songTemplateLabel">Year</td>
          <td class="songTemplateInfo">$Details(Year)</td>
          <td class="songTemplateLabel">Genre</td>
          <td class="songTemplateInfo">$Details(Genre)</td>
        </tr>
        <tr>
          <td class="songTemplateLabel">Rate</td>
          <td class="songTemplateInfo">$Details(Rate)</td>
          <td class="songTemplateLabel">Duration</td>
          <td class="songTemplateInfo">$Details(duration)</td>
        </tr>
      </table>
   }
   variable trackSummaryTemplate {
      <table class="infoTable">
        <tr>
          <td rowspan="5" valign="top">$Details(Cover)</td>
        </tr>
        <tr>
          <td class="songTitle">$Details(Title)</td>
        </tr>
        <tr>
          <td class="songArtist">$Details(Artist)</td>
        </tr>
        <tr>
          <td class="songAlbum">$Details(Album) $Details(Year)</td>
        </tr>
        <tr>
          <td class="songAlbum">[HTH::playlistRatingMenu "Current Rating" [soundControl::fileName]] $HTH::ratingStar($Details(Rating))</td>
        </tr>
      </table>
   }

   variable trackSummaryTemplateNR {
      <table class="infoTable">
        <tr>
          <td rowspan="5" valign="top">$Details(Cover)</td>
        </tr>
        <tr>
          <td class="songTitle">$Details(Title)</td>
        </tr>
        <tr>
          <td class="songArtist">$Details(Artist)</td>
        </tr>
        <tr>
          <td class="songAlbum">$Details(Album) $Details(Year)</td>
        </tr>
        <tr>
          <td class="songAlbum">Current Rating $HTH::ratingStar($Details(Rating))</td>
        </tr>
      </table>
   }

   variable playlistPopupMenu_title
   variable playlistPopupMenu_file
   variable playlistPopupMenu_counter
   variable playlistRatingMenu {
      <a href="#" onClick="return clickreturnvalue()" onMouseover="dropdownmenu(this, event, 'alm$HTH::playlistPopupMenu_counter')">$HTH::playlistPopupMenu_title</a>
      <div id="alm$HTH::playlistPopupMenu_counter" class="anylinkcss">
      <a href="?bt_track_name=$HTH::playlistPopupMenu_file&bt_track_rating=--">--</a>
      <a href="?bt_track_name=$HTH::playlistPopupMenu_file&bt_track_rating=*">&#149;</a>
      <a href="?bt_track_name=$HTH::playlistPopupMenu_file&bt_track_rating=**">&#149;&#149;</a>
      <a href="?bt_track_name=$HTH::playlistPopupMenu_file&bt_track_rating=***">&#149;&#149;&#149;</a>
      <a href="?bt_track_name=$HTH::playlistPopupMenu_file&bt_track_rating=****">&#149;&#149;&#149;&#149;</a>
      <a href="?bt_track_name=$HTH::playlistPopupMenu_file&bt_track_rating=*****">&#149;&#149;&#149;&#149;&#149;</a>
      </div>
   }
   variable browserPopupMenu_title
   variable browserPopupMenu_file
   variable browserPopupMenu_counter
   variable browserPopupMenu_folder
   #    <a href="#" onClick="return clickreturnvalue()" onMouseover="dropdownmenu(this, event, 'brs$HTH::browserPopupMenu_counter')">$HTH::browserPopupMenu_title</a>
   #    <a href="#" onMouseout="delayhidemenu()" onClick="return dropdownmenu(this, event, 'brs$HTH::browserPopupMenu_counter')">$HTH::browserPopupMenu_title</a>
   # cl <a href="#" onClick="return dropdownmenu(this, event, menu2, '120px')" onMouseout="delayhidemenu()">News Sites</a> (onclick)
   # Todo: Need to append &folder=
   variable browserRatingMenu {
      <a href="#" onClick="return clickreturnvalue()" onMouseover="dropdownmenu(this, event, 'brs$HTH::browserPopupMenu_counter')">$HTH::browserPopupMenu_title</a>
      <div id="brs$HTH::browserPopupMenu_counter" class="anylinkcss">
      <a href="?bt_track_name=$HTH::browserPopupMenu_file&bt_track_rating=--&folder=$HTH::browserPopupMenu_folder">--</a>
      <a href="?bt_track_name=$HTH::browserPopupMenu_file&bt_track_rating=*&folder=$HTH::browserPopupMenu_folder">&#149;</a>
      <a href="?bt_track_name=$HTH::browserPopupMenu_file&bt_track_rating=**&folder=$HTH::browserPopupMenu_folder">&#149;&#149;</a>
      <a href="?bt_track_name=$HTH::browserPopupMenu_file&bt_track_rating=***&folder=$HTH::browserPopupMenu_folder">&#149;&#149;&#149;</a>
      <a href="?bt_track_name=$HTH::browserPopupMenu_file&bt_track_rating=****&folder=$HTH::browserPopupMenu_folder">&#149;&#149;&#149;&#149;</a>
      <a href="?bt_track_name=$HTH::browserPopupMenu_file&bt_track_rating=*****&folder=$HTH::browserPopupMenu_folder">&#149;&#149;&#149;&#149;&#149;</a>
      </div>
   }

   variable head {
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>
   <head>
      <meta http-equiv="Content-Language" content="en-us">
      <meta name="GENERATOR"              content="SnackAmp Web Server">
      <meta name="ProgId"                 content="SnackAmp.Document">
      <LINK REL=StyleSheet                HREF="$css" TYPE="text/css" MEDIA=screen>
      <link rel=StyleSheet                HREF="/anylink.css" type="text/css" MEDIA=screen>

      <script type="text/javascript" src="/anylink.js">
      /***********************************************
      * AnyLink CSS Menu script-  Dynamic Drive DHTML code library (www.dynamicdrive.com)
      * This notice MUST stay intact for legal use
      * Visit Dynamic Drive at http://www.dynamicdrive.com/ for full source code
      ***********************************************/
      </script>
      <title>$title</title>
   }
   #<LINK REL=StyleSheet                HREF="/popupmenu.css" TYPE="text/css" MEDIA=screen>
   proc head {title css} {
      variable head
      set result [subst $head]
   }
   ##
   # Return the shorter of time to next track change or scheduled refresh
   #
   proc nextRefresh {args} {
      set rem [soundControl::remaining]
      if {[soundControl::isPlaying] && ($rem < $::snackAmpSettings(refreshRate))} {
         return [expr {$rem+1}]
      } else {
         return $::snackAmpSettings(refreshRate)
      }
   }
   proc getRefreshScript {period url} {
      Trace
      variable refreshScript
      return [subst $js::countDown]
   }
   proc coverArtShadow {content} {
         return [subst {
         <div id = "shadow-container">
            <div class = "shadow1">
               <div class = "shadow2">
                  <div class = "shadow3">
                     <div class = "container">
                        $content
                     </div>
                  </div>
               </div>
            </div>
         </div>
         }]
         #return [subst {<div class = "img-shadow">$content</div>}]
         # I will restore this when I get a drop shadow to both allow wrapping of text underneath and
         # works in ie
         #return [subst {<div>$content</div>}]
   }
   proc coverArt {folder size} {
      Trace
      foreach {ca valid} [albumart::covername $folder] {break}
      if {$valid} {
         return [subst {<img border=0 width="$size" height="$size" src="/coverart?get=[ENCODE $folder]" alt="[file tail $folder]">}]
      } else {
         if {$::snackAmpSettings(jukebox,nocoverart)} {
            return [subst {<img border=0 width="$size" height="$size" src="/remote/images/nocoverart.png" alt="No Cover Art">}]
         } else {
            return ""
         }
      }
   }
   proc rowStyle {_counter {mod 1}} {
      variable row_styles
      upvar $_counter counter
      if {[catch {incr counter} result] } then {
         set counter 0
      }
      return $row_styles([expr {($counter/$mod)&0x1}])
   }
   proc tableRow {cols {class ""}} {
      if {$class eq ""} {
         set result {<tr>}
      } else {
         set result [subst {<tr class='$class'>}]
         # This highlights the row
         #set result [subst {<tr class='$class' onMouseOver="this.className='highlight'" onMouseOut="this.className='$class'">}]
      }
      foreach {element} $cols {
         append result "<td>$element</td>"
      }
      append result "</tr>\n"
      return $result
   }
   proc inLinePlay {song} {
      return [subst {<a href="/remote/control?play=[ENCODE $song]" title="Play Track Immediately">$HTH::playbutton</a>}]
   }
   proc popUpProcess {what url song} {
      switch -- $what {
         play {set im $HTH::playbutton;set title "Play Track Immediately"}
         append {set im $HTH::appendbutton;set title "Append Track To Playlist"}
         insert {set im $HTH::insertbutton;set title "Queue Track to Play Next"}
         default {}
      }
      return [subst {<a href="javascript:textWindow('$url?$what=[ENCODE [ESCAPE $song]]')" title="$title">$im</A>}]
   }
   proc noamp {data} {
      return [string map {& %26} $data]
   }
   proc streamInfo {{class ""}} {
      set result ""
      foreach {file ip} [streamBuffer::allStreamInfo] {
         append result "[tableRow [list $file $ip] $class]\n"
      }
      return $result
   }
   proc displayFolder {File} {
      global snackAmpSettings
      if {$snackAmpSettings(ListDepth) > 0} {
         # File join doesnt work due to single list
         set File [string map {// /} $File]      ;# remove UNC stuff
         set name [file dirname [join [lrange [file split $File] end-$snackAmpSettings(ListDepth) end] /]]
      } else {
         set name [file tail [file dirname $File]]
      }
      return $name
   }
   proc popupInfo {ref {info "Show Track Details in Pop-up"}} {
      # Escape single quotes for javascript popup window,
      # Encode & to prevent confusion with parameter delimiter
      set ref "/remote/search?question=[ENCODE [ESCAPE $ref]]"
      return [subst {<a href="javascript:popupInfo('$ref',250,600,'Info')" title='$info'>$HTH::infobutton</a>}]
   }

   #------------------------------------------------------------------------------
   #  Clear data for current track, forcing a refresh
   #------------------------------------------------------------------------------
   proc clearCache {{File {}}} {
      variable InfoCache
      variable id3Cache
      if {$File ne ""} {
         unset -nocomplain id3Cache($File)
         unset -nocomplain InfoCache($File)
      } else {
         unset -nocomplain InfoCache
         unset -nocomplain id3Cache
      }
   }

   #------------------------------------------------------------------------------
   # Pop-up menu html, this subst scheme is kludgy
   #------------------------------------------------------------------------------
   proc playlistRatingMenu {text file} {
      set HTH::playlistPopupMenu_title [string map {* &#149;} $text]
      set HTH::playlistPopupMenu_file [ENCODE $file]
      set HTH::playlistPopupMenu_counter [clock clicks]
      set track [subst -nocommands $HTH::playlistRatingMenu]
   }
   proc browserRatingMenu {text file folder} {
      set HTH::browserPopupMenu_title [string map {* &#149;} $text]
      set HTH::browserPopupMenu_file [ENCODE $file]
      set HTH::browserPopupMenu_folder [ENCODE $folder]
      set HTH::browserPopupMenu_counter [clock clicks]
      set track [subst -nocommands $HTH::browserRatingMenu]
   }

}
##
# Schedule a variable to be unset some time in the future
# TODO: Could improve to not litter with a ton of afters,
# but have one that fires every 10 seconds to iterate
# over an internal list.
proc HTH::cleanLater {_var min} {
   after cancel [list HTH::clean $_var]
   after [expr {$min*60000}] [list HTH::clean $_var]
}

#------------------------------------------------------------------------------
# Return pop-up text, requires some css from
# http://psacake.com/web/jl.asp
# css added to anylink.css
#------------------------------------------------------------------------------
proc HTH::popInfo {shown popup} {
   set popup [string map {\n <BR>} $popup]
   return [subst -nocommands {<a class=info href="#">$shown<span>$popup</span></a>}]
}
#----------------------------------------------------------------------------
# Return track data into an array, use a cache if it exists
# With metakit we prob. don't need a cache
#----------------------------------------------------------------------------
proc HTH::getTrackData {file _Info} {
   upvar $_Info Info
   variable id3Cache
   if {[info exists id3Cache($file)]} {
      array set Info $id3Cache($file)
   } else {
      unset -nocomplain info
      ;# Get from database if possible
      set hasTag [db::lookup $file Info 1]
      # Insert the tag data if available
      foreach item {Title Artist Album} {
         if {$Info($item) eq ""} {
            set Info($item) [id3Tag::guess${item} $file]
         }
      }
      # Save for next time
      set HTH::id3Cache($file) [array get Info]
   }
   HTH::cleanLater HTH::id3Cache($file) 10
}
proc HTH::leader {title css {url "xx"}} {
   Trace
   # title is used in subst
   append result [head $title $css]
   if {$url ne "xx"} {
     append result \n
     append result [subst {    <META HTTP-EQUIV="Refresh" CONTENT="$::snackAmpSettings(refreshRate);URL=$url">}]
     append result \n
   }
   append result "  </head>\n"
}
#------------------------------------------------------------------------------
# Function   : ::HTH::trackDetails
# Description: Return HTML for the song details, long version
# Author     : Tom Wilkason
# Date       : 4/29/2002
#------------------------------------------------------------------------------
proc HTH::trackDetails {_Details} {
   upvar $_Details Details
   variable trackDetailsTemplate
   set Details(duration) [HMSfrac $Details(Duration)]
   set result [subst $trackDetailsTemplate]
   return $result
}
#----------------------------------------------------------------------------
# Short Version with user selectable ratin
#----------------------------------------------------------------------------
proc HTH::trackSummary {_Details} {
   upvar $_Details Details
   variable trackSummaryTemplate
   set result [subst $trackSummaryTemplate]
   return $result
}
#----------------------------------------------------------------------------
# Short Version without user rating
#----------------------------------------------------------------------------
proc HTH::trackSummaryNR {_Details} {
   upvar $_Details Details
   variable trackSummaryTemplateNR
   set result [subst $trackSummaryTemplateNR]
   return $result
}

#------------------------------------------------------------------------------
# Function   : trackInfo
# Description: Return, in a table, the details on some file
#              Called each update
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc HTH::trackInfo {fileName {short 1}} {
   variable InfoCache
   variable initDetails
   if {[info exists InfoCache($fileName)]} {
      array set Details $InfoCache($fileName)
   } else {
      # Init data with html blanks
      array set Details [array get initDetails]
      if {[db::lookup $fileName Details 1]<=0} {
         HttpD::Log "trackInfo:$error"
      }
      # Stuff not in the database
      set Details(Cover) [HTH::coverArtShadow [HTH::coverArt $Details(Folder) $::snackAmpSettings(coversize)]]
      set Details(Name) [file tail $fileName]
      set Details(Position) "N/A"
      # Fill in non tagged data based on file name
      foreach item {Title Artist Album} {
         if {$Details($item)=="&nbsp;"} {
            set Details($item) [id3Tag::guess${item} $fileName]
         }
      }
      ##
      # File info
      #
      if {![catch {file stat $fileName Stat}]} {
         set Details(fSize) "[commify $Stat(size)] Bytes"
         set Details(fDate) [timeFormat $Stat(mtime)]
      }
      # Store all this good stuff for fast access later
      set InfoCache($fileName) [array get Details]
   }
   # Don't keep it around forever
   HTH::cleanLater HTH::InfoCache($fileName) 10
   # Return properly formatted details
   switch -- $short {
      0 {set result [HTH::trackDetails Details]}
      1 {set result [HTH::trackSummary Details]}
      2 {set result [HTH::trackSummaryNR Details]}
      default {}
   }
   return $result
}
##
# Show play status
#
proc HTH::playStatus {{showWhere 1}} {
   set Active [string tolower [soundControl::playState]]
   if {$showWhere} {
      append result [H2 "SnackAmp currently $Active on [info hostname]:$::HttpD::Httpd(port)"]
   } else {
      append result [H2 "SnackAmp currently $Active"]
   }
   return $result
}
#------------------------------------------------------------------------------
# Function   : Return the current playlist, with appropriate controls
# Description:
# Author     : Tom Wilkason
# Date       : 7/2/2003
#------------------------------------------------------------------------------
proc HTH::trackList {control top newEnd {parms ""}} {
   Trace
   set result ""
   if {$::snackAmpSettings(listbytag)} {
      set fHeader "Artist"
   } else {
      set fHeader "Folder"
   }
   switch -- $control {
      0 {append result [HTH::tableRow [list $fHeader "Track Name"] headerRow]}
      1 {append result [HTH::tableRow [list ? X Rating $fHeader "Track Name"] headerRow]}
      2 {
         if {$::snackAmpSettings(jukebox,popup)} {
            append result [HTH::tableRow [list ? $fHeader "Track Name"] headerRow]
         } else {
            append result [HTH::tableRow [list $fHeader "Track Name"] headerRow]
         }
      }
      default {}
   }
   set ctr 0
   set lbt $::snackAmpSettings(listbytag)
   for {set i $top} {$i < $newEnd} {incr i} {
      set track ""
      if {$i == [LB::idxPlaying]} {
         set bgc activeRow
         incr ctr
      } else {
         set bgc [HTH::rowStyle ctr 1]
      }
      set song [LB::keyFromIndex $i]
      set file [LB::nameFromIdx $i]

      # Show either tags or folder/file names
      if {$lbt} {
         HTH::getTrackData $file Info
         set track $Info(Title)
         set folder $Info(Artist)
         set rating $Info(Rating)
         set MoreInfo "Show Track Details in Pop-up, Album=$Info(Album)"
      } else {
         set folder [HTH::displayFolder $file]
         set track [ROOTTAIL $file]
         set rating "--"
         set MoreInfo "Show Track Details in Pop-up"
      }
      ##
      # Show the appropriate row with appropriate controls
      #
      switch -- $control {
         0 {
            # No Control
            if {$folder eq ""} {
               set flink [SP]
            } else {
# Note: hyperlink for /display only
               set flink $folder
            }
            append result [HTH::tableRow [list $flink $track] $bgc]
         }
         1 {
            # Full Control
            if {$folder eq ""} {
               set flink [SP]
            } else {
               # Hyperlink only if remote controling (i.e. control= 1)
               # /remote is current (need to fix to use remote/browser as current)
               set flink [AREF "browser?folder=[ENCODE [file dirname $file]]" $folder "Browse This Folder"]
            }
            ##
            # Return the linked info
            #
            set rating [playlistRatingMenu $rating $file]
            set track "[HTH::inLinePlay $file][NOAMP $track]"
            set imInfo [HTH::popupInfo $file $MoreInfo]
            #set imInfo "[HTH::popupInfo $file][HTH::popInfo xxx {This is a test}]"
            set imDel [AREF "?delete=$i$parms" $HTH::deleteimage "Remove Track From Playlist"]
            append result [HTH::tableRow [list $imInfo $imDel $rating $flink $track] $bgc]
         }
         2 {
            # Just hyperlink (jukebox)
            if {$folder eq ""} {
               set flink [SP]
            } else {
               if {$::snackAmpSettings(jukebox,linkartist)} {
                  set flink [AREF "?folder=[ENCODE [file dirname $file]]" $folder "Browse This folder"]
               } else {
                  set flink $folder
               }
               # /jukebox/browser is current
            }
            if {$::snackAmpSettings(jukebox,popup)} {
               set imInfo [HTH::popupInfo $file]
               append result [HTH::tableRow [list $imInfo $flink $track] $bgc]
            } else {
               set imInfo [HTH::popupInfo $file]
               append result [HTH::tableRow [list $flink $track] $bgc]
            }
         }
         default {}
      }
   }
   return $result
}
#------------------------------------------------------------------------------
# Function   : HttpHelpers
# Description: Create a temp file, remove it 30 seconds later
# Author     : Tom Wilkason
# Date       : 7/8/2003
#------------------------------------------------------------------------------
proc HTH::WriteTempFile {ext content} {
   set mypath [file join $::TA(TEMP) [clock clicks].$ext]
   # temp file
   set fid [open $mypath w]
   fconfigure $fid -translation binary
   puts -nonewline $fid $content
   close $fid
   after 30000 [list file delete -force $mypath]
   return $mypath
}

#------------------------------------------------------------------------------
# Return a option list of sorting modes
#------------------------------------------------------------------------------
proc HTH::sortModes {args} {
   foreach {mode} [subSort::allSortModes] {
      append result "<option>$mode</option>\n"
   }
   return $result
}
#----------------------------------------------------------------------------
# Return a list of search modes
#----------------------------------------------------------------------------
proc HTH::searchModes {args} {
   foreach {mode} {Any File Title Artist Album Note Genre Folder Track Rate Tag} {
      if {$mode eq "Any"} {
         append result "<option selected>$mode</option>\n"
      } else {
         append result "<option>$mode</option>\n"
      }
   }
   return $result
}
proc HTH::plsTracks {num} {
   return "\[playlist\]\nnumberofentries=$num\n"
}
#------------------------------------------------------------------------------
# Function   : nextToPlay
# Description: Print out the next N tracks in the playlist
#
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc HTH::nextToPlay {playButton {N 20} {parms ""}} {
   Trace
   global LB
   set offset 2
   set result {}
   if {[catch {
      set top [expr {[LB::idxPlaying]-$offset}]
      set end [LB::index end]
      set newEnd [expr {$top+$N+$offset}]
      if {$newEnd > $end} {set newEnd $end}
      if {$top < 0} {set top 0}
      set current [file tail $::TA(currentMF)]
      append result [H2 "Coming up next (of [commify $end] tracks)"]
      append result "<table class='songPlaylist'>"
      append result [HTH::trackList $playButton $top $newEnd $parms]
      append result "</table>"
   } error]} {
      return $error
   } else {
      return $result
   }
}
##
# Return list of tracks that are queued by the user to play
#
proc HTH::queuedToPlay {playButton {N 20} {parms ""}} {
   Trace
   global LB
   set offset $N
   set result {}
   if {[catch {
      set top [expr {[LB::idxPlaying]+1}]
      set end [LB::userInsert]
      if {$top < 0} {set top 0}
      if {$end-$top <= 0} {
         #append result [H2 "No tracks queued"]
      } else {
         # Limit the number of queued tracks
         if {$end-$top > $::snackAmpSettings(httpNextLength)} {
            set end [expr {$top+$::snackAmpSettings(httpNextLength)}]
         }
         #set current [file tail $::TA(currentMF)]
         append result [H2 "User queued tracks"]
         append result "<table class='songPlaylist'>"
         append result [HTH::trackList $playButton $top $end $parms]
         append result "</table>"
      }
#      append result </div>
   } error]} {
      return $error
   } else {
      return $result
   }
}
##
# Unset a variable
#
proc HTH::clean {_var} {
   upvar $_var var
   if {[info exists var]} {
      unset var
   }
}
##
# Verify a file is within one of the media folders
#
proc HTH::isMedia {track} {
   foreach media $::TA(mediaDirs) {
      if {[string match $media* $track]} {
         return 1
      }
   }
   return 0
}
#------------------------------------------------------------------------------
# Function   : ::remote/control::plFiles
# Description: Callback from TML file to return a list of all playlist files
# Author     : Tom Wilkason
# Date       : 4/28/2002
#------------------------------------------------------------------------------
proc HTH::plFiles {args} {
   Trace
   set result ""
   # Create option list of all AutoPlaylist files
   foreach {file} [catalog::getPlaylists] {
      set choice [ROOTTAIL $file]
      if {$choice eq $::TA(PlayListSelection)} {
         append result "<option selected>$choice</option>\n"
      } else {
         append result "<option>$choice</option>\n"
      }
   }
   return [string trim $result]
}
#------------------------------------------------------------------------------
# Function   : ::remote/control::aplFiles
# Description: Callback from TML file to return a list of all AutoPlaylist files
# Author     : Tom Wilkason
# Date       : 4/28/2002
#------------------------------------------------------------------------------
proc HTH::aplFiles {args} {
   Trace
   set result ""
   # Create option list of all AutoPlaylist files
   foreach {file} [catalog::getAutoPlaylists] {
      set choice [ROOTTAIL $file]
      if {$choice eq $::TA(AutoPlayListSelection)} {
         append result "<option selected>$choice</option>\n"
      } else {
         append result "<option>$choice</option>\n"
      }
   }
   return [string trim $result]
}

#------------------------------------------------------------------------------
#                              j u s t c o n t r o l
#------------------------------------------------------------------------------

namespace eval remote/justcontrol {
   proc init { } {
     return 1
   }
}
proc remote/justcontrol::process {sock {params ""}} {
   Trace
   foreach param [split $params &]  {
      foreach {CmdType Cmd} [split $param =] {break}
      set Cmd [HttpD::Decode $Cmd]
      switch -- $CmdType {
         bt_Level {
            # Would be nice not to have to push data back
            # maybe hold to end to push it out if needed.
            if {[catch {::macros::bumpLevel $Cmd} result] } then {
               saLog $result
            }
         }
         bt_skip {
            if {[catch {::macros::skipPlay $Cmd} result] } then {
               saLog $result
            }

         }
         default {}
      }
   }
   return [list dhtml html 0]
}
#------------------------------------------------------------------------------
# Function   : streaming/live namespace
# Description: Process streaming files
# Author     : Tom Wilkason
# Date       : 3/13/2003
#------------------------------------------------------------------------------
namespace eval streaming/live {
   proc init {args} {
      package require urlStreaming
      return 1
   }
}
#------------------------------------------------------------------------------
#                              s t r e a m i n g  f i l e  b r o w s e r
#------------------------------------------------------------------------------
namespace eval streaming/browser {
   proc init { } {
     package require urlStreaming
     return $::snackAmpSettings(EnableStreaming)
   }
   variable brform {
  <H2>Enter Play Folder Option</H2>
  <form name=option_form>
  <TABLE>
    <TR>
      <TD>Sorting</TD>
      <TD>Streaming</TD>
      <TD>Subfolders</TD>
    </TR>
    <TR>
     <TD width="300em">
       <select size="1" name="cmb_Sort" class="formText">
         [HTH::sortModes]
       </select>
     </TD>
     <TD>
       <select size="1" name="cmb_How" class="formText">
         <option selected>stream</option>
         <option>tracks</option>
       </select>
     </TD>
     <TD>
       <select size="1" name="cmb_Sub" class="formText">
         <option selected>no</option>
         <option>yes</option>
       </select>
     </TD>
    </TR>
  </TABLE>
  </form>
   }
}
namespace eval streaming/playlists {
   proc init { } {
     package require urlStreaming
     return 1
   }
}
##
# playlist namespace, returns a nicly formatted list of playlist and other
# to build up streaming playlists.
#
namespace eval streaming/getplaylist.pls {
   proc init { } {
     package require urlStreaming
     return 1
   }
}
#------------------------------------------------------------------------------
# Function   : search namespace
# Description: Handles interface to request songs to be added to playlist
#              Can be called from different URL types and reacts accordingly
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
namespace eval search {
   variable myns
   variable control
   variable search_form {
    <!-- Search Database -->
    <table border="1" width="100%" bordercolor="darkblue" cellspacing="0" cellpadding="0">
      <tr>
        <td width="100%" bgcolor="darkblue" align="center"><font color="white" face="Helvetica">
        <b>Search All Tracks for Matching Text</b></font></td>
      </tr>
      <tr>
        <td width="100%" align="left">
          <form method="GET" action="/search" TITLE="Enter text then hit Search to find all songs that contain the text">
            <input type="submit" value="Search" name="Button">
            <font face="Helvetica">
              <input type="text" name="jb_search" size="30">
            </font>
          </form>
        </td>
      </tr>
    </table>
    <p>
   }
   variable close_window {<a href="" onClick="window.close()">Click this to close the window</A>}
   ##
   # Javascript helper routines
   #
   proc init { } {
     package require urlSearch
     return 1
   }
}

#------------------------------------------------------------------------------
# Wrapper for remote/search function, makes the search namespace react based
# on calling function
#------------------------------------------------------------------------------

namespace eval remote/browser {
   proc init { } {
     package require urlRemote
     return 1
   }
}
namespace eval remote/search {
   proc init { } {
     package require urlSearch
     return 1
   }
}
namespace eval remote/control {
   variable Details
     # <meta http-equiv="Content-Type" content="text/html; charset=windows-1252">
    #  <meta http-equiv="pragma"           content="no-cache">

   variable head {
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>
   <head>
      <meta http-equiv="Content-Language" content="en-us">
      <meta name="GENERATOR"              content="SnackAmp Web Server">
      <meta name="ProgId"                 content="SnackAmp.Document">
      <LINK REL=StyleSheet                HREF="$css" TYPE="text/css" MEDIA=screen>
      <title>$title</title>
   }
   proc init { } {
     package require urlRemote
     return 1
   }
}
#------------------------------------------------------------------------------
#                              m i n i
#------------------------------------------------------------------------------
# mini remote control (for P80x phones :-)
# returns a mini playlist
#
namespace eval remote/mini {
   proc init { } {
     package require urlMinicontrol
     return 1
   }
}
# mini remote control (for sony psp :-)
# returns a mini playlist
namespace eval remote/psp {
   proc init { } {
     package require urlMinicontrol
     return 1
   }
}
#------------------------------------------------------------------------------
#                              j u k e b o x
#------------------------------------------------------------------------------
namespace eval jukebox/display {
   proc init { } {
     package require urlJukebox
     return 1
   }
}
##
# Jukebox current playing
#
namespace eval jukebox/header {
   proc init { } {
     package require urlJukebox
     return 1
   }
}
##
# Jukebox next
#
namespace eval jukebox/next {
   proc init { } {
     package require urlJukebox
     return 1
   }
}
##
# Jukebox next
#
namespace eval jukebox/last {
   proc init { } {
     package require urlJukebox
     return 1
   }
}
namespace eval jukebox/folder {
   proc init { } {
     package require urlJukebox
     return 1
   }
}
namespace eval jukebox/browser {
   proc init { } {
     package require urlJukebox
     return 1
   }
}
namespace eval jukebox/search {
   proc init { } {
     package require urlJukebox
     return 1
   }
}


#------------------------------------------------------------------------------
#                              coverart server processing
#------------------------------------------------------------------------------
namespace eval coverart {
   proc init { } {
     return 1
   }
   variable cache
}
# Returns a local file that contains coverart
# For security only handles specific file names
proc coverart::process {sock {params ""}} {
   Trace
   if {[string length $params]} {
      foreach param [split $params &]  {
         foreach {CmdType Cmd} [split $param =] {break}
         switch -- $CmdType {
            get {
               set folder [HttpD::Decode $Cmd]
               foreach {img valid} [albumart::covername $folder] {break}
               if {[string length $img]} {
                  set ext [string trimleft [file extension $img] .]
                  return [list file $ext $img]
               }
            }
         } ;# end switch
      }
   }
   return [list dhtml html ""]
}
#------------------------------------------------------------------------------
#                              d i s p l a y
#------------------------------------------------------------------------------
namespace eval display {
   proc init { } {
     return 1
   }
}
#------------------------------------------------------------------------------
# Function   : display
# Description: Main function that is called by web server when /display
#              is entered.
#              e.g. /display
# Author     : Tom Wilkason
# Date       : 12/15/2002
#------------------------------------------------------------------------------
proc display::process {sock {params ""}} {
   Trace
   append Result [HTH::leader "SnackAmp Playlist" $::snackAmpSettings(cssName)]
   append Result "<body>\n"
   append Result [HTH::getRefreshScript [HTH::nextRefresh] "/display"]
   # Wait to let player take hold
   #append Result ${search::close_window}
   append Result [HTH::playStatus 0]
   append Result [HTH::trackInfo [soundControl::fileName] 1]
   append Result [HTH::nextToPlay 0 $::snackAmpSettings(httpNextLength)]
   append Result "</body>\n</html>"
   return [list dhtml html $Result]
}


package provide urlRedirect 1.0
##
# Utility routines
#
proc dumpSocket {} {
   foreach glb [info globals Httpdsock*] {
      global $glb
      puts "Array $glb--------------"
      catch {parray $glb}
   }
}
proc afters {args} {
   foreach after [after info] {puts [after info $after]}
}

