#------------------------------------------------------------------------------
# snackAmp - urlRemote package 1.0 for snackAmp
# provides dynamic http applications for SnackAmp
# web server.
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#                               c o n t r o l
#------------------------------------------------------------------------------
##
# control namespace to control the SnackAmp player from
# a web browser. When user enters /control ... in to the browser
# The process function of this namespace will be called passing any
# user parameters. To add a new command then create another namespace
# and provide an init and process function.
#
#------------------------------------------------------------------------------
# Function   : remote/control::process state machine
# Description: Main function that is called by web server when /remote/control
#              is entered. Inside is a loop that iterates over each parameter
#              on the POST/GET and handles each one.
#              e.g. /remote/control?a=2&b=c  <= two params a= and b=
# Author     : Tom Wilkason
# Date       : 4/24/2002
#------------------------------------------------------------------------------
proc remote/control::process {sock {params ""}} {
   Trace
   ##
   # Replace puts with appending to string then doing as a batch if needed
   #
   append Result [HTH::leader "SnackAmp Control" $::snackAmpSettings(cssName)]
   append Result "<body>\n"
   append Result [js::scripts popupInfo]
   append Result [HTH::getRefreshScript [HTH::nextRefresh] "/remote/control?Status=Refresh+Data"]
   foreach param [split $params &]  {
      foreach {CmdType Cmd} [split $param =] {break}
      set Cmd [HttpD::Decode $Cmd]
      switch -- $CmdType {
         bt_Control {
            if {[catch {::macros::$Cmd} result] } then {
               append Result "<p>$::errorInfo</p>\n"
            }
         }
         bt_skip {
            if {[catch {::macros::skipPlay $Cmd} result] } then {
               append Result "<p>$::errorInfo</p>\n"
            }
         }
         bt_vote {
            # TODO: Add code to increment the vote count
            # Force a re-read of database for this track to display updates
            HTH::clearCache [soundControl::fileName]
         }
         bt_rating {
            # Set the track rating
            # Force a re-read of database for this track to display updates
            HTH::clearCache [soundControl::fileName]
            db::setVal [soundControl::fileName] Rating $Cmd
            trackList::updateFiles [list [soundControl::fileName]]
            set ::songArray(Rating) [mrating $Cmd]
         }
         bt_track_name {
            # step 1
            set bt_track_name $Cmd
         }
         bt_track_rating {
            # step 2
            # Set the track rating for some track
            HTH::clearCache $bt_track_name
            db::setVal $bt_track_name Rating $Cmd
            trackList::updateFiles [list $bt_track_name]
            if {[soundControl::fileName] eq $bt_track_name} {
               set ::songArray(Rating) [mrating $Cmd]
            }

         }
         bt_Level {
            if {[catch {::macros::bumpLevel $Cmd} result] } then {
               append Result "<p>$::errorInfo</p>\n"
            }
         }
         play {
            # Play a specific song
            # /remote/control?play=some+name.mp3
            if {[catch {playSpecificSong $Cmd} error] } then {
               debug $result
               append Result $result
            }
         }
         delete {
            LB::deleteIndex $Cmd
            gui::updateFileDisplay
            # Remove from playlist
         }
         cmb_APL {
            # Select an AutoPlaylist file
            # TFW: If a real + is in file name, this may fail
            after 200 [list macros::readAPLFile [file join $::HOME $Cmd.tafs] 1]
         }
         cmb_Sort {
            # Sort/Shuffle the files
            set ::snackAmpSettings(sortModeTxt) $Cmd
            set subSort::sortSelection $::snackAmpSettings(sortModeTxt)
            # Hack, if a file then a custom sort must be read in
            set fname [file join $::HOME $Cmd.ss]
            if {[file exists $fname]} {
               subSort::readFile $Cmd
            }
            after 200 filtering::sortPlaylist
         }
         filter_text {
            # stage 1
            set ::TA(filter) $Cmd
            # Note: These will happen in this order, search_replace can process
            # Button=Filter&filter_text=ab&filter_what=All+Files&search_replace=Replace
         }
         filter_what {
            # stage 2
            switch -glob -- [string tolower $Cmd] {
               "all*files" {set Search_What 1}
               "play*" {set Search_What playList}
               "auto*" {set Search_What 0}
               default {set Search_What 1}
            }
         }
         search_replace {
            # stage 3, do the filtering
            if {![info exists Search_What]} {set Search_What 1}
            set Search_Replace [string tolower $Cmd]
            filtering::pushFilter
            macros::filterMusic $Search_Replace $Search_What
            # stage 3
         }
         cmb_PL {
            # Stage 1
            # Set the playlist file
            foreach ext {m3u taf} {
               set Playlist_File [file join $::HOME $Cmd.$ext]
               if {[file exists $Playlist_File]} {
                  break
               }
            }
         }
         PL_search_replace {
            # Stage 2
            # And then apply it
            macros::ProcessPlaylistFile $Playlist_File [string tolower $Cmd] 0
         }
         bt_Submit {
            # This isn't handled specifically, it comes first
         }
         default {}
      }
   }
   # Wait to let player take hold
   append Result [HTH::playStatus 1]
   append Result [HTH::trackInfo [soundControl::fileName] 1]

   append Result [HTH::nextToPlay 1 $::snackAmpSettings(httpNextLength)]
   append Result "</body>\n</html>"
   return [list dhtml html $Result]
}
#----------------------------------------------------------------------------
# File browser, when user clicks on artist name
#----------------------------------------------------------------------------
proc remote/browser::process {sock {params ""}} {
   Trace
   package require urlBrowser
   dirBrowser::doit $sock $params
}

package provide urlRemote 1.0
