#------------------------------------------------------------------------------
# snackAmp - urlSearch package 1.0 for snackAmp
# provides dynamic http applications for SnackAmp
# web server.
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
# Function   : search::doit state machine
# Description: Main function that is called by web server when /search
#              is entered.
#              e.g. /search?jb_search_pl=*abc*
# Author     : Tom Wilkason
# Date       : 12/15/2002
#------------------------------------------------------------------------------

# TFW: /search is used now but search needs to be called as /remote/search or /jukebox/search
proc remote/search::process {sock {params ""}} {
   search::doit $sock $params
}
proc search::doit {sock {params ""}} {
   Trace
   variable search_form
   variable javascript
   upvar #0 Httpd$sock data
   variable myns [uplevel {string trim [namespace current] :}]
   append Result [HTH::leader "SnackAmp Search Page" $::snackAmpSettings(cssName)]
   append Result "<body>\n"
   append Result [js::scripts popupInfo toggleAll_jb_add toggleAll_jb_search_pl foldersel getselected]
   append Result [subst {<a href="/$myns?$params">Refresh</a>|}]
   append Result ${search::close_window}
   switch -glob -- $myns {
      jukebox* {
         set cont 0
      }
      remote* {
         set cont 1
      }
      default {}
   }
   variable control $cont
   # Appropriate link to browser url e.g. /jukebox/browser
   set myBrowser [file join [file dirname $myns] browser]
   ##
   # Need to have search_form optional
   #
   #append Result [subst $search_form]
   set Search_What pl
   set Search_Fields Any
   set Tracks [list]
   set ratings [list]
   foreach param [split $params &]  {
      foreach {CmdType Cmd} [split $param =] {break}
      set Cmd [HttpD::Decode $Cmd]
      switch -- $CmdType {
         jb_search {
            # stage 1
            set filter $Cmd
         }
         rating {
            # stage 1
            lappend ratings $Cmd
         }
         cmb_Search {
            if {$Cmd == "Any"} {
               set Search_Fields [list Title Artist Album File]
            } else {
               set Search_Fields $Cmd
            }
         }
         search_what {
            # stage 2
            # TODO: search by ratings
            set re [db::buildRE $ratings]
            switch -glob -- [string tolower $Cmd] {
               "all*files" {
                  # User specified text to search catalog
                  set maxfind [expr {$::snackAmpSettings(httpNextLength)*5}]
                  append Result [H2 "The following tracks matched your search criteria '$filter', limited to [commify $maxfind] matches"]
                  append Result [H2 "Please check desired tracks and press <i>Queue Selected</i> at the bottom to add them to the playlist"]
                  # Pass params for putting in back button on next form
                  append Result [search::searchCatalog $re $Search_Fields $myBrowser $filter $maxfind $params]
               }
               "play*" {
                  # User specifed some text to search within the playlist
                  # This goes in a pop-up
                  append Result [H2 "The following tracks from the Play List matched your search criteria '$filter'"]
                  append Result [search::searchPlaylist $re $Search_Fields $myBrowser $filter 0]
               }
               "auto*" {
                  # User specifed some text to search within the playlist
                  # This goes in a pop-up
                  append Result [H2 "The following tracks from the AutoPlaylist matched your search criteria '$filter'"]
                  append Result [search::searchPlaylist $re $Search_Fields $myBrowser $filter 1]
               }
               default {
                  saLog "[me] Unknown:$Cmd"
               }
            }
         }
         play {
            # Play a specific song
            # /control?play=some+name.mp3
            if {[catch {LB::InsertPlay $Cmd} error] } then {
               debug $result
               append Result $result
            }
         }
         delete {
            LB::deleteIndex $Cmd
            gui::updateFileDisplay
            # Remove from playlist
         }
         question {
            # This goes in a pop-up
            # javascript pop-up track details (like playinfo)
            if {![file exists $Cmd]} {
               set Cmd [LB::nameFromKey $Cmd]
            }
            # TFW: would be nice to make the fonts smaller
            append Result [HTH::trackInfo $Cmd 0]
         }
         jb_add_start {
            # When processing a playlist add, this is the first param
            set playing [LB::idxActive]
            set old_end [LB::idxEnd]
            # The back reference uses the jb_add_start param (which is the params from the
            # request that originated the search) which was passed in here. It is saved
            # In a hidden field from the prior form.
            append Result |[AREF "/$myns?$Cmd" Back]\n
            append Result [H2 "Added the following tracks to the playlist"]
            append Result "<table class='songPlaylist'>\n"
         }
         jb_add {
            # When processing a playlist add, these are the middle params, one per item to add
            set next [LB::idxFile $Cmd]
            if {$control==1 || $next==-1 || $next<=$playing} {
               set track [NOAMP [file rootname [file tail $Cmd]]]
               set folder [NOAMP [HTH::displayFolder $Cmd]]
               append Result [HTH::tableRow [list $folder $track] headerRow]
               lappend Tracks $Cmd
            }
         }
         jb_append_process {
            # When processing a playlist append, this is the final param, signals end of list
            append Result "</table>\n"
            if {[llength $Tracks] > 0} {
               append Result [H2 "Your requested tracks were queued to play"]
               LB::userInsertList $Tracks
            } else {
               append Result [H2 "No tracks selected or tracks already in playlist"]
            }
         }
         jb_insert_process {
            ;# When processing a playlist insert, this is the final param, signals end of list
            append Result "</table>\n"
            if {[llength $Tracks] > 0} {
               append Result [H2 "Your requested tracks are now queued to play next"]
               set cp [LB::idxPlaying]
               LB::insertList $cp $Tracks
            } else {
               append Result [H2 "No tracks selected or tracks already in playlist"]
            }
         }
      }
   }
   # Insert Display Playlist link
   append Result "</body>\n</html>"
   return [list dhtml html $Result]
}
#------------------------------------------------------------------------------
# Function   : searchPlaylist
# Description: Based on a pattern from the web entry box, filter the current
#              playlist and return the matching results
#
# Author     : Tom Wilkason
# TODO: Filter results by regexp on the passed in re for ratings
#------------------------------------------------------------------------------
proc search::searchPlaylist {re fields parNs filter {aplOnly}} {
   variable control
   variable myns
   set i 0
   set result {<table class='songPlaylist'>}
   if {$aplOnly} {
      set files [aplFolder::getAPLFileList]
   } else {
      set files [LB::keyOrNamesAll 1]
   }
   if {$control==1} {
      append result [HTH::tableRow [list Num ? P X Rating Artist Album "Track Name"] headerRow]
   } else {
      append result [HTH::tableRow [list Num ? Rating Artist Album "Track Name"] headerRow]
   }
   foreach {row} [db::filteredView $files $fields $filter $re] {
      foreach {file album artist track folder rating} $row {break}
      set bgc [HTH::rowStyle j 5]
      if {$album eq ""} {
         set album [file tail $folder]
      }
      if {$folder eq ""} {
         set flink $album
      } else {
         set flink [AREF "/$parNs?folder=[ENCODE $folder]" $album]
      }
      set Info [HTH::popupInfo $file]
      # The second param just performs another search using the same filter
      set rating [string map {* &#149;} $rating]
      if {$control==1} {
         set Play [subst {<a href="/$myns?play=[ENCODE $file]&jb_search=$filter&search_what=Playlist" title="Play Track Immediately">$HTH::playbutton</a>}]
         set Delete [subst {<a href="/$myns?delete=$i&jb_search=$filter&search_what=Playlist">$HTH::deleteimage</a>}]
         append result [HTH::tableRow [list $i $Info $Play $Delete $rating $artist $flink $track] $bgc]
      }  else {
         append result [HTH::tableRow [list $i $Info $rating $artist $flink $track] $bgc]
      }
      incr i
   }
   append result "</table>"
   #append result "<a href='javascript:toggleAll_jb_search_pl(true)'>Select All</a> | <a href='javascript:toggleAll_jb_search_pl(false)'>Unselect All"
   return $result
}

#------------------------------------------------------------------------------
# Function   : search::searchCatalog
# Description: Search all songs (up to N) for results
#              Returns HTML shown all the tracks that matched the filter
#              Info and checkboxes are included
# Author     : Tom Wilkason
# Date       : 12/22/2002
#------------------------------------------------------------------------------
proc search::searchCatalog {re fields parNs filter max parms} {
   variable control
   variable myns
   append result "<FORM METHOD='GET' ACTION='/$myns' NAME='jb_add_form'>"
   append result "<input type='hidden' name='jb_add_start' value='$parms'>"
   # Initial param will be jb_add_start, followed by each checked item as jb_add, ending with jb_append_process=Submit
   append result {<table class='songPlaylist'>}
   if {$control==1} {
      append result [HTH::tableRow [list Chk ? P Rating Artist Album "Track Name"] headerRow]
   } else {
      append result [HTH::tableRow [list Chk ? Rating Artist Album "Track Name"] headerRow]
   }
   ##
   # Search all files for match
   #
   set i 0
   # TODO: Move this mk::select into metakit.tcl
   foreach {row} [mk::select $db::DB -count $max -globnc $fields *$filter* -regexp Rating $re -exact Mounted 1] {
      foreach {file artist album track folder rating} [mk::get $db::DB!$row File Artist Album Title Folder Rating] {break}
      set bgc [HTH::rowStyle i 5]
      if {$album eq ""} {
         set album [file tail $folder]
      }
      if {$folder eq ""} {
         set flink $album
      } else {
         set flink [AREF "/$parNs?folder=[ENCODE $folder]" $album]
      }
      set cb [subst -nocommands {<INPUT TYPE=checkbox NAME="jb_add" VALUE="$file">}]
      set infoBtn [HTH::popupInfo $file]
      # The second param just performs another search using the same filter
      set rating [string map {* &#149;} $rating]
      if {$control} {
         set playBtn [subst {<a href="/$myns?play=[ENCODE $file]&jb_search=$filter&search_what=AllFiles"  title="Play Track Immediately">$HTH::playbutton</a>}]
         append result [HTH::tableRow [list $cb $infoBtn $playBtn $rating $artist $flink $track] $bgc]
      }  else {
         append result [HTH::tableRow [list $cb $infoBtn $rating $artist $flink $track] $bgc]
      }
   }
   append result "</table>\n"
   append result "<INPUT TYPE=submit VALUE='Queue Selected' NAME=jb_append_process>\n"
   append result "<INPUT TYPE=submit VALUE='Insert Selected' NAME=jb_insert_process>\n"
   append result "</FORM>\n"
   append result [subst {<a href="javascript:toggleAll_jb_add(true)">Select All</a> | <a href="javascript:toggleAll_jb_add(false)">Unselect All</a><br>}]
   return $result
}
package provide urlSearch 1.0
