#------------------------------------------------------------------------------
# snackAmp - urlStreaming package 1.0 for snackAmp
# provides dynamic http applications for SnackAmp
# web server.
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#                              s t r e a m i n g  f i l e  b r o w s e r
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#                              l i v e
#------------------------------------------------------------------------------
##
# the /streaming/live process method just retuns a file name from an existing file.
# used to initiate a stream to a client. This is run from shortcuts
# AFTER the playlist has been built and selected.
# Note: TFW: Add another option to sort, shuffle or sort/shuffle
#
set TA(useputs) 0
proc streaming/live::process {sock {params ""}} {
   Trace
   upvar #0 Httpd$sock data
   set Tracks [list]
   set sortMethod "shuffle"
   if {[string length $params]} {
      foreach param [split $params &]  {
         foreach {CmdType Cmd} [split $param =] {break}
         set Cmd [HttpD::Decode $Cmd]
         switch -- $CmdType {
            apl {
               # Get some apl file list of indicies
               if {[catch {streaming/live::apl $Cmd $sortMethod} result] } then {
                  return -code error $result
               } else {
                  if {[llength $result] > 0} {
                     streamBuffer::setList $sock $result
                     set data(streamingLength) 44000000
                     set data(Title) "AutoPlaylist From $data(mime,host)"
                     set data(Album) "$Cmd ([llength $result] trks)"
                     return [list file html [streamBuffer::popFile $sock]]
                  } else {
                     return -code error "No apl files for '$Cmd'"
                  }
               }
            }
            playlist {
               # Get some tracks from a playlist
               if {[catch {streaming/live::playlist $Cmd $sortMethod} result]} then {
                  return -code error $result
               } else {
                  ##
                  # Don't return track, return a m3u file
                  #
                  if {[llength $result] > 0} {
                     streamBuffer::setList $sock $result
                     set data(streamingLength) 44000000
                     set data(Title) "Playlist From $data(mime,host)"
                     set data(Album) "$Cmd ([llength $result] trks)"
                     return [list file html [streamBuffer::popFile $sock]]
                  } else {
                     return -code error "No playlist files for '$Cmd'"
                  }
               }
            }
            active {
               # Get some tracks from a playlist
               if {[catch {streaming/live::activeplaylist $sortMethod} result]} then {
                  return -code error $result
               } else {
                  ##
                  # Don't return track, return a m3u file
                  #
                  if {[llength $result] > 0} {
                     streamBuffer::setList $sock $result
                     set data(streamingLength) 44000000
                     set data(Title) "Active Playlist From $data(mime,host)"
                     set data(Album) "$Cmd ([llength $result] trks)"
                     return [list file html [streamBuffer::popFile $sock]]
                  } else {
                     return -code error "No active playlist files for '$Cmd'"
                  }
               }
            }
            folder {
               set result [list]
               set files [sortByMethod [db::filesInFolder $Cmd 0] $sortMethod]
               foreach file $files {
                  set loc [db::whereInDb $file]
                  if {$loc >= 0} {
                     lappend result $loc
                  }
               }

               if {[llength $result] > 0} {
                  streamBuffer::setList $sock $result
                  set data(streamingLength) 44000000
                  set data(Title) "Folder $Cmd From $data(mime,host)"
                  set data(Album) "$Cmd ([llength $result] trks)"
                  return [list file html [streamBuffer::popFile $sock]]
               } else {
                  return -code error "No folder files found for '$Cmd'"
               }
            }
            treefolder {
               set files [sortByMethod [db::filesMatchingPattern [file join $Cmd *]] $sortMethod]
               set result [list]
               foreach file $files {
                  set loc [db::whereInDb $file]
                  if {$loc >= 0} {
                     lappend result $loc
                  }
               }

               ##
               # Don't return track, return a m3u file
               #
               if {[llength $result] > 0} {
                  streamBuffer::setList $sock $result
                  set data(streamingLength) 44000000
                  set data(Title) "Treefolder $Cmd From $data(mime,host)"
                  set data(Album) "$Cmd ([llength $result] trks)"
                  return [list file html [streamBuffer::popFile $sock]]
               } else {
                  return -code error "streaming/live::process: No folder files found for '$Cmd'"
               }
            }
            track {
               # Process a single track in a playlist
               set loc [db::whereInDb $Cmd]
               # for DB case can improve by querying all at once
               if {$loc >= 0} {
                  set file [db::TrackFromDb $loc]
                  streamBuffer::setList $sock $loc
                  set data(streaming) $loc
                  if {[db::lookup $file Details 1]<=0} {
                     saLog $error
                  } else {
                     # Fill in data that will go into icy-meta tags
                     set data(streamingLength) $Details(size)
                     foreach {index} {Title Album Artist Genre Rate} {
                        if {[info exists Details($index)]} {
                           set data($index) $Details($index)
                        } else {
                           set data($index) ""
                        }
                     }
                  }
                  return [list file html [streamBuffer::popFile $sock]]
               } else {
                  return -code error "streaming/live::process: '$Cmd' not found in database"
               }
            }
            sort        {set sortMethod $Cmd}
            default {
               return -code error "streaming/live::process: Invalid subcommand - '$CmdType'"
            }
         }
      }
   } else {
      return -code error "streaming/live::process: No Parameter '$params'"
   }
}
#----------------------------------------------------------------------------
#  Read in an autoplaylist file and return the tracks,
#  properly sorted, within it to the stream
#----------------------------------------------------------------------------
proc streaming/live::apl {name sortMethod} {
   Trace
   set fName [file join $::HOME $name.tafs]
   if {![file exists $fName]} {
      return -code error "$fName does not exist"
   } else {
      if {[catch {open $fName r} fid]} then {
         return -code error $fid
      }
      fconfigure $fid -buffersize 100000
      set full [split [read $fid] \n]
      close $fid

      # Grab copy of teFolders and update it per the APL file
      array set myte [aplFolder::folderArray]
      foreach line $full {
         foreach {Index Value} $line {
            set myte($Index) $Value
         }
      }
      # Flag database with this APL list
      db::syncToAPLVar myte
      # Grab each flagged APL file
      set sortList [db::allAplFiles]
      # sort tracks then grab their location in database
      set tracks [sortByMethod $sortList $sortMethod]
      set biglist [list]
      foreach track $tracks {
         lappend biglist [db::whereInDb $track]
      }
      # restore APL
      db::syncToAPL

   }
   return $biglist
}
#----------------------------------------------------------------------------
# Return a list of indicies for a playlist
#----------------------------------------------------------------------------
proc streaming/live::playlist {name sortMethod} {
   Trace
   set name [string trim $name \"] ;#"
   foreach ext {m3u taf} {
      set fName [file join $::HOME $name.$ext]
      if {[file exists $fName]} {
         break
      }
   }
   if {![file exists $fName]} {
      saLog "streaming/live::playlist: $fName does not exist"
      return -code error "$fName does not exist"
   } else {

      # Here we have a file case issue
      if {[catch {open $fName r} fid]} then {
         return -code error $fid
      }
      fconfigure $fid -buffersize 100000
      set full [split [read $fid] \n]

      close $fid
      set tracks [list]
      foreach {tk} $full {
         if {![string match "#*" $tk] && ($tk ne "")} {
            lappend tracks [file join $tk]   ;# file join makes forward slashes
         }
      }
      set tracks [sortByMethod $tracks $sortMethod]
      # Get folder list
      set biglist [list]
      foreach line $tracks {
         set line [file join $line] ;# make tcl name
         set loc [db::whereInDb $line]
         if {$loc >= 0} {
            lappend biglist $loc
         } else {
 #           saLog "$line not in database"
         }
      }
   }
   return $biglist
}
#----------------------------------------------------------------------------
# Return a list of indicies to current playlist on GUI
#----------------------------------------------------------------------------
proc streaming/live::activeplaylist {sortMethod} {
   set biglist [list]
   ##
   # Note: May want to get from cursor foward.
   #
   set tracks [sortByMethod [LB::keyOrNamesAll 1] $sortMethod]
   foreach file $tracks {
      set loc [db::whereInDb $file]
      if {$loc >= 0} {
         lappend biglist $loc
      }
   }
   return $biglist
}
proc streaming/live::folder {name} {
}
proc streaming/live::recurfolder {name} {
}
#------------------------------------------------------------------------------
# Function   : sortByMethod
# Description: Sort a list by the appropriate method
# Author     : Tom Wilkason
# Date       : 3/24/2003
#------------------------------------------------------------------------------
proc streaming/live::sortByMethod {Items how} {
   set index [lsearch -exact $::filtering::sortModes $how]
   switch -- $index {
      0  {return [lsort -dictionary $Items]}
      1  {return [shuffle $Items]}
      2  {return [shuffleByFolder $Items ldsort]}
      3  {return [shuffleByFolder $Items shuffle]}
      default {
         package require subSorting
         set tosort [subSort::getModes]
         return [db::subSorting $tosort $Items]
      }
   }
}

#----------------------------------------------------------------------------
# url to show a directory listing to allow browsing
# of files by folder.
# todo:
# button to play folder
# button in insert file/folder into playlist
# todo: Add form controls to define how streaming/tracks/sorting is to be used.
#----------------------------------------------------------------------------
namespace eval streaming/live {
}
proc streaming/browser::process {sock {params ""}} {
   Trace
   variable brform
   upvar #0 Httpd$sock data
   set title "SnackAmp Jukebox"
   set result [HTH::leader $title $::snackAmpSettings(cssName)]
   append result <body>
   append result [js::scripts popupInfo foldersel]

   set sortMethod "shuffle"
   if {[string length $params]} {
      foreach param [split $params &]  {
         foreach {CmdType Cmd} [split $param =] {break}
         switch -- $CmdType {
            sort {
               set sortMethod $Cmd
            }
            folder {
               append result [H1 "SnackAmp media folders on [info hostname]"]
               set folders [list]
               set files [list]
               set folder [HttpD::Decode $Cmd]
               set validFolder 0
               foreach media $::TA(mediaDirs) {
                  if {[string match $media* $folder ]} {
                     set validFolder 1
                     break
                  }
               }
               if {$validFolder==0} {
                  set folder [lindex $::TA(mediaDirs) 0]
               }
               set files [db::filesInFolder $folder 0]
               set folders [lsort -dictionary [glob -nocomplain -types d -directory $folder *]]
               #----------------------------
               ##
               # Create a list of all folders, includling toplevel ones
               #
               set fldr $folder
               while {1} {
                  # Get list of all parent folders
                  set under [file dirname $fldr]
                  if {$under eq $fldr} {
                     break
                  } else {
                     lappend folders $fldr ;########
                     set fldr $under
                  }
               }
               # List of toplevel folder
               foreach item $::TA(mediaDirs) {
                  lappend folders $item
                  if {[string match "$item*" $folder]} {
                     set icon $HTH::folderopen
                  } else {
                     set icon $HTH::folderclosed
                  }
                  append result [subst {$icon <a href="?folder=[ENCODE $item]">$item</a><br>\n}]
               }
               #----------------------------
               set folders [lsort -unique $folders]
               # Show the folders
               if {[llength $folders]} {
                  append result "<hr>\n"
                  append result [subst $brform]
                  append result [H2 "Current folder $folder"]
                  append result "<table class='songPlaylist'>"
                  # Find minimum folder length for indent offsets
                  foreach item $folders {
                     lappend reps [llength [file split $item]]
                  }
                  set minrep [lindex [lsort -integer $reps] 0]
                  foreach item $folders {
                     set icon $HTH::folderclosed
                     if {$item eq $folder} {
                        set icon $HTH::folderopen
                     }
                     set rep [expr {[llength [file split $item]]-$minrep}]
                     if {[string length [file tail $item]]} {
                        set titum [file tail $item]
                     } else {
                        set titum $item
                     }
                     set link "[string repeat [SP] $rep]$icon <a href=\"?sort=$sortMethod&folder=[ENCODE $item]\">$titum</a><br>\n"
                     # S = Stream folder contents
                     # T = Tracks for folder contents
                     set base "/streaming/getplaylist.pls"
                     set F [subst {<a href="javascript:foldersel('$base','[ESCAPE $item]')" title="Play Folder Immediately">$HTH::playbutton</A>}]
                     append result [HTH::tableRow [list $F $link] darkRows]
                  }
                  append result "</table>\n"
               }
               # Show the files
               if {[llength $files]} {
                  append result "<hr>\n"
                  append result {<table><tr><td valign="top">}
                  append result "<table class='songPlaylist'>"
                  foreach file $files {
                     set I [HTH::popupInfo $file]
                     set link [AREF "/streaming/getplaylist.pls?sort=$sortMethod&track=[ENCODE $file]" $HTH::playbutton "Play Track Immediately"]<br>
                     set size "[commify [expr {[db::getVal $file Size]/1024}]] KB"
                     append result [HTH::tableRow [list $I $link [DBINFO $file] $size] [HTH::rowStyle i]]
                  }
                  append result "</table>\n"
                  append result {</td><td valign="top">}
                  append result [HTH::coverArtShadow [HTH::coverArt $folder $::snackAmpSettings(coverbrowsersize)]]
                  append result "</td></tr></table>"
               } else {
                  append result "<hr>\n"
                  append result [H1 "No tracks within $folder"]
               }
            }
         }
      }
   } else {
      append result [H1 "SnackAmp folder browser"]
      append result [H2 "Please select a folder below to browse"]
      foreach folder $::TA(mediaDirs) {
         append result $HTH::folderclosed[AREF "?folder=[ENCODE $folder]" $folder]<br>
      }
   }
   append result "</body>\n</html>\n"
   return [list dhtml html [string trim $result]]
}
#------------------------------------------------------------------------------
#                 s t r e a m i n g   p l a y l i s t  e x a m p l e
#------------------------------------------------------------------------------
##
# Example code to display a selection of sources on this server
#
proc streaming/playlists::process {sock {params ""}} {
   Trace
   upvar #0 Httpd$sock data
   ##
   # This can be prettied up to select the various
   # playlists that are created.
   # Can add options to use sorting options
   #
   set title "SnackAmp Jukebox"
   set result [HTH::leader $title $::snackAmpSettings(cssName)]
   append result <body>
   ##
   # Need to add code to show folder trees
   #
   append result {[H1 <a href='/streaming/getplaylist.pls?active=active'>Current playlist</a>]}
   set Files [catalog::getAutoPlaylists]
   # Create option list of all AutoPlaylist files
   if {[llength $Files] > 0} {
      append result [H1 "AutoPlaylist files"]
      foreach {file} $Files {
         set choice [ROOTTAIL $file]
         append result [AREF "/streaming/getplaylist.pls?apl=[ENCODE $choice]" $choice]<br>
      }
   }
   # Create option list of all playlist files
   set Files [catalog::getPlaylists]
   if {[llength $Files] > 0} {
      append result [H1 "Playlist files"]
      foreach {file} $Files {
         set choice [ROOTTAIL $file]
         append result [AREF "/streaming/getplaylist.pls?playlist=[ENCODE $choice]" $choice]<br>
      }
   }
   append result "</body>\n</html>\n"
   return [list dhtml html [string trim $result]]
}
#------------------------------------------------------------------------------
#                             g e t p l a y l i s t
#------------------------------------------------------------------------------
##
# Playlist code, from a web browser
# Processes a request to play a track, playlist, active playlist or AutoPlaylist
# Creates a winamp type pls dynamic file that when processed by the /streaming/live command
# will create a queue of songs then play them.
# Note: Add ability to send the entire stream length down as icy data. Good for tracks
# The PlayFile command will be called to process the files (via the registry setting)
##
# Note: TFW: Add another option to sort, shuffle or shufflesort
#            pass it on to the /streaming/live parameter, the sort param
#            should be specified first, otherwise it will be
#            ignored.
proc streaming/getplaylist.pls::process {sock {params ""}} {
   Trace
   upvar #0 Httpd$sock data
   #sort=no%20sort&tree=no&how=tracks&folder=//Bertha/Media/Blues/Albert%20King/Blues%20Power
   set result ""
   set sortMethod "shuffle"
   set how "stream"
   set param ""
   set method "unknown"
   set fmethod folder
   set max 100 ;# TODO: User setting, maximum number of tracks to build up
   set tracks [list]
   if {$::snackAmpSettings(useIP)} {
      set refer "http://$::snackAmpSettings(ipAddress):$::snackAmpSettings(usedPort)"
   } elseif {[string length $::snackAmpSettings(altStSrc)]>0} {
      set refer "http://$::snackAmpSettings(altStSrc):$::snackAmpSettings(usedPort)"
   } else {
      set refer $::snackAmpSettings(internalWebBase)
   }
   if {[string length $params]} {
      foreach pr [split $params &]  {
         foreach {CmdType Cmd} [split $pr =] {break}
         # Create a dynamic playlist
         switch -- $CmdType {
            active -
            playlist -
            track -
            folder -
            apl {
               set method $CmdType
               set param $Cmd
            }
            tree {
               switch -- $Cmd {
                  no {set fmethod folder}
                  yes {set fmethod treefolder}
               }
            }
            sort {set sortMethod [HttpD::Decode $Cmd]}
            how  {set how [HttpD::Decode $Cmd]}
            submit  {}
            default {saLog "streaming/getplaylist.pls::process: Invalid subcommand - '$CmdType'"}
         }
      }
      ##
      # We have all the params, now process the options
      #
#puts "how=$how;sortMethod=$sortMethod;method=$method;fmethod=$fmethod"
      switch -- [string tolower $how] {
         stream {
            # Create a stream of multiple tracks (except track option)
            switch -- $method {
               active {
                  set    result [HTH::plsTracks 1]
                  append result "File1=$refer/streaming/live?sort=[ENCODE $sortMethod]&active=$param\n"
                  append result "Title1=Active Playlist"
               }
               apl {
                  set    result [HTH::plsTracks 1]
                  append result "File1=$refer/streaming/live?sort=[ENCODE $sortMethod]&apl=$param\n"
                  append result "Title1=Auto Playlist [HttpD::Decode $param]"
               }
               playlist {
                  set    result [HTH::plsTracks 1]
                  append result "File1=$refer/streaming/live?sort=[ENCODE $sortMethod]&playlist=$param\n"
                  append result "Title1=Playlist [HttpD::Decode $param]"
               }
               treefolder -
               folder {
                  set    result [HTH::plsTracks 1]
                  append result "File1=$refer/streaming/live?sort=[ENCODE $sortMethod]&$fmethod=$param\n"
                  append result "Title1=Folder [nameToShort [HttpD::Decode $param]]"
               }
               track {
                  set    result [HTH::plsTracks 1]
                  append result "File1=$refer/streaming/live?track=$param\n"
                  append result "Title1=Track [nameToShort [HttpD::Decode $param]]"
               }
               default {saLog "streaming/getplaylist.pls::process: Invalid subcommand method=$method"}
            }
         }
         tracks {
            # Create individual tracks
            switch -- $method {
               active {
                  if {[catch {streaming/live::activeplaylist $sortMethod} tracks]} then {
                     saLog $tracks
                  } else {
                     set i 1
                     set tracks [lrange $tracks 0 $max]
                     set result [HTH::plsTracks [llength $tracks]]
                     foreach index $tracks {
                        set track [db::TrackFromDb $index]
                        append result "File$i=$refer/streaming/live?track=[ENCODE $track]\n"
                        append result "Title$i=[nameToShort $track]\n"
                        incr i
                     }
                  }
               }
               apl {
                  if {[catch {streaming/live::apl [HttpD::Decode $param] $sortMethod} tracks] } then {
                     saLog $tracks
                  } else {
                     set i 1
                     set tracks [lrange $tracks 0 $max]
                     set result [HTH::plsTracks [llength $tracks]]
                     foreach index $tracks {
                        set track [db::TrackFromDb $index]
                        append result "File$i=$refer/streaming/live?track=[ENCODE $track]\n"
                        append result "Title$i=[nameToShort $track]\n"
                        incr i
                     }
                  }
               }
               playlist {
                  if {[catch {streaming/live::playlist [HttpD::Decode $param] $sortMethod} tracks]} then {
                     saLog "[me] $tracks\n$::errorInfo"
                  } else {
                     set i 1
                     set tracks [lrange $tracks 0 $max]
                     set result [HTH::plsTracks [llength $tracks]]
                     # shuffle
                     foreach index $tracks {
                        set track [db::TrackFromDb $index]
                        append result "File$i=$refer/streaming/live?track=[ENCODE $track]\n"
                        append result "Title$i=[nameToShort $track]\n"
                        incr i
                     }
                  }
               }
               folder {
                  set i 1
                  set path [HttpD::Decode $param]
                  set tracks [lrange [db::filesInFolder $path 0] 0 $max]
                  set result [HTH::plsTracks [llength $tracks]]
                  foreach track $tracks {
                     append result "File$i=$refer/streaming/live?track=[ENCODE $track]\n"
                     append result "Title$i=[nameToShort $track]\n"
                     incr i
                  }
               }
               treefolder {
                  set i 1
                  set path [HttpD::Decode $param]
                  set tracks [lrange [db::filesMatchingPattern [file join $path *]] 0 $max]
                  set result [HTH::plsTracks [llength $tracks]]
                  foreach track $tracks {
                     append result "File$i=$refer/streaming/live?track=[ENCODE $track]\n"
                     append result "Title$i=[nameToShort $track]\n"
                     incr i
                     if {$i > $max} {break}
                  }
               }
               track {
                  set result [HTH::plsTracks 1]
                  append result "File1=$refer/streaming/live?track=$param\n"
                  append result "Title1=Track [nameToShort [HttpD::Decode $param]]"
               }
               default {saLog "streaming/getplaylist.pls::process: Invalid subcommand method=$method"}
            }
         }
         default {saLog "streaming/getplaylist.pls::process: Invalid subcommand how=$how"}
      }
      set result [string trim $result]
      return [list dhtml pls $result]
#       if {$::TA(useputs)} {
#          set mypath [HTH::WriteTempFile pls $result]
#          return [list file pls $mypath]
#       } else {
#          return [list dhtml pls $result]
#       }
   } else {
      set title "SnackAmp Jukebox"
      set result [HTH::leader $title $::snackAmpSettings(cssName)]
      append result [H1 "Please specify a parameter"]
      append result "</body>\n</html>"
      if {$::TA(useputs)} {
         set mypath [HTH::WriteTempFile pls $result]
         return [list file pls $mypath]
      } else {
         return [list dhtml html $mypath]
      }
   }
}

package provide urlStreaming 1.0
